/*
 * Copyright 2008, 2009, 2010 Boris Popov <popov.b@gmail.com>
 */

#ifndef __xmldocument_hpp__
#define __xmldocument_hpp__

#include <XMLHeaders.hpp>
#include <XMLBuilder.hpp>
#include <XMLParser.hpp>
#include <XMLSearchResultCommon.hpp>
#include <XMLSteps.hpp>

namespace xmlxpath {

   class ChildLexem;
   class ChildLexemAll;
   class ChildAny;
   class ChildAnyAll;
   class Step;
}

namespace xmlxpathcnt {

   class Container;
}

struct XMLParserResult;

/**
 * \brief Уровень вложенности.
 *
 *	Расстояние любого элемента
 *	до корневого элемента в данных
 *	величинах.
 */
typedef size_t level;

/**
 * \brief Строка XPath.
 *
 * \anchor xpath
 *
 * Необходима для поиска в документе.
 */
typedef std::string xpath;

typedef std::string String;

/**
 * \brief Структура атрибута.
 *
 * \anchor atr
 *
 * Атрибут представляет собой пару имя-значения:
 * a = "6". Может быть у декларации и контейнера.
 */
struct XMLAttr
{
	std::string att;///<Имя атрибута.
	std::string val;///<Значение атрибута.
};

class XMLBodyIter;
class XMLNode;
class XMLAttrIter;
class XMLI;
class XMLO;
class XMLContain;
class XMLComment;
class XMLText;
class XMLProlog;
class XMLBody;
class XMLXpath;
class XMLDeclar;
class XMLSearchResult;
class XMLChildIter;

/**
 * \brief Класс документа.
 *
 * Создание, удаление документа.
 * Загрузка, выгрузка xml-файла.
 * Предоставление доступа к другим классам
 * для манипулирования \ref XMLProlog "заголовком" и
 * \ref XMLBody "телом" документа.
 */
class XMLDocument
{
	public:
		/**
		 * \brief Создание документа.
		 *
		 * Создание документа из контейнера
		 * \b c. Контейнер \b c становится корневым
		 * элементом документа.
		 */
		XMLDocument (XMLContain& c);			

		/**
		 * \brief Создание документа.
		 *
		 * В данном случае создается документ
		 * с корневым элементом по имени \b root.
		 * Данный конструктор применяется в том
		 * случае если следующей операцией будет
		 * вызов метода XMLDocument::Load (XMLI& i).
		 */
		XMLDocument ();

		/**
		 * \brief Создание документа.
		 *
		 * Создание документа из
		 * строки символов.
		 */
		XMLDocument (const String&);

		/**
		 * \brief Создание документа.
		 *
		 * Создание документа из
		 * строки символов.
		 */
		XMLDocument (const char*);

		~XMLDocument ();
		//@{
		///Получить ссылку на пролог.
		XMLProlog& GetProlog () const;
		///Получить ссылку на тело.
		XMLBody& GetBody () const;
		//@}
		//@{
		/**
		 * \brief Загрузить документ.
		 *
		 * Вызов виртуального метода
		 * абстрактного класса XMLI.
		 */
		bool Load (XMLI& i);
		/**
		 * \brief Загрузить документ из файла.
		 *
		 */
		bool Load (std::string FileName);
		/**
		 * \brief Сохранить документ.
		 *
		 * Вызов виртуального метода
		 * абстрактного класса XMLO.
		 */
		bool Save (XMLO& o);		
		///Информация об ошибке парсера.
		std::string GetMessage ();							
		//@}
		//@{
		/**
		 * \brief Удалить созданную структуру.
		 *
		 * Удаляется вся структура документа. 
		 * Корневой контейнер не удаляется.
		 */
		void Del ();										
		//@}
		
		/**
		 * \brief Тип элемента
		 *
		 * \anchor typecont
		 *
		 * - Комментарий представляет собой: <b>\<!--бла-бла-бла--\></b>.
		 * - Контейнер представляет собой <b>\<A\>бла-бла\</A\></b>. Внутри
		 *   него могут быть вложены другие элементы. Могут присутствовать
		 *   \ref XMLAttr "атрибуты".
		 * - Текст - просто текстовый фрагмент.
		 */
		enum TypeCont
		{
			Comment,	///<Комментарий
			Container,	///<Контейнер
			Text		///<Текст
		};

	private:
		std::string message;
		void correct (XMLParserResult pr);
		XMLProlog& Prolog;
		XMLBody& Body;
		XMLDocument& operator = (XMLDocument& d);

		const char* m_InputString;
		void m_ParseString ();
};


/**
 * \brief Класс декларации.
 *
 * Управление декларацией.
 */
class XMLDeclar
{

	friend class XMLDocument;
	friend class XMLBuilderOfProlog;
	friend class XMLBuildOfTree;

	public:
		/**
		 * \brief Конструктор
		 *
		 * Создать декларацию
		 * с содержимым <b>\<?xml version="1.0"?\></b>.
		 */
		XMLDeclar ();
		~XMLDeclar ();
		//@{
		///Очистить значение декларации. 
		void UnSet();		
		/**
		 * \brief Установить значение декларации по умолчанию.
		 *
		 * Установить значение декларации:
		 * <b>\<?xml version="1.0"?\></b>.
		 */
		void Set ();
		///Получить значение декларации.
		std::string Get ();
		//@}
		//@{
		///Получить значение итератора на первый атрибут.
		XMLAttrIter Begin ();							
		///Получить значение итератора на последний атрибут.
		XMLAttrIter End ();								
		//@}
		//@{
		///Удалить атрибут с именем \b a.
		void DelAtt (std::string a);					
		///Добавить атрибут \b a со значением \b v.
		void AddAtt (std::string a, std::string v);
		/**
		 * \brief Получить значение атрибута.
		 *
		 * Получить значение атрибута \b a.
		 *
		 */
		std::string GetAtt (std::string a) const;
		//@}

	private:
		std::list <XMLAttr> ListOfAtt;
		static std::string part_one;
		static std::string part_three;
		enum SWITCH {ON, OFF};
		SWITCH SW;
};

/**
 * \brief Класс пролога.
 *
 * Предоставляет доступ к:
 * -# декларации; 
 * -# инструкции по обработке;
 * -# определению типа документа.
 */
class XMLProlog
{
	public:
		///Конструктор.
		XMLProlog ();
		///Деструктор.
		~XMLProlog ();
		//@{
		/**
		 * \brief Получить ссылку на \ref XMLDeclar "декларацию".
		 */
		XMLDeclar& GetDeclar () const;
		//@}

	private:
		XMLDeclar& Declar;
};


/**
 * \brief Класс тела.
 *
 * Все что находится ниже корневого элемента,
 * включая этот корневой элемент.
 */
class XMLBody
{
	friend class XMLDocument;
	friend class XMLBuilder;
	
      public:
		/**
		 * \brief Конструктор.
		 *
		 * Создаётся тело документа, при этом в качестве
		 * корневого элемента 
		 * создаётся контейнер со значением \b root.
		 */
		XMLBody ();
		/**
		 * \brief Конструктор с параметром.
		 *
		 * Создаётся тело документа, при этом в качестве
		 * корневого элемента используется
		 * контейнер \b c.
		 */
		XMLBody (XMLContain& c);
		///Деструктор.
		~XMLBody ();
		//@{
		/**
		 * \brief Удалить узел на который указывает итератор.
		 *
		 * Данный метод удаляет контейнер, комментарий либо
		 * текстовый фрагмент, в зависимости от того,
		 * на что смотрит итератор \b i. В случае
		 * если это контейнер, то удаляются также
		 * все вложенные в него элементы. Корневой
		 * элемент удалить нельзя. В случае удачного
		 * удаления итератор \b i перемещается на родительский
		 * узел.
		 */
		void DelNode (XMLBodyIter& i);				

		/**
		 * \brief Удалить контейнер.
		 *
		 * Удаляется контейнер \b cn и все
		 * вложенные элементы. Корневой
		 * элемент удалить нельзя.
		 */
		void DelNode (XMLContain& cn);
		///Удалить комментарий.
		void DelNode (XMLComment& cm);
		///Удалить текстовый элемент.
		void DelNode (XMLText& tx);
		//@}
		//@{
		///Получить значение итератора на начало.
		XMLBodyIter BeginBodyIter ();
		///Получить значение итератора на конец.
		XMLBodyIter EndBodyIter ();
		//@}

		//@{
		/**
		 * \brief Поиск.
		 *
		 * Осуществить поиск от корневого контейнера.
		 * Результат поиска поместить
		 * в result.
		 */
		void Search (const xpath path, XMLSearchResult& result);
		//@}

		///Получить корневой контейнер
		XMLContain* GetRoot () const;
		
	private:
		XMLContain* Root;
		void delnode (XMLNode* pN);
};


/**
 * \brief Класс итератора по элементам тела.
 *
 * Обеспечивает последовательное перемещение
 * по элементам тела.
 */
class XMLBodyIter
{
	friend class XMLBody;
	friend class XMLContain;

	public:
		///Разыменовывать итератор.
		XMLNode* operator * ();		

		///Разыменовывать итератор.
		XMLNode* operator -> ();		

		//@{ 
		///Передвинуть итератор вперед.
		XMLBodyIter& operator ++ ();						
		///Передвинуть итератор вперед.
		XMLBodyIter& operator ++ (int t);					
		//@}
		//@{ 
		///Сравнить значения итераторов.
		friend bool operator != (XMLBodyIter a, XMLBodyIter b);	
		///Сравнить значения итераторов.
		friend bool operator == (XMLBodyIter a, XMLBodyIter b);	
		//@}

		XMLBodyIter (const XMLBodyIter&);
		XMLBodyIter& operator= (const XMLBodyIter&);

	private:
		XMLNode* pNode;
		
		std::queue <XMLNode*> Queue;
		XMLNode* plus_normal_iter ();
		XMLBodyIter ();
		XMLBodyIter (XMLNode* pN);
};

/**
 * \brief Класс итератора по атрибутам.
 *
 * Обеспечивает перемещение по атрибутам
 * контейнера либо декларации и обеспечивает
 * доступ к ним.
 */
class XMLAttrIter
{
	friend class XMLContain;
	friend class XMLDeclar;

	public:
		///Получить атрибут.
		XMLAttr& Get ();											
		//@{ 
		///Передвинуть итератор назад.
		XMLAttrIter& operator -- ();							
		///Передвинуть итератор назад.
		XMLAttrIter& operator -- (int t);						
		///Передвинуть итератор вперед.
		XMLAttrIter& operator ++ ();							
		///Передвинуть итератор вперед.
		XMLAttrIter& operator ++ (int t);						
		//@}
		//@{ 
		///Сравнить значения итераторов.
		friend bool operator != (XMLAttrIter a, XMLAttrIter b);	
		///Сравнить значения итераторов.
		friend bool operator == (XMLAttrIter a, XMLAttrIter b);	
		//@}
		
		XMLAttrIter ();											
		XMLAttrIter (const XMLAttrIter&);
		XMLAttrIter& operator= (const XMLAttrIter&);

	private:
		std::list <XMLAttr>::iterator IPA;
};

///Абстрактный класс ввода документа.
class XMLI
{
	public:
		virtual ~XMLI ();
		/**
		 * \brief Получить текст документа.
		 *
		 * Этот метод должен уметь забирать
		 * откуда-то (файл...) текст.
		 */
		virtual const char* Get () = 0;
};

///Абстрактный класс вывода документа.
class XMLO
{
	public:
		virtual ~XMLO ();
		/**
		 * \brief Отправить текст документа.
		 *
		 * Этот метод должен уметь сохранять
		 * документ, находящийся в
		 * const char* ch, в файле,
		 * потоке, базе данных...
		 */
		virtual void Put (const char* ch) = 0;
};

/**
 * \brief Класс узла.
 *
 * По объектам этого класса бегают итераторы
 * XMLBodyIter, XMLIter.
 */
class XMLNode
{

	friend class XMLBody;
	friend class XMLContain;
	

	public:
		/**
		 * \brief Деструктор.
		 *
		 * Если данный элемент
		 * вложен во внешний  контейнер, то
		 * из контейнера удаляется
		 * вся информация о данном
		 * элементе.
		 */
		virtual ~XMLNode ();
		/**
		 * \brief Конструктор.
		 *
		 * Создание нужного элемента в
		 * зависимости от параметра
		 * конструктора.
		 */
		XMLNode (XMLDocument::TypeCont type);
		//@{
		/**
		 * \brief Установить значение.
		 *
		 * В зависимости от типа узла, этим
		 * методом устанавливается декларация,
		 * комментарий либо имя контейнера.
		 */
		void Set (std::string str);		
		//@}
		//@{
		/**
		 * \brief Взять значение.
		 *
		 * В зависимости от типа узла, этим
		 * методом берётся декларация,
		 * комментарий либо имя контейнера.
		 */
		std::string Get () const ;
		///Получить тип узла.
		XMLDocument::TypeCont GetType () const ;
		///Получить удаленность узла.
		level GetLevel () const ;
		//@}

	protected:
		XMLContain* pParent;
		std::string Content;
		XMLDocument::TypeCont TypeNode;
		enum Place
		{
			in,
			out,
			undef
		};
		Place place;
		virtual std::string build ();
};


///Класс контейнера.
class XMLContain : public XMLNode
{
		friend class XMLBody;
		friend class XMLBodyIter;
		friend class XMLIter;
		friend class XMLAttrIter;
		friend class XMLBuildOfTree;
		friend class XMLXpath;
		friend class XMLBuilder;
		friend class xmlxpath::ChildLexem;
		friend class xmlxpath::ChildLexemAll;
		friend class xmlxpath::ChildAny;
		friend class xmlxpath::ChildAnyAll;

	public:
		/**
		 * \brief Конструктор.
		 *
		 * Создать контейнерный элемент с именем
		 * тега str.
		 */
		XMLContain (std::string str);
		//@{
		/**
		 * \brief Удалить атрибут.
		 *
		 * Удаление атрибута с именем \b a.
		 */
		void DelAtt (std::string a);
		/**
		 * \brief Добавить атрибут.
		 *
		 * Добавить атрибут \b a со значением
		 * \b v.
		 */
		void AddAtt (std::string a, std::string v);
		///Есть ли атрибуты?
		bool HasItAttr () const ;
		/**
		 * \brief Получить значение атрибута.
		 *
		 * Получить значение атрибута "a".
		 *
		 */
		std::string GetAtt (std::string a) const;
		//@}

		//@{
		/**
		 * \brief Значение итератора на первый элемент.
		 *
		 * Этот метод нужен для прохождения по "детям".
		 */
		XMLChildIter BeginChildIter ();						
		/**
		 * \brief Значение итератора на элемент после последнего.
		 *
		 * Этот метод нужен для прохождения по "детям".
		 */
		XMLChildIter EndChildIter ();	
		//@}

		//@{
		/**
		 * \brief Значение итератора на первый элемент.
		 *
		 * Этот метод нужен для прохождения по элементам,
		 * находящимся внутри элемента.
		 */
		XMLBodyIter BeginBodyIter ();						
		/**
		 * \brief Значение итератора на последний элемент.
		 *
		 * Этот метод нужен для прохождения по элементам,
		 * находящимся внутри элемента.
		 */
		XMLBodyIter EndBodyIter ();	
		//@}

		//@{
		///Значение итератора на первый атрибут.
		XMLAttrIter BeginAttr ();					
		///Значение итератора на последний атрибут.
		XMLAttrIter EndAttr ();						
		//@}
		
		
		//@{
		/**
		 * \brief Сцепить элементы.
		 *
		 * \anchor zepka
		 *
		 * Вложить контейнер e в контейнер,
		 * стоящий слева от <<.
		 */
		XMLContain& operator << (XMLContain& e);
		/**
		 * \brief Сцепить элементы.
		 *
		 * Вложить контейнер e в контейнер,
		 * стоящий слева от <<.
		 */
		XMLContain& operator << (XMLContain const &e);
		/**
		 * \brief Сцепить элементы.
		 *
		 * Вложить контейнер e в контейнер,
		 * стоящий слева от +=.
		 */
		XMLContain& operator += (XMLContain& e);
		/**
		 * \brief Сцепить элементы.
		 *
		 * Вложить контейнер e в контейнер,
		 * стоящий слева от +=.
		 */
		XMLContain& operator += (XMLContain const &e);
		//@}


		//@{
		/**
		 * \brief Сцепить элементы.
		 *
		 * Вложить текст t в контейнер,
		 * стоящий слева от <<.
		 */
		XMLContain& operator << (XMLText& t);
		/**
		 * \brief Сцепить элементы.
		 *
		 * Вложить текст t в контейнер,
		 * стоящий слева от <<.
		 */
		XMLContain& operator << (XMLText const &t);
		/**
		 * \brief Сцепить элементы.
		 *
		 * Вложить текст t в контейнер,
		 * стоящий слева от +=.
		 */
		XMLContain& operator += (XMLText& t);
		/**
		 * \brief Сцепить элементы.
		 *
		 * Вложить текст t в контейнер,
		 * стоящий слева от +=.
		 */
		XMLContain& operator += (XMLText const &t);
		//@}


		/**
		 * \brief Деструктор.
		 *
		 * Удалить контейнер и все
		 * вложенные в него
		 * элементы.
		 */
		~XMLContain ();

		//@{
		/**
		 * \brief Поиск.
		 *
		 * Осуществить поиск от контейнера.
		 * Результат поиска поместить
		 * в result.
		 */
		void Search (const xpath path, XMLSearchResult& result);
		//@}

	private:
		std::list <XMLNode*> ListOfChildren;
		std::list <XMLAttr> ListOfAttr;
		XMLContain ();
		XMLContain& operator = (XMLContain& c);
		std::string build ();
		std::string get_string_attr ();
		void m_SearchFromRoot (const xpath path, XMLSearchResult& result);
		void m_SearchFromContain (const xpath path, XMLSearchResult& result);
};

///Класс текста.
class XMLText : public XMLNode
{
		friend class XMLBuildOfTree;
		friend class XMLContain;

	public:
		/**
		 * \brief Конструктор.
		 *
		 * Создать элемент текста с содержимым
		 * str.
		 */
		XMLText (std::string str);
		~XMLText ();
		
	private:
		XMLText ();
		XMLText& operator = (XMLText& t);
		std::string build ();
};


class XMLSearchResultIter;

/**
 * \brief Класс результата поиска.
 *
 * Фактически это контейнер с
 * результатми поиска.
 */
class XMLSearchResult {

   friend class XMLContain;
   friend class xmlxpath::ChildLexem;
   friend class xmlxpath::ChildLexemAll;
   friend class xmlxpath::ChildAny;
   friend class xmlxpath::ChildAnyAll;
   friend class xmlxpath::Step;

   public:
      ///Создание объекта.
      XMLSearchResult ();
      ~XMLSearchResult ();

      ///Очистка контейнера
      void Clear ();

      ///Получить значение итератора на начало контейнера.
      XMLSearchResultIter Begin ();

      ///Получить значение итератора на конец контейнера.
      XMLSearchResultIter End ();

   private:
      SearchContainer m_container;
};

/**
 * \brief Класс итератора по результатам поиска.
 *
 * Итератор по классу XMLSearchResult.
 */
class XMLSearchResultIter {
   
   friend class XMLSearchResult;

   public:
       XMLSearchResultIter ();
       ~XMLSearchResultIter ();

      //@{ 
      ///Разыменовывать итератор.
      XMLContain* operator * ();		

      ///Разыменовывать итератор.
      XMLContain* operator -> ();		
      //@}

      //@{
      ///Передвинуть итератор вперед.
      XMLSearchResultIter& operator ++ ();						
      ///Передвинуть итератор вперед.
      XMLSearchResultIter& operator ++ (int t);					
      //@}

      //@{ 
      ///Сравнить значения итераторов.
      friend bool operator != (XMLSearchResultIter a, XMLSearchResultIter b);	
      ///Сравнить значения итераторов.
      friend bool operator == (XMLSearchResultIter a, XMLSearchResultIter b);	
      //@}

	  XMLSearchResultIter (const XMLSearchResultIter&);
	  XMLSearchResultIter& operator= (const XMLSearchResultIter&);

   private:
      SearchContainerIterator m_current;
};

typedef std::list<XMLNode*>::iterator IteratorOnChild;
bool operator != (XMLChildIter a, XMLChildIter b);	
bool operator == (XMLChildIter a, XMLChildIter b);	

/**
 * \brief Класс итератора для перемещения по "детям" контейнера.
 *
 */
class XMLChildIter
{
	friend class XMLContain;
	
	public:
		XMLChildIter ();
		~XMLChildIter ();
		///Разыменовывать итератор.
		XMLNode* operator * ();				
	
		///Разыменовывать итератор.
		XMLNode* operator -> ();				
		
		//@{ 
		///Передвинуть итератор вперед.
		XMLChildIter& operator ++ ();								
		///Передвинуть итератор вперед.
		XMLChildIter& operator ++ (int t);							
		//@}

		//@{ 
		///Передвинуть итератор назад.
		XMLChildIter& operator -- ();								
		///Передвинуть итератор назад.
		XMLChildIter& operator -- (int t);							
		//@}

		//@{ 
		///Сравнить значения итераторов.
		friend bool operator != (XMLChildIter a, XMLChildIter b);	
		///Сравнить значения итераторов.
		friend bool operator == (XMLChildIter a, XMLChildIter b);	
		//@}
		
		XMLChildIter (const XMLChildIter&);								
		XMLChildIter& operator= (const XMLChildIter&);

	private:
		IteratorOnChild m_IOC;
};
#endif

/**\file
 * \brief Заголовочный файл.
 *
 * Интерфейс работы
 * с xml-документом.
 *
 * \author Борис Попов.
 */

/**
  \mainpage Интерфейс библиотеки работы с XML-файлом

\section change7 Изменения в интерфейсе (9.12.2010):
- добавлен метод XMLBody::GetRoot () const.

\section change6 Изменения в интерфейсе (9.12.2009):
- метод XMLBody::Find (xpath path) удалён.

\section change5 Изменения в интерфейсе (3.11.2009):
- Методы XMLContain::GetAtt и XMLDeclar::GetAtt теперь константные.

\section change4 Изменения в интерфейсе (30.10.2009):
- Добавлены конструкторы XMLDocument::XMLDocument (const char*), XMLDocument::XMLDocument (const String&)

\section change3 Изменения в интерфейсе (13.06.2009):
- Исходный код приведён в соответствие с g++ версии 4.3.2.	

\section change2 Изменения в интерфейсе (17.04.2009):
- новый класс XMLChildIter (итератор по "детям"),
- новый метод XMLContain::BeginChildIter (),
- новый метод XMLContain::EndChildIter ().	

\section change1 Изменения в интерфейсе (20.01.2009):
- новый метод поиска от тела документа XMLBody::Search (const xpath path, XMLSearchResult& result),
- новый метод поиска от контейнерного элемента XMLContain::Search (const xpath path, XMLSearchResult& result),
- новый класс XMLSearchResult (результат поиска),
- новый класс XMLSearchResultIter (итератор по результату поиска),
- метод XMLBody::Find (xpath path) работает, но устаревший и применять не рекомендую.

\section content Содержание:
  -# \ref intro   "Общие сведения"
  -# \ref make    "Создание документа в памяти"
  -# \ref io      "Ввод - вывод"
  -# \ref decl    "Работа с декларацией"
  -# \ref pidtd   "Работа с инструкцией по обработке и определением типа документа"
  -# \ref move    "Перемещение по документу"
  -# \ref get     "Получение информации из элементов тела"
  -# \ref edit    "Редактирование"
  -# \ref find    "Поиск в документе"

  \author Борис Попов <popov.b@gmail.com> &copy;
 */

/**
 * \page intro Общие сведения.
 
\section about Сведения о документе xml.
XML документ состоит из двух частей:
- пролога;
- тела.

\subsection prolog Пролог.
Пролог располагается в начале документа
и состоит из:
- декларации;
- инструкции по обработке документа;
- определения типа документа.

Пролог может отсутствовать.

\subsubsection declar Декларация.
Может отсутствовать. Если есть, то располагается
в начале документа. Пример:
\code
<?xml version="1.0" ?>
\endcode

\subsubsection pi Инструкция по обработке документа.
Может отсутствовать. Если есть, то располагается
после декларации. Пример:
\code
<?xml-stylesheet type="text/css" href="greeting.css" ?>
\endcode

\subsubsection dtd Определение типа документа.
Может отсутствовать. Если есть, то располагается
после инструкции по обработке документа. Пример:
\code
 <!DOCTYPE test [
 <!ELEMENT test (#PCDATA) >
 <!ENTITY % xx '&#37;zz;'>
 <!ENTITY % zz '&#60;!ENTITY tricky "error-prone" >' > %xx; ]>
\endcode

\warning
Между декларацией и инструкцией по применению,
между инструкцией по применению и определением
типа документа, а также после типа документа
и до начала тела, могут присутствовать
комментарии.

\subsection body Тело.
Тело присутствует всегда. Если есть пролог, то
тело располагается за прологом.
Тело состоит из следующих элементов:
- контейнеров;
- текстовых фрагментов или проще говоря текста;
- комментариев.

\subsubsection contain Контейнеры.
Минимум из чего может состоять xml документ
это пустой контейнер. Примеры:
\code
<a>
</a>
\endcode
либо тоже, но проще:
\code
<a />
\endcode
Контейнер &mdash; иерархи-образующий элемент. Внутри него могут
располагаться подобные элементы, а также текст или комментарии:
\code
<a>
	<b>text</b>
</a>
\endcode
У контейнера могут быть атрибуты:
\code
<a x="1">
	<b x="2">text</b>
</a>
\endcode
\subsubsection text Текст.
Просто текстовый фрагмент, вложенный в контейнерный элемент.
Пример смотри выше.

\subsubsection comment Комментарии.
Комментарий представляет собой:
\code
<!-- comment -->
\endcode

\warning
За телом и до конца файла могут располагаться
комментарии.

\section classes Классы интерфейса.
Основные структурные единицы:
- класс документа;
- классы итераторов;
- классы элементов.

\subsection classdoc Класс документа и другие вспомогательные классы.
XMLDocument.

Данный класс осуществляет следующее:
- создание начальной структуры документа;
- удаление структуры документа;
- загрузка / выгрузка документа (через
  абстрактные классы ввода и вывода XMLI, XMLO);
- доступ к комментариям находящимся вне тела
  (через итератор по комментариям класс XMLCommentIter) и удаление их;
- доступ к прологу (через класс пролога XMLProlog);
- доступ к телу (через класс тела XMLBody).

XMLI, XMLO. Для создания в интерфейсе функций ввода-вывода,
необходимо написать собственную реализацию унаследованных
виртуальных методов данных классов.

XMLProlog. Предоставляет доступ к:
- декларации (через класс XMLDeclar);
- инструкции по обработке (через класс XMLPi);
- определению типа документа (через класс XMLDtd).

XMLBody. Позволяет:
- предоставлять доступ к элементам тела
  (через класс итератора XMLBodyIter);
- удалять отдельные элементы из тела;
- искать по критерию элементы в теле;
- вставлять комментарии после тела.

XMLDeclar. Позволяет:
- устанавливать / очищать / получать значение;
- получать доступ к отдельным атрибутам декларации
  через класс итератора XMLAttrIter;
- удалять / добавлять / получать атрибут декларации;
- вставлять комментарий после себя.

XMLPi и XMLDtd. Позволяют:
- устанавливать / очищать / получать значение;
- вставлять комментарий после себя.


\subsection classiter Классы итераторов.
В интерфейсе библиотеки существуют 4 класса итераторов:
- для последовательного перемещения по всем элементам
  тела (XMLBodyIter);
- для перемещения по контейнерам одного уровня,
  то есть равноудаленных от корневого элемента
  (XMLIter);
- для перемещения по атрибутам контейнера либо
  декларации (XMLAttrIter);
- для перемещения по комментариям, находящимся
  вне тела (XMLCommentIter).

\subsection classelmnt Классы элементов тела.
В интерфейсе доступны следующие классы
элементов тела:
- класс контейнера (XMLContain);
- класс текста (XMLText);
- класс комментария (XMLComment).

Базовым классом для указанных выше
классов является XMLNode. По объектам
этого класса перемещаются итераторы
XMLBodyIter и XMLIter, которые могут
разыменованы "звёздочкой".

Общими методами этих классов являются:
- получить / установить значение;
- получить тип элемента;
- получить удаленность от корневого элемента.

Кроме указанных методов, класс контейнера
позволяет следующее:
- добавлять / удалять / брать атрибут;
- получать значения итератора на атрибуты;
- получать значения итератора для прохождения по
  вложенным элементам;
- получать значения итератора для прохождения по
  элементам тела того же уровня;
- вкладывать в себя другие контейнеры, текстовые
  фрагменты или комментарии.
*/

/**
 * \page make Создание документа в памяти.

\section make1 Вариант 1

В одну строку.

Пример:
\code
	XMLDocument DOC (XMLContain ("newROOT") << (XMLContain ("R", "t", "1") << XMLText ("text") << XMLComment ("comment")));
\endcode

После вышеуказанного кода,
	документ принимает следующий вид:
\code
<?xml version="1.0" ?>
<newROOT>
  <R t="1">
    text
    <!--comment-->
  </R>
</newROOT>
\endcode

\section make2 Вариант 2

- создать необходимое число объектов
  XMLContain, XMLComment, XMLText;
- в конструкторе класса XMLDocument
  сцепить вышеперечисленные объекты
  методом XMLContain::operator +.

Пример:
\code
	XMLContain ROOT ("ROOT");
	XMLContain C1 ("C1");
	XMLContain C2 ("C2");
	C2.AddAtt ("arg1", "1");
	C2.AddAtt ("arg2", "2");
	XMLContain C555 ("555");
	XMLComment COM ("Comment");
	XMLText TXT ("Text");
	XMLDocument TEST ((ROOT += COM) += (C1 += (C2 += TXT)) += C555);
\endcode

После вышеуказанного кода,
	документ принимает следующий вид:
\code
<?xml version="1.0" ?>
<ROOT>
  <!--Comment-->
  <C1>
    <C2 arg1="1" arg2="2">
      Text
    </C2>
  </C1>
  <555 />
</ROOT>
\endcode

\section make3 Вариант 3

По мере поступления данных в программу,
можно создавать документ таким образом:

Пример:
\code
	// фрагмент кода...
	
	XMLContain ONE ("one");
	XMLContain UP ("up");

	// фрагмент кода...

	{

		// фрагмент кода...

		XMLContain TWO ("two");
		ONE << TWO;

		// фрагмент кода...

		{

			// фрагмент кода...

			XMLContain THREE ("three");
			XMLText TEXT ("This is text!");
			XMLComment COMM ("This is comment!");
			THREE << TEXT;
			THREE << COMM;
			ONE << THREE;

			// фрагмент кода...

		}

		// фрагмент кода...
	}

	// фрагмент кода...
	
	UP << ONE;

	// фрагмент кода...

	XMLDocument DOC (UP);
\endcode

После вышеуказанного кода,
	документ принимает следующий вид:
\code
<?xml version="1.0" ?>
<up>
  <one>
    <two />
    <three>
      This is text!
      <!--This is comment!-->
    </three>
  </one>
</up>
\endcode
 */

/**
 * \page io Ввод-вывод.
 
Операции ввода-вывода организованы с помощью
чисто виртуальных методов абстрактных классов
XMLI, XMLO.

Когда вызываются методы XMLDocument::Load (XMLI &i)
и XMLDocument::Save (XMLO &o), то они (методы) в свою очередь
вызывают виртуальные методы XMLI::Get () и
XMLO::Put (const char *ch) соответственно.

Таким образом, необходимо написать собственный
класс ввода или вывода, унаследованный соответственно
от XMLI или XMLO, и в нём переопределить
функции \b Get() или \b Put().
 
Пример:
\code
	// пропущено...

class XMLFileOut : public XMLO
{
	public:
		XMLFileOut (std::string file);
		void Put (const char* ch);
};

	// пропущено...

	XMLDocument DOC (ROOT);
	
	// пропущено...
	
	XMLFileOut FILES ("test1.xml");
	DOC.Save (FILES);
	
	// пропущено...
\endcode
 */

/**
 * \page decl Работа с декларацией.
 
За работу с декларацией отвечает класс XMLDeclar.
Получить доступ к объекту этого класса можно так:
\code
XMLDocument DOC;
XMLDeclar& DECLAR = DOC.GetProlog ().GetDeclar ();
\endcode

 Несколько особенностей при работе с декларацией:

 - декларация в документе одна;
 - при создании документа, создается
 декларация вида:
\code
<?xml version="1.0" ?>
\endcode
 - декларацию можно очистить (удалить) и создать вновь:
\code
XMLDocument DOC;
XMLDeclar& DECLAR = DOC.GetProlog ().GetDeclar ();

//распечатать декларацию:
std::cout << DECLAR.Get ();

//удалить декларацию:
DECLAR.UnSet ();

//установить декларацию по умолчанию
DECLAR.Set ();
\endcode
 - также можно взять / добавить / удалить атрибуты:
 \code
 //получить значение атрибута lang
 std::cout << DECLAR.GetAtt ("lang");

 //удалить атрибут
 DECLAR.DelAtt ("lang");

 //добавить атрибут
 DECLAR.AddAtt ("lang", "en");
 \endcode
 - доступ к отдельным атрибутам декларации
 осуществляется через итератор XMLAttrIter:
 \code
//пройти по всем атрибутам и распечатать их
XMLAttrIter AI;
AI = DECLAR.Begin ();
while (AI != DECLAR.End ())
{
	std::cout << (AI.Get ()).att << "  " <<(AI.Get ()).val;
	AI++;
}
 \endcode
 - вставить комментарий после декларации:
 \code
 DECLAR.InsertCommentAfter ("Это комментарий после декларации!");
 \endcode
 */

/**
 * \page pidtd Работа с инструкцией по обработке и определением типа документа.

На сегодняшний день функционал методов классов XMLPi и XMLDtd практически идентичен:
- получить доступ к объектам этого класса можно так:
\code
XMLDocument DOC;
XMLPi& PI = DOC.GetProlog ().GetPi ();
XMLDtd& DTD = DOC.GetProlog ().GetDtd ();
\endcode

- инструкцию по обработке и определение типа документа можно очистить / установить / взять:
\code
//установить значения
PI.Set ("xml-stylesheet type=\"text/css\" href=\"greeting.css\"");
DTD.Set ("test [
 <!ELEMENT test (#PCDATA) >
 <!ENTITY % xx '&#37;zz;'>
 <!ENTITY % zz '&#60;!ENTITY tricky \"error-prone\" >' > %xx; ]");

//распечатать
std::cout << PI.Get ();
std::cout << DTD.Get ();

//удалить декларацию:
PI.UnSet ();
DTD.UnSet ();
\endcode

- вставить комментарий после себя:
\code
PI.InsertCommentAfter ("Это комментарий после инструкции по обработке!");
DTD.InsertCommentAfter ("Это комментарий после определения типа документа!");
\endcode
 */

/**
 * \page move Перемещение по документу.

Для перемещения по документу используются
соответствующие итераторы.

\section allmove Последовательное перемещение по всем элементам внутри корневого контейнера.
Для данного перемещения используется объект класса XMLBodyIter.
Движение только вперед. Можно разыменовывать "звёздочкой".
Значение итератора на начало и конец надо получить у
объекта класса XMLBody.
\code
XMLDocument DOC;
XMLBodyIter BI = DOC.GetBody ().BeginBodyIter ();
while (BI != DOC.GetBody ().EndBodyIter ())
{
	std::cout << "Значение элемента: " << (*BI)->Get () << std::endl;	
	BI++;
}
\endcode
 
\section inmove Перемещение по элементам внутри некорневого контейнера.
Для данного перемещения используется объект класса XMLBodyIter.
Движение только вперед. Можно разыменовывать "звёздочкой".
Значение итератора на начало и конец надо получить у
объекта класса XMLContain.
\code
XMLDocument DOC;
XMLContain Contain;
XMLBodyIter BI = Contain.BeginBodyIter ();
while (BI != Contain.EndBodyIter ())
{
	std::cout << "Значение элемента: " << (*BI)->Get () << std::endl;	
	BI++;
}
\endcode

\section levelmove Перемещение по контейнерам одного уровня.
Используется итератор, объект класса XMLIter.
Движение вперед и назад. Можно разыменовывать "звёздочкой".
Значение итератора на начало и конец надо получить у
объекта класса XMLContain.
\code
XMLContain Contain;
XMLIter IT = Contain.Begin ();
while (IT != Contain.End ())
{
	std::cout << "Значение элемента: " << (*IT)->Get () << std::endl;	
	IT++;
}
\endcode

\section attmove Перемещение по атрибутам контейнера либо декларации.
Используется итератор, объект класса XMLAttrIter.
Движение вперед и назад.
Значение итератора на начало и конец надо получить у
объекта класса XMLContain либо XMLDeclar (в зависимости
от того по объекту какого класса требуется
перемещаться).
\code
XMLContain Contain;
XMLAttrIter AI = Contain.BeginAttr ();
while (AI != Contain.EndAttr ())
{
	std::cout << (AI.Get ()).att << "  " << (AI.Get ()).val;
	AI++;
}
\endcode

\section hedermove Перемещение по комментариям вне корневого контейнера.
Используется итератор, объект класса XMLCommentIter.
Движение вперед и назад.
Значение итератора на начало и конец надо получить у
объекта класса XMLDocument.
\code
XMLDocument DOC;
XMLCommentIter CI = DOC.BeginCommentIter ();
while (CI != DOC.EndCommentIter ())
{
	//вывести комментарий
	std::cout << "Значение комментария: " << СI.Get () << std::endl;
	//установить комментарий
	CI.Set ("Новый комментарий");
	CI++;
}
\endcode
 */

/**
 * \page get Получение информации из элементов тела.

\section gettype Получение типа элемента.
Метод XMLNode::GetType():
\code
XMLNode* pNODE;
switch (pNODE->GetType ())
{
	case XMLDocument::Comment
	{
		std::cout << "Это комментарий!";
		break;
	}
	case XMLDocument::Container
	{
		std::cout << "Это контейнер!";
		break;
	}
	case XMLDocument::Text
	{
		std::cout << "Это текст!";
		break;
	}
}
\endcode

\section getlevel Получение уровня вложенности элемента.
Метод XMLNode::GetLevel():
\code
XMLNode* pNODE;
std::cout << "Вложенность: " << pNode->GetLevel ();
\endcode

\section getval Получение значения элемента.
Метод XMLNode::Get():
\code
XMLNode* pNODE;
switch (pNODE->GetType ())
{
	case XMLDocument::Comment
	{
		//для элемента комментарий
		//значением является комментарий без <!--  -->
		std::cout << pNode->Get ();
		break;
	}
	case XMLDocument::Container
	{
		//для элемента контейнер
		//значением является
		//имя тега
		std::cout << pNode->Get ();
		break;
	}
	case XMLDocument::Text
	{
		//для элемента текст
		//значением является текст
		std::cout << pNode->Get ();
		break;
	}
}
\endcode

\section getatt Получение атрибутов контейнера.

\subsection getatthasa Проверка наличия атрибутов.
Метод XMLContain::HasItAttr():
\code
XMLContain Container;
if (Container.HasItAttr ())
{
	std::cout << "Атрибуты есть!";
}
\endcode

\subsection getattval Получение значения атрибута.
Метод XMLContain::GetAtt(std::string a):
\code
XMLContain Container;
std::string a;
if (Container.HasItAttr ())
{
	std::cout << "Атрибут: " << a << " Значение: " << Container.GetAtt (a);
}
\endcode

\subsection getattpair Получение пары атрибут-значение.
Метод XMLAttrIter::Get():
\code
XMLAttrIter ITA;
std::cout << "Атрибут: " << (ITA.Get ()).att << " Значение: " << (ITA.Get ()).val;
\endcode
 */

/**
 * \page edit Редактирование элеметов внутри тела.

\section edittex Текст.

- Создание. Конструктор XMLText:
\code
XMLText TEXT("Это текст");
\endcode

- Удаление.
 - Метод XMLBody::DelNode(XMLText& t):
\code
XMLBody BODY;
XMLText TEXT;
BODY.DelNode(TEXT);
\endcode
 - Метод XMLBody::DelNode(XMLIter& it):
\code
XMLBody BODY;
XMLIter IT;
//IT указывает на текст
BODY.DelNode(IT);
\endcode
 - Метод XMLBody::DelNode(XMLBodyIter& it):
\code
XMLBody BODY;
XMLBodyIter ITB;
//ITB указывает на текст
BODY.DelNode(ITB);
\endcode

- Изменение. Метод XMLNode::Set(std::string t):
\code
XMLNode* pText;
pText->Set("Новый текст!");
\endcode


\section editcom Комментарий.

- Создание. Конструктор XMLComment:
\code
XMLComment COMMENT("Это комментарий");
\endcode

- Удаление.
 - Метод XMLBody::DelNode(XMLComment& c):
\code
XMLBody BODY;
XMLComment COMMENT;
BODY.DelNode(COMMENT);
\endcode
 - Метод XMLBody::DelNode(XMLIter& it):
\code
XMLBody BODY;
XMLIter IT;
//IT указывает на комментарий
BODY.DelNode(IT);
\endcode
 - Метод XMLBody::DelNode(XMLBodyIter& it):
\code
XMLBody BODY;
XMLBodyIter ITB;
//ITB указывает на комментарий
BODY.DelNode(ITB);
\endcode
 
- Изменение. Метод XMLNode::Set(std::string c):
\code
XMLNode* pComment;
//pComment указывает на комментарий
pComment->Set("Новый комментарий!");
\endcode

\section editcon Контейнер.

- Создание. Конструктор XMLContain:
\code
XMLContain CONTAIN("Container");
\endcode

- Вложение.
 - Вложение текста. Метод XMLContain::operator+(XMLText t):
	 \code
	  XMLContain CONTAIN("Container");
	  XMLText TEXT("Текст");
	  CONTAIN << TEXT;
	 
	  // или так:
	  CONTAIN += TEXT;
	 \endcode
 - Вложение комментария. Метод XMLContain::operator+(XMLComment c):
	 \code
	  XMLContain CONTAIN("Container");
	  XMLText COMMENT("Комментарий");
	  CONTAIN << COMMENT;

	  // или так:
	  CONTAIN += COMMENT;
	 \endcode
 - Вложение контейнера. Метод XMLContain::operator+(XMLContain e):
	 \code
	  XMLContain CONTAIN("Container");
	  XMLContain A("A");
	  XMLContain B("B");
	  CONTAIN << A << B;

	  // или так:
	  CONTAIN += A += B;
	 \endcode

- Удаление.
 - Метод XMLBody::DelNode(XMLContain& c):
\code
XMLBody BODY;
XMLContain CONTAIN;
BODY.DelNode(CONTAIN);
\endcode
 - Метод XMLBody::DelNode(XMLIter& it):
\code
XMLBody BODY;
XMLIter IT;
//IT указывает на контейнер
BODY.DelNode(IT);
\endcode
 - Метод XMLBody::DelNode(XMLBodyIter& it):
\code
XMLBody BODY;
XMLBodyIter ITB;
//ITB указывает на контейнер
BODY.DelNode(ITB);
\endcode

- Изменение. Метод XMLNode::Set(std::string c):
\code
XMLNode* pContainer;
//pContainer указывает на контейнер
pContainer->Set("Новое значение тега!");
\endcode

\warning Удаление контейнера приводит к удалению
всех вложенных в него элементов.

\section editconatt Атрибуты контейнера.
- Добавить атрибут. Метод XMLContain::AddAtt(std::string a, std::string v):
\code
XMLContain CONTAIN;
CONTAIN.AddAtt("атрибут", "значение");
\endcode

- Удалить атрибут. Метод XMLContain::DelAtt(std::string a):
\code
XMLContain CONTAIN;
//удаляется "атрибут" и его значение
CONTAIN.DelAtt("атрибут");
\endcode

\section edel Удаление структуры документа.
Удаление всего документа из памяти, за исключением
корневого контейнера, осуществляется методом XMLDocument::Del():
\code
XMLDocument DOC;
DOC.Del();
\endcode
 */

/**
 * \page find Поиск в документе.

\section findintro Введение.

\subsection newfind Новый способ.
Идея &mdash; создать контейнер для результатов поиска
(XMLSearchResult),
создать строку поиска (xpath), предать их в метод
(
void XMLBody::Search (const xpath path, XMLSearchResult& result) или
void XMLContain::Search (const xpath path, XMLSearchResult& result)
)
метод
сохранит найденные контейнерные узлы в контейнере
для результатов поиска. Потом, итератором
(XMLSearchResultIter)
пройти по контейнеру
для результатов поиска и вытащить найденные результаты
(
XMLContain* XMLSearchResultIter::operator * ();		
).
\code
	XMLSearchResult SR;
	XMLBody& Body = DOC.GetBody ();
	xpath search = "// *[not(@*)]";
	Body.Search (search, SR);
	XMLSearchResultIter current;
	XMLSearchResultIter end;
	current = SR.Begin ();
	end = SR.End ();
	while (current != end) {

		std::cout << std::endl;
		std::cout << "Result: ";
		std::cout << (*current)->Get ();

	   current++;
	}
\endcode

\warning
При поиске от контейнерного элемента, строка поиска
не должна содержать лидирующий слэш. При поиске от тела
документа, строка поиска должна начинаться со слэш.

Более подробно о языке XPath можно посмотреть здесь:
<p> http://www.w3schools.com/xpath/xpath_intro.asp </p>
<p> http://www.zvon.org/xxl/XPathTutorial/General_rus/examples.html </p>

\section findstring Строка поиска.

\subsection findstringcom Общие положения.

Строка поиска представляет собой путь локализации контейнера, состоящий
из одного или более шагов локализации разделённых знаком:
\code
/
\endcode
Строка поиска всегда начинается с этого энака.

Шаг локализации представляет собой:
\code
ось::имя_контейнера[предикат]
\endcode
Ось это направление в котором происходит локализация (поиск).
Ось child, ось по умолчанию, может не указываться, предусматривает
поиск от родителя к ребёнку. В данной библиотеке на момент
генерации документации реализована только ось child.

Имя контейнера обязательно должно присутствовать в строке
либо может быть заменено на значок "все": \b *.

Предикат может отсутствовать. Предикат является дополнительным условием поиска.
В настоящий момент реализован предикат атрибута: \b \@.

\subsection findstringexm Примеры.
Отбор всех контейнеров:
\verbatim
// *
\endverbatim

Отбор всех контейнеров C, которые являются детьми корневого
контейнера ROOT:
\verbatim
/ROOT/C
\endverbatim

Отбор всех контейнеров C:
\verbatim
//C
\endverbatim

Отбор всех контейнеров C, которые являются потомками корневого
контейнера ROOT вне зависимости от близости родства:
\verbatim
/ROOT//C
\endverbatim

Отбор любого контейнера, имеющего любой атрибут:
\verbatim
// *[@*]
\endverbatim

Отбор любого контейнера, не имеющего любой атрибут:
\verbatim
// *[not(@*)]
\endverbatim

Отбор всех контейнеров C с атрибутом at:
\verbatim
//C[@at]
\endverbatim

Отбор всех контейнеров C с атрибутом at и значением
атрибута 6:
\verbatim
//C[@at='6']
\endverbatim
 */
