/*
 * Copyright 2008, Boris Popov <popov.b@gmail.com>
 * Copyright 2009, Boris Popov <popov.b@gmail.com>
 */

/*
 * utf-8:
 * Шаги локализации.
 * Этот файл описывает возможные объекты
 * шагов локализации.
 *
 */

#define NO_DEBUG
#ifdef DEBUG
#include <iostream>
#endif

#ifndef __steps__hpp__
#define __steps__hpp__

#include <string>
#include <list>
#include <XMLDocument.hpp>

class XMLSearchResult;
class XMLNode;

namespace xmlxpath {

typedef std::string string;
typedef std::list<XMLContain*> TempContain;
typedef std::list<XMLContain*>::iterator
   TempContainIterator;
typedef std::list<XMLNode*>::iterator
   NodeIterator;

class Predicat;

// =========================================== //
class Step {

   // каждый объект имеет указатель
   // на предикат. каждый найденный
   // контейнер должен быть проверен
   // на предикат. в случае успеха,
   // контейнер помещается в результирущий
   // контейнер.

   public:
      Step (const Predicat* predicat) :
	 m_predicat (predicat) {
	    
	    m_tempContainer.clear ();
	    return;
	 }

      Step () : m_predicat (0) {
	    
	 m_tempContainer.clear ();
	 return;
      }

      void SetPredicat (const Predicat* predicat) {
	 m_predicat = predicat; }
      
      const Predicat* GetPredicat () const {
	 return m_predicat; }

      virtual ~Step () {
	 
	 m_tempContainer.clear ();
	 return;
      }

      void Get (XMLSearchResult&
	    container);

      virtual void GetFirstFromRoot
	 (XMLContain* root, XMLSearchResult&
	  container) = 0;

   protected:
      const Predicat* m_predicat;
      TempContain m_tempContainer;
      void m_CopyToTemp (XMLSearchResult&
	    container);

   private:
      virtual void
	 m_Compare (TempContainIterator iter,
	    XMLSearchResult& container) = 0;
  
};
// =========================================== //

// =========================================== //
class ChildLexem : public Step {

   // [utf-8: пример: "/что-то"]

   public:
      ChildLexem (const string lexem,
	    const Predicat* predicat) :
	 Step (predicat),
	 m_lexem (lexem) {

      #ifdef DEBUG
      std::cerr << std::endl;
      std::cerr << "make ChildLexem";
      std::cerr << std::endl;
      #endif

	 }

      ChildLexem (const string lexem) :
	 Step (0),
	 m_lexem (lexem) {
	 
      #ifdef DEBUG
      std::cerr << std::endl;
      std::cerr << "make ChildLexem";
      std::cerr << std::endl;
      #endif

	 }
      
      ~ChildLexem ();

      void GetFirstFromRoot
	 (XMLContain* root, XMLSearchResult&
	    container);
     
   private:
      const string m_lexem;

      void m_Compare (TempContainIterator iter,
	    XMLSearchResult& container);

      void m_CompareContainer (XMLContain* cnt,
	    XMLSearchResult& container);

};
// =========================================== //

// =========================================== //
class ChildLexemAll : public Step {

   // [utf-8: пример: "//что-то"]
   
   public:
      ChildLexemAll (const string lexem,
	    const Predicat* predicat) :
	 Step (predicat),
	 m_lexem (lexem) {
	 
      #ifdef DEBUG
      std::cerr << std::endl;
      std::cerr << "make ChildLexemAll";
      std::cerr << std::endl;
      #endif

	 }

      ChildLexemAll (const string lexem) :
	 Step (0),
	 m_lexem (lexem) {
	 
      #ifdef DEBUG
      std::cerr << std::endl;
      std::cerr << "make ChildLexemAll";
      std::cerr << std::endl;
      #endif
	 
	 }
      
      ~ChildLexemAll ();
      
      void GetFirstFromRoot
	 (XMLContain* root, XMLSearchResult&
	    container);
   
   private:
      const string m_lexem;

      void m_Compare (TempContainIterator iter,
	    XMLSearchResult& container);
      
      void m_CompareContainer (XMLContain* cnt,
	    XMLSearchResult& container);
      
};
// =========================================== //

// =========================================== //
class ChildAny : public Step {

   // [utf-8: пример: "/*"]

   public:
      ChildAny (const Predicat* predicat) :
	 Step (predicat) {
	 
      #ifdef DEBUG
      std::cerr << std::endl;
      std::cerr << "make ChildAny";
      std::cerr << std::endl;
      #endif
	 
	 }
      
      ChildAny () :
	 Step (0) {
	 
      #ifdef DEBUG
      std::cerr << std::endl;
      std::cerr << "make ChildAny";
      std::cerr << std::endl;
      #endif
	 
	 }

      ~ChildAny ();

      void GetFirstFromRoot
	 (XMLContain* root, XMLSearchResult&
	    container);

   private:
      void m_Compare (TempContainIterator iter,
	    XMLSearchResult& container);
};
// =========================================== //

// =========================================== //
class ChildAnyAll : public Step {

   // [utf-8: пример: "//*"]
   
   public:
      ChildAnyAll (const Predicat* predicat) :
	 Step (predicat) {
	 
      #ifdef DEBUG
      std::cerr << std::endl;
      std::cerr << "make ChildAnyAll";
      std::cerr << std::endl;
      #endif
	 
	 }

      ChildAnyAll () :
	 Step (0) {
	 
      #ifdef DEBUG
      std::cerr << std::endl;
      std::cerr << "make ChildAnyAll";
      std::cerr << std::endl;
      #endif
	 }

      ~ChildAnyAll ();
     
      void GetFirstFromRoot
	 (XMLContain* root, XMLSearchResult&
	    container);

   private:
      void m_Compare (TempContainIterator iter,
	    XMLSearchResult& container);
};
// =========================================== //


//// ПРЕДИКАТЫ ////

// =========================================== //
enum Function {
   FUNC_FAKE,
   FUNC_NOT
};
// =========================================== //

// =========================================== //
class Predicat {

   public:
      Predicat (const Function f) :
	 m_func (f) {
	 #ifdef DEBUG
	    std::cerr << "func " << f;
	 #endif
	 }

      Predicat () :
	 m_func (FUNC_FAKE) {
	 #ifdef DEBUG
	    std::cerr << "func " << FUNC_FAKE;
	 #endif
	 }

      virtual ~Predicat () { return; }

      virtual bool MatchPredicat
	 (XMLContain& container) const = 0;

   protected:
      const Function m_func;
};
// =========================================== //

// =========================================== //
class Predicat_ATT : public Predicat {

   // [utf-8: пример "@at"]

   public:
      Predicat_ATT (const string at) :
	 Predicat (),
	 m_attribute (at) {
	 
      #ifdef DEBUG
      std::cerr << std::endl;
      std::cerr << "make Predicat_ATT";
      std::cerr << std::endl;
      #endif
	 
	 }

      Predicat_ATT (const string at,
	    const Function f) :
	 Predicat (f),
	 m_attribute (at) {
	 
      #ifdef DEBUG
      std::cerr << std::endl;
      std::cerr << "make Predicat_ATT";
      std::cerr << f;
      std::cerr << std::endl;
      #endif
	 
	 }
      
      bool MatchPredicat
	 (XMLContain& container) const;

      ~Predicat_ATT ();

   private:
      const string m_attribute;
};
// =========================================== //

// =========================================== //
class Predicat_ATTALL : public Predicat {
   
   // [utf-8: пример "@*"]

   public:
      Predicat_ATTALL () :
	 Predicat () {
	 
      #ifdef DEBUG
      std::cerr << std::endl;
      std::cerr << "make Predicat_ATTALL";
      std::cerr << std::endl;
      #endif
	 
	 }

      Predicat_ATTALL (const Function f) :
	 Predicat (f) {
	 
      #ifdef DEBUG
      std::cerr << std::endl;
      std::cerr << "make Predicat_ATTALL";
      std::cerr << f;
      std::cerr << std::endl;
      #endif
	 
	 }
      
      bool MatchPredicat
	 (XMLContain& container) const;

      ~Predicat_ATTALL ();

   private:
};
// =========================================== //

// =========================================== //
class Predicat_ATTandVAL : public Predicat {
   
   // [utf-8: пример "@at='6'"]

   public:
      Predicat_ATTandVAL (const string at,
	    const string val) :
	 Predicat (),
	 m_attribute (at),
	 m_value (val) {
	 
      #ifdef DEBUG
      std::cerr << std::endl;
      std::cerr << "make Predicat_ATTandVAL";
      std::cerr << std::endl;
      #endif
	 
	 }

      Predicat_ATTandVAL (const string at,
	    const string val,
	    const Function f) :
	 Predicat (f),
	 m_attribute (at),
	 m_value (val) {
	 
      #ifdef DEBUG
      std::cerr << std::endl;
      std::cerr << "make Predicat_ATTandVAL ";
      std::cerr << f;
      std::cerr << std::endl;
      #endif
	 
	 }
      
      bool MatchPredicat
	 (XMLContain& container) const;
      
      ~Predicat_ATTandVAL ();

   private:
      const string m_attribute;
      const string m_value;
};
// =========================================== //

// =========================================== //
class FunctionNOT {
  
   public:
      FunctionNOT ();
      ~FunctionNOT ();

      bool Get (bool in) const;

   private:

};
// =========================================== //

//// ПРЕДИКАТЫ ////

}

#endif
