/*
 * Copyright (c) 2010, Boris Popov <popov.b@gmail.com>
 *
 * Permission to use, copy, modify, and/or distribute
 * this software for any purpose with or without fee
 * is hereby granted, provided that the above copyright
 * notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "demirror.h"
#include "cbuffer.h"
#include "bool.h"

static void _DestroyAkk (DeMirror);
static void _ReNewAkk (DeMirror);

static Bool _IsItAmpersand (const char);
static Bool _IsItSemicolon (const char);

static Bool _IsIt_a (const char);
static Bool _IsIt_m (const char);
static Bool _IsIt_p (const char);

static Bool _IsIt_l (const char);
static Bool _IsIt_g (const char);
static Bool _IsIt_t (const char);

static Bool _IsIt_o (const char);
static Bool _IsIt_s (const char);

static Bool _IsIt_q (const char);
static Bool _IsIt_u (const char);

static void _AddToAkk (DeMirror, const char);
static void _AddToAkk_Amp (DeMirror);
static void _AddToAkk_AmpA (DeMirror);
static void _AddToAkk_AmpAP (DeMirror);
static void _AddToAkk_AmpAPO (DeMirror);
static void _AddToAkk_AmpAPOS (DeMirror);
static void _AddToAkk_AmpAM (DeMirror);
static void _AddToAkk_AmpAMP (DeMirror);
static void _AddToAkk_AmpL (DeMirror);
static void _AddToAkk_AmpLT (DeMirror);
static void _AddToAkk_AmpG (DeMirror);
static void _AddToAkk_AmpGT (DeMirror);
static void _AddToAkk_AmpQ (DeMirror);
static void _AddToAkk_AmpQU (DeMirror);
static void _AddToAkk_AmpQUO (DeMirror);
static void _AddToAkk_AmpQUOT (DeMirror);

static void _AddSignToAkk_Ampersand (DeMirror);
static void _AddSignToAkk_Lt (DeMirror);
static void _AddSignToAkk_Gt (DeMirror);
static void _AddSignToAkk_Apos (DeMirror);
static void _AddSignToAkk_Quot (DeMirror);

static void _ProcessAfterAmpersand (DeMirror, CBuffer);

static void _ProcessAfterAposL (DeMirror, CBuffer);
static void _ProcessAfterAposLt (DeMirror, CBuffer);

static void _ProcessAfterAmpG (DeMirror, CBuffer);
static void _ProcessAfterAmpGt (DeMirror, CBuffer);

static void _ProcessAfterAmpA (DeMirror, CBuffer);
static void _ProcessAfterAmpQ (DeMirror, CBuffer);
static void _ProcessAfterAmpQu (DeMirror, CBuffer);
static void _ProcessAfterAmpQuo (DeMirror, CBuffer);
static void _ProcessAfterAmpQuot (DeMirror, CBuffer);
static void _ProcessAfterAmpAm (DeMirror, CBuffer);
static void _ProcessAfterAmpAmp (DeMirror demirror, CBuffer buffer);
static void _ProcessAfterAmpAp (DeMirror, CBuffer);
static void _ProcessAfterAmpApo (DeMirror, CBuffer);
static void _ProcessAfterAmpApos (DeMirror, CBuffer);

static void _ProcessNext (DeMirror, CBuffer, void (*addtoakk) (DeMirror),
		Bool (*boolif) (const char), void (*pocessnext) (DeMirror, CBuffer));

static void _ProcessEnd (DeMirror, CBuffer, void (*addtoakk) (DeMirror),
		void (*addsign) (DeMirror));

static void _ProcessAfterAmpersand (DeMirror demirror, CBuffer buffer) {

	char ch;
	Result result = GetChar (buffer);
	if (HasItEof (result)) {
		
		_AddSignToAkk_Ampersand (demirror);
		return;
	}

	ch = GetCharFromResult (result);

	if (_IsIt_a (ch)) {

		_ProcessAfterAmpA (demirror, buffer);

	} else if (_IsIt_l (ch)) {

		_ProcessAfterAposL (demirror, buffer);
	
	} else if (_IsIt_g (ch)) {

		_ProcessAfterAmpG (demirror, buffer);

	} else if (_IsIt_q (ch)) {

		_ProcessAfterAmpQ (demirror, buffer);

	} else {

		_AddSignToAkk_Ampersand (demirror);
		_AddToAkk (demirror, ch);
	}
	return;
}

static void _ProcessAfterAmpQ (DeMirror demirror, CBuffer buffer) {

	_ProcessNext
		(demirror, buffer, _AddToAkk_AmpQ, _IsIt_u, _ProcessAfterAmpQu);
	return;
}

static void _ProcessAfterAmpQu (DeMirror demirror, CBuffer buffer) {

	_ProcessNext
		(demirror, buffer, _AddToAkk_AmpQU, _IsIt_o, _ProcessAfterAmpQuo);
	return;
}

static void _ProcessNext (DeMirror demirror, CBuffer buffer,
		void (*addtoakk) (DeMirror), Bool (*boolif) (const char),
		void (*pocessnext) (DeMirror, CBuffer)) {

	char ch;
	Result result = GetChar (buffer);
	if (HasItEof (result)) {

		addtoakk (demirror);
		return;
	}
			
	ch = GetCharFromResult (result);
	if (boolif (ch)) {

		pocessnext (demirror, buffer);

	} else {
		
		addtoakk (demirror);
		_AddToAkk (demirror, ch);
	}
	return;
}

static void _ProcessEnd (DeMirror demirror, CBuffer buffer,
		void (*addtoakk) (DeMirror), void (*addsign) (DeMirror)) {

	char ch;
	Result result = GetChar (buffer);
	if (HasItEof (result)) {

		addtoakk (demirror);
		return;
	}
			
	ch = GetCharFromResult (result);
	
	if (_IsItSemicolon (ch)) {

		addsign (demirror);

	} else {
		
		addtoakk (demirror);
		_AddToAkk (demirror, ch);
	}
	return;
}

static void _ProcessAfterAmpQuo (DeMirror demirror, CBuffer buffer) {

	_ProcessNext
		(demirror, buffer, _AddToAkk_AmpQUO, _IsIt_t, _ProcessAfterAmpQuot);
	return;
}

static void _ProcessAfterAmpQuot (DeMirror demirror, CBuffer buffer) {

	_ProcessEnd
		(demirror, buffer, _AddToAkk_AmpQUOT, _AddSignToAkk_Quot);
	return;
}

static void _ProcessAfterAmpA (DeMirror demirror, CBuffer buffer) {

	char ch;
	Result result = GetChar (buffer);
	if (HasItEof (result)) {
		
		_AddToAkk_AmpA (demirror);
		return;
	}

	ch = GetCharFromResult (result);

	if (_IsIt_m (ch)) {

		_ProcessAfterAmpAm (demirror, buffer);

	} else if (_IsIt_p (ch)) {
		
		_ProcessAfterAmpAp (demirror, buffer);

	} else {

		_AddToAkk_AmpA (demirror);
		_AddToAkk (demirror, ch);
	}
	return;
}

static void _ProcessAfterAmpAm (DeMirror demirror, CBuffer buffer) {

	_ProcessNext
		(demirror, buffer, _AddToAkk_AmpAM, _IsIt_p, _ProcessAfterAmpAmp);
	return;
}

static void _ProcessAfterAmpAmp (DeMirror demirror, CBuffer buffer) {

	_ProcessEnd
		(demirror, buffer, _AddToAkk_AmpAMP, _AddSignToAkk_Ampersand);
	return;
}

static void _ProcessAfterAmpAp (DeMirror demirror, CBuffer buffer) {

	_ProcessNext
		(demirror, buffer, _AddToAkk_AmpAP, _IsIt_o, _ProcessAfterAmpApo);
	return;
}

static void _ProcessAfterAmpApo (DeMirror demirror, CBuffer buffer) {

	_ProcessNext
		(demirror, buffer, _AddToAkk_AmpAPO, _IsIt_s, _ProcessAfterAmpApos);
	return;
}

static void _ProcessAfterAmpApos (DeMirror demirror, CBuffer buffer) {

	_ProcessEnd
		(demirror, buffer, _AddToAkk_AmpAPOS, _AddSignToAkk_Apos);
	return;
}

static void _ProcessAfterAmpG (DeMirror demirror, CBuffer buffer) {

	_ProcessNext
		(demirror, buffer, _AddToAkk_AmpG, _IsIt_t, _ProcessAfterAmpGt);
	return;
}

static void _ProcessAfterAmpGt (DeMirror demirror, CBuffer buffer) {

	_ProcessEnd
		(demirror, buffer, _AddToAkk_AmpGT, _AddSignToAkk_Gt);
	return;
}

static void _ProcessAfterAposL (DeMirror demirror, CBuffer buffer) {

	_ProcessNext
		(demirror, buffer, _AddToAkk_AmpL, _IsIt_t, _ProcessAfterAposLt);
	return;
}

static void _ProcessAfterAposLt (DeMirror demirror, CBuffer buffer) {

	_ProcessEnd
		(demirror, buffer, _AddToAkk_AmpLT, _AddSignToAkk_Lt);
	return;
}

DeMirror InitDeMirror () {

	DeMirror demirror = NULL;
	demirror = malloc (sizeof (struct _DeMirror));
	if (demirror != NULL) {

		StringAkkN** akk;
		demirror->Akk = NULL;
		
		akk = InitStringAkkN (4096); 
		if (akk != NULL) {
		
			demirror->Akk = akk;
		}
	}
	return demirror;
}

void DestroyDeMirror (DeMirror demirror) {

	_DestroyAkk (demirror);
	if (demirror != NULL) {

		free (demirror);
		demirror = NULL;
	}
	return;
}

CString DecodeDeMirror (DeMirror demirror, CString string) {

	CBuffer cbuffer;
	Result result;

	if (string == NULL) {

		return NULL;
	}
	
	if (demirror == NULL) {

		return NULL;
	}

	if (strlen (string) == 0) {

		return "";
	}

	_ReNewAkk (demirror);

	cbuffer = InitCBufferFromString (string);
	if (cbuffer == NULL) {

		return NULL;
	}

	result = GetChar (cbuffer);
	while (! HasItEof (result)) {
		
		char ch = GetCharFromResult (result);

		if (_IsItAmpersand (ch)) {

			_ProcessAfterAmpersand (demirror, cbuffer);

		} else {

			_AddToAkk (demirror, ch);
		}
	
		result = GetChar (cbuffer);
	}

	DestroyCBuffer (cbuffer);
	return (GetStringFromStringAkkN (demirror->Akk));
}

static void _DestroyAkk (DeMirror demirror) {

	if (demirror != NULL) {

		if ((demirror->Akk) != NULL) {
			
			DestoryStringAkkN (demirror->Akk);
			demirror->Akk = NULL;
		}
	}
	return;
}

static Bool _IsItAmpersand (const char ch) {

	return IsIt (ch, '&');
}

static Bool _IsItSemicolon (const char ch) {

	return IsIt (ch, ';');
}

static Bool _IsIt_a (const char ch) {

	return IsIt (ch, 'a');
}

static Bool _IsIt_m (const char ch) {

	return IsIt (ch, 'm');
}

static Bool _IsIt_p (const char ch) {

	return IsIt (ch, 'p');
}

static Bool _IsIt_l (const char ch) {

	return IsIt (ch, 'l');
}

static Bool _IsIt_g (const char ch) {

	return IsIt (ch, 'g');
}

static Bool _IsIt_t (const char ch) {

	return IsIt (ch, 't');
}

static Bool _IsIt_o (const char ch) {

	return IsIt (ch, 'o');
}

static Bool _IsIt_s (const char ch) {

	return IsIt (ch, 's');
}

static Bool _IsIt_q (const char ch) {

	return IsIt (ch, 'q');
}

static Bool _IsIt_u (const char ch) {

	return IsIt (ch, 'u');
}

static void _ReNewAkk (DeMirror demirror) {

	_DestroyAkk (demirror);
	if (demirror != NULL) {

		demirror->Akk = InitStringAkkN (4096);
	}
	return;
}

static void _AddToAkk (DeMirror demirror, const char ch) {

	AddByteToStringAkkN (demirror->Akk, ch);
	return;
}

static void _AddToAkk_Amp (DeMirror demirror) {

	_AddToAkk (demirror, '&');
	return;
}

static void _AddToAkk_AmpA (DeMirror demirror) {

	_AddToAkk_Amp (demirror);
	_AddToAkk (demirror, 'a');
	return;
}

static void _AddToAkk_AmpAP (DeMirror demirror) {

	_AddToAkk_AmpA (demirror);
	_AddToAkk (demirror, 'p');
	return;
}

static void _AddToAkk_AmpAPO (DeMirror demirror) {

	_AddToAkk_AmpAP (demirror);
	_AddToAkk (demirror, 'o');
	return;
}

static void _AddToAkk_AmpAPOS (DeMirror demirror) {

	_AddToAkk_AmpAPO (demirror);
	_AddToAkk (demirror, 's');
	return;
}

static void _AddToAkk_AmpAM (DeMirror demirror) {

	_AddToAkk_AmpA (demirror);
	_AddToAkk (demirror, 'm');
	return;
}

static void _AddToAkk_AmpAMP (DeMirror demirror) {


	_AddToAkk_AmpAM (demirror);
	_AddToAkk (demirror, 'p');
	return;
}

static void _AddToAkk_AmpL (DeMirror demirror) {

	_AddToAkk_Amp (demirror);
	_AddToAkk (demirror, 'l');
	return;
}

static void _AddToAkk_AmpLT (DeMirror demirror) {

	_AddToAkk_AmpL (demirror);
	_AddToAkk (demirror, 't');
	return;
}

static void _AddToAkk_AmpG (DeMirror demirror) {

	_AddToAkk_Amp (demirror);
	_AddToAkk (demirror, 'g');
	return;
}

static void _AddToAkk_AmpGT (DeMirror demirror) {


	_AddToAkk_AmpG (demirror);
	_AddToAkk (demirror, 't');
	return;
}

static void _AddToAkk_AmpQ (DeMirror demirror) {

	_AddToAkk_Amp (demirror);
	_AddToAkk (demirror, 'q');
	return;
}

static void _AddToAkk_AmpQU (DeMirror demirror) {

	_AddToAkk_AmpQ (demirror);
	_AddToAkk (demirror, 'u');
	return;
}

static void _AddToAkk_AmpQUO (DeMirror demirror) {
	
	_AddToAkk_AmpQU (demirror);
	_AddToAkk (demirror, 'o');
	return;
}

static void _AddToAkk_AmpQUOT (DeMirror demirror) {

	_AddToAkk_AmpQUO (demirror);
	_AddToAkk (demirror, 't');
	return;
}

static void _AddSignToAkk_Ampersand (DeMirror demirror) {
	
	_AddToAkk (demirror, '&');
	return;
}

static void _AddSignToAkk_Lt (DeMirror demirror) {
	
	_AddToAkk (demirror, '<');
	return;
}

static void _AddSignToAkk_Gt (DeMirror demirror) {
	
	_AddToAkk (demirror, '>');
	return;
}

static void _AddSignToAkk_Apos (DeMirror demirror) {
	
	_AddToAkk (demirror, '\'');
	return;
}

static void _AddSignToAkk_Quot (DeMirror demirror) {
	
	_AddToAkk (demirror, '\"');
	return;
}
