/*
 * Copyright (c) 2010, Boris Popov <popov.b@gmail.com>
 *
 * Permission to use, copy, modify, and/or distribute
 * this software for any purpose with or without fee
 * is hereby granted, provided that the above copyright
 * notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "rmvdelimiter.h"
#include "bool.h"

static Bool _ProbingLeftSpace (RemoverDelimiter);
static Bool _ProbingRightSpace (RemoverDelimiter);
static Bool _NoSpaces (RemoverDelimiter);
static Bool _IsItSpace (const char);
static void _ClearTempString (RemoverDelimiter);

static size_t _GetStart (RemoverDelimiter);
static size_t _GetEnd (RemoverDelimiter);
static void _GetStartEnd (RemoverDelimiter, size_t*, size_t*);
static void _MakeTempString (RemoverDelimiter, size_t, size_t);

RemoverDelimiter InitRemoverDelimiter () {

	RemoverDelimiter rd = NULL;
	rd = (RemoverDelimiter) malloc (sizeof (struct _RemoverDelimiter));

	if (rd != NULL) {

		rd->in_string = NULL;
		rd->out_string = NULL;
		rd->temp = NULL;
	}
	return rd;
}

void DestroyRemoverDelimiter (RemoverDelimiter rd) {

	if (rd != NULL) {

		_ClearTempString (rd);
		free (rd);
		rd = NULL;
	}
	return;
}

CString RemoveDelimiter (RemoverDelimiter rd,
		CString string) {

	if (rd == NULL) {

		return NULL;
	}

	if (string == NULL) {

		return "";
	}

	_ClearTempString (rd);
	
	rd->out_string = string;
	rd->in_string = string;

	if (! _NoSpaces (rd)) {
	
		size_t start;
		size_t end;
		_GetStartEnd (rd, &start, &end);
		_MakeTempString (rd, start, end);
		rd->out_string = rd->temp;
	}
	return (rd->out_string);
}

static Bool _ProbingLeftSpace (RemoverDelimiter rd) {

	const char ch = *(rd->in_string);
	return _IsItSpace (ch);
}

static Bool _ProbingRightSpace (RemoverDelimiter rd) {

	size_t len = strlen (rd->in_string);
	const char ch = *((rd->in_string) + len - 1);
	return _IsItSpace (ch);
}

static Bool _IsItSpace (const char ch) {
	
	if (isspace (ch < 0 ? 0 : ch)) {

		return SUCCESS;
	}
	return FAULT;
}

static Bool _NoSpaces (RemoverDelimiter rd) {

	if ( (! _ProbingLeftSpace (rd)) &&
			(! _ProbingRightSpace (rd)) ) {

		return SUCCESS;
	}
	return FAULT;
}

static void _ClearTempString (RemoverDelimiter rd) {

	if ((rd->temp) != NULL) {

		free (rd->temp);
		rd->temp = NULL;
	}
	return;
}

static size_t _GetStart (RemoverDelimiter rd) {

	size_t start = 0;
	size_t len;
	size_t i;
	const char* pch;

	if (! _ProbingLeftSpace (rd)) {

		return start;
	}
	
	pch = rd->in_string;
	len = strlen (rd->in_string);
	i = 0;
	
	for (i = 0; i < len; i++) {

		char ch = (*pch);
		if (! _IsItSpace (ch)) {

			break;
		}

		pch++;
		start++;
	}
	return start;
}

static size_t _GetEnd (RemoverDelimiter rd) {

	const char* pch;
	size_t i;
	size_t len = strlen (rd->in_string);
	size_t end = (len - 1);

	if (! _ProbingRightSpace (rd)) {

		return end;
	}

	pch = (rd->in_string) + len - 1;
	i = 0;
	
	for (i = (len - 1); i > 0; i--) {

		char ch = (*pch);
		if (! _IsItSpace (ch)) {

			break;
		}

		pch--;
		end--;
	}
	return end;
}

static void _GetStartEnd
(RemoverDelimiter rd, size_t* start, size_t* end) {

	*start = _GetStart (rd);
	*end = _GetEnd (rd);
	return;
}

static void _MakeTempString
(RemoverDelimiter rd, size_t start, size_t end) {

	if (start <= end) {

		rd->temp = (char*) malloc (end - start + 2);
		*((rd->temp) + (end - start + 1)) = '\0';
		strncpy ((rd->temp), ((rd->in_string) + start), (end - start + 1));

	} else {
		
		rd->temp = (char*) malloc (1);
		*(rd->temp) = '\0';
	}
	return;
}
