// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Класс обеспечивающий кеширование параметров при вычислении ускорений
///////////////////////////////////////////////////////////////////////////////
#include "AccAdapter.h"
#include "IAstroBase.h"
///////////////////////////////////////////////////////////////////////////////
// Инициализация
//-----------------------------------------------------------------------------

void AccAdapter::Init(const TimeJD& t, const MotionMode& m)
{
    T0 = t;

    IAstroBase* ab = GetAstroBase();
    ab->MtxJ20002GFCS(T0, BJ2000toGFCS);

    MMode = &m;
}

//-----------------------------------------------------------------------------
// Вычисление параметров
//-----------------------------------------------------------------------------
void AccAdapter::Set(const TimeJD& t, const Vect6& rv_j2000, bool precItrf)
{
    IAstroBase* ab = GetAstroBase();
    T_TDB = t;
    RJ2000 = rv_j2000;
    rv_j2000.GetVelocity(VJ2000);

    // Запрос параметров атмосферы

    bool fatm_sun = false;
    const IAtmosphereModel* atm = NULL;
    if (MMode->IsAtm) {
        atm = ab->GetEarthAtmosphere();
        if (atm != NULL && atm->IsSunPosNeed())
            fatm_sun = true;
    }

    // Положение Солнца в J2000
    // Оно нужно, если
    //   - учитывается гравитация Солнца
    //   - используется модель атмосферы, в которой нужно положение Солнца
    //   - учитывается световое давление

    if (MMode->IsPlanetSet(plIdSun) || fatm_sun || MMode->LightMode != lightNONE)
        ab->GetPosGeo(plIdSun, T_TDB, &SunJ2000);

    // Запрос характеристик гравитационного поля Земли
    bool feg_nc = false;
    const IGravityModel* eg = ab->GetPlanet(plIdEarth)->GetGravityModel();
    if (eg != NULL && (MMode->GrN > 0 || MMode->GrM > 0))
        feg_nc = true;

    // Если учитывается атмосфера или нецентральное поле, то будет
    // нужен переход в Гринвич
    if ((MMode->IsAtm && atm != NULL) || feg_nc) {
        double w = ab->GetEarthRotation();
        double dt = 0;
        if (precItrf) {
            ab->MtxJ20002GFCS(T_TDB, BJ2000toGFCS);
        }
        else {
            dt = (w * (T_TDB - T0)).GetDouble() * 86.4;
        }

        double cw = cos(dt);
        double sw = sin(dt);
        CJ2000toGFCS[0] = cw;
        CJ2000toGFCS[1] = sw;
        CJ2000toGFCS[2] = 0.0;
        CJ2000toGFCS[3] = -sw;
        CJ2000toGFCS[4] = cw;
        CJ2000toGFCS[5] = 0.0;
        CJ2000toGFCS[6] = 0.0;
        CJ2000toGFCS[7] = 0.0;
        CJ2000toGFCS[8] = 1.0;
        CJ2000toGFCS *= BJ2000toGFCS;

        CJ2000toGFCS.Transp(CGFCStoJ2000);
        // Вычисляем положение в Гринвиче
        RGr = rv_j2000;
        RGr *= CJ2000toGFCS;

        rv_j2000.GetVelocity(VGr);
        VGr *= CJ2000toGFCS;
        VGr.x += w * RGr.y;
        VGr.y -= w * RGr.x;
    }

    // Если используется атмосфера, вычисляем высоту и плотность

    if (MMode->IsAtm && atm != NULL) {
        if (fatm_sun) {
            SunGr = SunJ2000;
            SunGr *= CJ2000toGFCS;
        }
        Density = atm->Density(T_TDB, RGr, SunGr);
    }
}

///////////////////////////////////////////////////////////////////////////////
