///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Классы, используемые для задания априорных констант
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_APRIORICONST_H
#define BNOTOOL_APRIORICONST_H
///////////////////////////////////////////////////////////////////////////////
#include "Matrix.h"
#include "ITraject.h"
#include "twll.h"

///////////////////////////////////////////////////////////////////////////////
// Типы априорных констант
///////////////////////////////////////////////////////////////////////////////
enum AprioriConstType
{
    acUnknown = 0,
    // Априорный баллистический коэффициент
    acBC = 1,
    // Априорный коэффициент св. давления
    acLPC = 2
};

///////////////////////////////////////////////////////////////////////////////
// Общий интерфей априорных констант
///////////////////////////////////////////////////////////////////////////////
class IAprioriConst
{
public:
    // Конструктор
    IAprioriConst()
    {
        ACType = acUnknown;
    };
    // Деструктор
    virtual ~IAprioriConst()
    {
    };

    // Получить размерность
    virtual unsigned int GetSize() const = 0;
    // Получить вектор невязок
    virtual void GetDelVector(const DVector& u, DVector& del) const = 0;
    // Получить матрицу частных производных
    virtual void GetMatrix(const DVector& u, DMatrix& m) const = 0;
    // Получить весовую матрицу
    virtual void GetWeightMatrix(DSymmMatrix& m) const = 0;
    // Создать наследника
    virtual IAprioriConst* Clone() const = 0;

    virtual void Set(const DVector& u) = 0;

    // Получить тип
    AprioriConstType GetType() const
    {
        return ACType;
    };

    void SetType(AprioriConstType t)
    {
        ACType = t;
    };

protected:
    // Тип
    AprioriConstType ACType;
};

///////////////////////////////////////////////////////////////////////////////
// Представление априорной константы, как элемента связанного списка
///////////////////////////////////////////////////////////////////////////////
typedef class Twll_Entry_Class<IAprioriConst> IAprioriConstEntry;

///////////////////////////////////////////////////////////////////////////////
// Связанный список априорных констант
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
class DAprioriConst : public IAprioriConstEntry
{
public:
    // Конструктор
    DAprioriConst() : IAprioriConstEntry()
    {
    };
    // Деструктор
    virtual ~DAprioriConst()
    {
    };

    // Получить размерность
    virtual unsigned int GetSize() const
    {
        return 1;
    };
    // Получить вектор невязок
    virtual void GetDelVector(const DVector& u, DVector& del) const;
    // Получить матрицу частных производных
    virtual void GetMatrix(const DVector& u, DMatrix& m) const;
    // Получить весовую матрицу
    virtual void GetWeightMatrix(DSymmMatrix& m) const;
    // Создать наследника
    virtual IAprioriConst* Clone() const;
    // Установка нового значения по вектору
    virtual void Set(const DVector& u);

    // Установка параметров
    void SetParams(unsigned index, double v, double mse);

protected:
    // Индекс в векторе уточняемых параметров
    unsigned Index;
    // Значение константы
    double Value;
    // СКО константы
    double Mse;
};

///////////////////////////////////////////////////////////////////////////////
// Связанный список априорных констант
///////////////////////////////////////////////////////////////////////////////
class IAprioriConstList : public Twll_List_Class<IAprioriConstEntry>
{
public:
    // Добавление в список
    void Add(const IAprioriConst& a);
    // Удалените по типу
    void DeleteByType(AprioriConstType t);
    // Установка априорных констант по вектору
    void Set(const DVector& u);
};

///////////////////////////////////////////////////////////////////////////////
#endif
