///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Классы моделей земной атмосферы
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_ATMOSPHERE_H
#define BNOTOOL_ATMOSPHERE_H
///////////////////////////////////////////////////////////////////////////////
#include "IModel.h"

///////////////////////////////////////////////////////////////////////////////
// Класс статической атмосферы
///////////////////////////////////////////////////////////////////////////////
class StaticAtmosphere : public IAtmosphereModel
{
public:
    // Вычисление плотности. Положение точки и Солнца должны задаваться
    // в Гринвиче
    virtual double Density(const TimeJD& t, const Vect3& v, const Vect3& vSun)
    const;
    // Идентификатор модели
    virtual TAtmosphereId GetID() const
    {
        return atmIdStatic;
    };
    // Нужно ли для вычисления плотности положение Солнца?
    virtual bool IsSunPosNeed() const
    {
        return false;
    };
    // Нужна ли для вычисления плотности таблица атмосферных индексов?
    virtual bool IsTableNeed() const
    {
        return false;
    };
    // Получение коэффициентов статической атмосферы
    // Метод используется при расчете вариаций в численном интегрировании
    static void GetStaticAtmKoeff(double altitude, double* hi, double* k1,
                                  double* k2, double* a = NULL);
    // Расчет плотности в зависимости от высоты
    static double Density(double h);
};

///////////////////////////////////////////////////////////////////////////////
// Интерфейс динамических моделей атмосферы
///////////////////////////////////////////////////////////////////////////////
class IDynAtmosphere : public IAtmosphereModel
{
public:
    // Конструктор по умолчанию
    IDynAtmosphere(ITableGeoSolarFlux* g = NULL) : IAtmosphereModel()
    {
        Gsf = g;
    };
    // Нужно ли для вычисления плотности положение Солнца?
    virtual bool IsSunPosNeed() const
    {
        return true;
    };
    // Нужна ли для вычисления плотности таблица атмосферных индексов?
    virtual bool IsTableNeed() const
    {
        return true;
    };
    ITableGeoSolarFlux* Gsf;
};

///////////////////////////////////////////////////////////////////////////////
// Динамическая модель атмосферы по ГОСТ-77
///////////////////////////////////////////////////////////////////////////////
class Gost77Atmosphere : public IDynAtmosphere
{
public:
    // Конструктор по умолчанию
    Gost77Atmosphere(ITableGeoSolarFlux* g = NULL) : IDynAtmosphere(g)
    {
    };
    // Вычисление плотности. Положение точки и Солнца должны задаваться
    // в Гринвиче
    virtual double Density(const TimeJD& t, const Vect3& v, const Vect3& vSun)
    const;
    // Идентификатор модели
    virtual TAtmosphereId GetID() const
    {
        return atmIdDMA77;
    };
};

///////////////////////////////////////////////////////////////////////////////
// Динамическая модель атмосферы по ГОСТ-90
///////////////////////////////////////////////////////////////////////////////
class Gost90Atmosphere : public IDynAtmosphere
{
public:
    // Конструктор по умолчанию
    Gost90Atmosphere(ITableGeoSolarFlux* g = NULL, bool flag3h = false) : IDynAtmosphere(g)
    {
        is3HourInd = flag3h;
    };
    // Вычисление плотности. Положение точки и Солнца должны задаваться
    // в Гринвиче
    virtual double Density(const TimeJD& t, const Vect3& v, const Vect3& vSun)
    const;
    // Идентификатор модели
    virtual TAtmosphereId GetID() const
    {
        return atmIdDMA90;
    };

    bool is3HourInd;
};

///////////////////////////////////////////////////////////////////////////////
// Динамическая модель атмосферы по ГОСТ-2000
///////////////////////////////////////////////////////////////////////////////
class Gost2000Atmosphere : public IDynAtmosphere
{
public:
    // Конструктор по умолчанию
    Gost2000Atmosphere(ITableGeoSolarFlux* g = NULL, bool flag3h = false) : IDynAtmosphere(g)
    {
        is3HourInd = flag3h;
    };
    // Вычисление плотности. Положение точки и Солнца должны задаваться
    // в Гринвиче
    virtual double Density(const TimeJD& t, const Vect3& v, const Vect3& vSun)
    const;
    // Идентификатор модели
    virtual TAtmosphereId GetID() const
    {
        return atmIdDMA2000;
    };

    bool is3HourInd;
};

///////////////////////////////////////////////////////////////////////////////
// Динамическая модель атмосферы по ГОСТ-2004
///////////////////////////////////////////////////////////////////////////////
class Gost2004Atmosphere : public IDynAtmosphere
{
public:
    // Конструктор.
    // g - таблица индексов солнечной активности и геомагнитной возмущенности.
    // flag3h  - Признак использования трехчасовых индексов.
    // flagMod - Признак использования модификации ГОСТ-2004.
    Gost2004Atmosphere(ITableGeoSolarFlux* g = NULL, bool flag3h = false, bool flagMod = false)
        : IDynAtmosphere(g)
    {
        is3HourInd = flag3h;
        fmod = flagMod;
    };
    // Получить плотность атмосферы [кг/м^3]. Положение точки и
    // Солнца должны задаваться в Гринвиче.
    virtual double Density(const TimeJD& t, const Vect3& v, const Vect3& vSun) const;
    // Получить идентификатор модели.
    virtual TAtmosphereId GetID() const
    {
        return (fmod ? atmIdDMA2004m : atmIdDMA2004);
    };
    // Получить плотность ночной атмосферы [кг/м^3].
    // h   - высота над поверхностью Земного эллипсоида [тыс.км].
    // f81 - средневзвешенный индекс солнечной активности за 81 сутки.
    double NightDensity(double h, double f81) const;
    // Признак использования трехчасовых индексов геомагнитной возмущенности
    // true  - использовать трехчасовые индексы,
    // false - использовать среднесуточные индексы.
    bool is3HourInd;
    // Флаг использования модификации ГОСТ-2004, обеспечивающей
    // непрерывность зависимости плотности атмосферы от
    // средневзвешенного индекса солнечной активности.
    // true  - модифицировать ГОСТ-2004,
    // false - использовать "чистый" ГОСТ-2004.
    bool fmod;
};

///////////////////////////////////////////////////////////////////////////////
#endif
