///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Астробаллистические данные и алгоритмы, используемые в комплексе
///////////////////////////////////////////////////////////////////////////////
#include "Defines.h"
#include "BNOAstroBase.h"
#include "Consts.h"
#include <math.h>
#include <iostream>

///////////////////////////////////////////////////////////////////////////////
static PlanetConsts PConst[8] =
{
    {plIdMercury, "MERCURY", Ephemeride::idMercury, MercuryMeanRot_Vallado / 86.4, 2.4390, 0.0, 0.0},
    {plIdVenus, "VENUS", Ephemeride::idVenus, VenusMeanRot_Vallado / 86.4, 6.0520, 0.0, 0.0},
    {plIdMars, "MARS", Ephemeride::idMars, MarsMeanRot_Vallado / 86.4, 3.3972, 0.00647630, 0.0},
    {plIdJupiter, "JUPITER", Ephemeride::idJupiter, JupiterMeanRot_Vallado / 86.4, 71.4920, 0.0648744, 0.0},
    {plIdSaturn, "SATURN", Ephemeride::idSaturn, SaturnMeanRot_Vallado / 86.4, 60.2680, 0.0979624, 0.0},
    {plIdUranus, "URANUS", Ephemeride::idUranus, UranusMeanRot_Vallado / 86.4, 25.5590, 0.0229273, 0.0},
    {plIdNeptune, "NEPTUNE", Ephemeride::idNeptune, NeptuneMeanRot_Vallado / 86.4, 24.7640, 0.0171, 0.0},
    {plIdPluto, "PLUTO", Ephemeride::idPluto, PlutoMeanRot_Vallado / 86.4, 1.1510, 0.0, 0.0},
};
///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
BNOAstroBase::BNOAstroBase()
{
    int i;

    for (i = 0; i < 8; i++)
        Planets[i].Init(PConst[i]);

    char* name = getenv("BNO_CFG");
    if (name == NULL)
        CfgFName = "/mnt/server/BNO/FTABLES/FTables.cfg";
    else
        CfgFName = name;

    time_t today = time(NULL);
    time_t ts = today - 366 * 86400 * 5;

    TimeGD gs = TimeGD(ts);
    int d;
    double f;
    convertTimeGJ(gs.Year, gs.Month, gs.Day, gs.Hour, gs.Min, gs.Sec,
                  gs.Fraction, d, f);
    StartTime = TimeJD(d, f);
    time_t te = today + 366 * 86400 * 5;
    gs = TimeGD(te);
    convertTimeGJ(gs.Year, gs.Month, gs.Day, gs.Hour, gs.Min, gs.Sec,
                  gs.Fraction, d, f);
    EndTime = TimeJD(d, f);
}

//-----------------------------------------------------------------------------
// Получение указателя на модель планеты
IPlanet* BNOAstroBase::GetPlanet(TPlanetId pl)
{
    switch (pl) {
    case plIdEarth:
        return &Earth;
    case plIdMoon:
        return &Moon;
    case plIdSun:
        return &Sun;
    default:
        int i;
        for (i = 0; i < 8; i++) {
            if (Planets[i].Id() == pl)
                return Planets + i;
        }
        TAssert(false);
        return NULL;
    }
}

const IPlanet* BNOAstroBase::GetPlanet(TPlanetId pl) const
{
    switch (pl) {
    case plIdEarth:
        return &Earth;
    case plIdMoon:
        return &Moon;
    case plIdSun:
        return &Sun;
    default:
        int i;
        for (i = 0; i < 8; i++) {
            if (Planets[i].Id() == pl)
                return Planets + i;
        }
        TAssert(false);
        return NULL;
    }
}

//-----------------------------------------------------------------------------
// Преобразование времени между шкалами
void BNOAstroBase::ConvertTimeScaleJD(TmScale src, TmScale dst, int& d,
                                      double& f) const
{
    long double d1;
    const struct TRecTableTAI2UTC* block;
    double t1, dt;
    int i;

    if (src == dst)
        return;

    if (dst > src) {
        // enum TmScale { tmScaleTDB, tmScaleTDT, tmScaleTAI, tmScaleUTC, tmScaleUT1 };
        if (src <= tmScaleTDB && dst > tmScaleTDB) {
            // TDB -> TDT, TM04
            double ME = 357.5277233 * deg2rad + 35999.05034 * deg2rad *
                ((d + f - 2451545.0l) / 36525.0l);
            f -= ((long double)(0.001658 * sin(ME + 0.016708617 * sin(ME)))) / 86400.0l;
            int dd = (int)floor(f);
            d += dd;
            f -= dd;
        }

        if (src <= tmScaleTDT && dst > tmScaleTDT) {
            // TDT -> TAI, TM07a
            f -= 32.184l * sec2day;
            int dd = (int)floor(f);
            d += dd;
            f -= dd;
        }

        if (src <= tmScaleTAI && dst > tmScaleTAI) {
            // TAI -> UTC, TM06
            TAssert(TableTAI2UTC.isInit());
            i = TableTAI2UTC.Len() - 1;
            block = TableTAI2UTC.getBlock(i);
            for (; i > 0; i--) {
                if (d + f >= block->JD_TAI)
                    break;
                block--;
            };
            f -= (block->B + (d + f - modJD - block->MJD) * block->S) * sec2day;
            int dd = (int)floor(f);
            d += dd;
            f -= dd;
        }

        if (src <= tmScaleUTC && dst > tmScaleUTC) {
            // UTC -> UT1, TM08
            TAssert(TableUTC2UT1.isInit());
            d1 = d + f - TableUTC2UT1.JDBeg() - modJD;
            i = int(d1);
            if (i < 0)
                f += TableUTC2UT1.dUT1UTC(0) * sec2day;
            else if (i >= TableUTC2UT1.Len() - 1)
                f += TableUTC2UT1.dUT1UTC(TableUTC2UT1.Len() - 1) * sec2day;
            else {
                t1 = TableUTC2UT1.dUT1UTC(i);
                dt = TableUTC2UT1.dUT1UTC(i + 1) - t1;
                if (dt > 0.5) {
                    t1 += 1.0;
                    dt -= 1.0;
                }
                else if (dt < -0.5) {
                    t1 -= 1.0;
                    dt += 1.0;
                };
                f += (t1 + dt * (d1 - i)) * sec2day;
            }
            int dd = (int)floor(f);
            d += dd;
            f -= dd;
        }

    }
    else {
        if (src >= tmScaleUT1 && dst < tmScaleUT1) {
            TAssert(TableUTC2UT1.isInit());
            d1 = d + f - TableUTC2UT1.JDBeg() - modJD;
            i = int(d1);
            if (i < 0)
                f -= TableUTC2UT1.dUT1UTC(0) * sec2day;
            else if (i >= TableUTC2UT1.Len() - 1)
                f -= TableUTC2UT1.dUT1UTC(TableUTC2UT1.Len() - 1) * sec2day;
            else {
                t1 = TableUTC2UT1.dUT1UTC(i);
                dt = TableUTC2UT1.dUT1UTC(i + 1) - t1;
                if (dt > 0.5) {
                    t1 += 1.0;
                    dt -= 1.0;
                }
                else if (dt < -0.5) {
                    t1 -= 1.0;
                    dt += 1.0;
                }
                f -= (t1 + dt * (d1 - i)) * sec2day;
            }
            int dd = (int)floor(f);
            d += dd;
            f -= dd;
        }

        if (src >= tmScaleUTC && dst < tmScaleUTC) {
            // UTC -> TAI, TM05
            TAssert(TableTAI2UTC.isInit());
            i = TableTAI2UTC.Len() - 1;
            block = TableTAI2UTC.getBlock(i);
            for (; i > 0; i--) {
                if (d + f >= block->JD)
                    break;
                block--;
            }
            f += (block->B + (d + f - modJD - block->MJD) * block->S) * sec2day;
            int dd = (int)floor(f);
            d += dd;
            f -= dd;
        }

        if (src >= tmScaleTAI && dst < tmScaleTAI) {
            // TAI -> TDT, TM07
            f += 32.184l * sec2day;
            int dd = (int)floor(f);
            d += dd;
            f -= dd;
        }

        if (src >= tmScaleTDT && dst < tmScaleTDT) {
            // TDT -> TDB, TM04
            double ME = 357.5277233 * deg2rad + 35999.05034 * deg2rad * ((d + f - 2451545.0l)
                / 36525.0l);
            f += ((long double)(0.001658 * sin(ME + 0.016708617 * sin(ME)))) / 86400.0l;
            int dd = (int)floor(f);
            d += dd;
            f -= dd;
        }
    }
}

//-----------------------------------------------------------------------------
// Вычисление матрицы, учитывающей прецессию оси вращения Земли
double BNOAstroBase::MtxPrecession(const TimeJD& dtStart, const TimeJD& dtEnd,
                                   TransMtx& mtx) const
{
    const double toRad = (Pi / (180 * 3600));

    double t, t0;
    double dzeta, zet, teta, cos_dzeta, cos_zet, cos_teta;
    double Zp;

    TimeJD tdbe(dtEnd), tdbs(dtStart);
    double dtEnd_ = tdbe.GetDays() + tdbe.GetFraction();
    double dtStart_ = tdbs.GetDays() + tdbs.GetFraction();

    t = (dtEnd_ - dtStart_);
    t /= 36525.0;

    if (dtStart_ == doubleJ2000) {
        dzeta = t * (1.1180860865024400e-02 + t * (1.4635555405334672e-06 + t * 8.7256766326094288e-08));
        zet = t * (1.1180860865024400e-02 + t * (5.3071584043698689e-06 + t * 8.8250634372368836e-08));
        teta = t * (9.7171734551696701e-03 - t * (2.0684575704538350e-06 + t * 2.0281210721855220e-07));
    }
    else {
        t0 = (dtStart_ - doubleJ2000) / 36525.0;
        teta = (2306.2181 * toRad) + t0 * ((1.39656 * toRad) - t0 * (0.000139 * toRad));
        dzeta = t * (teta + t * (((0.30188 * toRad) -
            t0 * (0.000344 * toRad)) + t * (0.017998 * toRad)));
        zet = t * (teta + t * (((1.09468 * toRad) + t0 * (0.000066 * toRad)) + t * (0.018203 * toRad)));
        teta = t * (((2004.3109 * toRad) - t0 * ((0.85330 * toRad) + t0 * (0.000217 * toRad))) +
            t * (((-0.42665 * toRad) - t0 * (0.000217 * toRad)) - t * (0.041833 * toRad)));
    };

    Zp = dzeta + zet;

    if (&mtx != NULL) {
        cos_dzeta = cos(dzeta);
        dzeta = sin(dzeta);
        cos_zet = cos(zet);
        zet = sin(zet);
        cos_teta = cos(teta);
        teta = sin(teta);
        mtx[0] = cos_dzeta * cos_zet * cos_teta - dzeta * zet;
        mtx[1] = -dzeta * cos_zet * cos_teta - cos_dzeta * zet;
        mtx[2] = -cos_zet * teta;
        mtx[3] = cos_dzeta * zet * cos_teta + dzeta * cos_zet;
        mtx[4] = -dzeta * zet * cos_teta + cos_dzeta * cos_zet;
        mtx[5] = -zet * teta;
        mtx[6] = cos_dzeta * teta;
        mtx[7] = -dzeta * teta;
        mtx[8] = cos_teta;
    };

    return Zp;
}

//-----------------------------------------------------------------------------
// Данные для расчета нутации
static const double toRadN = (Pi * 1e-4) / (180 * 3600);

static double nutacKoeff[][9] = {
    {0.0, 0.0, 0.0, 0.0, 1.0, -171996.0 * toRadN, -174.2 * toRadN, 92025.0 * toRadN, 8.9 * toRadN},
    {0.0, 0.0, 2.0, -2.0, 2.0, -13187.0 * toRadN, -1.6 * toRadN, 5736.0 * toRadN, -3.1 * toRadN},
    {0.0, 0.0, 2.0, 0.0, 2.0, -2274.0 * toRadN, -0.2 * toRadN, 977.0 * toRadN, -0.5 * toRadN},
    {0.0, 0.0, 0.0, 0.0, 2.0, 2062.0 * toRadN, 0.2 * toRadN, -895.0 * toRadN, 0.5 * toRadN},
    {0.0, 1.0, 0.0, 0.0, 0.0, 1426.0 * toRadN, -3.4 * toRadN, 54.0 * toRadN, -0.1 * toRadN},
    {1.0, 0.0, 0.0, 0.0, 0.0, 712.0 * toRadN, 0.1 * toRadN, -7.0 * toRadN, 0.0 * toRadN},
    {0.0, 1.0, 2.0, -2.0, 2.0, -517.0 * toRadN, 1.2 * toRadN, 224.0 * toRadN, -0.6 * toRadN},
    {0.0, 0.0, 2.0, 0.0, 1.0, -386.0 * toRadN, -0.4 * toRadN, 200.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, 2.0, 0.0, 2.0, -301.0 * toRadN, 0.0 * toRadN, 129.0 * toRadN, -0.1 * toRadN},
    {0.0, -1.0, 2.0, -2.0, 2.0, 217.0 * toRadN, -0.5 * toRadN, -95.0 * toRadN, 0.3 * toRadN},
    {1.0, 0.0, 0.0, -2.0, 0.0, -158.0 * toRadN, 0.0 * toRadN, -1.0 * toRadN, 0.0 * toRadN},
    {0.0, 0.0, 2.0, -2.0, 1.0, 129.0 * toRadN, 0.1 * toRadN, -70.0 * toRadN, 0.0 * toRadN},
    {-1.0, 0.0, 2.0, 0.0, 2.0, 123.0 * toRadN, 0.0 * toRadN, -53.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, 0.0, 0.0, 1.0, 63.0 * toRadN, 0.1 * toRadN, -33.0 * toRadN, 0.0 * toRadN},
    {0.0, 0.0, 0.0, 2.0, 0.0, 63.0 * toRadN, 0.0 * toRadN, -2.0 * toRadN, 0.0 * toRadN},
    {-1.0, 0.0, 2.0, 2.0, 2.0, -59.0 * toRadN, 0.0 * toRadN, 26.0 * toRadN, 0.0 * toRadN},
    {-1.0, 0.0, 0.0, 0.0, 1.0, -58.0 * toRadN, -0.1 * toRadN, 32.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, 2.0, 0.0, 1.0, -51.0 * toRadN, 0.0 * toRadN, 27.0 * toRadN, 0.0 * toRadN},
    {2.0, 0.0, 0.0, -2.0, 0.0, 48.0 * toRadN, 0.0 * toRadN, 1.0 * toRadN, 0.0 * toRadN},
    {-2.0, 0.0, 2.0, 0.0, 1.0, 46.0 * toRadN, 0.0 * toRadN, -24.0 * toRadN, 0.0 * toRadN},
    {0.0, 0.0, 2.0, 2.0, 2.0, -38.0 * toRadN, 0.0 * toRadN, 16.0 * toRadN, 0.0 * toRadN},
    {2.0, 0.0, 2.0, 0.0, 2.0, -31.0 * toRadN, 0.0 * toRadN, 13.0 * toRadN, 0.0 * toRadN},
    {2.0, 0.0, 0.0, 0.0, 0.0, 29.0 * toRadN, 0.0 * toRadN, -1.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, 2.0, -2.0, 2.0, 29.0 * toRadN, 0.0 * toRadN, -12.0 * toRadN, 0.0 * toRadN},
    {0.0, 0.0, 2.0, 0.0, 0.0, 26.0 * toRadN, 0.0 * toRadN, -1.0 * toRadN, 0.0 * toRadN},
    {0.0, 0.0, 2.0, -2.0, 0.0, -22.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {-1.0, 0.0, 2.0, 0.0, 1.0, 21.0 * toRadN, 0.0 * toRadN, -10.0 * toRadN, 0.0 * toRadN},
    {0.0, 2.0, 0.0, 0.0, 0.0, 17.0 * toRadN, -0.1 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {0.0, 2.0, 2.0, -2.0, 2.0, -16.0 * toRadN, 0.1 * toRadN, 7.0 * toRadN, 0.0 * toRadN},
    {-1.0, 0.0, 0.0, 2.0, 1.0, 16.0 * toRadN, 0.0 * toRadN, -8.0 * toRadN, 0.0 * toRadN},
    {0.0, 1.0, 0.0, 0.0, 1.0, -15.0 * toRadN, 0.0 * toRadN, 9.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, 0.0, -2.0, 1.0, -13.0 * toRadN, 0.0 * toRadN, 7.0 * toRadN, 0.0 * toRadN},
    {0.0, -1.0, 0.0, 0.0, 1.0, -12.0 * toRadN, 0.0 * toRadN, 6.0 * toRadN, 0.0 * toRadN},
    {2.0, 0.0, -2.0, 0.0, 0.0, 11.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {-1.0, 0.0, 2.0, 2.0, 1.0, -10.0 * toRadN, 0.0 * toRadN, 5.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, 2.0, 2.0, 2.0, -8.0 * toRadN, 0.0 * toRadN, 3.0 * toRadN, 0.0 * toRadN},
    {0.0, -1.0, 2.0, 0.0, 2.0, -7.0 * toRadN, 0.0 * toRadN, 3.0 * toRadN, 0.0 * toRadN},
    {0.0, 0.0, 2.0, 2.0, 1.0, -7.0 * toRadN, 0.0 * toRadN, 3.0 * toRadN, 0.0 * toRadN},
    {1.0, 1.0, 0.0, -2.0, 0.0, -7.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {0.0, 1.0, 2.0, 0.0, 2.0, 7.0 * toRadN, 0.0 * toRadN, -3.0 * toRadN, 0.0 * toRadN},
    {-2.0, 0.0, 0.0, 2.0, 1.0, -6.0 * toRadN, 0.0 * toRadN, 3.0 * toRadN, 0.0 * toRadN},
    {0.0, 0.0, 0.0, 2.0, 1.0, -6.0 * toRadN, 0.0 * toRadN, 3.0 * toRadN, 0.0 * toRadN},
    {2.0, 0.0, 2.0, -2.0, 2.0, 6.0 * toRadN, 0.0 * toRadN, -3.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, 0.0, 2.0, 0.0, 6.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, 2.0, -2.0, 1.0, 6.0 * toRadN, 0.0 * toRadN, -3.0 * toRadN, 0.0 * toRadN},
    {0.0, 0.0, 0.0, -2.0, 1.0, -5.0 * toRadN, 0.0 * toRadN, 3.0 * toRadN, 0.0 * toRadN},
    {0.0, -1.0, 2.0, -2.0, 1.0, -5.0 * toRadN, 0.0 * toRadN, 3.0 * toRadN, 0.0 * toRadN},
    {2.0, 0.0, 2.0, 0.0, 1.0, -5.0 * toRadN, 0.0 * toRadN, 3.0 * toRadN, 0.0 * toRadN},
    {1.0, -1.0, 0.0, 0.0, 0.0, 5.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, 0.0, -1.0, 0.0, -4.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {0.0, 0.0, 0.0, 1.0, 0.0, -4.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {0.0, 1.0, 0.0, -2.0, 0.0, -4.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, -2.0, 0.0, 0.0, 4.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {2.0, 0.0, 0.0, -2.0, 1.0, 4.0 * toRadN, 0.0 * toRadN, -2.0 * toRadN, 0.0 * toRadN},
    {0.0, 1.0, 2.0, -2.0, 1.0, 4.0 * toRadN, 0.0 * toRadN, -2.0 * toRadN, 0.0 * toRadN},
    {1.0, 1.0, 0.0, 0.0, 0.0, -3.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {1.0, -1.0, 0.0, -1.0, 0.0, -3.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {-1.0, -1.0, 2.0, 2.0, 2.0, -3.0 * toRadN, 0.0 * toRadN, 1.0 * toRadN, 0.0 * toRadN},
    {0.0, -1.0, 2.0, 2.0, 2.0, -3.0 * toRadN, 0.0 * toRadN, 1.0 * toRadN, 0.0 * toRadN},
    {1.0, -1.0, 2.0, 0.0, 2.0, -3.0 * toRadN, 0.0 * toRadN, 1.0 * toRadN, 0.0 * toRadN},
    {3.0, 0.0, 2.0, 0.0, 2.0, -3.0 * toRadN, 0.0 * toRadN, 1.0 * toRadN, 0.0 * toRadN},
    {-2.0, 0.0, 2.0, 0.0, 2.0, -3.0 * toRadN, 0.0 * toRadN, 1.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, 2.0, 0.0, 0.0, 3.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {-1.0, 0.0, 2.0, 4.0, 2.0, -2.0 * toRadN, 0.0 * toRadN, 1.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, 0.0, 0.0, 2.0, -2.0 * toRadN, 0.0 * toRadN, 1.0 * toRadN, 0.0 * toRadN},
    {-1.0, 0.0, 2.0, -2.0, 1.0, -2.0 * toRadN, 0.0 * toRadN, 1.0 * toRadN, 0.0 * toRadN},
    {0.0, -2.0, 2.0, -2.0, 1.0, -2.0 * toRadN, 0.0 * toRadN, 1.0 * toRadN, 0.0 * toRadN},
    {-2.0, 0.0, 0.0, 0.0, 1.0, -2.0 * toRadN, 0.0 * toRadN, 1.0 * toRadN, 0.0 * toRadN},
    {2.0, 0.0, 0.0, 0.0, 1.0, 2.0 * toRadN, 0.0 * toRadN, -1.0 * toRadN, 0.0 * toRadN},
    {3.0, 0.0, 0.0, 0.0, 0.0, 2.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {1.0, 1.0, 2.0, 0.0, 2.0, 2.0 * toRadN, 0.0 * toRadN, -1.0 * toRadN, 0.0 * toRadN},
    {0.0, 0.0, 2.0, 1.0, 2.0, 2.0 * toRadN, 0.0 * toRadN, -1.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, 0.0, 2.0, 1.0, -1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, 2.0, 2.0, 1.0, -1.0 * toRadN, 0.0 * toRadN, 1.0 * toRadN, 0.0 * toRadN},
    {1.0, 1.0, 0.0, -2.0, 1.0, -1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {0.0, 1.0, 0.0, 2.0, 0.0, -1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {0.0, 1.0, 2.0, -2.0, 0.0, -1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {0.0, 1.0, -2.0, 2.0, 0.0, -1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, -2.0, 2.0, 0.0, -1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, -2.0, -2.0, 0.0, -1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, 2.0, -2.0, 0.0, -1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {1.0, 0.0, 0.0, -4.0, 0.0, -1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {2.0, 0.0, 0.0, -4.0, 0.0, -1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {0.0, 0.0, 2.0, 4.0, 2.0, -1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {0.0, 0.0, 2.0, -1.0, 2.0, -1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {-2.0, 0.0, 2.0, 4.0, 2.0, -1.0 * toRadN, 0.0 * toRadN, 1.0 * toRadN, 0.0 * toRadN},
    {2.0, 0.0, 2.0, 2.0, 2.0, -1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {0.0, -1.0, 2.0, 0.0, 1.0, -1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {0.0, 0.0, -2.0, 0.0, 1.0, -1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {0.0, 0.0, 4.0, -2.0, 2.0, 1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {0.0, 1.0, 0.0, 0.0, 2.0, 1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {1.0, 1.0, 2.0, -2.0, 2.0, 1.0 * toRadN, 0.0 * toRadN, -1.0 * toRadN, 0.0 * toRadN},
    {3.0, 0.0, 2.0, -2.0, 2.0, 1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {-2.0, 0.0, 2.0, 2.0, 2.0, 1.0 * toRadN, 0.0 * toRadN, -1.0 * toRadN, 0.0 * toRadN},
    {-1.0, 0.0, 0.0, 0.0, 2.0, 1.0 * toRadN, 0.0 * toRadN, -1.0 * toRadN, 0.0 * toRadN},
    {0.0, 0.0, -2.0, 2.0, 1.0, 1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {0.0, 1.0, 2.0, 0.0, 1.0, 1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {-1.0, 0.0, 4.0, 0.0, 2.0, 1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {2.0, 1.0, 0.0, -2.0, 0.0, 1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {2.0, 0.0, 0.0, 2.0, 0.0, 1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {2.0, 0.0, 2.0, -2.0, 1.0, 1.0 * toRadN, 0.0 * toRadN, -1.0 * toRadN, 0.0 * toRadN},
    {2.0, 0.0, -2.0, 0.0, 1.0, 1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {1.0, -1.0, 0.0, -2.0, 0.0, 1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {-1.0, 0.0, 0.0, 1.0, 1.0, 1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {-1.0, -1.0, 0.0, 2.0, 1.0, 1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN},
    {0.0, 1.0, 0.0, 1.0, 0.0, 1.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN, 0.0 * toRadN}
};
//-----------------------------------------------------------------------------
// Вычисление матрицы, учитывающей нутации оси вращения Земли
double BNOAstroBase::MtxNutation(const TimeJD& dt, TransMtx& mtx) const
{
    double t;
    int i;
    double Ml, Ms, uL, Ds, OL;
    double dPsi, dEps, eps0, t1;
    double* line;

    TimeJD tdb(dt);
    t = (tdb.GetDays() - doubleJ2000 + tdb.GetFraction()) / 36525.0;

    Ml = 2.355548394 + t * ((1325 * Pi2 + 3.470890873) + t * (1.517952e-4 + t * 3.103e-7));
    Ms = 6.24003594 + t * ((99 * Pi2 + 6.2666106) - t * (2.7974e-6 + t * 5.82e-8));
    uL = 1.627901934 + t * ((1342 * Pi2 + 1.431476084) - t * (6.42717e-5 - t * 5.34e-8));
    Ds = 5.198469514 + t * ((1236 * Pi2 + 5.36010650) - t * (3.34086e-5 - t * 9.22e-8));
    OL = 2.182438624 - t * ((5 * Pi2 + 2.341119397) - t * (3.61429e-5 + t * 3.88e-8));

    dPsi = 0;
    dEps = 0;
    for (i = 0; i < 106; i++) {
        line = nutacKoeff[i];
        t1 = line[0] * Ml + line[1] * Ms + line[2] * uL + line[3] * Ds + line[4] * OL;
        dPsi += (line[5] + line[6] * t) * sin(t1);
        dEps += (line[7] + line[8] * t) * cos(t1);
    };

    eps0 = 0.4090928042 - t * (0.2269655e-3 + t * (2.86e-9 - t * 8.80e-9));
    dEps += eps0;

    if (&mtx != NULL) {
        Ml = cos(dPsi);
        Ms = sin(dPsi);
        uL = cos(dEps);
        Ds = sin(dEps);
        OL = cos(eps0);
        t1 = sin(eps0);

        mtx[0] = Ml;
        mtx[1] = -Ms * OL;
        mtx[2] = -Ms * t1;
        mtx[3] = Ms * uL;
        mtx[4] = Ml * uL * OL + Ds * t1;
        mtx[5] = Ml * uL * t1 - Ds * OL;
        mtx[6] = Ms * Ds;
        mtx[7] = Ml * Ds * OL - uL * t1;
        mtx[8] = Ml * Ds * t1 + uL * OL;
    }
    else
        uL = cos(dEps);

    return dPsi * uL;
}

//-----------------------------------------------------------------------------
// Вычисление матрицы перехода от J2000 к METE
void BNOAstroBase::MtxJ20002METE(const TimeJD& jdMETE, TransMtx& mtx) const
{
    TransMtx cur1;
    double t, Na;
    MtxPrecession(TimeJD(0, doubleJ2000), jdMETE, mtx);
    Na = MtxNutation(jdMETE, cur1);
    cur1 *= mtx;
    t = cos(Na);
    mtx[0] = t;
    mtx[4] = t;
    t = sin(Na);
    mtx[1] = t;
    mtx[3] = -t;
    mtx[2] = 0.0;
    mtx[5] = 0.0;
    mtx[6] = 0.0;
    mtx[7] = 0.0;
    mtx[8] = 1.0;
    mtx *= cur1;
}

//-----------------------------------------------------------------------------
// Вычисление истинного звездного времени
double BNOAstroBase::SaTime(const TimeJD& jd) const
{
    TransMtx* p = NULL;
    return Smean(jd) + MtxNutation(jd, *p);
}

//-----------------------------------------------------------------------------
// Вычисление истинного звездного времени при известной поправке
double BNOAstroBase::SaTime(const TimeJD& jd, double ds) const
{
    return Smean(jd) + ds;
}

//-----------------------------------------------------------------------------
// Вычисление среднего звездного времени
double BNOAstroBase::Smean(const TimeJD& dt) const
{
    double d, j, T, f = dt.GetFraction();
    int jd = dt.GetDays();

    GetAstroBase()->ConvertTimeScaleJD(tmScaleTDB, tmScaleUT1, jd, f);
    d = jd - doubleJ2000 + f;
    j = floor(d);
    T = (j - 0.5) / 36525.0;

    d = (1.7533685592 + T * (628.3319706889 + T * (6.7707139e-6 - T * 4.50876e-10))) +
        EarthMeanRot * (d - j + 0.5);
    d = fmod(d, Pi2);
    if (d < 0)
        d += Pi2;
    return d;
}

//-----------------------------------------------------------------------------
// Функция вычисляет матрицу, учитывающую смещение положения мгновенного 
// полюса Земли относительно международного условного начала
void BNOAstroBase::MtxPolus(double x, double y, TransMtx& mtx) const
{
    double cx, sx, cy, sy;

    cx = cos(x);
    sx = sin(x);
    cy = cos(y);
    sy = sin(y);
    mtx[0] = cx;
    mtx[1] = sx * sy;
    mtx[2] = sx * cy;
    mtx[3] = 0.0;
    mtx[4] = cy;
    mtx[5] = -sy;
    mtx[6] = -sx;
    mtx[7] = cx * sy;
    mtx[8] = cx * cy;
}

//-----------------------------------------------------------------------------
// Функция вычисляет матрицу, учитывающую смещение положения мгновенного 
// полюса Земли относительно международного условного начала
void BNOAstroBase::MtxPolus(const TimeJD& t, TransMtx& mtx) const
{
    double dl, t1;
    double x, y;
    int ind;

    TAssert(TablePolus.isInit());
    ind = TablePolus.Len();
    if (ind == 1) {
        x = TablePolus.X(0);
        y = TablePolus.Y(0);
    }
    else {
        int d = t.GetDays();
        double f = t.GetFraction();
        GetAstroBase()->ConvertTimeScaleJD(tmScaleTDB, tmScaleUTC, d, f);
        dl = d - TablePolus.JDBeg() - modJD + f;
        ind = int(floor(dl));
        if (ind < 0)
            ind = 0;
        if (ind >= TablePolus.Len() - 1)
            ind = TablePolus.Len() - 2;
        dl = fmod(dl, 1.0);
        x = TablePolus.X(ind);
        t1 = TablePolus.X(ind + 1);
        x += (t1 - x) * dl;
        y = TablePolus.Y(ind);
        t1 = TablePolus.Y(ind + 1);
        y += (t1 - y) * dl;
    };
    x *= sec2rad;
    y *= sec2rad;
    MtxPolus(x, y, mtx);
}

//-----------------------------------------------------------------------------
// Вычисление матрицы перехода от ECI к фиксированному Гринвичу
void BNOAstroBase::MtxECI2GFCS(const TimeJD& jdECI, const TimeJD& jdGFCS,
                               TransMtx& mtx) const
{
    TransMtx cur1, cur2;
    double dl;

    MtxPrecession(jdECI, jdGFCS, cur1);

    dl = MtxNutation(jdGFCS, cur2);
    cur2 *= cur1;

    MtxRotation(SaTime(jdGFCS, dl), cur1);

    MtxPolus(jdGFCS, mtx);

    mtx *= cur1;
    mtx *= cur2;
}

//-----------------------------------------------------------------------------
// Получение матрицы перехода от METE к Гринвичу
void BNOAstroBase::MtxMETE2GCS(const TimeJD& t, TransMtx& mtx) const
{
    TransMtx mtx1;
    MtxRotation(Smean(t), mtx1);
    MtxPolus(t, mtx);
    mtx.MulRight(mtx1);
}

//-----------------------------------------------------------------------------
// Получение положений планет и тел Солнечной системы
void BNOAstroBase::GetPosGelio(TPlanetId pl, const TimeJD& t, Vect3* crd, Vect3* vel,
                               Vect3* acc) const
{
    if (pl == plIdEarth || pl == plIdMoon) {
        Vect3 cem, vem, aem;
        Vect3 cm, vm, am;
        Vect3 cs, vs, as;
        GetPosGelio(plIdEarthMoon, t, crd == NULL ? NULL : &cem, vel == NULL ? NULL : &vem,
                    acc == NULL ? NULL : &aem);
        GetPosGeo(plIdMoon, t, crd == NULL ? NULL : &cm, vel == NULL ? NULL : &vm,
                  acc == NULL ? NULL : &am);
        if (crd != NULL) {
            GetGelio2GeoShift(cem, cm, cs);
            if (pl == plIdMoon)
                *crd = cm;
            else
                *crd = 0.0;
            (*crd) -= cs;
        }
        if (vel != NULL) {
            GetGelio2GeoShift(vem, vm, vs);
            if (pl == plIdMoon)
                *vel = vm;
            else
                *vel = 0.0;
            (*vel) -= vs;
        }
        if (acc != NULL) {
            GetGelio2GeoShift(aem, am, as);
            if (pl == plIdMoon)
                *acc = am;
            else
                *acc = 0.0;
            (*acc) -= as;
        }
    }
    else {
        const IEphemeride* iph;
        if (pl == plIdEarthMoon)
            iph = &EarthMoon;
        else
            iph = GetPlanet(pl)->GetEphemeride();
        TAssert(iph->GetCS() == csIdJ2000S);
        iph->GetPosition(t, crd, vel, acc);
    }
}

//-----------------------------------------------------------------------------
// Получение положений планет и тел Солнечной системы
void BNOAstroBase::GetPosGeo(TPlanetId pl, const TimeJD& t, Vect3* crd, Vect3* vel,
                             Vect3* acc) const
{
    if (pl == plIdEarth || pl == plIdMoon) {
        const IEphemeride* iph = GetPlanet(pl)->GetEphemeride();
        TAssert(iph->GetCS() == csIdJ2000);
        iph->GetPosition(t, crd, vel, acc);
    }
    else {
        Vect3 cem, vem, aem;
        Vect3 cm, vm, am;
        Vect3 cs, vs, as;
        GetPosGelio(pl, t, crd, vel, acc);
        GetPosGelio(plIdEarthMoon, t, crd == NULL ? NULL : &cem, vel == NULL ? NULL : &vem,
                    acc == NULL ? NULL : &aem);
        GetPosGeo(plIdMoon, t, crd == NULL ? NULL : &cm, vel == NULL ? NULL : &vm,
                  acc == NULL ? NULL : &am);
        if (crd != NULL) {
            GetGelio2GeoShift(cem, cm, cs);
            (*crd) += cs;
        }
        if (vel != NULL) {
            GetGelio2GeoShift(vem, vm, vs);
            (*vel) += vs;
        }
        if (acc != NULL) {
            GetGelio2GeoShift(aem, am, as);
            (*acc) += as;
        }
    }
}

//-----------------------------------------------------------------------------
// Инициализация базы
bool BNOAstroBase::Init(bool noAtm, bool loadPlanets)
{
    if (!Config.Load(CfgFName)) {
        std::cerr << "failed to load cfg";
        return false;
    }

    if (!LoadTAITable(Config.GetValue("tableTAI2UTC"))) {
        std::cerr << "failed to tableTAI2UTC";
        return false;
    }

    SetAstroBase(*this);

    if (!LoadUT1Table(Config.GetValue("tableUTC2UT1"))) {
        std::cerr << "failed to tableUTC2UT1";
        return false;
    }

    if (!LoadPolusTable(Config.GetValue("tablePolus"))) {
        std::cerr << "failed to tablePolus";
        return false;
    }

    if (!noAtm) {
        if (!LoadGeoSolarFlux(Config.GetValue("tableEarthAtm"))) {
            std::cerr << "failed to tableEarthAtm";
            return false;
        }
    }

    TString eph_fname = Config.GetValue("tableEPH");

    EarthMoon.Init(eph_fname, Ephemeride::idEarthMoon);
    if (!EarthMoon.LoadData(StartTime, EndTime)) {
        std::cerr << "failed to tableEPH";
        return false;
    }

    MoonEphemeride* me = (MoonEphemeride *)Moon.GetEphemeride();
    me->Init(eph_fname);

    if (!me->LoadData(StartTime, EndTime)) {
        std::cerr << "failed to moon";
        return false;
    }

    for (int i = 0; i < 8; i++) {
        PlanetEphemeride* ie = (PlanetEphemeride *)Planets[i].GetEphemeride();
        ie->Init(eph_fname, PConst[i].PlanetEphId);
        if (loadPlanets) {
            ie->LoadData(StartTime, EndTime);
        }
    }

    if (!noAtm) {
        Earth.SetAtmModel(atmIdDMA2004, &TableGeoSolarFlux);
    }

    return true;
}

//-----------------------------------------------------------------------------
// Инициализация базы с явным указанием времени для эфемерид
bool BNOAstroBase::Init(const TimeGD& s, const TimeGD& e, bool noAtm, bool loadPlanets)
{
    int d;
    double f;
    convertTimeGJ(s.Year, s.Month, s.Day, s.Hour, s.Min, s.Sec,
                  s.Fraction, d, f);
    StartTime = TimeJD(d, f);
    convertTimeGJ(e.Year, e.Month, e.Day, e.Hour, e.Min, e.Sec,
                  e.Fraction, d, f);
    EndTime = TimeJD(d, f);
    return Init(noAtm, loadPlanets);
}

//-----------------------------------------------------------------------------
// Инициализация базы с явным указанием времени для эфемерид
bool BNOAstroBase::Init(const TimeJD& s, const TimeJD& e, bool noAtm, bool loadPlanets)
{
    StartTime = s;
    EndTime = e;
    return Init(noAtm, loadPlanets);
}

//-----------------------------------------------------------------------------
// Загрузка в память данных по эфемеридам
bool BNOAstroBase::LoadEphData(TPlanetId pl)
{
    if (pl == plIdEarthMoon)
        return EarthMoon.LoadData(StartTime, EndTime);
    else {
        IEphemeride* ie = GetPlanet(pl)->GetEphemeride();
        return ie->LoadData(StartTime, EndTime);
    }
}

//-----------------------------------------------------------------------------
// Выгрузка из памяти данных по эфемеридам
void BNOAstroBase::UnLoadEphData(TPlanetId pl)
{
    if (pl == plIdEarthMoon)
        EarthMoon.UnLoadData();
    else {
        IEphemeride* ie = GetPlanet(pl)->GetEphemeride();
        ie->UnLoadData();
    }
}

//-----------------------------------------------------------------------------
// Структура соответствия гравитационных моделей и ключей в файле конфигурации
class OneGravityModel
{
public:
    const char* Key;
    TGravityModelId ID;
};

// Таблица соотвествия ключей моделям гравитации
static OneGravityModel GModels[] =
{
    {"tableEGM96", grIdEGM96},
    {"tableJGM3", grIdJGM3},
    {"tableWGS72", grIdWGS72},
    {"tableWGS84", grIdWGS84},
    {"tablePZ902", grIdPZ902f}
};
// Установка модели гравитации для Земли
bool BNOAstroBase::SetEarthGravityModel(TGravityModelId mg)
{
    if (mg == grIdPZ90 || mg == grIdPZ902s) {
        Earth.SetGravityModel(mg);
        return true;
    }

    unsigned i;

    for (i = 0; i < sizeof(GModels) / sizeof(*GModels); i++) {
        if (GModels[i].ID == mg)
            break;
    }

    if (i == sizeof(GModels) / sizeof(*GModels))
        return false;

    const char* fn = Config.GetValue(GModels[i].Key);
    if (fn == NULL)
        return false;

    return Earth.SetGravityModel(mg, fn);
};
///////////////////////////////////////////////////////////////////////////////
