///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Астробаллистические данные и алгоритмы, используемые в комплексе
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_BNOASTROBASE_H
#define BNOTOOL_BNOASTROBASE_H
///////////////////////////////////////////////////////////////////////////////
#include "IAstroBase.h"
#include "AstroTimeTables.h"
#include "GeoSolarFlux.h"
#include "Planet.h"
#include "TextConfig.h"

///////////////////////////////////////////////////////////////////////////////
// Базовый класс астробаллистической базы
///////////////////////////////////////////////////////////////////////////////
class BNOAstroBase : public IAstroBase
{
public:
    // Конструктор по умолчанию
    BNOAstroBase();

    // Принудительная установка файла конфигурации
    void SetCfgFile(const char* fn)
    {
        CfgFName = fn;
    };

    // Инициализация базы
    bool Init(bool noAtm = false, bool loadPlanets = false);
    // Инициализация базы с явным указанием времени для эфемерид
    bool Init(const TimeGD& s, const TimeGD& e, bool noAtm = false, bool loadPlanets = false);
    // Инициализация базы с явным указанием времени для эфемерид
    bool Init(const TimeJD& s, const TimeJD& e, bool noAtm = false, bool loadPlanets = false);

    // Преобразование времени между шкалами
    virtual void ConvertTimeScaleJD(TmScale src, TmScale dst, int& d,
                                    double& f) const;
    // Вычисление матрицы перехода от J2000 к METE
    virtual void MtxJ20002METE(const TimeJD& jdMETE, TransMtx& mtx) const;
    // Вычисление истинного звездного времени
    virtual double SaTime(const TimeJD& jd) const;
    // Вычисление истинного звездного времени при известной поправке
    virtual double SaTime(const TimeJD& jd, double ds) const;
    // Вычисление среднего звездного времени
    virtual double Smean(const TimeJD& dt) const;
    // Вычисление матрицы перехода от ECI к фиксированному Гринвичу
    virtual void MtxECI2GFCS(const TimeJD& jdECI, const TimeJD& jdGFCS,
                             TransMtx& mtx) const;
    // Вычисление матрицы, учитывающей прецессию оси вращения Земли
    virtual double MtxPrecession(const TimeJD& dtStart, const TimeJD& dtEnd,
                                 TransMtx& mtx) const;
    // Получение матрицы перехода от METE к Гринвичу
    virtual void MtxMETE2GCS(const TimeJD& t, TransMtx& mtx) const;

    // Получение указателя на модель планеты
    virtual IPlanet* GetPlanet(TPlanetId pl);
    virtual const IPlanet* GetPlanet(TPlanetId pl) const;
    // Получение указателя на таблицу атмосферных индексов Земли
    TTableGeoSolarFlux* GetTableGeoSolarFlux()
    {
        return &TableGeoSolarFlux;
    };

    const TTableGeoSolarFlux* GetTableGeoSolarFlux() const
    {
        return &TableGeoSolarFlux;
    };

    // Получение положений планет и тел Солнечной системы
    virtual void GetPosGelio(TPlanetId pl, const TimeJD& t, Vect3* crd, Vect3* vel = NULL,
                             Vect3* acc = NULL) const;
    // Получение положений планет и тел Солнечной системы
    virtual void GetPosGeo(TPlanetId pl, const TimeJD& t, Vect3* crd, Vect3* vel = NULL,
                           Vect3* acc = NULL) const;

    // Загрузка в память данных по эфемеридам
    bool LoadEphData(TPlanetId pl);
    // Выгрузка из памяти данных по эфемеридам
    void UnLoadEphData(TPlanetId pl);

    // Установка модели атмосферы
    void SetAtmosphereModel(TAtmosphereId mi, bool flag3h = false)
    {
        Earth.SetAtmModel(mi, &TableGeoSolarFlux, flag3h);
    };
    // Установка модели гравитации для Земли
    bool SetEarthGravityModel(TGravityModelId mg);

    void GetWGS84Geoid(double& EarthRadius, double& EarthOblatness) const
    {
        EarthRadius = 6.3781369999999997;
        EarthOblatness = 0.0033528106647474805;
    }

    void GetPZ9002Geoid(double& EarthRadius, double& EarthOblatness) const
    {
        EarthRadius = 6.378136;
        EarthOblatness = 1.0 / 0.29825784E+03;
    }

protected:
    TString CfgFName;
    TimeJD StartTime;
    TimeJD EndTime;
    TextConfig Config;

    TTableUTC2UT1 TableUTC2UT1;
    TTableTAI2UTC TableTAI2UTC;
    TTablePolus TablePolus;
    TTableGeoSolarFlux TableGeoSolarFlux;

    SPlanet Planets[8];
    MoonPlanet Moon;
    SunPlanet Sun;
    EarthPlanet Earth;
    PlanetEphemeride EarthMoon;

    // Загрузка таблицы TAI -> UTC
    bool LoadTAITable(const char* fname)
    {
        return TableTAI2UTC.Init(fname);
    };
    // Загрузка таблицы UTC -> UT1
    bool LoadUT1Table(const char* fname)
    {
        return TableUTC2UT1.Init(fname);
    };
    // Загрузка таблицы движения полюсов
    bool LoadPolusTable(const char* fname)
    {
        return TablePolus.Init(fname);
    };
    // Загрузка таблицы атмосферных индексов
    bool LoadGeoSolarFlux(const char* fname)
    {
        return TableGeoSolarFlux.Init(fname);
    };

    // Вычисление матрицы, учитывающей нутации оси вращения Земли
    double MtxNutation(const TimeJD& dt, TransMtx& mtx) const;
    // Функция вычисляет матрицу, учитывающую смещение положения мгновенного 
    // полюса Земли относительно международного условного начала
    void MtxPolus(double x, double y, TransMtx& mtx) const;
    // Функция вычисляет матрицу, учитывающую смещение положения мгновенного 
    // полюса Земли относительно международного условного начала
    void MtxPolus(const TimeJD& t, TransMtx& mtx) const;
};

///////////////////////////////////////////////////////////////////////////////
#endif
