#include "CloseAppr.h"
#include "Consts.h"
///////////////////////////////////////////////////////////////////////////////
// Характеристики одиночного обнаруженного сближения
///////////////////////////////////////////////////////////////////////////////
// Нормировка вектора
//-----------------------------------------------------------------------------
void CAParams::UnitVect(const Vect3& vect, Vect3& unit)
{
    double len = 1.0 / vect.Len();

    int i;
    for (i = 0; i < 3; i++)
        unit[i] = vect[i] * len;
}

//-----------------------------------------------------------------------------
// Построение единичных векторов вдоль системы RNB
//-----------------------------------------------------------------------------
void CAParams::BuildRNB(const PhasePoint6D& rp, Vect3& r, Vect3& n, Vect3& b)
{
    Vect3 wrk, rr = rp.CoordsVel, v;
    rp.CoordsVel.GetVelocity(v);

    wrk = rr.MulVect(v);
    UnitVect(wrk, b);
    UnitVect(rr, r);
    n = r.MulVect(b);
}

//-----------------------------------------------------------------------------
// Расчет корреляционных матриц объектов в RNB
//-----------------------------------------------------------------------------
void CAParams::BuildCM_RNB(const Vect3& r, const Vect3& n, const Vect3& b,
                           const DSymmMatrix& ksrc, DSymmMatrix& krnb)
{
    double s;

    s = r[0] * r[0] * ksrc(0, 0) +
        r[1] * r[1] * ksrc(1, 1) +
        r[2] * r[2] * ksrc(2, 2) +
        2 * r[0] * r[1] * ksrc(0, 1) +
        2 * r[0] * r[2] * ksrc(0, 2) +
        2 * r[1] * r[2] * ksrc(1, 2);
    krnb(0, 0) = s;

    s = n[0] * n[0] * ksrc(0, 0) +
        n[1] * n[1] * ksrc(1, 1) +
        n[2] * n[2] * ksrc(2, 2) +
        2 * n[0] * n[1] * ksrc(0, 1) +
        2 * n[0] * n[2] * ksrc(0, 2) +
        2 * n[1] * n[2] * ksrc(1, 2);
    krnb(1, 1) = s;

    s = b[0] * b[0] * ksrc(0, 0) +
        b[1] * b[1] * ksrc(1, 1) +
        b[2] * b[2] * ksrc(2, 2) +
        2 * b[0] * b[1] * ksrc(0, 1) +
        2 * b[0] * b[2] * ksrc(0, 2) +
        2 * b[1] * b[2] * ksrc(1, 2);
    krnb(2, 2) = s;

    s = r[0] * n[0] * ksrc(0, 0) +
        r[1] * n[1] * ksrc(1, 1) +
        r[2] * n[2] * ksrc(2, 2) +
        (r[0] * n[1] + r[1] * n[0]) * ksrc(0, 1) +
        (r[0] * n[2] + r[2] * n[0]) * ksrc(0, 2) +
        (r[1] * n[2] + r[2] * n[1]) * ksrc(1, 2);
    krnb(0, 1) = s;

    s = r[0] * b[0] * ksrc(0, 0) +
        r[1] * b[1] * ksrc(1, 1) +
        r[2] * b[2] * ksrc(2, 2) +
        (r[0] * b[1] + r[1] * b[0]) * ksrc(0, 1) +
        (r[0] * b[2] + r[2] * b[0]) * ksrc(0, 2) +
        (r[1] * b[2] + r[2] * b[1]) * ksrc(1, 2);
    krnb(0, 2) = s;

    s = n[0] * b[0] * ksrc(0, 0) +
        n[1] * b[1] * ksrc(1, 1) +
        n[2] * b[2] * ksrc(2, 2) +
        (n[0] * b[1] + n[1] * b[0]) * ksrc(0, 1) +
        (n[0] * b[2] + n[2] * b[0]) * ksrc(0, 2) +
        (n[1] * b[2] + n[2] * b[1]) * ksrc(1, 2);
    krnb(1, 2) = s;
}

//-----------------------------------------------------------------------------
// Вычисление определителя симметричной матрицы 3х3
//-----------------------------------------------------------------------------
double CAParams::SMDet(const DSymmMatrix& p)
{
    return 2 * p(0, 1) * p(0, 2) * p(1, 2) +
        +p(0, 0) * p(1, 1) * p(2, 2)
        - p(0, 2) * p(0, 2) * p(1, 1)
        - p(0, 1) * p(0, 1) * p(2, 2)
        - p(0, 0) * p(1, 2) * p(1, 2);
}

//-----------------------------------------------------------------------------
// Вычисление параметров по взаимному положению объектов
// в указанный момент времени
//-----------------------------------------------------------------------------
bool CAParams::Set(ICASpaceObject& ko1, ICASpaceObject& ko2, const TimeJD tca)
{
    SOID1 = ko1.GetID();
    SOID2 = ko2.GetID();
    Tca = tca;

    PhasePoint6D p1, p2;
    DSymmMatrix cm1, cm2;

    if (!ko1.GetPos(tca, p1, &cm1) || !ko2.GetPos(tca, p2, &cm2))
        return false;

    // Вычисляем вектора разности по положению
    // и по скорости
    Vect3 dr, dv, rp1, vp1, rp2, vp2;
    rp1 = p1.CoordsVel;
    p1.CoordsVel.GetVelocity(vp1);
    rp2 = p2.CoordsVel;
    p2.CoordsVel.GetVelocity(vp2);

    int i;
    for (i = 0; i < 3; i++) {
        dr[i] = rp2[i] - rp1[i];
        dv[i] = vp2[i] - vp1[i];
    }

    // Вычисление модулей расстояния и отн. скорости
    Dr = dr.Len();
    Dv = dv.Len();

    // Строим единичные вектора, расположенные вдоль
    // систем RNB объектов
    Vect3 r1, n1, b1, r2, n2, b2;
    BuildRNB(p1, r1, n1, b1);
    BuildRNB(p2, r2, n2, b2);

    // Проекция расстояния и скорости на системы RNB объектов
    DrRNB1[0] = dr * r1;
    DrRNB2[0] = -(dr * r2);
    DvRNB1[0] = dv * r1;
    DvRNB2[0] = -(dv * r2);
    DrRNB1[1] = dr * n1;
    DrRNB2[1] = -(dr * n2);
    DvRNB1[1] = dv * n1;
    DvRNB2[1] = -(dv * n2);
    DrRNB1[2] = dr * b1;
    DrRNB2[2] = -(dr * b2);
    DvRNB1[2] = dv * b1;
    DvRNB2[2] = -(dv * b2);

    // Расчет расстояния между орбитами
    double vp12 = vp1 * vp2;
    double vpm2 = vp12 * vp12;
    double v1dr = vp1 * dr;
    double v2dr = vp2 * dr;
    double v_12 = vp1 * vp1;
    double v_22 = vp2 * vp2;
    double del = v_12 * v_22 - vpm2;
    double u_min = (vp12 * v1dr - v_12 * v2dr) / del;
    double v_min = (v_22 * v1dr - vp12 * v2dr) / del;

    double dx = dr[0] + vp2[0] * u_min - vp1[0] * v_min;
    double dy = dr[1] + vp2[1] * u_min - vp1[1] * v_min;
    double dz = dr[2] + vp2[2] * u_min - vp1[2] * v_min;

    MinOrbDistance = sqrt(dx * dx + dy * dy + dz * dz);

    // Расчет углов подлета
    Vect3 dvu;
    UnitVect(dv, dvu);

    Alpha1 = Pi - acos(n1 * dvu);
    Alpha2 = acos(n2 * dvu);
    Beta1 = -Pid2 + acos(r1 * dvu);
    Beta2 = Pid2 - acos(r2 * dvu);

    // Расчет корреляционных матриц для RNB объектов
    DSymmMatrix k12, k1(3), k2(3), krnb1(3), krnb2(3), kprnb1(3), kprnb2(3);
    int j;
    for (i = 0; i < 3; i++)
        for (j = i; j < 3; j++) {
            k1(i, j) = cm1(i, j);
            k2(i, j) = cm2(i, j);
        }
    k12 = k1;
    k12 += k2;
    BuildCM_RNB(r1, n1, b1, k12, krnb1);
    BuildCM_RNB(r2, n2, b2, k12, krnb2);
    BuildCM_RNB(r1, n1, b1, k1, kprnb1);
    BuildCM_RNB(r2, n2, b2, k2, kprnb2);

    // Расчет оценок точности по расстоянию
    for (i = 0; i < 3; i++) {
        SigDrRNB1[i] = sqrt(krnb1(i, i));
        SigDrRNB2[i] = sqrt(krnb2(i, i));
        SigDpRNB1[i] = sqrt(kprnb1(i, i));
        SigDpRNB2[i] = sqrt(kprnb2(i, i));
    }

    Vect3 udr;
    UnitVect(dr, udr);
    SigDr = sqrt(udr[0] * udr[0] * k12(0, 0) +
        udr[1] * udr[1] * k12(1, 1) +
        udr[2] * udr[2] * k12(2, 2) +
        2 * udr[0] * udr[1] * k12(0, 1) +
        2 * udr[0] * udr[2] * k12(0, 2) +
        2 * udr[1] * udr[2] * k12(1, 2));

    // Расчет вероятности столкновения
    DSymmMatrix k12_inv;
    double s1 = ko1.GetSize() * 1.e-6; // тыс. км
    double s2 = ko2.GetSize() * 1.e-6; // тыс. км
    double s = 0.125 * (s1 + s2) * (s1 + s2);
    k12_inv = k12;
    k12_inv.Inverse();

    //////////////////////////////////////
    DSymmMatrix k1_inv, k2_inv, kk_inv;
    k1_inv = k1;
    k1_inv.Inverse();
    k2_inv = k2;
    k2_inv.Inverse();
    kk_inv = k1_inv;
    kk_inv += k2_inv;
    //////////////////////////////////////

    double krr = dr[0] * dr[0] * k12_inv(0, 0) +
        dr[1] * dr[1] * k12_inv(1, 1) +
        dr[2] * dr[2] * k12_inv(2, 2) +
        2 * dr[0] * dr[1] * k12_inv(0, 1) +
        2 * dr[0] * dr[2] * k12_inv(0, 2) +
        2 * dr[1] * dr[2] * k12_inv(1, 2);

    double kvv = dv[0] * dv[0] * k12_inv(0, 0) +
        dv[1] * dv[1] * k12_inv(1, 1) +
        dv[2] * dv[2] * k12_inv(2, 2) +
        2 * dv[0] * dv[1] * k12_inv(0, 1) +
        2 * dv[0] * dv[2] * k12_inv(0, 2) +
        2 * dv[1] * dv[2] * k12_inv(1, 2);

    double krv = dr[0] * dv[0] * k12_inv(0, 0) +
        dr[1] * dv[1] * k12_inv(1, 1) +
        dr[2] * dv[2] * k12_inv(2, 2) +
        (dr[0] * dv[1] + dr[1] * dv[0]) * k12_inv(0, 1) +
        (dr[0] * dv[2] + dr[2] * dv[0]) * k12_inv(0, 2) +
        (dr[1] * dv[2] + dr[2] * dv[1]) * k12_inv(1, 2);

    double kp = s * Dv / sqrt(kvv * SMDet(k1) * SMDet(k2) * SMDet(kk_inv));
    Probab = kp * exp(-0.5 * (krr - krv * krv / kvv));

    return true;
}

///////////////////////////////////////////////////////////////////////////////
// Список обнаруженных сближений
///////////////////////////////////////////////////////////////////////////////
// Уничтожение дублей
// dt - порог "нечувствительности" по времени (в тыс. секунд)
//-----------------------------------------------------------------------------
void CAParamsList::KillDoubles(double dt)
{
    CAParamsEntry *f, *s, *n;

    for (f = first(); f != NULL; f = f->next()) {
        s = f->next();
        while (s != NULL) {
            n = s->next();
            TmDay dd = f->Tca - s->Tca;
            double d = dd.GetDouble();
            if ((f->SOID1 == s->SOID1 && f->SOID2 == s->SOID2 && fabs(d) < dt / 86.4)
                || (f->SOID1 == s->SOID2 && f->SOID2 == s->SOID2 && fabs(d) < dt / 86.4))
                delete_node(s);
            s = n;
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
