///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Классы, предназначенные для расчета опасных сближений
///////////////////////////////////////////////////////////////////////////////
#ifndef CLOSE_APPR_H
#define CLOSE_APPR_H

#include "PhasePoint.h"
#include "Matrix.h"
#include "twll.h"

///////////////////////////////////////////////////////////////////////////////
// Описание одиночного КО, участвующего в расчете опасных сближений
///////////////////////////////////////////////////////////////////////////////
class ICASpaceObject
{
public:
    // Деструктор
    virtual ~ICASpaceObject();

    // Запрос идентификатора КО
    virtual unsigned GetID() const = 0;
    // Запрос ср. размеров КО (в метрах)
    virtual double GetSize() const = 0;
    // Запрос интервала применимости орбиты КО
    virtual void GetTimeRange(TimeJD& tmin, TimeJD& tmax) const;
    // Запрос диапазона высот, в котором заключена орбита
    // (м.б. либо приближенным, либо точным в зависимости от
    // реализации)
    virtual void GetHeightRange(double& hp, double& ha) const;

    // Расчет положения и корреляционной матрицы объекта
    // на указанный момент времени
    // Метод возвращает false, если не удалось выполнить прогноз
    // Корреляционная матрица должна быть задана в той же
    // системе отсчета, что и выходной фазовый вектор
    virtual bool GetPos(const TimeJD& t, PhasePoint6D& rv,
                        DSymmMatrix* cm = NULL);
};

///////////////////////////////////////////////////////////////////////////////
// Описание одиночного КО, участвующего в расчете опасных сближений,
// как элемента списка
// ВНИМАНИЕ! Конкретные реализации КО, участвующие в расчете сближений
// должны наследоваться от этого класса
///////////////////////////////////////////////////////////////////////////////
typedef Twll_Entry_Class<ICASpaceObject> ICASpaceObjectEntry;

///////////////////////////////////////////////////////////////////////////////
// Список КО, участвующих в расчете опасных сближений
///////////////////////////////////////////////////////////////////////////////
class CASpaceObjectList : public Twll_List_Class<ICASpaceObjectEntry>
{
public:
    // Вычисление диапазона высот, в котором находятся орбиты объектов
    void GetHeightRange(double& hp, double& ha) const;

    // Фильтрация списка по диапазону высот
    // (Из списка удаляются объекты, орбиты которых лежат вне указанного
    // диапазона высот)
    void FilterByHeight(double hp, double ha);
    // Фильтрация списка по временному диапазону
    // (Из списка удаляются объекты, у которых интервалы применимости 
    // лежат вне указанного диапазона)
    void FilterByTime(const TimeJD& tmin, const TimeJD& tmax);
};

///////////////////////////////////////////////////////////////////////////////
// Характеристики одиночного обнаруженного сближения
///////////////////////////////////////////////////////////////////////////////
class CAParams
{
public:
    //--------------------------------------------------------
    // Поля данных
    //--------------------------------------------------------
    // Идентификаторы объектов
    int SOID1, SOID2;
    // Время сближения
    TimeJD Tca;
    // Расстояние между объектами в момент сближения
    double Dr;
    // Проекции вектора разности в системах RNB объектов
    Vect3 DrRNB1, DrRNB2;
    // СКО ошибок определения расстояния
    double SigDr;
    Vect3 SigDrRNB1, SigDrRNB2;
    // СКО ошибок определения положения КО системах RNB объектов
    Vect3 SigDpRNB1, SigDpRNB2;
    // Модуль отн. скорости сближения
    double Dv;
    // Прооекции вектора отн. скорости в системах RNB объектов
    Vect3 DvRNB1, DvRNB2;
    // Расстояние между орбитами
    double MinOrbDistance;
    // Угла подлета
    double Alpha1, Beta1, Alpha2, Beta2;
    // Вероятность столкновения
    double Probab;
    // Положения объектов в момент максимального сближения
    PhasePoint6D Pos1, Pos2;
    // Корреляционные матрицы положений объектов
    DSymmMatrix Cm1, Cm2;

    //--------------------------------------------------------
    // Методы
    //--------------------------------------------------------
    // Вычисление параметров по взаимному положению объектов
    // в указанный момент времени
    bool Set(ICASpaceObject& ko1, ICASpaceObject& ko2, const TimeJD tca);

protected:
    //--------------------------------------------------------
    // Вспомогательные методы, используемые при расчетах
    //--------------------------------------------------------
    // Нормировка вектора
    static void UnitVect(const Vect3& vect, Vect3& unit);
    // Построение единичных векторов вдоль системы RNB
    static void BuildRNB(const PhasePoint6D& rp, Vect3& r, Vect3& n, Vect3& b);
    // Расчет корреляционных матриц объектов в RNB
    static void BuildCM_RNB(const Vect3& r, const Vect3& n, const Vect3& b,
                            const DSymmMatrix& ksrc, DSymmMatrix& krnb);
    // Вычисление определителя симметричной матрицы 3х3
    static double SMDet(const DSymmMatrix& p);
};

///////////////////////////////////////////////////////////////////////////////
// Список обнаруженных сближений
///////////////////////////////////////////////////////////////////////////////
typedef Twll_Entry_Class<CAParams> CAParamsEntry;

class CAParamsList : public Twll_List_Class<CAParamsEntry>
{
public:
    // Уничтожение дублей
    // dt - порог "нечувствительности" по времени (в тыс. секунд)
    void KillDoubles(double dt = 1.e-3);
};

///////////////////////////////////////////////////////////////////////////////
// Вспомогательный класс, моделирующий узлы орбит и 
// равноускоренное движение
// Используется только внутри алгоритма поиска опасных сближений
///////////////////////////////////////////////////////////////////////////////
class CAOrbitNodes
{
public:
    CAOrbitNodes()
    {
        NodeFlag = false;
    };

    virtual ~CAOrbitNodes()
    {
    }

    // Указатель на исходный объект
    ICASpaceObject* Ko;

    // Инициализация объекта
    bool Init(ICASpaceObject& ko)
    {
        Ko = &ko;
        return true;
    }

    // Методы для быстрого приближенного прогнозирования
    // Установка опорного узла
    void SetNode(const TimeJD& jd);
    // Проверка правильности узла
    bool IsNodeValid() const
    {
        return NodeFlag;
    };
    // Быстрый прогноз на заданное время
    void FastPredict(const TimeJD& jd);
    // Значение координаты для сортировки
    double GetSortFCoord() const
    {
        return F_r[0];
    };
    // Координаты полученные быстрым прогнозом
    const Vect3& GetFCoords() const
    {
        return F_r;
    };
    // Скорости полученные быстрым прогнозом
    const Vect3& GetFVelocity() const
    {
        return F_v;
    };
    // Время, на которое выполнялся прогноз
    const TimeJD& FPredictTime() const
    {
        return F_time;
    };

    // Точный прогноз. Возвращает true, если он был успешен
    bool GetCoordVelocity(const TimeJD& jd, Vect3& coord, Vect3& velocity);

    // Проверка вхождения указанного времени в интервал применимости
    // траекторных данных
    bool IsInsideInterval(const TimeJD& jd) const;
    // Получение левой границы применимости данных
    //const TimeJD &GetStartTime() const;
    // Получение правой границы применимости данных
    //const TimeJD &GetEndTime() const;

private:
    // Параметры узла интерполяции
    // параметры узла
    TimeJD Node_Time;
    bool NodeFlag;
    Vect3 Node_R;
    Vect3 Node_V;
    Vect3 Node_A;

    // Результаты приближенного прогноза
    TimeJD F_time;
    Vect3 F_r;
    Vect3 F_v;
};

///////////////////////////////////////////////////////////////
//  Класс, выполняющий расчет опасных сближений
///////////////////////////////////////////////////////////////
class CASeacher
{
public:
    // -----------------------------------------------------
    // Блок переменных, управляющих алгоритмом поиска
    // Все эти переменные устанавливаются в конструкторе
    // по умолчанию
    // -----------------------------------------------------
    double MinPermissDistRad;
    double MinPermissDistNorm;
    double MinPermissDistBinorm;
    double MinPermissDistRadAbs;
    double MinPermissDistNormAbs;
    double MinPermissDistBinormAbs;
    double MinProbability;
    double TimeProbStep;
    double CoordDiff;
    double VelocityDiff;
    double LinearInt;
    double RVectorDiff2;
    double RVectorDiff3;

    CASeacher(void);

    virtual ~CASeacher(void)
    {
    };

    // Собственно расчет сближений "список на список"
    bool Run(CASpaceObjectList& ListKO1, CASpaceObjectList& ListKO2,
             const TimeJD& tstart, const TimeJD& tend);
    // Расчет сближений "объект на объект"
    bool Run(ICASpaceObject& dko1, ICASpaceObject& dko2,
             const TimeJD& tstart, const TimeJD& te);
    // Получение списка сближений
    const CAParamsList& GetResult() const
    {
        return ListCA;
    };
    // Очистка списка сближений
    void Clear()
    {
        ListCA.clear();
    };

protected:
    // Список рассчитанных сближений
    CAParamsList ListCA;
    // Расчет параметров сближения на интервале
    // от left_t до left_t + TimeProbStep
    // при необходимости - занесение в список сближений
    void AddCA(CAOrbitNodes& do2, CAOrbitNodes& do1,
               const TimeJD& left_t, const TimeJD& tend);
    // Вычисление минимального расстояния между объектами по
    // приближенным формулам
    static void MinDistance(const Vect3& r1, const Vect3& v1, const Vect3& r2, const Vect3& v2,
                            double& min_gap, double& time);
    // Следует ли считать сближение опасным?
    virtual bool IsDangerous(const CAParams& ca,
                             const ICASpaceObject& ko1, const ICASpaceObject& ko2);
};

///////////////////////////////////////////////////////////////////////////////
#endif
