///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Класс, реализующий диагональную матрицу
///////////////////////////////////////////////////////////////////////////////
#include "Matrix.h"

#include <math.h>
///////////////////////////////////////////////////////////////////////////////
// Инициализация
//-----------------------------------------------------------------------------
void DDiagMatrix::Init(const DDiagMatrix& mt)
{
    TAssert(mt.Vector != NULL);
    DLineMem::Init(mt.Dimension);

    memcpy(Vector, mt.Vector, mt.Dimension * sizeof(double));
}

//-----------------------------------------------------------------------------
// Оператор сложения
//-----------------------------------------------------------------------------
DDiagMatrix& DDiagMatrix::operator+=(const DDiagMatrix& mt)
{
    TAssert(mt.Vector != NULL);
    TAssert(mt.Dim() == Dimension);
    TAssert(Vector != NULL);

    unsigned i;

    for (i = 0; i < Dimension; i++)
        Vector[i] += mt.Vector[i];

    return *this;
}

//-----------------------------------------------------------------------------
// Оператор вычитания
//-----------------------------------------------------------------------------
DDiagMatrix& DDiagMatrix::operator-=(const DDiagMatrix& mt)
{
    TAssert(mt.Vector != NULL);
    TAssert(mt.Dim() == Dimension);
    TAssert(Vector != NULL);

    unsigned i;

    for (i = 0; i < Dimension; i++)
        Vector[i] -= mt.Vector[i];

    return *this;
}

//-----------------------------------------------------------------------------
// Умножение на матрицу справа
//-----------------------------------------------------------------------------
DDiagMatrix& DDiagMatrix::MulRight(const DDiagMatrix& mt)
{
    TAssert(mt.Vector != NULL);
    TAssert(mt.Dim() == Dimension);
    TAssert(Vector != NULL);
    TAssert(Vector != mt.Vector);

    unsigned i;

    for (i = 0; i < Dimension; i++)
        Vector[i] *= mt.Vector[i];

    return *this;
}

//-----------------------------------------------------------------------------
// Разложение матрицы на верхнюю и диагональную
//-----------------------------------------------------------------------------
bool DDiagMatrix::Expand(DUppMatrix& s, DDiagMatrix& d) const
{
    TAssert(Vector != NULL);

    unsigned i;

    for (i = 0; i < Dimension; i++)
        if (fabs(Vector[i]) < ZERO_ACCURACY)
            return false;

    d = *this;
    s.Init(Dimension);

    s.Init(0.0);
    for (i = 0; i < Dimension; i++)
        s.Buffer[s.Index(i, i)] = 1.0;

    return true;
}

//-----------------------------------------------------------------------------
// Инвертирование (формирование обратной метрицы)
//-----------------------------------------------------------------------------
bool DDiagMatrix::Inverse()
{
    TAssert(Vector != NULL);

    unsigned i, j;

    for (i = 0; i < Dimension; i++) {
        if (fabs(Vector[i]) < ZERO_ACCURACY) {
            for (j = 0; j < i; j++)
                Vector[j] = 1.0 / Vector[j];
            return false;
        }
        Vector[i] = 1.0 / Vector[i];
    }

    return true;
}

//-----------------------------------------------------------------------------
// Реализация запроса значений строки
//-----------------------------------------------------------------------------
void DDiagMatrix::GetLine(unsigned index, double* line) const
{
    TAssert(Vector != NULL);
    TAssert(line != NULL);
    TAssert(index < Dimension);

    memset(line, 0, Dimension * sizeof(double));
    line[index] = Vector[index];
}

//-----------------------------------------------------------------------------
// Реализация запроса значений стoлбца
//-----------------------------------------------------------------------------
void DDiagMatrix::GetCol(unsigned index, double* col) const
{
    GetLine(index, col);
}

//-----------------------------------------------------------------------------
// Реализация запроса значений элементов в виде массива
//-----------------------------------------------------------------------------
void DDiagMatrix::GetArray(double* array) const
{
    TAssert(Vector != NULL);
    TAssert(array != NULL);

    memset(array, 0, Dimension * Dimension * sizeof(double));

    unsigned i;
    for (i = 0; i < Dimension; i++) {
        *array = Vector[i];
        array += Dimension + 1;
    }
}

///////////////////////////////////////////////////////////////////////////////
