///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Служебный класс хранения вектора в памяти
///////////////////////////////////////////////////////////////////////////////
#include "Matrix.h"

#include <math.h>
///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
//-----------------------------------------------------------------------------
DLineMem::DLineMem()
{
    Dimension = 0;
    DimMem = 0;
    MemSize = 0;
    Vector = NULL;
}

//-----------------------------------------------------------------------------
// Деструктор
//-----------------------------------------------------------------------------
DLineMem::~DLineMem(void)
{
    if (Vector != NULL)
        delete Vector;
}

//-----------------------------------------------------------------------------
// Обмен содержимого
//-----------------------------------------------------------------------------
void DLineMem::XChng(DLineMem& m)
{
    unsigned b;

    b = Dimension;
    Dimension = m.Dimension;
    m.Dimension = b;

    b = DimMem;
    DimMem = m.DimMem;
    m.DimMem = b;

    b = MemSize;
    MemSize = m.MemSize;
    m.MemSize = b;

    double* ptr;
    ptr = Vector;
    Vector = m.Vector;
    m.Vector = ptr;
}

//-----------------------------------------------------------------------------
// Инициализация
//-----------------------------------------------------------------------------
void DLineMem::Init(unsigned dim)
{
    if (dim > MemSize || Vector == NULL) {
        if (Vector != NULL)
            delete Vector;
        Vector = new double[dim];
        MemSize = dim;
    }

    DimMem = Dimension = dim;
}

//-----------------------------------------------------------------------------
// Инициализация
//-----------------------------------------------------------------------------
void DLineMem::Init(const double* vec, unsigned dim)
{
    Init(dim);
    if (Vector == NULL)
        return;

    memcpy(Vector, vec, sizeof(double) * dim);
}

//-----------------------------------------------------------------------------
// Сложение с матрицей (столбцом)
//-----------------------------------------------------------------------------
void DLineMem::Add(const CDMatrix& mt)
{
    TAssert(Vector != NULL);
    TAssert(mt.Cols() == 1);
    TAssert(Dimension == mt.Rows());

    unsigned i;
    for (i = 0; i < mt.Rows(); i++)
        Vector[i] += mt(i, 0);

    return;
}

//-----------------------------------------------------------------------------
// Вычитание матрицы (столбца)
//-----------------------------------------------------------------------------
void DLineMem::Sub(const CDMatrix& mt)
{
    TAssert(Vector != NULL);
    TAssert(mt.Cols() == 1);
    TAssert(Dimension == mt.Rows());

    unsigned i;
    for (i = 0; i < mt.Rows(); i++)
        Vector[i] -= mt(i, 0);

    return;
}

//-----------------------------------------------------------------------------
// Сложение с вектором
//-----------------------------------------------------------------------------
void DLineMem::AddVect(const DLineMem& mt)
{
    TAssert(Vector != NULL);
    TAssert(mt.Vector != NULL);
    TAssert(Dimension == mt.Dimension);

    unsigned i;
    for (i = 0; i < mt.Dimension; i++)
        Vector[i] += mt.Vector[i];

    return;
}

//-----------------------------------------------------------------------------
// Вычитание вектора
//-----------------------------------------------------------------------------
void DLineMem::SubVect(const DLineMem& mt)
{
    TAssert(Vector != NULL);
    TAssert(mt.Vector != NULL);
    TAssert(Dimension == mt.Dimension);

    unsigned i;
    for (i = 0; i < mt.Dimension; i++)
        Vector[i] -= mt.Vector[i];

    return;
}

//-----------------------------------------------------------------------------
// Изменение размерности
//-----------------------------------------------------------------------------
void DLineMem::ResetDimension(unsigned dim)
{
    TAssert(dim <= DimMem);
    Dimension = dim;
}

//-----------------------------------------------------------------------------
// Поиск максимального элемента по модулю
//-----------------------------------------------------------------------------
double DLineMem::GetMaxElementAbs() const
{
    TAssert(Vector != NULL);
    TAssert(Dimension != 0);

    double max_value = fabs(Vector[0]);
    double value_abs;

    for (unsigned i = 1; i < Dimension; ++i) {
        value_abs = fabs(Vector[i]);
        if (value_abs > max_value)
            max_value = value_abs;
    }

    return max_value;
}

//-----------------------------------------------------------------------------
// Поиск минимального элемента по модулю
//-----------------------------------------------------------------------------
double DLineMem::GetMinElementAbs() const
{
    TAssert(Vector != NULL);
    TAssert(Dimension != 0);

    double min_value = fabs(Vector[0]);
    double value_abs;

    for (unsigned i = 1; i < Dimension; ++i) {
        value_abs = fabs(Vector[i]);
        if (value_abs < min_value)
            min_value = value_abs;
    }

    return min_value;
}

//-----------------------------------------------------------------------------
// Поиск максимального элемента
//-----------------------------------------------------------------------------
double DLineMem::GetMaxElement() const
{
    TAssert(Vector != NULL);
    TAssert(Dimension != 0);

    double max_value = Vector[0];

    for (unsigned i = 1; i < Dimension; ++i) {
        max_value = (Vector[i] > max_value) ? Vector[i] : max_value;
    }

    return max_value;
}

//-----------------------------------------------------------------------------
// Поиск минимального элемента
//-----------------------------------------------------------------------------
double DLineMem::GetMinElement() const
{
    TAssert(Vector != NULL);
    TAssert(Dimension != 0);

    double min_value = Vector[0];

    for (unsigned i = 1; i < Dimension; ++i) {
        min_value = (Vector[i] < min_value) ? Vector[i] : min_value;
    }

    return min_value;
}

//-----------------------------------------------------------------------------
// Установка единого значения для всех элементов
//-----------------------------------------------------------------------------
void DLineMem::Init(double value)
{
    TAssert(Vector != NULL);

    unsigned i;

    for (i = 0; i < Dimension; i++)
        Vector[i] = value;
}

//-----------------------------------------------------------------------------
// Умножение всех элементов на число
//-----------------------------------------------------------------------------
void DLineMem::Mult(double mltp)
{
    TAssert(Vector != NULL);

    unsigned i;

    for (i = 0; i < Dimension; i++)
        Vector[i] *= mltp;
}

//-----------------------------------------------------------------------------
// Формирование нового вектора, как результат умножения матрицы на
// вектор-столбец
//-----------------------------------------------------------------------------
void DLineMem::MulRect(const DRectMem& mt, const DLineMem& vc)
{
    TAssert(mt.Matrix != NULL);
    TAssert(vc.Vector != NULL);
    TAssert(Vector != vc.Vector);
    TAssert(vc.Dimension == mt.NCols);

    Init(mt.NRows);
    if (Vector == NULL)
        return;

    unsigned i, j;
    double* line;

    for (i = 0; i < mt.NRows; i++) {
        Vector[i] = 0;
        line = mt.Matrix[i];
        for (j = 0; j < mt.NCols; j++)
            Vector[i] += line[j] * vc.Vector[j];
    }

    return;
}

//-----------------------------------------------------------------------------
// Умножение на матрицу слева
//-----------------------------------------------------------------------------
void DLineMem::MulLeftRect(const DRectMem& mt)
{
    TAssert(mt.Matrix != NULL);
    TAssert(Vector != NULL);
    TAssert(Dimension == mt.NCols);

    DLineMem ptr;
    ptr.Init(mt.NRows);
    XChng(ptr);
    if (Vector == NULL)
        return;

    unsigned i, j;
    double* line;

    for (i = 0; i < mt.NRows; i++) {
        Vector[i] = 0;
        line = mt.Matrix[i];
        for (j = 0; j < mt.NCols; j++)
            Vector[i] += line[j] * ptr.Vector[j];
    }

    return;
}

///////////////////////////////////////////////////////////////////////////////
