///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Класс, реализующий нижнюю треугольную матрицу
///////////////////////////////////////////////////////////////////////////////
#include "Matrix.h"
///////////////////////////////////////////////////////////////////////////////
// Инициализация
//-----------------------------------------------------------------------------
void DLowMatrix::Init(const DLowMatrix& mt)
{
    TAssert(mt.Buffer != NULL);
    DTriagMem::Init(mt.Dim());

    unsigned i, j;

    for (i = 0; i < Dimension; i++) {
        for (j = 0; j < i + 1; j++)
            Buffer[Index(j, i)] = mt.Buffer[mt.Index(j, i)];
    }
}

//-----------------------------------------------------------------------------
// Инициализация
//-----------------------------------------------------------------------------
void DLowMatrix::Init(const DDiagMatrix& mt)
{
    DTriagMem::Init(mt.Dim());

    DTriagMem::Init(0.0);
    unsigned i;

    for (i = 0; i < Dimension; i++)
        Buffer[Index(i, i)] = mt.Vector[i];
}

//-----------------------------------------------------------------------------
// Оператор сложения
//-----------------------------------------------------------------------------
DLowMatrix& DLowMatrix::operator+=(const DLowMatrix& mt)
{
    TAssert(mt.Buffer != NULL);
    TAssert(mt.Dim() == Dimension);
    TAssert(Buffer != NULL);

    unsigned i, j;

    for (i = 0; i < Dimension; i++) {
        for (j = 0; j < i + 1; j++)
            Buffer[Index(j, i)] += mt.Buffer[mt.Index(j, i)];
    }

    return *this;
}

//-----------------------------------------------------------------------------
// Оператор вычитания
//-----------------------------------------------------------------------------
DLowMatrix& DLowMatrix::operator-=(const DLowMatrix& mt)
{
    TAssert(mt.Buffer != NULL);
    TAssert(mt.Dim() == Dimension);
    TAssert(Buffer != NULL);

    unsigned i, j;

    for (i = 0; i < Dimension; i++) {
        for (j = 0; j < i + 1; j++)
            Buffer[Index(j, i)] -= mt.Buffer[mt.Index(j, i)];
    }

    return *this;
}

//-----------------------------------------------------------------------------
// Оператор сложения
//-----------------------------------------------------------------------------
DLowMatrix& DLowMatrix::operator+=(const DDiagMatrix& mt)
{
    TAssert(mt.Dim() == Dimension);
    TAssert(Buffer != NULL);

    unsigned i;

    for (i = 0; i < Dimension; i++)
        Buffer[Index(i, i)] += mt.Vector[i];

    return *this;
}

//-----------------------------------------------------------------------------
// Оператор вычитания
//-----------------------------------------------------------------------------
DLowMatrix& DLowMatrix::operator-=(const DDiagMatrix& mt)
{
    TAssert(mt.Dim() == Dimension);
    TAssert(Buffer != NULL);

    unsigned i;

    for (i = 0; i < Dimension; i++)
        Buffer[Index(i, i)] -= mt.Vector[i];

    return *this;
}

//-----------------------------------------------------------------------------
// Умножение на матрицу справа
//-----------------------------------------------------------------------------
DLowMatrix& DLowMatrix::MulRight(const DLowMatrix& mt)
{
    TAssert(mt.Buffer != NULL);
    TAssert(Buffer != NULL);
    TAssert(Buffer != mt.Buffer);
    TAssert(Dimension == mt.Dimension);

    DLowMatrix src(Dimension);
    XChng(src);

    unsigned i, j, k;
    double* p;

    for (i = 0; i < Dimension; i++) {
        for (j = 0; j < i + 1; j++) {
            p = Buffer + Index(j, i);
            *p = 0.0;
            for (k = j; k < i + 1; k++)
                *p += src.Buffer[src.Index(k, i)] * mt.Buffer[mt.Index(j, k)];
        }
    }

    return *this;
}

//-----------------------------------------------------------------------------
// Умножение на матрицу слева
//-----------------------------------------------------------------------------
DLowMatrix& DLowMatrix::MulLeft(const DLowMatrix& mt)
{
    TAssert(mt.Buffer != NULL);
    TAssert(Buffer != NULL);
    TAssert(Buffer != mt.Buffer);
    TAssert(Dimension == mt.Dimension);

    DLowMatrix src(Dimension);
    XChng(src);

    unsigned i, j, k;
    double* p;

    for (i = 0; i < Dimension; i++) {
        for (j = 0; j < i + 1; j++) {
            p = Buffer + Index(j, i);
            *p = 0.0;
            for (k = j; k < i + 1; k++)
                *p += mt.Buffer[mt.Index(k, i)] * src.Buffer[src.Index(j, k)];
        }
    }

    return *this;
}

//-----------------------------------------------------------------------------
// Умножение на матрицу справа
//-----------------------------------------------------------------------------
DLowMatrix& DLowMatrix::MulRight(const DDiagMatrix& mt)
{
    TAssert(mt.Vector != NULL);
    TAssert(Buffer != NULL);
    TAssert(Dimension == mt.Dim());

    DLowMatrix src(Dimension);
    XChng(src);

    unsigned i, j;

    for (i = 0; i < Dimension; i++) {
        for (j = 0; j < i + 1; j++)
            Buffer[Index(j, i)] = src.Buffer[src.Index(j, i)] * mt.Vector[j];
    }

    return *this;
}

//-----------------------------------------------------------------------------
// Умножение на матрицу слева
//-----------------------------------------------------------------------------
DLowMatrix& DLowMatrix::MulLeft(const DDiagMatrix& mt)
{
    TAssert(mt.Vector != NULL);
    TAssert(Buffer != NULL);
    TAssert(Dimension == mt.Dim());

    DLowMatrix src(Dimension);
    XChng(src);

    unsigned i, j;

    for (i = 0; i < Dimension; i++) {
        for (j = 0; j < i + 1; j++)
            Buffer[Index(j, i)] = mt.Vector[i] * src.Buffer[src.Index(j, i)];
    }

    return *this;
}

//-----------------------------------------------------------------------------
// Транспонирование
//-----------------------------------------------------------------------------
DLowMatrix& DLowMatrix::Transpose(const DUppMatrix& mt)
{
    TAssert(mt.Buffer != NULL);

    DTriagMem::Init(mt.Dim());

    unsigned i, j;

    for (i = 0; i < Dimension; i++)
        for (j = 0; j < i + 1; j++)
            Buffer[Index(j, i)] = mt.Buffer[mt.Index(j, i)];

    return *this;
}

//-----------------------------------------------------------------------------
// Реализация запроса значений стoлбца
//-----------------------------------------------------------------------------
void DLowMatrix::GetCol(unsigned index, double* line) const
{
    TAssert(Buffer != NULL);
    TAssert(line != NULL);
    TAssert(index < Dimension);

    if (index > 0)
        memset(line, 0, index * sizeof(double));
    memcpy(line + index, Buffer + Index(index, index), (Dimension - index) * sizeof(double));
}

//-----------------------------------------------------------------------------
// Реализация запроса значений строки
//-----------------------------------------------------------------------------
void DLowMatrix::GetLine(unsigned index, double* col) const
{
    TAssert(Buffer != NULL);
    TAssert(col != NULL);
    TAssert(index < Dimension);

    unsigned i;

    for (i = 0; i <= index; i++)
        col[i] = *(Buffer + Index(i, index));
    if (index < Dimension - 1)
        memset(col + index + 1, 0, (Dimension - index - 1) * sizeof(double));
}

//-----------------------------------------------------------------------------
// Реализация запроса значений элементов в виде массива
//-----------------------------------------------------------------------------
void DLowMatrix::GetArray(double* array) const
{
    TAssert(Buffer != NULL);
    TAssert(array != NULL);

    unsigned i;
    for (i = 0; i < Dimension; i++) {
        GetLine(i, array);
        array += Dimension;
    }
}

///////////////////////////////////////////////////////////////////////////////
