///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
#include "Matrix.h"

#include <string.h>
///////////////////////////////////////////////////////////////////////////////
// Служебный класс хранения прямоугольного массива в памяти
///////////////////////////////////////////////////////////////////////////////
// Конструктор
//-----------------------------------------------------------------------------
DRectMem::DRectMem()
{
    NRows = 0;
    NCols = 0;
    NRowsMem = 0;
    NColsMem = 0;
    BufSize = 0;
    Matrix = NULL;
}

//-----------------------------------------------------------------------------
// Деструктор
//-----------------------------------------------------------------------------
DRectMem::~DRectMem()
{
    if (Matrix != NULL)
        delete[] Matrix;
}

//-----------------------------------------------------------------------------
// Обмен содержимого
//-----------------------------------------------------------------------------
void DRectMem::XChng(DRectMem& m)
{
    unsigned b;

    b = NRows;
    NRows = m.NRows;
    m.NRows = b;

    b = NCols;
    NCols = m.NCols;
    m.NCols = b;

    b = NRowsMem;
    NRowsMem = m.NRowsMem;
    m.NRowsMem = b;

    b = NColsMem;
    NColsMem = m.NColsMem;
    m.NColsMem = b;

    b = BufSize;
    BufSize = m.BufSize;
    m.BufSize = b;

    double** ptr;
    ptr = Matrix;
    Matrix = m.Matrix;
    m.Matrix = ptr;
}

//-----------------------------------------------------------------------------
// Создание буфера
//-----------------------------------------------------------------------------
double** DRectMem::CreatePtr(unsigned n, unsigned m)
{
    if (n == 0 || m == 0)
        return NULL;

    char* point = new char[ReqSize(n, m)];
    double** ptr = (double **)point;
    if (ptr == NULL)
        return NULL;
    SetLinesPtrs(ptr, n, m);
    return ptr;
}

//-----------------------------------------------------------------------------
// Разметка указателей
//-----------------------------------------------------------------------------
void DRectMem::SetLinesPtrs(double** ptr, unsigned n, unsigned m)
{
    double* d_point = (double *)(ptr + n);

    unsigned i;
    for (i = 0; i < n; ++i, d_point += m)
        ptr[i] = d_point;
}

//-----------------------------------------------------------------------------
// Инициализация 
//-----------------------------------------------------------------------------
void DRectMem::Init(unsigned r, unsigned c)
{
    TAssert(r != 0 && c != 0);

    unsigned rsize = ReqSize(r, c);

    if (rsize > BufSize || Matrix == NULL) {
        if (Matrix != NULL)
            delete[] Matrix;
        Matrix = CreatePtr(r, c);
        BufSize = rsize;
    }
    else
        SetLinesPtrs(Matrix, r, c);

    NRowsMem = NRows = r;
    NColsMem = NCols = c;
}

//-----------------------------------------------------------------------------
// Изменение размерности
//-----------------------------------------------------------------------------
void DRectMem::ResetDimension(unsigned r, unsigned c)
{
    TAssert(r <= NRowsMem);
    TAssert(c <= NColsMem);

    NRows = r;
    NCols = c;
}

//-----------------------------------------------------------------------------
// Установка единого значения для всех элементов
//-----------------------------------------------------------------------------
void DRectMem::Init(double value)
{
    TAssert(Matrix != NULL);

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        for (j = 0; j < NCols; j++)
            line[j] = value;
    }
}

//-----------------------------------------------------------------------------
// Инициализация
//-----------------------------------------------------------------------------
void DRectMem::Init(const CDMatrix& mt)
{
    Init(mt.Rows(), mt.Cols());
    if (Matrix == NULL)
        return;

    unsigned i;
    double* line;

    line = new double[NCols];

    for (i = 0; i < NRows; i++) {
        mt.GetLine(i, line);
        memcpy(Matrix[i], line, sizeof(double) * NCols);
    }

    delete[] line;
}

//-----------------------------------------------------------------------------
// Инициализация
//-----------------------------------------------------------------------------
void DRectMem::InitRect(const DRectMem& mt)
{
    TAssert(mt.Matrix != NULL);
    Init(mt.NRows, mt.NCols);
    if (Matrix == NULL)
        return;

    unsigned i;

    if (NRows == NRowsMem && NCols == NColsMem) {
        memcpy(Matrix[0], mt.Matrix[0], sizeof(double) * NRows * NCols);
        return;
    }

    for (i = 0; i < NRows; i++)
        memcpy(Matrix[i], mt.Matrix[i], NCols * sizeof(double));
}

//-----------------------------------------------------------------------------
// Инициализация
//-----------------------------------------------------------------------------
void DRectMem::Init(const DSymmMatrix& mt)
{
    Init(mt.Dim(), mt.Dim());
    if (Matrix == NULL)
        return;

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        line[i] = mt(i, i);
        for (j = i + 1; j < NCols; j++) {
            double a = mt.Buffer[mt.Index(i, j)];
            line[j] = a;
            Matrix[j][i] = a;
        }
    }
};
//-----------------------------------------------------------------------------
// Инициализация
//-----------------------------------------------------------------------------
void DRectMem::Init(const DUppMatrix& mt)
{
    Init(mt.Dim(), mt.Dim());
    if (Matrix == NULL)
        return;

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        memset(line, 0, i * sizeof(double));
        for (j = i; j < NCols; j++)
            line[j] = mt.Buffer[mt.Index(i, j)];
    }
}

//-----------------------------------------------------------------------------
// Инициализация
//-----------------------------------------------------------------------------
void DRectMem::Init(const DLowMatrix& mt)
{
    Init(mt.Dim(), mt.Dim());
    if (Matrix == NULL)
        return;

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        for (j = 0; j < i + 1; j++)
            line[j] = mt.Buffer[mt.Index(j, i)];
        memset(line + i + 1, 0, (NCols - i - 1) * sizeof(double));
    }
}

//-----------------------------------------------------------------------------
// Инициализация
//-----------------------------------------------------------------------------
void DRectMem::Init(const DDiagMatrix& mt)
{
    Init(mt.Dim(), mt.Dim());
    if (Matrix == NULL)
        return;

    unsigned i;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        memset(line, 0, NRows * sizeof(double));
        line[i] = mt.Vector[i];
    }
}

//-----------------------------------------------------------------------------
// Умножение всех элементов на число
//-----------------------------------------------------------------------------
void DRectMem::Mult(double mltp)
{
    TAssert(Matrix != NULL);

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        for (j = 0; j < NCols; j++)
            line[j] *= mltp;
    }
}

//-----------------------------------------------------------------------------
// Добавление матрицы
//-----------------------------------------------------------------------------
void DRectMem::Add(const CDMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(NRows == mt.Rows());
    TAssert(NCols == mt.Cols());

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        for (j = 0; j < NCols; j++)
            line[j] += mt(i, j);
    }

    return;
};
//-----------------------------------------------------------------------------
// Вычитание матрицы
//-----------------------------------------------------------------------------
void DRectMem::Sub(const CDMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(NRows == mt.Rows());
    TAssert(NCols == mt.Cols());

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        for (j = 0; j < NCols; j++)
            line[j] -= mt(i, j);
    }

    return;
};
//-----------------------------------------------------------------------------
// Добавление матрицы
//-----------------------------------------------------------------------------
void DRectMem::AddRect(const DRectMem& mt)
{
    TAssert(Matrix != NULL);
    TAssert(NRows == mt.NRows);
    TAssert(NCols == mt.NCols);

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        for (j = 0; j < NCols; j++)
            line[j] += mt.Matrix[i][j];
    }

    return;
};
//-----------------------------------------------------------------------------
// Вычитание матрицы
//-----------------------------------------------------------------------------
void DRectMem::SubRect(const DRectMem& mt)
{
    TAssert(Matrix != NULL);
    TAssert(NRows == mt.NRows);
    TAssert(NCols == mt.NCols);

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        for (j = 0; j < NCols; j++)
            line[j] -= mt.Matrix[i][j];
    }

    return;
};
//-----------------------------------------------------------------------------
// Добавление симметричной матрицы
//-----------------------------------------------------------------------------
void DRectMem::Add(const DSymmMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(NRows == mt.Dim());
    TAssert(NCols == mt.Dim());

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        line[i] += mt.Buffer[mt.Index(i, i)];
        for (j = i + 1; j < NCols; j++) {
            double a = mt.Buffer[mt.Index(i, j)];
            line[j] += a;
            Matrix[j][i] += a;
        }
    }

    return;
}

//-----------------------------------------------------------------------------
// Вычитание симметричной матрицы
//-----------------------------------------------------------------------------
void DRectMem::Sub(const DSymmMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(NRows == mt.Dim());
    TAssert(NCols == mt.Dim());

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        line[i] -= mt.Buffer[mt.Index(i, i)];
        for (j = i + 1; j < NCols; j++) {
            double a = mt.Buffer[mt.Index(i, j)];
            line[j] -= a;
            Matrix[j][i] -= a;
        }
    }

    return;
}

//-----------------------------------------------------------------------------
// Добавление верхней треугольной матрицы
//-----------------------------------------------------------------------------
void DRectMem::Add(const DUppMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(NRows == mt.Dim());
    TAssert(NCols == mt.Dim());

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        for (j = i; j < NCols; j++)
            line[j] += mt.Buffer[mt.Index(i, j)];
    }

    return;
}

//-----------------------------------------------------------------------------
// Вычитание верхней треугольной матрицы
//-----------------------------------------------------------------------------
void DRectMem::Sub(const DUppMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(NRows == mt.Dim());
    TAssert(NCols == mt.Dim());

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        for (j = i; j < NCols; j++)
            line[j] -= mt.Buffer[mt.Index(i, j)];
    }

    return;
}

//-----------------------------------------------------------------------------
// Добавление нижней треугольной матрицы
//-----------------------------------------------------------------------------
void DRectMem::Add(const DLowMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(NRows == mt.Dim());
    TAssert(NCols == mt.Dim());

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        for (j = 0; j < i + 1; j++)
            line[j] += mt.Buffer[mt.Index(j, i)];
    }

    return;
}

//-----------------------------------------------------------------------------
// Вычитание нижней треугольной матрицы
//-----------------------------------------------------------------------------
void DRectMem::Sub(const DLowMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(NRows == mt.Dim());
    TAssert(NCols == mt.Dim());

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        for (j = 0; j < i + 1; j++)
            line[j] -= mt.Buffer[mt.Index(j, i)];
    }

    return;
}

//-----------------------------------------------------------------------------
// Добавление диагональной матрицы
//-----------------------------------------------------------------------------
void DRectMem::Add(const DDiagMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(NRows == mt.Dim());
    TAssert(NCols == mt.Dim());

    unsigned i;

    for (i = 0; i < NRows; i++)
        Matrix[i][i] += mt.Vector[i];

    return;
}

//-----------------------------------------------------------------------------
// Вычитание диагональной матрицы
//-----------------------------------------------------------------------------
void DRectMem::Sub(const DDiagMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(NRows == mt.Dim());
    TAssert(NCols == mt.Dim());

    unsigned i;

    for (i = 0; i < NRows; i++)
        Matrix[i][i] -= mt.Vector[i];

    return;
}

//-----------------------------------------------------------------------------
// Формирование в результате перемножения двух матриц
//-----------------------------------------------------------------------------
void DRectMem::Mult(const DRectMem& m1, const DRectMem& m2)
{
    TAssert(m1.Matrix != NULL);
    TAssert(m2.Matrix != NULL);
    TAssert(m1.Matrix != Matrix);
    TAssert(m2.Matrix != Matrix);
    TAssert(m1.NCols == m2.NRows);

    Init(m1.NRows, m2.NCols);
    if (Matrix == NULL)
        return;

    unsigned i, j, k, f;
    double *p, *line, *lines;
    f = m2.NRows;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        for (j = 0; j < NCols; j++) {
            p = line + j;
            *p = 0.0;
            lines = m1.Matrix[i];
            for (k = 0; k < f; k++)
                *p += lines[k] * m2.Matrix[k][j];
        }
    }

    return;
}

//-----------------------------------------------------------------------------
// Умножение на матрицу справа
//-----------------------------------------------------------------------------
void DRectMem::MulRightRect(const DRectMem& mt)
{
    TAssert(mt.Matrix != NULL);
    TAssert(Matrix != NULL);
    TAssert(Matrix != mt.Matrix);
    TAssert(NCols == mt.NRows);

    DRectMem m;
    m.Init(NRows, mt.NCols);
    XChng(m);
    if (Matrix == NULL)
        return;

    Mult(m, mt);

    return;
}

//-----------------------------------------------------------------------------
// Умножение на матрицу слева
//-----------------------------------------------------------------------------
void DRectMem::MulLeftRect(const DRectMem& mt)
{
    TAssert(mt.Matrix != NULL);
    TAssert(Matrix != NULL);
    TAssert(Matrix != mt.Matrix);
    TAssert(mt.NCols == NRows);

    DRectMem m;
    m.Init(mt.NRows, NCols);
    XChng(m);
    if (Matrix == NULL)
        return;

    Mult(mt, m);

    return;
}

//-----------------------------------------------------------------------------
// Формирование в результате перемножения двух матриц
//-----------------------------------------------------------------------------
void DRectMem::Mult(const DRectMem& m, const DUppMatrix& mt)
{
    TAssert(m.Matrix != NULL);
    TAssert(m.Matrix != Matrix);
    TAssert(mt.Dim() == m.NCols);

    Init(m.NRows, mt.Dimension);
    if (Matrix == NULL)
        return;

    unsigned i, j, k;
    double *p, *line, *lines;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        lines = m.Matrix[i];
        for (j = 0; j < NCols; j++) {
            p = line + j;
            *p = 0.0;
            for (k = 0; k < j + 1; k++)
                *p += lines[k] * mt.Buffer[mt.Index(k, j)];
        }
    }

    return;
}

//-----------------------------------------------------------------------------
// Умножение на матрицу справа
//-----------------------------------------------------------------------------
void DRectMem::MulRight(const DUppMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(mt.Dim() == NCols);

    DRectMem m;
    m.Init(NRows, mt.Dimension);
    XChng(m);
    if (Matrix == NULL)
        return;

    Mult(m, mt);
    return;
}

//-----------------------------------------------------------------------------
// Формирование в результате перемножения двух матриц
//-----------------------------------------------------------------------------
void DRectMem::Mult(const DUppMatrix& mt, const DRectMem& m)
{
    TAssert(m.Matrix != Matrix);
    TAssert(m.Matrix != NULL);
    TAssert(mt.Cols() == m.NRows);

    Init(mt.Dimension, m.NCols);
    if (Matrix == NULL)
        return;

    unsigned i, j, k;
    double *p, *line;

    for (i = 0; i < mt.Dimension; i++) {
        line = Matrix[i];
        for (j = 0; j < NCols; j++) {
            p = line + j;
            *p = 0.0;
            for (k = i; k < mt.Dimension; k++)
                *p += mt.Buffer[mt.Index(i, k)] * m.Matrix[k][j];
        }
    }

    return;
}

//-----------------------------------------------------------------------------
// Умножение на матрицу слева
//-----------------------------------------------------------------------------
void DRectMem::MulLeft(const DUppMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(mt.Cols() == NRows);

    DRectMem m;
    m.Init(mt.Dimension, NCols);
    XChng(m);
    if (Matrix == NULL)
        return;

    Mult(mt, m);
}

//-----------------------------------------------------------------------------
// Формирование в результате перемножения двух матриц
//-----------------------------------------------------------------------------
void DRectMem::Mult(const DRectMem& m, const DLowMatrix& mt)
{
    TAssert(m.Matrix != NULL);
    TAssert(Matrix != m.Matrix);
    TAssert(mt.Dim() == m.NCols);

    Init(m.NRows, mt.Dimension);
    if (Matrix == NULL)
        return;

    unsigned i, j, k;
    double *p, *line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        for (j = 0; j < mt.Dimension; j++) {
            p = line + j;
            *p = 0.0;
            for (k = j; k < mt.Dimension; k++)
                *p += m.Matrix[i][k] * mt.Buffer[mt.Index(j, k)];
        }
    }

    return;
}

//-----------------------------------------------------------------------------
// Умножение на матрицу справа
//-----------------------------------------------------------------------------
void DRectMem::MulRight(const DLowMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(mt.Dim() == NCols);

    DRectMem m;
    m.Init(NRows, mt.Dimension);
    XChng(m);
    if (Matrix == NULL)
        return;

    Mult(m, mt);
    return;
}

//-----------------------------------------------------------------------------
// Формирование в результате перемножения двух матриц
//-----------------------------------------------------------------------------
void DRectMem::Mult(const DLowMatrix& mt, const DRectMem& m)
{
    TAssert(m.Matrix != NULL);
    TAssert(m.Matrix != Matrix);
    TAssert(mt.Cols() == m.NRows);

    Init(mt.Dimension, m.NCols);
    if (Matrix == NULL)
        return;

    unsigned i, j, k;
    double *p, *line;

    for (i = 0; i < mt.Dimension; i++) {
        line = Matrix[i];
        for (j = 0; j < NCols; j++) {
            p = line + j;
            *p = 0.0;
            for (k = 0; k < i + 1; k++)
                *p += mt.Buffer[mt.Index(k, i)] * m.Matrix[k][j];
        }
    }

    return;
}

//-----------------------------------------------------------------------------
// Умножение на матрицу слева
//-----------------------------------------------------------------------------
void DRectMem::MulLeft(const DLowMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(mt.Cols() == NRows);

    DRectMem m;
    m.Init(mt.Dimension, NCols);
    XChng(m);
    if (Matrix == NULL)
        return;

    Mult(mt, m);
    return;
}

//-----------------------------------------------------------------------------
// Формирование в результате перемножения двух матриц
//-----------------------------------------------------------------------------
void DRectMem::Mult(const DRectMem& m, const DDiagMatrix& mt)
{
    TAssert(m.Matrix != NULL);
    TAssert(m.Matrix != Matrix);
    TAssert(mt.Dim() == m.NCols);

    Init(m.NRows, mt.Dimension);
    if (Matrix == NULL)
        return;

    unsigned i, j;
    double* line;

    for (i = 0; i < NRows; i++) {
        line = Matrix[i];
        for (j = 0; j < mt.Dimension; j++)
            line[j] = m.Matrix[i][j] * mt.Vector[j];
    }

    return;
}

//-----------------------------------------------------------------------------
// Умножение на матрицу справа
//-----------------------------------------------------------------------------
void DRectMem::MulRight(const DDiagMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(mt.Dim() == NCols);

    DRectMem m;
    m.Init(NRows, mt.Dimension);
    XChng(m);
    if (Matrix == NULL)
        return;

    Mult(m, mt);
    return;
}

//-----------------------------------------------------------------------------
// Формирование в результате перемножения двух матриц
//-----------------------------------------------------------------------------
void DRectMem::Mult(const DDiagMatrix& mt, const DRectMem& m)
{
    TAssert(m.Matrix != NULL);
    TAssert(m.Matrix != Matrix);
    TAssert(mt.Cols() == m.NRows);

    Init(mt.Dimension, m.NCols);
    if (Matrix == NULL)
        return;

    unsigned i, j;
    double* line;

    for (i = 0; i < mt.Dimension; i++) {
        line = Matrix[i];
        for (j = 0; j < NCols; j++)
            line[j] = mt.Vector[i] * m.Matrix[i][j];
    }

    return;
}

//-----------------------------------------------------------------------------
// Умножение на матрицу слева
//-----------------------------------------------------------------------------
void DRectMem::MulLeft(const DDiagMatrix& mt)
{
    TAssert(Matrix != NULL);
    TAssert(mt.Cols() == NRows);

    DRectMem m;
    m.Init(mt.Dimension, NCols);
    XChng(m);
    if (Matrix == NULL)
        return;

    Mult(mt, m);

    return;
}

///////////////////////////////////////////////////////////////////////////////
