///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Класс, реализующий квадратную матрицу
///////////////////////////////////////////////////////////////////////////////
#include "Matrix.h"

#include <math.h>
///////////////////////////////////////////////////////////////////////////////
// Инициализация
//-----------------------------------------------------------------------------
void DSymmMatrix::Init(const DDiagMatrix& mt)
{
    TAssert(mt.Vector != NULL);
    DTriagMem::Init(mt.Dimension);
    if (Buffer == NULL)
        return;

    unsigned i;

    DTriagMem::Init(0.0);
    for (i = 0; i < Dimension; i++)
        Buffer[Index(i, i)] = mt.Vector[i];
}

//-----------------------------------------------------------------------------
// Инициализация
//-----------------------------------------------------------------------------
void DSymmMatrix::Init(const CDSymmMatrix& mt)
{
    DTriagMem::Init(mt.Dim());
    if (Buffer == NULL)
        return;

    unsigned i, j;

    for (i = 0; i < Dimension; i++) {
        for (j = i; j < Dimension; j++)
            Buffer[Index(i, j)] = mt(i, j);
    }
};
//-----------------------------------------------------------------------------
// Инициализация
//-----------------------------------------------------------------------------
void DSymmMatrix::Init(const DSymmMatrix& mt)
{
    DTriagMem::Init(mt.Dimension);
    if (Buffer == NULL)
        return;

    unsigned i, j;

    for (i = 0; i < Dimension; i++) {
        for (j = i; j < Dimension; j++)
            Buffer[Index(i, j)] = mt.Buffer[mt.Index(i, j)];
    }
};
//-----------------------------------------------------------------------------
// Оператор сложения
//-----------------------------------------------------------------------------
DSymmMatrix& DSymmMatrix::operator+=(const CDSymmMatrix& mt)
{
    TAssert(mt.Dim() == Dimension);
    TAssert(Buffer != NULL);

    unsigned i, j;

    for (i = 0; i < Dimension; i++) {
        for (j = i; j < Dimension; j++)
            Buffer[Index(i, j)] += mt(i, j);
    }

    return *this;
}

//-----------------------------------------------------------------------------
// Оператор сложения
//-----------------------------------------------------------------------------
DSymmMatrix& DSymmMatrix::operator+=(const DSymmMatrix& mt)
{
    TAssert(mt.Buffer != NULL);
    TAssert(mt.Dim() == Dimension);
    TAssert(Buffer != NULL);

    unsigned i, j;

    for (i = 0; i < Dimension; i++) {
        for (j = i; j < Dimension; j++)
            Buffer[Index(i, j)] += mt.Buffer[mt.Index(i, j)];
    }

    return *this;
}

//-----------------------------------------------------------------------------
// Оператор вычитания
//-----------------------------------------------------------------------------
DSymmMatrix& DSymmMatrix::operator-=(const CDSymmMatrix& mt)
{
    TAssert(mt.Dim() == Dimension);
    TAssert(Buffer != NULL);

    unsigned i, j;

    for (i = 0; i < Dimension; i++) {
        for (j = i; j < Dimension; j++)
            Buffer[Index(i, j)] -= mt(i, j);
    }

    return *this;
}

//-----------------------------------------------------------------------------
// Оператор вычитания
//-----------------------------------------------------------------------------
DSymmMatrix& DSymmMatrix::operator-=(const DSymmMatrix& mt)
{
    TAssert(mt.Buffer != NULL);
    TAssert(mt.Dim() == Dimension);
    TAssert(Buffer != NULL);

    unsigned i, j;

    for (i = 0; i < Dimension; i++) {
        for (j = i; j < Dimension; j++)
            Buffer[Index(i, j)] -= mt.Buffer[mt.Index(i, j)];
    }

    return *this;
}

//-----------------------------------------------------------------------------
// Оператор сложения
//-----------------------------------------------------------------------------
DSymmMatrix& DSymmMatrix::operator+=(const DDiagMatrix& mt)
{
    TAssert(mt.Dim() == Dimension);
    TAssert(Buffer != NULL);

    unsigned i;

    for (i = 0; i < Dimension; i++)
        Buffer[Index(i, i)] += mt.Vector[i];

    return *this;
}

//-----------------------------------------------------------------------------
// Оператор вычитания
//-----------------------------------------------------------------------------
DSymmMatrix& DSymmMatrix::operator-=(const DDiagMatrix& mt)
{
    TAssert(mt.Dim() == Dimension);
    TAssert(Buffer != NULL);

    unsigned i;

    for (i = 0; i < Dimension; i++)
        Buffer[Index(i, i)] -= mt.Vector[i];

    return *this;
}

//-----------------------------------------------------------------------------
// Оператор вычитания
//-----------------------------------------------------------------------------
bool DSymmMatrix::Expand(DUppMatrix& s, DDiagMatrix& d) const
{
    TAssert(Buffer != NULL);

    s.DTriagMem::Init(Dimension);
    d.Init(Dimension);

    s.Buffer[0] = 1.0;

    double d1;
    d.Vector[0] = d1 = Buffer[0];

    if (Dimension == 1)
        return true;

    double sum, element;
    unsigned i, j, k;

    for (j = 1; j < Dimension; ++j) {
        if (fabs(Buffer[j] * ZERO_ACCURACY) > fabs(d1))
            return false;
        s.Buffer[s.Index(0, j)] = Buffer[j] / d1;
    }

    for (i = 1; i < Dimension; ++i) {
        sum = 0;
        for (k = 0; k < i; ++k) {
            element = s.Buffer[s.Index(k, i)];
            sum += element * element * d.Vector[k];
        }
        d.Vector[i] = d1 = Buffer[Index(i, i)] - sum;
        for (j = i + 1; j < Dimension; ++j) {
            sum = 0;
            for (k = 0; k < i; ++k)
                sum += s.Buffer[s.Index(k, i)] * s.Buffer[s.Index(k, j)] * d.Vector[k];
            element = Buffer[Index(i, j)] - sum;
            if (fabs(element * ZERO_ACCURACY) > fabs(d1))
                return false;
            s.Buffer[s.Index(i, j)] = element / d1;
        }
    }

    for (i = 1; i < Dimension; ++i)
        s.Buffer[s.Index(i, i)] = 1.0;

    return true;
}

//-----------------------------------------------------------------------------
void DSymmMatrix::JacobiExpand(DDiagMatrix& S, DMatrix& R)
{
    unsigned n = Dimension;
    S.Init(n);
    R.Init(n, n);
    DSymmMatrix A(*this);

    static int i, j, k;
    int ns;
    double rc;
    static double rr, rs, ss, an1, an2, an3;
    static double x, y, z, c2, r1, s1, r2, r3, s2, t1, t2, t3, r8, r9, r6, r7;
    double d1;

    /* Function Body */
    rr = 1.e-20;
    rs = 1.0;
    ns = n - 1;

    for (i = 0; i < n; ++i) {
        for (j = 0; j < n; ++j) {
            if (i == j) {
                R(i, j) = 1.;
            }
            else {
                R(i, j) = 0.;
            }
        }
    }

    while (rs >= rr) {
        ss = 0.;
        s1 = 0.;
        for (i = 0; i < ns; ++i) {
            /* ============ */
            for (j = i + 1; j < n; ++j) {
                /* ========= */
                r1 = A(i, j);
                r2 = A(i, i);
                r3 = A(j, j);
                x = r1 * 2.;
                y = r2 - r3;
                z = sqrt(x * x + y * y);
                if (z <= rr) {
                    s2 = 0.;
                    c2 = 1.;
                }
                else {
                    c2 = y / z;
                    s2 = x / z;
                }
                rs = sqrt((1. - c2) * .5);
                d1 = sqrt((c2 + 1.) * .5);
                rc = fabs(d1);
                if (x < 0.) rc = -rc;
                t1 = (r2 + r3) * .5;
                t2 = (r2 - r3) * .5;
                t3 = r1;
                an1 = t1 + t2 * c2 + t3 * s2;
                an2 = t1 - t2 * c2 - t3 * s2;
                an3 = t3 * c2 - t2 * s2;
                for (k = 0; k < n; ++k) {
                    r8 = R(i, k);
                    r9 = R(j, k);
                    R(j, k) = -r8 * rs + r9 * rc;
                    R(i, k) = r8 * rc + r9 * rs;

                    r6 = (k > i) ? A(i, k) : A(k, i);
                    r7 = (k > j) ? A(j, k) : A(k, j);
                    if (k > i) {
                        A(i, k) = r6 * rc + r7 * rs;
                    }
                    else
                        A(k, i) = r6 * rc + r7 * rs;
                    if (k > j) {
                        A(j, k) = -r6 * rs + r7 * rc;
                    }
                    else
                        A(k, j) = -r6 * rs + r7 * rc;
                    A(i, j) = an3;
                    A(i, i) = an1;
                    A(j, j) = an2;
                }
            }
        }

        for (i = 0; i < n; ++i) {
            S(i, i) = A(i, i);
        }

        for (i = 0; i < ns; ++i) {
            s1 = (d1 = A(i, i), fabs(d1)) + s1;
            for (j = i + 1; j < n; ++j) {
                ss = (d1 = A(i, j), fabs(d1)) + ss;
            }
        }

        rs = ss / s1;
    }
}

//-----------------------------------------------------------------------------
// Обращение матрицы
//-----------------------------------------------------------------------------
bool DSymmMatrix::Inverse()
{
    TAssert(Buffer != NULL);

    DDiagMatrix d;
    DUppMatrix u;

    if (!Expand(u, d))
        return false;

    DLowMatrix l;
    DVector vc_unit(Dimension), x(Dimension), tmp(Dimension);

    l.Transpose(u);

    unsigned i, j;

    for (i = 0; i < Dimension; ++i) {
        vc_unit.Init(0.0);
        vc_unit.Vector[i] = 1.0;

        if (!TDMSolution(vc_unit, l, tmp)) {
            return false;
        }
        if (!TDMSolution(tmp, d, x)) {
            return false;
        }
        if (!TDMSolution(x, u, vc_unit)) {
            return false;
        }
        for (j = i; j < Dimension; ++j)
            Buffer[Index(i, j)] = vc_unit.Vector[j];
    }

    return true;
}

//-----------------------------------------------------------------------------
// Тройное умножение матриц
//-----------------------------------------------------------------------------
DSymmMatrix& DSymmMatrix::RectSymmTRecmt(const DRectMem& A, const DSymmMatrix& B)
{
    TAssert(A.Matrix != NULL);
    TAssert(B.Buffer != NULL);
    TAssert(A.NCols == B.Dimension);

    DTriagMem::Init(A.NRows);

    double* p_mt3 = new double[B.Dimension];

    double element;

    unsigned i, j, k;

    for (i = 0; i < A.NRows; i++) {
        for (j = 0; j < B.Dimension; j++) {
            element = 0.0;
            for (k = 0; k < B.Dimension; k++)
                element += A.Matrix[i][k] * B(j, k);
            p_mt3[j] = element;
        }

        for (j = i; j < A.NRows; j++) {
            element = 0.0;
            for (k = 0; k < B.Dimension; k++)
                element += p_mt3[k] * A.Matrix[j][k];
            Buffer[Index(i, j)] = element;
        }
    }

    delete[] p_mt3;

    return *this;
}

//-----------------------------------------------------------------------------
// Тройное умножение матриц
//-----------------------------------------------------------------------------
DSymmMatrix& DSymmMatrix::RectSymmTRecmt(const CDSqrMatrix& A, const DSymmMatrix& B)
{
    TAssert(B.Buffer != NULL);
    TAssert(A.Dim() == B.Dimension);

    DTriagMem::Init(A.Dim());

    double* p_mt3 = new double[B.Dimension];

    double element;

    unsigned i, j, k;

    for (i = 0; i < A.Dim(); i++) {
        for (j = 0; j < B.Dimension; j++) {
            element = 0.0;
            for (k = 0; k < B.Dimension; k++)
                element += A(i, k) * B(j, k);
            p_mt3[j] = element;
        }

        for (j = i; j < A.Dim(); j++) {
            element = 0.0;
            for (k = 0; k < B.Dimension; k++)
                element += p_mt3[k] * A(j, k);
            Buffer[Index(i, j)] = element;
        }
    }

    delete[] p_mt3;

    return *this;
}

//-----------------------------------------------------------------------------
// Тройное умножение матриц
//-----------------------------------------------------------------------------
DSymmMatrix& DSymmMatrix::TRectSymmRecmt(const DRectMem& A, const DSymmMatrix& B)
{
    TAssert(A.Matrix != NULL);
    TAssert(B.Buffer != NULL);
    TAssert(A.NRows == B.Dimension);

    DTriagMem::Init(A.NCols);

    double* p_mt3 = new double[B.Dimension];

    double element;

    unsigned i, j, k;

    for (i = 0; i < A.NCols; i++) {
        for (j = 0; j < B.Dimension; j++) {
            element = 0.0;
            for (k = 0; k < B.Dim(); k++)
                element += A.Matrix[k][i] * B(j, k);
            p_mt3[j] = element;
        }

        for (j = i; j < A.NCols; j++) {
            element = 0.0;
            for (k = 0; k < B.Dimension; k++)
                element += p_mt3[k] * A.Matrix[k][j];
            Buffer[Index(i, j)] = element;
        }
    }

    delete[] p_mt3;

    return *this;
}

//-----------------------------------------------------------------------------
// Тройное умножение матриц
//-----------------------------------------------------------------------------
DSymmMatrix& DSymmMatrix::TRectSymmRecmt(const CDSqrMatrix& A, const DSymmMatrix& B)
{
    TAssert(B.Buffer != NULL);
    TAssert(A.Dim() == B.Dimension);

    DTriagMem::Init(A.Dim());

    double* p_mt3 = new double[B.Dimension];

    double element;

    unsigned i, j, k;

    for (i = 0; i < A.Dim(); i++) {
        for (j = 0; j < B.Dimension; j++) {
            element = 0.0;
            for (k = 0; k < B.Dim(); k++)
                element += A(k, i) * B(j, k);
            p_mt3[j] = element;
        }

        for (j = i; j < A.Dim(); j++) {
            element = 0.0;
            for (k = 0; k < B.Dimension; k++)
                element += p_mt3[k] * A(k, j);
            Buffer[Index(i, j)] = element;
        }
    }

    delete[] p_mt3;

    return *this;
}

//-----------------------------------------------------------------------------
// Конструирование матрицы путем умножения на траспонированную 
//-----------------------------------------------------------------------------
DSymmMatrix& DSymmMatrix::RectTRecmt(const DRectMem& A)
{
    TAssert(A.Matrix != NULL);

    DTriagMem::Init(A.NRows);

    unsigned i, j, k;

    double sum;

    for (i = 0; i < A.NRows; i++) {
        for (j = i; j < A.NRows; j++) {
            sum = 0.0;
            for (k = 0; k < A.NCols; ++k)
                sum += A.Matrix[i][k] * A.Matrix[j][k];
            Buffer[Index(i, j)] = sum;
        }
    }

    return *this;
}

//-----------------------------------------------------------------------------
// Реализация запроса значений строки
//-----------------------------------------------------------------------------
void DSymmMatrix::GetLine(unsigned index, double* line) const
{
    TAssert(Buffer != NULL);
    TAssert(line != NULL);
    TAssert(index < Dimension);

    unsigned i;
    for (i = 0; i < index; i++)
        line[i] = *(Buffer + Index(i, index));
    memcpy(line + index, Buffer + Index(index, index), (Dimension - index) * sizeof(double));
}

//-----------------------------------------------------------------------------
// Реализация запроса значений элементов в виде массива
//-----------------------------------------------------------------------------
void DSymmMatrix::GetArray(double* array) const
{
    TAssert(Buffer != NULL);
    TAssert(array != NULL);

    unsigned i, j;

    for (i = 0; i < Dimension; i++) {
        memcpy(array, Buffer + Index(i, i), (Dimension - i) * sizeof(double));
        double* col = array + Dimension;
        for (j = i + 1; j < Dimension; j++, col += Dimension)
            *col = array[j - i];
        array += Dimension + 1;
    }
}

///////////////////////////////////////////////////////////////////////////////
// Тройное умножение
///////////////////////////////////////////////////////////////////////////////
double TRecmtSymmRecmt(const DVector& A, const DSymmMatrix& B)
{
    TAssert(A.Vector != NULL);
    TAssert(B.Buffer != NULL);
    TAssert(A.Dimension == B.Dimension);

    double* p_mt3 = new double[B.Dimension];

    double element;

    unsigned j, k;

    for (j = 0; j < B.Dimension; j++) {
        element = 0.0;
        for (k = 0; k < B.Dimension; k++)
            element += A.Vector[k] * B(j, k);
        p_mt3[j] = element;
    }

    element = 0.0;
    for (k = 0; k < B.Dimension; k++)
        element += p_mt3[k] * A.Vector[k];

    delete[] p_mt3;

    return element;
}

///////////////////////////////////////////////////////////////////////////////
