///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Определение 3- и 6-мерных декартовых координат и матриц преобразования
////////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL__DVECTORS_H
#define BNOTOOL__DVECTORS_H
////////////////////////////////////////////////////////////////////////////////
#include <string.h>
#include <stdio.h>
#include <math.h>
////////////////////////////////////////////////////////////////////////////////
// Предопределение класса матрицы 3x3
class TransMtx;

////////////////////////////////////////////////////////////////////////////////
// Трехмерный вектор в декартовых координатах
////////////////////////////////////////////////////////////////////////////////
class Vect3
{
public:
    Vect3(double x_ = 0, double y_ = 0, double z_ = 0)
    {
        Init(x_, y_, z_);
    };

    // Методы инициализации
    Vect3& Init(double x_ = 0, double y_ = 0, double z_ = 0)
    {
        x = x_;
        y = y_;
        z = z_;
        return *this;
    };

    // Вычисление длины вектора
    double Len() const
    {
        return sqrt(x * x + y * y + z * z);
    };

    // Квадрат длины вектора
    double Len2() const
    {
        return x * x + y * y + z * z;
    };

    // Умножение вектора на матрицу справа
    // (т.е. результат - вектор-строка)
    Vect3& MulRight(const TransMtx& mtx);
    // Умножение вектора на матрицу слева
    // (т.е. результат - вектор-столбец)
    Vect3& MulLeft(const TransMtx& mtx)
    {
        *this *= mtx;
        return *this;
    };
    Vect3& operator*=(const TransMtx& mtx);

    // Операторы
    // Изменение вектора на противоположный
    Vect3& operator-()
    {
        x = -x;
        y = -y;
        z = -z;
        return *this;
    };
    // Сравнение на равенство
    bool operator==(const Vect3& p) const
    {
        return (x == p.x) && (y == p.y) && (z == p.z);
    };
    // Присвоение всем координатам одинакового числа
    Vect3& operator=(double t)
    {
        x = t;
        y = t;
        z = t;
        return *this;
    };
    // Доступ к элементу по индексу
    double& operator[](int i)
    {
        return (i == 0 ? x : (i == 1 ? y : z));
    };
    // Получение компоненты вектора по индексу i
    double operator[](int i) const
    {
        return (i == 0 ? x : (i == 1 ? y : z));
    };
    // Сложение векторов 
    Vect3& operator+=(const Vect3& v0)
    {
        x += v0.x;
        y += v0.y;
        z += v0.z;
        return *this;
    };

    // Вычитание векторов 
    Vect3& operator-=(const Vect3& v0)
    {
        x -= v0.x;
        y -= v0.y;
        z -= v0.z;
        return *this;
    };
    // Умножение вектора на число
    Vect3& operator*=(double mul)
    {
        x *= mul;
        y *= mul;
        z *= mul;
        return *this;
    };
    // Деление вектора на число
    Vect3& operator/=(double div)
    {
        x /= div;
        y /= div;
        z /= div;
        return *this;
    };
    // Скалярное умножение
    double operator*(const Vect3& v0) const
    {
        return x * v0.x + v2 * v0.y + z * v0.z;
    };

    // Векторное умножение
    Vect3& MulVect(const Vect3& v0)
    {
        double x_ = x, y_ = y, z_ = z;
        x = y_ * v0.z - z_ * v0.y;
        y = z_ * v0.x - x_ * v0.z;
        z = x_ * v0.y - y_ * v0.x;
        return *this;
    };

    // Вывод на печать
    void print(const char* s, FILE* out = stdout) const
    {
        fprintf(out, "%s%23.16e,%23.16e,%23.16e\n", s, x, y, z);
    };

    // Сами координаты
    void Normalize()
    {
        double a = 1.0 / Len();
        x *= a;
        y *= a;
        z *= a;
    }

    // Сами координаты
    union
    {
        double x;
        double v1;
    };

    union
    {
        double y;
        double v2;
    };

    union
    {
        double z;
        double v3;
    };
};

////////////////////////////////////////////////////////////////////////////////
// Шестимерный вектор в декартовых координатах (координаты + скорости)
////////////////////////////////////////////////////////////////////////////////
class Vect6 : public Vect3
{
public:
    // Конструкторы
    Vect6(double x_ = 0, double y_ = 0, double z_ = 0,
          double vx_ = 0, double vy_ = 0, double vz_ = 0)
    {
        Init(x_, y_, z_, vx_, vy_, vz_);
    };

    Vect6(const Vect3& pt, double vx_, double vy_, double vz_)
    {
        Init(pt, vx_, vy_, vz_);
    };

    Vect6(const Vect3& pt, const Vect3& v)
    {
        Init(pt, v);
    };

    // Методы инициализации
    Vect6& Init(double x_ = 0, double y_ = 0, double z_ = 0,
                double vx_ = 0, double vy_ = 0, double vz_ = 0)
    {
        Vect3::Init(x_, y_, z_);
        vx = vx_;
        vy = vy_;
        vz = vz_;
        return *this;
    };

    Vect6& Init(const Vect3& pt, double vx_, double vy_, double vz_)
    {
        Vect3::Init(pt.x, pt.y, pt.z);
        vx = vx_;
        vy = vy_;
        vz = vz_;
        return *this;
    };

    Vect6& Init(const Vect3& pt, const Vect3& v)
    {
        Vect3::Init(pt.x, pt.y, pt.z);
        vx = v.x;
        vy = v.y;
        vz = v.z;
        return *this;
    };

    // Вычисление модуля скорости
    double LenV() const
    {
        return sqrt(vx * vx + vy * vy + vz * vz);
    };
    // Вычисление квадрата модуля скорости
    double LenV2() const
    {
        return vx * vx + vy * vy + vz * vz;
    };

    // Изменение вектора на противоположный
    Vect6& operator-()
    {
        x = -x;
        y = -y;
        z = -z;
        vx = -vx;
        vy = -vy;
        vz = -vz;
        return *this;
    };
    // Сравнение на равенство
    bool operator==(const Vect6& p) const
    {
        return (x == p.x) && (y == p.y) && (z == p.z) && (vx == p.vx) && (vy == p.vy)
            && (vz == p.vz);
    };
    // Доступ к элементу по индексу
    double& operator[](int i)
    {
        return i == 0 ? x : i == 1 ? y : i == 2 ? z : i == 3 ? vx : i == 4 ? vy : vz;
    };
    // Получение компоненты вектора по индексу i
    double operator[](int i) const
    {
        return i == 0 ? x : i == 1 ? y : i == 2 ? z : i == 3 ? vx : i == 4 ? vy : vz;
    };

    // Извлечение скорости, как трехмерного вектора
    Vect3& GetVelocity(Vect3& v) const
    {
        v.Init(vx, vy, vz);
        return v;
    };
    // Установка компонент скорости из трехмерного вектора
    void SetVelocity(const Vect3& v)
    {
        vx = v.x;
        vy = v.y;
        vz = v.z;
    };

    // Вывод на печать
    void print(const char* s, FILE* out = stdout) const;

    // Компоненты скоростей
    double vx, vy, vz;
};

////////////////////////////////////////////////////////////////////////////////
class TransMtx
{
public:
    // Конструкторы
    TransMtx()
    {
    };
    // Инициализация из массива
    TransMtx(const double* p)
    {
        Init(p);
    };

    // Иницализация (элементы в массиве p вытянуты по строкам
    TransMtx& Init(const double* p)
    {
        memcpy(mtx, p, sizeof(double) * 9);
        return *this;
    };
    // Формирование единичной матрицы
    TransMtx& Init1()
    {
        mtx[0] = 1.0;
        mtx[1] = 0.0;
        mtx[2] = 0.0;
        mtx[3] = 0.0;
        mtx[4] = 1.0;
        mtx[5] = 0.0;
        mtx[6] = 0.0;
        mtx[7] = 0.0;
        mtx[8] = 1.0;
        return *this;
    };

    // Оператор получения значения элемента матрицы 
    double operator()(int row, int col) const
    {
        return mtx[row * 3 + col];
    };
    // Оператор доступа к элементу 
    double& operator()(int row, int col)
    {
        return mtx[row * 3 + col];
    };
    // Оператор доступа к элементу 
    // (индексация по массиву строк)
    double& operator[](int pos)
    {
        return mtx[pos];
    };

    double operator[](int pos) const
    {
        return mtx[pos];
    };

    // Транспонирование
    TransMtx& Transp();
    // Транспонирование, результат в m
    // Исходная матрица не меняется
    void Transp(TransMtx& m) const;

    // Умножение на матрицу справа: A = A * B
    TransMtx& operator*=(double a);
    // Умножение на матрицу справа: A = A * B
    TransMtx& operator*=(const TransMtx& mRight);

    TransMtx& MulRight(const TransMtx& mRight)
    {
        *this *= mRight;
        return *this;
    };
    // Умножение на матрицу слева: A = B * A
    TransMtx& MulLeft(const TransMtx& mLeft);
    // Умножение матрицы на матрицу m
    void mul(const TransMtx& m, TransMtx& res) const;
    // Определитель матрицы
    double Det() const;

    // Сложение матриц
    TransMtx& operator+=(const TransMtx& mtx);
    // Вычитание матриц
    TransMtx& operator-=(const TransMtx& mtx);
    // Присвоение числа всем элементам матрицы
    TransMtx& operator =(double val);
    // Оператор сравнения матриц на совпадение
    bool operator==(const TransMtx& mtx) const;

    // Вывод на печать
    void print(const char* name, FILE* out = stdout) const;

    // Получение указателя на содержимое матрицы (по строкам)
    double* GetPtr()
    {
        return mtx;
    };
    // Получение указателя на содержимое матрицы (по строкам)
    const double* GetConstPtr() const
    {
        return mtx;
    };
    // Сохранение содержимого матрицы по указанному указателю
    void Save(double* p) const
    {
        memcpy(p, mtx, sizeof(double) * 9);
    };

private:
    double mtx[9];
};

////////////////////////////////////////////////////////////////////////////////
#endif
