///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация численного прогноза в прямоугольных координатах
///////////////////////////////////////////////////////////////////////////////
#include "DigPredict.h"
///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
//-----------------------------------------------------------------------------
DigPredict::DigPredict()
{
    Integrator = NULL;
    Adapter = NULL;
    KOProp = NULL;
    FlagInit = false;
}

//-----------------------------------------------------------------------------
// Установка внешнего интегратора
//-----------------------------------------------------------------------------
void DigPredict::SetIntegrator(IODESolver* slv)
{
    Integrator = slv;
    Integrator->SetSystem(&System);
}

//-----------------------------------------------------------------------------
// Установка внешней модели КО
//-----------------------------------------------------------------------------
void DigPredict::SetKOProp(IKOProp* prop)
{
    KOProp = prop;
    KOProp->Init(MMode);
};
//-----------------------------------------------------------------------------
// Установка начальных условий для интегрирования
//-----------------------------------------------------------------------------
void DigPredict::SetInitPoint(const PhasePoint6D& src)
{
    TAssert(Integrator != NULL);
    TAssert(Adapter != NULL);
    TAssert(KOProp != NULL);

    // Инициализация системы уравнений
    T0 = src.T;
    System.Init(T0, MMode, *KOProp, *Adapter);

    // Конвертируем нач. условия в J2000
    PhasePoint6D j2000;
    J2000Frame f2000;
    j2000.SetFrame(f2000);
    j2000.Assign(src);


    double Y0[6];

    Y0[0] = j2000.CoordsVel.x;
    Y0[1] = j2000.CoordsVel.y;
    Y0[2] = j2000.CoordsVel.z;
    Y0[3] = j2000.CoordsVel.vx;
    Y0[4] = j2000.CoordsVel.vy;
    Y0[5] = j2000.CoordsVel.vz;

    Integrator->SetInitial(0.0, Y0);

    FlagInit = true;
}

//-----------------------------------------------------------------------------
// "Переразгон" прогноза от текущей точки
//-----------------------------------------------------------------------------
void DigPredict::ReRun()
{
    PhasePoint6D tmp;
    GetDestPoint(tmp);
    SetInitPoint(tmp);
}

//-----------------------------------------------------------------------------
// Повторная инициализация прогноза
//-----------------------------------------------------------------------------
//2012 void DigPredict::ReInit()
//  {
//  Integrator->SetInitial(0.0, Y0);
//  }
//-----------------------------------------------------------------------------
// Собственно прогнозирование
//-----------------------------------------------------------------------------
bool DigPredict::Run(const TimeJD& t)
{
    TAssert(FlagInit);
    // Вычисляем точку, до которой должно идти интегрирование
    // (в тысячах секунд)
    double dt = (t - T0).GetDouble() * 86.4;
    return Integrator->Run(dt);
}

//-----------------------------------------------------------------------------
// Получение конечной точки
//-----------------------------------------------------------------------------
void DigPredict::GetDestPoint(PhasePoint6D& dest)
{
    TAssert(FlagInit);

    dest.T = T0;
    dest.T += Integrator->GetX() / 86.4;

    J2000Frame f2000;
    dest.SetFrame(f2000);

    const double* y = Integrator->GetSolution();

    dest.CoordsVel.x = y[0];
    dest.CoordsVel.y = y[1];
    dest.CoordsVel.z = y[2];
    dest.CoordsVel.vx = y[3];
    dest.CoordsVel.vy = y[4];
    dest.CoordsVel.vz = y[5];
}

///////////////////////////////////////////////////////////////////////////////
// Стандартный численный прогноз:
// - интегратор В.А.Степаньянца
// - адаптер координат, не учитвающий нутацию и прецессию
// - КО с постоянным балл. коэффициентом и коэф. светового давления
///////////////////////////////////////////////////////////////////////////////
// Допустимая погрешность на шаге интегрирования (тыс. км, тыс. секунд)
//-----------------------------------------------------------------------------
//static double Eps[] =
//  {
//  1.e-9, 1.e-9, 1.e-9, 1.e-9, 1.e-9, 1.e-9
//  };
//-----------------------------------------------------------------------------
// Конструктор по умолчанию
//-----------------------------------------------------------------------------
DigStPredict::DigStPredict()
    :
    DigPredict(),
    m_eps(1E-9),
    m_stepinit(10E-3),
    m_stepmin(.1E-3)
{
    // Установка интегратора
    DigPredict::SetIntegrator(&SIntegrator);
    // Установка адаптера
    DigPredict::SetAdapter(&SAdapter);
    // Установка модели КО
    DigPredict::SetKOProp(&KOModel);
    // Установка точностей по умолчанию
    SetAccuracy(m_eps, m_stepinit, m_stepmin, false);
}

DigStPredict::DigStPredict(double eps, double stepinit, double stepmin, bool precItrf)
    :
    DigPredict(),
    m_eps(eps),
    m_stepinit(stepinit),
    m_stepmin(stepmin)
{
    // Установка интегратора
    DigPredict::SetIntegrator(&SIntegrator);
    // Установка адаптера
    DigPredict::SetAdapter(&SAdapter);
    // Установка модели КО
    DigPredict::SetKOProp(&KOModel);
    // Установка точностей по умолчанию
    SetAccuracy(m_eps, m_stepinit, m_stepmin, false);
}

void DigStPredict::SetKOParams(double kb, double kappa)
{
    KOModel.Kb = kb;
    KOModel.Kappa = kappa;
}

void DigStPredict::SetAccuracy(double eps, double stepinit, double stepmin, bool precItrf)
{
    double epsd[6];

    for (int i = 0; i < 6; i++) {
        epsd[i] = eps;
    }

    SIntegrator.SetParams(epsd, stepinit /* нач. шаг */, stepmin /* мин. шаг*/);

    System.SetPrecItrf(precItrf);
}
