///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация численного прогноза в прямоугольных координатах
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_DIGPREDICT_H
#define BNOTOOL_DIGPREDICT_H
///////////////////////////////////////////////////////////////////////////////
#include "IPredict.h"
#include "IODE.h"
#include "ODESystem.h"
#include "ODEStpnSolver.h"
#include "KOConst.h"

///////////////////////////////////////////////////////////////////////////////
// Численный прогноз с произвольным интегратором, адаптером и моделью КО
///////////////////////////////////////////////////////////////////////////////
class DigPredict : public IPredict
{
public:
    DigPredict();

    virtual ~DigPredict()
    {
    };

    // Методы настройки

    // Установка внешнего интегратора
    void SetIntegrator(IODESolver* slv);
    // Установка внешнего адаптера координат
    void SetAdapter(AccAdapter* acc)
    {
        Adapter = acc;
    };
    // Установка внешней модели КО
    void SetKOProp(IKOProp* prop);
    // Собственно прогнозирование
    virtual bool Run(const TimeJD& t);
    // Получение конечной точки
    virtual void GetDestPoint(PhasePoint6D& dest);

    // Установка начальных условий для интегрирования
    void SetInitPoint(const PhasePoint6D& src);
    // "Переразгон" прогноза от текущей точки
    virtual void ReRun();

    // Используемая модель движения
    MotionMode MMode;

protected:
    // Система уравнений
    ODEPhaseSystem System;
    // Интегратор
    IODESolver* Integrator;
    // Используемый адаптер ускорений
    AccAdapter* Adapter;
    // Модель космического объекта
    IKOProp* KOProp;
    // Флаг наличия начальных условий
    bool FlagInit;
    // Момент времени, от которого идет отсчет
    TimeJD T0;
};

///////////////////////////////////////////////////////////////////////////////
// Стандартный численный прогноз:
// - интегратор В.А.Степаньянца
// - адаптер координат, не учитвающий нутацию и прецессию
// - КО с постоянным балл. коэффициентом и коэф. светового давления
///////////////////////////////////////////////////////////////////////////////
class DigStPredict : public DigPredict
{
public:
    DigStPredict();
    DigStPredict(double eps, double stepinit, double stepmin, bool precItrf);

    // Установка параметров КО
    void SetKOParams(double kb, double kappa);
    // Установка точности интегратора и шагов
    void SetAccuracy(double eps, double stepinit, double stepmin, bool precItrf);

    // Интегратор
    ODEStpnSolver SIntegrator;
    // Модель КО
    KOConst KOModel;

protected:
    // Адаптер
    AccAdapter SAdapter;

private:
    // Запрет установки внешнего интегратора
    void SetIntegrator(IODESolver* slv);
    // Запрет установки внешнего адаптера координат
    void SetAdapter(AccAdapter* acc);
    // Запрет установки внешней модели КО
    void SetKOProp(IKOProp* prop);

    //точность интегрирования
    double m_eps;
    //начальный шаг интегрирования
    double m_stepinit;
    //минимальный шаг интегрирования
    double m_stepmin;
};

#endif
