///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация численного прогноза с уравнениями в вариациях 
// в прямоугольных координатах
///////////////////////////////////////////////////////////////////////////////
#include "DigPredictVar.h"
///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
//-----------------------------------------------------------------------------
DigPredictVar::DigPredictVar()
{
    Integrator = NULL;
    Adapter = NULL;
    KOProp = NULL;
    FlagInit = false;
}

//-----------------------------------------------------------------------------
// Установка внешнего интегратора
//-----------------------------------------------------------------------------
void DigPredictVar::SetIntegrator(IODESolver* slv)
{
    Integrator = slv;
    Integrator->SetSystem(&System);
}

//-----------------------------------------------------------------------------
// Установка внешней модели КО
//-----------------------------------------------------------------------------
void DigPredictVar::SetKOProp(IKOPropVar* prop)
{
    KOProp = prop;
    KOProp->Init(MMode);
    System.SetKO(*KOProp);
};
//-----------------------------------------------------------------------------
// Установка начальных условий для интегрирования
//-----------------------------------------------------------------------------
void DigPredictVar::SetInitPoint(const PhasePoint6D& src, const DMatrix* vars)
{
    TAssert(Integrator != NULL);
    TAssert(Adapter != NULL);
    TAssert(KOProp != NULL);

    // Инициализация системы уравнений
    T0 = src.T;
    System.Init(T0, MMode, *Adapter);

    // Конвертируем нач. условия в J2000
    double* y;
    PhasePoint6D j2000;
    J2000Frame f2000;
    j2000.SetFrame(f2000);
    j2000.Assign(src);

    y = new double[6 + 36 + 6 * KOProp->GetNVars()];

    y[0] = j2000.CoordsVel.x;
    y[1] = j2000.CoordsVel.y;
    y[2] = j2000.CoordsVel.z;
    y[3] = j2000.CoordsVel.vx;
    y[4] = j2000.CoordsVel.vy;
    y[5] = j2000.CoordsVel.vz;

    int i, j;

    if (vars == NULL) {
        // Начальные условия по вариациям не были заданы явно
        for (i = 0; i < 6; i++) {
            for (j = 0; j < 6; j++)
                y[6 + i * 6 + j] = (i == j) ? 1.0 : 0.0;
        }

        for (i = 0; i < KOProp->GetNVars(); i++) {
            for (j = 0; j < 6; j++)
                y[6 + 36 + i * 6 + j] = 0.0;
        }
    }
    else {
        // Начальные условия по вариациям заданы явно
        TAssert(vars->Cols() == 6);
        TAssert((int)vars->Rows() == 6 + KOProp->GetNVars());
        for (i = 0; i < (int)vars->Rows(); i++) {
            for (j = 0; j < 6; j++)
                y[6 + i * 6 + j] = (*vars)(i, j);
        }
    }

    Integrator->SetInitial(0.0, y);

    FlagInit = true;
    delete[] y;
}

//-----------------------------------------------------------------------------
// "Переразгон" прогноза от текущей точки
//-----------------------------------------------------------------------------
void DigPredictVar::ReRun()
{
    PhasePoint6D v;
    DMatrix m;

    GetDestPoint(v);
    GetDestMatrix(m);
    SetInitPoint(v, &m);
}

//-----------------------------------------------------------------------------
// Собственно прогнозирование
//-----------------------------------------------------------------------------
bool DigPredictVar::Run(const TimeJD& t)
{
    TAssert(FlagInit);
    // Вычисляем точку, до которой должно идти интегрирование
    // (в тысячах секунд)
    double dt = (t - T0).GetDouble() * 86.4;
    return Integrator->Run(dt);
}

//-----------------------------------------------------------------------------
// Получение конечной точки
//-----------------------------------------------------------------------------
void DigPredictVar::GetDestPoint(PhasePoint6D& dest)
{
    TAssert(FlagInit);

    dest.T = T0;
    dest.T += Integrator->GetX() / 86.4;

    J2000Frame f2000;
    dest.SetFrame(f2000);

    const double* y = Integrator->GetSolution();

    dest.CoordsVel.x = y[0];
    dest.CoordsVel.y = y[1];
    dest.CoordsVel.z = y[2];
    dest.CoordsVel.vx = y[3];
    dest.CoordsVel.vy = y[4];
    dest.CoordsVel.vz = y[5];
}

//-----------------------------------------------------------------------------
// Получение матрицы частных производных
//-----------------------------------------------------------------------------
void DigPredictVar::GetDestMatrix(DMatrix& m)
{
    TAssert(FlagInit);

    int i, j, n;

    n = 6 + KOProp->GetNVars();
    m.Init(n, n);
    //m.Init(0);
    const double* y = Integrator->GetSolution() + 6;

    for (i = 0; i < n; i++) {
        for (j = 0; j < 6; j++) {
            m(i, j) = y[6 * i + j];
        }
        for (j = 6; j < 6 + KOProp->GetNVars(); j++) {
            m(i, j) = i == j ? 1 : 0;
        }
    }
}


///////////////////////////////////////////////////////////////////////////////
// Стандартный численный прогноз уравнений в вариациях:
// - интегратор В.А.Степаньянца
// - адаптер координат, не учитвающий нутацию и прецессию
// - КО с постоянным балл. коэффициентом и коэф. светового давления
///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
//-----------------------------------------------------------------------------
DigStPredictVar::DigStPredictVar()
    :
    DigPredictVar(),
    m_eps(1E-9),
    m_stepinit(10E-3),
    m_stepmin(.1E-3)
{
    // Установка адаптера
    DigPredictVar::SetAdapter(&SAdapter);
    // Установка модели КО
    DigPredictVar::SetKOProp(&KOModel);
    // Установка интегратора
    DigPredictVar::SetIntegrator(&SIntegrator);
    // Установка точностей по умолчанию
    SetAccuracy(m_eps, m_stepinit, m_stepmin, false);
}

DigStPredictVar::DigStPredictVar(double eps, double stepinit, double stepmin, bool precItrf)
    :
    DigPredictVar(),
    m_eps(eps),
    m_stepinit(stepinit),
    m_stepmin(stepmin)
{
    // Установка адаптера
    DigPredictVar::SetAdapter(&SAdapter);
    // Установка модели КО
    DigPredictVar::SetKOProp(&KOModel);
    // Установка интегратора
    DigPredictVar::SetIntegrator(&SIntegrator);
    // Установка точностей по умолчанию
    SetAccuracy(m_eps, m_stepinit, m_stepmin, precItrf);
}

void DigStPredictVar::SetKOParams(double kb, double kappa, bool iskbvar, bool iskappavar)
{
    KOModel.Kb = kb;
    KOModel.Kappa = kappa;

    if ((iskbvar != KOModel.IsKbVar) || (iskappavar != KOModel.IsKappaVar)) {
        KOModel.IsKbVar = iskbvar;
        KOModel.IsKappaVar = iskappavar;
        // Переустанавливаем систему ур-ний в интеграторе,
        // поскольку ее размерность поменялась
        SIntegrator.SetSystem(&System);
        // Установка точностей
        SetAccuracy(m_eps, m_stepinit, m_stepmin, System.getPrecItrf());
    }
}

void DigStPredictVar::SetAccuracy(double eps, double stepinit, double stepmin, bool precItrf)
{
    double* epsd = new double[6 + 36 + KOModel.GetNVars() * 6];

    for (int i = 0; i < 6; i++) {
        epsd[i] = eps;
    }

    for (int i = 0; i < 36 + KOModel.GetNVars() * 6; i++) {
        epsd[i + 6] = 1.e30;
    }

    SIntegrator.SetParams(epsd, stepinit /* нач. шаг */, stepmin /* мин. шаг*/);

    delete[] epsd;

    System.SetPrecItrf(precItrf);
}
