///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация численного прогноза с уравнениями в вариациях 
// в прямоугольных координатах
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_DIGPREDICTVAR_H
#define BNOTOOL_DIGPREDICTVAR_H
///////////////////////////////////////////////////////////////////////////////
#include "IPredict.h"
#include "IODE.h"
#include "ODESystem.h"
#include "ODEStpnSolver.h"
#include "KOConstVar.h"

///////////////////////////////////////////////////////////////////////////////
// Численный прогноз уравнений в вариациях с произвольным интегратором, 
// адаптером и моделью КО
///////////////////////////////////////////////////////////////////////////////
class DigPredictVar : public IPredictVar
{
public:
    // Конструктор по умолчанию
    DigPredictVar();
    // Деструктор
    virtual ~DigPredictVar()
    {
    };

    // Методы настройки

    // Установка внешнего интегратора
    void SetIntegrator(IODESolver* slv);
    // Установка внешнего адаптера координат
    void SetAdapter(AccAdapter* acc)
    {
        Adapter = acc;
    };

    // Установка внешней модели КО
    void SetKOProp(IKOPropVar* prop);

    // Собственно прогнозирование
    virtual bool Run(const TimeJD& t);
    // Получение конечной точки
    virtual void GetDestPoint(PhasePoint6D& dest);
    // Получение матрицы частных производных
    virtual void GetDestMatrix(DMatrix& m);

    // Установка начальных условий для интегрирования
    void SetInitPoint(const PhasePoint6D& src, const DMatrix* vars = NULL);
    // "Переразгон" прогноза от текущей точки
    virtual void ReRun();

    // Используемая модель движения
    MotionMode MMode;

protected:
    // Система уравнений
    ODEVarSystem System;
    // Интегратор
    IODESolver* Integrator;
    // Используемый адаптер ускорений
    AccAdapter* Adapter;
    // Модель космического объекта
    IKOPropVar* KOProp;
    // Флаг наличия начальных условий
    bool FlagInit;
    // Момент времени, от которого идет отсчет
    TimeJD T0;
};

///////////////////////////////////////////////////////////////////////////////
// Стандартный численный прогноз уравнений в вариациях:
// - интегратор В.А.Степаньянца
// - адаптер координат, не учитвающий нутацию и прецессию
// - КО с постоянным балл. коэффициентом и коэф. светового давления
///////////////////////////////////////////////////////////////////////////////
class DigStPredictVar : public DigPredictVar
{
public:
    DigStPredictVar();
    DigStPredictVar(double eps, double stepmin, double stepinit, bool precItrf);

    // Установка параметров КО
    void SetKOParams(double kb, double kappa, bool iskbvar = false, bool iskappavar = false);
    // Установка точности интегратора и шагов
    void SetAccuracy(double eps, double stepinit, double stepmin, bool precItrf);

    // Интегратор
    ODEStpnSolver SIntegrator;

protected:
    // Адаптер
    AccAdapter SAdapter;
    // Модель КО
    KOConstVar KOModel;

private:
    // Запрет установки внешнего интегратора
    void SetIntegrator(IODESolver* slv);
    // Запрет установки внешнего адаптера координат
    void SetAdapter(AccAdapter* acc);
    // Запрет установки внешней модели КО
    void SetKOProp(IKOPropVar* prop);

    //точность интегрирования
    double m_eps;
    //начальный шаг интегрирования
    double m_stepinit;
    //минимальный шаг интегрирования
    double m_stepmin;
};

///////////////////////////////////////////////////////////////////////////////
#endif
