///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Утилиты для определения драконического периода и его падения
///////////////////////////////////////////////////////////////////////////////
#include "DrPeriod.h"
#include "J2000Frame.h"
#include "Kepler.h"
#include "Consts.h"
#include <iostream>
#include "TimeFunctions.h"

///////////////////////////////////////////////////////////////////////////////
// Поиск восходящего узла
///////////////////////////////////////////////////////////////////////////////
bool GetAscNodeYZ(IPredict& prg, const TimeJD& tprim, PhasePoint6D& vect, double (*fParam)(const Vect6&), double (*fvParam)(const Vect6&))
{
    TimeJD t0;
    int i, k;
    double t1;

    t0 = tprim;

    // Определяем оскулирующий период орбиты
    //	std::cout << DateToStr(t0) << "\n";
    if (!prg.Run(t0)) {
        return false;
    }

    prg.GetDestPoint(vect);

    J2000Frame f;
    if (vect.GetFrame()->GetType() != csIdJ2000) {
        vect.Convert(f);
    }

    Kepler kp;
    kp.SetXYZ(vect.CoordsVel);
    double tp = kp.GetPeriod();


    for (k = 0; k < 2; ++k) {
        for (i = 0; i < 20; ++i) {
            if (!prg.Run(t0)) {
                return false;
            }

            prg.GetDestPoint(vect);
            if (vect.GetFrame()->GetType() != csIdJ2000) {
                vect.Convert(f);
            }

            if (fabs(fParam(vect.CoordsVel)) < 1.0e-6) {
                // Достигнута точность в 1м
                break;
            }

            t1 = fParam(vect.CoordsVel) / fvParam(vect.CoordsVel);

            // Шаг не больше четверти периода
            if (fabs(t1) > 0.25 * tp) {
                t1 = t1 > 0.0 ? 0.25 * tp : -0.25 * tp;
            }

            t0 -= t1 / 86.4;
        }

        if (fvParam(vect.CoordsVel) < 0.0) {
            // Пришли в нисходящий узел
            kp.SetXYZ(vect.CoordsVel);
            double dt = kp.GetTimeFromAN();
            if (kp.CosI < 0.995) {
                t0 -= 0.5 * tp / 86.400;
            }
            else {
                kp.U -= Pi;
                double dt1 = kp.GetTimeFromAN();
                if (dt1 < dt) {
                    t0 -= (dt - dt1) / 86.4;
                }
                else {
                    t0 -= (dt + (kp.GetPeriod() - dt1)) / 86400.0;
                }
            }
            continue;
        }
        break;
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////
// Вспомогательные функции
///////////////////////////////////////////////////////////////////////////////
static double GetZ(const Vect6& v) { return v.z; };
static double GetVZ(const Vect6& v) { return v.vz; };
static double GetY(const Vect6& v) { return v.y; };
static double GetVY(const Vect6& v) { return v.vy; };
///////////////////////////////////////////////////////////////////////////////
bool GetAscNode(IPredict& prg, PhasePoint6D& vect, const TimeJD& tprim, std::string ascendingNodeRule)
{
    if (!prg.Run(tprim)) {
        return false;
    }

    prg.GetDestPoint(vect);

    J2000Frame f;
    if (vect.GetFrame()->GetType() != csIdJ2000) {
        vect.Convert(f);
    }

    Kepler kp;
    kp.SetXYZ(vect.CoordsVel);

    if (ascendingNodeRule == "vz" || ascendingNodeRule == "default" && kp.CosI <= 0.995) {
        return GetAscNodeYZ(prg, tprim, vect, GetZ, GetVZ);
    }
    else {
        return GetAscNodeYZ(prg, tprim, vect, GetY, GetVY);
    }
}

///////////////////////////////////////////////////////////////////////////////
// Определение драконического периода и его падения
///////////////////////////////////////////////////////////////////////////////
bool GetOrbPeriod(IPredict& prg, const TimeJD& t, double& tp, double& dt, std::string ascendingNodeRule)
{
    PhasePoint6D vect2;
    TimeJD ta, ta2, tprim;

    PhasePoint6D vect;
    if (!GetAscNode(prg, vect, t, ascendingNodeRule)) {
        return false;
    }

    ta = vect.T;

    int l = 1;

    // Определяем оскулирующий период орбиты
    if (!prg.Run(ta)) {
        return false;
    }

    prg.GetDestPoint(vect);
    J2000Frame f;
    if (vect.GetFrame()->GetType() != csIdJ2000) {
        vect.Convert(f);
    }

    Kepler kp;
    kp.SetXYZ(vect.CoordsVel);
    double to = kp.GetPeriod();

    tprim = ta;
    tprim -= l * to / 86.4;

    if (!GetAscNode(prg, vect2, tprim, ascendingNodeRule)) {
        return false;
    }

    TmDay tmd = ta - vect2.T;
    tp = tmd.GetDouble() / l;

    ta2 = vect2.T;
    tprim = ta2;
    tprim -= l * to / 86.4;
    if (!GetAscNode(prg, vect2, tprim, ascendingNodeRule)) {
        return false;
    }

    tmd = ta - ta2;
    TmDay tmd1 = vect2.T - ta2;
    dt = (tmd.GetDouble() + tmd1.GetDouble()) / (l * l);

    return true;
};
///////////////////////////////////////////////////////////////////////////////
