///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Класс гравитационных моделей Земли
///////////////////////////////////////////////////////////////////////////////
#include "EarthGravity.h"
#include "Defines.h"
///////////////////////////////////////////////////////////////////////////////
#define MAXC 134
///////////////////////////////////////////////////////////////////////////////
// Вспомогательная функция, используемая при разборке файла гравитационной
// модели
#define MAX_LENGTH_LINE 128

static void convert_DE(char* s)
{
    char* s1;
    int len;

    s1 = s;
    while (s1 = strpbrk(s1, "Dd"), s1 != NULL) {
        switch (s1[1]) {
        case '+': ;
        case '-':
            s1[0] = 'e';
            break;
        default:
            s1++;
        }
    };

    s1 = s;
    len = strlen(s);
    while (s1 = strchr(s1, '.'), s1 != NULL) {
        if (s1[-1] < '0' || s1[-1] > '9') {
            /*
            if (len>=MAX_LENGTH_LINE)
              exceptBikop(excIdOutOfArray,"строка при разборе файла грав. модели", 1);*/
            memmove(s1 + 1, s1, len - (s1 - s));
            s[++len] = 0;
            s1[0] = '0';
        };
        s1++;
    };
};
///////////////////////////////////////////////////////////////////////////////
// Конструктор, инициализирующий модель из файла
EarthGravityModel::EarthGravityModel(TGravityModelId Id_, const char* name)
{
    initEmpty(Id_);
    FName = name;
};
//-----------------------------------------------------------------------------
// инициализация модели из файла /*fold00*/
bool EarthGravityModel::Load()
{
    //   вариант для чтения коэфф. из файла по имени name
    // 1 - N, 2 - M, 3 - Мю Земли, 4 - Re - экв. радиус Земли
    // n**2+1     - cn0, коэфф. при зональных гармониках
    // n**2+2*m   - cnm, коэфф. при тесс. и сект. гармониках
    // n**2+2*m+1 - dnm, коэфф. при тесс. и сект. гармониках
    FILE* f;
    unsigned int N, M, len, n, m;
    long double t2, t3, t4, t5;
    char s[MAX_LENGTH_LINE + 2], *s1, tit[9];

    freeAll();
    f = fopen(FName, "r");

    if (!f)
        return false;

    s1 = fgets(s + 1, MAX_LENGTH_LINE + 1, f);
    s1 = fgets(s + 1, MAX_LENGTH_LINE + 1, f);

    if (s1 == NULL)
        return false;

    s[0] = ' ';
    convert_DE(s);

    len = sscanf(s1, "%*s %d %d %Le %Le %Le", &N, &M, &t3, &t4, &t5);

    if (len < 5) {
        fclose(f);
        return false;
    };

    mu = t3 * 1e-12;
    Re = t4 * 1e-6;
    alpha = 1 / t5;

    if (N > MAXC)
        N = MAXC;
    if (M > MAXC)
        M = MAXC;

    SetSize(N, M);
    null();

    while (feof(f) == 0) {
        s1 = fgets(s + 1, MAX_LENGTH_LINE, f);
        if (s1 == NULL)
            break;
        s[0] = ' ';
        convert_DE(s);
        len = sscanf(s, "%7s %d %d %Le", tit, &n, &m, &t2);
        if (len < 4)
            break;
        if (n < 2 || n > N_)
            continue;
        if (m < 0 || m > M_)
            continue;
        if (strcmp(tit, "GCOEFC1") == 0)
            CC[m][n] = t2;
        else
            DD[m][n] = t2;
    };

    fclose(f);
    isInit = true;
    return true;
};
//-----------------------------------------------------------------------------
// конвертация в/из нормированных коэффициентов
void EarthGravityModel::convert(bool To)
{
    unsigned int n, m, M1;
    double t1, t2, Qn02, Qnm2, Qnm3;

    for (n = 2; n <= N_; n++) {
        Qn02 = 1.0 / (2 * n + 1);
        t1 = sqrt(Qn02);
        if (To)
            CC[0][n] *= t1;
        else
            CC[0][n] /= t1;
        Qnm2 = Qn02 * n * (n + 1) / 2.0;
        Qnm3 = 1.0;
        M1 = (n < M_ ? n : M_);
        for (m = 1; m <= M1; m++) {
            t1 = sqrt(Qnm2);
            t2 = sqrt(Qnm3);
            if (To) {
                CC[m][n] *= t1;
                CC[m][n] *= t2;
                DD[m][n] *= t1;
                DD[m][n] *= t2;
            }
            else {
                CC[m][n] /= t1;
                CC[m][n] /= t2;
                DD[m][n] /= t1;
                DD[m][n] /= t2;
            };
            Qnm2 *= (n + m + 1);
            Qnm3 *= (n - m);
        };
    };
};
///////////////////////////////////////////////////////////////////////////////
