///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Класс, описывающий модели Земли
///////////////////////////////////////////////////////////////////////////////
#include "Defines.h"
#include "Planet.h"
#include "Consts.h"
#include "EarthGravity.h"
///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
EarthPlanet::EarthPlanet()
{
    GrModel = new EarthPZ902GravityModel;
    AtmModel = NULL;

    Geometry.Re = GrModel->Re;
    Geometry.alpha = GrModel->alpha;
    omega = EarthMeanRot_Vallado / 86.4;
    RShad = Geometry.Re;
}

//-----------------------------------------------------------------------------
// Деструктор
EarthPlanet::~EarthPlanet()
{
    if (GrModel != NULL)
        delete GrModel;
    if (AtmModel != NULL)
        delete AtmModel;

    GrModel = NULL;
    AtmModel = NULL;
}

//-----------------------------------------------------------------------------
// Установка гравитационной модели
bool EarthPlanet::SetGravityModel(TGravityModelId m, const char* fname)
{
    TGravityModelId old = GrModel->GetID();

    switch (m) {
    case grIdPZ90:
        if (old != m) {
            delete GrModel;
            GrModel = new EarthPZ90GravityModel;
        }
        break;
    case grIdPZ902s:
        if (old != m) {
            delete GrModel;
            GrModel = new EarthPZ902GravityModel;
        }
        break;
    case grIdEGM96:
    case grIdJGM3:
    case grIdWGS84:
    case grIdWGS72:
    case grIdPZ902f:
        if (old != m) {
            EarthGravityModel* md = new EarthGravityModel(m, fname);
            if (!md->Load())
                return false;
            const char* p = strrchr(fname, '.');
            if (!strcmp(p, ".nrm"))
                md->fromNorm();
            delete GrModel;
            GrModel = md;
        }
        break;
    default:
        TAssert(false);
        return false;
    };

    Geometry.Re = GrModel->Re;
    Geometry.alpha = GrModel->alpha;
    omega = EarthMeanRot_Vallado / 86.4;
    RShad = Geometry.Re;
    return true;
}

//-----------------------------------------------------------------------------
// Установка модели атмосферы
void EarthPlanet::SetAtmModel(TAtmosphereId m, ITableGeoSolarFlux* gsf, bool flag3h)
{
    TAtmosphereId old = (AtmModel == NULL) ? atmIdNONE : AtmModel->GetID();

    switch (m) {
    case atmIdNONE:
        if (m != old) {
            delete AtmModel;
            AtmModel = NULL;
        }
        break;

    case atmIdStatic:
        if (m != old) {
            if (AtmModel != NULL)
                delete AtmModel;
            AtmModel = new StaticAtmosphere;
        }
        break;

    case atmIdDMA90:
        if (m != old) {
            if (AtmModel != NULL)
                delete AtmModel;
            AtmModel = new Gost90Atmosphere(gsf, flag3h);
        }
        else {
            Gost90Atmosphere* datm = (Gost90Atmosphere *)AtmModel;
            datm->Gsf = gsf;
            datm->is3HourInd = flag3h;
        }
        break;

    case atmIdDMA2000:
        if (m != old) {
            if (AtmModel != NULL)
                delete AtmModel;
            AtmModel = new Gost2000Atmosphere(gsf, flag3h);
        }
        else {
            Gost2000Atmosphere* datm = (Gost2000Atmosphere *)AtmModel;
            datm->Gsf = gsf;
            datm->is3HourInd = flag3h;
        }
        break;

    case atmIdDMA77:
        if (m != old) {
            if (AtmModel != NULL)
                delete AtmModel;
            AtmModel = new Gost77Atmosphere(gsf);
        }
        else {
            Gost77Atmosphere* datm = (Gost77Atmosphere *)AtmModel;
            datm->Gsf = gsf;
        }
        break;

    case atmIdDMA2004:
        if (m != old && m != atmIdDMA2004m) {
            if (AtmModel != NULL)
                delete AtmModel;
            AtmModel = new Gost2004Atmosphere(gsf, flag3h);
        }
        else {
            if (!AtmModel) {
                AtmModel = new Gost2004Atmosphere(gsf, flag3h, false);
            }
            else {
                Gost2004Atmosphere* datm = (Gost2004Atmosphere *)AtmModel;
                datm->Gsf = gsf;
                datm->is3HourInd = flag3h;
                datm->fmod = false;
            }
        }
        break;

    case atmIdDMA2004m:
        if (m != old && m != atmIdDMA2004m) {
            if (AtmModel != NULL)
                delete AtmModel;
            AtmModel = new Gost2004Atmosphere(gsf, flag3h, true);
        }
        else {
            if (!AtmModel) {
                AtmModel = new Gost2004Atmosphere(gsf, flag3h, true);
            }
            else {
                Gost2004Atmosphere* datm = (Gost2004Atmosphere *)AtmModel;
                datm->Gsf = gsf;
                datm->is3HourInd = flag3h;
                datm->fmod = true;
            }
        }
        break;

    default:
        TAssert(false);
    };
}

///////////////////////////////////////////////////////////////////////////////
