///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Определение класса Гринвичской замороженной системы координат
////////////////////////////////////////////////////////////////////////////////
#include "FGreenwichFrame.h"
#include "METEFrame.h"
#include "NZEFrame.h"
#include "IAstroBase.h"
////////////////////////////////////////////////////////////////////////////////
// Установка параметров системы
void FGreenwichFrame::SetParameters(const IFrameParameters& p)
{
    TAssert(p.GetFrameType() == csIdGFCS);
    const FGreenwichFrameParameters& pm = (const FGreenwichFrameParameters &)p;
    Pars = pm;
}

//------------------------------------------------------------------------------
// Получение объекта, который соответствует переходу от текущей системы отсчета 
// к системе отсчета J2000 на момент времени t0 (для 3-мерных векторов)
void FGreenwichFrame::Get3DJ2000Converter(const TimeJD& t0, FrameConverter3D& result) const
{
    TAssert(Pars.IsInit());

    result.Init();

    GetAstroBase()->MtxJ20002GFCS(Pars.GetTime(), result.Rotation);
    result.Rotation.Transp();
    return;
}

//------------------------------------------------------------------------------
// Получение объекта, который соответствует переходу от текущей системы отсчета 
// к системе отсчета J2000 на момент времени t0 (для 6-мерных векторов)
void FGreenwichFrame::Get6DJ2000Converter(const TimeJD& t0, FrameConverter6D& result) const
{
    TAssert(Pars.IsInit());
    result.Init();
    Get3DJ2000Converter(t0, result.Converter3D);
}

//------------------------------------------------------------------------------
// Получение объекта, который соответствует переходу от текущей системы отсчета 
// к системе отсчета rf на момент времени t0 (для 3-мерных векторов)
void FGreenwichFrame::Get3DConverter(const IFrame& rf, const TimeJD& t0,
                                     FrameConverter3D& result) const
{
    TAssert(Pars.IsInit());

    if (rf.GetType() == csIdGCS) {
        if (t0 == Pars.GetTime()) {
            result.Init();
            return;
        }
    }

    if (rf.GetType() == csIdMETE) {
        const METEFrame& mf = (const METEFrame &)rf;
        if (Pars.GetTime() == mf.GetFixTime()) {
            result.Init();
            GetAstroBase()->MtxMETE2GCS(Pars.GetTime(), result.Rotation);
            result.Rotation.Transp();
            return;
        }
    }

    if (rf.GetType() == csIdNZE) {
        const NZEFrame& lf = (const NZEFrame &)rf;
        if (Pars.GetTime() == t0) {
            lf.Get3DGreenwichConverter(result);
            result.Invert();
            return;
        }
    }

    IFrame::Get3DConverter(rf, t0, result);
}

//------------------------------------------------------------------------------
// Получение объекта, который соответствует переходу от текущей системы отсчета 
// к системе отсчета rf на момент времени t0 (для 6-мерных векторов)
void FGreenwichFrame::Get6DConverter(const IFrame& rf, const TimeJD& t0,
                                     FrameConverter6D& result) const
{
    TAssert(Pars.IsInit());

    if (rf.GetType() == csIdGCS) {
        if (t0 == Pars.GetTime()) {
            result.Init();
            GetAstroBase()->GCSVelocityMatrix(result.VRotation);
            return;
        }
    }

    if (rf.GetType() == csIdMETE) {
        const METEFrame& mf = (const METEFrame &)rf;
        if (Pars.GetTime() == mf.GetFixTime()) {
            result.Init();
            Get3DConverter(rf, t0, result.Converter3D);
            return;
        }
    }

    if (rf.GetType() == csIdNZE) {
        const NZEFrame& lf = (const NZEFrame &)rf;
        if (Pars.GetTime() == t0) {
            FrameConverter6D gl, fg;
            lf.Get6DGreenwichConverter(gl);
            gl.Invert();
            fg.Init();
            GetAstroBase()->GCSVelocityMatrix(fg.VRotation);
            fg.Multiply(gl, result);
            return;
        }
    }

    IFrame::Get6DConverter(rf, t0, result);
}

////////////////////////////////////////////////////////////////////////////////
