///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Определение операций преобразования декартовых координат при переходе
// в другую систему отсчета
////////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL__FRAMECONVERTER_H
#define BNOTOOL__FRAMECONVERTER_H
////////////////////////////////////////////////////////////////////////////////
#include "DVectors.h"

////////////////////////////////////////////////////////////////////////////////
class FrameConverter3D
{
public:
    // Конструктор по умолчанию. Создает тождественное преобразование. 
    FrameConverter3D()
    {
        Init();
    };
    // Конструктор, создающий конвертер из матрицы и вектора
    FrameConverter3D(const TransMtx& m, const Vect3& r)
    {
        Init(m, r);
    };

    // Очистка - создание тождественного преобразования
    FrameConverter3D& Init()
    {
        Rotation.Init1();
        Shift = 0;
        return *this;
    };
    // Конструктор, создающий конвертер из матрицы и вектора
    FrameConverter3D& Init(const TransMtx& m, const Vect3& r)
    {
        Rotation = m;
        Shift = r;
        return *this;
    };

    // Преобразовать прямоугольные координаты в исходной системе отсчета
    // в прямоугольные координаты в системе отсчета назначения.
    void Convert(const Vect3& source, Vect3& target) const;

    // Сформировать в объекте result обратное преобразование
    void Invert(FrameConverter3D& result) const;

    // Заменить текущее преобразование на обратное
    FrameConverter3D& Invert();

    // Получить в result композицию двух преобразований.
    // Левый операнд композиции  - текущий объект
    // Правый операнд композиции - преобразование c.
    void Multiply(const FrameConverter3D& c, FrameConverter3D& result) const;
    // Сформировать в текущем объекте его композицию с объектом с
    FrameConverter3D& Multiply(const FrameConverter3D& c);

    // Публичные переменные 

    // Матрица поворота
    TransMtx Rotation;
    // Вектор смещения. Содержит координаты начала исходной системы
    // отсчета в системе отсчета назначения.
    Vect3 Shift;
};

////////////////////////////////////////////////////////////////////////////////
class FrameConverter6D
{
public:
    // Конструктор по умолчанию. Создает тождественное преобразование. 
    FrameConverter6D()
    {
        Init();
    };
    // Конструктор с явным указанием параметров
    FrameConverter6D(const FrameConverter3D& c, const TransMtx& m,
                     const Vect3& v)
    {
        Init(c, m, v);
    };

    FrameConverter6D(const TransMtx& mc, const TransMtx& mv,
                     const Vect6& rv)
    {
        Init(mc, mv, rv);
    };

    // Очистка - создание тождественного преобразования
    FrameConverter6D& Init()
    {
        Converter3D.Init();
        VShift = 0;
        VRotation = 0;
        return *this;
    };
    // Инициализация с явным указанием параметров
    FrameConverter6D& Init(const FrameConverter3D& c, const TransMtx& m,
                           const Vect3& v)
    {
        Converter3D = c;
        VRotation = m;
        VShift = v;
        return *this;
    };
    // Инициализация с явным указанием матриц и вектора сдвига
    FrameConverter6D& Init(const TransMtx& mc, const TransMtx& mv,
                           const Vect6& rv);

    // Преобразовать прямоугольные координаты в исходной системе отсчета
    // в прямоугольные координаты в системе отсчета назначения.
    void Convert(const Vect6& source, Vect6& target) const;

    // Сформировать в объекте result обратное преобразование
    void Invert(FrameConverter6D& result) const;
    // Заменить текущее преобразование на обратное
    FrameConverter6D& Invert();

    // Получить в result композицию двух преобразований.
    // Левый операнд композиции  - текущий объект
    // Правый операнд композиции - преобразование c.
    void Multiply(const FrameConverter6D& c, FrameConverter6D& result) const;
    // Сформировать в текущем объекте его композицию с объектом с
    FrameConverter6D& Multiply(const FrameConverter6D& c);


    // Конвертер координат
    FrameConverter3D Converter3D;
    // Матрица поворота
    TransMtx VRotation;
    // Вектор смещения. Содержит координаты начала исходной системы
    // отсчета в системе отсчета назначения.
    Vect3 VShift;
};

////////////////////////////////////////////////////////////////////////////////
#endif
