///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Динамическая модель атмосферы по ГОСТ-2000
///////////////////////////////////////////////////////////////////////////////
#include "Defines.h"
#include "Atmosphere.h"
#include "IAstroBase.h"
///////////////////////////////////////////////////////////////////////////////
// Данные для модели
///////////////////////////////////////////////////////////////////////////////
// Структура коэффициентов нижней атмосферы
#define R0    1.58868e-8

struct LowAtmCoeffStruct
{
    double h, a0, k1, k2;
};

//-----------------------------------------------------------------------------
// Значения коэффициентов нижней атмосферы
//-----------------------------------------------------------------------------
static struct LowAtmCoeffStruct LowAtmCoeffs[] =
{
    {0, 1.228, -9.0764e-2, -2.0452e-3},
    {20, 9.013e-2, -0.16739, 6.2669e-4},
    {60, 3.0447e-4, -0.1356, -8.13e-4},
    {100, 3.66e-7, -0.1633, 4.621e-4}
};
//-----------------------------------------------------------------------------
// Вычисление плотности нижней атмосферы
//-----------------------------------------------------------------------------
static double LowAtmosphereDensity(double h)
{
    unsigned i;

    if (h < 0)
        h = 0.0;

    for (i = 0; i < sizeof(LowAtmCoeffs) / sizeof(*LowAtmCoeffs) - 1; i++)
        if (h < LowAtmCoeffs[i + 1].h)
            break;

    double dh = h - LowAtmCoeffs[i].h;

    return LowAtmCoeffs[i].a0 * exp(LowAtmCoeffs[i].k1 * dh + LowAtmCoeffs[i].k2 * dh * dh);
}

//-----------------------------------------------------------------------------
// Таблица уровней высот по слоям
//-----------------------------------------------------------------------------
static double TableHeightLevels[] =
{
    // 75  100  125  150  175  200   250 
    500, 500, 500, 500, 500, 500, 500, // a
    600, 700, 800, 800, 900, 900, 1000, // b 
    700, 700, 800, 800, 900, 900, 1000, // c 
    800, 800, 800, 800, 800, 800, 800, // d 
    600, 700, 800, 800, 800, 900, 800, // e 
    700, 700, 800, 800, 800, 900, 900, // k
    600, 700, 800, 800, 900, 900, 900 // l
};
//-----------------------------------------------------------------------------
// Вычисление уровня слоя для заданного коэффициента
//-----------------------------------------------------------------------------
static double GetHeightLevel(int ind, int cf)
{
    return TableHeightLevels[cf * 7 + ind];
}

//-----------------------------------------------------------------------------
// Таблицы коэффициентов
//-----------------------------------------------------------------------------
static double ACoefLowTab[] =
{
    //    a0             a1           a2          a3           a4           a5            a6
    2.57309e+01, -4.23306e-01, 2.62336e-03, -9.29071e-06, 1.83968e-08, -1.91628e-11, 8.20168e-15, // 75
    2.75605e+01, -4.66242e-01, 3.00008e-03, -1.08873e-05, 2.20776e-08, -2.3551e-11, 1.03109e-14, // 100
    2.85344e+01, -4.88303e-01, 3.17933e-03, -1.15414e-05, 2.32988e-08, -2.46356e-11, 1.06444e-14, // 125
    2.91482e+01, -5.02424e-01, 3.29326e-03, -1.19351e-05, 2.39615e-08, -2.51099e-11, 1.07148e-14, // 150
    2.94966e+01, -5.10952e-01, 3.36392e-03, -1.21725e-05, 2.43369e-08, -2.53384e-11, 1.07169e-14, // 175
    2.97564e+01, -5.1782e-01, 3.42511e-03, -1.23973e-05, 2.47498e-08, -2.56933e-11, 1.08203e-14, // 200
    3.00796e+01, -5.27712e-01, 3.52182e-03, -1.27769e-05, 2.54982e-08, -2.64066e-11, 1.10731e-14, // 250
};

static double ACoefHighTab[] =
{
    //    a0             a1           a2          a3             a4           a5            a6
    4.98651e+01, -3.34292e-01, 7.39139e-04, -8.88418e-07, 5.99515e-10, -2.13818e-13, 3.13667e-17, // 75
    5.81031e+00, -3.87883e-02, -3.52838e-05, 1.51843e-07, -1.63797e-10, 7.81677e-14, -1.43386e-17, // 100
    -4.78056e+00, 2.50162e-02, -1.75154e-04, 3.04471e-07, -2.5259e-10, 1.0452e-13, -1.7487e-17, // 125
    -3.36444e+01, 2.0446e-01, -6.11416e-04, 8.49908e-07, -6.26417e-10, 2.3843e-13, -3.71372e-17, // 150
    -2.65082e+01, 1.50087e-01, -4.42887e-04, 5.91948e-07, -4.17581e-10, 1.52315e-13, -2.28438e-17, // 175
    -2.3453e+01, 1.24743e-01, -3.58309e-04, 4.57871e-07, -3.07324e-10, 1.06444e-13, -1.51634e-17, // 200
    -1.62816e+01, 7.28713e-02, -2.04003e-04, 2.32554e-07, -1.333e-10, 3.74919e-14, -4.0457e-18, // 250
};

static double BCoefLowTab[] =
{
    //    b0             b1           b2          b3           b4
    6.49977e-2, -3.06873e-3, 1.95655e-5, -6.66691e-9, -2.07608e-11, // 75
    6.50652e-2, -3.02445e-3, 1.93214e-5, -9.73263e-9, -1.22376e-11, // 100
    0.165612, -4.35142e-3, 2.51698e-5, -2.2581e-8, -3.53038e-13, // 125
    -5.89144e-2, -1.3585e-3, 1.21856e-5, -1.9479e-9, -9.95541e-12, // 150
    4.44639e-3, -2.19289e-3, 1.58524e-5, -9.50062e-9, -3.7914e-12, // 175
    -0.140165, -4.65788e-4, 9.2958e-6, -3.94258e-10, -7.30491e-12, // 200
    -2.04263e-1, 1.36515e-4, 7.80717e-6, 3.04886e-10, -5.95365e-12, // 250

};

static double BCoefHighTab[] =
{
    //    b0             b1           b2          b3           b4
    19.1597, -6.80873e-2, 9.15651e-5, -5.39353e-8, 1.16731e-11, // 75
    26.9895, -8.95941e-2, 1.124e-4, -6.18513e-8, 1.25729e-11, // 100
    34.4274, -1.07173e-1, 1.26243e-4, -6.5469e-8, 1.26232e-11, // 125
    29.6371, -8.1444e-2, 8.45828e-5, -3.83526e-8, 6.41585e-12, // 150
    36.9576, -9.75364e-2, 9.74067e-5, -4.27077e-8, 6.96063e-12, // 175
    11.3165, -5.0416e-3, -2.222e-5, 2.40539e-8, -6.71207e-12, // 200
    -14.6575, 8.64075e-2, -1.34306e-4, 8.24585e-8, -1.77803e-11, // 250
};

static double CCoefLowTab[] =
{
    //    c0          c1           c2          c3          c4        n0     n1          n2       f1
    -0.132579, 2.9853e-3, -2.25849e-5, 1.20703e-7, -1.24478e-10, 2.058, 5.887e-3, -4.012e-6, 0.5411, // 75
    -0.132579, 2.9853e-3, -2.25849e-5, 1.20703e-7, -1.24478e-10, 2.058, 5.887e-3, -4.012e-6, 0.5515, // 100 
    -0.548245, 8.20121e-3, -4.01156e-5, 1.14321e-7, -8.71732e-11, 2.058, 5.887e-3, -4.012e-6, 0.5585, // 125
    -0.548245, 8.20121e-3, -4.01156e-5, 1.14321e-7, -8.71732e-11, 2.058, 5.887e-3, -4.012e-6, 0.5585, // 150
    -0.545624, 7.57748e-3, -3.34243e-5, 8.38662e-8, -5.59389e-11, 2.058, 5.887e-3, -4.012e-6, 0.5585, // 175
    -0.545624, 7.57748e-3, -3.34243e-5, 8.38662e-8, -5.59389e-11, 2.058, 5.887e-3, -4.012e-6, 0.5585, // 200 
    -0.552086, 7.16294e-3, -2.92525e-5, 6.48366e-8, -3.80755e-11, 2.058, 5.887e-3, -4.012e-6, 0.5585, // 250 
};

static double CCoefHighTab[] =
{
    //    c0          c1           c2          c3          c4        n0     n1          n2       f1
    26.9231, -6.85685e-2, 6.14236e-5, -2.00812e-8, 1.22787e-12, 2.058, 5.887e-3, -4.012e-6, 0.5411, // 75
    26.9231, -6.85685e-2, 6.14236e-5, -2.00812e-8, 1.22787e-12, 2.058, 5.887e-3, -4.012e-6, 0.5515, // 100
    29.9244, -5.61513e-2, 2.7733e-5, 4.23239e-9, -4.16164e-12, 2.058, 5.887e-3, -4.012e-6, 0.5585, // 125
    29.9244, -5.61513e-2, 2.7733e-5, 4.23239e-9, -4.16164e-12, 2.058, 5.887e-3, -4.012e-6, 0.5585, // 150
    17.7196, -5.01253e-3, -3.51915e-5, 3.38521e-8, -8.67852e-12, 2.058, 5.887e-3, -4.012e-6, 0.5585, // 175
    17.7196, -5.01253e-3, -3.51915e-5, 3.38521e-8, -8.67852e-12, 2.058, 5.887e-3, -4.012e-6, 0.5585, // 200
    1.8336, 3.78361e-2, -6.45005e-5, 3.48538e-8, -5.87135e-12, 2.058, 5.887e-3, -4.012e-6, 0.5585, // 250
};

static double DCoefLowTab[] =
{
    //    d0             d1           d2          d3
    -0.0619071, 3.53963e-3, 4.98912e-6, -6.89196e-9, // 75
    -0.0619071, 3.53963e-3, 4.98912e-6, -6.89196e-9, // 100
    -0.217106, 6.30628e-3, -5.65614e-6, 2.85734e-9, // 125
    -0.217106, 6.30628e-3, -5.65614e-6, 2.85734e-9, // 150
    -0.199696, 6.69745e-3, -7.68966e-6, 4.04246e-9, // 175
    -0.199696, 6.69745e-3, -7.68966e-6, 4.04246e-9, // 200
    -0.199696, 6.69745e-3, -7.68966e-6, 4.04246e-9, // 250
};

static double DCoefHighTab[] =
{
    //    d0          d1           d2          d3
    0.887001, 6.2522e-3, -6.91964e-6, 1.88445e-9, // 75
    0.887001, 6.2522e-3, -6.91964e-6, 1.88445e-9, // 100
    -0.451254, 7.8777e-3, -6.07417e-6, 1.35098e-9, // 125
    -0.451254, 7.8777e-3, -6.07417e-6, 1.35098e-9, // 150
    -3.07766, 1.29428e-2, -9.67358e-6, 2.39266e-9, // 175
    -3.07766, 1.29428e-2, -9.67358e-6, 2.39266e-9, // 200
    -3.07766, 1.29428e-2, -9.67358e-6, 2.39266e-9, // 250
};

static double ECoefLowTab[] =
{
    //    e0          e1           e2          e3          e4        e5       e6          e7          e8        et5       et6        et7          et8
    -0.777916, 6.0407e-3, -5.88566e-6, 6.92343e-8, -9.6119e-11, -0.20902, 9.7533e-2, -1.1817e-2, 1.6145e-3, -0.20807, 9.4449e-2, -8.7953e-3, 8.8385e-4, // 75
    -0.777739, 5.6909e-3, 1.81832e-6, 3.35689e-8, -5.16839e-11, -0.17156, 7.9830e-2, -9.4393e-3, 1.2622e-3, -0.17087, 7.7599e-2, -7.1375e-3, 6.9025e-4, // 100
    -0.570476, 2.95802e-3, 1.68896e-5, -4.74756e-9, -1.72711e-11, -0.14825, 6.8808e-2, -7.9836e-3, 1.0535e-3, -0.14770, 6.7052e-2, -6.0951e-3, 5.7456e-4, // 125
    -0.949573, 8.13121e-3, -3.87813e-6, 2.37694e-8, -2.77469e-11, -0.13287, 6.1603e-2, -7.0866e-3, 9.2813e-4, -0.13240, 6.0105e-2, -5.4388e-3, 5.0585e-4, // 150
    -0.967598, 8.41991e-3, -3.585e-6, 1.74801e-8, -1.96221e-11, -0.122241, 5.6538e-2, -6.4324e-3, 8.3723e-4, -0.12183, 5.5232e-2, -4.9580e-3, 4.5512e-4, // 175
    -1.02278, 9.23633e-3, -6.10128e-6, 1.78211e-8, -1.70073e-11, -0.11460, 5.3178e-2, -6.0436e-3, 7.7982e-4, -0.11423, 5.1994e-2, -4.6876e-3, 4.2548e-4, // 200
    -0.694824, 5.20601e-3, 1.16681e-5, -1.65494e-8, 5.75301e-12, -0.10524, 4.8551e-2, -5.3567e-3, 6.8809e-4, -0.10493, 4.7573e-2, -4.1711e-3, 3.7068e-4, // 250 
};

static double ECoefHighTab[] =
{
    //    e0      e1            e2          e3          e4           e5       e6          e7          e8          et5       et6        et7          et8
    39.0335, -0.1337, 1.77478e-4, -1.03627e-7, 2.24068e-11, -0.20902, 9.7533e-2, -1.1817e-2, 1.6145e-3, -0.20807, 9.4449e-2, -8.7953e-3, 8.8385e-4, // 75
    52.6286, -0.172381, 2.18499e-4, -1.2211e-7, 2.53442e-11, -0.17156, 7.9830e-2, -9.4393e-3, 1.2622e-3, -0.17087, 7.7599e-2, -7.1375e-3, 6.9025e-4, // 100
    68.4746, -2.15659e-1, 2.62273e-4, -1.40972e-7, 2.82285e-11, -0.14825, 6.8808e-2, -7.9836e-3, 1.0535e-3, -0.14770, 6.7052e-2, -6.0951e-3, 5.7456e-4, // 125
    58.422, -1.66664e-1, 1.85486e-4, -9.12345e-8, 1.67118e-11, -0.13287, 6.1603e-2, -7.0866e-3, 9.2813e-4, -0.13240, 6.0105e-2, -5.4388e-3, 5.0585e-4, // 150
    7.20188, 2.16109e-2, -6.52882e-5, 5.37077e-8, -1.4095e-11, -0.122241, 5.6538e-2, -6.4324e-3, 8.3723e-4, -0.12183, 5.5232e-2, -4.9580e-3, 4.5512e-4, // 175
    21.5948, -2.02239e-2, -1.72029e-5, 2.83017e-8, -8.94486e-12, -0.11460, 5.3178e-2, -6.0436e-3, 7.7982e-4, -0.11423, 5.1994e-2, -4.6876e-3, 4.2548e-4, // 200
    -103.303, 0.391488, -5.15204e-4, 2.91844e-7, -6.05856e-11, -0.10524, 4.8551e-2, -5.3567e-3, 6.8809e-4, -0.10493, 4.7573e-2, -4.1711e-3, 3.7068e-4, // 250
};

static double KCoefLowTab[] =
{
    //    k0          k1           k2          k3         k4
    0.404286, -6.53925e-3, 3.04751e-5, -3.34115e-8, 5.74523e-12, // 75
    0.046976, -5.88767e-4, 1.19128e-7, 2.08445e-8, -2.50236e-11, // 100
    0.0218454, 2.8609e-4, -5.94913e-6, 2.8003e-8, -2.47774e-11, // 125
    -0.0555525, 1.43513e-3, -1.11906e-5, 3.47539e-8, -2.6852e-11, // 150
    -0.171068, 3.18684e-3, -1.86472e-5, 4.44791e-8, -3.02625e-11, // 175
    -0.156258, 2.78232e-3, -1.56926e-5, 3.48516e-8, -2.19024e-11, // 200
    -0.27906, 4.37978e-3, -2.13459e-5, 4.02871e-8, -2.28034e-11, // 250
};

static double KCoefHighTab[] =
{
    //    k0          k1           k2          k3         k4
    8.91415, -2.46654e-2, 2.60909e-5, -1.24141e-8, 2.22751e-12, // 75
    7.95597, -1.92438e-2, 1.70933e-5, -6.43059e-9, 8.20432e-13, // 100
    8.88844, -2.06477e-2, 1.73431e-5, -5.97542e-9, 6.38979e-13, // 125
    3.2744, -5.4588e-4, -8.38537e-6, 8.1867e-9, -2.21511e-12, // 150
    -2.33606, 1.95394e-2, -3.40863e-5, 2.2329e-8, -5.06401e-12, // 175
    -3.76622, 2.51806e-2, -4.13477e-5, 2.62237e-8, -5.82154e-12, // 200
    -14.9992, 6.24112e-2, -8.64019e-5, 5.00368e-8, -1.04833e-11, // 250
};

static double LCoefLowTab[] =
{
    //    l0          l1           l2          l3         l4
    -0.988325, 9.43604e-3, -2.71945e-5, 1.1954e-7, -1.36507e-10, // 75
    -0.975185, 8.40266e-3, -1.27491e-5, 6.22831e-8, -7.19493e-11, // 100
    -0.759054, 5.26528e-3, 6.39465e-6, 1.09718e-8, -2.62991e-11, // 125
    -1.37655, 1.34777e-2, -2.5876e-5, 5.50738e-8, -4.39172e-11, // 150
    -1.21265, 1.14399e-2, -1.47726e-5, 2.78799e-8, -2.26938e-11, // 175
    -1.54396, 1.57732e-2, -3.04193e-5, 4.59684e-8, -2.84172e-11, // 200
    -1.67664, 1.77194e-2, -3.69498e-5, 5.09134e-8, -2.82878e-11, // 250
};

static double LCoefHighTab[] =
{
    //    l0          l1           l2          l3         l4
    45.1992, -1.56873e-1, 2.0707e-4, -1.2016e-7, 2.58235e-11, // 75
    59.8733, -1.98084e-1, 2.49366e-4, -1.38394e-7, 2.85472e-11, // 100
    68.3614, -2.11828e-1, 2.49462e-4, -1.2945e-7, 2.49748e-11, // 125
    55.2418, -1.53167e-1, 1.61763e-4, -7.49391e-8, 1.28582e-11, // 150
    59.6039, -1.57958e-1, 1.59472e-4, -7.08188e-8, 1.16949e-11, // 175
    23.9618, -3.42003e-2, 3.40553e-6, 1.4565e-8, -5.50719e-12, // 200
    -15.5728, 9.36704e-2, -1.49036e-4, 9.42151e-8, -2.0961e-11, // 250
};

static double* LowTables[] =
    {ACoefLowTab, BCoefLowTab, CCoefLowTab, DCoefLowTab, ECoefLowTab, KCoefLowTab, LCoefLowTab};
static double* HighTables[] =
    {ACoefHighTab, BCoefHighTab, CCoefHighTab, DCoefHighTab, ECoefHighTab, KCoefHighTab, LCoefHighTab};

#define F0_AMOUNT (sizeof(LowTables) / sizeof(*LowTables))

static unsigned CfLen[] =
{
    sizeof(ACoefLowTab) / sizeof(*ACoefLowTab) / F0_AMOUNT,
    sizeof(BCoefLowTab) / sizeof(*BCoefLowTab) / F0_AMOUNT,
    sizeof(CCoefLowTab) / sizeof(*CCoefLowTab) / F0_AMOUNT,
    sizeof(DCoefLowTab) / sizeof(*DCoefLowTab) / F0_AMOUNT,
    sizeof(ECoefLowTab) / sizeof(*ECoefLowTab) / F0_AMOUNT,
    sizeof(KCoefLowTab) / sizeof(*KCoefLowTab) / F0_AMOUNT,
    sizeof(LCoefLowTab) / sizeof(*LCoefLowTab) / F0_AMOUNT,
};

static const double* GetCoefs(double h, int find, int cf)
{
    double hr = GetHeightLevel(find, cf);
    double* table = (h >= hr) ? HighTables[cf] : LowTables[cf];
    return table + CfLen[cf] * find;
}

static double ADCoefs[] =
    {-1.66718e-2, -2.44075e-3, 3.08392e-6, 2.90115e-6, -4.99606e-8, 3.36327e-10, -1.0966e-12, 1.73227e-15, -1.06271e-18};

double Gost2000Atmosphere::Density(const TimeJD& stm, const Vect3& v,
                                   const Vect3& vSun) const
{
    TAssert(Gsf != NULL);

    double rv, h, f0;
    int ind;
    rv = v.Len();
    h = v.v3 / rv;

    double R = GetAstroBase()->GetEarthRadius();
    double ALPHA = GetAstroBase()->GetEarthOblatness();

    h = (rv - R * (1.0 - ALPHA * h * h)) * 1000.0;

    if (h > 1500.0)
        return 0.0;

    if (h < 120.0)
        return LowAtmosphereDensity(h);

    TimeJD utc = stm;

    TimeJD t17 = utc;
    t17 -= 1.7;
    const TGeoSolarFlux* pgsf = Gsf->Get_F107(t17);

    if (pgsf->f81 < 200.0) {
        ind = (int)((pgsf->f81 - 62.5) / 25.0);
        if (ind < 0)
            ind = 0;
        f0 = 75.0 + ind * 25;
    }
    else {
        ind = (int)((pgsf->f81 - 175.0) / 50.0);
        ind = (ind > 1 ? 6 : ind + 5);
        f0 = -50.0 + ind * 50;
    };
    double f107 = pgsf->f01;
    double f81 = pgsf->f81;

    const double* a = GetCoefs(h, ind, 0);
    const double* b = GetCoefs(h, ind, 1);
    const double* c = GetCoefs(h, ind, 2);
    const double* d = GetCoefs(h, ind, 3);
    const double* e = GetCoefs(h, ind, 4);
    const double* k = GetCoefs(h, ind, 5);
    const double* l = GetCoefs(h, ind, 6);

    double f1 = c[8];
    double cf1 = cos(f1);
    double sf1 = sin(f1);

    double cf = (v.v1 * (vSun.v1 * cf1 - vSun.v2 * sf1) + v.v2 * (vSun.v2 * cf1 + vSun.v1 * sf1) + v.v3 * vSun.v3)
        / (rv * vSun.Len());
    double cf2 = sqrt((1.0 + cf) / 2.0);

    double dd;
    TimeTLE tle(stm);
    dd = tle.Day - 1.0;

    double ad = ADCoefs[0] +
        ADCoefs[1] * dd +
        ADCoefs[2] * dd * dd +
        ADCoefs[3] * dd * dd * dd +
        ADCoefs[4] * dd * dd * dd * dd +
        ADCoefs[5] * dd * dd * dd * dd * dd +
        ADCoefs[6] * dd * dd * dd * dd * dd * dd +
        ADCoefs[7] * dd * dd * dd * dd * dd * dd * dd +
        ADCoefs[8] * dd * dd * dd * dd * dd * dd * dd * dd;

    double ah = k[0] +
        k[1] * h +
        k[2] * h * h +
        k[3] * h * h * h +
        k[4] * h * h * h * h;

    double zr = v.z / rv;
    double k5 = 1.0 + ah * zr * zr * zr * zr * zr * zr * zr * zr *
        (1.0 + zr * sin(Pi2 * (dd - 355) / 365.24));

    double k4 = e[0] +
        e[1] * h +
        e[2] * h * h +
        e[3] * h * h * h +
        e[4] * h * h * h * h;
    if (!is3HourInd) {
        TimeJD t06 = utc;
        t06 -= 0.6;
        pgsf = Gsf->Get_F107(t06);
        double kp = pgsf->kp;
        k4 *= e[5] +
            e[6] * kp +
            e[7] * kp * kp +
            e[8] * kp * kp * kp;
    }
    else {
        TimeJD t25 = utc;
        t25 -= 0.25;
        pgsf = Gsf->Get_F107(t25);
        double kpp = Gsf->Get_Kp3hourMod(t25);
        const double* et = e + 4;
        k4 *= et[5] +
            et[6] * kpp +
            et[7] * kpp * kpp +
            et[8] * kpp * kpp * kpp;
    }
    k4 += 1.0;

    double k3 = 1.0 + (f107 - f81) / (f81 + fabs(f107 - f81)) * (b[0] +
        b[1] * h +
        b[2] * h * h +
        b[3] * h * h * h +
        b[4] * h * h * h * h);

    double k2 = 1.0 + ad * (d[0] +
        d[1] * h +
        d[2] * h * h +
        d[3] * h * h * h);

    const double* n = c + 5;
    double k1 = 1.0 + pow(cf2, n[0] + n[1] * h + n[2] * h * h) * (c[0] +
        c[1] * h +
        c[2] * h * h +
        c[3] * h * h * h +
        c[4] * h * h * h * h);

    double k0 = 1.0 + (f81 - f0) / f0 * (l[0] +
        l[1] * h +
        l[2] * h * h +
        l[3] * h * h * h +
        l[4] * h * h * h * h);

    double rh = R0 * exp(a[0] +
        a[1] * h +
        a[2] * h * h +
        a[3] * h * h * h +
        a[4] * h * h * h * h +
        a[5] * h * h * h * h * h +
        a[6] * h * h * h * h * h * h);

    return rh * k0 * k1 * k2 * k3 * k4 * k5;
}

///////////////////////////////////////////////////////////////////////////////
