///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Динамическая модель атмосферы по ГОСТ-2004
///////////////////////////////////////////////////////////////////////////////
#include "Defines.h"
#include "Atmosphere.h"
#include "IAstroBase.h"
#include "G2004AtmCoef.h"
///////////////////////////////////////////////////////////////////////////////
// Реализация вспомогательных функций для
// динамической модели атмосферы ГОСТ-2004.
///////////////////////////////////////////////////////////////////////////////
// Получить плотность атмосферы для высот менее 120 км.
// h - высота над поверхностью Земного эллипсоида [км].
// Возвращаемое значение : плотность атмосферы [кг/м^3].
//-----------------------------------------------------------------------------
static double LowAtmosphereDensity(double h)
{
    if (h < 0)
        h = 0.0;

    const G2004AtmCoef::LowAtmCoefStruct& LowAtmCoef =
        G2004AtmCoef::getLowAtmCoef(h);

    double dh = h - LowAtmCoef.h;

    return LowAtmCoef.a0 *
        exp(LowAtmCoef.k1 * dh + LowAtmCoef.k2 * dh * dh);
}

//-----------------------------------------------------------------------------
// Получить плотность ночной атмосферы.
// h   - высота над поверхностью Земного эллипсоида [км].
// f81 - средневзвешенный индекс солнечной активности за 81 сутки.
// mod - флаг необходимости использования модификации ГОСТ-2004
//       для обеспечения непрерывности зависимости плотности
//       атмосферы от средневзвешенного индекса солнечной активности:
//       true - модифицировать ГОСТ-2004, false - не модифицировать.
// ind - порядковый номер фиксированного уровня солнечной активности :
//       75(0), 100(1), 125(2), 150(3), 175(4), 200(5), 250(6).
// Возвращаемое значение : плотность ночной атмосферы [кг/м^3].
//-----------------------------------------------------------------------------
static double getNightDensity(
    double h,
    double f81,
    bool mod,
    int& ind)
{
    // Плотность ночной атмосферы на высоте 120 км [кг/м^3].

    static const double R0 = 1.58868e-08;

    ind = G2004AtmCoef::getF0Index(f81);

    if (mod) {
        int ind_1, ind_2;

        if (f81 > G2004AtmCoef::F0Level[ind]) {
            ind_1 = ind;

            ind_2 = (ind < ((G2004AtmCoef::NumF0Levels) - 1)) ? (ind + 1) : ((G2004AtmCoef::NumF0Levels) - 1);
        }
        else {
            ind_1 = (ind > 0) ? (ind - 1) : 0;

            ind_2 = ind;
        }

        if (ind_1 != ind_2) {
            double z_1, z_2;

            double f0_1, f0_2;

            f0_1 = G2004AtmCoef::F0Level[ind_1];

            f0_2 = G2004AtmCoef::F0Level[ind_2];

            const double* a_1 = G2004AtmCoef::getCoef_a(h, ind_1);

            const double* a_2 = G2004AtmCoef::getCoef_a(h, ind_2);

            z_1 = a_1[0] + h * (
                a_1[1] + h * (
                    a_1[2] + h * (
                        a_1[3] + h * (
                            a_1[4] + h * (
                                a_1[5] + h * (
                                    a_1[6]))))));

            z_2 = a_2[0] + h * (
                a_2[1] + h * (
                    a_2[2] + h * (
                        a_2[3] + h * (
                            a_2[4] + h * (
                                a_2[5] + h * (
                                    a_2[6]))))));

            return R0 * exp((z_1 * (f0_2 - f81) +
                z_2 * (f81 - f0_1)) / (f0_2 - f0_1));
        }
    }

    const double* a = G2004AtmCoef::getCoef_a(h, ind);

    return R0 * exp(a[0] + h * (
        a[1] + h * (
            a[2] + h * (
                a[3] + h * (
                    a[4] + h * (
                        a[5] + h * (
                            a[6])))))));
}

///////////////////////////////////////////////////////////////////////////////
// Реализация класса динамической модели атмосферы ГОСТ-2004.
///////////////////////////////////////////////////////////////////////////////
double Gost2004Atmosphere::Density(
    const TimeJD& stm,
    const Vect3& v,
    const Vect3& vSun) const
{
    TAssert(Gsf != NULL);

    //-------------------------------------------------------------------------

    TimeJD tm;

    int ind;

    double h, f107, f81, f0;

    double ro_night;

    double K0, K1, K2, K3, K4;

    tm = stm;

    //-------------------------------------------------------------------------

    // Вычисляем высоту над поверхностью Земного эллипсоида.

    double rv = v.Len();

    h = v.v3 / rv;

    double R = GetAstroBase()->GetEarthRadius();

    double ALPHA = GetAstroBase()->GetEarthOblatness();

    h = (rv - R * (1.0 - ALPHA * h * h)) * 1.0e+03;

    if (h > 1500.0) {
        return 0.0;
    }

    if (h < 120.0) {
        return LowAtmosphereDensity(h);
    }

    //-------------------------------------------------------------------------

    // Определяем из таблицы индексы солнечной активности.

    TimeJD t170 = tm;

    t170 -= 1.7;

    const TGeoSolarFlux* pgsf = Gsf->Get_F107(t170);

    f107 = pgsf->f01;

    f81 = pgsf->f81w;

    ro_night = getNightDensity(h, f81, fmod, ind);

    f0 = G2004AtmCoef::F0Level[ind];

    //-------------------------------------------------------------------------

    const double* b = G2004AtmCoef::getCoef_b(h, ind);
    const double* c = G2004AtmCoef::getCoef_c(h, ind);
    const double* d = G2004AtmCoef::getCoef_d(h, ind);
    const double* e = G2004AtmCoef::getCoef_e(h, ind);
    const double* l = G2004AtmCoef::getCoef_l(h, ind);

    //-------------------------------------------------------------------------

    K4 = e[0] + h * (
        e[1] + h * (
            e[2] + h * (
                e[3] + h * (
                    e[4]))));

    if (!is3HourInd) // суточные индексы
    {
        TimeJD t060 = tm;

        t060 -= 0.6;

        double Kp = Gsf->Get_F107(t060)->kp;

        K4 *= (e[5] + Kp * (
            e[6] + Kp * (
                e[7] + Kp * (
                    e[8]))));
    }
    else // трехчасовые индексы
    {
        TimeJD t025 = tm;

        t025 -= 0.25;

        double kpp = Gsf->Get_Kp3hourMod(t025);

        K4 *= (e[9] + kpp * (
            e[10] + kpp * (
                e[11] + kpp * (
                    e[12]))));
    }

    //-------------------------------------------------------------------------

    K3 = (f107 - f81) / (f81 + fabs(f107 - f81)) *
    (b[0] + h * (
        b[1] + h * (
            b[2] + h * (
                b[3] + h * (
                    b[4])))));

    //-------------------------------------------------------------------------

    TimeTLE tle(tm);

    double dd = tle.Day - 1.0;

    const double* ADCoefs = G2004AtmCoef::getCoef_A();

    double Ad = ADCoefs[0] + dd * (
        ADCoefs[1] + dd * (
            ADCoefs[2] + dd * (
                ADCoefs[3] + dd * (
                    ADCoefs[4] + dd * (
                        ADCoefs[5] + dd * (
                            ADCoefs[6] + dd * (
                                ADCoefs[7] + dd * (
                                    ADCoefs[8]))))))));

    K2 = (d[0] + h * (
        d[1] + h * (
            d[2] + h * (
                d[3] + h * (
                    d[4]))))) * Ad;

    //-------------------------------------------------------------------------

    double cf1 = cos(c[8]);

    double sf1 = sin(c[8]);

    double cf = (v.v1 * (vSun.v1 * cf1 - vSun.v2 * sf1) +
        v.v2 * (vSun.v2 * cf1 + vSun.v1 * sf1) +
        v.v3 * (vSun.v3)) / (rv * vSun.Len());

    double cf2 = sqrt((1.0 + cf) / 2.0);

    K1 = pow(cf2, c[5] + c[6] * h + c[7] * h * h) *
    (c[0] + h * (
        c[1] + h * (
            c[2] + h * (
                c[3] + h * (
                    c[4])))));

    //-------------------------------------------------------------------------

    K0 = 1.0 + (l[0] + h * (
        l[1] + h * (
            l[2] + h * (
                l[3] + h * (
                    l[4]))))) * (f81 - f0) / f0;

    //-------------------------------------------------------------------------

    return ro_night * K0 * (1.0 + K1 + K2 + K3 + K4);

    //-------------------------------------------------------------------------
}

//-----------------------------------------------------------------------------
double Gost2004Atmosphere::NightDensity(double h, double f81) const
{
    int ind;

    // Перевод [тыс.км] в [км].

    return getNightDensity(h * 1.0e+03, f81, fmod, ind);
}

///////////////////////////////////////////////////////////////////////////////
