///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Коэффициенты динамической модели атмосферы ГОСТ-2004.
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_G2004ATMCOEF_H
#define BNOTOOL_G2004ATMCOEF_H

///////////////////////////////////////////////////////////////////////////////
// Класс, обеспечивающий доступ к коэффициентам
// динамической модели атмосферы ГОСТ-2004.
///////////////////////////////////////////////////////////////////////////////
class G2004AtmCoef
{
private:
    // Конструктор.
    G2004AtmCoef();
public:
    // Число фиксированных уровней солнечной активности F0 в ГОСТ-2004.
    enum { NumF0Levels = 7 };

    // Фиксированные уровни солнечной активности F0 в ГОСТ-2004 :
    // 75(0), 100(1), 125(2), 150(3), 175(4), 200(5), 250(6).
    // В скобках указан порядковый индекс фиксированных
    // уровней солнечной активности.
    static const double F0Level[NumF0Levels];
    // Получить порядковый индекс ближайшего фиксированного
    // уровня солнечной активности.
    // f81 - cредневзвешенный индекс солнечной активности.
    // Возвращаемое значение : порядковый индекс ближайшего
    // фиксированного уровня солнечной активности.
    static int getF0Index(double f81);

    // Cтруктура коэффициентов для расчета
    // плотности атмосферы на высотах менее 120 км.
    struct LowAtmCoefStruct
    {
        double h;
        double a0;
        double k1;
        double k2;
    };

    // Получить коэффициенты для расчета плотности
    // атмосферы на высотах менее 120 км.
    // h - Высота над поверхностью Земного эллипсоида [км].
    // Возвращаемое значение : ссылка на структуру коэффициентов.
    static const LowAtmCoefStruct& getLowAtmCoef(double h);
    // Получить коэффициенты : a0, a1, a2, a3, a4, a5, a6.
    // h   - Высота над поверхностью земного эллипсоида [км].
    // ind - Порядковый индекс фиксированного уровня
    //       солнечной активности (нумерация с нуля).
    // Возвращаемое значение : указатель на массив коэффициентов.
    static const double* getCoef_a(double h, int ind);
    // Получить коэффициенты : b0, b1, b2, b3, b4.
    // h   - Высота над поверхностью земного эллипсоида [км].
    // ind - Порядковый индекс фиксированного уровня
    //       солнечной активности (нумерация с нуля).
    // Возвращаемое значение : указатель на массив коэффициентов.
    static const double* getCoef_b(double h, int ind);
    // Получить коэффициенты : c0, c1, c2, c3, c4, n0, n1, n2, phi1.
    // h   - Высота над поверхностью земного эллипсоида [км].
    // ind - Порядковый индекс фиксированного уровня
    //       солнечной активности (нумерация с нуля).
    // Возвращаемое значение : указатель на массив коэффициентов.
    static const double* getCoef_c(double h, int ind);
    // Получить коэффициенты : d0, d1, d2, d3, d4.
    // h   - Высота над поверхностью земного эллипсоида [км].
    // ind - Порядковый индекс фиксированного уровня
    //       солнечной активности (нумерация с нуля).
    // Возвращаемое значение : указатель на массив коэффициентов.
    static const double* getCoef_d(double h, int ind);
    // Получить коэффициенты : e0, e1, e2, e3, e4, e5, e6, e7, e8, et5, et6, et7, et8.
    // h   - Высота над поверхностью земного эллипсоида [км].
    // ind - Порядковый индекс фиксированного уровня
    //       солнечной активности (нумерация с нуля).
    // Возвращаемое значение : указатель на массив коэффициентов.
    static const double* getCoef_e(double h, int ind);
    // Получить коэффициенты : l0, l1, l2, l3, l4.
    // h   - Высота над поверхностью земного эллипсоида [км].
    // ind - Порядковый индекс фиксированного уровня
    //       солнечной активности (нумерация с нуля).
    // Возвращаемое значение : указатель на массив коэффициентов.
    static const double* getCoef_l(double h, int ind);
    // Получить коэффициенты аппроксимации зависимости
    // полугодового эффекта от времени года (9 штук).
    // Возвращаемое значение : указатель на массив коэффициентов.
    static const double* getCoef_A();
};

///////////////////////////////////////////////////////////////////////////////
#endif // BNOTOOL_G2004ATMCOEF_H
///////////////////////////////////////////////////////////////////////////////
