///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Класс географических координат
///////////////////////////////////////////////////////////////////////////////
#include "GeoCoords.h"
#include "Consts.h"
#include "IAstroBase.h"
///////////////////////////////////////////////////////////////////////////////
GeoCoords::GeoCoords()
{
    IAstroBase* ab = GetAstroBase();

    Re = ab->GetEarthRadius();
    Alpha = ab->GetEarthOblatness();
}

GeoCoords::GeoCoords(double EarthRadius, double EarthOblatness)
{
    Re = EarthRadius;
    Alpha = EarthOblatness;
}

GeoCoords::GeoCoords(const IGravityModel& model)
{
    Re = model.Re;
    Alpha = model.alpha;
}

//1/298.257223563

//------------------------------------------------------------------------------
// Установка координат по декартовым
//------------------------------------------------------------------------------
void GeoCoords::SetXYZ(const Vect3& src)
{
    double x2 = src.x * src.x, y2 = src.y * src.y, z2 = src.z * src.z;
    double r_xy = sqrt(x2 + y2);
    double r2 = x2 + y2 + z2;
    double r_abs = sqrt(r2);

    H = r_abs - Re * (1.0 - Alpha * z2 / r2);

    Phi = atan(src.z / r_xy);
    Phi += 2.0 * Alpha * (1.0 - H / r_abs) * src.z * r_xy / r2;

    Lam = atan2(src.y / r_xy, src.x / r_xy);
    if (Lam < 0)
        Lam += Pi2;
}

//------------------------------------------------------------------------------
// Расстояние от центра эллипсоида до его поверхности
//------------------------------------------------------------------------------
double GeoCoords::SDistance() const
{
    double sin_lat = sin(Phi);
    return Re / sqrt(1.0 - (2.0 - Alpha) * Alpha * sin_lat * sin_lat);
}

//------------------------------------------------------------------------------
// Получение декартовых координат
//------------------------------------------------------------------------------
void GeoCoords::GetXYZ(Vect3& dst) const
{
    double sd = SDistance();
    double ch = H + sd;
    double cos_long = cos(Lam);
    double sin_long = sin(Lam);
    double cos_lat = cos(Phi);
    double sin_lat = sin(Phi);
    double one_flatt = 1.0 - Alpha;
    one_flatt *= one_flatt;

    dst.x = ch * cos_lat * cos_long;
    dst.y = ch * cos_lat * sin_long;
    dst.z = (one_flatt * sd + H) * sin_lat;
}

//------------------------------------------------------------------------------
// Установка координат по декартовым
//------------------------------------------------------------------------------
void GeoCoordsPrec::SetXYZ(const Vect3& src)
{
    double rxy = sqrt(src.x * src.x + src.y * src.y);
    double r = src.Len();

    Lam = atan2(src.y, src.x);

    if (rxy / Re < 1.6e-6) {
        Phi = asin(src.z / r);
        H = r - Re * (1 - Alpha * (src.z / r) * (src.z / r));
    }
    else {
        double b = Re * (1 - Alpha);
        if (src.z < 0)
            b = -b;
        double e = ((src.z + b) * b / Re - Re) / rxy;
        double f = ((src.z - b) * b / Re + Re) / rxy;
        double p = 4.0 * (e * f + 1.0) / 3.0;
        double q = 2.0 * (e * e - f * f);
        double d = p * p * p + q * q;
        double v;
        if (d >= 0) {
            double s = sqrt(d) + q;
            double s3 = pow(fabs(s), 1.0 / 3.0);
            if (s < 0)
                s3 = -s3;
            v = p / s3 - s3;
            v = -(2.0 * q + v * v * v) / 3.0 / p;
        }
        else
            v = 2.0 * sqrt(-p) * cos(acos(q / p / sqrt(-p)) / 3.0);
        double g = (e + sqrt(e * e + v)) * 0.5;
        double t = sqrt(g * g + (f - v * g) / (2.0 * g - e)) - g;
        Phi = atan(Re * (1 - t * t) / (2.0 * b * t));
        H = (rxy - Re * t) * cos(Phi) + (src.z - b) * sin(Phi);
    }
}

///////////////////////////////////////////////////////////////////////////////
