///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Класс гравитационных моделей Земли
///////////////////////////////////////////////////////////////////////////////
#include "SGravity.h"
#include "Defines.h"
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
//-----------------------------------------------------------------------------
GravityModel::GravityModel()
{
    initEmpty(grIdNONE);
};
//-----------------------------------------------------------------------------
// Конструктор, создающий модель копированием ограниченного
// числа гармоник из другой модели
//-----------------------------------------------------------------------------
GravityModel::GravityModel(TGravityModelId Id_, unsigned int N,
                           unsigned int M, const IGravityModel* c_d)
{
    initEmpty(Id_);
    Init(N, M, c_d);
};
//-----------------------------------------------------------------------------
// Инициализация пустой модели
//-----------------------------------------------------------------------------
void GravityModel::initEmpty(TGravityModelId Id_)
{
    Id = Id_;
    N_ = 0;
    M_ = 0;
    CC = NULL;
    DD = NULL;
    isInit = false;
};
//-----------------------------------------------------------------------------
// Очистка модели
//-----------------------------------------------------------------------------
void GravityModel::freeAll()
{
    unsigned int i;

    if (CC) {
        for (i = 0; i <= M_; i++)
            delete[] CC[i];
        delete[] CC;
    };

    if (DD) {
        for (i = 0; i <= M_; i++)
            delete[] DD[i];
        delete[] DD;
    };

    initEmpty(Id);
};
//-----------------------------------------------------------------------------
// обнуление всех коэффициентов
//-----------------------------------------------------------------------------
void GravityModel::null()
{
    unsigned int i, i1;

    for (i = 0; i <= M_; i++) {
        for (i1 = 2; i1 <= N_; i1++) {
            CC[i][i1] = 0.0;
            DD[i][i1] = 0.0;
        };
    };
};
//-----------------------------------------------------------------------------
// Инициализация модели копированием ограниченного
// числа гармоник из другой модели
//-----------------------------------------------------------------------------
void GravityModel::Init(unsigned int N, unsigned int M,
                        const IGravityModel* c_d)
{
    TAssert(N <= c_d->N() && M <= c_d->M());
    mu = c_d->mu;
    Re = c_d->Re;
    alpha = c_d->alpha;
    Id = Id;

    SetSize(N, M);

    for (unsigned int n = 2; n <= N; n++) {
        for (unsigned int m = 0; m <= M; m++) {
            CC[m][n] = c_d->c(n, m);
            if (m > 0)
                DD[m][n] = c_d->d(n, m);
        };
    };

    isInit = true;
};
//-----------------------------------------------------------------------------
// печать содержимого модели
//-----------------------------------------------------------------------------
void GravityModel::print(char* comment, FILE* out) const
{
    unsigned int M1, n, m;
    fprintf(out, "%s", comment);

    if (!isInit) {
        fprintf(out, " - модель не инициализирована\n");
        return;
    };

    fprintf(out, " параметры модели: N=%3d, M=%3d; мю = %19.12e, экв.радиус =%19.12e, сжатие =%19.12e",
            N_, M_, mu, Re, alpha);
    for (n = 2; n <= N_; n++) {
        fprintf(out, "\n%3d ", n);
        M1 = (n < M_ ? n : M_);
        for (m = 0; m <= n; m++) {
            if ((m > 0) && ((m % 10) == 0))
                fprintf(out, "\n    ");
            fprintf(out, "%3d:%10.2e", m, c(n, m));
        };
        fprintf(out, "\n                  ");
        for (m = 1; m <= M1; m++) {
            if ((m > 0) && ((m % 10) == 0))
                fprintf(out, "\n    ");
            fprintf(out, "%3d:%10.2e", m, d(n, m));
        };
    };
    fprintf(out, "\n");
};
//-----------------------------------------------------------------------------
// установить размерность модели (все данные при этом пропадут!!!)
//-----------------------------------------------------------------------------
void GravityModel::SetSize(unsigned int N, unsigned int M)
{
    if ((N == N_) && (M == M_))
        return;

    unsigned int i;
    if (N != N_) {
        if (CC) {
            for (i = 0; i <= M_; i++) {
                delete[] CC[i];
                delete[] DD[i];
            };
        };
    };

    if (M != M_) {
        if (CC)
            delete[] CC;
        CC = new double*[M + 1];
        if (DD)
            delete[] DD;
        DD = new double*[M + 1];
    };

    for (i = 0; i <= M; i++) {
        CC[i] = new double[N + 1];
        DD[i] = new double[N + 1];
    };

    N_ = N;
    M_ = M;
    isInit = false;
};
///////////////////////////////////////////////////////////////////////////////
