///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Интерфейсы доступа к астробаллистическим данным и алгоритмам
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_IASTROBASE_H
#define BNOTOOL_IASTROBASE_H
///////////////////////////////////////////////////////////////////////////////
#include "TM.h"
#include "DVectors.h"
#include "Consts.h"
#include "Ids.h"
#include "IModel.h"
#include "TMScale.h"

// Базовый класс астробаллистической базы
///////////////////////////////////////////////////////////////////////////////
class IAstroBase
{
public:
    virtual ~IAstroBase()
    {
    };

    // Преобразование времени между шкалами
    virtual void ConvertTimeScaleJD(TmScale src, TmScale dst, int& d,
                                    double& f) const = 0;

    // Получение матрицы перехода от J2000 к METE
    virtual void MtxJ20002METE(const TimeJD& jdMETE, TransMtx& mtx) const = 0;
    // Вычисление истинного звездного времени
    virtual double SaTime(const TimeJD& jd) const = 0;
    // Вычисление истинного звездного времени при известной поправке
    virtual double SaTime(const TimeJD& jd, double ds) const = 0;
    // Вычисление среднего звездного времени
    virtual double Smean(const TimeJD& dt) const = 0;
    // Вычисление матрицы перехода от ECI к фиксированному Гринвичу
    virtual void MtxECI2GFCS(const TimeJD& jdECI, const TimeJD& jdGFCS,
                             TransMtx& mtx) const = 0;
    // Вычисление матрицы, учитывающей прецессию оси вращения Земли
    virtual double MtxPrecession(const TimeJD& dtStart, const TimeJD& dtEnd,
                                 TransMtx& mtx) const = 0;
    // Получение матрицы перехода от METE к Гринвичу
    virtual void MtxMETE2GCS(const TimeJD& t, TransMtx& mtx) const = 0;

    // Получение указателя на модель планеты
    virtual IPlanet* GetPlanet(TPlanetId pl) = 0;
    virtual const IPlanet* GetPlanet(TPlanetId pl) const = 0;

    // Получение положений планет и тел Солнечной системы
    virtual void GetPosGelio(TPlanetId pl, const TimeJD& t, Vect3* crd, Vect3* vel = NULL,
                             Vect3* acc = NULL) const = 0;
    // Получение положений планет и тел Солнечной системы
    virtual void GetPosGeo(TPlanetId pl, const TimeJD& t, Vect3* crd, Vect3* vel = NULL,
                           Vect3* acc = NULL) const = 0;
    // Вычисление вектора сдвига при переходе от гелиоцентрической системы
    // к геоцентрической
    void GetGelio2GeoShift(const Vect3& vEarthMoon, const Vect3& vMoon, Vect3& result) const;

    // Вычисление матрицы перехода от J2000 к фиксированному Гринвичу
    void MtxJ20002GFCS(const TimeJD& jdGFCS, TransMtx& mtx) const
    {
        TimeJD tf(0, doubleJ2000);
        MtxECI2GFCS(tf, jdGFCS, mtx);
    };
    // Вычисление матрицы скоростей при переходе от фиксированного Гринвича
    // к вращающемуся
    void GCSVelocityMatrix(TransMtx& m) const;
    // Вычисление матрицы перехода от Гринвича к системе NEZ
    void MtxGCS2NZE(const Vect3& p, TransMtx& mtx) const;

    // Получение параметров Земли 
    // Скорость вращения [рад / тыс. сек.]
    double GetEarthRotation() const
    {
        return GetPlanet(plIdEarth)->omega;
    };
    // Коэффициент сжатия
    double GetEarthOblatness() const
    {
        return GetPlanet(plIdEarth)->Geometry.alpha;
    };
    // Экваториальный радиус
    double GetEarthRadius() const
    {
        return GetPlanet(plIdEarth)->Geometry.Re;
    };
    // Экваториальный радиус
    double GetEarthMu() const
    {
        return GetPlanet(plIdEarth)->GetGravityModel()->mu;
    };
    // Модель атмосферы
    const IAtmosphereModel* GetEarthAtmosphere() const
    {
        return GetPlanet(plIdEarth)->GetAtmosphere();
    };

protected:
    // Вычисление матрицы поворота системы координат в плоскости 
    // истинного экватора на угол dt
    void MtxRotation(double dt, TransMtx& mtx) const;
    // Истинное звездное время в радианах
    // На входе Гринвичское среднее звездное время [рад] и
    // поправка к гринвичскому среднему звездному времени [рад]
    double Sa(double Smean, double dSmean)
    {
        return Smean + dSmean;
    }
};

///////////////////////////////////////////////////////////////////////////////
// Функции установки и получения текущего экземпляра астробаллистической базы
///////////////////////////////////////////////////////////////////////////////
// Установка
void SetAstroBase(IAstroBase& base);
// Получение 
IAstroBase* GetAstroBase();
///////////////////////////////////////////////////////////////////////////////
#endif
