///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Определение систем отсчета и их параметров
////////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_IFRAMES_H
#define BNOTOOL_IFRAMES_H
////////////////////////////////////////////////////////////////////////////////
#include "Ids.h"
#include "FrameConverter.h"
#include "TM.h"

////////////////////////////////////////////////////////////////////////////////
// Интерфейс параметров систем отсчета
////////////////////////////////////////////////////////////////////////////////
class IFrameParameters
{
public:
    // Деструктор
    virtual ~IFrameParameters()
    {
    };
    // Возвращение типа системы отсчета
    virtual TCoordSysId GetFrameType() const = 0;
    // Проверка на совпадение параметров
    virtual bool IsEqual(const IFrameParameters& c) const = 0;
};

////////////////////////////////////////////////////////////////////////////////
// Интерфейс системы отсчета
////////////////////////////////////////////////////////////////////////////////
class IFrame
{
public:
    // Деструктор
    virtual ~IFrame()
    {
    };

    // Получение ссылки на параметры системы
    virtual const IFrameParameters& GetParameters() const = 0;
    // Установка параметров системы
    virtual void SetParameters(const IFrameParameters& p) = 0;
    // Является ли система инерциальной ?
    virtual bool IsInertial() const = 0;
    // Является ли система геоцентрической ?
    virtual bool IsGeocentric() const = 0;

    // Возвращение типа системы отсчета
    TCoordSysId GetType() const
    {
        return GetParameters().GetFrameType();
    };
    // Проверка совпадения систем отсчета
    bool IsEqual(const IFrame& rf) const
    {
        return GetParameters().IsEqual(rf.GetParameters());
    };

    // Получение объекта, который соответствует переходу от текущей системы отсчета 
    // к системе отсчета J2000 на момент времени t0 (для 3-мерных векторов)
    virtual void Get3DJ2000Converter(const TimeJD& t0, FrameConverter3D& result) const = 0;
    // Получение объекта, который соответствует переходу от текущей системы отсчета 
    // к системе отсчета J2000 на момент времени t0 (для 6-мерных векторов)
    virtual void Get6DJ2000Converter(const TimeJD& t0, FrameConverter6D& result) const = 0;

    // Получение объекта, который соответствует переходу от текущей системы отсчета 
    // к системе отсчета rf на момент времени t0 (для 3-мерных векторов)
    virtual void Get3DConverter(const IFrame& rf, const TimeJD& t0,
                                FrameConverter3D& result) const;
    // Получение объекта, который соответствует переходу от текущей системы отсчета 
    // к системе отсчета rf на момент времени t0 (для 6-мерных векторов)
    virtual void Get6DConverter(const IFrame& rf, const TimeJD& t0, FrameConverter6D& result) const;
};

////////////////////////////////////////////////////////////////////////////////
IFrame* CreateIFrame(const IFrameParameters& p);
////////////////////////////////////////////////////////////////////////////////
#endif
