///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Интерфейс измерений
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_IMEASUREMENT_H
#define BNOTOOL_IMEASUREMENT_H
///////////////////////////////////////////////////////////////////////////////
#include "ITraject.h"
#include <string>

///////////////////////////////////////////////////////////////////////////////
// Типы измерений.
///////////////////////////////////////////////////////////////////////////////
// Неизвестный тип.
#define mtUnknown 0
// 6-мерный вектор прямоугольных координат и скоростей в
// некоторой системе отсчета.
#define mtOrt 1
// Оптическое измерение.
#define mtOptic 2
// Априорная информация в виде элементов Кеплера.
#define mtKApriori 3
// Радиолокационные измерения.
#define mtRLocation 4
// Радиотехнические измерения КАМЫ
#define mtRTKama 5
// Зарезервированные значения для пользовательских типов измерений
#define mtRTKamaDt 6
// йЪНЕТЕОЙЕ Ч ЧЙДЕ TLE
#define mtTLE 7
// ъБТЕЪЕТЧЙТПЧБООЩЕ ЪОБЮЕОЙС ДМС РПМШЪПЧБФЕМШУЛЙИ ФЙРПЧ ЙЪНЕТЕОЙК
#define mtUser 1000
///////////////////////////////////////////////////////////////////////////////
// Класс определяющий интерфейс измерения
// Реализованы наиболее общие методы для всех типов измерений
///////////////////////////////////////////////////////////////////////////////
#include "ObjectId.h"
#include "MeasId.h"
#include "SiteId.h"

class IMeasurement
{
public:
    // Конструктор по умолчанию
    IMeasurement();
    // Деструктор.
    virtual ~IMeasurement()
    {
    };
    // Метод возвращает размерность вектора измерений
    virtual unsigned GetSize() const = 0;
    // Метод возвращает тип измерения
    virtual int GetType() const
    {
        return Type;
    };
    // Метод возвращает время измерения
    virtual const TimeJD& GetTime() const
    {
        return MTime;
    };
    // Метод возвращает в объекте v вектор измерений
    virtual void GetRealVector(DVector& v) const = 0;
    // Метод возвращает в объекте v вектор СКО измерений
    virtual void GetMSEVector(DVector& v) const = 0;
    // Метод возвращает идентификатор измерения
    virtual TMeasId GetID() const
    {
        return Id;
    };
    // Метод возвращает идентификатор наблюдательной станции,
    // к которой относится измерение.
    // Метод возвращает 0, если станция не определена.
    virtual TSiteId GetStationID() const
    {
        return StId;
    };

    virtual TNightId GetNightID() const
    {
        return NightId;
    };

    // Метод возвращает идентификатор космического объекта,
    // к которому относится измерение.
    // Метод возвращает 0, если объект не определен или не известен.
    virtual TObjectId GetObjectID() const
    {
        return ObjId;
    };

    virtual TPacketId GetPacketID() const
    {
        return PacketId;
    };
    // Метод устанавливает время измерения
    virtual void SetTime(const TimeJD& t)
    {
        MTime = t;
    };
    // Метод устанавливает идентификатор измерения
    virtual void SetID(TMeasId id)
    {
        Id = id;
    };
    // Метод устанавливает идентификатор наблюдательной станции,
    // к которой относится измерение
    virtual void SetStationID(TSiteId id)
    {
        StId = id;
    };

    virtual void SetNightID(TNightId id)
    {
        NightId = id;
    };
    // Метод устанавливает идентификатор космического объекта,
    // к которому относится измерение
    virtual void SetObjectID(TObjectId id)
    {
        ObjId = id;
    };
    // Метод возвращает в переменной p весовую матрицу измерения,
    virtual void SetPacketID(TPacketId id)
    {
        PacketId = id;
    };
    // нЕФПД ЧПЪЧТБЭБЕФ Ч РЕТЕНЕООПК p ЧЕУПЧХА НБФТЙГХ ЙЪНЕТЕОЙС,
    // которая является квадратной симметричной матрицей
    virtual void GetWeightMatrix(DSymmMatrix& p) const;
    // Метод осуществляет моделирование измерения для космического объекта,
    // движение которого описывается с помощью переменной trj.
    // Метод вычисляет следующие параметры:
    // t	-	время, которое соответствует положению космического объекта
    //      на орбите для текущего измерения (это время равно времени измерения,
    //      если не учитывается запаздывание сигнала)
    // fr	-	система отсчета, в которой задаются положение и скорости аппарата
    //      в момент измерения, а также матрица частных производных (см. далее параметры f и m)
    // v	-	указатель вектор моделируемого измерения. Если этот параметр равен 0,
    //      вектор моделируемых измерений не вычисляется
    // m	-	матрица производных размерности Nx6, где N - размерность вектора измерений
    //      (т.е. матрица состоит из N строк и 6 столбцов). В столбцах расположены
    //      производные по Rx Ry Rz Vx  Vy Vz, а в строках - компоненты вектора измерений.
    //      Если этот параметр равен 0, вектор моделируемых измерений не вычисляется
    // del	-	вектор невязок измеряемых параметров. Если этот параметр равен 0,
    //        вектор невязок не вычисляется
    virtual bool Simulate(int matrixSize, ITraject& trj, TimeJD& t, const IFrame& fr, Vect6* f = 0,
                          DVector* v = 0, DMatrix* m = 0, DVector* del = 0) const = 0;

    // Если m1 или m2 == NULL, принмается текущее измерение
    virtual void Diff(DVector& d, const DVector* m1 = NULL, const DVector* m2 = NULL) const = 0;

    // Создать копию объекта с помощью new()
    virtual IMeasurement* Clone() const = 0;

    virtual bool GetIan(unsigned k) const
    {
        return false;
    };
    // Запрос аномальности компоненты
    virtual void SetIan(unsigned k, bool f = true)
    {
        return;
    };

    bool IsRearranged() const { return m_rearranged; }
    bool SetRearranged(bool value) { return m_rearranged = value; }
    bool IsDroppable() const { return m_droppable; }
    bool SetDroppable(bool value) { return m_droppable = value; }

    int counter;

    bool m_skip;
protected:
    // Идентификатор измерения
    TMeasId Id;
    // Идентификатор наблюдательной станции
    TSiteId StId;
    // Идентификатор наблюдательной ночи
    TNightId NightId;
    // Идентификатор космического объекта,
    // к которому относится измерение
    TObjectId ObjId;

    TPacketId PacketId;

    int Type;

    // Время измерения
    TimeJD MTime;

    bool m_rearranged;
    bool m_droppable;
};

///////////////////////////////////////////////////////////////////////////////
// Класс, реализующий наиболее общие методы для всех типов измерений
///////////////////////////////////////////////////////////////////////////////
class AMeasurement : public IMeasurement
{
public:
    // Конструктор по умолчанию
    AMeasurement();
    // Деструктор.
    virtual ~AMeasurement();
    // Получение исходной системы отсчета
    virtual const IFrame* GetSrcFrame() const = 0;
    // Метод осуществляет моделирование измерения для космического объекта,
    // движение которого описывается с помощью переменной trj.
    // Метод вычисляет следующие параметры:
    // t	-	время, которое соответствует положению космического объекта
    //      на орбите для текущего измерения (это время равно времени измерения,
    //      если не учитывается запаздывание сигнала)
    // fr	-	система отсчета, в которой задаются положение и скорости аппарата
    //      в момент измерения, а также матрица частных производных (см. далее параметры f и m)
    // v	-	указатель вектор моделируемого измерения. Если этот параметр равен 0,
    //      вектор моделируемых измерений не вычисляется
    // m	-	матрица производных размерности Nx6, где N - размерность вектора измерений
    //      (т.е. матрица состоит из N строк и 6 столбцов). В столбцах расположены
    //      производные по Rx Ry Rz Vx  Vy Vz, а в строках - компоненты вектора измерений.
    //      Если этот параметр равен 0, вектор моделируемых измерений не вычисляется
    // del	-	вектор невязок измеряемых параметров. Если этот параметр равен 0,
    //        вектор невязок не вычисляется
    Vect3 GetRecPos() const { return RecPos; }
    const FrameConverter6D& GetFrameConverter() const { return FrameConverter; }
protected:
    // Рабочая система отсчета (которая использовалась при последнем вызове
    // метода simulate()
    //	mutable IFrame const *WrkFrame;
    // Конвертер преобразования от рабочей системы отсчета к исходную
    mutable FrameConverter6D FrameConverter;
    // Метод определяет время t, которое соответствует моменту излучения 
    // сигнала с объекта, а также его положение на орбите в этот момент времени
    virtual bool DetectPosition(ITraject& trj, TimeJD& t, const IFrame* fr = 0, Vect6* f = 0) const = 0;
    // Метод установки рабочей системы отсчета
    void SetWrkFrame(const IFrame& fr) const;

    Vect3 RecPos;

private:
};

///////////////////////////////////////////////////////////////////////////////
#endif
