///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Интерфейс пакета измерений
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_IMEASUREMENTPACKET_H
#define BNOTOOL_IMEASUREMENTPACKET_H
///////////////////////////////////////////////////////////////////////////////
#include "IMeasurement.h"
#include "IMeasurementFilter.h"

///////////////////////////////////////////////////////////////////////////////
class IMeasurementPacket
{
public:
    // Метод возвращает число измерений в пакете
    virtual ~IMeasurementPacket()
    {
    };
    // Метод возвращает число измерений в пакете
    virtual unsigned GetSize() const = 0;
    // Метод возвращает true, если пакет пуст (не содержит ни одного измерения)
    // и возвращает false в противном случае
    virtual bool IsEmpty() const = 0;
    // Метод возвращает число активных измерений в пакете
    virtual unsigned GetActiveSize() const = 0;
    // Метод возвращает true, если пакет содержит хотя бы одно активное измерение
    // и возвращает false в противном случае
    virtual bool IsActive() const = 0;
    // Метод делает текущим первое измерение пакета
    virtual void SetFirst() = 0;
    // Метод делает текущим последнее измерение пакета
    virtual void SetLast() = 0;
    // Метод делает текущим первое активное измерение пакета.
    // Если активных измерений нет, то текущее измерение становится неопределенным
    virtual void SetFirstActive() = 0;
    // Метод делает текущим последнее активное измерение пакета.
    // Если активных измерений нет, то текущее измерение становится неопределенным
    virtual void SetLastActive() = 0;
    // Метод делает текущим первое измерение пакета, время которого больше или равно 
    // указанного времени t. 
    // Если таких измерений нет, то текущее измерение становится неопределенным
    virtual void SetByTime(const TimeJD& t) = 0;
    // Метод делает текущим первое активное измерение пакета, время которого больше
    // или равно указанного времени t.
    // Если таких измерений нет, то текущее измерение становится неопределенным
    virtual void SetByTimeActive(const TimeJD& t) = 0;
    // Метод позволяет явно установить текущее измерение пакета по ссылке m.
    // Если указанное измерение отсутствует в пакете, текущее измерение становится неопределенным.
    virtual void SetCurrent(const IMeasurement& m) = 0;
    // Метод позволяtт явно установить текущее измерение пакета либо  по идентификатору измерения id.
    // Если указанное измерение отсутствует в пакете, текущее измерение становится неопределенным.
    virtual void SetCurrent(TMeasId id) = 0;
    // Метод возвращает указатель на текущее измерение пакета.
    // Если текущее измерение не определено, метод возвращает NULL.
    virtual const IMeasurement* GetCurrent() const = 0;
    virtual IMeasurement* GetCurrent() = 0;
    // Метод делает текущим измерение, которое располагается непосредственно сразу за ранее
    // установленным текущим измерением.
    // Если текущим было последнее измерение пакета,
    // то после работы метода текущее измерение становится неопределенным.
    virtual void GoNext() = 0;
    // Метод делает текущим измерение, которое располагается непосредственно перед ранее
    // установленным текущим измерением.
    // Если текущим было первое измерение пакета,
    // то после работы метода текущее измерение становится неопределенным.
    virtual void GoPrev() = 0;
    // Метод делает текущим первое активное измерение, которое расположено в пакете после
    // ранее установленного текущего измерения.
    // Если таких измерений нет, то после работы метода текущее измерение становится неопределенным.
    virtual void GoNextActive() = 0;
    // Метод делает текущим последнее активное измерение, которое расположено в пакете перед
    // ранее установленным текущим измерением.
    // Если таких измерений нет, то после работы метода текущее измерение становится неопределенным.
    virtual void GoPrevActive() = 0;
    // Метод возвращает указатель на измерение пакета, идентификатор которого равен id.
    // Если такое измерение отсутствует в пакете, то метод возвращает NULL.
    // Метод не изменяет текущего измерения пакета.
    virtual const IMeasurement* Find(TMeasId id) const = 0;
    // Метод возвращаtт true, если указанное измерение является активным (т.е. включено в обработку)
    virtual bool IsActive(const IMeasurement& m) const = 0;
    // Метод возвращаtт true, если измерение с указанным id является активным (т.е. включено в обработку)
    virtual bool IsActive(TMeasId id) const = 0;
    // Метод возвращает значение признака ручного исключения измерения m из обработки.
    // Это значение равно true, если этот признак установлен и, следовательно,
    // измерение исключено из обработки.
    virtual bool GetManualFlag(const IMeasurement& m) const = 0;
    // Метод возвращает значение признака ручного исключения измерения с
    // идентификатором id из обработки.
    // Это значение равно true, если этот признак установлен и, следовательно,
    // измерение исключено из обработки.
    virtual bool GetManualFlag(TMeasId id) const = 0;
    // Метод устанавливает значение f признака ручного исключения измерения m из обработки.
    // Если f равно true, то признак установливается и, следовательно, измерение исключено из обработки.
    virtual void SetManualFlag(const IMeasurement& m, bool f) = 0;
    // Метод устанавливает значение f признака ручного исключения измерения с
    // идентификатором id из обработки.
    // Если f равно true, то признак установливается и, следовательно, измерение исключено из обработки.
    virtual void SetManualFlag(TMeasId id, bool f) = 0;
    // Метод возвращает значение признака автоматического исключения измерения m из обработки.
    // Это значение равно true, если этот признак установлен и, следовательно,
    // измерение исключено из обработки.
    virtual bool GetAutoFlag(const IMeasurement& m) const = 0;
    // Метод возвращает значение признака автоматического исключения измерения с
    // идентификатором id из обработки.
    // Это значение равно true, если этот признак установлен и, следовательно,
    // измерение исключено из обработки.
    virtual bool GetAutoFlag(TMeasId id) const = 0;
    // Метод очищает флаги автоматического исключения из обработки для всех измерений пакета
    virtual void ClearAutoFlags() = 0;
    // Метод устанавливает объект, выполняющий автоматическое исключение измерений из обработки
    virtual void SetFilter(const IMeasurementFilter* f) = 0;
    // Метод возвращает указатель на объект, выполняющий автоматическое исключение
    // измерений из обработки
    virtual const IMeasurementFilter* GetFilter() = 0;
    // Метод расставляет признаки автоматического исключения из обработки для текущего измерения meas
    virtual void FilterMeasurement(const IMeasurement& m, const DVector& del) = 0;
    // Метод возвращает true, если число измеренных параметров в пакете
    // превышает указанное число kmin
    // При подсчете учитываются только активные измерения
    virtual bool IsEnough(unsigned kmin) = 0;

    virtual void SetDeleteFlag(const IMeasurement& m, bool f) = 0;
    virtual bool GetDeleteFlag(TMeasId id) const = 0;
    virtual bool GetDeleteFlag(const IMeasurement& m) const = 0;
};

///////////////////////////////////////////////////////////////////////////////
#endif
