///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
#ifndef IMINMNK_H
#define IMINMNK_H
///////////////////////////////////////////////////////////////////////////////
// Алгоритмы минимизации функции нескольких переменных
///////////////////////////////////////////////////////////////////////////////
#include "Matrix.h"

///////////////////////////////////////////////////////////////////////////////
// Признаки дальнейшего пути
///////////////////////////////////////////////////////////////////////////////
typedef enum
{
    UNDEFINED_PATH = 0,
    // Путь не назначен
    DESC_PATH = 1,
    // Назначен спуск
    GN_PATH = 2 // Назначено движение по Гауссу-Ньютону
} EnMinMnkPath;

///////////////////////////////////////////////////////////////////////////////
// Режимы печати для отладки
///////////////////////////////////////////////////////////////////////////////
typedef enum
{
    NONE = -1,
    // Нет печати
    TRACE = 1,
    // Короткая трасса: входные невязки и добавки
} EnMinMnkPrintReg;

///////////////////////////////////////////////////////////////////////////////
// Интерфейс функции нескольких переменных, используемый при
// минимизации
///////////////////////////////////////////////////////////////////////////////
class IMinFunct
{
public:
    // Деструктор
    virtual ~IMinFunct()
    {
    };

    // Запрос размерности аргументов функции
    virtual unsigned Size() const = 0;
    // Метод запроса вычисления функционала, градиента и
    // матрицы частных производных
    // Метод возвращает false, если при указанных аргументах
    // не удается вычислить значение функции
    virtual bool Fun(const DVector& u, double& f, DVector* df = NULL,
                     DSymmMatrix* fm = NULL) = 0;
    // Запрос приемлимости полученного решения
    virtual bool CanBeAccepted(const DVector& u) = 0;
    // Реализация операции сложения для уточняемого вектора
    virtual void AddU(const DVector& u0, const DVector& du,
                      DVector& u) const = 0;
    // Реализация алгоритма проверки сходимости
    // Возврат true, если сошлись
    virtual bool IsMin(const DVector& u, double f,
                       const DVector& u1, double f1,
                       const DSymmMatrix& s, double alpha, unsigned nit,
                       EnMinMnkPath& path) = 0;
};

///////////////////////////////////////////////////////////////////////////////
// Класс, выполняющий минимизацию функции нескольких переменных
///////////////////////////////////////////////////////////////////////////////
class DMinMnk
{
public:
    // Признаки сходимости/расходимости
    typedef enum
    {
        SUCCESS_RES = 0,
        // Достигнута сходимость
        NITER_RES = 1,
        // Вышли по исчерпанию числа итераций
        STOP_RES = 2,
        // Застряли на месте
        FUNC_RES = 3,
        // Ошибка при вычислении функционала
        BAD_IANS = 4,
        FAIL_RES = 5
    } EnRes;


    // Конструктор
    DMinMnk();
    // Деструктор
    virtual ~DMinMnk();

    // Инициаализация: все рабочие переменные алгоритма
    // приводятся в начальное состояние
    void Reset();

    // Методы установки и запроса констант алгоритма
    // Запрос максимального числа итераций
    unsigned GetMaxIter() const
    {
        return MaxIt;
    };
    // Установка максимального числа итераций
    void SetMaxIter(unsigned n)
    {
        MaxIt = n;
    };
    // Запрос максимального числа итераций спуском
    unsigned GetMaxIterD() const
    {
        return MaxItD;
    };
    // Установка максимального числа итераций спуском
    void SetMaxIterD(unsigned n)
    {
        MaxItD = n;
    };
    // Запрос максимального числа делений отрезка пополам
    unsigned GetMaxN05() const
    {
        return MaxN05;
    };
    // Установка максимального числа делений отрезка пополам
    void SetMaxN05(unsigned n)
    {
        MaxN05 = n;
    };
    // Запрос ограничения на размер шага
    double GetCn1() const
    {
        return Cn1;
    };
    // Установка ограничения на размер шага
    void SetCn1(double c)
    {
        Cn1 = c;
    };
    // Запрос ограничения на размер шага
    double GetCn2() const
    {
        return Cn2;
    };
    // Установка ограничения на размер шага
    void SetCn2(double c)
    {
        Cn2 = c;
    };
    // Запрос параметра сходимости
    double GetAlphaS() const
    {
        return AlphaS;
    };
    // Установка ограничения на размер шага
    void SetAlphaS(double a)
    {
        AlphaS = a;
    };

    // Выполнение минимизации
    // u0 - начальное приближение
    // f - минимизируемая функция
    void Minimize(const DVector& u0, IMinFunct& f);

    // Методы запроса результатов минимизации
    // Запрос признака успешности решения
    EnRes GetRes() const
    {
        return Res;
    };
    // Запрос текущего вектора
    const DVector& GetVector() const
    {
        return U;
    };
    // Запрос конечного функционала
    const double GetFk() const
    {
        return Fk;
    };
    // Запрос информационной матрицы
    const DSymmMatrix& GetSMatrix() const
    {
        return Sm;
    };
    //итоговое число операций
    const unsigned GetNit() const
    {
        return Nit;
    };

    // текущий режим печати минимизации
    EnMinMnkPrintReg RegPrint;

protected:
    // Константы алгоритма
    // Максимально допустимое число итераций
    unsigned MaxIt;
    // Максимально допустимое число итераций спуском
    unsigned MaxItD;
    // Максимально допустимое число делений отрезка пополам
    unsigned MaxN05;
    // Ограничения шага
    double Cn1, Cn2;
    // Параметр сходимости
    double AlphaS;

    // Рабочие переменные
    // Признак достижения минимума
    bool IsMin;
    // Признак дальнейшего пути
    EnMinMnkPath Path;
    // Признак удачности спуска
    bool IsDesc;
    // Общее число итераций
    unsigned Nit;
    // Число итераций спуска
    unsigned NitD;
    // Признак успешности решения
    EnRes Res;
    // Начальное и конечное значения функционала
    double F0, Fk;
    // Обратная матрица
    DSymmMatrix Sm;
    // Текущее значение уточняемого вектора
    DVector U;

    // Операции алгоритма
    // Обращение матрицы
    bool InvertMatrix(const DSymmMatrix& fm, DSymmMatrix& s);
    // Обращение матрицы с использованием регуляризации и нормировки
    bool InvertByReg(const DSymmMatrix& fm, double r, DSymmMatrix& s);
    // Реализация TA ПАРАБОЛА
    void Parabola(double f0, double f1, double fp0, double fp1, double kc0,
                  double& umin, double& fmin);
    // Реализация одного шага TA СПУСК
    bool DescOneStep(IMinFunct& fc, double& f, DVector& u, DVector& df,
                     DSymmMatrix& fm, DSymmMatrix& s);
    // Реализация алгоритма спуска
    bool Descent(IMinFunct& fc, double& f, DVector& u, DVector& df,
                 DSymmMatrix& fm, DSymmMatrix& s);
    // Реализация шага по Гауссу-Ньютону
    bool GNStep(IMinFunct& fc, double& f, DVector& u, DVector& df,
                DSymmMatrix& fm, DSymmMatrix& s, const DVector& du);
    // Реализизация алгоритма умножения Vt*S*V, где V-вектор,
    // S - симметричная матрица
    double VtMSV(const DVector& v, const DSymmMatrix& s);
    // Отладочная печать вектора
    void PrintVector(const DVector& u) const;
};

///////////////////////////////////////////////////////////////////////////////
#endif
