// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация класса, вычисляющего функционал, градииент и матрицу производных 
// при оптимизации орбиты по измерениям и априорной информации
///////////////////////////////////////////////////////////////////////////////
#include "IMinMnkTr.h"
#include "J2000Frame.h"
///////////////////////////////////////////////////////////////////////////////
// Конструктор
//-----------------------------------------------------------------------------
IMinTrFun::IMinTrFun()
{
    UseDigMtx = false;
    PrintFlag = false;
    PrintMatrix = false;
    Del = NULL;
}

//-----------------------------------------------------------------------------
// Деструктор
//-----------------------------------------------------------------------------
IMinTrFun::~IMinTrFun()
{
    if (Del != NULL)
        delete[] Del;
}

//-----------------------------------------------------------------------------
// Метод запроса вычисления функционала, градиента и
// матрицы частных производных
// Метод возвращает false, если при указанных аргументах
// не удается вычислить значение функции
//-----------------------------------------------------------------------------
bool IMinTrFun::Fun(const DVector& u, double& f, DVector* df, DSymmMatrix* fm)
{
    if (!IsValid(u))
        return false;

    DMatrix* matr = NULL;

    TAssert(MPacket != NULL);

    J2000Frame jf;

    if ((df != NULL || fm != NULL) && UseDigMtx) {
        // Блок расчета частных производных методом конечных разностей
        matr = new DMatrix[MPacket->GetSize()];
        if (!CreatePDM(u, matr)) {
            delete[] matr;
            return false;
        }
    }

    Init(u);

    // Инициализация
    f = 0.0;
    unsigned i, j, l, n = Size();
    int k;
    if (df != NULL) {
        df->Init(n);
        df->Init(0.0);
    }
    if (fm != NULL) {
        fm->Init(n);
        fm->Init(0.0);
    }

    // Цикл по измерениям
    MPacket->SetLast();
    for (k = MPacket->GetSize() - 1; k >= 0; k--) {
        const IMeasurement* meas = MPacket->GetCurrent();
        MPacket->GoPrev();
        if (meas->GetSize() == 0)
            continue;
        DVector m, d;
        DSymmMatrix w;
        DMatrix* pdm;
        if (UseDigMtx || (df == NULL && fm == NULL))
            pdm = NULL;
        else
            pdm = new DMatrix;
        TimeJD tm;
        Vect6 fp;
        DMatrix mm, mx;
        if (!meas->Simulate(6, *this, tm, jf, &fp, &m, &mm, &d)) {
            if (pdm != NULL)
                delete pdm;
            return false;
        }
        if (pdm != NULL) {
            if (!PredictMatrix(tm, jf, fp, mx)) {
                if (pdm != NULL)
                    delete pdm;
                return false;
            }
            /*
            if (PrintMatrix)
              {
              PrintMatr(mm);
              PrintMatr(mx);
              }*/
            Mult(*pdm, mm, mx);
        }
        Del[k] = d;
        meas->GetWeightMatrix(w);

        double ff = 0.0;
        for (j = 0; j < meas->GetSize(); j++) {
            if (!meas->GetIan(j))
                ff += w(j, j) * d(j) * d(j);
        }
        f += ff;

        // Отладочная печать
        if (PrintFlag)
            PrintMeas(k, *meas, d, w);

        if (df != NULL || fm != NULL) {
            DMatrix *e, em;
            if (UseDigMtx)
                e = matr + k;
            else {
                em.Init(meas->GetSize(), n);
                for (i = 0; i < meas->GetSize(); i++) {
                    for (j = 0; j < Size(); j++)
                        em(i, j) = (*pdm)(i, j);
                }
                e = &em;
            }
            if (PrintMatrix)
                PrintMatr(*e);
            if (df != NULL) {
                for (i = 0; i < n; i++) {
                    for (j = 0; j < meas->GetSize(); j++) {
                        if (!meas->GetIan(j))
                            (*df)(i) += (*e)(j, i) * w(j, j) * d(j);
                    }
                }
            }
            if (fm != NULL) {
                for (i = 0; i < n; i++) {
                    for (j = i; j < n; j++)
                        for (l = 0; l < meas->GetSize(); l++) {
                            if (!meas->GetIan(l))
                                (*fm)(i, j) += (*e)(l, i) * (*e)(l, j) * w(l, l);
                        }
                }
            }
        }
        if (pdm != NULL)
            delete pdm;
    }

    if (matr != NULL)
        delete[] matr;

    Fa = 0.0;
    IAprioriConstEntry* apr;
    for (apr = ACList.first(), k = 0; apr != NULL; apr = apr->next(), k++) {
        unsigned na = apr->GetSize();
        DVector da;
        DSymmMatrix wa;
        DMatrix ma;
        apr->GetDelVector(u, da);
        apr->GetWeightMatrix(wa);
        apr->GetMatrix(u, ma);
        for (j = 0; j < na; j++)
            Fa += wa(j, j) * da(j) * da(j);
        f += Fa;
        if (df != NULL) {
            for (i = 0; i < n; i++) {
                for (j = 0; j < na; j++)
                    (*df)(i) += ma(j, i) * wa(j, j) * da(j);
            }
        }
        if (fm != NULL) {
            for (i = 0; i < n; i++) {
                for (j = i; j < n; j++)
                    for (l = 0; l < na; l++)
                        (*fm)(i, j) += ma(l, i) * ma(l, j) * wa(l, l);
            }
        }
        // Отладочная печать
        if (PrintFlag)
            PrintApr(k, *apr, da, wa);
    }

    return true;
}

//-------------------------------------------------------------
// Расчет матриц частных производных методом конечных разностей
// на моменты измерений
//-------------------------------------------------------------
bool IMinTrFun::CreatePDM(const DVector& u, DMatrix* matr)
{
    unsigned j, k, n, nm;
    int i;

    J2000Frame jf;

    n = Size();
    nm = MPacket->GetSize();

    DVector* m = new DVector[nm];
    // Цикл по измерениям
    MPacket->SetLast();
    for (i = nm - 1; i >= 0; i--) {
        const IMeasurement* meas = MPacket->GetCurrent();
        MPacket->GoPrev();
        if (meas->GetSize() == 0)
            continue;
        m[i].Init(meas->GetSize());
        matr[i].Init(meas->GetSize(), n);
    }

    DVector eps;
    GetEps(eps);
    for (j = 0; j < n; j++) {
        DVector uc(n), epsm(n);
        epsm.Init(0.0);
        epsm(j) = eps(j);
        TAssert(eps.Dim() == n);
        AddU(u, epsm, uc);
        Init(uc);
        TimeJD t;
        Vect6 f;
        MPacket->SetLast();
        for (i = nm - 1; i >= 0; i--) {
            const IMeasurement* meas = MPacket->GetCurrent();
            MPacket->GoPrev();
            if (!meas->Simulate(6, *this, t, jf, &f, m + i)) {
                delete[] m;
                return false;
            }
        }
        epsm.Init(0.0);
        epsm(j) = -eps(j);
        AddU(u, epsm, uc);
        Init(uc);
        MPacket->SetLast();
        for (i = nm - 1; i >= 0; i--) {
            const IMeasurement* meas = MPacket->GetCurrent();
            MPacket->GoPrev();
            DVector v(meas->GetSize()), dv(meas->GetSize());
            if (!meas->Simulate(6, *this, t, jf, &f, &v)) {
                delete[] m;
                return false;
            }
            dv = m[i];
            dv -= v;
            for (k = 0; k < meas->GetSize(); k++)
                matr[i](k, j) = dv(k) * 0.5 / eps[j];
        }
    }

    delete[] m;

    return true;
}

//-----------------------------------------------------------------------------
// Установка измерений
//-----------------------------------------------------------------------------
void IMinTrFun::SetPacket(IMeasurementPacket* p)
{
    MPacket = p;

    if (Del != NULL)
        delete[] Del;

    Del = new DVector[MPacket->GetSize()];
}

//-----------------------------------------------------------------------------
// Отладочная печать невязок измерений
//-----------------------------------------------------------------------------
void IMinTrFun::PrintMeas(int k, const IMeasurement& meas, const DVector& d,
                          const DSymmMatrix& w)
{
    TimeJD t = meas.GetTime();
    double td = t.GetDays() + t.GetFraction();

    printf("%3d %16.7f ", k, td);
    TimeGD gd(t);
    printf(" %04d/%02d/%02d %02d/%02d/%06.3f ", gd.Year, gd.Month, gd.Day, gd.Hour,
           gd.Min, gd.Sec + gd.Fraction);

    unsigned i;
    printf("тип %d ", meas.GetType());
    printf("нев. ");
    double f = 0.0;
    for (i = 0; i < meas.GetSize(); i++) {
        printf("%c", meas.GetIan(i) ? '*' : ' ');
        printf("%+9.3e ", d(i));
        f += d(i) * d(i) * w(i, i);
    }
    printf("f = %+9.3e ", f);

    printf("  веса: ");
    for (i = 0; i < meas.GetSize(); i++)
        printf("%+9.3e ", w(i, i));
    printf("\n");
}

//-----------------------------------------------------------------------------
// Отладочная печать невязок апр. констант
//-----------------------------------------------------------------------------
void IMinTrFun::PrintApr(int k, const IAprioriConst& apr, const DVector& d,
                         const DSymmMatrix& w)
{
    printf("Апр. %3d ", k);

    unsigned i;

    printf("нев. ");
    double f = 0.0;
    for (i = 0; i < apr.GetSize(); i++) {
        printf("%c", w(i, i) == 0.0 ? '*' : ' ');
        printf("%+9.3e ", d(i));
        f += d(i) * d(i) * w(i, i);
    }
    printf("f = %+9.3e ", f);

    printf("  веса: ");
    for (i = 0; i < apr.GetSize(); i++)
        printf("%+9.3e ", w(i, i));
    printf("\n");
}

//-----------------------------------------------------------------------------
// Отладочная печать матрицы 
//-----------------------------------------------------------------------------
void IMinTrFun::PrintMatr(const DMatrix& m) const
{
    unsigned i, j;

    for (i = 0; i < m.Rows(); i++) {
        for (j = 0; j < m.Cols(); j++)
            printf("%+14.8e ", m(i, j));
        printf("\n");
    }

    printf("\n");
}

///////////////////////////////////////////////////////////////////////////////
