///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация класса, вычисляющего функционал, градииент и матрицу производных 
// при оптимизации орбиты по измерениям и априорной информации
///////////////////////////////////////////////////////////////////////////////
#ifndef IMINMNK_TRAJECT_H
#define IMINMNK_TRAJECT_H
///////////////////////////////////////////////////////////////////////////////
#include "IMinMnk.h"
#include "ITraject.h"
#include "IMeasurementPacket.h"
#include "AprioriConst.h"

///////////////////////////////////////////////////////////////////////////////
class IMinTrFun : public IMinFunct, public ITrajectVar
{
public:
    // Конструктор
    IMinTrFun();
    // Деструктор
    virtual ~IMinTrFun();

    // Запрос размерности аргументов функции
    virtual unsigned Size() const = 0;
    // Метод запроса вычисления функционала, градиента и
    // матрицы частных производных
    // Метод возвращает false, если при указанных аргументах
    // не удается вычислить значение функции
    virtual bool Fun(const DVector& u, double& f, DVector* df = NULL,
                     DSymmMatrix* fm = NULL);

    // Установка измерений
    void SetPacket(IMeasurementPacket* p);
    // Получение указателя на пакет измерений
    IMeasurementPacket* GetPacket() const
    {
        return MPacket;
    };

    // Установка флага отладочной печати
    void SetPrintFlag(bool f)
    {
        PrintFlag = f;
    };
    // Установка флага печати матриц частных производных
    void SetPrintMatrix(bool f)
    {
        PrintMatrix = f;
    };

    // Установка флага расчета частных производных в конечных разностях
    void SetDigMtx(bool f)
    {
        UseDigMtx = f;
    };
    // Установка флага расчета частных производных в конечных разностях
    bool GetDigMtx() const
    {
        return UseDigMtx;
    };

    // Запрос невязок измерений
    const DVector& GetDel(unsigned k) const
    {
        return Del[k];
    };

    // Запрос наличия априорных констант
    bool HasApr() const
    {
        return !ACList.isempty();
    };
    // Запрос функционала на априорных константах
    double GetFa() const
    {
        return Fa;
    };
    // Изменение априорных констант по текущему вектору
    void SetApr(const DVector& u)
    {
        ACList.Set(u);
    };
    // Добавление априорной константы
    void AddApr(const IAprioriConstEntry& a)
    {
        ACList.Add(*a.Clone());
    };
    // Очистка априорных констант
    void ClearApr()
    {
        ACList.clear();
    };

protected:
    // Флаг использования численного расчета для матрицы частных производных
    bool UseDigMtx;
    // Приращения, используемые при численном расчете матрицы частных производных
    virtual void GetEps(DVector& eps) const = 0;
    // Контроль входных параметров
    virtual bool IsValid(const DVector& u) const = 0;
    // Функционал априорной орбиты
    double Fa;
    // Указатель на объект, описывающий пакет измерений
    IMeasurementPacket* MPacket;
    // Список априорных констант
    IAprioriConstList ACList;
    // Невязки измерений
    DVector* Del;
    // Флаг отладочной печати
    bool PrintFlag;
    // Флаг печати матриц частных производных
    bool PrintMatrix;

    // Расчет матриц частных производных методом конечных разностей на
    // моменты измерений
    bool CreatePDM(const DVector& u, DMatrix* matr);
    // Отладочная печать невязок измерений
    void PrintMeas(int k, const IMeasurement& meas, const DVector& d, const DSymmMatrix& w);
    // Отладочная печать невязок апр. констант
    void PrintApr(int k, const IAprioriConst& apr, const DVector& d, const DSymmMatrix& w);
    // Отладочная печать матрицы 
    void PrintMatr(const DMatrix& m) const;
};

///////////////////////////////////////////////////////////////////////////////
#endif
