///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Интерфейсы моделей
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_IMODELS_H
#define BNOTOOL_IMODELS_H
///////////////////////////////////////////////////////////////////////////////
#include "TM.h"
#include "DVectors.h"
#include "Ids.h"

///////////////////////////////////////////////////////////////////////////////
// Интерфейс доступа к объекту, вычисляющему положение планеты или тела
// Солнечной системы
class IEphemeride
{
public:
    // Загрузить данные для указанного интервала
    virtual ~IEphemeride()
    {
    };
    // Загрузить данные для указанного интервала
    virtual bool LoadData(const TimeJD& s, const TimeJD& e) = 0;
    // Освободить память
    virtual void UnLoadData() = 0;
    // Вычислить положение тела.
    virtual void GetPosition(const TimeJD& t, Vect3* pos = NULL,
                             Vect3* vel = NULL, Vect3* acc = NULL) const = 0;
    // Вернуть идентификатор системы в которой вычисляются положения
    virtual TCoordSysId GetCS() const = 0;
};

///////////////////////////////////////////////////////////////////////////////
// Класс описывающий геометрическую модель небесного тела
class TGeometryModel
{
public:
    double Re; // экваториальный радиус
    double alpha; // сжатие
};

///////////////////////////////////////////////////////////////////////////////
// Интерфейс модели атмосферы
class IAtmosphereModel
{
public:
    // Вычисление плотности. Положение точки и Солнца должны задаваться
    virtual ~IAtmosphereModel()
    {
    };
    // Вычисление плотности. Положение точки и Солнца должны задаваться
    // в Гринвиче
    virtual double Density(const TimeJD& t, const Vect3& v, const Vect3& vSun)
    const = 0;
    // Идентификатор модели
    virtual TAtmosphereId GetID() const = 0;
    // Нужно ли для вычисления плотности положение Солнца?
    virtual bool IsSunPosNeed() const = 0;
    // Нужна ли для вычисления плотности таблица атмосферных индексов?
    virtual bool IsTableNeed() const = 0;
};

///////////////////////////////////////////////////////////////////////////////
// Атмосферные индексы, привязанные к конкретной дате
class TGeoSolarFlux
{
public:
    double f135; // средний индекс солнечной активности за 135 дней  - гост 77
    double f90; // средний индекс солнечной активности за 90 дней   - ???
    double f81; // средний индекс солнечной активности за 81 день   - ???
    double f135w; // средневзвешенный индекс солнечной активности за 135 дней
    double f81w; // средневзвешенный индекс солнечной активности за 81 день - гост 90
    double f01; // средний индекс солнечной активности за 1 день     - во всех
    double kp; // планетарный трехчасовой индекс геомагнитной возмущенности (среднее из восьми значений)
    int ap; // средняя планетарная амплитуда геомагнитной возмущенности
    double three_hours_kp[8]; // трехчасовые индексы
    bool iskp3;
};

///////////////////////////////////////////////////////////////////////////////
// Интерфейс таблицы индексов солнечной активности и геомагнитной возмущенности
class ITableGeoSolarFlux
{
public:
    // Получение индексов на указанную дату
    virtual ~ITableGeoSolarFlux()
    {
    };
    // Получение индексов на указанную дату
    virtual const TGeoSolarFlux* Get_F107(const TimeJD& tm) const = 0;
    // Получение трехчасовых индексов на указанную дату
    virtual const double* Get_Kp3hour(const TimeJD& tm) const = 0;
    // Получение модифицированного трехчасового индекса
    virtual double Get_Kp3hourMod(const TimeJD& tm) const = 0;
};

///////////////////////////////////////////////////////////////////////////////
// Интерфейс моделей гравитации
class IGravityModel
{
public:
    // Получение значений гармоник
    virtual ~IGravityModel()
    {
    };
    // Получение значений гармоник
    virtual double c(unsigned int n, unsigned int m) const = 0;
    virtual double d(unsigned int n, unsigned int m) const = 0;
    // Получение числа гармоник
    virtual unsigned int N() const = 0;
    virtual unsigned int M() const = 0;
    // Идентификатор модели
    virtual TGravityModelId GetID() const = 0;
    virtual const double* const* get_cc() const { return nullptr; }
    virtual const double* const* get_dd() const { return nullptr; }

    double mu; // Гравитационная постоянная
    double Re; // Экваториальный радиус
    double alpha; // Коэффициент сжатия планеты
};

///////////////////////////////////////////////////////////////////////////////
// Интерфейс данных о планете и ее моделях
class IPlanet
{
public:
    // Конструктор по умолчанию
    IPlanet()
    {
    };
    //
    virtual ~IPlanet()
    {
    };

    // имя планеты
    virtual const char* Name() const = 0;
    // числовой идентификатор
    virtual TPlanetId Id() const = 0;
    // Указатель на гравитационную модель
    virtual const IGravityModel* GetGravityModel() const = 0;
    // Указатель на модель атмосферы
    virtual const IAtmosphereModel* GetAtmosphere() const = 0;
    // Указатель на эфемериды
    virtual IEphemeride* GetEphemeride() = 0;
    virtual const IEphemeride* GetEphemeride() const = 0;

    double omega; // Угловая скорость вращения
    double RShad; // Радиус тени
    TGeometryModel Geometry;
};

///////////////////////////////////////////////////////////////////////////////
#endif
