///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Интерфейсы доступа к моделям движения с уточняемыми параметрами
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_ITRAJECT_H
#define BNOTOOL_ITRAJECT_H
///////////////////////////////////////////////////////////////////////////////
#include "TM.h"
#include "Matrix.h"
#include "DVectors.h"
#include "IFrames.h"
#include "IPredict.h"

///////////////////////////////////////////////////////////////////////////////
// Класс интерфейса доступа к модели движения с уточняемыми параметрами
///////////////////////////////////////////////////////////////////////////////
class ITraject
{
public:
    ITraject() : m_cache(false)
    {
    }

    // Деструктор.
    virtual ~ITraject()
    {
    };
    // Метод устанавливает значение вектора уточняемых параметров
    virtual unsigned Size() const = 0;
    // Метод устанавливает значение вектора уточняемых параметров
    virtual bool Init(const DVector& u0) = 0;
    // Метод устанавливает время начальной точки траектории
    virtual void SetTime(const TimeJD& t) = 0;
    // Метод возвращает время начальной точки траектории
    virtual const TimeJD& GetTime() const = 0;
    // Метод выполняет прогнозирование траектории на указанный момент времени t. 
    // Параметр fr определяет систему отсчету, в которой должен быть возвращен 
    // фазовый вектор f (т.е. координаты и скорости)
    virtual bool Predict(const TimeJD& t, const IFrame& fr, Vect6& f) = 0;
    virtual bool PredictA(const TimeJD& t, Vect6& f) { return 0; };
    virtual bool PredictMatrix(const TimeJD& t, const IFrame& fr, Vect6& f, DMatrix& m) = 0;
    // Метод выполняет вычисление вектора уточняемых параметров u для указанного 
    // момента времени t.
    virtual bool Move(const TimeJD& t, DVector& u) = 0;

    //	virtual IPredict* GetPredictorEx() {return 0;}
    virtual IPredict* GetPredictor() = 0;

    bool PredictEx(const TimeJD& t, Vect6& f)
    {
        if (m_cache && m_cacheTime == t) {
            f = m_cachePos;
            return true;
        }

        IPredict* predict = GetPredictor();
        if (!predict) { return false; }

        if (!predict->Run(t)) {
            return false;
        }

        predict->GetDestPointEX(f);

        m_cacheTime = t;
        m_cache = true;
        m_cachePos = f;

        return true;
    }

private:
    TimeJD m_cacheTime;
    Vect6 m_cachePos;
    bool m_cache;
};

///////////////////////////////////////////////////////////////////////////////
// Класс интерфейса доступа к модели движения с уточняемыми параметрами
// и возможностью расчета частных производных по уточняемым параметрам
///////////////////////////////////////////////////////////////////////////////
class ITrajectVar : public ITraject
{
public:
    // Метод выполняет прогнозирование траектории на указанный момент времени t и
    // подсчет матрицы частных от фазового вектора по уточняемым параметрам.
    // Параметр fr определяет систему отсчету, в которой должен быть возвращен
    // фазовый вектор f (т.е. координаты и скорости) и матрица m.
    virtual bool PredictMatrix(const TimeJD& t, const IFrame& fr, Vect6& f, DMatrix& m) = 0;
    // Получение ссылки на экземпляр прогноза (для внешних настроек)
    //	virtual IPredictVar *GetPredictor() = 0;
};

///////////////////////////////////////////////////////////////////////////////
#endif
