///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Модель измерений "Камы"
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_KAMAMEAS_H
#define BNOTOOL_KAMAMEAS_H
///////////////////////////////////////////////////////////////////////////////
#include "IMeasurement.h"

///////////////////////////////////////////////////////////////////////////////
// Модель измерений "Камы"
///////////////////////////////////////////////////////////////////////////////
class KAMAMeasurement : public IMeasurement
{
public:
    // Конструктор по умолчанию
    KAMAMeasurement();
    // Деструктор.
    virtual ~KAMAMeasurement()
    {
    };
    // Метод возвращает размерность вектора измерений
    virtual unsigned GetSize() const
    {
        return 1;
    };
    // Метод возвращает в объекте v вектор измерений
    virtual void GetRealVector(DVector& v) const
    {
        v.Init(1U);
        v(0) = D;
    };
    // Метод возвращает в объекте v вектор СКО измерений
    virtual void GetMSEVector(DVector& v) const
    {
        v.Init(1U);
        v(0) = SigD;
    };
    // Метод осуществляет моделирование измерения для космического объекта,
    // движение которого описывается с помощью переменной trj.
    // Метод вычисляет следующие параметры:
    // t	-	время, которое соответствует положению космического объекта
    //      на орбите для текущего измерения (это время равно времени измерения,
    //      если не учитывается запаздывание сигнала)
    // fr	-	система отсчета, в которой задаются положение и скорости аппарата
    //      в момент измерения, а также матрица частных производных (см. далее параметры f и m)
    // v	-	указатель вектор моделируемого измерения. Если этот параметр равен 0,
    //      вектор моделируемых измерений не вычисляется
    // m	-	матрица производных размерности Nx6, где N - размерность вектора измерений
    //      (т.е. матрица состоит из N строк и 6 столбцов). В столбцах расположены
    //      производные по Rx Ry Rz Vx  Vy Vz, а в строках - компоненты вектора измерений.
    //      Если этот параметр равен 0, вектор моделируемых измерений не вычисляется
    // del	-	вектор невязок измеряемых параметров. Если этот параметр равен 0,
    //        вектор невязок не вычисляется
    virtual bool Simulate(int matrixSize, ITraject& trj, TimeJD& t, const IFrame& fr, Vect6* f = 0,
                          DVector* v = 0, DMatrix* m = 0, DVector* del = 0) const;
    // Создать копию объекта с помощью new()
    // Если m1 или m2 == NULL, принмается текущее измерение
    virtual void Diff(DVector& d, const DVector* m1 = NULL, const DVector* m2 = NULL) const;
    // Создать копию объекта с помощью new()
    virtual IMeasurement* Clone() const;

    // Иницицализация параметров измерения
    void Init(double d, double sigd, const Vect3& psend, const Vect3& prec)
    {
        D = d;
        SigD = sigd;
        PSend = psend;
        PRec = prec;
    };

    virtual void SetIan(unsigned k, bool f = true)
    {
        m_isBad = f;
    };

    virtual bool GetIan(unsigned k) const
    {
        return m_isBad;
    };

protected:
    // Измеренная дальность
    double D;
    // СКО дальности
    double SigD;
    // Координаты передающей антенны в Гринвиче
    Vect3 PSend;
    // Координаты принимающей антенны в Гринвиче
    Vect3 PRec;

    bool m_isBad;
};

///////////////////////////////////////////////////////////////////////////////
// Модель измерений Ddot
///////////////////////////////////////////////////////////////////////////////
class KAMADtMeasurement : public IMeasurement
{
public:
    // Конструктор по умолчанию
    KAMADtMeasurement() : IMeasurement()
    {
        Type = mtRTKamaDt;
    };
    // Деструктор.
    virtual ~KAMADtMeasurement()
    {
    };
    // Метод возвращает размерность вектора измерений
    virtual unsigned GetSize() const
    {
        return 1;
    };
    // Метод возвращает в объекте v вектор измерений
    virtual void GetRealVector(DVector& v) const
    {
        v.Init(1U);
        v(0) = Dt;
    };
    // Метод возвращает в объекте v вектор СКО измерений
    virtual void GetMSEVector(DVector& v) const
    {
        v.Init(1U);
        v(0) = SigDt;
    };
    // Метод осуществляет моделирование измерения для космического объекта,
    // движение которого описывается с помощью переменной trj.
    // Метод вычисляет следующие параметры:
    // t	-	время, которое соответствует положению космического объекта
    //      на орбите для текущего измерения (это время равно времени измерения,
    //      если не учитывается запаздывание сигнала)
    // fr	-	система отсчета, в которой задаются положение и скорости аппарата
    //      в момент измерения, а также матрица частных производных (см. далее параметры f и m)
    // v	-	указатель вектор моделируемого измерения. Если этот параметр равен 0,
    //      вектор моделируемых измерений не вычисляется
    // m	-	матрица производных размерности Nx6, где N - размерность вектора измерений
    //      (т.е. матрица состоит из N строк и 6 столбцов). В столбцах расположены
    //      производные по Rx Ry Rz Vx  Vy Vz, а в строках - компоненты вектора измерений.
    //      Если этот параметр равен 0, вектор моделируемых измерений не вычисляется
    // del	-	вектор невязок измеряемых параметров. Если этот параметр равен 0,
    //        вектор невязок не вычисляется
    virtual bool Simulate(int matrixSize, ITraject& trj, TimeJD& t, const IFrame& fr, Vect6* f = 0,
                          DVector* v = 0, DMatrix* m = 0, DVector* del = 0) const;
    // Создать копию объекта с помощью new()
    // Если m1 или m2 == NULL, принмается текущее измерение
    virtual void Diff(DVector& d, const DVector* m1 = NULL, const DVector* m2 = NULL) const;
    // Создать копию объекта с помощью new()
    // Создать копию объекта с помощью new()
    virtual IMeasurement* Clone() const;

    // Иницицализация параметров измерения
    void Init(double dt, double sigdt, const Vect3& psend, const Vect3& prec)
    {
        Dt = dt;
        SigDt = sigdt;
        PSend = psend;
        PRec = prec;
    };

protected:
    // Измеренная скорость измения дальности
    double Dt;
    // СКО скорости измения дальности
    double SigDt;
    // Координаты передающей антенны в Гринвиче
    Vect3 PSend;
    // Координаты принимающей антенны в Гринвиче
    Vect3 PRec;
};

///////////////////////////////////////////////////////////////////////////////
#endif
