///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация модели КО для уравнений в вариациях с постоянными баллистическим
// коэффициентом и коэффициентом светового давления
///////////////////////////////////////////////////////////////////////////////
#include "KOConstVar.h"
#include "KOConst.h"
#include "IAstroBase.h"
#include "Atmosphere.h"
#include "GeoCoords.h"
///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
//-----------------------------------------------------------------------------
KOConstVar::KOConstVar() : IKOPropVar()
{
    IsKbVar = false;
    IsKappaVar = false;
}

//-----------------------------------------------------------------------------
// Вычисление полного вектора ускорения в J2000
//-----------------------------------------------------------------------------
void KOConstVar::GetAcc(const AccAdapter& p, Vect3& acc) const
{
    KOConst ko;

    ko.Kappa = Kappa;
    ko.Kb = Kb;
    ko.Init(*MMode);
    ko.GetAcc(p, acc);
}

//-----------------------------------------------------------------------------
// Получение числа варьируемых параметров
//-----------------------------------------------------------------------------
int KOConstVar::GetNVars() const
{
    int ret = 0;
    if (IsKbVar)
        ret++;
    if (IsKappaVar)
        ret++;
    return ret;
}

//-----------------------------------------------------------------------------
// Вычисление добавок к правым частям уравнений в вариациях, зависящих
// от свойств KO
//-----------------------------------------------------------------------------
void KOConstVar::GetRPVars(const AccAdapter& p, TransMtx& dfdr, TransMtx& dfdv,
                           DMatrix& dfdc) const
{
    TAssert(MMode != NULL);

    int i, j;
    for (i = 0; i < 3; i++)
        for (j = 0; j < 3; j++) {
            dfdr(i, j) = 0.0;
            dfdv(i, j) = 0.0;
        }

    if (IsKbVar || IsKappaVar)
        dfdc.Init(GetNVars(), 3);

    if (MMode->IsAtm)
        GetAtmVars(p, dfdr, dfdv);

    if (MMode->LightMode != lightNONE) {
        TransMtx m1;
        GetSolarVars(p, m1);
        dfdr += m1;
    }

    if (IsKbVar) {
        Vect3 var;
        GetKbVars(p, var);
        for (j = 0; j < 3; j++)
            dfdc(0, j) = var[j];
    }

    if (IsKappaVar) {
        Vect3 var;
        GetKappaVars(p, var);
        i = IsKbVar ? 1 : 0;
        for (j = 0; j < 3; j++)
            dfdc(i, j) = var[j];
    }
}

//-----------------------------------------------------------------------------
// Вычисление вариаций от атмосферы по координатам и скоростям
//-----------------------------------------------------------------------------
void KOConstVar::GetAtmVars(const AccAdapter& p, TransMtx& dfdr, TransMtx& dfdv) const
{
    double req, alpha;

    req = GetAstroBase()->GetEarthRadius();
    alpha = GetAstroBase()->GetEarthOblatness();
    const IAtmosphereModel* atm = GetAstroBase()->GetEarthAtmosphere();

    // Вычисляем высоту точки над поверхностью Земли (CS10).
    double rlen, vlen, h, ez;
    rlen = p.RGr.Len();
    vlen = p.VGr.Len();
    ez = p.RGr.z / rlen;
    h = rlen - req * (1.0 - alpha * ez * ez);

    Vect3 rg;
    double dd, dd1;
    double gv, gr;
    gv = -Kb * vlen * p.Density * 1.e6;
    int i, j;

    double epsr = 1.e-6;
    for (i = 0; i < 3; i++) {
        rg = p.RGr;
        rg[i] += epsr;
        dd = atm->Density(p.T_TDB, rg, p.SunGr);
        rg = p.RGr;
        rg[i] -= epsr;
        dd1 = atm->Density(p.T_TDB, rg, p.SunGr);
        gr = -(dd - dd1) * 0.5 / epsr * 1.e6 * Kb * vlen;
        for (j = 0; j < 3; j++)
            dfdr(i, j) = p.VGr[j] * gr;
    }

    Vect3 ve;
    ve.x = p.VGr.x / vlen;
    ve.y = p.VGr.y / vlen;
    ve.z = p.VGr.z / vlen;

    dfdv(0, 0) = gv * (ve.x * ve.x + 1.0);
    dfdv(0, 1) = gv * (ve.x * ve.y);
    dfdv(0, 2) = gv * (ve.x * ve.z);
    dfdv(1, 0) = dfdv(0, 1);
    dfdv(1, 1) = gv * (ve.y * ve.y + 1.0);
    dfdv(1, 2) = gv * (ve.y * ve.z);
    dfdv(2, 0) = dfdv(0, 2);
    dfdv(2, 1) = dfdv(1, 2);
    dfdv(2, 2) = gv * (ve.z * ve.z + 1.0);

    // Мы получили матрицы в Гринвиче
    // Теперь переводим их в J2000

    TransMtx omega;
    for (i = 0; i < 3; i++)
        for (j = 0; j < 3; j++)
            omega(i, j) = 0;

    omega(0, 1) = -GetAstroBase()->GetEarthRotation();
    omega(1, 0) = GetAstroBase()->GetEarthRotation();

    omega.MulLeft(dfdv);
    dfdr += omega;

    dfdr.MulLeft(p.CGFCStoJ2000);
    dfdr.MulRight(p.CJ2000toGFCS);

    dfdv.MulLeft(p.CGFCStoJ2000);
    dfdv.MulRight(p.CJ2000toGFCS);
}

//-----------------------------------------------------------------------------
// Вычисление вариаций от светового давления по координатам и скоростям
//-----------------------------------------------------------------------------
void KOConstVar::GetSolarVars(const AccAdapter& p, TransMtx& dfdr) const
{
    KOConst ko;
    Vect3 acc;

    int i, j;

    ko.Kappa = Kappa;
    ko.Kb = Kb;
    ko.Init(*MMode);
    double psi = ko.LightAcc(p, acc);

    if (psi == 0.0) {
        for (i = 0; i < 3; i++)
            for (j = 0; j < 3; j++)
                dfdr(i, j) = 0;
        return;
    }

    Vect3 rsp = p.RJ2000;
    rsp -= p.SunJ2000;
    double dsp = rsp.Len();
    rsp *= 1.0 / dsp;
    double gsp = -3.0 * Kappa * GetAstroBase()->GetPlanet(plIdSun)->GetGravityModel()->mu /
        (dsp * dsp * dsp);

    dfdr(0, 0) = psi * gsp * (rsp.x * rsp.x - 1.0 / 3.0);
    dfdr(0, 1) = psi * gsp * rsp.x * rsp.y;
    dfdr(0, 2) = psi * gsp * rsp.x * rsp.z;
    dfdr(1, 0) = dfdr(0, 1);
    dfdr(1, 1) = psi * gsp * (rsp.y * rsp.y - 1.0 / 3.0);
    dfdr(1, 2) = psi * gsp * rsp.y * rsp.z;
    dfdr(2, 0) = dfdr(0, 2);
    dfdr(2, 1) = dfdr(1, 2);
    dfdr(2, 2) = psi * gsp * (rsp.z * rsp.z - 1.0 / 3.0);
}

//-----------------------------------------------------------------------------
// Вычисление явных производных от баллистического коэффициента
//-----------------------------------------------------------------------------
void KOConstVar::GetKbVars(const AccAdapter& p, Vect3& var) const
{
    TAssert(MMode->IsAtm);
    KOConst ko;

    ko.Kappa = Kappa;
    ko.Kb = 1.0;
    ko.Init(*MMode);
    ko.AtmAcc(p, var);
}

//-----------------------------------------------------------------------------
// Вычисление явных производных от коэффициента св. давления
//-----------------------------------------------------------------------------
void KOConstVar::GetKappaVars(const AccAdapter& p, Vect3& var) const
{
    TAssert(MMode->LightMode != lightNONE);
    KOConst ko;

    ko.Kappa = 1.0;
    ko.Kb = Kb;
    ko.Init(*MMode);
    ko.LightAcc(p, var);
}

///////////////////////////////////////////////////////////////////////////////
