///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация прогноза по законам Кеплера
///////////////////////////////////////////////////////////////////////////////
#include "KeplPredict.h"
#include "Consts.h"
#include "IAstroBase.h"

///////////////////////////////////////////////////////////////////////////////
// Вспомогательные функции 
///////////////////////////////////////////////////////////////////////////////
double Debut(double g, double alpha1, double alpha2)
{
    double a0 = -1.916905827235138e-3;
    double b0 = 1.0 - a0;
    double a1 = 1.045620035932856;
    double a2 = -2.970499758665694e-1;
    double c = 1.309972403091835e-1;
    double b1 = -c - 3.965153894019779e-2;
    double b2 = -3.813781995676532e-1;
    double q = floor(g / Pi2) * Pi2;
    double h = g + alpha2 - q;
    double t1 = Sqrt2 * 0.5;
    double teta1 = (alpha2 + alpha1) * t1;
    double teta2 = (alpha2 - alpha1) * t1;
    double gamma, delta, z, fi, n;

    if (h < Pi - alpha2) {
        if (h < Pid2 - alpha1) {
            if (h < Pid4 + teta2) {
                gamma = alpha1;
                delta = -alpha2;
                z = h;
                fi = 0.0;
                n = 1.0;
            }
            else {
                gamma = alpha2;
                delta = -alpha1;
                z = Pid2 - h;
                fi = Pid2;
                n = -1.0;
            }
        }
        else {
            if (h < Pi3d4 - teta1) {
                gamma = alpha2;
                delta = alpha1;
                z = h - Pid2;
                fi = Pid2;
                n = 1.0;
            }
            else {
                gamma = -alpha1;
                delta = -alpha2;
                z = Pi - h;
                fi = Pi;
                n = -1.0;
            }
        }
    }
    else {
        if (h < 3.0 * Pid2 + alpha1) {
            if (h < 5.0 * Pid4 - teta2) {
                gamma = -alpha1;
                delta = alpha2;
                z = h - Pi;
                fi = Pi;
                n = 1.0;
            }
            else {
                gamma = -alpha2;
                delta = alpha1;
                z = 3.0 * Pid2 - h;
                fi = 3.0 * Pid2;
                n = -1.0;
            };
        }
        else {
            if (h < 7.0 * Pid4 + teta1) {
                gamma = -alpha2;
                delta = -alpha1;
                z = h - 3.0 * Pid2;
                fi = 3.0 * Pid2;
                n = 1.0;
            }
            else {
                gamma = alpha1;
                delta = alpha2;
                z = Pi2 - h;
                fi = Pi2;
                n = -1.0;
            };
        }
    };

    double c0 = z + a0 * gamma + b0 * delta;
    double c1 = (1.0 + c * z - a1 * gamma - b1 * delta) * 0.5;
    double c2 = c + a2 * gamma + b2 * delta;
    if (fabs(c2) < 1e-15)
        c0 = c0 / (2.0 * c1);
    else {
        c0 = c1 * c1 - c0 * c2;
        c0 = (c0 < 0.0 ? c1 / c2 : (c1 - sqrt(c0)) / c2);
    };
    if (c0 < 0.0)
        c0 = 0.0;
    else if (c0 > Pid4)
        c0 = Pid4;

    return n * c0 + fi + q;
}

//-----------------------------------------------------------------------------
double F2u(double f, double alpha1, double alpha2)
{
    double e2 = alpha1 * alpha1 + alpha2 * alpha2;
    double _1_e2 = 1.0 - e2;
    _1_e2 = _1_e2 < 0.0 ? 0.0 : sqrt(_1_e2);
    double beta = 1.0 / (1.0 + _1_e2);
    double sinf = sin(f);
    double cosf = cos(f);
    return f + 2.0 * atan2(
        beta * ((alpha1 - beta * e2) * sinf + alpha2 * (1.0 - cosf)),
        1.0 - beta * ((alpha1 - beta * e2) * cosf + alpha2 * sinf + alpha1));
};
///////////////////////////////////////////////////////////////////////////////
// Установка начальных условий для прогноза
//-----------------------------------------------------------------------------
void KeplPredict::SetInitPoint(const PhasePoint6D& src)
{
    PhasePoint6D j2000;
    J2000Frame f2000;
    j2000.SetFrame(f2000);
    j2000.Assign(src);

    Kepl0.SetXYZ(j2000.CoordsVel);
    T0 = src.T;
    T0 -= Kepl0.GetTimeFromAN() / 86.4;
    Tk = T0;
    Kepl0.U = 0;

    FlagInit = true;
}

//-----------------------------------------------------------------------------
// Собственно прогнозирование
//-----------------------------------------------------------------------------
bool KeplPredict::Run(const TimeJD& t)
{
    TAssert(FlagInit);

    double e2 = Kepl0.Ex * Kepl0.Ex + Kepl0.Ey * Kepl0.Ey;
    double e = sqrt(e2);
    double _1_e2 = 1.0 - e2;
    _1_e2 = _1_e2 < 0.0 ? 0.0 : sqrt(_1_e2);
    double alpha1 = (Kepl0.Ex + e2) / (1.0 + Kepl0.Ex);
    double alpha2 = sqrt(1.0 - e2) * (Kepl0.Ey / (1.0 + Kepl0.Ex));

    double mu = GetAstroBase()->GetEarthMu();
    double dt = (t - T0).GetDouble() * 86.4;

    double n0 = sqrt(mu / (Kepl0.A * Kepl0.A * Kepl0.A));
    double g = n0 * dt;
    double f = Debut(g, alpha1, alpha2);

    double sinf, cosf, f0, f1, f2, df;
    int iit = 5;

    while (iit-- > 0) {
        sinf = sin(f);
        cosf = cos(f);
        f0 = f - alpha1 * sinf - alpha2 * (1.0 - cosf) - g;
        f1 = 1.0 - alpha1 * cosf - alpha2 * sinf;
        f2 = alpha1 * sinf - alpha2 * cosf;
        df = (f0 * f1) / (f1 * f1 - 0.5 * f0 * f2);
        f -= df;
        if (e * (df * df * df * df) / (5.0 * fabs(f0)) < 1.e-8)
            break;
    };
    Kepl0.U = F2u(f, alpha1, alpha2);
    Tk = t;

    return true;
}

//-----------------------------------------------------------------------------
// Получение конечной точки
//-----------------------------------------------------------------------------
void KeplPredict::GetDestPoint(PhasePoint6D& dest)
{
    TAssert(FlagInit);

    /*	J2000Frame f2000;
        dest.SetFrame(f2000);
        dest.T = Tk;
        Kepl0.GetXYZ(dest.CoordsVel);*/

    Vect6 pos;
    Kepl0.GetXYZ(pos);

    dest.Init(static_cast<const IFrame&>(J2000Frame()), pos, Tk);
}

void KeplPredict::GetDestPointEX(Vect6& pos)
{
    TAssert(FlagInit);

    Kepl0.GetXYZ(pos);

    //	dest.Init(static_cast<const IFrame&>(J2000Frame()),pos,Tk);
}


///////////////////////////////////////////////////////////////////////////////
