///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Классы для оптимизации траектории в элементах Лагранжа
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_LAGRTRAJECTVAR_H
#define BNOTOOL_LAGRTRAJECTVAR_H
///////////////////////////////////////////////////////////////////////////////
#include "IMinMnkTr.h"
#include "TrjNumVar.h"
#include "Lagrange.h"

///////////////////////////////////////////////////////////////////////////////
class LagrTrajectVar : public IMinTrFun
{
public:
    // Конструктор
    LagrTrajectVar();
    // Деструктор
    virtual ~LagrTrajectVar();

    // Установка флагов
    void SetFlags(const bool* f);
    // Получение флагов
    const bool* GetFlags() const
    {
        return Flags;
    };
    // Установка параметров в элементах Лагранжа
    void SetParams(const TimeJD& t, const Lagrange& lagr, double kb, double kappa);
    // Получение текущих параметров в элементах Лагранжа
    void GetParams(Lagrange& lagr, double& kb, double& kappa, const DVector& u) const;

    // Возврат размерности вектора уточняемых параметров
    virtual unsigned Size() const;
    // Метод устанавливает значение вектора уточняемых параметров
    virtual bool Init(const DVector& u0);
    // Метод устанавливает время начальной точки траектории
    virtual void SetTime(const TimeJD& t)
    {
        GetTrj().SetTime(t);
    };
    // Метод возвращает время начальной точки траектории
    virtual const TimeJD& GetTime() const
    {
        return GetTrj().GetTime();
    }

    // Метод выполняет прогнозирование траектории на указанный момент времени t. 
    // Параметр fr определяет систему отсчету, в которой должен быть возвращен 
    // фазовый вектор f (т.е. координаты и скорости)
    virtual bool Predict(const TimeJD& t, const IFrame& fr, Vect6& f)
    {
        return GetTrj().Predict(t, fr, f);
    };
    // Метод выполняет вычисление вектора уточняемых параметров u для указанного 
    // момента времени t.
    virtual bool Move(const TimeJD& t, DVector& u);
    // Метод выполняет прогнозирование траектории на указанный момент времени t и
    // подсчет матрицы частных от фазового вектора по уточняемым параметрам.
    // Параметр fr определяет систему отсчету, в которой должен быть возвращен
    // фазовый вектор f (т.е. координаты и скорости) и матрица m.
    virtual bool PredictMatrix(const TimeJD& t, const IFrame& fr, Vect6& f, DMatrix& m);
    // Запрос приемлимости полученного решения
    virtual bool CanBeAccepted(const DVector& u);
    // Реализация алгоритма проверки сходимости
    // Возврат true, если сошлись
    virtual bool IsMin(const DVector& u, double f,
                       const DVector& u1, double f1,
                       const DSymmMatrix& s, double alpha, unsigned nit,
                       EnMinMnkPath& path);
    // Реализация операции сложения для уточняемого вектора
    virtual void AddU(const DVector& u0, const DVector& du,
                      DVector& u) const;
    // Получение объекта прогноза
    virtual ITrajectVar& GetTrj()
    {
        return Trj;
    };

    virtual const ITrajectVar& GetTrj() const
    {
        return Trj;
    };
    // Получение ссылки на экземпляр прогноза (для внешних настроек)
    virtual IPredictVar* GetPredictor()
    {
        return (IPredictVar*)GetTrj().GetPredictor();
    };

    bool allowPoorQuality;

protected:
    // Флаги оптмизируемых параметров
    bool Flags[8];
    // Объект прогноза
    TrjNumVar Trj;
    // Элементы Лагранжа (полный набор)
    Lagrange Lagr;
    // Балл. коэффициент
    double Kb;
    // Коэффициент свет. давления
    double Kappa;
    // Матрица частных производных для нач. условий
    DSqrMatrix DrDl;

    // Метод возвращает индекс элемента Лагранжа (0-7),
    // k-той компоненте соответствующего вектора
    virtual unsigned LIndex(unsigned k) const;
    // Приращения, используемые при численном расчете матрицы частных производных
    virtual void GetEps(DVector& eps) const;
    // Формирование выходной матрицы частных производных
    virtual void FormPDM(const DMatrix& mr, DMatrix& m);
    // Контроль входных параметров
    virtual bool IsValid(const DVector& u) const;
};

///////////////////////////////////////////////////////////////////////////////
#endif
