///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Класс орбитальных элементов Лагранжа
///////////////////////////////////////////////////////////////////////////////
#include "Lagrange.h"
#include "IAstroBase.h"
#include "Defines.h"
#include <iostream>
///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
//-----------------------------------------------------------------------------
Lagrange::Lagrange()
{
    IAstroBase* ab = GetAstroBase();

    Mu = ab->GetEarthMu();

    Prec = 1.e-9;
    NMax = 10;
}

//-----------------------------------------------------------------------------
// Установка координат и скоростей по декартовым
//-----------------------------------------------------------------------------
void Lagrange::SetXYZ(const Vect6& src)
{
    double c1, c2, c3, c, d, f1, f2;

    c1 = src[1] * src[5] - src[2] * src[4];
    c2 = src[2] * src[3] - src[0] * src[5];
    c3 = src[0] * src[4] - src[1] * src[3];
    c = sqrt(c1 * c1 + c2 * c2 + c3 * c3);
    d = 1 / c;

    c3 = sqrt((c3 * d + 1) * 0.5);
    f1 = 0.5 * d / c3;

    P = -c2 * f1;
    Q = c1 * f1;

    c1 = Mu / sqrt(src[0] * src[0] + src[1] * src[1] + src[2] * src[2]);
    c2 = src[3] * src[3] + src[4] * src[4] + src[5] * src[5] - c1;
    c1 = (c1 - c2) / Mu;
    d = src[0] * src[3] + src[1] * src[4] + src[2] * src[5];
    f1 = c2 * src[0] - d * src[3];
    f2 = c2 * src[1] - d * src[4];
    c2 = ((c2 * src[2] - d * src[5]) * c3 + f1 * Q - f2 * P) * 2;
    K = (f1 - c2 * Q) / Mu;
    H = (c2 * P + f2) / Mu;
    c2 = sqrt(c1 / Mu);
    L = 1.0 / c2;
    c = d / (c + L);
    c3 = 2 * (src[0] * Q - src[1] * P + src[2] * c3);
    f1 = (src[0] - c3 * Q) * c1 + K - H * c;
    f2 = (src[1] + c3 * P) * c1 + H + K * c;

    if ((f1 * f1 + f2 * f2) > 0.0)
        Lambda = atan2(f2, f1) - d * c2;
    else
        Lambda = -d * c2;

    if (Lambda < 0)
        Lambda += Pi2;
}

//-----------------------------------------------------------------------------
// Получение декартовых координат и скоростей
//-----------------------------------------------------------------------------
bool Lagrange::GetXYZ(Vect6& dst) const
{
    double a1, nu, n, nu1;
    double a11, x1, y1, vx1, vy1, f, f1, ce, ce1;

    a1 = L * L / Mu;
    nu = sqrt(1.0 - H * H - K * K);
    n = Mu * Mu / (L * L * L);
    nu1 = sqrt(1.0 - P * P - Q * Q);

    double oke[4];
    if (!SolveKeplerEq(oke)) {
        return false;
    }

    ce1 = (1.0 - oke[1]) / (1.0 + nu);
    ce = oke[0] / (1.0 + nu);
    x1 = a1 * (oke[3] - K + H * ce);
    y1 = a1 * (oke[2] - H - K * ce);
    a11 = a1 * n;
    vx1 = a11 * (-oke[2] + H * ce1) / oke[1];
    vy1 = a11 * (oke[3] - K * ce1) / oke[1];
    f = 2 * (P * y1 - Q * x1);
    f1 = 2 * (P * vy1 - Q * vx1);
    dst[0] = x1 + f * Q;
    dst[1] = y1 - f * P;
    dst[2] = f * nu1;
    dst[3] = vx1 + f1 * Q;
    dst[4] = vy1 - f1 * P;
    dst[5] = f1 * nu1;

    return true;
}

//-----------------------------------------------------------------------------
// Решение общего кеплеровского уравнения
//-----------------------------------------------------------------------------
bool Lagrange::SolveKeplerEq(double* oke) const
{
    double de, s1, sinl, cosl, s, sinde, cosde;
    double abs_de;
    int i;

    sinl = sin(Lambda);
    cosl = cos(Lambda);
    s = (K * sinl - H * cosl) * (1 + K * cosl + H * sinl);

    oke[2] = sin(Lambda + s);
    oke[3] = cos(Lambda + s);

    for (i = 0; i < NMax; i++) {
        oke[0] = K * oke[2] - H * oke[3];
        oke[1] = 1 - H * oke[2] - K * oke[3];
        de = (s - oke[0]) / oke[1];
        abs_de = fabs(de);
        if (abs_de < Prec)
            return true;
        s -= de;
        sinde = sin(de);
        cosde = cos(de);
        s1 = oke[2] * cosde - oke[3] * sinde;
        oke[3] = oke[3] * cosde + oke[2] * sinde;
        oke[2] = s1;
    }

    return false;
}

//-----------------------------------------------------------------------------
// Период орбиты
//-----------------------------------------------------------------------------
double Lagrange::GetPeriod() const
{
    return Pi2 * L * L * L / (Mu * Mu);
}

//-----------------------------------------------------------------------------
// Эксцентриситет орбиты
//-----------------------------------------------------------------------------
double Lagrange::GetEcc() const
{
    return sqrt(H * H + K * K);
}

//-----------------------------------------------------------------------------
// Радиус перицентра
//-----------------------------------------------------------------------------
double Lagrange::GetRP() const
{
    double e = GetEcc();
    return L * L * (1.0 - e) / Mu;
}

//-----------------------------------------------------------------------------
// Сохранение в файле
//-----------------------------------------------------------------------------
void Lagrange::Save(FILE* f) const
{
    TAssert(f != NULL);

    fwrite(&Lambda, sizeof(double), 1, f);
    fwrite(&L, sizeof(double), 1, f);
    fwrite(&P, sizeof(double), 1, f);
    fwrite(&Q, sizeof(double), 1, f);
    fwrite(&H, sizeof(double), 1, f);
    fwrite(&K, sizeof(double), 1, f);
}

//-----------------------------------------------------------------------------
// Загрузка из файла
//-----------------------------------------------------------------------------
void Lagrange::Load(FILE* f)
{
    TAssert(f != NULL);

    fread(&Lambda, sizeof(double), 1, f);
    fread(&L, sizeof(double), 1, f);
    fread(&P, sizeof(double), 1, f);
    fread(&Q, sizeof(double), 1, f);
    fread(&H, sizeof(double), 1, f);
    fread(&K, sizeof(double), 1, f);
}

///////////////////////////////////////////////////////////////////////////////
// Вычисление матрицы производных от по прямоугольных координат
// по элементат Лагранжа
///////////////////////////////////////////////////////////////////////////////
bool DXYZDLagrange(const Lagrange& lagr, DSqrMatrix& d)
{
    Vect6 src;
    Vect3 v;
    double el[6];

    if (!lagr.GetXYZ(src)) {
        return false;
    }

    src.GetVelocity(v);

    double r2 = src.Len2();
    double v2 = v.Len2();
    double rv = src * v;
    double r = sqrt(r2);

    // Гравитационная постоянная
    double mu = lagr.GetMu();

    // Интеграл энергии
    double h = v2 - 2.0 * mu / r;

    double c[3], fv[3];

    // Вектор орбитального момента
    c[0] = src[1] * v[2] - src[2] * v[1];
    c[1] = src[2] * v[0] - src[0] * v[2];
    c[2] = src[0] * v[1] - src[1] * v[0];

    // Вектор Лапласа
    fv[0] = -mu * src[0] / r + v[1] * c[2] - v[2] * c[1];
    fv[1] = -mu * src[1] / r + v[2] * c[0] - v[0] * c[2];
    fv[2] = -mu * src[2] / r + v[0] * c[1] - v[1] * c[0];

    // Их квадраты
    double c2 = c[0] * c[0] + c[1] * c[1] + c[2] * c[2];
    double cm = sqrt(c2);
    double fv2 = fv[0] * fv[0] + fv[1] * fv[1] + fv[2] * fv[2];

    // Расчет эксцентрической аномалии
    double cosE = v2 - mu / r;
    double e;
    if (rv >= 0 && cosE > 0)
        e = asin(rv * sqrt(fabs(h)) / sqrt(fv2));
    else if (cosE > 0)
        e = Pi2 + asin(rv * sqrt(fabs(h)) / sqrt(fv2));
    else
        e = Pi - asin(rv * sqrt(fabs(h)) / sqrt(fv2));

    // Расчет элементов Кеплера

    // Большая полуось

    el[0] = -mu / h;

    double cxy = sqrt(c[0] * c[0] + c[1] * c[1]);
    double fi1 = cm * fv[2] / (mu * cxy);
    double fi2 = (c[0] * fv[1] - c[1] * fv[0]) / (mu * cxy);

    // Эксцентриситет

    el[1] = sqrt(fi1 * fi1 + fi2 * fi2);

    // Аргумент перицентра

    if (el[1] == 0)
        el[4] = 0;
    else if (fi1 >= 0 && fi2 > 0)
        el[4] = asin(fi1 / el[1]);
    else if (fi2 > 0)
        el[4] = Pi2 + asin(fi1 / el[1]);
    else
        el[4] = Pi - asin(fi1 / el[1]);

    // Наклонение

    el[2] = acos(c[2] / cm);

    // Долгота восходящего узла

    if (c[0] >= 0 && c[1] < 0)
        el[3] = asin(c[0] / cxy);
    else if (c[1] <= 0)
        el[3] = Pi2 + asin(c[0] / cxy);
    else
        el[3] = Pi - asin(c[0] / cxy);

    // Время прохождения узла

    el[5] = -pow(fabs(h), -1.5) * (mu * e - sqrt(fabs(h)) * rv);

    // Вычисление вспомогательных величин
    unsigned i;

    double vr = rv / r;
    double l1 = mu / (r * r * r);
    double l2 = c2 / (2.0 * mu * sqrt(fv2));
    double l3 = h / (2.0 * mu * sqrt(fv2));
    double l4 = cm / (c[0] * fv[1] - c[1] * fv[0]);
    double l5 = c[2] * fv[2] * cm / (cxy * cxy);
    double l6 = mu * fv[2] / sqrt(fv2);
    double l7 = c[2] / (2 * cm * cm * cm);
    double l8 = mu / (h * r * vr);
    double l9 = l8 * (3 * vr * (-el[5]) / r2 - v2 / mu);
    double l10 = l8 * r2 / mu * (3 * vr * (-el[5]) / r - 2.0);
    double l11 = l8 * (c2 - mu * r) / sqrt(fv2);
    double l12 = -v[2] / (cxy * cxy);
    double l13 = src[2] / (cxy * cxy);

    // Вспомогательные вектора
    double wr[3], wv[3], yr[3], yv[3];

    wr[0] = wr[1] = 0;
    wr[2] = v2 - mu / r;

    wv[0] = wv[1] = 0;
    wv[2] = -rv;

    yr[0] = v[1] / cm;
    yr[1] = -v[0] / cm;
    yr[2] = 0;

    yv[0] = -src[1] / cm;
    yv[1] = src[0] / cm;
    yv[2] = 0;

    // Вспомогательные производные
    double dhdr[3], dhdv[3], dc2dr[3], dc2dv[3];

    for (i = 0; i < 3; i++) {
        dhdr[i] = 2 * l1 * src[i];
        dhdv[i] = 2 * v[i];
    }
    dc2dr[0] = 2 * (v[1] * c[2] - v[2] * c[1]);
    dc2dr[1] = 2 * (v[2] * c[0] - v[0] * c[2]);
    dc2dr[2] = 2 * (v[0] * c[1] - v[1] * c[0]);
    dc2dv[0] = 2 * (c[1] * src[2] - c[2] * src[1]);
    dc2dv[1] = 2 * (c[2] * src[0] - c[0] * src[2]);
    dc2dv[2] = 2 * (c[0] * src[1] - c[1] * src[0]);

    double dcosidr[3], dcosidv[3], dedr[3], dedv[3];
    double dadr[3], dadv[3];
    double dwdr[3], dwdv[3], didr[3], didv[3];
    double domdr[3], domdv[3], dtpdr[3], dtpdv[3];

    for (i = 0; i < 3; i++) {
        dcosidr[i] = yr[i] - l7 * dc2dr[i];
        dcosidv[i] = yv[i] - l7 * dc2dv[i];
        dedr[i] = l2 * dhdr[i] + l3 * dc2dr[i];
        dedv[i] = l2 * dhdv[i] + l3 * dc2dv[i];
        dadr[i] = mu / (h * h) * dhdr[i];
        dadv[i] = mu / (h * h) * dhdv[i];
        dwdr[i] = l4 * (l5 * dcosidr[i] - l6 * dedr[i] + l1 * src[2] * src[i] - v[2] * v[i] + wr[i]);
        dwdv[i] = l4 * (l5 * dcosidv[i] - l6 * dedv[i] + 2 * src[2] * v[i] -
            v[2] * src[i] + wv[i]);
        didr[i] = -1 / sin(el[2]) * dcosidr[i];
        didv[i] = -1 / sin(el[2]) * dcosidv[i];
        domdr[i] = l12 * c[i];
        domdv[i] = l13 * c[i];
        dtpdr[i] = l9 * src[i] + l11 * dedr[i];
        dtpdv[i] = l10 * v[i] + l11 * dedv[i];
    }

    // Заполнение матрицы производных от элементов Кеплера
    // по прямоугольным координатам
    DSqrMatrix res(6);

    for (i = 0; i < 3; i++) {
        res(0, i) = dadr[i];
        res(0, i + 3) = dadv[i];
        res(1, i) = dedr[i];
        res(1, i + 3) = dedv[i];
        res(2, i) = didr[i];
        res(2, i + 3) = didv[i];
        res(3, i) = domdr[i];
        res(3, i + 3) = domdv[i];
        res(4, i) = dwdr[i];
        res(4, i + 3) = dwdv[i];
        res(5, i) = dtpdr[i];
        res(5, i + 3) = dtpdv[i];
    }

    // Вычисление вспомогательных величин для матрицы
    // от элементов Лагранжа по элементам Кеплера
    double ma = sqrt(mu / el[0]);
    double ci2 = cos(el[2] * 0.5);
    double si2 = sin(el[2] * 0.5);
    double com = cos(el[3]);
    double som = sin(el[3]);
    double cp = cos(el[3] + el[4]);
    double sp = sin(el[3] + el[4]);

    DSqrMatrix dldk(6);

    // Заполнение матрицы производных
    // от элементов Лагранжа по элементам Кеплера
    dldk.Init(0.0);

    dldk(0, 0) = 1.5 * ma * el[5] / (el[0] * el[0]);
    dldk(0, 3) = 1.0;
    dldk(0, 4) = 1.0;
    dldk(0, 5) = -ma / el[0];

    dldk(1, 0) = 0.5 * ma;

    dldk(2, 2) = 0.5 * ci2 * com;
    dldk(2, 3) = -si2 * som;

    dldk(3, 2) = 0.5 * ci2 * som;
    dldk(3, 3) = si2 * com;

    dldk(4, 1) = sp;
    dldk(4, 3) = el[1] * cp;
    dldk(4, 4) = el[1] * cp;

    dldk(5, 1) = cp;
    dldk(5, 3) = -el[1] * sp;
    dldk(5, 4) = -el[1] * sp;

    // Теперь получаем матрицу производных от элементов Лагранжа
    // по прямоугольным координатам
    d = dldk;
    d.MulRight(res);

    // Интересующая нас матрица получается обращением последней
    d.Inverse();

    return true;
}

///////////////////////////////////////////////////////////////////////////////
