///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Классы для оптимизации траектории в элементах Лагранжа
///////////////////////////////////////////////////////////////////////////////
#include "LagrTrajectVar.h"
#include "Angles.h"
#include "DrPeriod.h"
#include <iostream>
///////////////////////////////////////////////////////////////////////////////
// Конструктор
//-----------------------------------------------------------------------------
LagrTrajectVar::LagrTrajectVar()
    : allowPoorQuality(false)
{
    int i;

    for (i = 0; i < 8; i++)
        Flags[i] = true;
}

//-----------------------------------------------------------------------------
// Деструктор
//-----------------------------------------------------------------------------
LagrTrajectVar::~LagrTrajectVar()
{
}

//-----------------------------------------------------------------------------
// Установка флагов
//-----------------------------------------------------------------------------
void LagrTrajectVar::SetFlags(const bool* f)
{
    int i;

    for (i = 0; i < 8; i++)
        Flags[i] = f[i];

    Trj.SetKOProp(Kb, Kappa, Flags[6], Flags[7]);
}

//-----------------------------------------------------------------------------
// Установка параметров в элементах Лагранжа
//-----------------------------------------------------------------------------
void LagrTrajectVar::SetParams(const TimeJD& t, const Lagrange& lagr, double kb,
                               double kappa)
{
    Trj.SetTime(t);
    Lagr = lagr;
    Kb = kb;
    Kappa = kappa;
    Trj.SetKOProp(Kb, Kappa, Flags[6], Flags[7]);
}

//-----------------------------------------------------------------------------
// Возврат размерности вектора уточняемых параметров
//-----------------------------------------------------------------------------
unsigned LagrTrajectVar::Size() const
{
    unsigned i, k = 0;

    for (i = 0; i < 8; i++) {
        if (Flags[i])
            k++;
    }

    return k;
}

//-----------------------------------------------------------------------------
// Метод возвращает индекс элемента Лагранжа (0-7),
// k-той компоненте соответствующего вектора
//-----------------------------------------------------------------------------
unsigned LagrTrajectVar::LIndex(unsigned k) const
{
    unsigned c = 0, i;

    for (i = 0; i < 8; i++) {
        if (Flags[i]) {
            if (c == k)
                return i;
            c++;
        }
    }

    TAssert(false);
    return 0;
}

//-----------------------------------------------------------------------------
// Получение текущих параметров в элементах Лагранжа
//-----------------------------------------------------------------------------
void LagrTrajectVar::GetParams(Lagrange& lagr, double& kb, double& kappa,
                               const DVector& u) const
{
    TAssert(u.Dim() == Size());

    unsigned i, k;

    lagr = Lagr;
    kb = Kb;
    kappa = Kappa;

    for (i = 0; i < Size(); i++) {
        k = LIndex(i);
        switch (k) {
        case 0:
            lagr.Lambda = u(i);
            break;
        case 1:
            lagr.L = u(i);
            break;
        case 2:
            lagr.P = u(i);
            break;
        case 3:
            lagr.Q = u(i);
            break;
        case 4:
            lagr.H = u(i);
            break;
        case 5:
            lagr.K = u(i);
            break;
        case 6:
            kb = u(i);
            break;
        case 7:
            kappa = u(i);
            break;
        };
    }
}

//-----------------------------------------------------------------------------
// Контроль входных параметров
//-----------------------------------------------------------------------------
bool LagrTrajectVar::IsValid(const DVector& u) const
{
    Lagrange lagr;
    double kb;
    double kappa;

    GetParams(lagr, kb, kappa, u);

    //	double rp = lagr.GetRP();
    //	std::cout << "rp:" << rp << "\n";
    //	std::cout << "ecc:" << lagr.GetEcc() << "\n";
    //	std::cout << "period:" << lagr.GetPeriod() << "\n\n";
    //	double ecc = lagr.GetEcc();
    //	double period = lagr.GetPeriod();

    if (lagr.GetEcc() > 0.99) {
        return false;
    }

    if (lagr.L < 0.0)
        return false;

    if (lagr.P * lagr.P + lagr.Q * lagr.Q > 1.0)
        return false;
    if (lagr.H * lagr.H + lagr.K * lagr.K > 1.0)
        return false;

    return true;
}

//-----------------------------------------------------------------------------
// Метод устанавливает значение вектора уточняемых параметров
//-----------------------------------------------------------------------------
bool LagrTrajectVar::Init(const DVector& u0)
{
    GetParams(Lagr, Kb, Kappa, u0);

    Vect6 rv;
    if (!Lagr.GetXYZ(rv)) {
        return false;
    }

    Trj.SetKOProp(Kb, Kappa, Flags[6], Flags[7]);
    unsigned nt = Trj.Size();

    DVector ur(nt);

    unsigned i;
    for (i = 0; i < 6; i++)
        ur(i) = rv[i];

    if (Flags[6])
        ur(i++) = Kb;
    if (Flags[7])
        ur(i++) = Kappa;

    TAssert(i == Trj.Size());
    Trj.Init(ur);

    DXYZDLagrange(Lagr, DrDl);

    return true;
}

//-----------------------------------------------------------------------------
// Метод выполняет вычисление вектора уточняемых параметров u для указанного 
// момента времени t.
//-----------------------------------------------------------------------------
bool LagrTrajectVar::Move(const TimeJD& t, DVector& u)
{
    J2000Frame jf;
    Vect6 f;

    if (!GetTrj().Predict(t, jf, f))
        return false;

    unsigned n = Size(), i, k;
    u.Init(n);

    Lagrange lagr;
    lagr.SetXYZ(f);

    for (i = 0; i < Size(); i++) {
        k = LIndex(i);
        switch (k) {
        case 0:
            u(i) = lagr.Lambda;
            break;
        case 1:
            u(i) = lagr.L;
            break;
        case 2:
            u(i) = lagr.P;
            break;
        case 3:
            u(i) = lagr.Q;
            break;
        case 4:
            u(i) = lagr.H;
            break;
        case 5:
            u(i) = lagr.K;
            break;
        case 6:
            u(i) = Kb;
            break;
        case 7:
            u(i) = Kappa;
            break;
        };
    }

    return true;
}

//-----------------------------------------------------------------------------
// Метод выполняет прогнозирование траектории на указанный момент времени t и
// подсчет матрицы частных от фазового вектора по уточняемым параметрам.
// Параметр fr определяет систему отсчету, в которой должен быть возвращен
// фазовый вектор f (т.е. координаты и скорости) и матрица m.
//-----------------------------------------------------------------------------
bool LagrTrajectVar::PredictMatrix(const TimeJD& t, const IFrame& fr, Vect6& f, DMatrix& m)
{
    DMatrix mr;
    if (!GetTrj().PredictMatrix(t, fr, f, mr)) {
        return false;
    }

    FormPDM(mr, m);

    return true;
}

//-----------------------------------------------------------------------------
// Формирование выходной матрицы частных производных
//-----------------------------------------------------------------------------
void LagrTrajectVar::FormPDM(const DMatrix& mr, DMatrix& m)
{
    DSqrMatrix ml(6);

    int i, j, k = 0;;
    for (i = 0; i < 6; i++) {
        for (j = 0; j < 6; j++)
            ml(i, j) = mr(i, j);
    }

    ml.MulRight(DrDl);

    m.Init(6U, Size());

    for (i = 0; i < 6; i++) {
        if (Flags[i]) {
            for (j = 0; j < 6; j++)
                m(j, k) = ml(j, i);
            k++;
        }
    }

    if (Flags[6]) {
        for (j = 0; j < 6; j++)
            m(j, k) = mr(j, 6);
        k++;
    }

    if (Flags[7]) {
        int p = Flags[6] ? 7 : 6;
        for (j = 0; j < 6; j++)
            m(j, k) = mr(j, p);
        k++;
    }
}

//-----------------------------------------------------------------------------
// Запрос приемлимости полученного решения
//-----------------------------------------------------------------------------
bool LagrTrajectVar::CanBeAccepted(const DVector& u)
{
    if (allowPoorQuality) {
        return true;
    }

    unsigned i, nb = 0, np = 0, nbadmax;
    unsigned j;
    double f = 0, fmax;

    if (MPacket->GetSize() <= 3)
        nbadmax = 0;
    else if (MPacket->GetSize() <= 5)
        nbadmax = 1;
    else if (MPacket->GetSize() <= 7)
        nbadmax = 2;
    else if (MPacket->GetSize() <= 10)
        nbadmax = 3;
    else
        nbadmax = (int)floor(0.3 * MPacket->GetSize());

    MPacket->SetFirst();
    for (i = 0; i < MPacket->GetSize(); i++) {
        const IMeasurement* meas = MPacket->GetCurrent();
        MPacket->GoNext();
        DSymmMatrix w;
        meas->GetWeightMatrix(w);
        for (j = 0; j < meas->GetSize(); j++) {
            if (meas->GetIan(j)) {
                nb++;
                break;
            }
        }
        if (nb > nbadmax)
            return false;
        for (j = 0; j < meas->GetSize(); j++) {
            if (meas->GetIan(j))
                continue;
            f += Del[i](j) * Del[i](j) * w(j, j);
            np++;
        }
    }

    if (MPacket->GetSize() <= 3)
        fmax = 3.0;
    else if (MPacket->GetSize() <= 6)
        fmax = 6.0;
    else
        fmax = 15.0;

    return f / np < fmax;
}

//-----------------------------------------------------------------------------
// Реализация алгоритма проверки сходимости
// Возврат true, если сошлись
//-----------------------------------------------------------------------------
static double Km[] = {0.0001, 0.1e-3, 0.00003, 0.00006, 0.00005, 0.00005};
static double Ks[] = {0.0008, 0.9e-3, 0.0001, 0.0001, 0.001, 0.001};
static double Kn[] = {0.001, 1.0e-3, 0.001, 0.001, 0.001, 0.001};

bool LagrTrajectVar::IsMin(const DVector& u, double f,
                           const DVector& u1, double f1,
                           const DSymmMatrix& s, double alpha, unsigned nit,
                           EnMinMnkPath& path)
{
    path = UNDEFINED_PATH;

    if (fabs(f - f1) > 0.03 * f)
        return false;

    unsigned i, k, n = Size();
    DVector du(n);
    for (i = 0; i < n; i++) {
        k = LIndex(i);
        if (k == 0)
            du[i] = fabs(AngleDiff(u[i], u1[i]));
        else
            du[i] = fabs(u[i] - u1[i]);
    }

    for (i = 0; i < n; i++) {
        k = LIndex(i);
        if (k >= 6)
            continue;
        double d = alpha * s(i, i);
        if (Km[k] > d)
            d = Km[k];
        if (du[i] * du[i] > d)
            break;
    }

    if (i == n)
        return true;

    Lagrange lagr;
    double kb, kappa;
    GetParams(lagr, kb, kappa, u1);

    double tp, dt;
    GetOrbPeriod(*GetTrj().GetPredictor(), GetTime(), tp, dt, "default");

    if (fabs(dt) > 0.000006 && nit > 1) {
        for (i = 0; i < n; i++) {
            k = LIndex(i);
            if (k == 6) {
                if (du[i] > 0.03 * fabs(kb))
                    break;
            }
            else if (k == 7)
                continue;
            else if (k > 7)
                continue;
            else if (du[i] > Ks[k])
                break;
        }
        if (i == n)
            return true;
    }

    for (i = 0; i < n; i++) {
        k = LIndex(i);
        if (k == 6) {
            if (du[i] > 2 * fabs(kb))
                break;
        }
        else if (k == 7) {
            if (du[i] > 2 * fabs(kappa))
                break;
        }
        else if (k > 7)
            continue;
        else if (du[i] > Kn[k])
            break;
    }

    if (i == n)
        path = GN_PATH;

    return false;
}

//-----------------------------------------------------------------------------
// Реализация операции сложения для уточняемого вектора
//-----------------------------------------------------------------------------
void LagrTrajectVar::AddU(const DVector& u0, const DVector& du,
                          DVector& u) const
{
    unsigned n = Size(), i, k;

    for (i = 0; i < n; i++) {
        k = LIndex(i);
        u[i] = u0[i] + du[i];
        if (k == 0)
            u[i] = Angle2Pi(u[i]);
    }
}

//-----------------------------------------------------------------------------
// Приращения, используемые при численном расчете матрицы частных производных
//-----------------------------------------------------------------------------
void LagrTrajectVar::GetEps(DVector& eps) const
{
    eps.Init(Size());

    unsigned i;
    for (i = 0; i < Size(); i++) {
        int k = LIndex(i);
        switch (k) {
        case 0:
        case 2:
        case 3:
        case 4:
        case 5:
            eps(i) = 1e-6;
            break;
        case 1:
            eps(i) = 1e-6;
            break;
        case 6:
            eps(i) = 1e-5;
            break;
        case 7:
            eps(i) = 1e-8;
            break;
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
