///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Определение класса описания геоцентрической системы METE
////////////////////////////////////////////////////////////////////////////////
#include "METEFrame.h"
#include "NZEFrame.h"
#include "FGreenwichFrame.h"
#include "IAstroBase.h"
////////////////////////////////////////////////////////////////////////////////
// Установка параметров системы
void METEFrame::SetParameters(const IFrameParameters& p)
{
    TAssert(p.GetFrameType() == csIdMETE);
    const METEFrameParameters& pm = (const METEFrameParameters &)p;
    Pars = pm;
}

//------------------------------------------------------------------------------
// Получение объекта, который соответствует переходу от текущей системы отсчета 
// к системе отсчета J2000 на момент времени t0 (для 3-мерных векторов)
void METEFrame::Get3DJ2000Converter(const TimeJD& t0, FrameConverter3D& result) const
{
    TAssert(Pars.IsInit());

    result.Init();

    GetAstroBase()->MtxJ20002METE(Pars.GetTime(), result.Rotation);
    result.Rotation.Transp();
    return;
}

//------------------------------------------------------------------------------
// Получение объекта, который соответствует переходу от текущей системы отсчета 
// к системе отсчета J2000 на момент времени t0 (для 6-мерных векторов)
void METEFrame::Get6DJ2000Converter(const TimeJD& t0, FrameConverter6D& result) const
{
    TAssert(Pars.IsInit());
    result.Init();
    Get3DJ2000Converter(t0, result.Converter3D);
}

//------------------------------------------------------------------------------
// Получение объекта, который соответствует переходу от текущей системы отсчета 
// к системе отсчета rf на момент времени t0 (для 3-мерных векторов)
void METEFrame::Get3DConverter(const IFrame& rf, const TimeJD& t0,
                               FrameConverter3D& result) const
{
    TAssert(Pars.IsInit());

    if (rf.GetType() == csIdGFCS) {
        const FGreenwichFrame& gf = (const FGreenwichFrame &)rf;
        if (Pars.GetTime() == gf.GetFixTime()) {
            result.Init();
            GetAstroBase()->MtxMETE2GCS(Pars.GetTime(), result.Rotation);
            return;
        }
    }

    if (rf.GetType() == csIdGCS) {
        if (t0 == Pars.GetTime()) {
            result.Init();
            GetAstroBase()->MtxMETE2GCS(t0, result.Rotation);
            return;
        }
    }

    if (rf.GetType() == csIdNZE) {
        const NZEFrame& lf = (const NZEFrame &)rf;
        if (Pars.GetTime() == t0) {
            FrameConverter3D mg, gl;
            mg.Init();
            GetAstroBase()->MtxMETE2GCS(t0, mg.Rotation);
            lf.Get3DGreenwichConverter(gl);
            gl.Invert();
            mg.Multiply(gl, result);
            return;
        }
    }

    IFrame::Get3DConverter(rf, t0, result);
}

//------------------------------------------------------------------------------
// Получение объекта, который соответствует переходу от текущей системы отсчета 
// к системе отсчета rf на момент времени t0 (для 6-мерных векторов)
void METEFrame::Get6DConverter(const IFrame& rf, const TimeJD& t0,
                               FrameConverter6D& result) const
{
    TAssert(Pars.IsInit());

    if (rf.GetType() == csIdGFCS) {
        const FGreenwichFrame& gf = (const FGreenwichFrame &)rf;
        if (Pars.GetTime() == gf.GetFixTime()) {
            result.Init();
            Get3DConverter(rf, t0, result.Converter3D);
            return;
        }
    }

    if (rf.GetType() == csIdGCS) {
        if (t0 == Pars.GetTime()) {
            result.Init();
            Get3DConverter(rf, t0, result.Converter3D);
            TransMtx w;
            GetAstroBase()->GCSVelocityMatrix(w);
            w.mul(result.Converter3D.Rotation, result.VRotation);
            return;
        }
    }

    if (rf.GetType() == csIdNZE) {
        const NZEFrame& lf = (const NZEFrame &)rf;
        if (Pars.GetTime() == t0) {
            FrameConverter6D mg, gl;
            mg.Init();
            GetAstroBase()->MtxMETE2GCS(t0, mg.Converter3D.Rotation);
            TransMtx w;
            GetAstroBase()->GCSVelocityMatrix(w);
            w.mul(mg.Converter3D.Rotation, mg.VRotation);
            lf.Get6DGreenwichConverter(gl);
            gl.Invert();
            mg.Multiply(gl, result);
            return;
            return;
        }
    }

    IFrame::Get6DConverter(rf, t0, result);
}

////////////////////////////////////////////////////////////////////////////////
