///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Класс, реализующий вектор-столбец
///////////////////////////////////////////////////////////////////////////////
#include "Matrix.h"

#include <math.h>
///////////////////////////////////////////////////////////////////////////////
// Получение квадратной матрицы в результате перемножения 
// нижней и верхней треугольных матриц
///////////////////////////////////////////////////////////////////////////////
DSqrMatrix& Mult(DSqrMatrix& dst, const DLowMatrix& src1, const DUppMatrix& src2)
{
    TAssert(src1.Dim() == src2.Dim());
    TAssert(src1.Buffer != NULL);
    TAssert(src2.Buffer != NULL);

    dst.Init(src1.Dimension);

    unsigned i, j, k;
    double* line;

    for (i = 0; i < src1.Dimension; i++) {
        line = dst.Matrix[i];
        for (j = 0; j < src2.Dimension; j++) {
            line[j] = 0.0;
            for (k = 0; k < j + 1 && k < i + 1; k++)
                line[j] += src1.Buffer[src1.Index(k, i)] * src2.Buffer[src2.Index(k, j)];
        }
    }

    return dst;
}

///////////////////////////////////////////////////////////////////////////////
// Получение квадратной матрицы в результате перемножения 
// верхней и нижней треугольных матриц
///////////////////////////////////////////////////////////////////////////////
DSqrMatrix& Mult(DSqrMatrix& dst, const DUppMatrix& src1, const DLowMatrix& src2)
{
    TAssert(src1.Dim() == src2.Dim());
    TAssert(src1.Buffer != NULL);
    TAssert(src2.Buffer != NULL);

    dst.Init(src1.Dimension);

    unsigned i, j, k;
    double* line;

    for (i = 0; i < src1.Dim(); i++) {
        line = dst.Matrix[i];
        for (j = 0; j < src2.Dim(); j++) {
            line[j] = 0.0;
            for (k = (i > j) ? i : j; k < src2.Dim(); k++)
                line[j] += src1.Buffer[src1.Index(i, k)] * src2.Buffer[src2.Index(j, k)];
        }
    }

    return dst;
}

///////////////////////////////////////////////////////////////////////////////
// Решение системы уравнений с квадратной матрицей произвольного вида
///////////////////////////////////////////////////////////////////////////////
bool TDMSolution(const DVector& B, const CDSqrMatrix& A, DVector& X)
{
    TAssert(A.Dim() == B.Dim());

    DSqrMatrix inv(A);

    if (!inv.Inverse()) {
        X.Init((unsigned)0);
        return false;
    }

    Mult(X, inv, B);
    return true;
}

///////////////////////////////////////////////////////////////////////////////
// Решение системы уравнений с квадратной симметричной матрицей произвольного 
// вида
///////////////////////////////////////////////////////////////////////////////
bool TDMSolution(const DVector& B, const CDSymmMatrix& A, DVector& X)
{
    TAssert(A.Dim() == B.Dim());

    DUppMatrix up_mt;
    DDiagMatrix di_mt;

    if (!A.Expand(up_mt, di_mt)) {
        if (X.Vector != NULL) {
            delete X.Vector;
            X.Vector = NULL;
        }
        return false;
    }

    DLowMatrix lo_mt;
    lo_mt.Transpose(up_mt);

    DVector x, tmp;

    if (!TDMSolution(B, lo_mt, tmp))
        return false;
    if (!TDMSolution(tmp, di_mt, x))
        return false;
    if (!TDMSolution(x, up_mt, X))
        return false;

    return true;
}

///////////////////////////////////////////////////////////////////////////////
// Решение системы уравнений с диагональной матрицей
///////////////////////////////////////////////////////////////////////////////
bool TDMSolution(const DVector& B, const DDiagMatrix& A, DVector& X)
{
    TAssert(A.Dim() == B.Dim());

    X.Init(A.Dimension);

    unsigned i;

    for (i = 0; i < A.Dimension; i++) {
        if (fabs(A.Vector[i]) < ZERO_ACCURACY) {
            X.Init((unsigned)0);
            return false;
        }
        X.Vector[i] = B.Vector[i] / A.Vector[i];
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////
// Решение системы уравнений с верхней треугольной матрицей
///////////////////////////////////////////////////////////////////////////////
bool TDMSolution(const DVector& b, const DUppMatrix& a, DVector& x)
{
    TAssert(b.Dim() == a.Dim());

    unsigned n = a.Dimension;

    x.Init(n);

    if (fabs(a.Buffer[a.Index(n - 1, n - 1)]) <= ZERO_ACCURACY) {
        x.Init((unsigned)0);
        return false;
    }

    x.Vector[n - 1] = b.Vector[n - 1] / a.Buffer[a.Index(n - 1, n - 1)];

    double sum;
    int i;
    unsigned j;

    for (i = n - 2; i >= 0; --i) {
        sum = 0.0;
        for (j = i + 1; j < n; ++j)
            sum += x[j] * a.Buffer[a.Index(i, j)];
        if (fabs(a.Buffer[a.Index(i, i)]) <= ZERO_ACCURACY) {
            x.Init((unsigned)0);
            return false;
        }
        x.Vector[i] = (b.Vector[i] - sum) / a.Buffer[a.Index(i, i)];
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////
// Решение системы уравнений с нижней треугольной матрицей
///////////////////////////////////////////////////////////////////////////////
bool TDMSolution(const DVector& b, const DLowMatrix& a, DVector& x)
{
    TAssert(b.Dim() == a.Dim());

    unsigned n = a.Dimension;

    x.Init(n);

    if (fabs(a.Buffer[0]) <= ZERO_ACCURACY) {
        x.Init((unsigned)0);
        return false;
    }

    x.Vector[0] = b.Vector[0] / a.Buffer[0];

    double sum;
    unsigned i, j;

    for (i = 1; i < n; i++) {
        sum = 0.0;
        for (j = 0; j < i; ++j)
            sum += x.Vector[j] * a.Buffer[a.Index(j, i)];
        if (fabs(a.Buffer[a.Index(i, i)]) <= ZERO_ACCURACY) {
            n = 0;
            return false;
        }
        x.Vector[i] = (b.Vector[i] - sum) / a.Buffer[a.Index(i, i)];
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////
// Решение системы уравнений с нижней треугольной матрицей
///////////////////////////////////////////////////////////////////////////////
bool TDMSolutionJ(const DVector& B, DSymmMatrix& A, DVector& X)
{
    TAssert(A.Dim() == B.Dim());

    DMatrix R;
    DDiagMatrix S;
    DSymmMatrix a(A);

    a.JacobiExpand(S, R);

    if (S(0) <= 0) return false;
    double Smin = S(0) * 1E-9;
    for (unsigned i = 0; i < S.Dim(); i++)
        if (S(i) <= Smin)
            S(i) = Smin;
    //return false;
    // if (S(0)/S(S.Dim()-1)>1.0e18)
    //	return false;

    DSymmMatrix D(S);

    for (unsigned i = 0; i < S.Dim(); i++)
        D(i, i) = 1 / S(i);

    ///////////////////////////
    a.TRectSymmRecmt(R, D);
    X = B;
    X.MulLeft(a);

    return true;
}

///////////////////////////////////////////////////////////////////////////////
// Реализация скалярного умножений
///////////////////////////////////////////////////////////////////////////////
double operator*(const DVector& a, const DVector& b)
{
    TAssert(a.Dimension == b.Dimension);
    TAssert(a.Vector != NULL);
    TAssert(b.Vector != NULL);

    double s = 0.0;
    unsigned i;

    for (i = 0; i < a.Dimension; i++)
        s += a.Vector[i] * b.Vector[i];

    return s;
}

///////////////////////////////////////////////////////////////////////////////
