///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Классы представления матриц
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_MATRIX_H
#define BNOTOOL_MATRIX_H
///////////////////////////////////////////////////////////////////////////////
#include "Defines.h"

#include <string.h>
///////////////////////////////////////////////////////////////////////////////
// Предопределения классов
///////////////////////////////////////////////////////////////////////////////
class DMatrix;
class DVector;
class DSqrMatrix;
class DSymmMatrix;
class DUppMatrix;
class DLowMatrix;
class DDiagMatrix;

///////////////////////////////////////////////////////////////////////////////
// Общий интерфейс прямоугольной матрицы, доступной только для чтения
///////////////////////////////////////////////////////////////////////////////
class CDMatrix
{
public:
    // Запрос числа строк
    virtual ~CDMatrix()
    {
    };
    // Запрос числа строк
    virtual unsigned Rows() const = 0;
    // Запрос числа столбцов
    virtual unsigned Cols() const = 0;
    // Получение значения элемента
    virtual double operator()(unsigned r, unsigned c) const
    {
        return Val(r, c);
    };
    // Запрос значений элементов в виде массива
    // (массив развертнут по строкам)
    virtual void GetArray(double* array) const = 0;
    // Запрос значений строки
    virtual void GetLine(unsigned index, double* line) const = 0;
    // Запрос значений столбца
    virtual void GetCol(unsigned index, double* col) const = 0;

protected:
    virtual double Val(unsigned r, unsigned c) const = 0;
};

///////////////////////////////////////////////////////////////////////////////
// Общий интерфейс квадратной матрицы, доступной только для чтения
///////////////////////////////////////////////////////////////////////////////
class CDSqrMatrix : public CDMatrix
{
public:
    unsigned Dim(void) const
    {
        return Rows();
    };
};

///////////////////////////////////////////////////////////////////////////////
// Общий интерфейс симметричной матрицы, доступной только для чтения
///////////////////////////////////////////////////////////////////////////////
class CDSymmMatrix : public CDSqrMatrix
{
public:
    // Разложение матрицы на верхнюю и диагональную
    virtual bool Expand(DUppMatrix& s, DDiagMatrix& d) const = 0;
};

///////////////////////////////////////////////////////////////////////////////
// Служебный класс хранения прямоугольного массива в памяти
///////////////////////////////////////////////////////////////////////////////
class DRectMem
{
    // Дружественные классы
    friend class DTriagMem;
    friend class DLineMem;
    friend class DSymmMatrix;

    // Функции создания массива путем умножения 
    friend void MultRect(DRectMem& dst, const DRectMem& src1, const DRectMem& src2);
    friend void MultRect(DRectMem& dst, const DRectMem& src1, const DUppMatrix& src2);
    friend void MultRect(DRectMem& dst, const DUppMatrix& src1, const DRectMem& src2);
    friend void MultRect(DRectMem& dst, const DRectMem& src1, const DLowMatrix& src2);
    friend void MultRect(DRectMem& dst, const DLowMatrix& src1, const DRectMem& src2);
    friend void MultRect(DRectMem& dst, const DRectMem& src1, const DDiagMatrix& src2);
    friend void MultRect(DRectMem& dst, const DDiagMatrix& src1, const DRectMem& src2);

public:
    // Конструктор
    DRectMem();
    // Дестируктор
    virtual ~DRectMem();

    // Установка единого значения для всех элементов
    void Init(double value);

protected:
    // Число строк
    unsigned NRows;
    // Число колонок
    unsigned NCols;
    // Число строк в памяти
    unsigned NRowsMem;
    // Число колонок в памяти
    unsigned NColsMem;
    // Размер буфера
    unsigned BufSize;
    // Буфер
    double** Matrix;

    // Вычисление требуемого размера буфера
    static unsigned ReqSize(unsigned n, unsigned m)
    {
        return sizeof(double) * m * n + sizeof(double *) * n;
    };
    // Создание буфера
    virtual double** CreatePtr(unsigned n, unsigned m);
    // Разметка указателей
    virtual void SetLinesPtrs(double** ptr, unsigned n, unsigned m);

    // Обмен содержимого
    void XChng(DRectMem& m);
    // Инициализация 
    void Init(unsigned r, unsigned c);
    // Изменение размерности
    void ResetDimension(unsigned r, unsigned c);

    // Инициализация
    void Init(const CDMatrix& mt);
    void InitRect(const DRectMem& mt);
    void Init(const DSymmMatrix& mt);
    void Init(const DUppMatrix& mt);
    void Init(const DLowMatrix& mt);
    void Init(const DDiagMatrix& mt);

    // Умножение всех элементов на число
    void Mult(double mltp);

    // Операции сложения и вычитания
    void Add(const CDMatrix& mt);
    void Sub(const CDMatrix& mt);
    void AddRect(const DRectMem& mt);
    void SubRect(const DRectMem& mt);
    void Add(const DSymmMatrix& mt);
    void Sub(const DSymmMatrix& mt);
    void Add(const DUppMatrix& mt);
    void Sub(const DUppMatrix& mt);
    void Add(const DLowMatrix& mt);
    void Sub(const DLowMatrix& mt);
    void Add(const DDiagMatrix& mt);
    void Sub(const DDiagMatrix& mt);

    // Бинарные операции умножения
    void Mult(const DRectMem& m1, const DRectMem& m2);
    void Mult(const DRectMem& m, const DUppMatrix& mt);
    void Mult(const DUppMatrix& mt, const DRectMem& m);
    void Mult(const DRectMem& m, const DLowMatrix& mt);
    void Mult(const DLowMatrix& mt, const DRectMem& m);
    void Mult(const DRectMem& m, const DDiagMatrix& mt);
    void Mult(const DDiagMatrix& mt, const DRectMem& m);

    // Унарные операции умножения
    void MulRightRect(const DRectMem& mt);
    void MulLeftRect(const DRectMem& mt);
    void MulRight(const DUppMatrix& mt);
    void MulLeft(const DUppMatrix& mt);
    void MulRight(const DLowMatrix& mt);
    void MulLeft(const DLowMatrix& mt);
    void MulRight(const DDiagMatrix& mt);
    void MulLeft(const DDiagMatrix& mt);
private:
    DRectMem(const DRectMem&);
};

///////////////////////////////////////////////////////////////////////////////
// Служебный класс хранения вектора в памяти
///////////////////////////////////////////////////////////////////////////////
class DLineMem
{
    // Дружественные классы
    friend class DRectMem;
    friend class DTriagMem;
    friend class DSymmMatrix;
    friend class DUppMatrix;
    friend class DLowMatrix;

    // Функции создания вектора путем умножения 
    friend DVector& Mult(DVector& dst, const DVector& src1, const CDMatrix& src2);
    friend DVector& Mult(DVector& dst, const CDMatrix& src1, const DVector& src2);
    friend DVector& Mult(DVector& dst, const DVector& src1, const DMatrix& src2);
    friend DVector& Mult(DVector& dst, const DMatrix& src1, const DVector& src2);
    friend DVector& Mult(DVector& dst, const DVector& src1, const DSqrMatrix& src2);
    friend DVector& Mult(DVector& dst, const DSqrMatrix& src1, const DVector& src2);

public:
    // Конструктор
    DLineMem();
    // Деструктор
    virtual ~DLineMem();

protected:
    unsigned Dimension, DimMem;
    unsigned MemSize;
    double* Vector;

    // Инициализация
    void Init(unsigned dim);
    void Init(const double* vec, unsigned dim);
    // Изменение размерности
    void ResetDimension(unsigned dim);

    // Обмен содержимого
    void XChng(DLineMem& m);

    // Поиск максимального и минимального элемента 
    double GetMaxElementAbs() const;
    double GetMinElementAbs() const;
    double GetMaxElement() const;
    double GetMinElement() const;

    // Установка единого значения для всех элементов
    void Init(double value);
    // Умножение всех элементов на число
    void Mult(double mltp);

    // Операции сложения и вычитания
    void Add(const CDMatrix& mt);
    void Sub(const CDMatrix& mt);
    void AddVect(const DLineMem& mt);
    void SubVect(const DLineMem& mt);

    // Операции умножения
    void MulLeftRect(const DRectMem& mt);
    void MulRect(const DRectMem& mt, const DLineMem& vc);
private:
    DLineMem(const DLineMem&);
};

///////////////////////////////////////////////////////////////////////////////
// Служебный класс хранения треугольной матрицы в памяти
///////////////////////////////////////////////////////////////////////////////
class DTriagMem
{
    // Дружественные классы
    friend class DRectMem;
    friend class DLineMem;

public:
    // Конструктор
    DTriagMem();
    // Деструктор
    virtual ~DTriagMem();

protected:
    unsigned Dimension, DimMem, MemSize;
    double* Buffer;

    // Обмен содержимого
    void XChng(DTriagMem& m);
    // Инициализация
    void Init(unsigned dim);
    // Изменение размерности
    void ResetDimension(unsigned dim);

    // Установка единого значения для всех элементов
    void Init(double value);
    // Умножение всех элементов на число
    void Mult(double mltp);

    // Вычисление требуемого размера буфера
    static inline unsigned ReqMemSize(unsigned n)
    {
        return (n * (n - 1) >> 1) + n;
    };
    // Вычисление индекса элемента в массиве
    unsigned Index(unsigned x, unsigned y) const
    {
        return ((x * ((DimMem << 1) - x - 1)) >> 1) + y;
    };
private:
    DTriagMem(const DTriagMem&);
};

///////////////////////////////////////////////////////////////////////////////
// Класс, реализующий произвольную прямоугольную матрицу
///////////////////////////////////////////////////////////////////////////////
class DMatrix : public CDMatrix, public DRectMem
{
    friend class DSqrMatrix;

public:
    // Конструктор по умолчанию
    DMatrix() : DRectMem()
    {
    };
    // Конструктор с указанием размерности
    DMatrix(unsigned r, unsigned c)
    {
        DRectMem::Init(r, c);
    };
    // Конструкторы копирования
    DMatrix(const DMatrix& mt) : CDMatrix(), DRectMem()
    {
        DRectMem::InitRect(mt);
    };
    DMatrix(const DSqrMatrix& mt);

    DMatrix(const CDMatrix& mt)
    {
        DRectMem::Init(mt);
    };

    DMatrix(const CDSymmMatrix& mt)
    {
        DRectMem::Init(mt);
    };

    DMatrix(const DSymmMatrix& mt)
    {
        DRectMem::Init(mt);
    };

    DMatrix(const DUppMatrix& mt)
    {
        DRectMem::Init(mt);
    };

    DMatrix(const DLowMatrix& mt)
    {
        DRectMem::Init(mt);
    };

    DMatrix(const DDiagMatrix& mt)
    {
        DRectMem::Init(mt);
    };

    // Деструктор
    virtual ~DMatrix(void)
    {
    };

    // Получение значения элемента
    virtual double operator()(unsigned r, unsigned c) const
    {
        return Val(r, c);
    };
    // Доступ к элементу матрицы
    double& operator()(unsigned r, unsigned c)
    {
        TAssert(r < NRows && c < NCols);
        TAssert(Matrix != NULL);
        return Matrix[r][c];
    };

    // Операторы присваивания
    DMatrix& operator=(const DMatrix& mt)
    {
        InitRect(mt);
        return *this;
    };
    DMatrix& operator=(const DSqrMatrix& mt);

    DMatrix& operator=(const CDMatrix& mt)
    {
        DRectMem::Init(mt);
        return *this;
    };

    DMatrix& operator=(const CDSymmMatrix& mt)
    {
        DRectMem::Init(mt);
        return *this;
    };

    DMatrix& operator=(const DSymmMatrix& mt)
    {
        DRectMem::Init(mt);
        return *this;
    };

    DMatrix& operator=(const DUppMatrix& mt)
    {
        DRectMem::Init(mt);
        return *this;
    };

    DMatrix& operator=(const DLowMatrix& mt)
    {
        DRectMem::Init(mt);
        return *this;
    };

    DMatrix& operator=(const DDiagMatrix& mt)
    {
        DRectMem::Init(mt);
        return *this;
    };

    // Инициализация 
    void Init(unsigned r, unsigned c)
    {
        DRectMem::Init(r, c);
    };
    // Установка единого значения для всех элементов
    void Init(double value)
    {
        DRectMem::Init(value);
    };

    // Изменение размерности
    void ResetDimension(unsigned r, unsigned c)
    {
        DRectMem::ResetDimension(r, c);
    };

    // Умножение на число
    DMatrix& operator*=(double mltp)
    {
        Mult(mltp);
        return *this;
    }

    // Операции сложения и вычитания
    DMatrix& operator+=(const CDMatrix& mt)
    {
        Add(mt);
        return *this;
    };

    DMatrix& operator+=(const DMatrix& mt)
    {
        AddRect(mt);
        return *this;
    };
    DMatrix& operator+=(const DSqrMatrix& mt);

    DMatrix& operator-=(const CDMatrix& mt)
    {
        Sub(mt);
        return *this;
    };

    DMatrix& operator-=(const DMatrix& mt)
    {
        SubRect(mt);
        return *this;
    };
    DMatrix& operator-=(const DSqrMatrix& mt);

    DMatrix& operator+=(const CDSymmMatrix& mt)
    {
        Add(mt);
        return *this;
    };

    DMatrix& operator+=(const DSymmMatrix& mt)
    {
        Add(mt);
        return *this;
    };

    DMatrix& operator-=(const CDSymmMatrix& mt)
    {
        Sub(mt);
        return *this;
    };

    DMatrix& operator-=(const DSymmMatrix& mt)
    {
        Sub(mt);
        return *this;
    };
    DMatrix& operator+=(const DUppMatrix& mt);
    DMatrix& operator-=(const DUppMatrix& mt);
    DMatrix& operator+=(const DLowMatrix& mt);
    DMatrix& operator-=(const DLowMatrix& mt);
    DMatrix& operator+=(const DDiagMatrix& mt);
    DMatrix& operator-=(const DDiagMatrix& mt);

    // Операции умножения
    DMatrix& MulRight(const DMatrix& mt)
    {
        MulRightRect(mt);
        return *this;
    };

    inline DMatrix& operator*=(const DMatrix& mt)
    {
        return MulRight(mt);
    };
    DMatrix& MulRight(const DSqrMatrix& mt);

    inline DMatrix& operator*=(const DSqrMatrix& mt)
    {
        return MulRight(mt);
    };

    DMatrix& MulRight(const CDMatrix& mt)
    {
        MulRightRect(DMatrix(mt));
        return *this;
    };

    inline DMatrix& operator*=(const CDMatrix& mt)
    {
        return MulRight(mt);
    };

    DMatrix& MulLeft(const DMatrix& mt)
    {
        MulLeftRect(mt);
        return *this;
    };
    DMatrix& MulLeft(const DSqrMatrix& mt);

    DMatrix& MulLeft(const CDMatrix& mt)
    {
        MulLeftRect(DMatrix(mt));
        return *this;
    };

    DMatrix& MulRight(const DUppMatrix& mt)
    {
        DRectMem::MulRight(mt);
        return *this;
    };

    DMatrix& MulLeft(const DUppMatrix& mt)
    {
        DRectMem::MulLeft(mt);
        return *this;
    };

    DMatrix& operator*=(const DUppMatrix& mt)
    {
        return MulRight(mt);
    };

    DMatrix& MulRight(const DLowMatrix& mt)
    {
        DRectMem::MulRight(mt);
        return *this;
    };

    DMatrix& MulLeft(const DLowMatrix& mt)
    {
        DRectMem::MulLeft(mt);
        return *this;
    };

    DMatrix& operator*=(const DLowMatrix& mt)
    {
        return MulRight(mt);
    };

    DMatrix& MulRight(const DDiagMatrix& mt)
    {
        DRectMem::MulRight(mt);
        return *this;
    };

    DMatrix& MulLeft(const DDiagMatrix& mt)
    {
        DRectMem::MulLeft(mt);
        return *this;
    };

    DMatrix& operator*=(const DDiagMatrix& mt)
    {
        return MulRight(mt);
    };

    // Транспонирование
    void Transpose();

    // Реализация запроса числа строк
    virtual unsigned Rows() const
    {
        return NRows;
    };
    // Реализация запроса числа столбцов
    virtual unsigned Cols() const
    {
        return NCols;
    };
    // Реализация получения элемента
    virtual double Val(unsigned r, unsigned c) const
    {
        TAssert(r < NRows && c < NCols);
        TAssert(Matrix != NULL);
        return Matrix[r][c];
    };
    // Реализация запроса значений элементов в виде массива
    virtual void GetArray(double* array) const;
    // Реализация запроса значений строки
    virtual void GetLine(unsigned index, double* line) const
    {
        TAssert(index < NRows);
        TAssert(Matrix != NULL);
        TAssert(line != NULL);
        memcpy(line, Matrix[index], sizeof(double) * NCols);
    }

    // Реализация запроса значений столбца
    virtual void GetCol(unsigned index, double* col) const;
};

///////////////////////////////////////////////////////////////////////////////
// Класс, реализующий квадратную матрицу
///////////////////////////////////////////////////////////////////////////////
class DSqrMatrix : public CDSqrMatrix, public DRectMem
{
    friend class DMatrix;
    friend DSqrMatrix& Mult(DSqrMatrix& dst, const DLowMatrix& src1, const DUppMatrix& src2);
    friend DSqrMatrix& Mult(DSqrMatrix& dst, const DUppMatrix& src1, const DLowMatrix& src2);

public:
    // Конструктор по умолчанию
    DSqrMatrix()
    {
    };
    // Конструктор с указанием размерности
    DSqrMatrix(unsigned n)
    {
        DRectMem::Init(n, n);
    };
    // Конструкторы копирования
    DSqrMatrix(const CDSqrMatrix& mt)
    {
        DRectMem::Init(mt);
    };

    DSqrMatrix(const DSqrMatrix& mt) : CDSqrMatrix(), DRectMem()
    {
        InitRect(mt);
    };

    DSqrMatrix(const CDSymmMatrix& mt)
    {
        DRectMem::Init(mt);
    };

    DSqrMatrix(const DSymmMatrix& mt)
    {
        DRectMem::Init(mt);
    };

    DSqrMatrix(const DUppMatrix& mt)
    {
        DRectMem::Init(mt);
    };

    DSqrMatrix(const DLowMatrix& mt)
    {
        DRectMem::Init(mt);
    };

    DSqrMatrix(const DDiagMatrix& mt)
    {
        DRectMem::Init(mt);
    };

    // Деструктор
    virtual ~DSqrMatrix(void)
    {
    };

    // Получение значения элемента
    virtual double operator()(unsigned r, unsigned c) const
    {
        return Val(r, c);
    };
    // Доступ к элементу матрицы
    double& operator()(unsigned r, unsigned c)
    {
        TAssert(r < NRows && c < NCols);
        TAssert(Matrix != NULL);
        return Matrix[r][c];
    };

    // Операторы присваивания
    DSqrMatrix& operator=(const DSqrMatrix& mt)
    {
        InitRect(mt);
        return *this;
    };

    DSqrMatrix& operator=(const CDSqrMatrix& mt)
    {
        DRectMem::Init(mt);
        return *this;
    };

    DSqrMatrix& operator=(const CDSymmMatrix& mt)
    {
        DRectMem::Init(mt);
        return *this;
    };

    DSqrMatrix& operator=(const DSymmMatrix& mt)
    {
        DRectMem::Init(mt);
        return *this;
    };

    DSqrMatrix& operator=(const DUppMatrix& mt)
    {
        DRectMem::Init(mt);
        return *this;
    };

    DSqrMatrix& operator=(const DLowMatrix& mt)
    {
        DRectMem::Init(mt);
        return *this;
    };

    DSqrMatrix& operator=(const DDiagMatrix& mt)
    {
        DRectMem::Init(mt);
        return *this;
    };

    // Инициализация 
    void Init(unsigned n)
    {
        DRectMem::Init(n, n);
    };
    // Установка единого значения для всех элементов
    void Init(double value)
    {
        DRectMem::Init(value);
    };
    // Изменение размерности
    void ResetDimension(unsigned n)
    {
        DRectMem::ResetDimension(n, n);
    };

    // Умножение на число
    DSqrMatrix& operator*=(double mltp)
    {
        Mult(mltp);
        return *this;
    }

    // Операции сложения и вычитания
    DSqrMatrix& operator+=(const CDSqrMatrix& mt)
    {
        Add(mt);
        return *this;
    };

    DSqrMatrix& operator+=(const DSqrMatrix& mt)
    {
        AddRect(mt);
        return *this;
    };

    DSqrMatrix& operator-=(const CDSqrMatrix& mt)
    {
        Sub(mt);
        return *this;
    };

    DSqrMatrix& operator-=(const DSqrMatrix& mt)
    {
        SubRect(mt);
        return *this;
    };

    DSqrMatrix& operator+=(const CDSymmMatrix& mt)
    {
        Add(mt);
        return *this;
    };

    DSqrMatrix& operator+=(const DSymmMatrix& mt)
    {
        Add(mt);
        return *this;
    };

    DSqrMatrix& operator-=(const CDSymmMatrix& mt)
    {
        Sub(mt);
        return *this;
    };

    DSqrMatrix& operator-=(const DSymmMatrix& mt)
    {
        Sub(mt);
        return *this;
    };

    DSqrMatrix& operator+=(const DUppMatrix& mt)
    {
        Add(mt);
        return *this;
    };

    DSqrMatrix& operator-=(const DUppMatrix& mt)
    {
        Sub(mt);
        return *this;
    };

    DSqrMatrix& operator+=(const DLowMatrix& mt)
    {
        Add(mt);
        return *this;
    };

    DSqrMatrix& operator-=(const DLowMatrix& mt)
    {
        Sub(mt);
        return *this;
    };

    DSqrMatrix& operator+=(const DDiagMatrix& mt)
    {
        Add(mt);
        return *this;
    };

    DSqrMatrix& operator-=(const DDiagMatrix& mt)
    {
        Sub(mt);
        return *this;
    };

    // Операции умножения
    DSqrMatrix& MulRight(const CDSqrMatrix& mt)
    {
        MulRightRect(DSqrMatrix(mt));
        return *this;
    };

    inline DSqrMatrix& operator*=(const CDSqrMatrix& mt)
    {
        return MulRight(mt);
    };

    DSqrMatrix& MulRight(const DSqrMatrix& mt)
    {
        MulRightRect(mt);
        return *this;
    };

    inline DSqrMatrix& operator*=(const DSqrMatrix& mt)
    {
        return MulRight(mt);
    };

    DSqrMatrix& MulLeft(const DSqrMatrix& mt)
    {
        MulLeftRect(mt);
        return *this;
    };

    DSqrMatrix& MulLeft(const CDSqrMatrix& mt)
    {
        MulLeftRect(DSqrMatrix(mt));
        return *this;
    };

    DSqrMatrix& MulRight(const DUppMatrix& mt)
    {
        DRectMem::MulRight(mt);
        return *this;
    };

    DSqrMatrix& MulLeft(const DUppMatrix& mt)
    {
        DRectMem::MulLeft(mt);
        return *this;
    };

    DSqrMatrix& operator*=(const DUppMatrix& mt)
    {
        return MulRight(mt);
    };

    DSqrMatrix& MulRight(const DLowMatrix& mt)
    {
        DRectMem::MulRight(mt);
        return *this;
    };

    DSqrMatrix& MulLeft(const DLowMatrix& mt)
    {
        DRectMem::MulLeft(mt);
        return *this;
    };

    DSqrMatrix& operator*=(const DLowMatrix& mt)
    {
        return MulRight(mt);
    };

    DSqrMatrix& MulRight(const DDiagMatrix& mt)
    {
        DRectMem::MulRight(mt);
        return *this;
    };

    DSqrMatrix& MulLeft(const DDiagMatrix& mt)
    {
        DRectMem::MulLeft(mt);
        return *this;
    };

    DSqrMatrix& operator*=(const DDiagMatrix& mt)
    {
        return MulRight(mt);
    };

    // Транспонирование
    void Transpose();
    // Обращение
    bool Inverse();

    // Реализация запроса числа строк
    virtual unsigned Rows() const
    {
        return NRows;
    };
    // Реализация запроса числа столбцов
    virtual unsigned Cols() const
    {
        return NCols;
    };
    // Реализация получения элемента
    virtual double Val(unsigned r, unsigned c) const
    {
        TAssert(r < NRows && c < NCols);
        TAssert(Matrix != NULL);
        return Matrix[r][c];
    };
    // Реализация запроса значений элементов в виде массива
    virtual void GetArray(double* array) const;
    // Реализация запроса значений строки
    virtual void GetLine(unsigned index, double* line) const
    {
        TAssert(index < NRows);
        TAssert(Matrix != NULL);
        TAssert(line != NULL);
        memcpy(line, Matrix[index], sizeof(double) * NCols);
    }

    // Реализация запроса значений столбца
    virtual void GetCol(unsigned index, double* col) const;
};

///////////////////////////////////////////////////////////////////////////////
// Класс, реализующий симметтичную матрицу
///////////////////////////////////////////////////////////////////////////////
class DSymmMatrix : public CDSymmMatrix, public DTriagMem
{
    friend double TRecmtSymmRecmt(const DVector& A, const DSymmMatrix& B);

public:
    // Конструкторы
    DSymmMatrix()
    {
    };

    DSymmMatrix(unsigned dimension)
    {
        Init(dimension);
    };

    DSymmMatrix(const DSymmMatrix& mt) : CDSymmMatrix(), DTriagMem()
    {
        Init(mt);
    };

    DSymmMatrix(const CDSymmMatrix& mt)
    {
        Init(mt);
    };

    DSymmMatrix(const DDiagMatrix& mt)
    {
        Init(mt);
    };

    virtual ~DSymmMatrix(void)
    {
    };

    // Получение значения элемента
    virtual double operator()(unsigned r, unsigned c) const
    {
        return Val(r, c);
    };
    // Доступ к элементу матрицы
    double& operator()(unsigned r, unsigned c)
    {
        TAssert(r < Dimension && c < Dimension);
        TAssert(Buffer != NULL);
        unsigned ind = (c >= r) ? Index(r, c) : Index(c, r);
        return Buffer[ind];
    };

    // Операторы присваивания
    DSymmMatrix& operator=(const DSymmMatrix& mt)
    {
        Init(mt);
        return *this;
    };

    DSymmMatrix& operator=(const CDSymmMatrix& mt)
    {
        Init(mt);
        return *this;
    };

    DSymmMatrix& operator=(const DDiagMatrix& mt)
    {
        Init(mt);
        return *this;
    };

    // Умножение на число
    DSymmMatrix& operator*=(double mltp)
    {
        Mult(mltp);
        return *this;
    };
    // Инициализация
    void Init(unsigned dim)
    {
        DTriagMem::Init(dim);
    };
    // Установка единого значения для всех элементов
    void Init(double value)
    {
        DTriagMem::Init(value);
    };

    // Операции сложения и вычитания
    DSymmMatrix& operator+=(const CDSymmMatrix& mt);
    DSymmMatrix& operator-=(const CDSymmMatrix& mt);
    DSymmMatrix& operator+=(const DSymmMatrix& mt);
    DSymmMatrix& operator-=(const DSymmMatrix& mt);
    DSymmMatrix& operator+=(const DDiagMatrix& mt);
    DSymmMatrix& operator-=(const DDiagMatrix& mt);

    // Обращение
    bool Inverse();

    // Тройное умножение матриц
    DSymmMatrix& RectSymmTRecmt(const DRectMem& A, const DSymmMatrix& B);
    DSymmMatrix& TRectSymmRecmt(const DRectMem& A, const DSymmMatrix& B);
    DSymmMatrix& RectSymmTRecmt(const CDSqrMatrix& A, const DSymmMatrix& B);
    DSymmMatrix& TRectSymmRecmt(const CDSqrMatrix& A, const DSymmMatrix& B);

    // Конструирование матрицы путем умножения на траспонированную 
    DSymmMatrix& RectTRecmt(const DRectMem& A);

    // Разложение матрицы на верхнюю и диагональную
    virtual bool Expand(DUppMatrix& s, DDiagMatrix& d) const;

    void JacobiExpand(DDiagMatrix& S, DMatrix& R);

    // Тройное умножение
    friend double TRectSymmRect(const DLineMem& A, const DSymmMatrix& B);

    // Реализация запроса числа строк
    virtual unsigned Rows() const
    {
        return Dimension;
    };
    // Реализация запроса числа столбцов
    virtual unsigned Cols() const
    {
        return Dimension;
    };
    // Реализация получения элемента
    virtual double Val(unsigned r, unsigned c) const
    {
        TAssert(r < Dimension && c < Dimension);
        TAssert(Buffer != NULL);
        unsigned ind = (c >= r) ? Index(r, c) : Index(c, r);
        return Buffer[ind];
    };
    // Реализация запроса значений элементов в виде массива
    virtual void GetArray(double* array) const;
    // Реализация запроса значений строки
    virtual void GetLine(unsigned index, double* line) const;
    // Реализация запроса значений столбца
    virtual void GetCol(unsigned index, double* col) const
    {
        GetLine(index, col);
    };

protected:
    // Инициализация
    void Init(const CDSymmMatrix& mt);
    void Init(const DDiagMatrix& mt);
    void Init(const DSymmMatrix& mt);
};

///////////////////////////////////////////////////////////////////////////////
// Класс, реализующий вектор-столбец (как частный случай матрицы)
///////////////////////////////////////////////////////////////////////////////
class DVector : public CDMatrix, public DLineMem
{
    friend bool TDMSolution(const DVector& b, const DDiagMatrix& a, DVector& x);
    friend bool TDMSolution(const DVector& b, const DUppMatrix& a, DVector& x);
    friend bool TDMSolution(const DVector& b, const CDSymmMatrix& a, DVector& x);
    friend bool TDMSolution(const DVector& b, const DLowMatrix& a, DVector& x);
    friend double operator*(const DVector& a, const DVector& b);
    friend double TRecmtSymmRecmt(const DVector& A, const DSymmMatrix& B);

public:

    // Конструкторы
    DVector()
    {
    };

    DVector(unsigned dim)
    {
        Init(dim);
    }

    DVector(const DVector& vec) : CDMatrix(), DLineMem()
    {
        DLineMem::Init(vec.Vector, vec.Dimension);
    };

    // Деструктор
    virtual ~DVector()
    {
    };

    // Получение значения элемента
    virtual double operator()(unsigned r, unsigned c) const
    {
        return Val(r, c);
    };

    double operator[](unsigned index) const
    {
        TAssert(index < Dimension);
        return Vector[index];
    };

    double operator()(unsigned index) const
    {
        TAssert(index < Dimension);
        return Vector[index];
    };

    unsigned Dim(void) const
    {
        return Dimension;
    };


    // Доступ к элементу матрицы
    double& operator[](unsigned index)
    {
        TAssert(index < Dimension);
        TAssert(Vector != NULL);
        return Vector[index];
    };

    double& operator()(unsigned index)
    {
        TAssert(index < Dimension);
        TAssert(Vector != NULL);
        return Vector[index];
    };

    // Инициализация
    void Init(unsigned dim)
    {
        DLineMem::Init(dim);
    };
    // Установка единого значения для всех элементов
    void Init(double value)
    {
        DLineMem::Init(value);
    };

    // Умножение на число
    DVector& operator*=(double mltp)
    {
        Mult(mltp);
        return *this;
    };

    // Операторы присваивания
    DVector& operator=(const CDMatrix& mt);

    DVector& operator=(const DVector& vec)
    {
        DLineMem::Init(vec.Vector, vec.Dimension);
        return *this;
    };

    // Длина вектора
    double Abs(void) const;

    // Операции сложения и вычитания
    DVector& operator+=(const DVector& mt)
    {
        AddVect(mt);
        return *this;
    };

    DVector& operator-=(const DVector& mt)
    {
        SubVect(mt);
        return *this;
    };

    // Операции умножения
    DVector& MulLeft(const CDMatrix& mt)
    {
        MulLeftRect(DMatrix(mt));
        return *this;
    };

    DVector& MulLeft(const DMatrix& mt)
    {
        MulLeftRect(mt);
        return *this;
    };

    DVector& MulLeft(const DSqrMatrix& mt)
    {
        MulLeftRect(mt);
        return *this;
    };

    DVector& MulLeft(const DUppMatrix& mt);
    DVector& MulLeft(const DLowMatrix& mt);
    DVector& MulLeft(const DDiagMatrix& mt);

    // Произведение двух матриц на вектор
    DVector& TRectRectVector(const CDMatrix& a, const CDMatrix& p, const DVector& b);

    friend double TRectSymmRecmt(const DVector& a, const DSymmMatrix& b);

    // Реализация запроса числа строк
    virtual unsigned Rows() const
    {
        return Dimension;
    };
    // Реализация запроса числа столбцов
    virtual unsigned Cols() const
    {
        return 1;
    };
    // Реализация получения элемента
    virtual double Val(unsigned r, unsigned c) const
    {
        TAssert(c == 0 && r < Dimension);
        TAssert(Vector != NULL);
        return Vector[r];
    };
    // Реализация запроса значений элементов в виде массива
    virtual void GetArray(double* array) const;
    // Реализация запроса значений строки
    virtual void GetLine(unsigned index, double* line) const;
    // Реализация запроса значений столбца
    virtual void GetCol(unsigned index, double* col) const;
};

///////////////////////////////////////////////////////////////////////////////
// Класс, реализующий верхнюю треугольную матрицу
///////////////////////////////////////////////////////////////////////////////
class DUppMatrix : public CDMatrix, public DTriagMem
{
    friend class DVector;
    friend class DSymmMatrix;
    friend class DLowMatrix;
    friend class DDiagMatrix;

    friend DSqrMatrix& Mult(DSqrMatrix& dst, const DLowMatrix& src1, const DUppMatrix& src2);
    friend DSqrMatrix& Mult(DSqrMatrix& dst, const DUppMatrix& src1, const DLowMatrix& src2);
    //friend DDiagMatrix &Mult(DDiagMatrix &dst, const DLowMatrix &src1, const DUppMatrix &src2);
    friend bool TDMSolution(const DVector& b, const DUppMatrix& a, DVector& x);

public:
    // Конструкторы
    DUppMatrix()
    {
    };

    DUppMatrix(unsigned dimension)
    {
        Init(dimension);
    };

    DUppMatrix(const DUppMatrix& mt) : CDMatrix(), DTriagMem()
    {
        Init(mt);
    };

    DUppMatrix(const DDiagMatrix& mt)
    {
        Init(mt);
    };
    // Деструктор
    virtual ~DUppMatrix(void)
    {
    };

    // Запрос размерности
    unsigned Dim() const
    {
        return Dimension;
    };

    // Получение значения элемента
    virtual double operator()(unsigned r, unsigned c) const
    {
        return Val(r, c);
    };
    // Доступ к элементу матрицы
    double& operator()(unsigned r, unsigned c)
    {
        TAssert(r < Dimension && c < Dimension);
        TAssert(Buffer != NULL);
        TAssert(c >= r);
        return Buffer[Index(r, c)];
    };

    // Операторы присваивания
    DUppMatrix& operator=(const DUppMatrix& mt)
    {
        Init(mt);
        return *this;
    };

    DUppMatrix& operator=(const DDiagMatrix& mt)
    {
        Init(mt);
        return *this;
    };

    // Инициализация
    void Init(unsigned dim)
    {
        DTriagMem::Init(dim);
    };

    void Init(double value)
    {
        DTriagMem::Init(value);
    };

    // Умножение на число
    DUppMatrix& operator*=(double mltp)
    {
        Mult(mltp);
        return *this;
    };

    // Операции сложения и вычитания
    DUppMatrix& operator+=(const DUppMatrix& mt);
    DUppMatrix& operator-=(const DUppMatrix& mt);
    DUppMatrix& operator+=(const DDiagMatrix& mt);
    DUppMatrix& operator-=(const DDiagMatrix& mt);

    // Операции умножения
    DUppMatrix& MulRight(const DUppMatrix& mt);
    DUppMatrix& MulLeft(const DUppMatrix& mt);

    DUppMatrix& operator*=(const DUppMatrix& mt)
    {
        return MulRight(mt);
    };

    DUppMatrix& MulRight(const DDiagMatrix& mt);
    DUppMatrix& MulLeft(const DDiagMatrix& mt);

    DUppMatrix& operator*=(const DDiagMatrix& mt)
    {
        return MulRight(mt);
    };

    // Формирование матрицы путем транспонирования
    DUppMatrix& Transpose(const DLowMatrix& A);

    // Реализация запроса числа строк
    virtual unsigned Rows() const
    {
        return Dimension;
    };
    // Реализация запроса числа столбцов
    virtual unsigned Cols() const
    {
        return Dimension;
    };
    // Реализация получения элемента
    virtual double Val(unsigned r, unsigned c) const
    {
        TAssert(r < Dimension && c < Dimension);
        TAssert(Buffer != NULL);
        return (c >= r) ? Buffer[Index(r, c)] : 0.0;
    };
    // Реализация запроса значений элементов в виде массива
    virtual void GetArray(double* array) const;
    // Реализация запроса значений строки
    virtual void GetLine(unsigned index, double* line) const;
    // Реализация запроса значений столбца
    virtual void GetCol(unsigned index, double* col) const;

protected:
    // Инициализация
    void Init(const DUppMatrix& mt);
    void Init(const DDiagMatrix& mt);
};

///////////////////////////////////////////////////////////////////////////////
// Класс, реализующий верхнюю нижнюю матрицу
///////////////////////////////////////////////////////////////////////////////
class DLowMatrix : public CDMatrix, public DTriagMem
{
    friend class DVector;
    friend class DSymmMatrix;
    friend class DUppMatrix;
    friend class DDiagMatrix;

    friend DSqrMatrix& Mult(DSqrMatrix& dst, const DLowMatrix& src1, const DUppMatrix& src2);
    friend DSqrMatrix& Mult(DSqrMatrix& dst, const DUppMatrix& src1, const DLowMatrix& src2);
    //friend DDiagMatrix &Mult(DDiagMatrix &dst, const DLowMatrix &src1, const DUppMatrix &src2);
    friend bool TDMSolution(const DVector& b, const DLowMatrix& a, DVector& x);

public:
    // Конструкторы
    DLowMatrix()
    {
    };

    DLowMatrix(unsigned dimension)
    {
        Init(dimension);
    };

    DLowMatrix(const DLowMatrix& mt) : CDMatrix(), DTriagMem()
    {
        Init(mt);
    };

    DLowMatrix(const DDiagMatrix& mt)
    {
        Init(mt);
    };
    // Деструктор
    virtual ~DLowMatrix(void)
    {
    };

    // Запрос размерности
    unsigned Dim() const
    {
        return Dimension;
    };

    // Получение значения элемента
    virtual double operator()(unsigned r, unsigned c) const
    {
        return Val(r, c);
    };
    // Доступ к элементу матрицы
    double& operator()(unsigned r, unsigned c)
    {
        TAssert(r < Dimension && c < Dimension);
        TAssert(Buffer != NULL);
        TAssert(c <= r);
        return Buffer[Index(c, r)];
    };

    // Операторы присваивания
    DLowMatrix& operator=(const DLowMatrix& mt)
    {
        Init(mt);
        return *this;
    };

    DLowMatrix& operator=(const DDiagMatrix& mt)
    {
        Init(mt);
        return *this;
    };

    // Умножение на число
    DLowMatrix& operator*=(double mltp)
    {
        Mult(mltp);
        return *this;
    };

    // Инициализация
    void Init(unsigned dim)
    {
        DTriagMem::Init(dim);
    };
    // Установка единого значения для всех элементов
    void Init(double value)
    {
        DTriagMem::Init(value);
    };

    // Операции сложения и вычитания
    DLowMatrix& operator+=(const DLowMatrix& mt);
    DLowMatrix& operator-=(const DLowMatrix& mt);
    DLowMatrix& operator+=(const DDiagMatrix& mt);
    DLowMatrix& operator-=(const DDiagMatrix& mt);

    // Операции умножения
    DLowMatrix& MulRight(const DLowMatrix& mt);
    DLowMatrix& MulLeft(const DLowMatrix& mt);

    DLowMatrix& operator*=(const DLowMatrix& mt)
    {
        return MulRight(mt);
    };

    DLowMatrix& MulRight(const DDiagMatrix& mt);
    DLowMatrix& MulLeft(const DDiagMatrix& mt);

    DLowMatrix& operator*=(const DDiagMatrix& mt)
    {
        return MulRight(mt);
    };

    // Формирование матрицы путем транспонирования
    DLowMatrix& Transpose(const DUppMatrix& A);

    // Реализация запроса числа строк
    virtual unsigned Rows() const
    {
        return Dimension;
    };
    // Реализация запроса числа столбцов
    virtual unsigned Cols() const
    {
        return Dimension;
    };
    // Реализация получения элемента
    virtual double Val(unsigned r, unsigned c) const
    {
        TAssert(r < Dimension && c < Dimension);
        TAssert(Buffer != NULL);
        return (c <= r) ? Buffer[Index(c, r)] : 0.0;
    };
    // Реализация запроса значений элементов в виде массива
    virtual void GetArray(double* array) const;
    // Реализация запроса значений строки
    virtual void GetLine(unsigned index, double* line) const;
    // Реализация запроса значений столбца
    virtual void GetCol(unsigned index, double* col) const;

protected:
    // Инициализация
    void Init(const DLowMatrix& mt);
    void Init(const DDiagMatrix& mt);
};

///////////////////////////////////////////////////////////////////////////////
// Класс, реализующий диагональную матрицу
///////////////////////////////////////////////////////////////////////////////
class DDiagMatrix : public CDSymmMatrix, public DLineMem
{
    friend class DVector;
    friend class DSymmMatrix;

    //friend DDiagMatrix &Mult(DDiagMatrix &dst, const DLowMatrix &src1, const DUppMatrix &src2);
    friend bool TDMSolution(const DVector& B, const DDiagMatrix& A, DVector& X);

public:

    // Конструкторы
    DDiagMatrix()
    {
    };

    DDiagMatrix(unsigned dimension)
    {
        Init(dimension);
    };

    DDiagMatrix(const DDiagMatrix& mt) : CDSymmMatrix(), DLineMem()
    {
        Init(mt);
    };
    // Деструктор
    virtual ~DDiagMatrix(void)
    {
    };

    // Запрос размерности
    unsigned Dim() const
    {
        return Dimension;
    };

    // Доступ по индексу
    double operator[](unsigned index) const
    {
        TAssert(index < Dimension);
        TAssert(Vector != NULL);
        return Vector[index];
    };
    // Доступ по индексу
    double operator()(unsigned index) const
    {
        TAssert(index < Dimension);
        TAssert(Vector != NULL);
        return Vector[index];
    };

    // Получение значения элемента
    virtual double operator()(unsigned r, unsigned c) const
    {
        return Val(r, c);
    };
    // Доступ к элементу матрицы
    double& operator()(unsigned r, unsigned c)
    {
        TAssert(r < Dimension && c < Dimension);
        TAssert(Vector != NULL);
        TAssert(r == c);
        return Vector[r];
    };

    double& operator[](unsigned index)
    {
        TAssert(index < Dimension);
        TAssert(Vector != NULL);
        return Vector[index];
    };

    double& operator()(unsigned index)
    {
        TAssert(index < Dimension);
        TAssert(Vector != NULL);
        return Vector[index];
    };

    // Оператор присваивания
    DDiagMatrix& operator=(const DDiagMatrix& mt)
    {
        Init(mt);
        return *this;
    };

    // Умножение на число
    DDiagMatrix& operator*=(double mltp)
    {
        Mult(mltp);
        return *this;
    };

    // Инициализация
    void Init(unsigned dim)
    {
        DLineMem::Init(dim);
    };
    // Установка единого значения для всех элементов
    void Init(double value)
    {
        DLineMem::Init(value);
    };

    // Операции сложения и вычитания
    DDiagMatrix& operator+=(const DDiagMatrix& mt);
    DDiagMatrix& operator-=(const DDiagMatrix& mt);

    // Операции умножения
    DDiagMatrix& MulRight(const DDiagMatrix& mt);
    DDiagMatrix& MulLeft(const DDiagMatrix& mt);

    DDiagMatrix& operator*=(const DDiagMatrix& mt)
    {
        return MulRight(mt);
    };

    // Инвертирование (формирование обратной метрицы)
    bool Inverse();

    // Разложение матрицы на верхнюю и диагональную
    virtual bool Expand(DUppMatrix& s, DDiagMatrix& d) const;

    // Реализация запроса числа строк
    virtual unsigned Rows() const
    {
        return Dimension;
    };
    // Реализация запроса числа столбцов
    virtual unsigned Cols() const
    {
        return Dimension;
    };
    // Реализация получения элемента
    virtual double Val(unsigned r, unsigned c) const
    {
        TAssert(r < Dimension && c < Dimension);
        TAssert(Vector != NULL);
        return (r == c) ? Vector[r] : 0.0;
    };
    // Реализация запроса значений элементов в виде массива
    virtual void GetArray(double* array) const;
    // Реализация запроса значений строки
    virtual void GetLine(unsigned index, double* line) const;
    // Реализация запроса значений столбца
    virtual void GetCol(unsigned index, double* col) const;

protected:
    // Инициализация
    void Init(const DDiagMatrix& mt);
};

///////////////////////////////////////////////////////////////////////////////
// Функции сложения и вычитания
///////////////////////////////////////////////////////////////////////////////
template <class SRC1, class SRC2>
DMatrix& Add(DMatrix& dst, const SRC1& src1, const SRC2& src2)
{
    dst = src1;
    dst += src2;
    return dst;
};

template <class SRC1, class SRC2>
DMatrix& Sub(DMatrix& dst, const SRC1& src1, const SRC2& src2)
{
    dst = src1;
    dst -= src2;
    return dst;
};

template <class SRC1, class SRC2>
DSqrMatrix& Add(DSqrMatrix& dst, const SRC1& src1, const SRC2& src2)
{
    dst = src1;
    dst += src2;
    return dst;
};

template <class SRC1, class SRC2>
DSqrMatrix& Sub(DSqrMatrix& dst, const SRC1& src1, const SRC2& src2)
{
    dst = src1;
    dst -= src2;
    return dst;
};

inline DVector& Add(DVector& dst, const DVector& src1, const DVector& src2)
{
    dst = src1;
    dst += src2;
    return dst;
};

inline DVector& Sub(DVector& dst, const DVector& src1, const DVector& src2)
{
    dst = src1;
    dst -= src2;
    return dst;
};

inline DSymmMatrix& Add(DSymmMatrix& dst, const CDSymmMatrix& src1, const CDSymmMatrix& src2)
{
    dst = src1;
    dst += src2;
    return dst;
};

inline DSymmMatrix& Sub(DSymmMatrix& dst, const CDSymmMatrix& src1, const CDSymmMatrix& src2)
{
    dst = src1;
    dst -= src2;
    return dst;
};

inline DUppMatrix& Add(DUppMatrix& dst, const DUppMatrix& src1, const DUppMatrix& src2)
{
    dst = src1;
    dst += src2;
    return dst;
};

inline DUppMatrix& Sub(DUppMatrix& dst, const DUppMatrix& src1, const DUppMatrix& src2)
{
    dst = src1;
    dst -= src2;
    return dst;
};

inline DLowMatrix& Add(DLowMatrix& dst, const DLowMatrix& src1, const DLowMatrix& src2)
{
    dst = src1;
    dst += src2;
    return dst;
};

inline DLowMatrix& Sub(DLowMatrix& dst, const DLowMatrix& src1, const DLowMatrix& src2)
{
    dst = src1;
    dst -= src2;
    return dst;
};

inline DDiagMatrix& Add(DDiagMatrix& dst, const DDiagMatrix& src1, const DDiagMatrix& src2)
{
    dst = src1;
    dst += src2;
    return dst;
};

inline DDiagMatrix& Sub(DDiagMatrix& dst, const DDiagMatrix& src1, const DDiagMatrix& src2)
{
    dst = src1;
    dst -= src2;
    return dst;
};
///////////////////////////////////////////////////////////////////////////////
// Функции умножения
///////////////////////////////////////////////////////////////////////////////
inline void MultRect(DRectMem& dst, const DRectMem& src1, const DRectMem& src2)
{
    dst.Mult(src1, src2);
};

inline void MultRect(DRectMem& dst, const DRectMem& src1, const DUppMatrix& src2)
{
    dst.Mult(src1, src2);
};

inline void MultRect(DRectMem& dst, const DUppMatrix& src1, const DRectMem& src2)
{
    dst.Mult(src1, src2);
};

inline void MultRect(DRectMem& dst, const DRectMem& src1, const DLowMatrix& src2)
{
    dst.Mult(src1, src2);
};

inline void MultRect(DRectMem& dst, const DLowMatrix& src1, const DRectMem& src2)
{
    dst.Mult(src1, src2);
};

inline void MultRect(DRectMem& dst, const DRectMem& src1, const DDiagMatrix& src2)
{
    dst.Mult(src1, src2);
};

inline void MultRect(DRectMem& dst, const DDiagMatrix& src1, const DRectMem& src2)
{
    dst.Mult(src1, src2);
};

inline DMatrix& Mult(DMatrix& dst, const CDMatrix& src1, const CDMatrix& src2)
{
    MultRect(dst, DMatrix(src1), DMatrix(src2));
    return dst;
};

inline DMatrix& Mult(DMatrix& dst, const CDMatrix& src1, const DMatrix& src2)
{
    MultRect(dst, DMatrix(src1), src2);
    return dst;
};

inline DMatrix& Mult(DMatrix& dst, const DMatrix& src1, const CDMatrix& src2)
{
    MultRect(dst, src1, DMatrix(src2));
    return dst;
};

inline DMatrix& Mult(DMatrix& dst, const DMatrix& src1, const DMatrix& src2)
{
    MultRect(dst, src1, src2);
    return dst;
};

inline DMatrix& Mult(DMatrix& dst, const DMatrix& src1, const DUppMatrix& src2)
{
    MultRect(dst, src1, src2);
    return dst;
};

inline DMatrix& Mult(DMatrix& dst, const CDMatrix& src1, const DUppMatrix& src2)
{
    MultRect(dst, DMatrix(src1), src2);
    return dst;
};

inline DMatrix& Mult(DMatrix& dst, const DUppMatrix& src1, const DMatrix& src2)
{
    MultRect(dst, src1, src2);
    return dst;
};

inline DMatrix& Mult(DMatrix& dst, const DUppMatrix& src1, const CDMatrix& src2)
{
    MultRect(dst, src1, DMatrix(src2));
    return dst;
};

inline DMatrix& Mult(DMatrix& dst, const DMatrix& src1, const DLowMatrix& src2)
{
    MultRect(dst, src1, src2);
    return dst;
};

inline DMatrix& Mult(DMatrix& dst, const CDMatrix& src1, const DLowMatrix& src2)
{
    MultRect(dst, DMatrix(src1), src2);
    return dst;
};

inline DMatrix& Mult(DMatrix& dst, const DLowMatrix& src1, const DMatrix& src2)
{
    MultRect(dst, src1, src2);
    return dst;
};

inline DMatrix& Mult(DMatrix& dst, const DLowMatrix& src1, const CDMatrix& src2)
{
    MultRect(dst, src1, DMatrix(src2));
    return dst;
};

inline DMatrix& Mult(DMatrix& dst, const DMatrix& src1, const DDiagMatrix& src2)
{
    MultRect(dst, src1, src2);
    return dst;
};

inline DMatrix& Mult(DMatrix& dst, const CDMatrix& src1, const DDiagMatrix& src2)
{
    MultRect(dst, DMatrix(src1), src2);
    return dst;
};

inline DMatrix& Mult(DMatrix& dst, const DDiagMatrix& src1, const DMatrix& src2)
{
    MultRect(dst, src1, src2);
    return dst;
};

inline DMatrix& Mult(DMatrix& dst, const DDiagMatrix& src1, const CDMatrix& src2)
{
    MultRect(dst, src1, DMatrix(src2));
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const CDSqrMatrix& src1, const CDSqrMatrix& src2)
{
    MultRect(dst, DSqrMatrix(src1), DSqrMatrix(src2));
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const CDSqrMatrix& src1, const DSqrMatrix& src2)
{
    MultRect(dst, DSqrMatrix(src1), src2);
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const DSqrMatrix& src1, const CDSqrMatrix& src2)
{
    MultRect(dst, src1, DSqrMatrix(src2));
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const DSqrMatrix& src1, const DSqrMatrix& src2)
{
    MultRect(dst, src1, src2);
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const DSqrMatrix& src1, const DUppMatrix& src2)
{
    MultRect(dst, src1, src2);
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const CDSqrMatrix& src1, const DUppMatrix& src2)
{
    MultRect(dst, DSqrMatrix(src1), src2);
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const DUppMatrix& src1, const DSqrMatrix& src2)
{
    MultRect(dst, src1, src2);
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const DUppMatrix& src1, const CDSqrMatrix& src2)
{
    MultRect(dst, src1, DSqrMatrix(src2));
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const DSqrMatrix& src1, const DLowMatrix& src2)
{
    MultRect(dst, src1, src2);
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const CDSqrMatrix& src1, const DLowMatrix& src2)
{
    MultRect(dst, DSqrMatrix(src1), src2);
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const DLowMatrix& src1, const DSqrMatrix& src2)
{
    MultRect(dst, src1, src2);
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const DLowMatrix& src1, const CDSqrMatrix& src2)
{
    MultRect(dst, src1, DSqrMatrix(src2));
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const DSqrMatrix& src1, const DDiagMatrix& src2)
{
    MultRect(dst, src1, src2);
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const CDSqrMatrix& src1, const DDiagMatrix& src2)
{
    MultRect(dst, DSqrMatrix(src1), src2);
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const DDiagMatrix& src1, const DSqrMatrix& src2)
{
    MultRect(dst, src1, src2);
    return dst;
};

inline DSqrMatrix& Mult(DSqrMatrix& dst, const DDiagMatrix& src1, const CDSqrMatrix& src2)
{
    MultRect(dst, src1, DSqrMatrix(src2));
    return dst;
};

DSqrMatrix& Mult(DSqrMatrix& dst, const DLowMatrix& src1, const DUppMatrix& src2);
DSqrMatrix& Mult(DSqrMatrix& dst, const DUppMatrix& src1, const DLowMatrix& src2);

inline DVector& Mult(DVector& dst, const CDMatrix& src1, const DVector& src2)
{
    dst.MulRect(DMatrix(src1), src2);
    return dst;
};

inline DVector& Mult(DVector& dst, const DMatrix& src1, const DVector& src2)
{
    dst.MulRect(src1, src2);
    return dst;
};

inline DVector& Mult(DVector& dst, const DSqrMatrix& src1, const DVector& src2)
{
    dst.MulRect(src1, src2);
    return dst;
};

inline DUppMatrix& Mult(DUppMatrix& dst, const DUppMatrix& src1, const DUppMatrix& src2)
{
    dst = src1;
    dst *= src2;
    return dst;
};

inline DUppMatrix& Mult(DUppMatrix& dst, const DUppMatrix& src1, const DDiagMatrix& src2)
{
    dst = src1;
    dst *= src2;
    return dst;
};

inline DUppMatrix& Mult(DUppMatrix& dst, const DDiagMatrix& src1, const DUppMatrix& src2)
{
    dst = src1;
    dst *= src2;
    return dst;
};

inline DLowMatrix& Mult(DLowMatrix& dst, const DLowMatrix& src1, const DLowMatrix& src2)
{
    dst = src1;
    dst *= src2;
    return dst;
};

inline DLowMatrix& Mult(DLowMatrix& dst, const DLowMatrix& src1, const DDiagMatrix& src2)
{
    dst = src1;
    dst *= src2;
    return dst;
};

inline DLowMatrix& Mult(DLowMatrix& dst, const DDiagMatrix& src1, const DLowMatrix& src2)
{
    dst = src1;
    dst *= src2;
    return dst;
};

inline DDiagMatrix& Mult(DDiagMatrix& dst, const DDiagMatrix& src1, const DDiagMatrix& src2)
{
    dst = src1;
    dst *= src2;
    return dst;
};
///////////////////////////////////////////////////////////////////////////////
// Функции решения линейных систем
///////////////////////////////////////////////////////////////////////////////
bool TDMSolution(const DVector& B, const CDSqrMatrix& A, DVector& X);
bool TDMSolution(const DVector& B, const CDSymmMatrix& A, DVector& X);
bool TDMSolution(const DVector& B, const DDiagMatrix& A, DVector& X);
bool TDMSolution(const DVector& B, const DUppMatrix& A, DVector& X);
bool TDMSolution(const DVector& B, const DLowMatrix& A, DVector& X);
bool TDMSolutionJ(const DVector& B, DSymmMatrix& A, DVector& X);
///////////////////////////////////////////////////////////////////////////////
// Скалярное умножение векторов
///////////////////////////////////////////////////////////////////////////////
double operator*(const DVector& a, const DVector& b);
///////////////////////////////////////////////////////////////////////////////
// Тройное умножение
///////////////////////////////////////////////////////////////////////////////
double TRecmtSymmRecmt(const DVector& A, const DSymmMatrix& B);
///////////////////////////////////////////////////////////////////////////////
// Реализация методов класса DMatrix
///////////////////////////////////////////////////////////////////////////////
inline DMatrix::DMatrix(const DSqrMatrix& mt)
{
    InitRect(mt);
}

inline DMatrix& DMatrix::operator=(const DSqrMatrix& mt)
{
    InitRect(mt);
    return *this;
};

inline DMatrix& DMatrix::operator+=(const DSqrMatrix& mt)
{
    AddRect(mt);
    return *this;
};

inline DMatrix& DMatrix::operator-=(const DSqrMatrix& mt)
{
    SubRect(mt);
    return *this;
};

inline DMatrix& DMatrix::operator+=(const DUppMatrix& mt)
{
    Add(mt);
    return *this;
};

inline DMatrix& DMatrix::operator-=(const DUppMatrix& mt)
{
    Sub(mt);
    return *this;
};

inline DMatrix& DMatrix::operator+=(const DLowMatrix& mt)
{
    Add(mt);
    return *this;
};

inline DMatrix& DMatrix::operator-=(const DLowMatrix& mt)
{
    Sub(mt);
    return *this;
};

inline DMatrix& DMatrix::operator+=(const DDiagMatrix& mt)
{
    Add(mt);
    return *this;
};

inline DMatrix& DMatrix::operator-=(const DDiagMatrix& mt)
{
    Sub(mt);
    return *this;
};

inline DMatrix& DMatrix::MulRight(const DSqrMatrix& mt)
{
    MulRightRect(mt);
    return *this;
};

inline DMatrix& DMatrix::MulLeft(const DSqrMatrix& mt)
{
    MulLeftRect(mt);
    return *this;
};
///////////////////////////////////////////////////////////////////////////////
// Константы, используемая для контроля нуля
///////////////////////////////////////////////////////////////////////////////
#define ZERO_ACCURACY (1.0e-100)
///////////////////////////////////////////////////////////////////////////////
#endif
