///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация системы дифференциальных уравнений, задающих движение КО
///////////////////////////////////////////////////////////////////////////////
#include "ODESystem.h"
#include "GeoCoords.h"
#include "IAstroBase.h"
#include "EarthGravity.h"

///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
//-----------------------------------------------------------------------------
ODEPhaseSystem::ODEPhaseSystem()
{
    MMode = NULL;
    KOModel = NULL;
    Ad = NULL;
    Pm = NULL;
    Pm1 = NULL;
    Ron = NULL;
    Nm1 = NULL;
    EarthGravity = NULL;
    m_precItrf = false;
};
//-----------------------------------------------------------------------------
// Деструктор
//-----------------------------------------------------------------------------
ODEPhaseSystem::~ODEPhaseSystem()
{
    if (Pm != NULL) {
        delete[] Pm;
        delete[] Pm1;
        delete[] Ron;
        delete[] Nm1;
    }
}

//-----------------------------------------------------------------------------
// Инициализация
//-----------------------------------------------------------------------------
void ODEPhaseSystem::Init(const TimeJD& t0, const MotionMode& m, IKOProp& ko, AccAdapter& a)
{
    T0 = t0;
    MMode = &m;
    KOModel = &ko;
    Ad = &a;

    Ad->Init(T0, *MMode);

    EarthGravity = GetAstroBase()->GetPlanet(plIdEarth)->GetGravityModel();

    Ng = MMode->GrN;
    if (Ng > (int)EarthGravity->N())
        Ng = EarthGravity->N();

    Mg = MMode->GrM;
    if (Mg > (int)EarthGravity->M())
        Mg = EarthGravity->M();

    int n = Ng > Mg ? Ng : Mg;
    if (n <= 0)
        return;

    if (Pm != NULL) {
        delete[] Pm;
        delete[] Pm1;
        delete[] Ron;
        delete[] Nm1;
    }

    Ron = new double[n + 1];
    Pm = new double[n + 3];
    Pm1 = new double[n + 3];
    Nm1 = new double[n];

    int i;

    for (i = 0; i < n; i++)
        Nm1[i] = 1.0 / (i + 2);
}

//-----------------------------------------------------------------------------
// Приведение аргумента интегрирования (тыс. секунд) к дате
//-----------------------------------------------------------------------------
void ODEPhaseSystem::SetTime(double x, TimeJD& t) const
{
    t = T0;
    t += x / 86.4;
}

//-----------------------------------------------------------------------------
// Вычисление разности между начальным и указанным временем
//-----------------------------------------------------------------------------
double ODEPhaseSystem::GetDiff(const TimeJD& t) const
{
    double x = (t - T0).GetDouble() * 86.4;
    return x;
}

//-----------------------------------------------------------------------------
// Установка адаптера ускорений
//-----------------------------------------------------------------------------
void ODEPhaseSystem::SetAdapter(double x, const double* ph, bool precItrf)
{
    TimeJD t;
    SetTime(x, t);
    Vect6 j2000;

    j2000.x = ph[0];
    j2000.y = ph[1];
    j2000.z = ph[2];
    j2000.vx = ph[3];
    j2000.vy = ph[4];
    j2000.vz = ph[5];

    Ad->Set(t, j2000, precItrf);
}

//-----------------------------------------------------------------------------
// Вычисление правых частей
// Метод возвращает false, если достигнута критичная точка,
// после которой дальнейшее решение невозможно
//-----------------------------------------------------------------------------
bool ODEPhaseSystem::GetRP(double t, const double* ph, double* f)
{
    // Взвод адаптера 
    SetAdapter(t, ph, m_precItrf);

    // Проверка на критическую высоту
    if (MMode->IsAtm) {
        GeoCoords g;
        g.SetXYZ(Ad->RGr);
        if (g.H < MMode->HCrit)
            return false;
    }

    f[0] = ph[3];
    f[1] = ph[4];
    f[2] = ph[5];
    f[3] = 0;
    f[4] = 0;
    f[5] = 0;

    // Расчет гравитационного ускорения от Земли
    if (MMode->IsPlanetSet(plIdEarth)) {
        // Расчет нецентрального поля
        if (Ng > 0 || Mg > 0) {
            Vect3 gacc;
            EarthNonCentralAcc(gacc, Ad->RGr);
            f[3] = gacc.x;
            f[4] = gacc.y;
            f[5] = gacc.z;
        }
        // Расчет центрального поля Земли
        double r3 = Ad->RJ2000.Len2();
        r3 *= sqrt(r3);
        r3 = -GetAstroBase()->GetEarthMu() / r3;
        f[3] += Ad->RJ2000.x * r3;
        f[4] += Ad->RJ2000.y * r3;
        f[5] += Ad->RJ2000.z * r3;
    }

    // Расчет гравитацилнного ускорения от планет
    int i;
    for (i = 0; i < MMode->GetNPlanet(); i++) {
        TPlanetId pid = MMode->GetPlanet(i);
        if (pid == plIdEarth)
            continue;
        Vect3 pacc;
        if (pid == plIdSun)
            PlanetGravityAcc(GetAstroBase()->GetPlanet(pid)->GetGravityModel()->mu,
                             Ad->RJ2000, Ad->SunJ2000, pacc);
        else {
            Vect3 pos;
            GetAstroBase()->GetPosGeo(pid, Ad->T_TDB, &pos);
            PlanetGravityAcc(GetAstroBase()->GetPlanet(pid)->GetGravityModel()->mu,
                             Ad->RJ2000, pos, pacc);
        }
        f[3] += pacc.x;
        f[4] += pacc.y;
        f[5] += pacc.z;
    }

    // Расчет ускоренuй, обусловленных свойствами КО
    Vect3 kacc;
    KOModel->GetAcc(*Ad, kacc);
    f[3] += kacc.x;
    f[4] += kacc.y;
    f[5] += kacc.z;

    return true;
}

//-----------------------------------------------------------------------------
// Расчет ускорений от нецентрального поля Земли
//-----------------------------------------------------------------------------
void ODEPhaseSystem::EarthNonCentralAcc(Vect3& acc, const Vect3& pos)
{
    double r = pos.Len();
    double ro = EarthGravity->Re / r;
    double xr = pos.x / r;
    double yr = pos.y / r;
    double zr = pos.z / r;

    // Зональные гармоники
    Pm[1] = 1;
    Pm[2] = 3 * zr;
    double p0_2 = zr;
    double p0_1 = 1.5 * zr * zr - 0.5;
    double ro1 = ro;
    Ron[1] = ro;

    double x0 = 0;
    double x2 = 0;
    double p0 = p0_1;
    double cn0, t1;

    const double* const* CC = EarthGravity->get_cc();
    const double* const* DD = EarthGravity->get_dd();

    int n;
    for (n = 2; n <= Ng; n++) {
        ro1 *= ro;
        Ron[n] = ro1;
        cn0 = EarthGravity->c(n, 0) * ro1;
        //cn0 = CC[0][n] * ro1;
        x0 += cn0 * (n + 1) * p0;
        x2 += cn0 * Pm[n];
        if (n == Ng)
            break;
        t1 = (2 * n + 1) * zr;
        Pm[n + 1] = (t1 * Pm[n] - (n + 1) * Pm[n - 1]) / n;
        p0 = (t1 * p0_1 - n * p0_2) / (n + 1);
        p0_2 = p0_1;
        p0_1 = p0;
    };

    // Секториальные и тессеральные гармоники
    double c1 = xr;
    double s1 = yr;

    double z0 = x0;
    double z1 = 0.0;
    double z2 = x2;
    double z11 = 0;
    double z12 = 0;

    int m;
    double cm = 0.0, sm = 0.0, cc;
    for (m = 1; m <= Mg; m++) {
        double x01 = 0;
        double x11 = 0;
        double x21 = 0;
        double x02 = 0;
        double x12 = 0;
        double x22 = 0;

        p0_2 = Pm[m];
        Pm1[m + 1] = (2 * m + 1) * p0_2;
        Pm1[m + 2] = (2 * m + 3) * zr * Pm1[m + 1];

        n = m < 2 ? 2 : m;
        const double* pn = Nm1;

        for (; n <= Mg; n++) {
            t1 = Ron[n];

            double cnm = EarthGravity->c(n, m) * t1;
            //double cnm = CC[m][n] * t1;

            double dnm = EarthGravity->d(n, m) * t1;
            //double dnm = DD[m][n] * t1;

            t1 = Pm[n];

            double t2 = cnm * t1;
            x11 += t2;
            t1 *= dnm;
            x12 += t1;
            x01 += t2 * (n + 1);
            x02 += t1 * (n + 1);

            if (n > m + 2) {
                Pm1[n] = ((2 * n - 1) * zr * Pm1[n - 1] - (n + m) * Pm1[n - 2]) * (*pn);
                pn++;
            }

            if (n != m) {
                t1 = Pm1[n];
                x21 += cnm * t1;
                x22 += dnm * t1;
            }
        }

        if (m == 1) {
            z11 = x11;
            c1 = xr;
            cm = c1;
            z12 = x12;
            s1 = yr;
            sm = s1;
        }
        else {
            z11 += m * (x11 * cm + x12 * sm);
            z12 += m * (x12 * cm - x11 * sm);
            cc = cm * c1 - sm * s1;
            sm = sm * c1 + cm * s1;
            cm = cc;
        };
        z0 += x01 * cm + x02 * sm;
        z1 += m * (x11 * cm + x12 * sm);
        z2 += x21 * cm + x22 * sm;

        // ???
        double* Pmt = Pm;
        Pm = Pm1;
        Pm1 = Pmt;
    }

    double x01 = z0 + z1 + zr * z2;
    double x11 = EarthGravity->mu / (r * r);

    acc.x = x11 * (z11 - xr * x01);
    acc.y = x11 * (z12 - yr * x01);
    acc.z = x11 * (z2 - zr * x01);

    // Получили ускорения в Гринвиче. Теперь приводим их к J2000

    acc.MulLeft(Ad->CGFCStoJ2000);
}

//-----------------------------------------------------------------------------
// Расчет ускорений от небесных тел
//-----------------------------------------------------------------------------
void ODEPhaseSystem::PlanetGravityAcc(double mu, const Vect3& r, const Vect3& rp,
                                      Vect3& acc) const
{
    double len, len1;
    double x, y, z, x1, y1, z1;

    x1 = rp.x;
    len1 = x1 * x1;
    x = x1 - r.x;
    len = x * x;
    y1 = rp.y;
    len1 += y1 * y1;
    y = y1 - r.y;
    len += y * y;
    z1 = rp.z;
    len1 += z1 * z1;
    z = z1 - r.z;
    len += z * z;

    len1 *= sqrt(len1);
    len1 = mu / len1;
    len *= sqrt(len);
    len = mu / len;

    acc.x = x * len - x1 * len1;
    acc.y = y * len - y1 * len1;
    acc.z = z * len - z1 * len1;
}

void ODEPhaseSystem::SetPrecItrf(bool value)
{
    m_precItrf = value;
}

///////////////////////////////////////////////////////////////////////////////
