///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_ODEPHASERP_H
#define BNOTOOL_ODEPHASERP_H
///////////////////////////////////////////////////////////////////////////////
#include "IODE.h"
#include "MMode.h"
#include "IKOProp.h"
#include "AccAdapter.h"
#include "IModel.h"

///////////////////////////////////////////////////////////////////////////////
// Реализация системы дифференциальных уравнений, задающих движение КО
///////////////////////////////////////////////////////////////////////////////
class ODEPhaseSystem : public IODESystem
{
public:
    // Конструктор по умолчанию
    ODEPhaseSystem();
    // Деструктор
    virtual ~ODEPhaseSystem();

    // Инициализация
    void Init(const TimeJD& t0, const MotionMode& m, IKOProp& ko, AccAdapter& a);
    void SetPrecItrf(bool value);
    // Получение размерности системы
    virtual int GetDimension() const
    {
        return 6;
    };
    // Вычисление правых частей
    // Метод возвращает false, если достигнута критичная точка,
    // после которой дальнейшее решение невозможно
    virtual bool GetRP(double t, const double* ph, double* f);

    bool getPrecItrf() const { return m_precItrf; }
protected:
    // Описание модели движения
    const MotionMode* MMode;
    // Описание модели КО
    const IKOProp* KOModel;
    // Адаптер ускорений
    AccAdapter* Ad;
    // Начальное время, от которого ведется отсчет процедуры
    // интегрирования
    TimeJD T0;

    // Приведение аргумента интегрирования (тыс. секунд) к дате
    void SetTime(double x, TimeJD& t) const;
    // Вычисление разности между начальным и указанным временем
    double GetDiff(const TimeJD& t) const;
    // Установка адаптера ускорений
    void SetAdapter(double x, const double* ph, bool precItrf);

    // Расчет ускорений от нецентрального поля Земли
    void EarthNonCentralAcc(Vect3& acc, const Vect3& pos);
    // Расчет ускорений от небесных тел
    void PlanetGravityAcc(double mu, const Vect3& r, const Vect3& rp,
                          Vect3& acc) const;
    // Модель гравитации Земли
    // (фиксируется при инициализации) 
    const IGravityModel* EarthGravity;
    // Служебные массивы, зависящие от размерности нецентрального
    // гравитационного поля Земли
    double *Pm, *Pm1, *Ron, *Nm1;
    // Фактическая размерность гармоник
    int Ng, Mg;
    // расчет матрицы перехода J2000->GFCS на каждом шаге
    bool m_precItrf;
};

///////////////////////////////////////////////////////////////////////////////
// Реализация системы дифференциальных уравнений, задающих движение KO и 
// матрицу частных производных 
///////////////////////////////////////////////////////////////////////////////
class ODEVarSystem : public ODEPhaseSystem
{
public:
    // Конструктор по умолчанию
    ODEVarSystem();
    // Деструктор
    virtual ~ODEVarSystem()
    {
    };

    // Установка модели КО
    void SetKO(IKOPropVar& ko)
    {
        KOVar = &ko;
    };
    // Инициализация
    void Init(const TimeJD& t0, const MotionMode& m, AccAdapter& a);

    // Получение размерности системы
    virtual int GetDimension() const
    {
        TAssert(KOVar != NULL);
        return 6 + (6 + KOVar->GetNVars()) * 6;
    };
    // Вычисление правых частей
    // Метод возвращает false, если достигнута критичная точка,
    // после которой дальнейшее решение невозможно
    virtual bool GetRP(double t, const double* ph, double* f);

protected:
    // Описание модели КО c уравнениями в вариациях
    IKOPropVar* KOVar;

    // Вычисление вариаций от центрального гравитационного поля Земли
    void VarCentralAccel(TransMtx& m);

    // Вычисление вариаций от центрального гравитационного поля Земли
    void VarNonCentralAccel(TransMtx& m);

    // Вычисление вариаций от центрального гравитационного поля планеты
    void VarPlanetAccel(double mu, Vect3& r, Vect3& r0, TransMtx& m);
};

///////////////////////////////////////////////////////////////////////////////
#endif
