///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация системы дифференциальных уравнений, задающих движение KO и
// матрицу частных производных
///////////////////////////////////////////////////////////////////////////////
#include "ODESystem.h"
#include "IAstroBase.h"
///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
//-----------------------------------------------------------------------------
ODEVarSystem::ODEVarSystem() : ODEPhaseSystem()
{
    KOVar = NULL;
}

//-----------------------------------------------------------------------------
// Инициализация
//-----------------------------------------------------------------------------
void ODEVarSystem::Init(const TimeJD& t0, const MotionMode& m, AccAdapter& a)
{
    ODEPhaseSystem::Init(t0, m, *KOVar, a);
}

//-----------------------------------------------------------------------------
// Вычисление правых частей
// Метод возвращает false, если достигнута критичная точка,
// после которой дальнейшее решение невозможно
//-----------------------------------------------------------------------------
bool ODEVarSystem::GetRP(double t, const double* ph, double* f)
{
    if (!ODEPhaseSystem::GetRP(t, ph, f))
        return false;

    //  if (DOVAR != t)
    //  {
    //	return true;
    //  }

    // dr/dr0 -> rr
    // dv/dr0 -> vr
    // dr/dv0 -> rv
    // dv/dv0 -> vv
    //
    // dr/ds  -> rs
    // dv/ds  -> vs
    // dr/dk  -> rk
    // dv/dk  -> vk
    //
    TransMtx rr, vr, rv, vv;
    TransMtx fr, fv;
    TransMtx m1, m2;
    DMatrix dfdc;
    int i, j;

    //===========================================================
    // Входной вектор ph cодержит минимум 42 значения:
    //===========================================================
    //      x        y        z        vx        vy        vz
    //
    //    dx/dx0    dy/dx0   dz/dx0   dvx/dx0   dvy/dx0   dvz/dx0
    //
    //    dx/dy0    dy/dy0   dz/dy0   dvx/dy0   dvy/dy0   dvz/dy0
    //
    //    dx/dz0    dy/dz0   dz/dz0   dvx/dz0   dvy/dz0   dvz/dz0
    //
    //    dx/dvx0   dy/dvx0  dz/dvx0  dvx/dvx0  dvy/dvx0  dvz/dvx0
    //
    //    dx/dvy0   dy/dvy0  dz/dvy0  dvx/dvy0  dvy/dvy0  dvz/dvy0
    //
    //    dx/dvz0   dy/dvz0  dz/dvz0  dvx/dvz0  dvy/dvz0  dvz/dvz0
    //
    //===========================================================
    // внутренняя, используемая в прогнозе нумерация, следующая:
    //===========================================================
    //       0   1   2   3   4   5
    //       6   7   8   9  10  11
    //      12  13  14  15  16  17
    //      18  19  20  21  22  23
    //      24  25  26  27  28  29
    //      30  31  32  33  34  35
    //      36  37  38  39  40  41
    //=================================================================

    for (i = 0; i < 3; i++) {
        for (j = 0; j < 3; j++) {
            rr(i, j) = ph[6 + 6 * i + j];
            f[6 + 6 * i + j] = vr(i, j) = ph[6 + 6 * i + j + 3];
            rv(i, j) = ph[24 + 6 * i + j];
            f[24 + 6 * i + j] = vv(i, j) = ph[24 + 6 * i + j + 3];
        }
    }

    // Производим обнуление матриц ускорений
    for (i = 0; i < 3; i++)
        for (j = 0; j < 3; j++) {
            fr(i, j) = 0.0;
            fv(i, j) = 0.0;
        }

    // Расчет вариаций от центрального поля Земли
    if (MMode->IsPlanetSet(plIdEarth)) {
        VarCentralAccel(fr);
        // Расчет нецентрального поля
        if (Ng > 0 || Mg > 0) {
            VarNonCentralAccel(m1);
            fr += m1;
        }
    }

    // Расчет вариаций от планет
    // Расчет гравитацилнного ускорения от планет
    for (i = 0; i < MMode->GetNPlanet(); i++) {
        TPlanetId pid = MMode->GetPlanet(i);
        if (pid == plIdEarth)
            continue;
        if (pid == plIdSun)
            VarPlanetAccel(GetAstroBase()->GetPlanet(pid)->GetGravityModel()->mu, Ad->RJ2000, Ad->SunJ2000, m2);
        else {
            Vect3 pos;
            GetAstroBase()->GetPosGeo(pid, Ad->T_TDB, &pos);
            VarPlanetAccel(GetAstroBase()->GetPlanet(pid)->GetGravityModel()->mu, Ad->RJ2000, pos, m2);
        }
        fr += m2;
    }

    KOVar->GetRPVars(*Ad, m1, fv, dfdc);
    fr += m1;

    // Вычисляем матрицу df/dr0
    m1 = rr;
    m1.MulRight(fr);
    m2 = vr;
    m2.MulRight(fv);
    m1 += m2;

    // "Расфасовываем" эту матрицу в правые части
    for (i = 0; i < 3; i++)
        for (j = 0; j < 3; j++)
            f[6 + i * 6 + j + 3] = m1(i, j);

    // Вычисляем матрицу df/dv0
    m1 = rv;
    m1.MulRight(fr);
    m2 = vv;
    m2.MulRight(fv);
    m1 += m2;

    // "Расфасовываем" эту матрицу в правые части
    for (i = 0; i < 3; i++)
        for (j = 0; j < 3; j++)
            f[24 + i * 6 + j + 3] = m1(i, j);

    // Если нужны вариации по доп. параметрам
    if (KOVar->GetNVars() > 0) {
        // Перенос производных от скоростей в правые части
        for (i = 0; i < KOVar->GetNVars(); i++) {
            for (j = 0; j < 3; j++)
                f[42 + 6 * i + j] = ph[42 + 6 * i + j + 3];
        }
        // Вычисление вектора df/dc(i)
        for (i = 0; i < KOVar->GetNVars(); i++) {
            Vect3 v1, v2;
            for (j = 0; j < 3; j++) {
                v1[j] = ph[42 + 6 * i + j];
                v2[j] = ph[42 + 6 * i + j + 3];
            }
            v1.MulRight(fr);
            v2.MulRight(fv);
            v1 += v2;
            for (j = 0; j < 3; j++)
                f[42 + 6 * i + j + 3] = v1[j] + dfdc(i, j);
        }
    }

    return true;
}

//-----------------------------------------------------------------------------
// Вычисление вариаций от центрального гравитационного поля планеты
//-----------------------------------------------------------------------------
void ODEVarSystem::VarPlanetAccel(double mu, Vect3& r, Vect3& r0, TransMtx& m)
{
    double len, len1, coef;
    Vect3 rd = r;
    rd -= r0;
    len = rd.Len();
    len1 = 1.0 / len;
    rd *= len1;
    coef = 3 * mu / (len * len * len);
    m(0, 0) = coef * (rd[0] * rd[0] - 1.0 / 3.0);
    m(0, 1) = coef * (rd[0] * rd[1]);
    m(0, 2) = coef * (rd[0] * rd[2]);
    m(1, 0) = m(0, 1);
    m(1, 1) = coef * (rd[1] * rd[1] - 1.0 / 3.0);
    m(1, 2) = coef * (rd[1] * rd[2]);
    m(2, 0) = m(0, 2);
    m(2, 1) = m(1, 2);
    m(2, 2) = coef * (rd[2] * rd[2] - 1.0 / 3.0);
}

//-----------------------------------------------------------------------------
// Вычисление вариаций от центрального гравитационного поля Земли
//-----------------------------------------------------------------------------
void ODEVarSystem::VarCentralAccel(TransMtx& m)
{
    Vect3 r0;

    VarPlanetAccel(GetAstroBase()->GetEarthMu(), Ad->RJ2000, r0, m);
}

//-----------------------------------------------------------------------------
// Вычисление вариаций от центрального гравитационного поля Земли
//-----------------------------------------------------------------------------
void ODEVarSystem::VarNonCentralAccel(TransMtx& m)
{
    Vect3 dr(1.e-5, 1.e-5, 1.e-5);

    int i, j;
    Vect3 ra, fp, fm;

    for (i = 0; i < 3; i++) {
        for (j = 0; j < 3; j++)
            ra[j] = (i == j) ? Ad->RGr[j] + dr[j] : Ad->RGr[j];
        EarthNonCentralAcc(fp, ra);
        for (j = 0; j < 3; j++)
            ra[j] = (i == j) ? Ad->RGr[j] - dr[j] : Ad->RGr[j];
        EarthNonCentralAcc(fm, ra);
        for (j = 0; j < 3; j++)
            m(j, i) = (fp[j] - fm[j]) / (2.0 * dr[i]);
    }

    // Получили матрицу в Гринвиче
    // Теперь переводим ее в J2000
    m.MulRight(Ad->CJ2000toGFCS);
}

///////////////////////////////////////////////////////////////////////////////
