///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
#include "ObservePoint.h"
#include "NZEFrame.h"
#include "LocalJ2000Frame.h"
#include "GreenwichFrame.h"
#include "IAstroBase.h"
#include "globals.h"
#include "METEFrame.h"

///////////////////////////////////////////////////////////////////////////////
// Класс наблюдательного пункта
// Методы вычисляют параметры положения указанного объекта относительно
// пункта
///////////////////////////////////////////////////////////////////////////////
// Вычисление дальности, азимута, угла места в системе NEZ
//-----------------------------------------------------------------------------
void ObservePoint::GetRAG(const PhasePoint3D& p, RAzGammaCoords& rag) const
{
    TAssert(InitFlag);

    NZEFrame fr;
    fr.Init(PCoords);

    PhasePoint3D pt;
    pt.SetFrame(fr);
    pt.Assign(p);
    rag.SetXYZ(pt.Coords);
}

void ObservePoint::GetRAG(const PhasePoint6D& p, RAzGammaCoords& rag) const
{
    TAssert(InitFlag);

    NZEFrame fr;
    fr.Init(PCoords);

    PhasePoint6D pt;
    pt.SetFrame(fr);
    pt.Assign(p);
    rag.SetXYZ(pt.CoordsVel);
}

void ObservePoint::GetRAG(const PhasePoint3D& p, const FrameConverter6D& cnv, RAzGammaCoords& rag) const
{
    TAssert(InitFlag);

    Vect6 res;
    cnv.Convert(Vect6(p.Coords, Vect3()), res);

    rag.SetXYZ(Vect3(res.x, res.y, res.z));
}

void ObservePoint::GetRAGt(const PhasePoint6D& p, const FrameConverter6D& cnv, RAzGamma6Coords& rag) const
{
    TAssert(InitFlag);

    Vect6 res;
    cnv.Convert(p.CoordsVel, res);
    rag.SetXYZ(res);
}

void ObservePoint::GetRAGt(const PhasePoint6D& p, RAzGamma6Coords& rag) const
{
    TAssert(InitFlag);

    NZEFrame fr;
    fr.Init(PCoords);

    PhasePoint6D pt;
    pt.SetFrame(fr);
    pt.Assign(p);
    rag.SetXYZ(pt.CoordsVel);
}

void ObservePoint::GetRAGtGCS(Vect6 gcsPos, RAzGamma6Coords& rag) const
{
    TAssert(InitFlag);

    NZEFrame fr;
    fr.Init(PCoords);

    FrameConverter6D lg;
    fr.Get6DGreenwichConverter(lg);
    lg.Invert();

    Vect6 topo;
    lg.Convert(gcsPos, topo);
    rag.SetXYZ(topo);
}

void ObservePoint::GetRAGtFast(const Vect6& p, const FrameConverter6D& cnv, RAzGamma6Coords& rag) const
{
    TAssert(InitFlag);

    Vect6 res;
    cnv.Convert(p, res);

    //	if (res.z < 0){
    //		rag.Coords.Gamma = -1;
    //		return;
    //	}

    rag.SetXYZ(res);

    //	J2000Frame j2000;
    //	FrameConverter6D cnv;
    //	j2000.Get6DConverter(fr, p.T, cnv);
    //	

    //	PhasePoint6D px(GreenwichFrame(),PCoords,Vect3(),p.T);
    //	px.Convert(METEFrame(p.T));

    //	{

    //	Vect3 xx = p.CoordsVel;
    //	xx -= px.CoordsVel;

    /*	double omega_E = 1.00273790934;
    double secday = 86400;
    double mfactor = 2.0*Pi*omega_E/secday;
    double Dtheta = 360/secday*deg2rad;
    double ex_earth_flat = 1/298.26;
    double ex_earth_radius_in_km = 6378.137;
    double sin_lat,cos_lat,sin_theta,cos_theta;
    double top_s,top_e,top_z;
    double rangex,rangey,rangez,rangew;
    double sposx,sposy,sposz;
    double svelx,svely,svelz;
    double uposx,uposy,uposz;
    double uvelx,uvely,uvelz;
    double c,s,achcp,theta;
    double srange,svrange;
    double urange,uvrange;
    double vrangex,vrangey,vrangez,vrangew;
    //	double snposx,snposy,snposz;
    //	double rnangex,rnangey,rnangez,rnangew;
    double beta;

    double lat = geo_userlat;  //rads
    double lon = geo_userlon;  //rads
    double alt = geo_useralt;  //km

    theta = gmst + lon;

    sin_lat   = sin(lat);
    cos_lat   = cos(lat);
    sin_theta = sin(theta);
    cos_theta = cos(theta);

    //	c = 1.0/sqrt(1.0 + ex_earth_flat*(ex_earth_flat - 2.0)*sqr(sin_lat));
    //	s = sqr(1.0 - ex_earth_flat)*c;
    //	achcp = (ex_earth_radius_in_km*c + alt)*cos_lat;

    ///user XYZ ECI coords
    uposx = spos.x*1E3;
    uposy = spos.y*1E3;
    uposz = spos.z*1E3;
    urange=  sqrt(uposx*uposx + uposy*uposy + uposz*uposz);

    //user XZY ECI vels
    uvelx   = spos.vx;
    uvely   = spos.vx;
    uvelz   = spos.vz;
    uvrange = sqrt(uvelx*uvelx + uvely*uvely);

    //satellites XZY ECI coords
    sposx   = p.CoordsVel.x*1E3;
    sposy   = p.CoordsVel.y*1E3;
    sposz   = p.CoordsVel.z*1E3;
    srange  = p.CoordsVel.Len();

    //satellites XYZ ECI vels
    svelx   = p.CoordsVel.vx;
    svely   = p.CoordsVel.vy;
    svelz   = p.CoordsVel.vz;
    svrange = p.CoordsVel.LenV();

    //ranges beetwen user and sat
    rangex = sposx - uposx;
    rangey = sposy - uposy;
    rangez = sposz - uposz;
    rangew = sqrt(rangex*rangex + rangey*rangey + rangez*rangez);

    //vranges beetwen user and sat
    vrangex = svelx - uvelx;
    vrangey = svely - uvely;
    vrangez = svelz - uvelz;
    vrangew = (rangex*vrangex + rangey*vrangey + rangez*vrangez)/rangew;

    top_s = sin_lat*cos_theta*rangex + sin_lat*sin_theta*rangey - cos_lat*rangez;
    top_e =-sin_theta*rangex + cos_theta*rangey;
    top_z = cos_lat*cos_theta*rangex + cos_lat*sin_theta*rangey + sin_lat*rangez;

    double Pcos_theta = -sin_theta*Dtheta;
    double Psin_theta =  cos_theta*Dtheta;

    double Ptop_z = cos_lat*(cos_theta*vrangex + Pcos_theta*rangex) + cos_lat*(sin_theta*vrangey + Psin_theta*rangey) + sin_lat*vrangez;
    double Ptop_e = -(Psin_theta*rangex + sin_theta*vrangex) + (Pcos_theta*rangey + cos_theta*vrangey);
    double Ptop_s = sin_lat*(Pcos_theta*rangex + cos_theta*vrangex) + sin_lat*(Psin_theta*rangey + sin_theta*vrangey) - cos_lat*vrangez;

    /*	double tpz = sqrt(top_s*top_s + top_e*top_e + top_z*top_z);

    double azim = atan(-top_e/top_s);
    if (top_s > 0.0) azim += Pi;
    if (azim < 0.0) azim += 2.0*Pi;

    double el = asin(top_z/rangew);* /
    //	}
    //

    rag.SetXYZ(Vect6(-top_s*1E-3,top_e*1E-3,top_z*1E-3, -Ptop_s, Ptop_e, Ptop_z));
    */
}

//-----------------------------------------------------------------------------
// Вычисление дальности, alpha, delta в местной экв. системе (J2000) 
//-----------------------------------------------------------------------------
void ObservePoint::GetDAD(const PhasePoint3D& p, AlphDelDCoords& dad) const
{
    GeoCoordsPrec g;
    g.SetXYZ(PCoords);

    LocalJ2000Frame fr;
    fr.Init(g);

    PhasePoint3D pt;
    pt.SetFrame(fr);
    pt.Assign(p);
    dad.SetXYZ(pt.Coords);
}

void ObservePoint::GetDAD(const PhasePoint6D& p, AlphDelDCoords& dad) const
{
    GeoCoordsPrec g;
    g.SetXYZ(PCoords);

    LocalJ2000Frame fr;
    fr.Init(g);

    PhasePoint6D pt;
    pt.SetFrame(fr);
    pt.Assign(p);
    dad.SetXYZ(pt.CoordsVel);
}

//-----------------------------------------------------------------------------
// Вычисление дальности, alpha, delta и их производных 
// в местной экв. системе (J2000) 
//-----------------------------------------------------------------------------
void ObservePoint::GetDADt(const Vect6& p, const FrameConverter6D& cnv, AlphDelD6Coords& dad) const
{
    Vect6 res;
    cnv.Convert(p, res);
    dad.SetXYZ(res);
}

void ObservePoint::GetDADt(const PhasePoint6D& p, AlphDelD6Coords& dad) const
{
    GeoCoordsPrec g;
    g.SetXYZ(PCoords);

    LocalJ2000Frame fr;
    fr.Init(g);

    PhasePoint6D pt;
    pt.SetFrame(fr);
    pt.Assign(p);
    dad.SetXYZ(pt.CoordsVel);
}

void ObservePoint::GetDADtPrtDrv(const PhasePoint6D& p, AlphDelD6Coords& dad, DMatrix& prtDrv) const
{
    GeoCoordsPrec g;
    g.SetXYZ(PCoords);

    LocalJ2000Frame fr;
    fr.Init(g);

    PhasePoint6D pt;
    pt.SetFrame(fr);
    pt.Assign(p);
    dad.SetXYZ(pt.CoordsVel);

    J2000Frame j2000;
    FrameConverter6D frameConverter;
    j2000.Get6DConverter(fr, p.T, frameConverter);

    dad.GetPrtDrv(pt.CoordsVel, frameConverter, prtDrv);
}

//-----------------------------------------------------------------------------
// Вычисление фазового угла на Солнце
//-----------------------------------------------------------------------------
double ObservePoint::PhaseSunAngle(const PhasePoint3D& p, int* sign) const
{
    IAstroBase* ab = GetAstroBase();

    J2000Frame fr;
    GreenwichFrame fg;

    PhasePoint3D pg, ps, pt;

    // Положение пункта наблюдения в Гринвиче
    pg.SetFrame(fg);
    pg.T = p.T;
    pg.Coords = PCoords;

    // Положение пункта наблюдения в J2000
    ps.SetFrame(fr);
    ps.Assign(pg);

    // Положение объекта в J2000
    pt.SetFrame(fr);
    pt.Assign(p);

    // Положение Солнца в J2000 в момент p.T
    Vect3 rs;
    ab->GetPosGeo(plIdSun, p.T, &rs);

    // Расстояние от пункта до Солнца
    rs -= ps.Coords;
    double ds = rs.Len();

    // Время прохождения света от пункта до Солнца
    double tau = ds / Clight;

    // Положение Солнца с учетом времени распространения света
    TimeJD t = p.T;
    t -= tau / 86.4;
    ab->GetPosGeo(plIdSun, p.T, &rs);

    // Вектора положения Солнца и объекта относительно пункта
    rs -= ps.Coords;
    Vect3 rt = pt.Coords;
    rt -= ps.Coords;

    if (sign != 0) {
        *sign = (rs[0] * rt[1] - rs[1] * rt[0]) > 0 ? 1 : -1;
    }

    // Сам фазовый угол
    return acos((rs * rt) / (rs.Len() * rt.Len()));
}

double ObservePoint::PhaseSunAngle(const Vect3& sunJ2000, const Vect3& siteJ2000, const Vect3& objectJ2000, int* sign)
{
    //	IAstroBase *ab = GetAstroBase();

    Vect3 rs = sunJ2000;
    Vect3 ps = siteJ2000;
    Vect3 pt = objectJ2000;

    // Расстояние от пункта до Солнца
    //	rs -= ps;
    //	double ds = rs.Len();

    // Время прохождения света от пункта до Солнца
    /*	double tau = ds / Clight;

    // Положение Солнца с учетом времени распространения света
    TimeJD t = p.T;
    t -= tau / 86.4;
    ab->GetPosGeo(plIdSun, p.T, &rs);
    */
    // Вектора положения Солнца и объекта относительно пункта
    rs -= ps;
    Vect3 rt = pt;
    rt -= ps;

    if (sign != 0) {
        *sign = (rs[0] * rt[1] - rs[1] * rt[0]) > 0 ? 1 : -1;
    }

    return acos((rs * rt) / (rs.Len() * rt.Len()));
}


///////////////////////////////////////////////////////////////////////////////
