///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Средства предварительного построения орбит по оптическим измерениям
// методом Лапласа
///////////////////////////////////////////////////////////////////////////////

#include "OptOrbit.h"
#include "GreenwichFrame.h"
#include "IAstroBase.h"
#include <iostream>
#include "Kepler.h"

///////////////////////////////////////////////////////////////////////////////
// Класс упорядочивания элементов в списке по времени
///////////////////////////////////////////////////////////////////////////////
class OptCompare : public TSListCompareDef<OpticMeasurement>
{
public:
    virtual int compare(const OpticMeasurement* p1, const OpticMeasurement* p2)
    {
        if (p1->GetTime() < p2->GetTime())
            return -1;
        if (p1->GetTime() > p2->GetTime())
            return 1;
        return 0;
    };
};

///////////////////////////////////////////////////////////////////////////////
// Класс, выполняющий построение орбиты методом Лапласа
OptOrbit::OptOrbit()
{
    Clear();
}

///////////////////////////////////////////////////////////////////////////////
// Добавление измерения
//-----------------------------------------------------------------------------
bool OptOrbit::Add(const OpticMeasurement* m)
{
    if (OList.Len() == 0) {
        OList.Add(m);
        return true;
    }

    Vect3 stpos = OList[0]->GetStationCoord();
    if (!(stpos == m->GetStationCoord())) {
        return false;
    }

    const IFrame* f = OList[0]->GetSrcFrame();
    if (!f->IsEqual(*m->GetSrcFrame())) {
        return false;
    }

    OList.Add(m);

    return true;
}

bool OptOrbit::CalcVector(double a0, double a1, double a2,
                          double d0, double d1, double d2,
                          const Vect3& wip, const Vect3& vip, const Vect3& rip, double earthRadius, double earthMu,
                          Vect6& res, Vect3& unitVector, double& r0)
{
    Vect3 e, ed;
    DSqrMatrix l1;
    if (!CreateLMatrix(a0, a1, a2, d0, d1, d2, e, ed, l1)) {
        return false;
    }

    // Подготовка к решению уравнения
    r0 = 1.1 * earthRadius;;
    double r1 = 31.1 * earthRadius;;
    double del0 = DeltaL(r0, l1, e, rip, wip, earthMu);
    double del1 = DeltaL(r1, l1, e, rip, wip, earthMu);

    int maxCount = 9999;
    int count = 0;
    // Итерационный цикл решения уравнения 
    while (fabs(r0 - r1) > 1.e-15) {
        double r2 = r1 - del1 * (r1 - r0) / (del1 - del0);
        double del2 = DeltaL(r2, l1, e, rip, wip, earthMu);
        r0 = r1;
        del0 = del1;
        r1 = r2;
        del1 = del2;
        if (++count > maxCount) {
            return false;
        }
    }

    Vect3 rov;
    DeltaKO(l1, r0, rip, wip, earthMu, rov);

    for (int i = 0; i < 3; i++) {
        res[i] = rov[0] * e[i] + rip[i];
        res[i + 3] = rov[1] * e[i] + rov[0] * ed[i] + vip[i];
    }

    unitVector = e;

    return true;
}

void OptOrbit::CalcSiteCoords(const IFrame* srcfrm, TimeJD t0, Vect3 sitecoords, double earthRotationVel, Vect3& wip, Vect3& vip, Vect3& rip)
{
    FrameConverter3D conv;
    srcfrm->Get3DConverter(static_cast<const IFrame&>(GreenwichFrame()), t0, conv);

    // Вычисление скорости и ускорения изм. пункта в Гринвиче
    double w = earthRotationVel;
    rip = sitecoords;

    vip.x = -w * rip.y;
    vip.y = w * rip.x;
    vip.z = 0.0;
    wip.x = -w * w * rip.x;
    wip.y = -w * w * rip.y;
    wip.z = 0.0;
    conv.Invert();

    // Перевод положения, скорости и ускорения ИП в систему пункта
    rip.MulLeft(conv.Rotation);
    vip.MulLeft(conv.Rotation);
    wip.MulLeft(conv.Rotation);
}

bool OptOrbit::CalcDvs(const OptList& meas, TimeJD t0, int m, double& a0, double& a1, double& a2, double& d0, double& d1, double& d2)
{
    DVector aalpha, adelta;
    if (!BuildPolynom(meas, t0, 0, m, aalpha) || !BuildPolynom(meas, t0, 1, m, adelta)) {
        return false;
    }

    // Вычисление alpha, delta и их производных в момент t0
    a0 = aalpha[0];
    a1 = aalpha[1];
    d0 = adelta[0];
    d1 = adelta[1];
    a2 = m > 2 ? 2.0 * aalpha[2] : 0;
    d2 = m > 2 ? 2.0 * adelta[2] : 0;

    return true;
};

void OptOrbit::CalcDvsSec(double a0, double a1, double d0, double d1, Vect3 e, double r0, double mu, double& a2, double& d2)
{
    Vect3 e2;
    for (int i = 0; i < 3; i++) {
        e2[i] = e[i] * mu / (r0 * r0);
    }

    e2a2d2(a0, a1, d0, d1, e2, a2, d2);
}

int OptOrbit::CalcPolyOrder(int vecsize, double interval)
{
    // Назначение порядка полинома
    int m;
    if (vecsize <= 2) {
        m = 2;
    }
    else if (vecsize <= 3) {
        m = 3;
    }
    else if (vecsize <= 6) {
        if (interval * 86.4 < 10.8) {
            m = 3;
        }
        else {
            m = 4;
        }
    }
    else {
        if (interval * 86.4 < 3.6) {
            m = 3;
        }
        else if (interval * 86.4 < 10.8) {
            m = 4;
        }
        else {
            m = 5;
        }
    }

    return m;
}

void OptOrbit::GetMeasTimes(const OptList& meas, TimeJD& t0, double& tau)
{
    tau = (meas[meas.Len() - 1]->GetTime() - meas[0]->GetTime()).GetDouble();
    t0 = meas[0]->GetTime();
    t0 += tau * 0.5;
}

bool OptOrbit::Build(Vect6& res)
{
    if (OList.Len() < 2) {
        return false;
    }

    const Vect3 sitecoords = OList[0]->GetStationCoord();
    const int vecsize = OList.Len();
    const IFrame* srcfrm = OList[0]->GetSrcFrame();
    const double earthRadius = GetAstroBase()->GetEarthRadius();
    const double earthRotationVel = GetAstroBase()->GetEarthRotation();
    const double earthMu = GetAstroBase()->GetEarthMu();

    OptCompare comp;
    OList.SetSortFunction(&comp);

    TimeJD t0;
    double tau;
    GetMeasTimes(OList, t0, tau);

    int m = CalcPolyOrder(vecsize, tau);

    Vect3 wip, vip, rip;
    CalcSiteCoords(srcfrm, t0, sitecoords, earthRotationVel, wip, vip, rip);

    double a0, a1, a2, d0, d1, d2;
    if (!CalcDvs(OList, t0, m, a0, a1, a2, d0, d1, d2)) {
        return false;
    }

    bool ret = false;
    int flag = 0;

    while (true) {

        Vect3 unitVector;
        double r0;
        if (!CalcVector(a0, a1, a2, d0, d1, d2, wip, vip, rip, earthRadius, earthMu, res, unitVector, r0)) {
            ret = false;
            break;
        }

        Kepler kep;
        kep.SetXYZ(res);

        if (kep.A < 0) {
            if (flag == 0) {
                if (!CalcDvs(OList, t0, 2, a0, a1, a2, d0, d1, d2)) {
                    ret = false;
                    break;
                }
                flag = 1;
            }
            else if (flag == 1) {
                CalcDvsSec(a0, a1, d0, d1, unitVector, r0, earthMu, a2, d2);
                flag = 2;
            }
            else {
                ret = false;
                break;
            }
        }
        else {
            ret = true;
            break;
        }
    }

    return ret;
}

int GetQuadrand(double angle)
{
    double csin = sin(angle);
    double ccos = cos(angle);
    int qudrant;
    if (csin > 0) {
        if (ccos > 0) {
            qudrant = 1;
        }
        else {
            qudrant = 2;
        }
    }
    else {
        if (ccos > 0) {
            qudrant = 4;
        }
        else {
            qudrant = 3;
        }
    }
    return qudrant;
}

//-----------------------------------------------------------------------------
// Построение полинома, аппроксиморуещего одну из компонент измерения
//-----------------------------------------------------------------------------
bool OptOrbit::BuildPolynom(const OptList& meas, const TimeJD& t0, int k, int m, DVector& res)
{
    DSymmMatrix a(m);
    DVector b(m);

    for (int l = 0; l < m; l++) {
        for (int j = l; j < m; j++) {
            a(l, j) = 0.0;
        }
        b(l) = 0.0;
    }

    int quadrants[4] = {0, 0, 0, 0};
    bool correct = false;
    if (k == 0) {
        for (int i = 0; i < meas.Len(); i++) {
            DVector fv;
            meas[i]->GetRealVector(fv);
            double f = fv(k);
            quadrants[GetQuadrand(f) - 1] = 1;
        }
        if (quadrants[0] == 1 && quadrants[3] == 1) {
            correct = true;
        }
    }

    // Цикл по измерениям
    for (int i = 0; i < meas.Len(); i++) {
        double dt = (meas[i]->GetTime() - t0).GetDouble() * 86.4;
        DVector fv;
        DSymmMatrix pm;
        meas[i]->GetRealVector(fv);
        meas[i]->GetWeightMatrix(pm);
        double f = fv(k);
        if (k == 0 && correct && GetQuadrand(f) == 4) {
            f -= Pi2;
        }
        double p = pm(k, k);
        for (int l = 0; l < m; l++) {
            b(l) += p * f * pow(dt, l);
            for (int j = l; j < m; j++) {
                a(l, j) += p * pow(dt, l + j);
            }
        }

    }

    // Решение системы уравнений
    return TDMSolution(b, a, res);
}

//-----------------------------------------------------------------------------
// Расчет единичного вектора и обратной матрицы Лапласа
//-----------------------------------------------------------------------------
bool OptOrbit::CreateLMatrix(double a0, double a1, double a2,
                             double d0, double d1, double d2,
                             Vect3& e, Vect3& ed, DSqrMatrix& l1)
{
    double cd = cos(d0);
    double sd = sin(d0);
    double ca = cos(a0);
    double sa = sin(a0);

    e.x = cd * ca;
    e.y = cd * sa;
    e.z = sd;

    ed.x = -d1 * sd * ca - a1 * cd * sa;
    ed.y = -d1 * sd * sa + a1 * cd * ca;
    ed.z = d1 * cd;

    Vect3 ed2;

    ed2.x = (-d1 * d1 * cd - d2 * sd - a1 * a1 * cd) * ca + (2.0 * a1 * d1 * sd - a2 * cd) * sa;
    ed2.y = (-d1 * d1 * cd - d2 * sd - a1 * a1 * cd) * sa + (-2.0 * a1 * d1 * sd + a2 * cd) * ca;
    ed2.z = -d1 * d1 * sd + d2 * cd;

    l1.Init(3U);

    for (int i = 0; i < 3; i++) {
        l1(i, 0) = e[i];
        l1(i, 1) = 2 * ed[i];
        l1(i, 2) = ed2[i];
    }

    return l1.Inverse();
}

void OptOrbit::e2a2d2(double a0, double a1, double d0, double d1, Vect3 e2, double& a2, double& d2)
{
    double cd = cos(d0);
    double sd = sin(d0);
    double ca = cos(a0);
    double sa = sin(a0);

    a2 = (-e2[0] * sa + e2[1] * ca + 2 * a1 * d1 * sd) / cd;
    d2 = (e2[2] + d1 * d1 * sd) / cd;
}

//-----------------------------------------------------------------------------
// Вычисление расстояния от наблюдателя до КО и двух его производных
// по времени
//-----------------------------------------------------------------------------
void OptOrbit::DeltaKO(const DSqrMatrix& l1, double r, const Vect3& rip,
                       const Vect3& wip, double earthMu, Vect3& ro)
{
    double r3 = r * r * r;

    DVector b(3U);
    for (int i = 0; i < 3; i++) {
        b(i) = -wip[i] - earthMu * rip[i] / r3;
    }

    b.MulLeft(l1);

    ro[2] = b(0);
    ro[1] = b(1);
    ro[0] = fabs(b(2));
}

//-----------------------------------------------------------------------------
// Расчет невязки уравнения Лапласа
//-----------------------------------------------------------------------------
double OptOrbit::DeltaL(double r, const DSqrMatrix& l1, const Vect3& e, const Vect3& rip, const Vect3& wip, double earthMu)
{
    Vect3 ro;
    DeltaKO(l1, r, rip, wip, earthMu, ro);
    double rip2 = rip.Len2();
    double erip = e * rip;
    return r * r - ro[0] * ro[0] - rip2 - 2 * ro[0] * erip;
}

TimeJD OptOrbit::GetTime() const
{
    double tau = (OList[OList.Len() - 1]->GetTime() - OList[0]->GetTime()).GetDouble();
    TimeJD jd = OList[0]->GetTime();
    jd += tau * 0.5;
    return jd;
}

void OptOrbit::Clear()
{
    OList.Clear();
}

///////////////////////////////////////////////////////////////////////////////
