///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Класс, моделирующий оптические измерения c учетом времени прохождения
// сигнала.
// Класс является абстрактным, поскольку в нем не реализован метод
// GetSrcFrame()
///////////////////////////////////////////////////////////////////////////////
#include "OpticDelayMeasurement.h"
#include "GreenwichFrame.h"
#include "Consts.h"
#include "PhasePoint.h"
#include <iostream>
#include <iomanip>

///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
//-----------------------------------------------------------------------------
OpticDelayMeasurement::OpticDelayMeasurement() : OpticMeasurement()
{
    TimePrec = 0.0001e-3; // По умолчанию назначем точность определения времени 
    // распространения сигнала равной одной сотой секунды
}

//-----------------------------------------------------------------------------
// Метод определяет время t, которое соответствует моменту излучения сигнала с 
// объекта, а также его положение на орбите в этот момент времени
//-----------------------------------------------------------------------------
bool OpticDelayMeasurement::DetectPosition(ITraject& trj, TimeJD& t, const IFrame* fr, Vect6* f) const
{
    Vect6 fo;
    Vect3 rc;
    double tau = 0, tau1 = 0;
    GreenwichFrame gr;

    if (fr == NULL) {
        fr = &gr;
    }

    if (f == NULL)
        f = &fo;

    if (RecPosFrInit) {
        rc = RecPosFr;
    }
    else {
        if (!DetectRecieverPos(t, *fr, rc)) {
            return false;
        }
    }

    int count = 0;
    for (;;) {
        t = GetTime();
        t -= tau / 86.4;

        if (!trj.Predict(t, *fr, *f)) {
            return false;
        }

        Vect3 del = rc;
        del -= *f;
        tau1 = del.Len() / Clight;
        if (fabs(tau - tau1) < TimePrec) {
            return true;
        }

        tau = tau1;

        if (count++ > 100) {
            return false;
        }
    }
}

bool OpticDelayMeasurement::DetectPositionEx(ITraject& trj, TimeJD& t, Vect6& f) const
{
    double tau = 0, tau1 = 0;
    Vect3 rc = RecPosFr;

    for (;;) {
        t = GetTime();
        t -= tau / 86.4;

        if (!trj.PredictEx(t, f)) {
            return false;
        }

        //	return true;
        //	break; //TODO

        Vect3 del = rc;
        del -= f;
        tau1 = del.Len() / Clight;
        if (fabs(tau - tau1) < TimePrec) {
            return true;
        }

        tau = tau1;
    }
}

//-----------------------------------------------------------------------------
// Метод определяет положение приемника сигнала f в момент времени t
// в системе отсчета, определяемой параметром fr
//-----------------------------------------------------------------------------
bool OpticDelayMeasurement::DetectRecieverPos(const TimeJD& t, const IFrame& fr,
                                              Vect3& f) const
{
    /*  PhasePoint3D dst;
    dst.SetFrame(fr);

    PhasePoint3D src;
    GreenwichFrame gr;
    src.Init(gr, RecPos, MTime);

    dst.Assign(src);

    f = dst.Coords;
    */

    PhasePoint3D dst(static_cast<const IFrame&>(GreenwichFrame()), RecPos, MTime);

    dst.Convert(fr);
    f = dst.Coords;

    return true;
}

bool OpticDelayMeasurement::PrepareDstFrame(const IFrame& fr)
{
    SetWrkFrame(fr);

    if (!DetectRecieverPos(GetTime(), fr, RecPosFr)) {
        return false;
    }

    RecPosFrInit = true;

    return true;
}

bool OpticDelayMeasurement::PrepareDstFrame_gr(const IFrame& fr)
{
    SetWrkFrame(J2000Frame());

    if (!DetectRecieverPos(GetTime(), fr, RecPosFr)) {
        return false;
    }

    RecPosFrInit = true;

    return true;
}

///////////////////////////////////////////////////////////////////////////////
