///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Класс, моделирующий оптические измерения без учета времени прохождения
// времени сигнала.
///////////////////////////////////////////////////////////////////////////////
#include "OpticMeasurement.h"
#include "Angles.h"
#include "Consts.h"
#include "LocalJ2000Frame.h"
#include <iostream>
#include "GreenwichFrame.h"

///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
//-----------------------------------------------------------------------------

class CAA3DCoordinate
{
public:
    //Constructors / Destructors
    CAA3DCoordinate() : X(0), Y(0), Z(0)
    {
    };

    //member variables
    double X;
    double Y;
    double Z;
};

class CAA2DCoordinate
{
public:
    //Constructors / Destructors
    CAA2DCoordinate() : X(0), Y(0)
    {
    };

    //member variables
    double X;
    double Y;
};

struct AberrationCoefficient
{
    int L2;
    int L3;
    int L4;
    int L5;
    int L6;
    int L7;
    int L8;
    int Ldash;
    int D;
    int Mdash;
    int F;
    int xsin;
    int xsint;
    int xcos;
    int xcost;
    int ysin;
    int ysint;
    int ycos;
    int ycost;
    int zsin;
    int zsint;
    int zcos;
    int zcost;
};

static AberrationCoefficient g_AberrationCoefficients[] =
{
    //L2   L3   L4  L5  L6  L7  L8  Ldash D   Mdash F   xsin      xsint xcos    xcost ysin   ysint ycos     ycost zsin   zsint zcos    zcost
    {0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1719914, -2, -25, 0, 25, -13, 1578089, 156, 10, 32, 684185, -358},
    {0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6434, 141, 28007, -107, 25697, -95, -5904, -130, 11141, -48, -2559, -55},
    {0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 715, 0, 0, 0, 6, 0, -657, 0, -15, 0, -282, 0},
    {0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 715, 0, 0, 0, 0, 0, -656, 0, 0, 0, -285, 0},
    {0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 486, -5, -236, -4, -216, -4, -446, 5, -94, 0, -193, 0},
    {0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 159, 0, 0, 0, 2, 0, -147, 0, -6, 0, -61, 0},
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 26, 0, 0, 0, -59, 0},
    {0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 39, 0, 0, 0, 0, 0, -36, 0, 0, 0, -16, 0},
    {0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 33, 0, -10, 0, -9, 0, -30, 0, -5, 0, -13, 0},
    {0, 2, 0, -1, 0, 0, 0, 0, 0, 0, 0, 31, 0, 1, 0, 1, 0, -28, 0, 0, 0, -12, 0},
    {0, 3, -8, 3, 0, 0, 0, 0, 0, 0, 0, 8, 0, -28, 0, 25, 0, 8, 0, 11, 0, 3, 0},
    {0, 5, -8, 3, 0, 0, 0, 0, 0, 0, 0, 8, 0, -28, 0, -25, 0, -8, 0, -11, 0, -3, 0},
    {2, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 21, 0, 0, 0, 0, 0, -19, 0, 0, 0, -8, 0},
    {1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -19, 0, 0, 0, 0, 0, 17, 0, 0, 0, 8, 0},
    {0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 17, 0, 0, 0, 0, 0, -16, 0, 0, 0, -7, 0},
    {0, 1, 0, -2, 0, 0, 0, 0, 0, 0, 0, 16, 0, 0, 0, 0, 0, 15, 0, 1, 0, 7, 0},
    {0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 16, 0, 0, 0, 1, 0, -15, 0, -3, 0, -6, 0},
    {0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 11, 0, -1, 0, -1, 0, -10, 0, -1, 0, -5, 0},
    {2, -2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -11, 0, -10, 0, 0, 0, -4, 0, 0, 0},
    {0, 1, 0, -1, 0, 0, 0, 0, 0, 0, 0, -11, 0, -2, 0, -2, 0, 9, 0, -1, 0, 4, 0},
    {0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7, 0, -8, 0, -8, 0, 6, 0, -3, 0, 3, 0},
    {0, 3, 0, -2, 0, 0, 0, 0, 0, 0, 0, -10, 0, 0, 0, 0, 0, 9, 0, 0, 0, 4, 0},
    {1, -2, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9, 0, 0, 0, 0, 0, -9, 0, 0, 0, -4, 0},
    {2, -3, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9, 0, 0, 0, 0, 0, -8, 0, 0, 0, -4, 0},
    {0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, -9, 0, -8, 0, 0, 0, -3, 0, 0, 0},
    {2, -4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -9, 0, 8, 0, 0, 0, 3, 0, 0, 0},
    {0, 3, -2, 0, 0, 0, 0, 0, 0, 0, 0, 8, 0, 0, 0, 0, 0, -8, 0, 0, 0, -3, 0},
    {0, 0, 0, 0, 0, 0, 0, 1, 2, -1, 0, 8, 0, 0, 0, 0, 0, -7, 0, 0, 0, -3, 0},
    {8, -12, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4, 0, -7, 0, -6, 0, 4, 0, -3, 0, 2, 0},
    {8, -14, 0, 0, 0, 0, 0, 0, 0, 0, 0, -4, 0, -7, 0, 6, 0, -4, 0, 3, 0, -2, 0},
    {0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, -6, 0, -5, 0, -4, 0, 5, 0, -2, 0, 2, 0},
    {3, -4, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, -1, 0, -2, 0, -7, 0, 1, 0, -4, 0},
    {0, 2, 0, -2, 0, 0, 0, 0, 0, 0, 0, 4, 0, -6, 0, -5, 0, -4, 0, -2, 0, -2, 0},
    {3, -3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -7, 0, -6, 0, 0, 0, -3, 0, 0, 0},
    {0, 2, -2, 0, 0, 0, 0, 0, 0, 0, 0, 5, 0, -5, 0, -4, 0, -5, 0, -2, 0, -2, 0},
    {0, 0, 0, 0, 0, 0, 0, 1, -2, 0, 0, 5, 0, 0, 0, 0, 0, -5, 0, 0, 0, -2, 0},
};


//////////////////////////////// Implementation ///////////////////////////////

CAA3DCoordinate EarthVelocity(double JD)
{
    double T = (JD - 2451545) / 36525;
    double L2 = 3.1761467 + 1021.3285546 * T;
    double L3 = 1.7534703 + 628.3075849 * T;
    double L4 = 6.2034809 + 334.0612431 * T;
    double L5 = 0.5995465 + 52.9690965 * T;
    double L6 = 0.8740168 + 21.3299095 * T;
    double L7 = 5.4812939 + 7.4781599 * T;
    double L8 = 5.3118863 + 3.8133036 * T;
    double Ldash = 3.8103444 + 8399.6847337 * T;
    double D = 5.1984667 + 7771.3771486 * T;
    double Mdash = 2.3555559 + 8328.6914289 * T;
    double F = 1.6279052 + 8433.4661601 * T;

    CAA3DCoordinate velocity;

    int nAberrationCoefficients = sizeof(g_AberrationCoefficients) / sizeof(AberrationCoefficient);
    for (int i = 0; i < nAberrationCoefficients; i++) {
        double Argument = g_AberrationCoefficients[i].L2 * L2 + g_AberrationCoefficients[i].L3 * L3 +
            g_AberrationCoefficients[i].L4 * L4 + g_AberrationCoefficients[i].L5 * L5 +
            g_AberrationCoefficients[i].L6 * L6 + g_AberrationCoefficients[i].L7 * L7 +
            g_AberrationCoefficients[i].L8 * L8 + g_AberrationCoefficients[i].Ldash * Ldash +
            g_AberrationCoefficients[i].D * D + g_AberrationCoefficients[i].Mdash * Mdash +
            g_AberrationCoefficients[i].F * F;
        velocity.X += (g_AberrationCoefficients[i].xsin + g_AberrationCoefficients[i].xsint * T) * sin(Argument);
        velocity.X += (g_AberrationCoefficients[i].xcos + g_AberrationCoefficients[i].xcost * T) * cos(Argument);

        velocity.Y += (g_AberrationCoefficients[i].ysin + g_AberrationCoefficients[i].ysint * T) * sin(Argument);
        velocity.Y += (g_AberrationCoefficients[i].ycos + g_AberrationCoefficients[i].ycost * T) * cos(Argument);

        velocity.Z += (g_AberrationCoefficients[i].zsin + g_AberrationCoefficients[i].zsint * T) * sin(Argument);
        velocity.Z += (g_AberrationCoefficients[i].zcos + g_AberrationCoefficients[i].zcost * T) * cos(Argument);
    }

    return velocity;
}

CAA2DCoordinate EquatorialAberration(double Alpha, double Delta, double JD)
{
    //Convert to radians
    //  Alpha = CAACoordinateTransformation::DegreesToRadians(Alpha*15);
    //  Delta = CAACoordinateTransformation::DegreesToRadians(Delta);

    double cosAlpha = cos(Alpha);
    double sinAlpha = sin(Alpha);
    double cosDelta = cos(Delta);
    double sinDelta = sin(Delta);

    CAA3DCoordinate velocity = EarthVelocity(JD);

    //What is the return value
    CAA2DCoordinate aberration;

    aberration.X = ((velocity.Y * cosAlpha - velocity.X * sinAlpha) / (17314463350.0 * cosDelta));
    aberration.Y = (-(((velocity.X * cosAlpha + velocity.Y * sinAlpha) * sinDelta - velocity.Z * cosDelta) / 17314463350.0));

    return aberration;
}

//=======================================
OpticMeasurement::OpticMeasurement() : AFreeMeasurement()
{
    Type = mtOptic;
    m_aberation_method = aberration_method::NONE;
    m_isBad = false;
    RecPosFrInit = false;
    VAlpha = 0;
    VDelta = 0;
}

//-----------------------------------------------------------------------------
// Метод возвращает в объекте v вектор измерений
//-----------------------------------------------------------------------------
void OpticMeasurement::GetRealVector(DVector& v) const
{
    v.Init(2U);
    v[0] = Alpha;
    v[1] = Delta;
};
//-----------------------------------------------------------------------------
// Метод возвращает в объекте v вектор СКО измерений
//-----------------------------------------------------------------------------
void OpticMeasurement::GetMSEVector(DVector& v) const
{
    v.Init(2U);
    v[0] = 1.0 / PAlpha;
    v[1] = 1.0 / PDelta;
};

//-----------------------------------------------------------------------------
// Метод осуществляет моделирование измерения для космического объекта, 
// движение которого описывается с помощью переменной trj. 
//-----------------------------------------------------------------------------
bool OpticMeasurement::Simulate(int matrixSize, ITraject& trj, TimeJD& t, const IFrame& fr, Vect6* f, DVector* v, DMatrix* m, DVector* del) const
{
    Vect6 pv;

    if (f == 0 && (v != 0 || m != 0 || del != 0)) {
        f = &pv;
    }

    J2000Frame wrkfrm;
    if (!DetectPosition(trj, t, (const IFrame *)&wrkfrm, f)) {
        return false;
    }
            
    if (v == 0 && m == 0 && del == 0) {
        return true;
    }

    DVector ang(2);
    Vect6 crd;
    FrameConverter.Convert(*f, crd);

    double r2 = crd.Len();
    ang(1) = r2 > 0 ? asin(crd.z / r2) : 0.0;
    ang(0) = (crd.x != 0 || crd.y != 0) ? atan2(crd.y, crd.x) : 0.0; // <-- [-PI,+PI]
    ang(0) = Angle2Pi(ang(0));

    if (v != 0) {
        *v = ang;
    }

    if (del != 0) {
        Diff(*del, NULL, &ang);
    }

    if (m != 0) {

        double ra2 = crd.Len2();
        double ras2 = crd.x * crd.x + crd.y * crd.y;
        double ras = sqrt(ras2);
        Vect3 dalfaa, ddeltaa, dalfaf, ddeltaf;
        dalfaa[0] = -crd.y / ras2;
        dalfaa[1] = crd.x / ras2;
        dalfaa[2] = 0.0;
        ddeltaa[0] = -crd.x * crd.z / (ras * ra2);
        ddeltaa[1] = -crd.y * crd.z / (ras * ra2);
        ddeltaa[2] = ras / ra2;
        FrameConverter3D inv;
        FrameConverter.Converter3D.Invert(inv);
        dalfaf = dalfaa;
        dalfaf.MulLeft(inv.Rotation);
        ddeltaf = ddeltaa;
        ddeltaf.MulLeft(inv.Rotation);
        m->Init(2, matrixSize);
        m->Init(0.0);
        unsigned i;
        for (i = 0; i < 3; i++) {
            (*m)(0, i) = dalfaf[i];
            (*m)(1, i) = ddeltaf[i];
        }
    }

    return true;
}

//-----------------------------------------------------------------------------
// Метод осуществляет моделирование измерения для космического объекта, 
// движение которого описывается с помощью переменной trj. 
//-----------------------------------------------------------------------------
bool OpticMeasurement::Simulate_gr(int matrixSize, ITraject& trj, TimeJD& t, const IFrame& fr, Vect6* f, DVector* v, DMatrix* m, DVector* del) const
{
    Vect6 pv;

    if (f == 0 && (v != 0 || m != 0 || del != 0)) {
        f = &pv;
    }

    //J2000Frame wrkfrm;
    GreenwichFrame wrkfrm;
    if (!DetectPosition(trj, t, (const IFrame *)&wrkfrm, f)) {
        return false;
    }

    PhasePoint6D pp(wrkfrm, *f, t);
    pp.Convert(fr);
    *f = pp.CoordsVel;

    if (v == 0 && m == 0 && del == 0) {
        return true;
    }

    DVector ang(2);
    Vect6 crd;
    FrameConverter.Convert(*f, crd);

    double r2 = crd.Len();
    ang(1) = r2 > 0 ? asin(crd.z / r2) : 0.0;
    ang(0) = (crd.x != 0 || crd.y != 0) ? atan2(crd.y, crd.x) : 0.0; // <-- [-PI,+PI]
    ang(0) = Angle2Pi(ang(0));

    if (v != 0) {
        *v = ang;
    }

    if (del != 0) {
        Diff(*del, NULL, &ang);
    }

    if (m != 0) {

        double ra2 = crd.Len2();
        double ras2 = crd.x * crd.x + crd.y * crd.y;
        double ras = sqrt(ras2);
        Vect3 dalfaa, ddeltaa, dalfaf, ddeltaf;
        dalfaa[0] = -crd.y / ras2;
        dalfaa[1] = crd.x / ras2;
        dalfaa[2] = 0.0;
        ddeltaa[0] = -crd.x * crd.z / (ras * ra2);
        ddeltaa[1] = -crd.y * crd.z / (ras * ra2);
        ddeltaa[2] = ras / ra2;
        FrameConverter3D inv;
        FrameConverter.Converter3D.Invert(inv);
        dalfaf = dalfaa;
        dalfaf.MulLeft(inv.Rotation);
        ddeltaf = ddeltaa;
        ddeltaf.MulLeft(inv.Rotation);
        m->Init(2, matrixSize);
        m->Init(0.0);
        unsigned i;
        for (i = 0; i < 3; i++) {
            (*m)(0, i) = dalfaf[i];
            (*m)(1, i) = ddeltaf[i];
        }
    }

    return true;
}
//-----------------------------------------------------------------------------
// Вычисление невязок между измерениями
// Если m1 или m2 == NULL, принмается текущее измерение
//-----------------------------------------------------------------------------
void OpticMeasurement::Diff(DVector& d, const DVector* m1, const DVector* m2) const
{
    d.Init(2U);

    d(0) = AngleDiff(m1 == NULL ? Alpha : (*m1)(0), m2 == NULL ? Alpha : (*m2)(0));
    d(1) = AngleDiff(m1 == NULL ? Delta : (*m1)(1), m2 == NULL ? Delta : (*m2)(1));
}

void OpticMeasurement::AberrationCorrect(double alpha, double delta, TimeJD date, bool backwardCorrection, aberration_method ab, double& correctedAlpha, double& correctedDelta)
{
    if (ab == aberration_method::AA) {
        CAA2DCoordinate r = EquatorialAberration(alpha, delta, date.GetDays() + date.GetFraction());
        if (backwardCorrection) {
            correctedAlpha = alpha - r.X;
            correctedDelta = delta - r.Y;
        }
        else {
            correctedAlpha = alpha + r.X;
            correctedDelta = delta + r.Y;
        }
    }
    else if(ab == aberration_method::ST){
        Vect3 v;

        // Вычисление вектора орбитального движения Земли
        double u = 0.0172021177 * (date - TimeJD(2451544, 0.5)).GetDouble();
        double cu = cos(u);
        double su = sin(u);
        v.x = -29.330 * cu + 5.157 * su;
        v.y = -4.731 * cu - 26.911 * su;
        v.z = -2.051 * cu - 11.667 * su;

        // Вычисление единичного вектора от наблюдателя на КО
        Vect3 e;
        e.x = cos(alpha) * cos(delta);
        e.y = sin(alpha) * cos(delta);
        e.z = sin(delta);

        // Вектор, направленный на КО с учетом поправок на аберрацию
        Vect3 r;
        r = e;
        r *= Clight;

        if (backwardCorrection) {
            r -= v;
        }
        else {
            r += v;
        }

        // Вычисление значений углов с поправками
        double rs = sqrt(r.x * r.x + r.y * r.y);
        double ca = r.x / rs;
        correctedAlpha = r.y >= 0.0 ? acos(ca) : Pi2 - acos(ca);
        double sd = r.z / r.Len();
        correctedDelta = asin(sd);
    }
}

//-----------------------------------------------------------------------------
// Коррекция с учетом звездной аберрации
//-----------------------------------------------------------------------------
void OpticMeasurement::Correct(const aberration_method& ab)
{
    AberrationCorrect(Alpha, Delta, MTime, false, ab, Alpha, Delta);
    m_aberation_method = ab;
}

/////////////////////////////////////////////////////////////////////////////////
//bool OpticMeasurement::SimulateXXX(int matrixSize, ITraject &trj, TimeJD &t, const IFrame &fr, Vect6 *f /*= 0*/, Vect6* ff /*= 0*/, DVector *v /*= 0*/, DMatrix *m /*= 0*/, DVector *del /*= 0*/ ) const
//{
//	Vect6 pv;
//
//	if (f == 0 && (v != 0 || m != 0 || del != 0))
//		f = &pv;
//
//	J2000Frame wrkfrm; 
//	if (!DetectPosition(trj, t, (const IFrame *)&wrkfrm, f))
//		return false;
//
//	*ff = *f;
//
//	if (v == 0 && m == 0 && del == 0)
//		return true;
//
//	DVector ang(2);
//	Vect6 crd;
//	FrameConverter.Convert(*f, crd);
//
//	*f = crd;
//
//	double r2 = crd.Len();
//	ang(1) = r2 > 0 ? asin(crd.z / r2) : 0.0; 
//	ang(0) = (crd.x !=0 || crd.y !=0 ) ? atan2(crd.y, crd.x) : 0.0 ; // <-- [-PI,+PI]
//	
//	ang(0) = Angle2Pi(ang(0));
//	
//	if (v != 0){
//		*v = ang;
//	}
//
//	if (del != 0){
//		Diff(*del, NULL, &ang);
//	}
//
//	if (m != 0)
//	{
//		double ra2 = crd.Len2();
//		double ras2 = crd.x * crd.x + crd.y * crd.y;
//		double ras = sqrt(ras2);
//		Vect3 dalfaa, ddeltaa, dalfaf, ddeltaf;
//		dalfaa[0] = -crd.y / ras2;
//		dalfaa[1] = crd.x / ras2;
//		dalfaa[2] = 0.0;
//		ddeltaa[0] = -crd.x * crd.z / (ras * ra2);
//		ddeltaa[1] = -crd.y * crd.z / (ras * ra2);
//		ddeltaa[2] = ras / ra2;
//		FrameConverter3D inv;
//		FrameConverter.Converter3D.Invert(inv);
//		dalfaf = dalfaa;
//		dalfaf.MulLeft(inv.Rotation);
//		ddeltaf = ddeltaa;
//		ddeltaf.MulLeft(inv.Rotation);
//
//		m->Init(2, matrixSize);
//		m->Init(0.0);
//		unsigned i;
//		for (i = 0; i < 3; i++)
//		{
//			(*m)(0, i) = dalfaf[i];
//			(*m)(1, i) = ddeltaf[i];
//		}
//	}
//
//	return true;
//}
/////////////////////////////////////////////////////////////////////////////////
