///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Класс, моделирующий оптические измерения без учета времени прохождения
// сигнала.
// Класс является абстрактным, поскольку в нем не реализован метод
// GetSrcFrame()
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_OPTICMESUREMENT_H
#define BNOTOOL_OPTICMESUREMENT_H
///////////////////////////////////////////////////////////////////////////////
#include "AFreeMeasurement.h"

///////////////////////////////////////////////////////////////////////////////

enum class aberration_method
{
    NONE,
    ST,
    AA
};

class OpticMeasurement : public AFreeMeasurement
{
public:
    // Измеренное значение alpha
    double Alpha;
    // Измеренное значение delta
    double Delta;
    // Вес для alpha
    double PAlpha;
    // Вес для delta
    double PDelta;

    //скорости
    double VAlpha;
    double VDelta;

    // Конструктор по умолчанию
    OpticMeasurement();
    // Метод возвращает размерность вектора измерений
    virtual unsigned GetSize() const
    {
        return 2;
    };
    // Метод возвращает в объекте v вектор измерений
    virtual void GetRealVector(DVector& v) const;
    // Метод возвращает в объекте v вектор СКО измерений
    virtual void GetMSEVector(DVector& v) const;
    // Метод осуществляет моделирование измерения для космического объекта, 
    // движение которого описывается с помощью переменной trj. 
    // Метод вычисляет следующие параметры:
    // t	-	время, которое соответствует положению космического объекта
    //      на орбите для текущего измерения (это время равно времени измерения, 
    //      если не учитывается запаздывание сигнала)
    // fr	-	система отсчета, в которой задаются положение и скорости аппарата 
    //      в момент измерения, а также матрица частных производных (см. далее параметры f и m)
    // v	-	указатель вектор моделируемого измерения. Если этот параметр равен 0, 
    //      вектор моделируемых измерений не вычисляется
    // m	-	матрица производных размерности Nx6, где N - размерность вектора измерений 
    //      (т.е. матрица состоит из N строк и 6 столбцов). В столбцах расположены
    //      производные по Rx Ry Rz Vx  Vy Vz, а в строках - компоненты вектора измерений. 
    //      Если этот параметр равен 0, вектор моделируемых измерений не вычисляется
    // del	-	вектор невязок измеряемых параметров. Если этот параметр равен 0, 
    //        вектор невязок не вычисляется
    virtual bool Simulate(int matrixSize, ITraject& trj, TimeJD& t, const IFrame& fr, Vect6* f = 0, DVector* v = 0, DMatrix* m = 0, DVector* del = 0) const;
    virtual bool Simulate_gr(int matrixSize, ITraject& trj, TimeJD& t, const IFrame& fr, Vect6* f = 0, DVector* v = 0, DMatrix* m = 0, DVector* del = 0) const;

    virtual void Diff(DVector& d, const DVector* m1 = NULL, const DVector* m2 = NULL) const;

    //	bool SimulateOrbitAnalyzer(int matrixSize, ITraject &trj, TimeJD &t, const IFrame &fr, Vect6 *f, Vect6 *topo, DVector *v, DMatrix *m, DVector *del) const;

    //	virtual bool Simulate2(ITraject &trj, TimeJD &t, const IFrame &fr, Vect6 *f = 0,
    //		DVector *v = 0, DMatrix *m = 0, DVector *del = 0) const;

    //	virtual bool SimulateXXX(int matrixSize, ITraject &trj, TimeJD &t, const IFrame &fr, Vect6 *f = 0, Vect6* ff = 0,
    //		DVector *v = 0, DMatrix *m = 0, DVector *del = 0) const;

    //	bool SimulateEx(int matrixSize, ITraject &trj, TimeJD &t, Vect6 &f, double* v,  double *del) const;

    virtual bool DetectPositionEx(ITraject& trj, TimeJD& t, Vect6& f) const { return false; };
    
    // Получение флага коррекции аберрации
    const aberration_method& get_aberation_method() const { return m_aberation_method; }
    // Коррекция с учетом звездной аберрации
    void Correct(const aberration_method& ab);
    // Получение координат станции в Гринвиче
    virtual const Vect3& GetStationCoord() const = 0;

    Vect3 GetRecPosFr() const { return RecPosFr; }

    virtual void SetIan(unsigned k, bool f = true)
    {
        m_isBad = f;
    };

    virtual bool GetIan(unsigned k) const
    {
        return m_isBad;
    };

    static void AberrationCorrect(double alpha, double delta, TimeJD date, bool backwardCorrection, aberration_method ab, double& correctedAlpha, double& correctedDelta);
protected:
    Vect3 RecPosFr;
    bool RecPosFrInit;
    bool m_isBad;

    // Флаг учета аберрации
    aberration_method m_aberation_method;
};

///////////////////////////////////////////////////////////////////////////////
#endif
