///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Классы фазовых точек
////////////////////////////////////////////////////////////////////////////////
#include "PhasePoint.h"
#include "J2000Frame.h"
////////////////////////////////////////////////////////////////////////////////
// Фазовая точка
// Деструктор
IPhasePoint::~IPhasePoint()
{
    if (Frame != NULL)
        delete Frame;

    Frame = NULL;
}

//------------------------------------------------------------------------------
// Установка системы отсчета
void IPhasePoint::SetFrame(const IFrame& f)
{
    if (Frame != NULL) {
        if (Frame->IsEqual(f)) {
            return;
        }
        else {
            delete Frame;
            Frame = CreateIFrame(f.GetParameters());
        }
    }
    else {
        Frame = CreateIFrame(f.GetParameters());
    }
}

//------------------------------------------------------------------------------
// Инициализация
void IPhasePoint::Init(const IPhasePoint& src)
{
    SetFrame(*src.Frame);
    T = src.T;
}

////////////////////////////////////////////////////////////////////////////////
// Трехмерная фазовая точка 
// Инициализация
void PhasePoint3D::Init(const IFrame& f, const Vect3& c, const TimeJD& t)
{
    SetFrame(f);
    Coords = c;
    T = t;
}

//------------------------------------------------------------------------------
// Инициализация
void PhasePoint3D::Init(const PhasePoint3D& src)
{
    SetFrame(*src.Frame);
    Coords = src.Coords;
    T = src.T;
}

//------------------------------------------------------------------------------
// Конвертирование в указанную систему отсчета
void PhasePoint3D::Convert(const IFrame& f)
{
    if (Frame != NULL && Frame->IsEqual(f)) {
        return;
    }

    FrameConverter3D cnv;
    Vect3 src = Coords;

    Frame->Get3DConverter(f, T, cnv);
    cnv.Convert(src, Coords);
    SetFrame(f);
}

//------------------------------------------------------------------------------
// Присвоение точки с конвертированием в текущую систему отсчета
void PhasePoint3D::Assign(const PhasePoint3D& src)
{
    FrameConverter3D cnv;

    T = src.T;

    if (Frame != NULL && Frame->IsEqual(*src.Frame)) {
        Coords = src.Coords;
        return;
    }

    src.Frame->Get3DConverter(*Frame, src.T, cnv);
    cnv.Convert(src.Coords, Coords);
}

////////////////////////////////////////////////////////////////////////////////
// Шестимерная фазовая точка 
// Инициализация
void PhasePoint6D::Init(const IFrame& f, const Vect6& c, const TimeJD& t)
{
    SetFrame(f);
    CoordsVel = c;
    T = t;
}

//------------------------------------------------------------------------------
// Инициализация
void PhasePoint6D::Init(const IFrame& f, const Vect3& c, const Vect3& v, const TimeJD& t)
{
    SetFrame(f);
    CoordsVel.Init(c, v);
    T = t;
}

//------------------------------------------------------------------------------
// Инициализация
void PhasePoint6D::Init(const PhasePoint6D& src)
{
    SetFrame(*src.Frame);
    CoordsVel = src.CoordsVel;
    T = src.T;
}

//------------------------------------------------------------------------------
// Конвертирование в указанную систему отсчета
void PhasePoint6D::Convert(const IFrame& f)
{
    if (Frame != NULL && Frame->IsEqual(f)) {
        return;
    }

    FrameConverter6D cnv;
    Vect6 src = CoordsVel;

    Frame->Get6DConverter(f, T, cnv);
    cnv.Convert(src, CoordsVel);
    SetFrame(f);
}

//------------------------------------------------------------------------------
// Присвоение точки с конвертированием в текущую систему отсчета
void PhasePoint6D::Assign(const PhasePoint6D& src)
{
    FrameConverter6D cnv;

    T = src.T;

    if (Frame != NULL && Frame->IsEqual(*src.Frame)) {
        CoordsVel = src.CoordsVel;
        return;
    }

    src.Frame->Get6DConverter(*Frame, src.T, cnv);
    cnv.Convert(src.CoordsVel, CoordsVel);
}

////////////////////////////////////////////////////////////////////////////////
