///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Классы, описывающие планеты и тела Солнечной системы
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_PLANET_H
#define BNOTOOL_PLANET_H
///////////////////////////////////////////////////////////////////////////////
#include "IModel.h"
#include "ephemeris.h"
#include "SGravity.h"
#include "Atmosphere.h"

///////////////////////////////////////////////////////////////////////////////
// Вспомогательный класс, описывающий набор констант для планеты
///////////////////////////////////////////////////////////////////////////////
class PlanetConsts
{
public:
    TPlanetId PlanetId;
    const char* PlanetName;
    Ephemeride::EnBodyID PlanetEphId;
    double MeanRot;
    double Re;
    double Flat;
    double Mu;
};

///////////////////////////////////////////////////////////////////////////////
// Класс для планет Солнечной системы
///////////////////////////////////////////////////////////////////////////////
class SPlanet : public IPlanet
{
public:
    // Конструктор по умолчанию
    SPlanet()
    {
    };
    // Конструктор с параметрами планеты
    SPlanet(const PlanetConsts& pc, const char* eph_name = NULL)
    {
        Init(pc, eph_name);
    };

    // Инициализация параметрами планеты
    void Init(const PlanetConsts& pc, const char* eph_name = NULL);

    // имя планеты
    virtual const char* Name() const
    {
        return PName;
    };
    // числовой идентификатор
    virtual TPlanetId Id() const
    {
        return ID;
    };
    // Указатель на гравитационную модель
    virtual const IGravityModel* GetGravityModel() const
    {
        return &GrModel;
    };
    // Указатель на модель атмосферы
    virtual const IAtmosphereModel* GetAtmosphere() const
    {
        return NULL;
    };
    // Указатель на эфемериды
    virtual IEphemeride* GetEphemeride()
    {
        return &Eph;
    };

    virtual const IEphemeride* GetEphemeride() const
    {
        return &Eph;
    };

protected:
    const char* PName;
    TPlanetId ID;
    SGravityModel GrModel;
    PlanetEphemeride Eph;
};

///////////////////////////////////////////////////////////////////////////////
// Класс для Луны
///////////////////////////////////////////////////////////////////////////////
class MoonPlanet : public IPlanet
{
public:
    // Конструктор с параметрами планеты
    MoonPlanet(const char* eph_name = NULL)
    {
        Init(eph_name);
    };

    // Инициализация параметрами планеты
    void Init(const char* eph_name = NULL);

    // имя планеты
    virtual const char* Name() const
    {
        return "Moon";
    };
    // числовой идентификатор
    virtual TPlanetId Id() const
    {
        return plIdMoon;
    };
    // Указатель на гравитационную модель
    virtual const IGravityModel* GetGravityModel() const
    {
        return &GrModel;
    };
    // Указатель на модель атмосферы
    virtual const IAtmosphereModel* GetAtmosphere() const
    {
        return NULL;
    };
    // Указатель на эфемериды
    virtual IEphemeride* GetEphemeride()
    {
        return &Eph;
    };

    virtual const IEphemeride* GetEphemeride() const
    {
        return &Eph;
    };

protected:
    SGravityModel GrModel;
    MoonEphemeride Eph;
};

///////////////////////////////////////////////////////////////////////////////
// Класс для Солнца
///////////////////////////////////////////////////////////////////////////////
class SunPlanet : public IPlanet
{
public:
    // Конструктор по умолчанию
    SunPlanet();

    // имя планеты
    virtual const char* Name() const
    {
        return "Sun";
    };
    // числовой идентификатор
    virtual TPlanetId Id() const
    {
        return plIdSun;
    };
    // Указатель на гравитационную модель
    virtual const IGravityModel* GetGravityModel() const
    {
        return &GrModel;
    };
    // Указатель на модель атмосферы
    virtual const IAtmosphereModel* GetAtmosphere() const
    {
        return NULL;
    };
    // Указатель на эфемериды
    virtual IEphemeride* GetEphemeride()
    {
        return &Eph;
    };

    virtual const IEphemeride* GetEphemeride() const
    {
        return &Eph;
    };

protected:
    SGravityModel GrModel;
    SunEphemeride Eph;
};

///////////////////////////////////////////////////////////////////////////////
// Класс для Земли
///////////////////////////////////////////////////////////////////////////////
class EarthPlanet : public IPlanet
{
public:
    // Конструктор по умолчанию
    EarthPlanet();
    // Деструктор
    virtual ~EarthPlanet();

    // имя планеты
    virtual const char* Name() const
    {
        return "Earth";
    };
    // числовой идентификатор
    virtual TPlanetId Id() const
    {
        return plIdEarth;
    };
    // Указатель на гравитационную модель
    virtual const IGravityModel* GetGravityModel() const
    {
        return GrModel;
    };
    // Указатель на модель атмосферы
    virtual const IAtmosphereModel* GetAtmosphere() const
    {
        return AtmModel;
    };
    // Указатель на эфемериды
    virtual IEphemeride* GetEphemeride()
    {
        return &Eph;
    };

    virtual const IEphemeride* GetEphemeride() const
    {
        return &Eph;
    };

    // Установка гравитационной модели
    bool SetGravityModel(TGravityModelId m, const char* fname = NULL);
    // Установка модели атмосферы
    void SetAtmModel(TAtmosphereId m, ITableGeoSolarFlux* gsf = NULL, bool flag3h = false);

protected:
    IGravityModel* GrModel;
    IAtmosphereModel* AtmModel;
    EarthEphemeride Eph;
};

///////////////////////////////////////////////////////////////////////////////
#endif
