///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация РЛ-измерения 
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_RLMEASUREMENT_H
#define BNOTOOL_RLMEASUREMENT_H
///////////////////////////////////////////////////////////////////////////////
#include "IMeasurement.h"
#include "LocalRLFrame.h"

///////////////////////////////////////////////////////////////////////////////
class RLMeasurement : public IMeasurement
{
public:
    // "Измеренные" значения в виде 6-мерного вектора
    // 0 - дальность [km]
    // 1 - eps [rad]
    // 2 - gamma [rad] 
    // 3 - рад. скорость [km/s] 
    // 4 - скорость по eps [rad/s] 
    // 5 - скорость по gamma [rad/s] 
    double Vect[6];
    // СКО компонентов
    double Mse[6];
    // Признаки аномальности
    bool Ians[6];

    // Конструктор
    RLMeasurement();

    // Метод возвращает размерность вектора измерений
    virtual unsigned GetSize() const
    {
        return 6;
    };
    // Метод возвращает в объекте v вектор измерений
    virtual void GetRealVector(DVector& v) const;
    // Метод возвращает в объекте v вектор СКО измерений
    virtual void GetMSEVector(DVector& v) const;
    // Метод возвращает в переменной p весовую матрицу измерения,
    // которая является квадратной симметричной матрицей
    virtual void GetWeightMatrix(DSymmMatrix& p) const;

    // Метод осуществляет моделирование измерения для космического объекта,
    // движение которого описывается с помощью переменной trj.
    // Метод вычисляет следующие параметры:
    // t	-	время, которое соответствует положению космического объекта
    //      на орбите для текущего измерения (это время равно времени измерения,
    //      если не учитывается запаздывание сигнала)
    // fr	-	система отсчета, в которой задаются положение и скорости аппарата
    //      в момент измерения, а также матрица частных производных (см. далее параметры f и m)
    // v	-	указатель вектор моделируемого измерения. Если этот параметр равен 0,
    //      вектор моделируемых измерений не вычисляется
    // m	-	матрица производных размерности Nx6, где N - размерность вектора измерений
    //      (т.е. матрица состоит из N строк и 6 столбцов). В столбцах расположены
    //      производные по Rx Ry Rz Vx  Vy Vz, а в строках - компоненты вектора измерений.
    //      Если этот параметр равен 0, вектор моделируемых измерений не вычисляется
    // del	-	вектор невязок измеряемых параметров. Если этот параметр равен 0,
    //        вектор невязок не вычисляется
    virtual bool Simulate(int matrixsize, ITraject& trj, TimeJD& t, const IFrame& fr, Vect6* f = 0,
                          DVector* v = 0, DMatrix* m = 0, DVector* del = 0) const;
    // Вычисление невязок между измерениями
    // Если m1 или m2 == NULL, принмается текущее измерение
    virtual void Diff(DVector& d, const DVector* m1 = NULL, const DVector* m2 = NULL) const;
    // Создать копию объекта с помощью new()
    virtual IMeasurement* Clone() const;

    // Установка положения станции в Гринвиче
    void SetStationCoordsAz(const GeoCoordsPrec& p, double az);

    // Получение РЛ-координат по прямоугольным в местной
    void RLCoords(const Vect6& v, DVector& rl) const;
    // Получение РЛ-координат и матрицы частных производных по прямоугольным в местной
    void RLCoordsPDM(const Vect6& v, DVector& rl, DMatrix& pdm) const;
    // Получение прямоугольных координат по измерению
    bool GetXYZ(Vect6& out, const double* v = NULL) const;

    // Запрос признака аномальности компоненты
    virtual bool GetIan(unsigned k) const
    {
        return Ians[k];
    };
    // Запрос аномальности компоненты
    virtual void SetIan(unsigned k, bool f = true)
    {
        Ians[k] = f;
    };

protected:
    // Система отсчета станиции
    LocalRLFrame Frame;
};

///////////////////////////////////////////////////////////////////////////////
#endif
