///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Определение классов параметров систем отсчета
////////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_SFRAMEPARS_H
#define BNOTOOL_SFRAMEPARS_H
////////////////////////////////////////////////////////////////////////////////
#include "IFrames.h"
#include "Defines.h"
#include "GeoCoords.h"

////////////////////////////////////////////////////////////////////////////////
// Базовый класс для всех систем, которые определяются только их типом
////////////////////////////////////////////////////////////////////////////////
class SFrameParameters : public IFrameParameters
{
public:
    // Возвращение типа системы отсчета
    virtual TCoordSysId GetFrameType() const
    {
        return FrameType;
    };
    // Проверка на совпадение параметров
    virtual bool IsEqual(const IFrameParameters& c) const
    {
        return FrameType == c.GetFrameType();
    };

protected:
    TCoordSysId FrameType;
};

////////////////////////////////////////////////////////////////////////////////
// Параметры системы J2000 (геоцентрической)
////////////////////////////////////////////////////////////////////////////////
class J2000FrameParameters : public SFrameParameters
{
public:
    J2000FrameParameters()
    {
        FrameType = csIdJ2000;
    };
};

////////////////////////////////////////////////////////////////////////////////
// Параметры системы J2000 (гелиоцентрической)
////////////////////////////////////////////////////////////////////////////////
class J2000SFrameParameters : public SFrameParameters
{
public:
    J2000SFrameParameters()
    {
        FrameType = csIdJ2000S;
    };
};

////////////////////////////////////////////////////////////////////////////////
// Параметры системы B1950
////////////////////////////////////////////////////////////////////////////////
class B1950FrameParameters : public SFrameParameters
{
public:
    B1950FrameParameters()
    {
        FrameType = csIdB1950;
    };
};

////////////////////////////////////////////////////////////////////////////////
// Параметры системы B1900
////////////////////////////////////////////////////////////////////////////////
class B1900FrameParameters : public SFrameParameters
{
public:
    B1900FrameParameters()
    {
        FrameType = csIdB1900;
    };
};

////////////////////////////////////////////////////////////////////////////////
// Параметры системы вращающегося Гринвича
////////////////////////////////////////////////////////////////////////////////
class GreenwichFrameParameters : public SFrameParameters
{
public:
    GreenwichFrameParameters()
    {
        FrameType = csIdGCS;
    };
};

////////////////////////////////////////////////////////////////////////////////
// Базовый класс для всех систем с временем фиксации
////////////////////////////////////////////////////////////////////////////////
class TFrameParameters : public SFrameParameters
{
public:
    // Конструктор по умолчанию
    TFrameParameters();
    // Конструктор с указанием времени фиксации системы
    TFrameParameters(const TimeJD& t)
    {
        Init(t);
    };

    // Проверка на совпадение параметров
    virtual bool IsEqual(const IFrameParameters& c) const;

    // Инициализация с указанием времени фиксации
    void Init(const TimeJD& t);

    // Возврат времени фиксации
    const TimeJD& GetTime() const
    {
        TAssert(InitFlag);
        return TFix;
    };
    // Проверка инициализации
    bool IsInit() const
    {
        return InitFlag;
    };

protected:
    TimeJD TFix;
    bool InitFlag;
};

////////////////////////////////////////////////////////////////////////////////
// Параметры системы METE
////////////////////////////////////////////////////////////////////////////////
class METEFrameParameters : public TFrameParameters
{
public:
    METEFrameParameters() : TFrameParameters()
    {
        FrameType = csIdMETE;
    };

    METEFrameParameters(const TimeJD& t) : TFrameParameters(t)
    {
        FrameType = csIdMETE;
    };
};

////////////////////////////////////////////////////////////////////////////////
// Параметры "замороженной" Гринвичской системы 
////////////////////////////////////////////////////////////////////////////////
class FGreenwichFrameParameters : public TFrameParameters
{
public:
    FGreenwichFrameParameters() : TFrameParameters()
    {
        FrameType = csIdGFCS;
    };

    FGreenwichFrameParameters(const TimeJD& t) : TFrameParameters(t)
    {
        FrameType = csIdGFCS;
    };
};

////////////////////////////////////////////////////////////////////////////////
// Параметры "замороженной" Гринвичской системы 
////////////////////////////////////////////////////////////////////////////////
class ECIFrameParameters : public TFrameParameters
{
public:
    ECIFrameParameters() : TFrameParameters()
    {
        FrameType = csIdECI;
    };

    ECIFrameParameters(const TimeJD& t) : TFrameParameters(t)
    {
        FrameType = csIdECI;
    };
};

////////////////////////////////////////////////////////////////////////////////
// Параметры топоцентрической системы NEZ
////////////////////////////////////////////////////////////////////////////////
class NZEFrameParameters : public SFrameParameters
{
public:
    // Конструктор по умолчанию
    NZEFrameParameters()
    {
        Init();
        InitFlag = false;
    };
    // Конструктор с указанием точки расположения системы в Гринвиче
    NZEFrameParameters(const Vect3& p)
    {
        Init(p);
    };

    // Проверка на совпадение параметров
    virtual bool IsEqual(const IFrameParameters& c) const;

    // Инициализация с указанием точки расположения системы в Гринвиче
    void Init(const Vect3& p)
    {
        Init();
        Point = p;
        InitFlag = true;
    };

    // Возврат времени фиксации
    const Vect3& GetPoint() const
    {
        TAssert(InitFlag);
        return Point;
    };
    // Проверка инициализации
    bool IsInit() const
    {
        return InitFlag;
    };

protected:
    void Init()
    {
        FrameType = csIdNZE;
    };

    Vect3 Point;
    bool InitFlag;
};

////////////////////////////////////////////////////////////////////////////////
// Параметры орбитальной системы RNB
////////////////////////////////////////////////////////////////////////////////
class RNBFrameParameters : public SFrameParameters
{
public:
    // Конструктор по умолчанию
    RNBFrameParameters()
    {
        Init();
        InitFlag = false;
    };
    // Конструкторы с указанием точки расположения КА
    // и его вектора скорости в J2000
    RNBFrameParameters(const Vect3& p, const Vect3& v)
    {
        Init(p, v);
    };

    RNBFrameParameters(const Vect6& rv)
    {
        Init(rv);
    };

    // Проверка на совпадение параметров
    virtual bool IsEqual(const IFrameParameters& c) const;

    // Инициализация с указанием точки расположения КА
    // и его вектора скорости в J2000
    void Init(const Vect3& p, const Vect3& v);
    void Init(const Vect6& rv);

    // Возврат времени фиксации
    const Vect3& GetPoint() const
    {
        TAssert(InitFlag);
        return Point;
    };
    // Проверка инициализации
    bool IsInit() const
    {
        return InitFlag;
    };

    const TransMtx& GetMatrix() const
    {
        TAssert(InitFlag);
        return Matrix;
    };

protected:
    void Init()
    {
        FrameType = csIdRNB;
    };

    Vect3 Point; // Точка положения КА в J2000
    bool InitFlag;
    TransMtx Matrix; // Матрица перехода от J2000 к RNB
};

////////////////////////////////////////////////////////////////////////////////
// Параметры локальной системы J2000
////////////////////////////////////////////////////////////////////////////////
class LocalJ2000FrameParameters : public SFrameParameters
{
public:
    // Конструктор по умолчанию
    LocalJ2000FrameParameters()
    {
        Init();
        InitFlag = false;
    };
    // Конструкторы с указанием точки расположения КА
    LocalJ2000FrameParameters(const GeoCoordsPrec& p)
    {
        Init(p);
    };

    // Проверка на совпадение параметров
    virtual bool IsEqual(const IFrameParameters& c) const;

    // Инициализация с указанием точки расположения системы
    void Init(const GeoCoordsPrec& p)
    {
        GPoint = p;
        InitFlag = true;
    };

    // Проверка инициализации
    bool IsInit() const
    {
        return InitFlag;
    };

    // Возврат координат точки начала системы
    const GeoCoordsPrec& GetPoint() const
    {
        return GPoint;
    };

protected:
    void Init()
    {
        FrameType = csIdLocalJ2000;
    };

    GeoCoordsPrec GPoint; // Географические координаты начала отсчета
    bool InitFlag;
};

////////////////////////////////////////////////////////////////////////////////
// Параметры локальной радиолокационной системы 
////////////////////////////////////////////////////////////////////////////////
class LocalRLFrameParameters : public SFrameParameters
{
public:
    // Конструктор по умолчанию
    LocalRLFrameParameters()
    {
        Init();
        InitFlag = false;
    };
    // Конструкторы с указанием точки расположения КА
    // и его вектора скорости в J2000
    LocalRLFrameParameters(const GeoCoordsPrec& p, double az)
    {
        Init(p, az);
    };

    // Проверка на совпадение параметров
    virtual bool IsEqual(const IFrameParameters& c) const;

    // Инициализация с указанием точки расположения системы
    void Init(const GeoCoordsPrec& p, double az)
    {
        GPoint = p;
        Az = az;
        InitFlag = true;
    };

    // Проверка инициализации
    bool IsInit() const
    {
        return InitFlag;
    };

    // Возврат координат точки начала системы
    const GeoCoordsPrec& GetPoint() const
    {
        return GPoint;
    };
    // Возврат азимута 
    double GetAz() const
    {
        return Az;
    };

protected:
    void Init()
    {
        FrameType = csIdLocalRL;
    };

    // Географические координаты начала отсчета
    GeoCoordsPrec GPoint;
    // Азимут [рад]
    double Az;
    bool InitFlag;
};

////////////////////////////////////////////////////////////////////////////////
#endif
