///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Классы гравитационных моделей
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_SGRAVITY_H
#define BNOTOOL_SGRAVITY_H
///////////////////////////////////////////////////////////////////////////////
#include "IModel.h"

///////////////////////////////////////////////////////////////////////////////
// Гравитационная модель без гармоник (только с гравитационной постоянной)
///////////////////////////////////////////////////////////////////////////////
class SGravityModel : public IGravityModel
{
public:
    // Конструктор по умолчанию
    SGravityModel()
    {
    }

    // Конструктор с инициализацией параметров
    SGravityModel(double m, double r, double a)
    {
        mu = m;
        Re = r;
        alpha = a;
    }

    // Получение значений гармоник
    virtual double c(unsigned int /*n*/, unsigned int /*m*/) const
    {
        return 0;
    }

    virtual double d(unsigned int /*n*/, unsigned int /*m*/) const
    {
        return 0;
    }

    // Получение числа гармоник
    virtual unsigned int N() const
    {
        return 0;
    }

    virtual unsigned int M() const
    {
        return 0;
    };

    // Идентификатор модели
    virtual TGravityModelId GetID() const
    {
        return grIdMu;
    };
};

///////////////////////////////////////////////////////////////////////////////
// Гравитационная модель Земли PZ90
///////////////////////////////////////////////////////////////////////////////
class EarthPZ90GravityModel : public IGravityModel
{
public:
    // Конструктор по умолчанию
    EarthPZ90GravityModel();

    // Получение значений гармоник
    virtual double c(unsigned int n, unsigned int m) const
    {
        return CC[m][n];
    };

    virtual double d(unsigned int n, unsigned int m) const
    {
        return DD[m][n];
    };

    virtual const double* const* get_cc() const { return CC; }

    virtual const double* const* get_dd() const { return DD; }

    // Получение числа гармоник
    virtual unsigned int N() const
    {
        return 36;
    };

    virtual unsigned int M() const
    {
        return 36;
    };

    // Идентификатор модели
    virtual TGravityModelId GetID() const
    {
        return grIdPZ90;
    };

protected:
    // Гармоники
    double** CC;
    double** DD;
};

///////////////////////////////////////////////////////////////////////////////
// Гравитационная модель с указанным числом гармоник гармониками (для Земли)
///////////////////////////////////////////////////////////////////////////////
class EarthPZ902GravityModel : public IGravityModel
{
public:
    // Конструктор по умолчанию
    EarthPZ902GravityModel();

    double** getCC() const
    {
        return CC;
    }

    double** getDD() const
    {
        return DD;
    }

    // Получение значений гармоник
    virtual double c(unsigned int n, unsigned int m) const
    {
        return CC[m][n];
    };

    virtual double d(unsigned int n, unsigned int m) const
    {
        return DD[m][n];
    };

    virtual const double* const* get_cc() const { return CC; }

    virtual const double* const* get_dd() const { return DD; }

    // Получение числа гармоник
    virtual unsigned int N() const
    {
        return 36;
    };

    virtual unsigned int M() const
    {
        return 36;
    };

    // Идентификатор модели
    virtual TGravityModelId GetID() const
    {
        return grIdPZ902s;
    };

protected:
    // Гармоники
    double** CC;
    double** DD;
};

///////////////////////////////////////////////////////////////////////////////
// Гравитационная модель с указанным числом гармоник гармониками (для Земли)
///////////////////////////////////////////////////////////////////////////////
class GravityModel : public IGravityModel
{
public:
    // Конструктор, создающий модель копированием ограниченного
    // числа гармоник из другой модели
    GravityModel(TGravityModelId Id, unsigned int N, unsigned int M,
                 const IGravityModel* cd);
    // Конструктор по умолчанию
    GravityModel();

    // Деструктор
    virtual ~GravityModel()
    {
        freeAll();
    };

    // Получение значений гармоник
    virtual double c(unsigned int n, unsigned int m) const
    {
        return CC[m][n];
    };

    virtual double d(unsigned int n, unsigned int m) const
    {
        return DD[m][n];
    };

    // Установка значений гармоник
    double c(unsigned int n, unsigned int m, double value)
    {
        return (CC[m][n] = value);
    };

    double d(unsigned int n, unsigned int m, double value)
    {
        return (DD[m][n] = value);
    };

    virtual const double* const* get_cc() const { return CC; }

    virtual const double* const* get_dd() const { return DD; }

    // Получение числа гармоник
    virtual unsigned int N() const
    {
        return N_;
    };

    virtual unsigned int M() const
    {
        return M_;
    };

    // Идентификатор модели
    virtual TGravityModelId GetID() const
    {
        return Id;
    };

    // Инициализация модели копированием ограниченного
    // числа гармоник из другой модели
    void Init(unsigned int N, unsigned int M, const IGravityModel* c_d);
    // Инициализация пустой модели
    void initEmpty(TGravityModelId Id);
    // обнуление всех коэффициентов
    void null();
    // Печать
    void print(char* comment, FILE* out = stdout) const;

    // Идентификатор модели
    TGravityModelId Id;

protected:
    // Гармоники
    double** CC;
    double** DD;
    // Размерность гармоник
    unsigned int N_;
    unsigned int M_;
    // Флаг инициализации
    bool isInit;
    // Очистка модели
    void freeAll();
    // Установка числа гармоник
    void SetSize(unsigned int N, unsigned int M);
};

///////////////////////////////////////////////////////////////////////////////
#endif
