///////////////////////////////////////////////////////////////////////////////
// Библиотека программных инструментов
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Упорядоченные списки указателей
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOLS_SLIST_H
#define BNOTOOLS_SLIST_H
///////////////////////////////////////////////////////////////////////////////
#include "Defines.h"
///////////////////////////////////////////////////////////////////////////////
// тип функции сравнения.
// Должна возвращать <0, если первый элемент меньше,\n
//                   >0 - если больше, 0 = если совп.\n
//                  (возвращать <0, если первый элемент перед вторым, >0 - если после)
typedef int (*TCompareFunction)(const void* pFromUser, const void* pFromList);
///////////////////////////////////////////////////////////////////////////////
class TSListCompare;

///////////////////////////////////////////////////////////////////////////////
// Базовый класс - упорядоченный список указателей
///////////////////////////////////////////////////////////////////////////////
class TSList
{
public:
    // конструктор 
    TSList();
    // деструктор
    virtual ~TSList();
    // установить длину буфера списка
    void SetLength(int pos);
    // добавить в конец списка или в место, определяемое функцией упорядочения
    void Add(const void* p);
    // вставить в последнюю найденную позицию            
    void Insert(const void* p);
    // вставить в указанную позицию
    bool Insert(const void* p, int pos);
    // удалить элемент номер pos
    bool Delete(int pos);
    // заменить элемент в указанной позиции
    bool Replace(const void* p, int pos);
    // получить n-ый элемент списка 
    void* operator[](int Ind) const;
    // функция для сравнения элементов для упорядочения списка
    void SetSortFunction(TCompareFunction fun);
    // функция для сравнения элементов для упорядочения списка
    void SetSortFunction(TSListCompare* fun);
    // поиск элемента в массиве. Индекс найденного элемента - в FoundPos
    bool Find(const void* p);
    // поиск элемента в массиве с указанием функции сравнения
    bool Find(const void* p, TCompareFunction fun);
    // поиск элемента в массиве с указанием функции сравнения
    bool Find(const void* p, TSListCompare& fun);
    // очистить список
    void Clear();
    // получить указатель на массив указателей
    const void** GetPtr() const { return List; };
    // возвращает длину списка
    int Len() const { return LastPos + 1; };
    // позиция после последнего поиска   
    int FoundPos;
protected:
    bool LowInsert(const void* p, int pos);
    // указатель на массив указателей
    const void** List;
    // сохранение указателя от пред.поиска
    const void* PPrev;
    // длина имеющегося массива        
    int Length;
    // указатель на функцию упорядочения
    TSListCompare* PFunSort;
    // указатель на функцию упорядочения
    TCompareFunction FunSort;
    // признак - последний раз поиск шел по процедуре Sort
    bool LastFindSort;
    // признак - последний поиск был успешным
    bool LastFindOk;
    // последняя занятая позиция 
    int LastPos;
private:
    // Запрет прямого копирования списка
    TSList(TSList& l)
    {
        TAssert(false);
    };

    void operator=(TSList& l)
    {
        TAssert(false);
    };
    // поиск элемента в массиве. bySort=true : по сортированному списку
    bool Find(const void* p, int& ind, TCompareFunction fun, bool bySort = true);
    // поиск элемента в массиве. bySort=true : по сортированному списку
    bool Find(const void* p, int& ind, TSListCompare& fun, bool bySort = true);
};

///////////////////////////////////////////////////////////////////////////////
/// класс сравнения элементов упорядоченного списка
///////////////////////////////////////////////////////////////////////////////
//  Функция compare должна возвращать -1 если первый элемент меньше,
//                                     1 если первый элемент больше,
//                                     0 если совпадают.
///////////////////////////////////////////////////////////////////////////////
class TSListCompare
{
public:
    // Деструктор
    virtual ~TSListCompare()
    {
    };
    virtual int compare(const void* pFromUser, const void* pFromList) = 0;
};

///////////////////////////////////////////////////////////////////////////////
/// Шаблон сравнения элементов упорядоченного списка
///////////////////////////////////////////////////////////////////////////////
template <class objType>
class TSListCompareDef : public TSListCompare
{
public:
    TSListCompareDef() : TSListCompare()
    {
    };

    virtual ~TSListCompareDef()
    {
    };

    virtual int compare(const void* p1, const void* p2)
    {
        return compare((objType *)p1, (objType *)p2);
    };
    virtual int compare(const objType* p1, const objType* p2) = 0;
};

///////////////////////////////////////////////////////////////////////////////
// Шаблон упорядоченного списка указателей
///////////////////////////////////////////////////////////////////////////////
template <class objType, bool _clearOnDelete>
class TSListDef : public TSList
{
private:
    TSListDef(TSListDef& l)
    {
        TAssert(false);
    };

    void operator=(TSListDef& l)
    {
        TAssert(false);
    };
    bool clearOnDelete;
public:
    // конструктор 
    TSListDef() : TSList()
    {
        clearOnDelete = _clearOnDelete;
    };
    // конструктор с указанием признака авт. уничтожения
    TSListDef(bool clearOnDelete) : TSList()
    {
        this->clearOnDelete = clearOnDelete;
    };
    // деструктор
    ~TSListDef()
    {
        if (clearOnDelete)
            for (int i = LastPos; i >= 0; i--)
                delete ((objType *)(List[i]));
    };
    // очистить список
    void Clear(bool forceDelete = false)
    {
        if (clearOnDelete || forceDelete)
            for (int i = LastPos; i >= 0; i--)
                delete ((objType *)(List[i]));
        TSList::Clear();
    };
    // удалить элемент номер pos
    bool Delete(int pos, bool forceDelete = false)
    {
        if (clearOnDelete || forceDelete)
            delete ((objType *)(List[pos]));
        return TSList::Delete(pos);
    };
    // заменить элемент в указанной позиции
    bool Replace(const void* p, int pos, bool forceDelete = false)
    {
        if (FunSort || PFunSort)
            return false;
        if (clearOnDelete || forceDelete)
            delete ((objType *)(List[pos]));
        return TSList::Replace(p, pos);
    };
    // получить n-ый элемент списка 
    objType* operator[](int Ind) const
    {
        if (Ind < 0 || Ind > LastPos)
            return NULL;
        return (objType *)(List[Ind]);
    };
    // получить указатель на массив указателей
    const objType** GetPtr() const
    {
        return (const objType **)List;
    };
    // установить длину буфера списка
    void SetLength(int newLen)
    {
        if (clearOnDelete) {
            while (LastPos >= newLen)
                delete ((objType *)(List[LastPos--]));
        };
        TSList::SetLength(newLen);
    };
    // функция для сравнения элементов для упорядочения списка
    void SetSortFunction(TSListCompareDef<objType>* fun)
    {
        TSList::SetSortFunction((TSListCompare *)fun);
    };
    // функция для сравнения элементов для упорядочения списка
    void SetSortFunction(int fun(const objType* pFromUser, const objType* pFromList))
    {
        TSList::SetSortFunction((TCompareFunction)&fun);
    };
    // поиск элемента в массиве с указанием функции сравнения
    bool Find(const void* p, TSListCompareDef<objType>* fun)
    {
        return TSList::Find(p, (TSListCompare &)fun);
    }

    // поиск элемента в массиве с указанием функции сравнения
    bool Find(const void* p, int* fun(const objType* pFromUser, const objType* pFromList))
    {
        return TSList::Find(p, (TCompareFunction)fun);
    }
};

#endif
