///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация класса земной статической атмосферы
///////////////////////////////////////////////////////////////////////////////
#include "Atmosphere.h"
#include "IAstroBase.h"

///////////////////////////////////////////////////////////////////////////////
// Структура представления атмосферных коэффициентов
struct StaticAtmCoeff
{
    double h;
    double A;
    double k1;
    double k2;
};

///////////////////////////////////////////////////////////////////////////////
// Коэффициенты статической атмосферы
static struct StaticAtmCoeff StaticAtmKoeff[8] =
{
    {0.0, 0.1250, -0.2639e-8, 0.7825e-4},
    {20.0, 0.9091e-2, 0.4407e-9, 0.1638e-3},
    {60.0, 0.2630e-4, -0.2560e-8, 0.5905e-4},
    {100.0, 0.4141e-7, 0.1469e-8, 0.1787e-3},
    {150.0, 0.2173e-9, 0.8004e-10, 0.3734e-4},
    {300.0, 0.4861e-11, 0.7111e-11, 0.1547e-4},
    {600.0, 0.8904e-13, 0.1831e-11, 0.9275e-5},
    {900.0, 0.6497e-14, 0.0, 0.9540e-5},
};
///////////////////////////////////////////////////////////////////////////////
// Получение коэффициентов статической атмосферы
void StaticAtmosphere::GetStaticAtmKoeff(double altitude, double* hi, double* k1,
                                         double* k2, double* a)
{
    int i;
    double hkm = altitude * 1.e3;

    if (hkm < 0.0) {
        hkm = 0.0;
        i = 0;
    }
    else if (hkm < 100.0) {
        i = ((int)(hkm + 20)) / 40;
    }
    else if (hkm < 150.0) {
        i = 3;
    }
    else {
        i = ((int)hkm) / 300 + 4;
        if (i > 7)
            i = 7;
    }

    struct StaticAtmCoeff& p = StaticAtmKoeff[i];

    if (hi != NULL)
        *hi = p.h * 1.e-3; // тыс. км
    if (k1 != NULL)
        *k1 = p.k1;
    if (k2 != NULL)
        *k2 = p.k2;
    if (a != NULL)
        *a = p.A;
}

//-----------------------------------------------------------------------------
// Расчет плотности в зависимости от высоты
double StaticAtmosphere::Density(double h)
{
    double hi, k1, k2, a;
    GetStaticAtmKoeff(h, &hi, &k1, &k2, &a);
    h = (h * 1.e3 - hi) * 1000.0; // высота нужна в метрах
    return 9.80665 * a * exp(h * (k1 * h - k2));
}

//-----------------------------------------------------------------------------
// Расчет плотности в зависимости от положения
double StaticAtmosphere::Density(const TimeJD& t, const Vect3& v,
                                 const Vect3& vSun) const
{
    double rv, h, re, alph;
    rv = v.Len();
    h = v.z / rv;
    re = GetAstroBase()->GetEarthRadius();
    alph = GetAstroBase()->GetEarthOblatness();
    h = (rv - re * (1.0 - alph * h * h)) * 1000.0;
    return (h > 1500.0 ? 0.0 : Density(h));
}

///////////////////////////////////////////////////////////////////////////////
