///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация TLE-измерения 
///////////////////////////////////////////////////////////////////////////////
#include "TLEMeasurement.h"
#include "METEFrame.h"
#include "Kepler.h"
#include "IAstroBase.h"
#include "Angles.h"
#include <iostream>
#include "TimeFunctions.h"

///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
//-----------------------------------------------------------------------------
TLEMeasurement::TLEMeasurement() : IMeasurement()
{
    Type = mtTLE;
    int i;
    for (i = 0; i < 6; i++)
        Ians[i] = false;
}

//-----------------------------------------------------------------------------
// Создать копию объекта с помощью new()
//-----------------------------------------------------------------------------
IMeasurement* TLEMeasurement::Clone() const
{
    TLEMeasurement* m = new TLEMeasurement;
    *m = *this;
    return m;
}

//-----------------------------------------------------------------------------
// Метод возвращает в объекте v вектор измерений
//-----------------------------------------------------------------------------
void TLEMeasurement::GetRealVector(DVector& v) const
{
    v.Init(6U);

    int i;
    for (i = 0; i < 6; i++)
        v[i] = Vect[i];
}

//-----------------------------------------------------------------------------
// Метод возвращает в объекте v вектор СКО измерений
//-----------------------------------------------------------------------------
void TLEMeasurement::GetMSEVector(DVector& v) const
{
    v.Init(6U);

    int i;
    for (i = 0; i < 6; i++)
        v[i] = 1.0 / PVect[i];
}

//-----------------------------------------------------------------------------
// Установка TLE-элементов с помощью набора TLE
//-----------------------------------------------------------------------------
void TLEMeasurement::SetByTLESet(const TLESet& tle)
{
    TimeTLE tt(tle.EpochYr, tle.EpochDays);
    MTime = tt;

    Vect[0] = tle.No;
    Vect[1] = tle.Ecco;
    Vect[2] = tle.Omegao;
    Vect[3] = tle.Inclo;
    Vect[4] = tle.Argpo;
    Vect[5] = tle.Mo;

    Normalize(Vect);
}

//-----------------------------------------------------------------------------
// Приведение угловых величин к диапазону 0-2Pi
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void TLEMeasurement::GetWeightMatrix(DSymmMatrix& p) const
{
    p.Init(6U);
    p.Init(0.0);

    int i;
    for (i = 0; i < 6; i++) {
        //		if (Ians[i])
        //			p(i, i) = 0.0;
        //		else
        p(i, i) = PVect[i] * PVect[i];
    }
}

//-----------------------------------------------------------------------------
// оПХБЕДЕМХЕ СЦКНБШУ БЕКХВХМ Й ДХЮОЮГНМС 0-2Pi
//-----------------------------------------------------------------------------
void TLEMeasurement::Normalize(double* v)
{
    if (v[1] < 0)
        v[1] = 0.0;

    v[2] = Angle2Pi(v[2]);

    while (v[3] >= Pi)
        v[3] -= Pi2;
    while (v[3] < 0)
        v[3] += Pi2;

    v[4] = Angle2Pi(v[4]);
    v[5] = Angle2Pi(v[5]);
}

//-----------------------------------------------------------------------------
// оПХБЕДЕМХЕ СЦКНБШУ БЕКХВХМ Й ДХЮОЮГНМС 0-2Pi ДКЪ БЯО. ЩКЕЛЕМРНБ
//-----------------------------------------------------------------------------
void TLEMeasurement::NormalizeAux(double* v)
{
    /*  v[2] = Angle2Pi(v[2]);

        while (v[3] >= Pi)
            v[3] -= Pi2;
        while (v[3] < 0)
            v[3] += Pi2;

        v[5] = Angle2Pi(v[5]);
    */
    //	 while (v[5] >= Pi)
    //	    v[5] -= Pi2;
    //	  while (v[5] < 0)
    //	    v[5] += Pi2;

    v[5] = Angle2Pi(v[5]);
}

//-----------------------------------------------------------------------------
// Пересчет вектора состояния в прямоугольной системе
// в TLE-элементы в первом приближении
//-----------------------------------------------------------------------------
const double xpdotp = 229.1831180523293;

void TLEMeasurement::CreateTLEVector(const Vect6& rv, double* q)
{
    Kepler kp;
    kp.SetXYZ(rv);

    double mu = GetAstroBase()->GetEarthMu();

    q[0] = 86.4 / Pi2 * sqrt(mu / (kp.A * kp.A * kp.A)) / xpdotp;
    q[1] = kp.GetEcc();
    q[2] = kp.Omega;
    q[3] = acos(kp.CosI);
    q[4] = kp.GetArgP();

    // Интервал между прохождением узла и перицентром

    double dtu = kp.GetTimeP();

    // Интервал от текущей точки до восх. узла

    double dto = kp.GetTimeFromAN();

    double dtp = dto - dtu;

    q[5] = dtp * sqrt(mu / (kp.A * kp.A * kp.A));

    Normalize(q);
}

//-----------------------------------------------------------------------------
// Метод осуществляет моделирование измерения для космического объекта,
// движение которого описывается с помощью переменной trj.
// Метод вычисляет следующие параметры:
// t	-	время, которое соответствует положению космического объекта
//      на орбите для текущего измерения (это время равно времени измерения,
//      если не учитывается запаздывание сигнала)
// fr	-	система отсчета, в которой задаются положение и скорости аппарата
//      в момент измерения, а также матрица частных производных (см. далее параметры f и m)
// v	-	указатель вектор моделируемого измерения. Если этот параметр равен 0,
//      вектор моделируемых измерений не вычисляется
// m	-	матрица производных размерности Nx6, где N - размерность вектора измерений
//      (т.е. матрица состоит из N строк и 6 столбцов). В столбцах расположены
//      производные по Rx Ry Rz Vx  Vy Vz, а в строках - компоненты вектора измерений.
//      Если этот параметр равен 0, вектор моделируемых измерений не вычисляется
// del	-	вектор невязок измеряемых параметров. Если этот параметр равен 0,
//        вектор невязок не вычисляется
//-----------------------------------------------------------------------------
bool TLEMeasurement::Simulate(int matrixSize, ITraject& trj, TimeJD& t, const IFrame& fr, Vect6* f, DVector* v, DMatrix* m, DVector* del) const
{
    Vect6 pv;

    if (v == 0 && m == 0 && del == 0)
        return true;

    if (f == 0 && (v != 0 || m != 0 || del != 0))
        f = &pv;

    if (!trj.Predict(MTime, fr, *f))
        return false;
    t = MTime;

    METEFrame mframe(MTime);
    PhasePoint6D src, mete;

    src.Init(fr, *f, t);
    mete.SetFrame(mframe);
    mete.Assign(src);

    double q0[6], q1[6], q[6];
    double aux0[6], aux1[6], aux[6];

    CreateTLEVector(mete.CoordsVel, q0);

    ToAux(q0, aux0);

    TLESet tle;
    CreateTLESet(q0, tle);
    MSGPPredict prg;
    prg.Init(tle);
    if (!prg.Run(MTime))
        return false;
    PhasePoint6D out;
    prg.GetDestPoint(out);

    CreateTLEVector(out.CoordsVel, q1);

    // PrintTLE("TLE из вектора - 2", q1);

    ToAux(q1, aux1);

    int i;
    for (i = 0; i < 6; i++)
        aux[i] = 2 * aux0[i] - aux1[i];
    NormalizeAux(aux);
    /* FromAux(aux, q);
       if (q[1] < 0.01)
       {
           if (!CorrectM(mete.CoordsVel, q))
               return false;
           ToAux(q, aux);
       }
   */


    //ToAux(q0, aux);

    int k = 0;
    DSqrMatrix dxdq(6);
    // double eps[] = {1.e-3 / xpdotp, 1.e-6, 1.e-3 * deg2rad, 1.e-3 * deg2rad, 1.e-6, 1.e-4 * deg2rad};
    //  double cnt[] = {1.e-3 / xpdotp, 1.e-6, 1.e-3 * deg2rad, 1.e-3 * deg2rad, 1.e-6, 1.e-3 * deg2rad};
    // double mdq[] = {1.e0 / xpdotp,  1.e-1, 1.e0  * deg2rad, 1.e0  * deg2rad, 1.e-1, 5.e-1 * deg2rad};
    double eps[] = {1.e-6, 1.e-6, 1.e-6, 1.e-6, 1.e-6, 1.e-6};
    double cnt[] = {1.e-6, 1.e-6, 1.e-6, 1.e-6, 1.e-6, 1.e-6};
    double mdq[] = {1.e-1, 1.e-1, 1.e-1, 1.e-1, 1.e-1, 1.e-1};
    DVector dxm(6), daux(6);

    for (;;) {
        TLESet tle;
        FromAux(aux, q);
        // PrintTLE("TLE", q);
        CreateTLESet(q, tle);
        MSGPPredict prg;
        prg.Init(tle);
        if (!prg.Run(MTime))
            return false;

        PhasePoint6D dst;
        //  DVector dxm(6), daux(6);
        prg.GetDestPoint(dst);
        // PrintVector("Вектор", dst.CoordsVel);
        for (i = 0; i < 6; i++)
            dxm[i] = mete.CoordsVel[i] - dst.CoordsVel[i];

        if (!CreatePDM(aux, eps, dxdq))
            return false;

        if (!dxdq.Inverse())
            return false;

        Mult(daux, dxdq, dxm);

        for (i = 0; i < 6; i++) {
            if (fabs(daux[i]) > cnt[i])
                break;
        }
        if (i == 6)
            break;

        double alph = 1.0;
        for (i = 0; i < 6; i++) {
            double a = fabs(mdq[i] / daux[i]);
            if (a < alph)
                alph = a;
        }

        for (i = 0; i < 6; i++)
            aux[i] += daux[i] * alph;
        NormalizeAux(aux);
        k++;
        if (k > 100)
            return false;
    }

    if (v) {
        v->Init(6U);
        for (i = 0; i < 6; i++)
            (*v)[i] = q[i];
    }

    if (m) {
        int j;
        TAssert(fr.IsGeocentric() && fr.IsInertial());
        m->Init(6, 6);
        FrameConverter3D res;
        DSqrMatrix cnv(6);
        cnv.Init(0.0);
        fr.Get3DConverter(mframe, MTime, res);
        for (i = 0; i < 3; i++) {
            for (j = 0; j < 3; j++) {
                cnv(i, j) = res.Rotation(i, j);
                cnv(i + 3, j + 3) = res.Rotation(i, j);
            }
        }
        dxdq.MulRight(cnv);
        /*
            DSqrMatrix dqaux(6);

            dqaux.Init(0.0);

            dqaux(0, 0) = 1.0;
            dqaux(2, 2) = 1.0;
            dqaux(3, 3) = 1.0;
            dqaux(5, 5) = 1.0;
            dqaux(1, 1) = aux[1] / q[1];
            dqaux(1, 4) = aux[4] / q[1];
            dqaux(4, 1) = +aux[4] / (q[1] * q[1]);
            dqaux(4, 4) = -aux[1] / (q[1] * q[1]);
        */

        for (i = 0; i < 6; i++) {
            for (j = 0; j < 6; j++)
                (*m)(i, j) = dxdq(i, j);
        }
    }

    if (del) {
        /*   DVector qv(6);
               for (i = 0; i < 6; i++)
                   qv(i) = q[i];
               Diff(*del, NULL, &qv);
           */
        (*del) = daux;
    }

    return true;
}

void TLEMeasurement::CalcTleXYZResiduals(const double* TLEElements1, const double* TLEElements2, TimeJD date, Vect6& residuals) const
{
    Vect6 xyzMETE1, xyzMETE2;

    CalcTleXYZResiduals(TLEElements1, date, xyzMETE1);
    CalcTleXYZResiduals(TLEElements2, date, xyzMETE2);

    double dx = xyzMETE2.x - xyzMETE1.x;
    double dy = xyzMETE2.y - xyzMETE1.y;
    double dz = xyzMETE2.z - xyzMETE1.z;
    double dvx = xyzMETE2.vx - xyzMETE1.vx;
    double dvy = xyzMETE2.vy - xyzMETE1.vy;
    double dvz = xyzMETE2.vz - xyzMETE1.vz;

    residuals = Vect6(dx, dy, dz, dvx, dvy, dvz);
}

void TLEMeasurement::CalcTleXYZResiduals(const double* TLEElements, TimeJD date, Vect6& xyzMETE) const
{
    TLESet tle;
    CreateTLESet(TLEElements, tle);
    MSGPPredict prg;
    prg.Init(tle);
    prg.Run(date);
    PhasePoint6D dst;
    prg.GetDestPoint(dst);
    xyzMETE = dst.CoordsVel;
}

//-----------------------------------------------------------------------------
// Создание набора TLE-элементов по текущему вектору
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void TLEMeasurement::Diff(DVector& d, const DVector* m1, const DVector* m2) const
{
    d.Init(6U);

    DVector s1, s2;
    int i;

    if (m1 != NULL)
        s1 = *m1;
    else {
        s1.Init(6U);
        for (i = 0; i < 6; i++)
            s1(i) = Vect[i];
    }

    if (m2 != NULL)
        s2 = *m2;
    else {
        s2.Init(6U);
        for (i = 0; i < 6; i++)
            s2(i) = Vect[i];
    }

    d(0) = s1(0) - s2(0);
    d(1) = s1(1) - s2(1);
    d(2) = AngleDiff(s1(2), s2(2));
    d(3) = AngleDiff(s1(3), s2(3));
    d(4) = AngleDiff(s1(4), s2(4));
    d(5) = AngleDiff(s1(5), s2(5));
}

//-----------------------------------------------------------------------------
// яНГДЮМХЕ МЮАНПЮ TLE-ЩКЕЛЕМРНБ ОН РЕЙСЫЕЛС БЕЙРНПС
//-----------------------------------------------------------------------------
void TLEMeasurement::CreateTLESet(const double* q, TLESet& tle) const
{
    TimeTLE tt = MTime;

    tle.EpochYr = tt.Year;
    tle.EpochDays = tt.Day;

    tle.No = q[0];
    tle.Ecco = q[1];
    tle.Omegao = q[2];
    tle.Inclo = q[3];
    tle.Argpo = q[4];
    tle.Mo = q[5];

    tle.BStar = 0;
    tle.NDot = 0;
    tle.NDDot = 0;
    tle.Satnum = 0;
    tle.ElementNum = 0;

    tle.Set();
}

//-----------------------------------------------------------------------------
// Вычисление матрицы частных производных при MSGP-прогнозе
//-----------------------------------------------------------------------------
bool TLEMeasurement::CreatePDM(const double* aux, const double* eps, DSqrMatrix& pdm) const
{
    Vect6 xp, xm;
    double qc[6];
    double auxc[6], daux[6];
    int i, j;

    for (i = 0; i < 6; i++) {

        for (j = 0; j < 6; j++) {
            auxc[j] = aux[j];
            if (i == j)
                auxc[j] += eps[j];
        }
        NormalizeAux(auxc);
        FromAux(auxc, qc);
        TLESet tle;
        CreateTLESet(qc, tle);
        MSGPPredict prg;
        prg.Init(tle);
        if (!prg.Run(MTime))
            return false;
        PhasePoint6D out;
        prg.GetDestPoint(out);
        xp = out.CoordsVel;

        for (j = 0; j < 6; j++) {
            auxc[j] = aux[j];
            if (i == j) {
                auxc[j] -= eps[j];
                daux[j] = 2.0 * eps[j];
            }

        }
        FromAux(auxc, qc);
        CreateTLESet(qc, tle);
        prg.Init(tle);
        if (!prg.Run(MTime))
            return false;
        prg.GetDestPoint(out);
        xm = out.CoordsVel;

        for (j = 0; j < 6; j++)
            pdm(j, i) = (xp[j] - xm[j]) / daux[i];
    }

    return true;
}

//-----------------------------------------------------------------------------
// оЕПЕБНД БН БЯОНЛНЦЮРЕКЭМШЕ ЩКЕЛЕМРШ
//-----------------------------------------------------------------------------
void TLEMeasurement::ToAux(const double* src, double* aux)
{
    /* aux[0] = src[0];
       aux[1] = src[1] * sin(src[4]);
       aux[2] = src[2];
       aux[3] = src[3];
       aux[4] = src[1] * cos(src[4]);
     aux[5] = src[5];*/

    IAstroBase* ab = GetAstroBase();
    double Mu = ab->GetEarthMu();
    double a1 = 86.4 / (Pi2 * src[0] * xpdotp);
    aux[0] = pow(Mu * a1 * a1, 1.0 / 3.0);
    aux[1] = cos(src[2]) * sin(src[3] / 2.0);
    aux[2] = sin(src[2]) * sin(src[3] / 2.0);
    aux[3] = src[1] * cos(src[2] + src[4]);
    aux[4] = src[1] * sin(src[2] + src[4]);
    aux[5] = src[5] + src[2] + src[4];
}

//-----------------------------------------------------------------------------
// оЕПЕБНД ХГ БЯОНЛНЦЮРЕКЭМШУ ЩКЕЛЕМРНБ Б ХЯУНДМШЕ
//-----------------------------------------------------------------------------
void TLEMeasurement::FromAux(const double* aux, double* src)
{
    /*  src[0] = aux[0];
        src[1] = sqrt(aux[1] * aux[1] + aux[4] * aux[4]);
        src[2] = aux[2];
        src[3] = aux[3];
        src[4] = atan2(aux[1] / src[1], aux[4] / src[1]);
        if (src[4] < 0)
            src[4] += Pi2;
        src[5] = aux[5];
      */
    IAstroBase* ab = GetAstroBase();
    double Mu = ab->GetEarthMu();
    src[0] = 86.4 / Pi2 * sqrt(Mu / (aux[0] * aux[0] * aux[0])) / xpdotp;
    src[1] = sqrt(aux[3] * aux[3] + aux[4] * aux[4]);
    src[3] = acos(1 - 2 * (aux[2] * aux[2] + aux[1] * aux[1])); // РЅР°РєР»РѕРЅРµРЅРёРµ
    src[2] = atan2(aux[2] / sin(src[3] / 2.0), aux[1] / sin(src[3] / 2.0));

    src[4] = atan2(aux[4] / src[1], aux[3] / src[1]) - src[2];
    if (src[4] < 0)
        src[4] += Pi2;
    src[5] = aux[5] - src[4] - src[2];

    Normalize(src);
}

//-----------------------------------------------------------------------------
// нРКЮДЙЮ - ОЕВЮРЭ ЩКЕЛЕМРНБ
//-----------------------------------------------------------------------------
void TLEMeasurement::PrintTLE(const char* str, const double* tle) const
{
    printf("%s\n", str);
    printf("n  = %f\n", tle[0]);
    printf("e  = %f\n", tle[1]);
    printf("OM = %f\n", tle[2] * rad2deg);
    printf("i  = %f\n", tle[3] * rad2deg);
    printf("om = %f\n", tle[4] * rad2deg);
    printf("M  = %f\n", tle[5] * rad2deg);
    printf("\n");
}

//-----------------------------------------------------------------------------
// нРКЮДЙЮ - ОЕВЮРЭ ОПЪЛНСЦНКЭМНЦН БЕЙРНПЮ
//-----------------------------------------------------------------------------
void TLEMeasurement::PrintVector(const char* str, const Vect6& v) const
{
    printf("%s\n", str);
    printf("x  = %f\n", v[0]);
    printf("y  = %f\n", v[1]);
    printf("z  = %f\n", v[2]);
    printf("vx = %f\n", v[3]);
    printf("vy = %f\n", v[4]);
    printf("vz = %f\n", v[5]);
    printf("\n");
}

//-----------------------------------------------------------------------------
// йНППЕЙЖХЪ M ОПХ ЛЮКШУ ЩЙЯЖЕМРПХЯЕРЮУ
//-----------------------------------------------------------------------------
bool TLEMeasurement::CorrectM(const Vect6& rv, double* q) const
{
    TLESet tle;
    CreateTLESet(q, tle);
    MSGPPredict prg;
    prg.Init(tle);
    if (!prg.Run(MTime))
        return false;

    PhasePoint6D dst;
    prg.GetDestPoint(dst);

    Vect6 drv;
    int i, j, k = -1, l = -20;

    double f0 = 0.0;

    //printf("%f %f\n", q[5] * rad2deg, f0);

    for (i = 0; i < 36; i++) {
        double qc[6];
        memcpy(qc, q, 6 * sizeof(double));
        qc[5] = i * 10.0 * deg2rad;
        CreateTLESet(qc, tle);
        prg.Init(tle);
        if (!prg.Run(MTime))
            return false;
        PhasePoint6D dst;
        prg.GetDestPoint(dst);
        for (j = 0; j < 6; j++)
            drv[j] = rv[j] - dst.CoordsVel[j];
        double fc = drv.Len() + drv.LenV();
        //printf("%f %f\n", qc[5] * rad2deg, fc);
        if (fc < f0 || f0 == 0.0) {
            f0 = fc;
            k = i;
        }
    }

    if (k == -1)
        return true;

    for (i = -10; i < 10; i++) {
        double qc[6];
        memcpy(qc, q, 6 * sizeof(double));
        qc[5] = (k * 10 + i) * deg2rad;
        CreateTLESet(qc, tle);
        prg.Init(tle);
        if (!prg.Run(MTime))
            return false;
        PhasePoint6D dst;
        prg.GetDestPoint(dst);
        for (j = 0; j < 6; j++)
            drv[j] = rv[j] - dst.CoordsVel[j];
        double fc = drv.Len() + drv.LenV();
        //printf("%f %f\n", qc[5] * rad2deg, fc);
        if (fc < f0) {
            f0 = fc;
            l = i;
        }
    }

    if (l == -20)
        q[5] = k * 10 * deg2rad;
    else
        q[5] = (k * 10 + l) * deg2rad;

    if (q[1] < 1.e-4) {
        k = -20;
        for (i = -10; i < 10; i++) {
            double qc[6];
            memcpy(qc, q, 6 * sizeof(double));
            qc[5] += i * 0.1 * deg2rad;
            CreateTLESet(qc, tle);
            prg.Init(tle);
            if (!prg.Run(MTime))
                return false;
            PhasePoint6D dst;
            prg.GetDestPoint(dst);
            for (j = 0; j < 6; j++)
                drv[j] = rv[j] - dst.CoordsVel[j];
            double fc = drv.Len() + drv.LenV();
            // printf("%f %f\n", qc[5] * rad2deg, fc);
            if (fc < f0) {
                f0 = fc;
                k = i;
            }
        }
        if (k != -20)
            q[5] += k * 0.1 * deg2rad;
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////
