///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация TLE-измерения 
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_TLEMEASUREMENT_H
#define BNOTOOL_TLEMEASUREMENT_H
///////////////////////////////////////////////////////////////////////////////
#include "IMeasurement.h"
#include "msgp.h"

///////////////////////////////////////////////////////////////////////////////
class TLEMeasurement : public IMeasurement
{
public:
    // "Измеренные" значения в виде 6-мерного вектора
    // 0 - n (среднее движение) [rad/min]
    // 1 - e (эксцентриситет)
    // 2 - долгота восходящего узла [rad] 
    // 3 - наклонение [rad] 
    // 4 - аргумент перигея [rad] 
    // 5 - Средняя аномалия [rad] 
    double Vect[6];
    // Веса компонентов
    double PVect[6];

    bool Ians[6];

    // Конструктор
    TLEMeasurement();

    // Метод возвращает размерность вектора измерений
    virtual unsigned GetSize() const
    {
        return 6;
    };
    // Метод возвращает время измерения
    virtual const TimeJD& GetTime() const
    {
        return MTime;
    };
    // Метод возвращает в объекте v вектор измерений
    virtual void GetRealVector(DVector& v) const;
    // Метод возвращает в объекте v вектор СКО измерений
    virtual void GetMSEVector(DVector& v) const;
    // Метод возвращает идентификатор наблюдательной станции,
    // к которой относится измерение.
    // Метод возвращает 0, если станция не определена.
    virtual TSiteId GetStationID() const
    {
        return TSiteId();
    };
    // Метод устанавливает идентификатор наблюдательной станции,
    // к которой относится измерение
    virtual void SetStationID(unsigned id)
    {
    };

    // Установка TLE-элементов с помощью набора TLE
    void SetByTLESet(const TLESet& tle);

    // Метод осуществляет моделирование измерения для космического объекта,
    // движение которого описывается с помощью переменной trj.
    // Метод вычисляет следующие параметры:
    // t	-	время, которое соответствует положению космического объекта
    //      на орбите для текущего измерения (это время равно времени измерения,
    //      если не учитывается запаздывание сигнала)
    // fr	-	система отсчета, в которой задаются положение и скорости аппарата
    //      в момент измерения, а также матрица частных производных (см. далее параметры f и m)
    // v	-	указатель вектор моделируемого измерения. Если этот параметр равен 0,
    //      вектор моделируемых измерений не вычисляется
    // m	-	матрица производных размерности Nx6, где N - размерность вектора измерений
    //      (т.е. матрица состоит из N строк и 6 столбцов). В столбцах расположены
    //      производные по Rx Ry Rz Vx  Vy Vz, а в строках - компоненты вектора измерений.
    //      Если этот параметр равен 0, вектор моделируемых измерений не вычисляется
    // del	-	вектор невязок измеряемых параметров. Если этот параметр равен 0,
    //        вектор невязок не вычисляется
    virtual bool Simulate(int matrixSize, ITraject& trj, TimeJD& t, const IFrame& fr, Vect6* f = 0,
                          DVector* v = 0, DMatrix* m = 0, DVector* del = 0) const;

    // Если m1 или m2 == NULL, принмается текущее измерение
    virtual void Diff(DVector& d, const DVector* m1 = NULL, const DVector* m2 = NULL) const;
    // Создать копию объекта с помощью new()
    virtual IMeasurement* Clone() const;

    // Запрос признака аномальности компоненты
    virtual bool GetIan(unsigned k) const
    {
        return Ians[k];
    };
    // Запрос аномальности компоненты
    virtual void SetIan(unsigned k, bool f = true)
    {
        Ians[k] = f;
    };
    // Метод возвращает в переменной p весовую матрицу измерения,
    // которая является квадратной симметричной матрицей
    virtual void GetWeightMatrix(DSymmMatrix& p) const;

    // Пересчет вектора состояния в прямоугольной системе
    // в TLE-элементы в первом приближении
    static void CreateTLEVector(const Vect6& rv, double* q);

    // Перевод во вспомогательные элементы
    static void ToAux(const double* src, double* aux);

    // Перевод из вспомогательных элементов в исходные
    static void FromAux(const double* aux, double* src);

    // Пересчет вектора состояния в прямоугольной системе
    static void NormalizeAux(double* v);
protected:
    // Создание набора TLE-элементов по текущему вектору
    void CreateTLESet(const double* q, TLESet& tle) const;
    // Вычисление матрицы частных производных при MSGP-прогнозе
    bool CreatePDM(const double* q, const double* eps, DSqrMatrix& pdm) const;
    // Приведение угловых величин к диапазону 0-2Pi
    static void Normalize(double* v);

    // Отладка - печать элементов
    void PrintTLE(const char* str, const double* tle) const;
    // Отладка - печать прямоугольного вектора
    void PrintVector(const char* str, const Vect6& v) const;
    // Коррекция M при малых эксцентрисетах
    bool CorrectM(const Vect6& rv, double* q) const;

private:
    void CalcTleXYZResiduals(const double* TLEElements, TimeJD date, Vect6& xyzMETE) const;
    void CalcTleXYZResiduals(const double* TLEElements1, const double* TLEElements2, TimeJD date, Vect6& residuals) const;
};

///////////////////////////////////////////////////////////////////////////////
#endif
