///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Представление времени и временных шкал
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_TM_H
#define BNOTOOL_TM_H
///////////////////////////////////////////////////////////////////////////////
#include "TMCF.h"
#include "TMUtil.h"
#include "TMStr.h"
//-----------------------------------------------------------------------------
#include <stdio.h>
#include <math.h>
#include <time.h>
///////////////////////////////////////////////////////////////////////////////
// Предопределение классов представления времени
///////////////////////////////////////////////////////////////////////////////
class TimeGD;
class TimeTLE;
class TimeJD;
///////////////////////////////////////////////////////////////////////////////
// Функции конвертирования представления времени
///////////////////////////////////////////////////////////////////////////////
void convertTime(const TimeTLE& tle, TimeJD& jd);
void convertTime(const TimeJD& jd, TimeTLE& tle);
void convertTime(const TimeGD& grd, TimeJD& jd);
void convertTime(const TimeJD& jd, TimeGD& grd);

///////////////////////////////////////////////////////////////////////////////
// Класс представления времени в виде Юлианской даты
///////////////////////////////////////////////////////////////////////////////
class TimeJD
{
public:
    // Конструкторы
    TimeJD()
    {
        Init(0, 0.0);
    };

    TimeJD(int d, double f)
    {
        Init(d, f);
    };

    TimeJD(const TimeJD& dt)
    {
        Init(dt);
    };

    TimeJD(const TimeGD& dt)
    {
        Init(dt);
    };

    TimeJD(const TimeTLE& dt)
    {
        Init(dt);
    };

    TimeJD(const char* str)
    {
        Init(str);
    };

    // Функции инициализации
    TimeJD& Init(int d, double f = 0)
    {
        Days = d;
        Fraction = f;
        normalize();
        return *this;
    };

    TimeJD& Init(const TimeJD& dt)
    {
        *this = dt;
        return *this;
    };

    TimeJD& Init(const TimeGD& dt)
    {
        convertTime(dt, *this);
        return *this;
    };

    TimeJD& Init(const TimeTLE& dt)
    {
        convertTime(dt, *this);
        return *this;
    };
    TimeJD& Init(const char* str);

    // Получение составляющих даты
    int GetDays() const
    {
        return Days;
    };

    double GetFraction() const
    {
        return Fraction;
    };

    // Операторы присваивания
    TimeJD& operator=(const TimeTLE& t)
    {
        Init(t);
        return *this;
    };

    TimeJD& operator=(const TimeGD& t)
    {
        Init(t);
        return *this;
    };

    // Арифметические операторы
    TimeJD& operator+=(const TmDay& ival)
    {
        Fraction += ival.GetDouble();
        normalize();
        return *this;
    };

    TimeJD& operator-=(const TmDay& ival)
    {
        Fraction -= ival.GetDouble();
        normalize();
        return *this;
    };

    TimeJD& operator+=(double ival)
    {
        Fraction += ival;
        normalize();
        return *this;
    };

    TimeJD& operator-=(double ival)
    {
        Fraction -= ival;
        normalize();
        return *this;
    };

    // Бинарные операторы
    friend TimeJD operator-(const TimeJD& dt, const TmDay& t)
    {
        TimeJD dt1(dt);
        dt1 -= t;
        return dt1;
    };
    friend TmDay operator-(const TimeJD& dt, const TimeJD& dt2);

    // Операторы сравнения
    friend bool operator<(const TimeJD& t1, const TimeJD& t2);

    friend bool operator>(const TimeJD& t1, const TimeJD& t2)
    {
        return (t2 < t1);
    };
    friend bool operator==(const TimeJD& t1, const TimeJD& t2);

    friend bool operator!=(const TimeJD& t1, const TimeJD& t2)
    {
        return !(t1 == t2);
    };

    friend bool operator<=(const TimeJD& t1, const TimeJD& t2)
    {
        return (t1 < t2) || (t1 == t2);
    };

    friend bool operator>=(const TimeJD& t1, const TimeJD& t2)
    {
        return (t1 > t2) || (t1 == t2);
    };

    // Методы печати и приведения к символьному виду
    void print(char* name = "", FILE* out = stdout) const;
    void printScale(FILE* out = stdout) const;
    void Str(char* str) const;

private:
    // Число полных суток
    int Days;
    // Дробная часть даты
    double Fraction;
    // Нормализация представления даты
    void normalize()
    {
        normalizeJD(Days, Fraction);
    };
};

///////////////////////////////////////////////////////////////////////////////
// Класс представления времени в виде Григорианской даты
///////////////////////////////////////////////////////////////////////////////
class TimeGD
{
public:
    // Конструкторы
    TimeGD()
    {
        Init(time(NULL));
    };

    TimeGD(int Day_, int Month_, int Year_, int Hour_ = 0, int Min_ = 0,
           int Sec_ = 0, double Fraction_ = 0.0)
    {
        Init(Day_, Month_, Year_, Hour_, Min_, Sec_, Fraction_);
    };

    TimeGD(const TimeJD& jd)
    {
        Init(jd);
    };

    TimeGD(const char* str)
    {
        Init(str);
    };

    TimeGD(const char* dt, const char* tm)
    {
        Init(dt, tm);
    };

    TimeGD(time_t t)
    {
        Init(t);
    };

    // Публичные переменные
    int Year;
    unsigned short int Month;
    unsigned short int Day;
    unsigned short int Hour;
    unsigned short int Min;
    unsigned short int Sec;
    double Fraction;

    // Методы инициализации
    TimeGD& Init(int Day_ = 1, int Month_ = 1, int Year_ = 1, int Hour_ = 0, int Min_ = 0,
                 int Sec_ = 0, double Fraction_ = 0.0)
    {
        Year = Year_;
        Month = Month_;
        Day = Day_;
        Hour = Hour_;
        Min = Min_;
        Sec = Sec_;
        Fraction = Fraction_;
        return *this;
    };

    TimeGD& Init(const TimeJD& jd)
    {
        convertTime(jd, *this);
        return *this;
    };

    TimeGD& Init(const char* str)
    {
        fromStrDate(str);
        return *this;
    };

    TimeGD& Init(const char* dt, const char* tm)
    {
        Init(dt, tm);
        return *this;
    };
    TimeGD& Init(time_t t);


    // Методы печати
    void print(char* name = "", FILE* out = stdout) const;
    void printScale(FILE* out = stdout) const;
    void StrDate(char* str) const;
    void StrTime(char* str) const;
    void StrDTime(char* str) const;

private:
    void fromStrDate(const char* str);
    void fromStrTime(const char* str);
};

///////////////////////////////////////////////////////////////////////////////
// Класс представления времени в виде Григорианской даты
///////////////////////////////////////////////////////////////////////////////
class TimeTLE
{
public:
    // Конструкторы
    TimeTLE(int Year_ = 1, double Day_ = 1.0)
    {
        Init(Year_, Day_);
    };

    TimeTLE(const char* str)
    {
        Init(str);
    };

    TimeTLE(const TimeJD& jd)
    {
        Init(jd);
    };

    // Публичные переменные
    int Year;
    double Day;

    // Методы инициализации
    TimeTLE& Init(int Year_ = 1, double Day_ = 1.0)
    {
        Year = Year_;
        Day = Day_;
        return *this;
    };

    TimeTLE& Init(const char* str)
    {
        fromStr(str);
        return *this;
    };

    TimeTLE& Init(const TimeJD& jd)
    {
        convertTime(jd, *this);
        return *this;
    };

    // Методы печати
    void print(char* name = "", FILE* out = stdout) const;
    void printScale(FILE* out = stdout) const;
    void Str(char* str) const;

protected:
    void fromStr(const char* str);
};

///////////////////////////////////////////////////////////////////////////////
#endif
