///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Представление времени и временных шкал
///////////////////////////////////////////////////////////////////////////////
#include "Defines.h"
#include "TM.h"
//-----------------------------------------------------------------------------
#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
///////////////////////////////////////////////////////////////////////////////
// Проверка формата юлианской даты
///////////////////////////////////////////////////////////////////////////////
bool IsJulianDate(const char* str)
{
    const char* p;
    unsigned k;

    for (p = str; isspace(*p); p++);

    k = 0;
    while (*p != '.') {
        if (!isdigit(*p))
            return false;
        p++;
        k++;
    }

    if (k < 7)
        return false;
    p++;
    k = 0;

    k = 0;
    while (*p != '\0') {
        if (!isdigit(*p))
            return false;
        p++;
        k++;
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////
// Проверка формата григорианской даты
///////////////////////////////////////////////////////////////////////////////
bool IsGrigorianDate(const char* str)
{
    const char* p;
    unsigned i, m, d;

    for (p = str; isspace(*p); p++);

    // Year

    for (i = 0; i < 4; i++, p++) {
        if (*p == '/')
            break;
        if (!isdigit(*p))
            return false;
    }

    if (*p != '/')
        return false;
    p++;


    for (i = 0, m = 0; i < 2; i++, p++) {
        if (*p == '/')
            break;
        if (!isdigit(*p))
            return false;
        m = m * 10 + (*p) - '0';
    }

    if (*p != '/')
        return false;
    if (m < 1 || m > 12)
        return false;
    p++;


    for (i = 0, d = 0; i < 2; i++, p++) {
        if (*p == '\0')
            break;
        if (!isdigit(*p))
            return false;
        d = d * 10 + (*p) - '0';
    }

    if (*p != '\0')
        return false;
    if (d < 1 && d > 31)
        return false;

    return true;
}

///////////////////////////////////////////////////////////////////////////////
// Проверка формата времени суток
///////////////////////////////////////////////////////////////////////////////
bool IsDTime(const char* str)
{
    const char *p, *ps;
    unsigned i, h, m, k = 0;
    double s;

    for (p = str; isspace(*p); p++);

    for (i = 0, h = 0; i < 2; i++, p++) {
        if (*p == ':' || *p == '\0')
            break;
        if (!isdigit(*p))
            return false;
        h = h * 10 + (*p) - '0';
    }

    if (*p != ':' && *p != '\0')
        return false;
    if (h > 23)
        return false;

    if (*p == '\0')
        return true;

    p++;

    for (i = 0, m = 0; i < 2; i++, p++) {
        if (*p == ':' || *p == '\0')
            break;
        if (!isdigit(*p))
            return false;
        m = m * 10 + (*p) - '0';
    }

    if (*p != ':' && *p != '\0')
        return false;
    if (m > 59)
        return false;

    if (*p == '\0')
        return true;

    p++;

    ps = p;
    while (*p != '.' && *p != '\0') {
        if (!isdigit(*p))
            return false;
        p++;
    }

    k = 0;

    if (*p != '\0') {
        p++;
        while (*p != '\0') {
            if (!isdigit(*p))
                return false;
            p++;
            k++;
        }
    }

    s = atof(ps);
    if (s < 0 || s > 60.0)
        return false;

    return true;
}

///////////////////////////////////////////////////////////////////////////////
// Проверка формата даты в формате TLE
///////////////////////////////////////////////////////////////////////////////
bool IsTLEDate(const char* str)
{
    const char *p, *ps;
    unsigned i, k;
    double d;

    for (p = str; isspace(*p); p++);

    for (i = 0; i < 4; i++, p++) {
        if (!isdigit(*p))
            return false;
    }

    if (*p != '+')
        return false;
    p++;


    ps = p;
    k = 0;
    while (*p != '.' && *p != '\0') {
        if (!isdigit(*p))
            return false;
        p++;
        k++;
    }

    if (k > 3)
        return false;

    if (*p != '\0') {
        p++;
        k = 0;
        while (*p != '.' && *p != '\0') {
            if (!isdigit(*p))
                return false;
            p++;
            k++;
        }
        if (k > 8)
            return false;
    }

    d = atof(ps);
    if (d < 1 || d >= 367.0)
        return false;

    return true;
}

///////////////////////////////////////////////////////////////////////////////
// Метод, конструирующий юлианскую дату из строки
///////////////////////////////////////////////////////////////////////////////
TimeJD& TimeJD::Init(const char* str)
{
    long double d;

    TAssert(IsJulianDate(str));

    sscanf(str, "%Lf", &d);
    Days = (int)floor(d);
    Fraction = (double)(d - Days);

    return *this;
}

///////////////////////////////////////////////////////////////////////////////
// Метод, конструирующий юлианскую дату в формате TLE из строки
///////////////////////////////////////////////////////////////////////////////
void TimeTLE::fromStr(const char* str)
{
    TAssert(IsTLEDate(str));

    Year = 0;

    const char* p;

    for (p = str; isspace(*p); p++);

    while (*p != '+') {
        Year = Year * 10 + (*p) - '0';
        p++;
    }

    p++;

    Day = 0.0;
    while (*p != '.' && *p != '\0') {
        Day = Day * 10.0 + (*p) - '0';
        p++;
    }

    if (*p == '\0')
        return;

    long double fr = 0.1;
    while (*p != '.' && *p != '\0') {
        Day += ((*p) - '0') * fr;
        fr *= 0.1;
        p++;
    }
}

///////////////////////////////////////////////////////////////////////////////
// Метод, конструирующий григорианскую дату из строки даты
///////////////////////////////////////////////////////////////////////////////
void TimeGD::fromStrDate(const char* str)
{
    TAssert(IsGrigorianDate(str));

    Init();

    const char* p;

    for (p = str; isspace(*p); p++);

    Year = 0;
    while (*p != '/') {
        Year = Year * 10 + (*p) - '0';
        p++;
    }

    p++;
    Month = 0;
    while (*p != '/') {
        Month = Month * 10 + (*p) - '0';
        p++;
    }

    p++;
    Day = 0;
    while (*p != '\0') {
        Day = Day * 10 + (*p) - '0';
        p++;
    }
}

///////////////////////////////////////////////////////////////////////////////
// Метод, конструирующий юлианскую дату из строки с датой и временем
///////////////////////////////////////////////////////////////////////////////
void TimeGD::fromStrTime(const char* str)
{
    TAssert(IsDTime(str));

    const char* p;

    for (p = str; isspace(*p); p++);

    Hour = 0;
    while (*p != ':' && *p != '\0') {
        Hour = Hour * 10 + (*p) - '0';
        p++;
    }

    if (*p == '\0')
        return;

    p++;
    Min = 0;
    while (*p != ':' && *p != '\0') {
        Min = Min * 10 + (*p) - '0';
        p++;
    }

    if (*p == '\0')
        return;

    p++;
    Sec = 0;
    while (*p != '.' && *p != '\0') {
        Sec = Sec * 10 + (*p) - '0';
        p++;
    }

    if (*p == '\0')
        return;

    double fr = 0.1;
    p++;
    Fraction = 0.0;
    while (*p != '\0') {
        Fraction += fr * ((*p) - '0');
        fr *= 0.1;
        p++;
    }
}

///////////////////////////////////////////////////////////////////////////////
// Переменные, устанавливающие точность при печати
///////////////////////////////////////////////////////////////////////////////
unsigned short int JLDATE_STR_LEN;
unsigned short int GRTIME_STR_LEN;
unsigned short int TLEDATE_STR_LEN;
unsigned short int timePrintPrecision = (SetTimePrecision(3), 3);
///////////////////////////////////////////////////////////////////////////////
// Функция возвращает точность вывода времени
// (количество знаков для вывода долей секунд)
///////////////////////////////////////////////////////////////////////////////
unsigned int GetTimePrecision()
{
    return timePrintPrecision;
};
///////////////////////////////////////////////////////////////////////////////
// Функция устанавливает точность вывода времени
// (количество знаков для вывода долей секунд)
///////////////////////////////////////////////////////////////////////////////
void SetTimePrecision(unsigned int timePrec_)
{
    timePrintPrecision = timePrec_;
    JLDATE_STR_LEN = timePrintPrecision + 14;
    GRTIME_STR_LEN = timePrintPrecision + 20;
    TLEDATE_STR_LEN = timePrintPrecision + 15;
};
///////////////////////////////////////////////////////////////////////////////
// Методы печати для юлианской даты
///////////////////////////////////////////////////////////////////////////////
void TimeJD::print(char* name, FILE* out) const
{
    fprintf(out, name);
    char* buffer = new char[JLDATE_STR_LEN + 1];
    Str(buffer);
    fprintf(out, buffer);
    delete[] buffer;
};
//-----------------------------------------------------------------------------
void TimeJD::Str(char* str) const
{
    char* buffer = new char[JLDATE_STR_LEN - 9 + 4];
    sprintf(buffer, "%*.*f", JLDATE_STR_LEN - 9 + 3, JLDATE_STR_LEN - 9, Fraction);
    double fp = atof(buffer);
    sprintf(str, "%d", Days + (int)floor(fp));
    char* p = strchr(buffer, '.');
    strcat(str, p);
    delete[] buffer;
};
///////////////////////////////////////////////////////////////////////////////
// Методы печати для юлианской даты в формате TLE
///////////////////////////////////////////////////////////////////////////////
void TimeTLE::print(char* name, FILE* out) const
{
    TimeJD jd(*this);
    double d = pow(10., -(TLEDATE_STR_LEN - 10)) * 0.5;
    jd += d;
    TimeTLE t(jd);
    fprintf(out, "%s%04d+%*.*f", name, t.Year, TLEDATE_STR_LEN - 6, TLEDATE_STR_LEN - 10, t.Day);
};
//-----------------------------------------------------------------------------
void TimeTLE::Str(char* str) const
{
    TimeJD jd(*this);
    double d = pow(10., -(TLEDATE_STR_LEN - 10)) * 0.5;
    jd += d;
    TimeTLE t(jd);
    sprintf(str, "%04d+%*.*f", t.Year, TLEDATE_STR_LEN - 6, TLEDATE_STR_LEN - 10, t.Day);
};
///////////////////////////////////////////////////////////////////////////////
// Методы печати для григорианской даты
///////////////////////////////////////////////////////////////////////////////
TimeGD& TimeGD::Init(time_t t)
{
    struct tm* tt = gmtime(&t);
    Fraction = 0;
    Sec = tt->tm_sec;
    Min = tt->tm_min;
    Hour = tt->tm_hour;
    Day = tt->tm_mday;
    Month = tt->tm_mon + 1;
    Year = tt->tm_year + 1900;
    return *this;
}

//-----------------------------------------------------------------------------
void TimeGD::print(char* name, FILE* out) const
{
    char* str = new char[GRTIME_STR_LEN + 1];
    StrDTime(str);
    fprintf(out, "%s%s", name, str);
    delete[] str;
};
//-----------------------------------------------------------------------------
void TimeGD::StrDate(char* str) const
{
    sprintf(str, "%04d/%02d/%02d", Year, Month, Day);
};
//-----------------------------------------------------------------------------
void TimeGD::StrTime(char* str) const
{
    double d;
    unsigned z;

    if (timePrintPrecision <= 0) {
        z = 0;
        d = 0.5;
    }
    else {
        z = timePrintPrecision;
        double w = z;
        d = pow(10, -w) * 0.5;
    }

    double fts = Hour * 3600 + Min * 60 + Sec + Fraction + d;
    int ifts = (int)floor(fts);
    int h = ifts / 3600;
    int m = (ifts - h * 3600) / 60;
    int s = ifts - h * 3600 - m * 60;
    while (h >= 24)
        h -= 24;

    sprintf(str, "%02d:%02d:%02d", h, m, s);

    if (timePrintPrecision == 0)
        return;

    strcat(str, ".");

    char format[10];
    char* txt = new char[timePrintPrecision + 1];
    sprintf(format, "%%0%dd", z);
    unsigned i, k = 1;
    for (i = 0; i < z; i++)
        k *= 10;
    unsigned ff = (unsigned)floor(Fraction * k);
    sprintf(txt, format, ff);
    strcat(str, txt);
    delete[] txt;

    return;
};
//-----------------------------------------------------------------------------
void TimeGD::StrDTime(char* str) const
{
    unsigned z;
    double d;

    if (timePrintPrecision <= 0) {
        z = 0;
        d = 0.5;
    }
    else {
        z = timePrintPrecision;
        double w = z;
        d = pow(10, -w) * 0.5;
    }

    TimeJD tp(*this);
    tp += d / 86400.;
    TimeGD gp(tp);

    sprintf(str, "%04d/%02d/%02d %02d:%02d:%02d",
            gp.Year, gp.Month, gp.Day, gp.Hour, gp.Min, gp.Sec);

    if (z == 0)
        return;

    strcat(str, ".");

    char format[10];
    char* txt = new char[timePrintPrecision + 1];
    sprintf(format, "%%0%dd", z);
    unsigned i, k = 1;
    for (i = 0; i < z; i++)
        k *= 10;
    unsigned ff = (unsigned)floor(gp.Fraction * k);
    sprintf(txt, format, ff);
    strcat(str, txt);
    delete[] txt;

    return;
};
///////////////////////////////////////////////////////////////////////////////
