///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация пакета измерений в виде связанного списка
///////////////////////////////////////////////////////////////////////////////
#include "TMeasurementPacket.h"
///////////////////////////////////////////////////////////////////////////////
// Реализацич методов одиночного измерения пакета
///////////////////////////////////////////////////////////////////////////////
// Конструктор
//-----------------------------------------------------------------------------
IMeasurementRef::IMeasurementRef()
{
    Ref = NULL;
    ManualFlag = false;
    AutoFlag = false;
    DeleteFlag = false;
}

//-----------------------------------------------------------------------------
// Деструктор
//-----------------------------------------------------------------------------
IMeasurementRef::~IMeasurementRef()
{
    if (Ref != NULL)
        delete Ref;
    Ref = NULL;
}

//-----------------------------------------------------------------------------
// Является ли измерение активным
//-----------------------------------------------------------------------------
bool IMeasurementRef::IsActive() const
{
    return !ManualFlag && !AutoFlag && !DeleteFlag;
}

bool IMeasurementRef::IsDeleted() const
{
    return DeleteFlag;
}

///////////////////////////////////////////////////////////////////////////////
// Класс, хранящий пакет измерений в форме связанного списка
///////////////////////////////////////////////////////////////////////////////
// Метод возвращает true, если пакет содержит хотя бы одно активное измерение
// и возвращает false в противном случае
//-----------------------------------------------------------------------------
bool TMeasurementPacket::IsActive() const
{
    IMeasurementEntry* f;

    for (f = first(); f != NULL; f = f->next()) {
        if (f->IsActive())
            return true;
    }

    return false;
}

//-----------------------------------------------------------------------------
// Метод возвращает число активных измерений в пакете
//-----------------------------------------------------------------------------
unsigned TMeasurementPacket::GetActiveSize() const
{
    IMeasurementEntry* e;
    unsigned ret = 0;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->IsActive())
            ret++;
    }

    return ret;
}

//-----------------------------------------------------------------------------
// Метод делает текущим первое активное измерение пакета.
// Если активных измерений нет, то текущее измерение становится неопределенным
//-----------------------------------------------------------------------------
void TMeasurementPacket::SetFirstActive()
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->IsActive()) {
            Current = e;
            return;
        }
    }

    Current = NULL;
}

//-----------------------------------------------------------------------------
// Метод делает текущим последнее активное измерение пакета.
// Если активных измерений нет, то текущее измерение становится неопределенным
//-----------------------------------------------------------------------------
void TMeasurementPacket::SetLastActive()
{
    IMeasurementEntry* e;

    for (e = last(); e != NULL; e = e->prev()) {
        if (e->IsActive()) {
            Current = e;
            return;
        }
    }

    Current = NULL;
}

//-----------------------------------------------------------------------------
// Метод делает текущим первое измерение пакета, время которого больше или равно
// указанного времени t.
// Если таких измерений нет, то текущее измерение становится неопределенным
//-----------------------------------------------------------------------------
void TMeasurementPacket::SetByTime(const TimeJD& t)
{
    IMeasurementEntry* e;
    TimeJD ta, tc;
    ta = t;

    for (e = first(); e != NULL; e = e->next()) {
        tc = e->Ref->GetTime();
        if (tc >= ta) {
            Current = e;
            return;
        }
    }

    Current = NULL;
}

//-----------------------------------------------------------------------------
// Метод делает текущим первое активное измерение пакета, время которого больше
// или равно указанного времени t.
// Если таких измерений нет, то текущее измерение становится неопределенным
//-----------------------------------------------------------------------------
void TMeasurementPacket::SetByTimeActive(const TimeJD& t)
{
    IMeasurementEntry* e;
    TimeJD ta, tc;
    ta = t;

    for (e = first(); e != NULL; e = e->next()) {
        tc = e->Ref->GetTime();
        if (tc >= ta)
            break;
    }

    while (e != NULL && !e->IsActive())
        e = e->next();

    Current = e;
}

//-----------------------------------------------------------------------------
// Метод позволяет явно установить текущее измерение пакета по ссылке m.
// Если указанное измерение отсутствует в пакете, текущее измерение становится 
// неопределенным.
//-----------------------------------------------------------------------------
void TMeasurementPacket::SetCurrent(const IMeasurement& m)
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->Ref == &m) {
            Current = e;
            return;
        }
    }

    Current = NULL;
}

//-----------------------------------------------------------------------------
// Метод позволяет явно установить текущее измерение пакета по идентификатору 
// измерения id.
// Если указанное измерение отсутствует в пакете, текущее измерение становится 
// неопределенным.
//-----------------------------------------------------------------------------
void TMeasurementPacket::SetCurrent(TMeasId id)
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->Ref->GetID() == id) {
            Current = e;
            return;
        }
    }

    Current = NULL;
}

//-----------------------------------------------------------------------------
// Метод делает текущим измерение, которое располагается непосредственно сразу 
// за ранее установленным текущим измерением.
// Если текущим было последнее измерение пакета,
// то после работы метода текущее измерение становится неопределенным.
//-----------------------------------------------------------------------------
void TMeasurementPacket::GoNext()
{
    if (Current == NULL)
        return;
    Current = Current->next();
}

//-----------------------------------------------------------------------------
// Метод делает текущим измерение, которое располагается непосредственно перед 
// ранее установленным текущим измерением.
// Если текущим было первое измерение пакета,
// то после работы метода текущее измерение становится неопределенным.
//-----------------------------------------------------------------------------
void TMeasurementPacket::GoPrev()
{
    if (Current == NULL)
        return;
    Current = Current->prev();
}

//-----------------------------------------------------------------------------
// Метод делает текущим первое активное измерение, которое расположено в пакете 
// после ранее установленного текущего измерения.
// Если таких измерений нет, то после работы метода текущее измерение 
// становится неопределенным.
//-----------------------------------------------------------------------------
void TMeasurementPacket::GoNextActive()
{
    if (Current == NULL)
        return;

    IMeasurementEntry* e = Current->next();

    while (e != NULL && !e->IsActive())
        e = e->next();

    Current = e;
}

//-----------------------------------------------------------------------------
// Метод делает текущим последнее активное измерение, которое расположено в 
// пакете перед ранее установленным текущим измерением.
// Если таких измерений нет, то после работы метода текущее измерение 
// становится неопределенным.
//-----------------------------------------------------------------------------
void TMeasurementPacket::GoPrevActive()
{
    if (Current == NULL)
        return;

    IMeasurementEntry* e = Current->prev();

    while (e != NULL && !e->IsActive())
        e = e->prev();

    Current = e;
}

//-----------------------------------------------------------------------------
// Метод возвращает указатель на измерение пакета, идентификатор которого 
// равен id. Если такое измерение отсутствует в пакете, то метод возвращает 
// NULL. Метод не изменяет текущего измерения пакета.
//-----------------------------------------------------------------------------
const IMeasurement* TMeasurementPacket::Find(TMeasId id) const
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->Ref->GetID() == id)
            return e->Ref;
    }

    return NULL;
}

//-----------------------------------------------------------------------------
// Метод возвращает true, если указанное измерение является активным 
// (т.е. включено в обработку)
//-----------------------------------------------------------------------------
bool TMeasurementPacket::IsActive(const IMeasurement& mr) const
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->Ref == &mr)
            break;
    }

    return e != NULL ? e->IsActive() : false;
}

//-----------------------------------------------------------------------------
// Метод возвращает true, если измерение с указанным id является активным 
// (т.е. включено в обработку)
//-----------------------------------------------------------------------------
bool TMeasurementPacket::IsActive(TMeasId id) const
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->Ref->GetID() == id)
            break;
    }

    return e != NULL ? e->IsActive() : false;
}

//-----------------------------------------------------------------------------
// Метод возвращает значение признака ручного исключения измерения m из обработки.
// Это значение равно true, если этот признак установлен и, следовательно,
// измерение исключено из обработки.
//-----------------------------------------------------------------------------
bool TMeasurementPacket::GetManualFlag(const IMeasurement& mr) const
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->Ref == &mr)
            break;
    }

    return e != NULL ? e->ManualFlag : false;
}

//-----------------------------------------------------------------------------
// Метод устанавливает значение f признака ручного исключения измерения m из 
// обработки. Если f равно true, то признак установливается и, следовательно, 
// измерение исключено из обработки.
//-----------------------------------------------------------------------------
bool TMeasurementPacket::GetManualFlag(TMeasId id) const
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->Ref->GetID() == id)
            break;
    }
    return e != NULL ? e->ManualFlag : false;
}

//-----------------------------------------------------------------------------
bool TMeasurementPacket::GetDeleteFlag(TMeasId id) const
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->Ref->GetID() == id)
            break;
    }
    return e != NULL ? e->DeleteFlag : false;
}

bool TMeasurementPacket::GetDeleteFlag(const IMeasurement& mr) const
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->Ref == &mr)
            break;
    }

    return e != NULL ? e->DeleteFlag : false;
}

void TMeasurementPacket::SetDeleteFlag(const IMeasurement& mr, bool f)
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->Ref == &mr)
            break;
    }

    if (e != NULL)
        e->DeleteFlag = f;
}

//-----------------------------------------------------------------------------
// Метод устанавливает значение f признака ручного исключения измерения m из 
// обработки. Если f равно true, то признак установливается и, следовательно, 
// измерение исключено из обработки.
//-----------------------------------------------------------------------------
void TMeasurementPacket::SetManualFlag(const IMeasurement& mr, bool f)
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->Ref == &mr)
            break;
    }

    if (e != NULL)
        e->ManualFlag = f;
}

//-----------------------------------------------------------------------------
// Метод устанавливает значение f признака ручного исключения измерения с
// идентификатором id из обработки. Если f равно true, то признак 
// установливается и, следовательно, измерение исключено из обработки.
//-----------------------------------------------------------------------------
void TMeasurementPacket::SetManualFlag(TMeasId id, bool f)
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->Ref->GetID() == id)
            break;
    }
    if (e != NULL)
        e->ManualFlag = f;
}

//-----------------------------------------------------------------------------
// Метод возвращает значение признака автоматического исключения измерения m 
// из обработки. Это значение равно true, если этот признак установлен и, 
// следовательно, измерение исключено из обработки.
//-----------------------------------------------------------------------------
bool TMeasurementPacket::GetAutoFlag(const IMeasurement& mr) const
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->Ref == &mr)
            break;
    }

    return e != NULL ? e->AutoFlag : false;
}

//-----------------------------------------------------------------------------
// Метод возвращает значение признака автоматического исключения измерения с
// идентификатором id из обработки.
// Это значение равно true, если этот признак установлен и, следовательно,
// измерение исключено из обработки.
//-----------------------------------------------------------------------------
bool TMeasurementPacket::GetAutoFlag(TMeasId id) const
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->Ref->GetID() == id)
            break;
    }
    return e != NULL ? e->AutoFlag : false;
}

//-----------------------------------------------------------------------------
// Метод очищает флаги автоматического исключения из обработки для всех 
// измерений пакета
//-----------------------------------------------------------------------------
void TMeasurementPacket::ClearAutoFlags()
{
    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next())
        e->AutoFlag = false;
}

//-----------------------------------------------------------------------------
// Добавление измерения в пакет
//-----------------------------------------------------------------------------
void TMeasurementPacket::Add(const IMeasurement* cm, bool man_flag)
{
    TimeJD tm, t;
    tm = cm->GetTime();

    IMeasurementEntry* n = new IMeasurementEntry;
    n->Ref = cm->Clone();
    n->ManualFlag = man_flag;

    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        t = e->Ref->GetTime();
        if (tm < t) {
            insert_prev(e, n, 0);
            return;
        }
    }

    insert_last(n, 0);
}

//-----------------------------------------------------------------------------
// Метод расставляет признаки автоматического исключения из обработки для 
// измерения m. Второй параметр - вектор невязок
//-----------------------------------------------------------------------------
void TMeasurementPacket::FilterMeasurement(const IMeasurement& m, const DVector& del)
{
    if (Filter == NULL)
        return;

    IMeasurementEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->Ref == &m)
            break;
    }

    if (e != NULL && !e->ManualFlag)
        e->AutoFlag = Filter->IsDeactivate(*e->Ref, del);
}

//-----------------------------------------------------------------------------
// Формирование пакета из исходного за счет извлечения измерений в заданном 
// интервале времени
//-----------------------------------------------------------------------------
void TMeasurementPacket::Create(IMeasurementPacket& src, const TimeJD& t1,
                                const TimeJD& t2)
{
    TimeJD tl, tr, tt;

    clear();

    tl = t1;
    tr = t2;

    if (tr < tl) {
        tt = tl;
        tl = tr;
        tr = tt;
    }

    src.SetFirst();

    while (src.GetCurrent() != NULL && src.GetCurrent()->GetTime() < tl)
        src.GoNext();

    const IMeasurement* m;

    while ((m = src.GetCurrent()) != NULL && m->GetTime() <= tr) {
        Add(m, GetManualFlag(*m));
        src.GoNext();
    }
}

//-----------------------------------------------------------------------------
// Формирование пакета из исходного за счет извлечения измерений в заданном 
// интервале времени
//-----------------------------------------------------------------------------
bool TMeasurementPacket::IsEnough(unsigned kmin)
{
    unsigned k = 0;

    SetFirstActive();

    const IMeasurement* m;

    while ((m = GetCurrent()) != NULL) {
        k += m->GetSize();
        if (k > kmin)
            return true;
        GoNextActive();
    }

    return false;
}

//-----------------------------------------------------------------------------
// Исключение из пакета неактивных измерений
//-----------------------------------------------------------------------------
void TMeasurementPacket::DeleteInActive()
{
    IMeasurementEntry *e, *en;

    e = first();
    while (e != NULL) {
        en = e->next();
        if (!e->IsActive())
            delete_node(e);
        e = en;
    }
}

///////////////////////////////////////////////////////////////////////////////
