///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация пакета измерений в виде связанного списка
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_TMEASUREMENTPACKET_H
#define BNOTOOL_TMEASUREMENTPACKET_H
///////////////////////////////////////////////////////////////////////////////
#include "IMeasurementPacket.h"
#include "twll.h"

///////////////////////////////////////////////////////////////////////////////
// Представление одного элемента пакета
///////////////////////////////////////////////////////////////////////////////
class IMeasurementRef
{
public:
    // Конструктор
    IMeasurementRef();
    // Деструктор
    virtual ~IMeasurementRef();
    // Является ли измерение активным
    bool IsActive() const;
    bool IsDeleted() const;
    // Указатель на измерение
    IMeasurement* Ref;
    // Флаг участия в обработке (ручной)
    bool ManualFlag;
    bool DeleteFlag;
    // Флаг участия в обработке (автоматический
    bool AutoFlag;
};

///////////////////////////////////////////////////////////////////////////////
// Вспомогательный класс для представления измерения в качестве элемента списка
///////////////////////////////////////////////////////////////////////////////
typedef Twll_Entry_Class<IMeasurementRef> IMeasurementEntry;

///////////////////////////////////////////////////////////////////////////////
// Класс, хранящий пакет измерений в форме связанного списка
///////////////////////////////////////////////////////////////////////////////
class TMeasurementPacket : public Twll_List_Class<IMeasurementEntry>, public IMeasurementPacket
{
public:
    // Конструктор по умолчанию
    TMeasurementPacket()
    {
        Current = NULL;
        Filter = NULL;
    };
    // Метод возвращает число измерений в пакете
    virtual unsigned GetSize() const
    {
        return size();
    };
    // Метод возвращает true, если пакет пуст (не содержит ни одного измерения)
    // и возвращает false в противном случае
    virtual bool IsEmpty() const
    {
        return isempty();
    };
    // Метод возвращает число активных измерений в пакете
    virtual unsigned GetActiveSize() const;
    // Метод возвращает true, если пакет содержит хотя бы одно активное измерение
    // и возвращает false в противном случае
    virtual bool IsActive() const;
    // Метод делает текущим первое измерение пакета
    virtual void SetFirst()
    {
        Current = first();
    };
    // Метод делает текущим последнее измерение пакета
    virtual void SetLast()
    {
        Current = last();
    };
    // Метод делает текущим первое активное измерение пакета.
    // Если активных измерений нет, то текущее измерение становится неопределенным
    virtual void SetFirstActive();
    // Метод делает текущим последнее активное измерение пакета.
    // Если активных измерений нет, то текущее измерение становится неопределенным
    virtual void SetLastActive();
    // Метод делает текущим первое измерение пакета, время которого больше или равно
    // указанного времени t.
    // Если таких измерений нет, то текущее измерение становится неопределенным
    virtual void SetByTime(const TimeJD& t);
    // Метод делает текущим первое активное измерение пакета, время которого больше
    // или равно указанного времени t.
    // Если таких измерений нет, то текущее измерение становится неопределенным
    virtual void SetByTimeActive(const TimeJD& t);
    // Метод позволяет явно установить текущее измерение пакета по ссылке m.
    // Если указанное измерение отсутствует в пакете, текущее измерение становится неопределенным.
    virtual void SetCurrent(const IMeasurement& m);
    // Метод позволяет явно установить текущее измерение пакета по идентификатору измерения id.
    // Если указанное измерение отсутствует в пакете, текущее измерение становится неопределенным.
    virtual void SetCurrent(TMeasId id);
    // Метод возвращает указатель на текущее измерение пакета.
    // Если текущее измерение не определено, метод возвращает NULL.
    virtual const IMeasurement* GetCurrent() const
    {
        return Current == NULL ? NULL : Current->Ref;
    };

    virtual IMeasurement* GetCurrent()
    {
        return Current == NULL ? NULL : Current->Ref;
    };
    // Метод делает текущим измерение, которое располагается непосредственно сразу за ранее
    // установленным текущим измерением.
    // Если текущим было последнее измерение пакета,
    // то после работы метода текущее измерение становится неопределенным.
    virtual void GoNext();
    // Метод делает текущим измерение, которое располагается непосредственно перед ранее
    // установленным текущим измерением.
    // Если текущим было первое измерение пакета,
    // то после работы метода текущее измерение становится неопределенным.
    virtual void GoPrev();
    // Метод делает текущим первое активное измерение, которое расположено в пакете после
    // ранее установленного текущего измерения.
    // Если таких измерений нет, то после работы метода текущее измерение становится неопределенным.
    virtual void GoNextActive();
    // Метод делает текущим последнее активное измерение, которое расположено в пакете перед
    // ранее установленным текущим измерением.
    // Если таких измерений нет, то после работы метода текущее измерение становится неопределенным.
    virtual void GoPrevActive();
    // Метод возвращает указатель на измерение пакета, идентификатор которого равен id.
    // Если такое измерение отсутствует в пакете, то метод возвращает NULL.
    // Метод не изменяет текущего измерения пакета.
    virtual const IMeasurement* Find(TMeasId id) const;
    // Метод возвращает true, если указанное измерение является активным (т.е. включено в обработку)
    virtual bool IsActive(const IMeasurement& m) const;
    // Метод возвращает true, если измерение с указанным id является активным (т.е. включено в обработку)
    virtual bool IsActive(TMeasId id) const;
    // Метод возвращает значение признака ручного исключения измерения m из обработки.
    // Это значение равно true, если этот признак установлен и, следовательно,
    // измерение исключено из обработки.

    virtual void SetDeleteFlag(const IMeasurement& m, bool f);
    virtual bool GetDeleteFlag(TMeasId id) const;
    virtual bool GetDeleteFlag(const IMeasurement& m) const;

    virtual bool GetManualFlag(const IMeasurement& m) const;
    // Метод возвращает значение признака ручного исключения измерения с
    // идентификатором id из обработки.
    // Это значение равно true, если этот признак установлен и, следовательно,
    // измерение исключено из обработки.
    virtual bool GetManualFlag(TMeasId id) const;
    // Метод устанавливает значение f признака ручного исключения измерения m из обработки.
    // Если f равно true, то признак установливается и, следовательно, измерение исключено из обработки.
    virtual void SetManualFlag(const IMeasurement& m, bool f);
    // Метод устанавливает значение f признака ручного исключения измерения с
    // идентификатором id из обработки.
    // Если f равно true, то признак установливается и, следовательно, измерение исключено из обработки.
    virtual void SetManualFlag(TMeasId id, bool f);
    // Метод возвращает значение признака автоматического исключения измерения m из обработки.
    // Это значение равно true, если этот признак установлен и, следовательно,
    // измерение исключено из обработки.
    virtual bool GetAutoFlag(const IMeasurement& m) const;
    // Метод возвращает значение признака автоматического исключения измерения с
    // идентификатором id из обработки.
    // Это значение равно true, если этот признак установлен и, следовательно,
    // измерение исключено из обработки.
    virtual bool GetAutoFlag(TMeasId id) const;
    // Метод очищает флаги автоматического исключения из обработки для всех измерений пакета
    virtual void ClearAutoFlags();
    // Метод устанавливает объект, выполняющий автоматическое исключение измерений из обработки
    virtual void SetFilter(const IMeasurementFilter* f)
    {
        Filter = f;
    };
    // Метод возвращает указатель на объект, выполняющий автоматическое исключение
    // измерений из обработки
    virtual const IMeasurementFilter* GetFilter()
    {
        return Filter;
    };
    // Метод расставляет признаки автоматического исключения из обработки для 
    // измерения m. Второй параметр - вектор невязок
    virtual void FilterMeasurement(const IMeasurement& m, const DVector& del);
    // Метод возвращает true, если число измеренных параметров в пакете
    // превышает указанное число kmin
    // При подсчете учитываются только активные измерения
    virtual bool IsEnough(unsigned kmin);

    // Добавление измерения в пакет
    // Измерения располагаются в порядке возрастания их времен
    void Add(const IMeasurement* m, bool man_flag = false);

    // Формирование пакета из исходного за счет извлечения
    // измерений в заданном интервале
    void Create(IMeasurementPacket& src, const TimeJD& t1, const TimeJD& t2);

    // Исключение из пакета неактивных измерений
    void DeleteInActive();

protected:
    // Текущее измерение потока
    IMeasurementEntry* Current;
    // Указатель на объект, описывающий процедуру автоматического отбора измерений
    // для обработки
    const IMeasurementFilter* Filter;
};

///////////////////////////////////////////////////////////////////////////////
#endif
