///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
#include "TrScaner.h"
///////////////////////////////////////////////////////////////////////////////
// Список времен
///////////////////////////////////////////////////////////////////////////////
// Вспомогательная функция сортировки по возрастанию
//-----------------------------------------------------------------------------
static int cmp_inc(TWLL_BENTRY_CLASS* b1, TWLL_BENTRY_CLASS* b2)
{
    TimeJDEntry* t1 = (TimeJDEntry *)b1;
    TimeJDEntry* t2 = (TimeJDEntry *)b2;

    if (*t1 < *t2)
        return -1;
    if (*t1 > *t2)
        return 1;

    return 0;
}

//-----------------------------------------------------------------------------
// Вспомогательная функция сортировки по убыванию
//-----------------------------------------------------------------------------
static int cmp_dec(TWLL_BENTRY_CLASS* b1, TWLL_BENTRY_CLASS* b2)
{
    TimeJDEntry* t1 = (TimeJDEntry *)b1;
    TimeJDEntry* t2 = (TimeJDEntry *)b2;

    if (*t1 > *t2)
        return -1;
    if (*t1 < *t2)
        return 1;

    return 0;
}

//-----------------------------------------------------------------------------
// Сортировка dir > 0 - по возрастанию
//-----------------------------------------------------------------------------
void TimeJDList::Sort(int dir)
{
    sort(dir > 0 ? cmp_inc : cmp_dec);
}

//-----------------------------------------------------------------------------
// Поиск 
//-----------------------------------------------------------------------------
TimeJDEntry* TimeJDList::Find(const TimeJD& t) const
{
    TimeJDEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (t == *e)
            return e;
    }

    return NULL;
}

///////////////////////////////////////////////////////////////////////////////
// Набор контрольных точек, задаваемых с помощью явных моментов времени
///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
//-----------------------------------------------------------------------------
TimeCPointSet::TimeCPointSet()
{
    CurDir = 0;
}

//-----------------------------------------------------------------------------
// Установка очередной точки
//-----------------------------------------------------------------------------
void TimeCPointSet::Set(const TimeJD& t)
{
    if (TList.Find(t) == NULL) {
        TimeJDEntry* e = new TimeJDEntry;
        *(TimeJD *)e = t;
        TList.insert_last(e, 0);
    }
}

//-----------------------------------------------------------------------------
// Удаление точки
//-----------------------------------------------------------------------------
void TimeCPointSet::Delete(const TimeJD& t)
{
    TimeJDEntry* e = TList.Find(t);
    if (e != NULL)
        TList.delete_node(e);
}

//-----------------------------------------------------------------------------
// Содержит ли набор указанную точку 
//-----------------------------------------------------------------------------
bool TimeCPointSet::HasPoint(const TimeJD& t) const
{
    TimeJDEntry* e = TList.Find(t);

    for (e = TList.first(); e != NULL; e = e->next()) {
        if (*e == t)
            return true;
    }

    return false;
}

//-----------------------------------------------------------------------------
// Есть ли точки в указанном диапазоне ?
//-----------------------------------------------------------------------------
bool TimeCPointSet::HasPoints(const TimeJD& t1, const TimeJD& t2)
{
    int dir = (t2 > t1) ? 1 : -1;

    if (CurDir != dir) {
        TList.Sort(dir);
        CurDir = dir;
    }

    if (dir > 0) {
        TimeJDEntry* e = TList.first();
        while (e != NULL && (*e) <= t1)
            e = e->next();

        if (e == NULL || *e > t2)
            return false;

        if (*e < t2)
            return true;

        return t1 == t2;
    }
    else {
        TimeJDEntry* e = TList.first();
        while (e != NULL && (*e) >= t1)
            e = e->next();

        if (e == NULL || *e < t2)
            return false;

        if (*e > t2)
            return true;

        return t1 == t2;
    }
}

//-----------------------------------------------------------------------------
// Вернуть точку, ближайшую к левой границе
//-----------------------------------------------------------------------------
TimeJD TimeCPointSet::GetPoint(const TimeJD& t1, const TimeJD& t2)
{
    int dir = (t2 > t1) ? 1 : -1;
    TAssert(dir == CurDir);

    if (dir > 0) {
        TimeJDEntry* e = TList.first();
        while (e != NULL && (*e) <= t1)
            e = e->next();

        TAssert(e != NULL && *e <= t2);

        return *e;
    }
    else {
        TimeJDEntry* e = TList.first();
        while (e != NULL && (*e) >= t1)
            e = e->next();

        TAssert(e != NULL && *e >= t2);

        return *e;
    }
}

///////////////////////////////////////////////////////////////////////////////
// Список наборов контрольных точек, задаваемых с помощью явных моментов времени
///////////////////////////////////////////////////////////////////////////////
// Есть ли точки в указанном диапазоне ?
//-----------------------------------------------------------------------------
bool TimeCPointList::HasPoints(const TimeJD& t1, const TimeJD& t2)
{
    if (isempty())
        return false;

    TimeCPointEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->HasPoints(t1, t2))
            return true;
    }

    return false;
}

//-----------------------------------------------------------------------------
// Вернуть точку, ближайшую к левой границе
//-----------------------------------------------------------------------------
TimeJD TimeCPointList::GetPoint(const TimeJD& t1, const TimeJD& t2)
{
    bool flag = false;
    TimeJD ret;

    TimeCPointEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (!e->HasPoints(t1, t2))
            continue;
        TimeJD t = e->GetPoint(t1, t2);
        if (!flag) {
            flag = true;
            ret = t;
        }
        else if ((t1 < t2 && t < ret) || (t1 > t2 && t > ret))
            ret = t;
    }

    TAssert(flag);
    return ret;
}

///////////////////////////////////////////////////////////////////////////////
// Набор контрольных точек, задаваемых постоянным шагом
///////////////////////////////////////////////////////////////////////////////
// Инициализация
//-----------------------------------------------------------------------------
void ITimeStepSet::Init(const TimeJD& t1, const TimeJD& t2)
{
    TAssert(Step != 0);

    Dir = t1 < t2 ? 1 : -1;

    if (Dir > 0) {
        if (TStart < t1) {
            double dt = floor((t1 - TStart).GetDouble() / fabs(Step / 86.4)) + 1.0;
            NSteps = (int)dt;
            TNext = TStart;
            TNext += fabs(Step) / 86.4 * NSteps;
            while (TNext <= t1) {
                NSteps++;
                TNext = TStart;
                TNext += fabs(Step) / 86.4 * NSteps;
            }
            return;
        }
        else {
            double dt = floor((TStart - t1).GetDouble() / fabs(Step / 86.4));
            NSteps = -(int)dt;
            TNext = TStart;
            TNext += fabs(Step) / 86.4 * NSteps;
            while (TNext <= t1) {
                NSteps++;
                TNext = TStart;
                TNext += fabs(Step) / 86.4 * NSteps;
            }
            return;
        }
    }
    else {
        if (TStart < t1) {
            double dt = floor((t1 - TStart).GetDouble() / fabs(Step / 86.4));
            NSteps = (int)dt;
            TNext = TStart;
            TNext += fabs(Step) / 86.4 * NSteps;
            while (TNext >= t1) {
                NSteps--;
                TNext = TStart;
                TNext += fabs(Step) / 86.4 * NSteps;
            }
            return;
        }
        else {
            double dt = floor((TStart - t1).GetDouble() / fabs(Step / 86.4)) + 1.0;
            NSteps = -(int)dt;
            TNext = TStart;
            TNext += fabs(Step) / 86.4 * NSteps;
            while (TNext >= t1) {
                NSteps++;
                TNext = TStart;
                TNext -= fabs(Step) / 86.4 * NSteps;
            }
            return;
        }
    }
}

//-----------------------------------------------------------------------------
// Есть ли точки в указанном диапазоне ?
//-----------------------------------------------------------------------------
bool ITimeStepSet::HasPoints(const TimeJD& t1, const TimeJD& t2)
{
    TAssert(Dir != 0 && Step != 0);
    TAssert((t1 < t2 && Dir > 0) || (t1 > t2 && Dir < 0));

    if (Dir > 0)
        return TNext > t1 && TNext <= t2;
    else
        return TNext < t1 && TNext >= t2;
}

//-----------------------------------------------------------------------------
// Вернуть точку, ближайшую к левой границе
//-----------------------------------------------------------------------------
TimeJD& ITimeStepSet::GetPoint(const TimeJD& t1, const TimeJD& t2)
{
    return TNext;
}

//-----------------------------------------------------------------------------
// Содержит ли набор указанную точку 
//-----------------------------------------------------------------------------
bool ITimeStepSet::HasPoint(const TimeJD& t) const
{
    return t == TNext;
}

//-----------------------------------------------------------------------------
// Метод, вызываемый при достижении очередной контрольной точки
//-----------------------------------------------------------------------------
bool ITimeStepSet::Reaction(const PhasePoint6D& src)
{
    NSteps += Dir;
    TNext = TStart;
    TNext += fabs(Step) / 86.4 * NSteps;
    return CReaction(src);
}

///////////////////////////////////////////////////////////////////////////////
// Список наборов контрольных точек, задаваемых с шагом по времени
///////////////////////////////////////////////////////////////////////////////
// Инициализация
//-----------------------------------------------------------------------------
void ITimeStepList::Init(const TimeJD& t1, const TimeJD& t2)
{
    ITimeStepEntry* e;

    for (e = first(); e != NULL; e = e->next())
        e->Init(t1, t2);
}

//-----------------------------------------------------------------------------
// Есть ли точки в указанном диапазоне ?
//-----------------------------------------------------------------------------
bool ITimeStepList::HasPoints(const TimeJD& t1, const TimeJD& t2)
{
    if (isempty())
        return false;

    ITimeStepEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (e->HasPoints(t1, t2))
            return true;
    }

    return false;
}

//-----------------------------------------------------------------------------
// Вернуть точку, ближайшую к левой границе
//-----------------------------------------------------------------------------
TimeJD ITimeStepList::GetPoint(const TimeJD& t1, const TimeJD& t2)
{
    bool flag = false;
    TimeJD ret;

    ITimeStepEntry* e;

    for (e = first(); e != NULL; e = e->next()) {
        if (!e->HasPoints(t1, t2))
            continue;
        TimeJD t = e->GetPoint(t1, t2);
        if (!flag) {
            flag = true;
            ret = t;
        }
        else if ((t1 < t2 && t < ret) || (t1 > t2 && t > ret))
            ret = t;
    }

    TAssert(flag);
    return ret;
}

///////////////////////////////////////////////////////////////////////////////
