///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
#include "TrScaner.h"
///////////////////////////////////////////////////////////////////////////////
// Универсальный сканер с использованием внешнего прогноза
///////////////////////////////////////////////////////////////////////////////
// Консруктор по умолчанию
//-----------------------------------------------------------------------------
TrScaner::TrScaner()
{
    FList = NULL;
    RList = NULL;
    TList = NULL;
    SList = NULL;

    Nf = Nr = 0;

    Fl = Fr = Fc = NULL;
    Rl = Rr = Rc = NULL;
};
//-----------------------------------------------------------------------------
// Деструктор
//-----------------------------------------------------------------------------
TrScaner::~TrScaner()
{
    Clear();
}

//-----------------------------------------------------------------------------
// Очистка заказанной памяти
//-----------------------------------------------------------------------------
void TrScaner::Clear()
{
    if (Nf > 0) {
        delete[] Fl;
        delete[] Fr;
        delete[] Fc;
    }

    if (Nr > 0) {
        delete[] Rl;
        delete[] Rr;
        delete[] Rc;
    }

    Nf = Nr = 0;
}

//-----------------------------------------------------------------------------
// Запуск сканера в заданном диапазоне времени
// Задается шаг сканирования и требуемая точность выхода на 
// точки конттрольные точки
//-----------------------------------------------------------------------------
bool TrScaner::Run(IPredict& prg, const TimeJD& t1, const TimeJD& t2, double step, double eps)
{
    // Подтягивание прогноза на начальный момент времени и проверка результата
    if (!prg.Run(t1))
        return false;

    prg.GetDestPoint(Pl);

    // Заказ памяти для рабочих переменных
    Clear();
    if (FList == NULL)
        Nf = 0;
    else {
        Nf = FList->size();
        Fl = new double[Nf];
        Fr = new double[Nf];
        Fc = new double[Nf];
    }

    if (RList == NULL)
        Nr = 0;
    else {
        Nr = RList->size();
        Rl = new int[Nr];
        Rr = new int[Nr];
        Rc = new int[Nr];
    }

    // Заполнение функций на левой границе
    SetFunctions(Pl, Fl, Rl);

    // Инициализация контрольных точек с постоянным шагом
    if (SList != NULL)
        SList->Init(t1, t2);

    // Основной цикл
    do {
        // Проверка временных контрольных точек на левой границе
        if (!CheckTPoints(Pl))
            break;
        // Выбор правой границы шага
        TimeJD tr = SelectStep(Pl.T, t2, step);
        // Прогноз на правую границу
        if (!prg.Run(tr))
            return false;
        prg.GetDestPoint(Pr);
        Pr.T = tr;
        // Заполнение функций на левой границе
        SetFunctions(Pr, Fr, Rr);
        // Если внутри есть контрольные точки по функциям, то
        // пытаемся к ним сойтись
        if (HasCPoints(Fl, Rl, Fr, Rr)) {
            if (!FindCPoint(prg, eps))
                return false;
            if (!CallReaction())
                return true;
        }
        // Сдвиг к следующему шагу
        DoStep(Pl, Fl, Rl, Pr, Fr, Rr);
    }
    while (((Pl.T < t2) && (t1 < t2)) || ((Pl.T > t2) && (t1 > t2)));

    // Проверка временных контрольных точек в конце
    CheckTPoints(Pl);

    return true;
}

//-----------------------------------------------------------------------------
// Заполнение массивов функций
//-----------------------------------------------------------------------------
void TrScaner::SetFunctions(const PhasePoint6D& p, double* f, int* r)
{
    int i;

    IFloatCPointEntry* fe;

    if (Nf > 0) {
        for (i = 0, fe = FList->first(); fe != NULL; fe = fe->next())
            f[i++] = fe->CFunction(p);
    }

    IRelCPointEntry* re;

    if (Nr > 0) {
        for (i = 0, re = RList->first(); re != NULL; re = re->next())
            r[i++] = re->CFunction(p);
    }
}

//-----------------------------------------------------------------------------
// Проверка временных контрольных точек
//-----------------------------------------------------------------------------
bool TrScaner::CheckTPoints(const PhasePoint6D& p) const
{
    bool ret = true;

    if (TList != NULL) {
        TimeCPointEntry* te;
        for (te = TList->first(); te != NULL; te = te->next()) {
            if (te->HasPoint(p.T) && (!te->CReaction(p)))
                ret = false;
        }
    }

    if (SList != NULL) {
        ITimeStepEntry* se;
        for (se = SList->first(); se != NULL; se = se->next()) {
            if (se->HasPoint(p.T) && (!se->Reaction(p)))
                ret = false;
        }
    }

    return ret;
}

//-----------------------------------------------------------------------------
// Выбор правой границы на шаге
//-----------------------------------------------------------------------------
TimeJD TrScaner::SelectStep(const TimeJD& t1, const TimeJD& t2, double snom)
{
    if (t1 < t2) {
        TimeJD tr;

        tr = t1;
        tr += fabs(snom) / 86.4;

        if (tr > t2)
            tr = t2;

        if (TList != NULL && TList->HasPoints(t1, tr))
            tr = TList->GetPoint(t1, tr);

        if (SList != NULL && SList->HasPoints(t1, tr)) {
            TimeJD t = SList->GetPoint(t1, tr);
            if (t < tr)
                tr = t;
        }

        return tr;
    }
    else {
        TimeJD tr;

        tr = t1;
        tr -= fabs(snom) * 86.4;

        if (tr < t2)
            tr = t2;

        if (TList != NULL && TList->HasPoints(t1, tr))
            tr = TList->GetPoint(t1, tr);

        if (SList != NULL && SList->HasPoints(t1, tr)) {
            TimeJD t = SList->GetPoint(t1, tr);
            if (t > tr)
                tr = t;
        }

        return tr;
    }
}

//-----------------------------------------------------------------------------
// Проверка наличия контрольных точек на интервале
//-----------------------------------------------------------------------------
bool TrScaner::HasCPoints(const double* fl, const int* rl, const double* fr, const int* rr)
{
    if (Nf > 0) {
        int i;
        for (i = 0; i < Nf; i++) {
            if (fl[i] * fr[i] < 0)
                return true;
        }
    }

    if (Nr > 0) {
        int i;
        for (i = 0; i < Nr; i++) {
            if (rl[i] * rr[i] < 0)
                return true;
        }
    }

    return false;
}

//-----------------------------------------------------------------------------
// Сдвиг к следующему шагу
//-----------------------------------------------------------------------------
void TrScaner::DoStep(PhasePoint6D& pl, double* fl, int* rl,
                      const PhasePoint6D& pr, const double* fr, const int* rr)
{
    pl = pr;

    if (Nf > 0) {
        int i;
        for (i = 0; i < Nf; i++)
            fl[i] = fr[i];
    }

    if (Nr > 0) {
        int i;
        for (i = 0; i < Nr; i++)
            rl[i] = rr[i];
    }
}

//-----------------------------------------------------------------------------
// Поиск контрольной точки на интервале с заданной точностью
//-----------------------------------------------------------------------------
bool TrScaner::FindCPoint(IPredict& prg, double eps)
{
    while (fabs((Pl.T - Pr.T).GetDouble()) * 86.4 > eps) {
        double sd = (Pr.T - Pl.T).GetDouble();
        double sc = sd;
        // Ищем новое приближение для времени
        int i;
        for (i = 0; i < Nf; i++) {
            if (Fl[i] * Fr[i] >= 0)
                continue;
            double s;
            s = Fl[i] * sd / (Fl[i] - Fr[i]);
            if (fabs(s) < 0.0001 * fabs(sd))
                s = 0.0001 * sd;
            else if (fabs(s - sd) < 0.0001 * fabs(sd))
                s = 0.9999 * sd;
            if (fabs(s) < fabs(sc))
                sc = s;
        }
        if (fabs(sc) > 0.5 * fabs(sd) && Nr > 0) {
            for (i = 0; i < Nf; i++) {
                if (Rl[i] * Rr[i] >= 0)
                    continue;
                sc = 0.5 * sd;
            }
        }
        TimeJD tc = Pl.T;
        tc += sc;
        if (!prg.Run(tc))
            return false;
        prg.GetDestPoint(Pc);
        // Заполнение функций на момент tc
        SetFunctions(Pc, Fc, Rc);
        // Проверка наличия контрольных точек на левом интервале
        if (HasCPoints(Fl, Rl, Fc, Rc))
            DoStep(Pr, Fr, Rr, Pc, Fc, Rc);
        else
            DoStep(Pl, Fl, Rl, Pc, Fc, Rc);
    }

    return true;
}

//-----------------------------------------------------------------------------
// Вызов методов реакции после того, как найдены контрольные точки
//-----------------------------------------------------------------------------
bool TrScaner::CallReaction()
{
    int i;
    bool ret = true;

    IFloatCPointEntry* fe;

    if (FList != NULL) {
        for (i = 0, fe = FList->first(); fe != NULL; fe = fe->next(), i++) {
            if (Fl[i] * Fr[i] >= 0)
                continue;
            if (!fe->CReaction(Pl))
                ret = false;
        }
    }

    IRelCPointEntry* re;
    if (RList != NULL) {
        for (i = 0, re = RList->first(); re != NULL; re = re->next(), i++) {
            if (Rl[i] * Rr[i] >= 0)
                continue;
            if (!re->CReaction(Pl))
                ret = false;
        }
    }

    return ret;
}

///////////////////////////////////////////////////////////////////////////////
