///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Классы для организации прогнозирования с установкой контрольных точек
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_TRSCANER_H
#define BNOTOOL_TRSCANER_H
///////////////////////////////////////////////////////////////////////////////
#include "twll.h"
#include "IPredict.h"

///////////////////////////////////////////////////////////////////////////////
// Средства обеспечения сканирования траектории движения с выходом на 
// контрольные точки
///////////////////////////////////////////////////////////////////////////////
// Интерфейс контрольной точки, задаваемой с помощью действительной функции
///////////////////////////////////////////////////////////////////////////////
class IFloatCPoint
{
public:
    // Метод, вычисляющий значение функции
    virtual ~IFloatCPoint()
    {
    };
    // Метод, вычисляющий значение функции
    // Ноль означает условие выхода на контрольную точку
    virtual double CFunction(const PhasePoint6D& src) const = 0;
    // Метод, который должен вызываться при достижении контрольной точки
    // Метод возвращает true, если можно продолжать сканирование
    virtual bool CReaction(const PhasePoint6D& src)
    {
        return true;
    };
};

///////////////////////////////////////////////////////////////////////////////
// Список контрольных точек с действительными функциями
///////////////////////////////////////////////////////////////////////////////
typedef Twll_Entry_Class<IFloatCPoint> IFloatCPointEntry;

class IFloatCPointList : public Twll_List_Class<IFloatCPointEntry>
{
public:
};

///////////////////////////////////////////////////////////////////////////////
// Интерфейс контрольной точки, задаваемой с помощью релейной функции
///////////////////////////////////////////////////////////////////////////////
class IRelCPoint
{
public:
    // Метод, вычисляющий значение функции
    virtual ~IRelCPoint()
    {
    };
    // Метод, вычисляющий значение функции
    // Ноль означает условие выхода на контрольную точку
    virtual int CFunction(const PhasePoint6D& src) const = 0;
    // Метод, который должен вызываться при достижении контрольной точки
    // Метод возвращает true, если можно продолжать сканирование
    virtual bool CReaction(const PhasePoint6D& src)
    {
        return true;
    };
};

///////////////////////////////////////////////////////////////////////////////
// Список контрольных точек с релейными функциями
///////////////////////////////////////////////////////////////////////////////
typedef Twll_Entry_Class<IRelCPoint> IRelCPointEntry;

class IRelCPointList : public Twll_List_Class<IRelCPointEntry>
{
public:
};

///////////////////////////////////////////////////////////////////////////////
// Набор контрольных точек, задаваемых с помощью явных моментов времени
///////////////////////////////////////////////////////////////////////////////
typedef Twll_Entry_Class<TimeJD> TimeJDEntry;

class TimeJDList : public Twll_List_Class<TimeJDEntry>
{
public:
    // Сортировка dir > 0 - по возрастанию
    void Sort(int dir);
    // Поиск 
    TimeJDEntry* Find(const TimeJD& t) const;
};

class TimeCPointSet
{
public:
    // Конструктор по умолчанию
    virtual ~TimeCPointSet()
    {
    };
    // Конструктор по умолчанию
    TimeCPointSet();

    // Установка очередной точки
    void Set(const TimeJD& t);
    // Удаление точки
    void Delete(const TimeJD& t);

    // Содержит ли набор указанную точку 
    bool HasPoint(const TimeJD& t) const;
    // Есть ли точки в указанном диапазоне ?
    bool HasPoints(const TimeJD& t1, const TimeJD& t2);
    // Вернуть точку, ближайшую к левой границе
    TimeJD GetPoint(const TimeJD& t1, const TimeJD& t2);

    // Метод, который должен вызываться при достижении контрольной точки
    // Метод возвращает true, если можно продолжать сканирование
    virtual bool CReaction(const PhasePoint6D& src)
    {
        return true;
    };

protected:
    // Список установленных точек
    TimeJDList TList;
    // Текущий порядок сортировки
    int CurDir;
};

typedef Twll_Entry_Class<TimeCPointSet> TimeCPointEntry;

class TimeCPointList : public Twll_List_Class<TimeCPointEntry>
{
public:
    // Есть ли точки в указанном диапазоне ?
    bool HasPoints(const TimeJD& t1, const TimeJD& t2);
    // Вернуть точку, ближайшую к левой границе
    TimeJD GetPoint(const TimeJD& t1, const TimeJD& t2);
};

///////////////////////////////////////////////////////////////////////////////
// Набор контрольных точек, задаваемых постоянным шагом
///////////////////////////////////////////////////////////////////////////////
class ITimeStepSet
{
public:
    // Конструктор
    virtual ~ITimeStepSet()
    {
    };
    // Конструктор
    ITimeStepSet()
    {
        Dir = 0;
        Step = 0.0;
    };
    // Начальная точка
    TimeJD TStart;
    // Шаг в тысячах секунд
    double Step;

    // Инициализация
    void Init(const TimeJD& t1, const TimeJD& t2);
    // Содержит ли набор указанную точку 
    bool HasPoint(const TimeJD& t) const;
    // Есть ли точки в указанном диапазоне ?
    bool HasPoints(const TimeJD& t1, const TimeJD& t2);
    // Вернуть точку, ближайшую к левой границе
    TimeJD& GetPoint(const TimeJD& t1, const TimeJD& t2);
    // Метод, вызываемый при достижении очередной контрольной точки
    bool Reaction(const PhasePoint6D& src);

protected:
    // Метод, который должен вызываться при достижении контрольной точки
    // Метод возвращает true, если можно продолжать сканирование
    virtual bool CReaction(const PhasePoint6D& src)
    {
        return true;
    };

    // Направление движения
    int Dir;
    // Число шагов
    int NSteps;
    // Очередной момент времени останова
    TimeJD TNext;
};

typedef Twll_Entry_Class<ITimeStepSet> ITimeStepEntry;

class ITimeStepList : public Twll_List_Class<ITimeStepEntry>
{
public:
    // Есть ли точки в указанном диапазоне ?
    bool HasPoints(const TimeJD& t1, const TimeJD& t2);
    // Вернуть точку, ближайшую к левой границе
    TimeJD GetPoint(const TimeJD& t1, const TimeJD& t2);
    // Инициализация
    void Init(const TimeJD& t1, const TimeJD& t2);
};

///////////////////////////////////////////////////////////////////////////////
// Универсальный сканер с использованием внешнего прогноза
///////////////////////////////////////////////////////////////////////////////
class TrScaner
{
public:
    // Консруктор по умолчанию
    TrScaner();
    // Деструктор
    virtual ~TrScaner();

    // Установка списка контрольных точек, заданных действительными функциями
    void SetPoints(IFloatCPointList* flist)
    {
        FList = flist;
    };
    // Установка списка контрольных точек, заданных релейными функциями
    void SetPoints(IRelCPointList* rlist)
    {
        RList = rlist;
    };
    // Установка списка контрольных точек, заданных явными моментами времени
    void SetPoints(TimeCPointList* tlist)
    {
        TList = tlist;
    };
    // Установка списка контрольных точек, задаваемых с шагом по времени
    void SetPoints(ITimeStepList* slist)
    {
        SList = slist;
    };

    // Запуск сканера в заданном диапазоне времени
    // Задается шаг сканирования и требуемая точность выхода на 
    // точки конттрольные точки
    bool Run(IPredict& prg, const TimeJD& t1, const TimeJD& t2, double step, double eps);

protected:
    // Список контрольных точек, заданных действительными функциями
    IFloatCPointList* FList;
    // Список контрольных точек, заданных релейными функциями
    IRelCPointList* RList;
    // Список контрольных точек, заданных явными моментами времени
    TimeCPointList* TList;
    // Список точек, задаваемых с шагом по времени
    ITimeStepList* SList;

    // Рабочие переменные для сохранения промежуточных результатов
    // Число контрольных точек с функциями
    int Nf, Nr;
    // Память для хранения значений действительых функций
    double *Fl, *Fr, *Fc;
    // Память для хранения значений целых функций
    int *Rl, *Rr, *Rc;
    // Значения вектора фазовых переменных
    PhasePoint6D Pl, Pr, Pc;

    // Внутренние методы
    // Очистка заказанной памяти
    void Clear();
    // Заполнение массивов функций
    void SetFunctions(const PhasePoint6D& p, double* f, int* r);
    // Проверка временных контрольных точек
    bool CheckTPoints(const PhasePoint6D& p) const;
    // Выбор правой границы на шаге
    TimeJD SelectStep(const TimeJD& t1, const TimeJD& t2, double snom);
    // Проверка наличия контрольных точек на интервале
    bool HasCPoints(const double* fl, const int* rl, const double* fr, const int* rr);
    // Сдвиг к следующему шагу
    void DoStep(PhasePoint6D& pl, double* fl, int* rl,
                const PhasePoint6D& pr, const double* fr, const int* rr);
    // Поиск контрольной точки на интервале с заданной точностью
    bool FindCPoint(IPredict& prg, double eps);
    // Вызов методов реакции после того, как найдены контрольные точки
    bool CallReaction();
};

///////////////////////////////////////////////////////////////////////////////
#endif
