///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация модели движения для краевой задачи на основе численного прогноза
// с постоянными балл. коэффициентом и коэффициентом светового давления
///////////////////////////////////////////////////////////////////////////////
#include "TrjNumVar.h"
///////////////////////////////////////////////////////////////////////////////
// Конструктор
//-----------------------------------------------------------------------------
TrjNumVar::TrjNumVar() : InitTime(0, 0.0)
{
    Kb = Kappa = 0.0;
    IsKbVar = IsKappaVar = false;
}

//-----------------------------------------------------------------------------
// Установка значений коэффициентов и признаков их уточнения
//-----------------------------------------------------------------------------
void TrjNumVar::SetKOProp(double kb, double kappa, bool is_kb_var,
                          bool is_kappa_var)
{
    Kb = kb;
    Kappa = kappa;
    IsKbVar = is_kb_var;
    IsKappaVar = is_kappa_var;
}

// Возврат размерности вектора уточняемых параметров
//-----------------------------------------------------------------------------
unsigned TrjNumVar::Size() const
{
    unsigned n = 6;
    if (IsKbVar)
        n++;
    if (IsKappaVar)
        n++;

    return n;
}

//-----------------------------------------------------------------------------
// Метод устанавливает значение вектора уточняемых параметров
//-----------------------------------------------------------------------------
bool TrjNumVar::Init(const DVector& u0)
{
    // Время должно быть установлено ранее с помощью SetTime() !!!!
    TAssert(InitTime != TimeJD(0, 0.0));

    // Проверка размерности переданного вектора
    unsigned n = Size();
    TAssert(u0.Dim() == n);

    // Установка Kb и Карра в прогнозе
    if (IsKbVar)
        Kb = u0[6];

    if (IsKappaVar)
        Kappa = IsKbVar ? u0[7] : u0[6];

    Predictor.SetKOParams(Kb, Kappa, IsKbVar, IsKappaVar);

    PhasePoint6D ph;
    J2000Frame j2000;
    Vect6 v;

    int i;
    for (i = 0; i < 6; i++)
        v[i] = u0[i];

    ph.Init(j2000, v, InitTime);

    Predictor.SetInitPoint(ph);

    return true;
}

//-----------------------------------------------------------------------------
// Метод выполняет прогнозирование траектории на указанный момент времени t.
// Параметр fr определяет систему отсчету, в которой должен быть возвращен
// фазовый вектор f (т.е. координаты и скорости)
//-----------------------------------------------------------------------------
bool TrjNumVar::Predict(const TimeJD& t, const IFrame& fr, Vect6& f)
{
    if (!GetPredictor()->Run(t))
        return false;

    PhasePoint6D pr, res;

    GetPredictor()->GetDestPoint(pr);
    res.SetFrame(fr);
    res.Assign(pr);

    f = res.CoordsVel;

    return true;
}

//-----------------------------------------------------------------------------
// Метод выполняет прогнозирование траектории на указанный момент времени t и
// подсчет матрицы частных от фазового вектора по уточняемым параметрам.
// Параметр fr определяет систему отсчету, в которой должен быть возвращен
// фазовый вектор f (т.е. координаты и скорости) и матрица m.
//-----------------------------------------------------------------------------
bool TrjNumVar::PredictMatrix(const TimeJD& t, const IFrame& fr, Vect6& f, DMatrix& m)
{
    if (!Predict(t, fr, f))
        return false;

    J2000Frame j2000;

    if (j2000.IsEqual(fr)) {
        GetPredictor()->GetDestMatrix(m);
        m.Transpose();

        return true;
    }

    // Алгоритм с получением матрицы в произвольно СО пока не реализован
    TAssert(false);

    return false;
}

//-----------------------------------------------------------------------------
// Метод выполняет вычисление вектора уточняемых параметров u для указанного
// момента времени t.
//-----------------------------------------------------------------------------
bool TrjNumVar::Move(const TimeJD& t, DVector& u)
{
    if (!GetPredictor()->Run(t))
        return false;

    unsigned n = Size();
    u.Init(n);

    PhasePoint6D ph;

    GetPredictor()->GetDestPoint(ph);
    int i;

    for (i = 0; i < 6; i++)
        u[i] = ph.CoordsVel[i];

    if (IsKbVar)
        u[6] = Kb;

    if (IsKappaVar) {
        if (IsKbVar)
            u[7] = Kappa;
        else
            u[6] = Kappa;
    }

    return true;
}

///////////////////////////////////////////////////////////////////////////////
