///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Реализация модели движения для краевой задачи на основе численного прогноза
// с постоянными балл. коэффициентом и коэффициентом светового давления
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_TRJNUMVAR_H
#define BNOTOOL_TRJNUMVAR_H
///////////////////////////////////////////////////////////////////////////////
#include "ITraject.h"
#include "DigPredictVar.h"

///////////////////////////////////////////////////////////////////////////////
class TrjNumVar : public ITrajectVar
{
public:
    // Конструктор
    TrjNumVar();
    // Деструктор
    virtual ~TrjNumVar()
    {
    };
    // Метод устанавливает значение вектора уточняемых параметров
    virtual unsigned Size() const;
    // Метод устанавливает значение вектора уточняемых параметров
    virtual bool Init(const DVector& u0);
    // Метод устанавливает время начальной точки траектории
    virtual void SetTime(const TimeJD& t)
    {
        InitTime = t;
    };
    // Метод возвращает время начальной точки траектории
    virtual const TimeJD& GetTime() const
    {
        return InitTime;
    };

    // Метод выполняет прогнозирование траектории на указанный момент времени t.
    // Параметр fr определяет систему отсчету, в которой должен быть возвращен
    // фазовый вектор f (т.е. координаты и скорости)
    virtual bool Predict(const TimeJD& t, const IFrame& fr, Vect6& f);
    // Метод выполняет прогнозирование траектории на указанный момент времени t и
    // подсчет матрицы частных от фазового вектора по уточняемым параметрам.
    // Параметр fr определяет систему отсчету, в которой должен быть возвращен
    // фазовый вектор f (т.е. координаты и скорости) и матрица m.
    virtual bool PredictMatrix(const TimeJD& t, const IFrame& fr, Vect6& f, DMatrix& m);
    // Метод выполняет вычисление вектора уточняемых параметров u для указанного
    // момента времени t.
    virtual bool Move(const TimeJD& t, DVector& u);
    // Получение ссылки на экземпляр прогноза (для внешних настроек)
    virtual DigStPredictVar* GetPredictor()
    {
        return &Predictor;
    };
    virtual IPredict* GetPredictorEx() { return &Predictor; }
    // Установка значений коэффициентов и признаков их уточнения
    void SetKOProp(double kb, double kappa, bool is_kb_var = false, bool is_kappa_var = false);

protected:
    // Класс численного прогноза уравнений в вариациях
    DigStPredictVar Predictor;
    // Время привязки начальных условий
    TimeJD InitTime;
    // Балл. коэффициент по умолчанию
    double Kb;
    // Коэффициент св. давления по умолчанию
    double Kappa;
    // Флаг уточнения Кb
    bool IsKbVar;
    // Флаг уточнения Кappa
    bool IsKappaVar;
};

///////////////////////////////////////////////////////////////////////////////
#endif
