///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Класс вычисления Эфемерид тел (Солнечной системы).
// Копирование и присваивание объектов этого класса невозможно.
///////////////////////////////////////////////////////////////////////////////
#include "ephemeris.h"
#include "Defines.h"
///////////////////////////////////////////////////////////////////////////////
//- Определенные ниже массивы отражают формат файла эфемерид
//- и не должны изменяться
//---------------------------------------------------------------------------
// Порядки полиномов для планет.
static int Order[] = {11, 11, 14, 9, 8, 7, 7, 5, 5, 11, 14};
//---------------------------------------------------------------------------
// Интервалы для полиномов в сутках по планетам.
static int DInterval[] = {8, 32, 16, 32, 32, 32, 32, 32, 32, 4, 32};
//---------------------------------------------------------------------------
#define J1950Date 2433282.5
///////////////////////////////////////////////////////////////////////////////
Ephemeride::Ephemeride()
{
    XCoefs = NULL;
    ClearCash();
}

//---------------------------------------------------------------------------
Ephemeride::~Ephemeride()
{
    UnLoadData();
}

//---------------------------------------------------------------------------
// Освободить память
void Ephemeride::UnLoadData()
{
    if (XCoefs != NULL)
        delete XCoefs;
    ClearCash();
    XCoefs = NULL;
}

//---------------------------------------------------------------------------
// Загрузить данные для указанного интервала
bool Ephemeride::LoadData(const TimeJD& s, const TimeJD& e)
{
    TimeJD stdb(s), etdb(e);
    TAssert(stdb <= etdb);

    if (XCoefs != NULL && stdb == StartTime && etdb == EndTime)
        return true;

    UnLoadData();

    return ReadData(stdb, etdb);
}

//---------------------------------------------------------------------------
bool Ephemeride::ReadData(const TimeJD& start, const TimeJD& end)
{
    TAssert(sizeof(double) == 8);

    // Здесь вычисляются следующие параметры:
    //  D32Len - Длина порции файла, покрывающей 32 суток
    //  BodyOffset - Смещение коэффициентов тела относительно начала 32-дневной порции
    //  CfLen - Число коэффициентов тела в одной порции

    int i;
    D32Len = 2;
    for (i = 0; i < (int)(sizeof(Order) / sizeof(int)); i++) {
        long cfl = 3 * (Order[i] + 1) * (32 / DInterval[i]);
        if (i == (int)BodyID) {
            BodyOffset = D32Len;
            CfLen = cfl;
        }
        D32Len += cfl;
    }

    // Открытие исходного файла

    FILE* f = fopen(FName, "rb");
    if (!f)
        return false;

    // Чтение из файла времени начала первой 32-суточной порции

    double cur[2];
    fread(cur, 2 * sizeof(double), 1, f);

    // Установка времени начала данных по эфемеридам

    int d = (int)floor(cur[1]) + (int)floor(J1950Date);
    double ff = cur[1] - floor(cur[1]) + J1950Date - floor(J1950Date);
    FileStartTime.Init(d, ff);

    // Определяем конечное время для данных в файле
    fseek(f, 0, SEEK_END);
    long flen = ftell(f);
    d += 32 * flen / D32Len / sizeof(double);
    FileEndTime.Init(d, ff);

    // Вычисление начального времени для чтения (отсчитывается
    // в сутках от времени первой поpции)

    int sdays = (int)floor(((start.GetDays() - FileStartTime.GetDays() +
            start.GetFraction() - FileStartTime.GetFraction())
        / DInterval[BodyID])) * DInterval[BodyID];
    StartTime.Init(sdays + FileStartTime.GetDays(), FileStartTime.GetFraction());

    // Вычисление конечного времени для чтения (отсчитывается
    // в сутках от времени первой поpции)

    int edays = (int)ceil(((end.GetDays() - FileStartTime.GetDays() +
            +end.GetFraction() - FileStartTime.GetFraction()) /
        DInterval[BodyID])) * DInterval[BodyID];
    EndTime.Init(edays + FileStartTime.GetDays(), FileStartTime.GetFraction());

    // Вычисление количества памяти, необходимой для хранения всех коэффициентов
    // и ее заказ

    int lmem = (edays - sdays) / DInterval[BodyID] * (Order[BodyID] + 1);
    XCoefs = new double[lmem * 3];
    YCoefs = XCoefs + lmem;
    ZCoefs = YCoefs + lmem;

    // Цикл чтения

    double *xp, *yp, *zp;
    xp = XCoefs;
    yp = YCoefs;
    zp = ZCoefs;
    for (d = sdays; d < edays; d += DInterval[BodyID]) {
        int np = d / 32;
        long offset = np * D32Len + BodyOffset + (d % 32) / DInterval[BodyID] * 3 * (Order[BodyID] + 1);
        fseek(f, offset * sizeof(double), SEEK_SET);
        fread(xp, sizeof(double) * (Order[BodyID] + 1), 1, f);
        fread(yp, sizeof(double) * (Order[BodyID] + 1), 1, f);
        fread(zp, sizeof(double) * (Order[BodyID] + 1), 1, f);
        xp += Order[BodyID] + 1;
        yp += Order[BodyID] + 1;
        zp += Order[BodyID] + 1;
    }

    if (ferror(f))
        return false;

    fclose(f);
    return true;
}

//---------------------------------------------------------------------------
void Ephemeride::CalculatePolynom(double t_cheb, double* coeff, double* coord,
                                  double* coord_dt, double* coord_dt2) const
{
    double b2 = 0.0;
    double b1 = 0.0;
    double b;
    double b2_s = 0.0;
    double b1_s = 0.0;
    double b_s = 0.0;
    double b2_ss = 0.0;
    double b1_ss = 0.0;
    double b_ss = 0.0;

    for (int i = Order[BodyID]; i > 0; --i) {
        b = coeff[i] - b2 + 2.0 * b1 * t_cheb;
        b2 = b1;
        b1 = b;

        if (coord_dt || coord_dt2) {
            b_s = 2.0 * (b1_s * t_cheb + b2) - b2_s;
            b2_s = b1_s;
            b1_s = b_s;
        }

        if (coord_dt2) {
            b_ss = 2.0 * (b1_ss * t_cheb + 2.0 * b2_s) - b2_ss;
            b2_ss = b1_ss;
            b1_ss = b_ss;
        }
    }

    if (coord)
        *coord = coeff[0] - b2 + b1 * t_cheb;
    if (coord_dt)
        *coord_dt = b1 + b_s * t_cheb - b2_s;
    if (coord_dt2)
        *coord_dt2 = 2.0 * b1_s + b1_ss * t_cheb - b2_ss;
}

//---------------------------------------------------------------------------
void Ephemeride::GetPosition(const TimeJD& t, Vect3* pos,
                             Vect3* vel, Vect3* acc) const
{
    // TAssert(XCoefs != NULL);
    // TAssert(tdb >= StartTime && tdb <= EndTime);

    if (GetFromCash(t, pos, vel, acc))
        return;

    TimeJD tdb(t);

    double diff = (tdb - StartTime).GetDouble();
    double tch = 2.0 * fmod(diff / DInterval[BodyID], 1.0) - 1.0;
    int offset = (int)floor(diff / DInterval[BodyID]) * (Order[BodyID] + 1);

    CalculatePolynom(tch, XCoefs + offset,
                     pos == NULL ? NULL : &pos->x,
                     vel == NULL ? NULL : &vel->x,
                     acc == NULL ? NULL : &acc->x);

    CalculatePolynom(tch, YCoefs + offset,
                     pos == NULL ? NULL : &pos->y,
                     vel == NULL ? NULL : &vel->y,
                     acc == NULL ? NULL : &acc->y);

    CalculatePolynom(tch, ZCoefs + offset,
                     pos == NULL ? NULL : &pos->z,
                     vel == NULL ? NULL : &vel->z,
                     acc == NULL ? NULL : &acc->z);

    double dt_scale = 2 / (86400.0 * DInterval[BodyID]);
    double dt2_scale = dt_scale * dt_scale * 1.e3;
    if (pos) {
        pos->x *= 1.e-3;
        pos->y *= 1.e-3;
        pos->z *= 1.e-3;
    }

    if (vel) {
        vel->x *= dt_scale;
        vel->y *= dt_scale;
        vel->z *= dt_scale;
    }

    if (acc) {
        acc->x *= dt2_scale;
        acc->y *= dt2_scale;
        acc->z *= dt2_scale;
    }

    PutToCash(t, pos, vel, acc);

    return;
}

//---------------------------------------------------------------------------
// Запрос из кэша
bool Ephemeride::GetFromCash(const TimeJD& t, Vect3* pos, Vect3* vel, Vect3* acc) const
{
    if (pos != NULL) {
        if (Tp == t)
            *pos = Cp;
        else
            return false;
    }

    if (vel != NULL) {
        if (Tv == t)
            *vel = Cv;
        else
            return false;
    }

    if (acc != NULL) {
        if (Ta == t)
            *acc = Ca;
        else
            return false;
    }

    return true;
}

//---------------------------------------------------------------------------
// Запоминание в кэше
void Ephemeride::PutToCash(const TimeJD& t, Vect3* pos, Vect3* vel, Vect3* acc) const
{
    if (pos != NULL) {
        Tp = t;
        Cp = *pos;
    }

    if (vel != NULL) {
        Tv = t;
        Cv = *vel;
    }

    if (acc != NULL) {
        Ta = t;
        Ca = *acc;
    }
}

///////////////////////////////////////////////////////////////////////////////
