///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Класс вычисления Эфемерид тел (Солнечной системы).
// Копирование и присваивание объектов этого класса невозможно.
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_EPHEMERIDE_H
#define BNOTOOL_EPHEMERIDE_H
///////////////////////////////////////////////////////////////////////////////
#include "TM.h"
#include "DVectors.h"
#include "IAstroBase.h"

///////////////////////////////////////////////////////////////////////////////
// Класс вычисления Эфемерид тел (Солнечной системы).
class Ephemeride : public IEphemeride
{
public:
    // Перечисление кодов результата инициализации объекта.
    // Перечисление тел Солнечной системы.
    enum EnBodyID
    {
        idMercury = 0,
        idVenus = 1,
        idEarthMoon = 2,
        idMars = 3,
        idJupiter = 4,
        idSaturn = 5,
        idUranus = 6,
        idNeptune = 7,
        idPluto = 8,
        idMoon = 9,
        idSun = 10,
        idUnknown = -1
    };

    // Конструктор.
    Ephemeride();
    // Конструктор с указанием параметров
    Ephemeride(const char* fname, EnBodyID body)
    {
        XCoefs = NULL;
        Init(fname, body);
    };

    // Деструктор.
    virtual ~Ephemeride();

    // Инициализация
    virtual void Init(const char* fname, EnBodyID body)
    {
        UnLoadData();
        strcpy(FName, fname);
        BodyID = body;
    };

    // Загрузить данные для указанного интервала
    virtual bool LoadData(const TimeJD& s, const TimeJD& e);

    // Освободить память
    virtual void UnLoadData();

    // Вычислить положение тела.
    virtual void GetPosition(const TimeJD& t, Vect3* pos = NULL,
                             Vect3* vel = NULL, Vect3* acc = NULL) const;
    // Получить идентификатор тела.
    EnBodyID GetBodyID() const
    {
        return BodyID;
    }

    // Получить начальное время интервала использования Эфемерид.
    const TimeJD& GetStartTime() const
    {
        return StartTime;
    };
    // Получить конечное время интервала использования Эфемерид.
    const TimeJD& GetEndTime() const
    {
        return EndTime;
    };
    // Получить начальное время полного интервала использования Эфемерид.
    const TimeJD& GetFileStartTime() const
    {
        return FileStartTime;
    };
    // Получить конечное время полного интервала использования Эфемерид.
    const TimeJD& GetFileEndTime() const
    {
        return FileEndTime;
    };

protected:
    // Идентификатор тела.
    EnBodyID BodyID;
    // Имя файла с данными
    char FName[256];

private:
    // Длина порции файла, покрывающей 32 суток.
    long D32Len;
    // Смещение коэффициентов тела относительно начала 32-дневной порции.
    long BodyOffset;
    // Число коэффициентов тела в одной порции.
    long CfLen;
    // Начало интервала.
    TimeJD StartTime;
    // Конец интервала.
    TimeJD EndTime;
    // Начало интервала в исходном файле.
    TimeJD FileStartTime;
    // Конец интервала в исходном файле.
    TimeJD FileEndTime;
    // Коэффициенты интерполяционных многочленов.
    double *XCoefs, *YCoefs, *ZCoefs;

    // Прочитать данные из файла, соответствующие заданному
    // интервалу времени.
    bool ReadData(const TimeJD& start, const TimeJD& end);
    // Вычислить значение координаты, скорость и ускорение в заданный
    // момент времени.
    void CalculatePolynom(double t_cheb, double* coeff, double* coord,
                          double* coord_dt, double* coord_dt2) const;

    // Реализация кэша 
    // Переменные для хранения
    mutable TimeJD Tp, Tv, Ta;
    mutable Vect3 Cp, Cv, Ca;
    // Запрос из кэша
    bool GetFromCash(const TimeJD& t, Vect3* pos, Vect3* vel, Vect3* acc) const;
    // Запоминание в кэше
    void PutToCash(const TimeJD& t, Vect3* pos, Vect3* vel, Vect3* acc) const;
    // Очистка кэша
    void ClearCash()
    {
        Tp = Tv = Ta = TimeJD(0, 0.0);
    };

    // Запретить копирование.
    Ephemeride(const Ephemeride& src);
    // Запретить присваивание.
    Ephemeride& operator=(const Ephemeride& src);
};

///////////////////////////////////////////////////////////////////////////////
// Класс эфемерид планет
///////////////////////////////////////////////////////////////////////////////
class PlanetEphemeride : public Ephemeride
{
public:
    // Конструктор.
    PlanetEphemeride() : Ephemeride()
    {
    };
    // Конструктор с указанием параметров
    PlanetEphemeride(const char* fname, EnBodyID body) : Ephemeride(fname, body)
    {
    };

    // Вернуть идентификатор системы в которой вычисляются положения
    virtual TCoordSysId GetCS() const
    {
        return csIdJ2000S;
    };
};

///////////////////////////////////////////////////////////////////////////////
// Класс эфемерид Луны
///////////////////////////////////////////////////////////////////////////////
class MoonEphemeride : public Ephemeride
{
public:
    // Конструктор.
    MoonEphemeride() : Ephemeride()
    {
        BodyID = idMoon;
    };
    // Конструктор с указанием параметров
    MoonEphemeride(const char* fname) : Ephemeride(fname, idMoon)
    {
    };

    // Вернуть идентификатор системы в которой вычисляются положения
    virtual TCoordSysId GetCS() const
    {
        return csIdJ2000;
    };

    // Инициализация
    virtual void Init(const char* fname)
    {
        Ephemeride::Init(fname, idMoon);
    };
    // Инициализация
    virtual void Init(const char* fname, EnBodyID body)
    {
        Ephemeride::Init(fname, idMoon);
    };
};

///////////////////////////////////////////////////////////////////////////////
// Класс эфемерид Солнца
///////////////////////////////////////////////////////////////////////////////
class SunEphemeride : public IEphemeride
{
public:
    // Вернуть идентификатор системы в которой вычисляются положения
    virtual TCoordSysId GetCS() const
    {
        return csIdJ2000S;
    };

    // Загрузить данные для указанного интервала
    virtual bool LoadData(const TimeJD& s, const TimeJD& e)
    {
        return true;
    };
    // Освободить память
    virtual void UnLoadData()
    {
    }

    // Вычислить положение тела.
    virtual void GetPosition(const TimeJD& t, Vect3* pos = NULL,
                             Vect3* vel = NULL, Vect3* acc = NULL) const
    {
        if (pos != NULL)
            *pos = 0.0;
        if (vel != NULL)
            *vel = 0.0;
        if (acc != NULL)
            *acc = 0.0;
    }
};

///////////////////////////////////////////////////////////////////////////////
// Класс эфемерид Солнца
///////////////////////////////////////////////////////////////////////////////
class EarthEphemeride : public IEphemeride
{
public:
    // Вернуть идентификатор системы в которой вычисляются положения
    virtual TCoordSysId GetCS() const
    {
        return csIdJ2000;
    };

    // Загрузить данные для указанного интервала
    virtual bool LoadData(const TimeJD& s, const TimeJD& e)
    {
        return true;
    };
    // Освободить память
    virtual void UnLoadData()
    {
    }

    // Вычислить положение тела.
    virtual void GetPosition(const TimeJD& t, Vect3* pos = NULL,
                             Vect3* vel = NULL, Vect3* acc = NULL) const
    {
        if (pos != NULL)
            *pos = 0.0;
        if (vel != NULL)
            *vel = 0.0;
        if (acc != NULL)
            *acc = 0.0;
    }
};

///////////////////////////////////////////////////////////////////////////////
#endif
