///////////////////////////////////////////////////////////////////////////////
// Библиотека алгоритмов небесной механики
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
#include "Kepler.h"
#include "IAstroBase.h"
///////////////////////////////////////////////////////////////////////////////
// Класс орбитальных элементов Кеплера
///////////////////////////////////////////////////////////////////////////////
// Конструктор по умолчанию
//-----------------------------------------------------------------------------
Kepler::Kepler()
{
    IAstroBase* ab = GetAstroBase();

    Mu = ab->GetEarthMu();
}

//------------------------------------------------------------------------------
// Установка координат по декартовым
//------------------------------------------------------------------------------
void Kepler::SetXYZ(const Vect6& src)
{
    double r2 = src.Len2();
    double r = sqrt(r2);
    double v2 = src.LenV2();
    double rrv = src.x * src.vx + src.y * src.vy + src.z * src.vz;
    double rv = rrv / r;
    double c2 = r2 * v2 - rrv * rrv;
    double p = c2 / Mu;
    double c = sqrt(c2);

    A = 1.0 / (2.0 / r - v2 / Mu);
    CosI = (src.x * src.vy - src.y * src.vx) / c;

    double sini = sqrt(1.0 - CosI * CosI);
    double eCosNu = p / r - 1.0;
    double eSinNu = (p * rv) / c;

    double eps = 1.0 - 1.0e-15;
    if (fabs(CosI) > eps) {
        if (CosI > 0)
            CosI = 1.0;
        else
            CosI = -1.0;
        Ex = eCosNu;
        Ey = -eSinNu;
        Omega = atan2(src.y, src.x);
    }
    else {
        double sinU = src.z / (r * sini);
        double cosU = (r * src.vz - src.z * rv) / (c * sini);
        Ex = eCosNu * cosU + eSinNu * sinU;
        Ey = eCosNu * sinU - eSinNu * cosU;
        Omega = atan2(src.y * cosU - src.x * CosI * sinU, src.x * cosU + src.y * CosI * sinU);
        U = atan2(sinU, cosU);
        if (U < 0.0)
            U += Pi2;
    };

    if (Omega < 0.0)
        Omega += Pi2;
}

//------------------------------------------------------------------------------
// Получение декартовых координат и скоростей
//------------------------------------------------------------------------------
bool Kepler::GetXYZ(Vect6& dst) const
{
    double p = Ex * Ex + Ey * Ey;
    p = A * (1.0 - p);
    double sinU = sin(U);
    double cosU = cos(U);
    double r = p / (1.0 + Ex * cosU + Ey * sinU);
    double c = sqrt(Mu * p) / r;
    double b = ((Ex * sinU - Ey * cosU) / p) * c;
    double sini = sqrt(1.0 - CosI * CosI);
    double sinOM = sin(Omega);
    double cosOM = cos(Omega);

    dst.x = r * (cosU * cosOM - CosI * sinU * sinOM);
    dst.y = r * (cosU * sinOM + CosI * sinU * cosOM);
    dst.z = r * sini * sinU;
    dst.vx = b * dst.x - c * (sinU * cosOM + CosI * cosU * sinOM);
    dst.vy = b * dst.y - c * (sinU * sinOM - CosI * cosU * cosOM);
    dst.vz = b * dst.z + c * sini * cosU;

    return true;
}

//------------------------------------------------------------------------------
// Вычисляемые параметры
//------------------------------------------------------------------------------
// Период орбиты
//------------------------------------------------------------------------------
double Kepler::GetPeriod() const
{
    return Pi2 * sqrt(A * A * A / Mu);
}

//------------------------------------------------------------------------------
// Радиус перицентра
//------------------------------------------------------------------------------
double Kepler::GetRP() const
{
    return A * (1.0 - sqrt(Ex * Ex + Ey * Ey));
}

//------------------------------------------------------------------------------
// Радиус апоцентра
//------------------------------------------------------------------------------
double Kepler::GetRA() const
{
    return A * (1.0 + sqrt(Ex * Ex + Ey * Ey));
}

//------------------------------------------------------------------------------
// Высота перицентра над эллипсоидом
//------------------------------------------------------------------------------
double Kepler::GetHP(const TGeometryModel* geom) const
{
    if (geom == NULL)
        geom = &GetAstroBase()->GetPlanet(plIdEarth)->Geometry;

    double ey2 = Ey * Ey;
    double ecc2 = Ex * Ex + ey2;
    double ecc = sqrt(ecc2);
    double re = geom->Re;
    double obl = geom->alpha;
    return A * (1.0 - ecc) - re * (1.0 - obl * (1.0 - CosI * CosI) * ey2 / ecc2);
}

//------------------------------------------------------------------------------
// Высота апоцентра над эллипсоидом
//------------------------------------------------------------------------------
double Kepler::GetHA(const TGeometryModel* geom) const
{
    if (geom == NULL)
        geom = &GetAstroBase()->GetPlanet(plIdEarth)->Geometry;

    double ey2 = Ey * Ey;
    double ecc2 = Ex * Ex + ey2;
    double ecc = sqrt(ecc2);
    double re = geom->Re;
    double obl = geom->alpha;
    return A * (1.0 + ecc) - re * (1.0 - obl * (1.0 - CosI * CosI) * ey2 / ecc2);
}

//------------------------------------------------------------------------------
// Эксцентриситет орбиты
//------------------------------------------------------------------------------
double Kepler::GetEcc() const
{
    return sqrt(Ex * Ex + Ey * Ey);
}

//------------------------------------------------------------------------------
// Аргумент перицентра
//------------------------------------------------------------------------------
double Kepler::GetArgP() const
{
    double buf = atan2(Ey, Ex);
    if (buf < 0.0)
        buf += Pi2;
    return buf;
}

//------------------------------------------------------------------------------
// Наклонение относительно плоскости Лапласа
//------------------------------------------------------------------------------
double Kepler::GetLaplaceInc() const
{
    double sin_i = 1.0 - CosI * CosI;
    sin_i = sqrt(sin_i);
    return acos(CosI * cos(LaplaceAngle) + sin_i * sin(LaplaceAngle) * cos(Omega));
}

//------------------------------------------------------------------------------
// Долгота восходящего узла в плоскости Лапласа
//------------------------------------------------------------------------------
double Kepler::GetLaplaceLAN() const
{
    double sin_i = 1.0 - CosI * CosI;
    sin_i = sqrt(sin_i);

    double longitude = atan2(sin_i * sin(Omega),
                             sin_i * cos(LaplaceAngle) * cos(Omega) -
                             CosI * sin(LaplaceAngle));
    if (longitude < 0.0)
        longitude += Pi2;

    return longitude;
}

//------------------------------------------------------------------------------
// Время движения от восх. узла до текущей точки
//------------------------------------------------------------------------------

TimeJD Kepler::GetTimeFromAN(const TimeJD& base) const
{
    TimeJD tp = base;
    tp -= GetTimeFromAN() / 86.4;

    return tp;
}

double Kepler::GetTimeFromAN() const
{
    double e2 = Ex * Ex + Ey * Ey;
    double one_ex = 1.0 + Ex;
    double alph1 = (Ex + e2) / one_ex;
    double one_e2 = fabs(1.0 - e2);
    double sqrt_one_e2 = sqrt(one_e2);
    double alph2 = sqrt_one_e2 * Ey / one_ex;
    double a3 = A * A * A;
    double alph0 = sqrt(a3 / Mu);
    double s_u = sin(U);
    double c_u = cos(U);
    double bet = 1.0 / (1.0 + sqrt_one_e2);
    double ex_bet_e2 = bet * e2 + Ex;
    double F = U - 2.0 * atan2(bet * (ex_bet_e2 * s_u + Ey * (1.0 - c_u)),
                               (1.0 + bet * (ex_bet_e2 * c_u + Ey * s_u + Ex)));

    return alph0 * (F - alph1 * sin(F) - alph2 * (1.0 - cos(F)));
}

//------------------------------------------------------------------------------
// Истиная аномалия
//------------------------------------------------------------------------------
double Kepler::GetTrueAnomaly() const
{
    double ret = U - GetArgP();
    while (ret < 0.0)
        ret += Pi2;

    return ret;
}

//------------------------------------------------------------------------------
// чЩЮЙУМЕОЙЕ ЙОФЕТЧБМБ ЧТЕНЕОЙ НЕЦДХ РЕТЙГЕОФТПН Й ХЪМПН (бМЗПТЙФН EL23)
//------------------------------------------------------------------------------
double Kepler::GetTimeP() const
{
    double e = GetEcc();
    double ap = GetArgP();
    double eo = -2.0 * atan(sqrt((1.0 - e) / (1.0 + e)) * tan(ap * 0.5));
    double n = sqrt(Mu / (A * A * A));
    double dt = (e * sin(eo) - eo) / n;
    return dt;
}

double Kepler::GetSPeriod(int NF) const
{
    double esinw2 = Ey * Ey;
    double E2 = esinw2 + Ex * Ex;
    double E20 = 1.0 - E2;
    double EZ = sqrt(E20);
    double EZ2 = Ey / EZ;
    double cosi2 = CosI * CosI;

    // Semi-Draconic period
    double q1 = (EZ2 * E20 / cosi2 + atan(EZ2)) / Pi;
    return (0.5 - NF * q1) * GetPeriod();
}

///////////////////////////////////////////////////////////////////////////////
