/*     ----------------------------------------------------------------


                                sgp4ext.cpp


    this file contains basic functions needed for the sgp4 test routine.

                            companion code for
               fundamentals of astrodyanmics and applications
                                    2001
                              by david vallado

       (h)               email valladodl@worldnet.att.net
       (w) 719-573-2600, email dvallado@agi.com

       *****************************************************************

    current :
              17 jun 04  david vallado
                           3rd edition baseline
    changes :
              14 may 01  david vallado
                           2nd edition baseline
              23 nov 87  david vallado
                           original baseline

       ----------------------------------------------------------------

                              implementation

       ----------------------------------------------------------------      */

#include <math.h>
#include <stdio.h>
#include "sgp4ext.h"

/*------------------------------------------------------------------------------
|
|                           procedure days2mdhms
|
|  this procedure converts the day of the year, days, to the equivalent month
|    day, hour, minute and second.
|
|  algorithm     : set up array for the number of days per month
|                  find leap year - use 1900 because 2000 is a leap year
|                  loop through a temp value while the value is < the days
|                  perform int conversions to the correct day and month
|                  convert remainder into h m s using type conversions
|
|  author        : david vallado                  303-344-6037    1 mar 2001
|
|  inputs          description                    range / units
|    year        - year                           1900 .. 2100
|    days        - julian day of the year         0.0  .. 366.0
|
|  outputs       :
|    mon         - month                          1 .. 12
|    day         - day                            1 .. 28,29,30,31
|    hr          - hour                           0 .. 23
|    min         - minute                         0 .. 59
|    sec         - second                         0.0 .. 59.999
|
|  locals        :
|    dayofyr     - day of year
|    temp        - temporary extended values
|    inttemp     - temporary int value
|    i           - index
|    lmonth[12]  - int array containing the number of days per month
|
|  coupling      :
|    none.
|
 -----------------------------------------------------------------------------*/
void days2mdhms
(
    int year, double days,
    int& mon, int& day, int& hr, int& min, double& sec
)
{
    int i, inttemp, dayofyr;
    double temp;
    int lmonth[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

    dayofyr = int(days);

    /* ----------------- find month and day of month ---------------- */
    if (((year % 4) == 0) && (((year % 400) != 0) || ((year % 400) == 0)))
        lmonth[1] = 29;

    i = 1;
    inttemp = 0;
    while ((dayofyr > inttemp + lmonth[i - 1]) && (i < 12)) {
        inttemp += lmonth[i - 1];
        i++;
    }
    mon = i;
    day = dayofyr - inttemp;

    /* ----------------- find hours minutes and seconds ------------- */
    temp = (days - dayofyr) * 24.0;
    hr = int(temp);
    temp = (temp - hr) * 60.0;
    min = int(temp);
    sec = (temp - min) * 60.0;
}

/*------------------------------------------------------------------------------
|
|                           procedure finddays
|
|  this procedure finds the fractional days through a year given the year,
|    month, day, hour, minute and second.
|
|  algorithm     : set up array for the number of days per month
|                  find leap year - use 1900 because 2000 is a leap year
|                  check for a leap year
|                  loop to find the elapsed days in the year
|
|  author        : david vallado                  303-344-6037
|
|  inputs          description                    range / units
|    year        - year                           1900 .. 2100
|    mon         - month                          1 .. 12
|    day         - day                            1 .. 28,29,30,31
|    hr          - hour                           0 .. 23
|    min         - minute                         0 .. 59
|    sec         - second                         0.0 .. 59.999
|
|  outputs       :
|    days        - day of year plus fraction of a
|                    day                          days
|
|  locals        :
|    lmonth      - length of months of year
|    i           - index
|
|  coupling      :
|    none.
|
|  references    :
|    vallado       2001, 202-203, ex 3-12
|
 -----------------------------------------------------------------------------*/
void finddays
(
    int year, int month, int day, int hr, int min,
    double sec, double& days
)
{
    int lmonth[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
    int i;

    if (((year - 1900) % 4) == 0)
        lmonth[1] = 29;

    i = 1;
    days = 0.0;
    while ((i < month) && (i < 12)) {
        days += lmonth[i - 1];
        i++;
    }
    days += day + hr / 24.0 + min / 1440.0 + sec / 86400.0;
}

/*------------------------------------------------------------------------------
|
|                           function gstime
|
|  this function finds the greenwich sidereal time.
|
|  author        : david vallado                  303-344-6037    1 mar 2001
|
|  inputs          description                    range / units
|    jdut1       - julian date in ut1             days from 4713 bc
|
|  outputs       :
|    gstime      - greenwich sidereal time        0 to 2pi rad
|
|  locals        :
|    temp        - temporary variable for doubles   rad
|    tut1        - julian centuries from the
|                  jan 1, 2000 12 h epoch (ut1)
|
|  coupling      :
|    doublemod     - mod function for double variables
|
|  references    :
|    vallado       2001, 191, eq 3-45
|
-----------------------------------------------------------------------------*/
#define pi 3.14159265358979323846

double gstime(double jdut1)
{
    const double twopi = 2.0 * pi;
    const double deg2rad = pi / 180.0;
    double temp, tut1;

    tut1 = (jdut1 - 2451545.0) / 36525.0;
    temp = -6.2e-6 * tut1 * tut1 * tut1 + 0.093104 * tut1 * tut1 +
        (876600.0 * 3600 + 8640184.812866) * tut1 + 67310.54841; // sec
    temp = modfunc(temp * deg2rad / 240.0, twopi); //360/86400 = 1/240, to deg, to rad

    /* ------------------------ check quadrants --------------------- */
    if (temp < 0.0)
        temp += twopi;

    //kbn  if (show == 'y')
    //    printf("%18.12f\n", tut1);

    return temp;
}

/*------------------------------------------------------------------------------
|
|                           procedure invjday
|
|  this procedure finds the year, month, day, hour, minute and second
|  given the julian date. tu can be ut1, tdt, tdb, etc.
|
|  algorithm     : set up starting values
|                  find leap year - use 1900 because 2000 is a leap year
|                  find the elapsed days through the year in a loop
|                  call routine to find each individual value
|
|  author        : david vallado                  303-344-6037    1 mar 2001
|
|  inputs          description                    range / units
|    jd          - julian date                    days from 4713 bc
|
|  outputs       :
|    year        - year                           1900 .. 2100
|    mon         - month                          1 .. 12
|    day         - day                            1 .. 28,29,30,31
|    hr          - hour                           0 .. 23
|    min         - minute                         0 .. 59
|    sec         - second                         0.0 .. 59.999
|
|  locals        :
|    days        - day of year plus fractional
|                  portion of a day               days
|    tu          - julian centuries from 0 h
|                  jan 0, 1900
|    temp        - temporary double values
|    leapyrs     - number of leap years from 1900
|
|  coupling      :
|    days2mdhms  - finds month, day, hour, minute and second given days and year
|
|  references    :
|    vallado       2001, 203-205, alg 22, ex 3-13
|
 -----------------------------------------------------------------------------*/
void invjday
(
    double jd,
    int& year, int& mon, int& day,
    int& hr, int& min, double& sec
)
{
    int leapyrs;
    double days, tu, temp;

    /* --------------- find year and days of the year --------------- */
    temp = jd - 2415019.5;
    tu = temp / 365.25;
    year = 1900 + int(tu);
    leapyrs = int((year - 1901) * 0.25);
    // nudge by 8.64x10-7 sec to get even outputs
    days = temp - ((year - 1900) * 365.0 + leapyrs) + 0.00000000001;

    /* ------------ check for case of beginning of a year ----------- */
    if (days < 1.0) {
        year = year - 1;
        leapyrs = int((year - 1901) * 0.25);
        days = temp - ((year - 1900) * 365.0 + leapyrs);
    }

    /* ----------------- find remaing data  ------------------------- */
    days2mdhms(year, days, mon, day, hr, min, sec);
    sec = sec - 0.00000086400;
}

/*------------------------------------------------------------------------------
|
|                           procedure julianday
|
|  this procedure finds the julian date given the year, month, day, and time.
|    the julian date is defined by each elapsed day since noon, jan 1, 4713 bc.
|
|  algorithm     : calculate the answer in one step for efficiency
|
|  author        : david vallado                  303-344-6037    1 mar 2001
|
|  inputs          description                    range / units
|    year        - year                           1900 .. 2100
|    mon         - month                          1 .. 12
|    day         - day                            1 .. 28,29,30,31
|    hr          - universal time hour            0 .. 23
|    min         - universal time min             0 .. 59
|    sec         - universal time sec             0.0 .. 59.999
|    whichtype   - julian or gregorian calender   'j' or 'g'
|
|  outputs       :
|    jd          - julian date                    days from 4713 bc
|
|  locals        :
|    b           - var to aid gregorian dates
|
|  coupling      :
|    none.
|
|  references    :
|    vallado       2001, 186-188, alg 14, ex 3-14
|
 -----------------------------------------------------------------------------*/
void julianday
(
    int year, int mon, int day, int hr, int min, double sec,
    double& jd
)
{
    jd = 367.0 * year -
        int((7 * (year + int((mon + 9) / 12))) * 0.25) +
        int(275 * mon / 9) +
        day + 1721013.5 +
        ((sec / 60.0 + min) / 60.0 + hr) / 24.0; // ut in days
    // - 0.5*sgn(100.0*year + mon - 190002.5) + 0.5;
}

void juliandayall
(
    int year, int mon, int day, int hr, int min, double sec,
    char whichtype, double& jd
)
{
    double b;

    if (mon <= 2) {
        year = year - 1;
        mon = mon + 12;
    }
    if (whichtype == 'j')
        /* --------- use for julian calender, every 4 years --------- */
        b = 0.0;
    else {
        /* ---------------------- use for gregorian ----------------- */
        b = 2 - int(year * 0.01) + int(int(year * 0.01) * 0.25);
        jd = int(365.25 * (year + 4716)) +
            int(30.6001 * (mon + 1)) +
            day + b - 1524.5 +
            ((sec / 60.0 + min) / 60.0 + hr) / 24.0; // ut in days
    }
}

/*------------------------------------------------------------------------------|
|                           procedure lstime
|
|  this procedure finds the local sidereal time at a given location.
|
|  author        : david vallado                  303-344-6037    1 mar 2001
|
|  inputs          description                    range / units
|    lon         - site longitude (west -)        -2pi to 2pi rad
|    jdut1       - julian date in ut1             days from 4713 bc
|
|  outputs       :
|    lst         - local sidereal time            0.0 to 2pi rad
|    gst         - greenwich sidereal time        0.0 to 2pi rad
|
|  locals        :
|    none.
|
|  coupling      :
|    doublemod       mod function for double variables
|    gstime        finds the greenwich sidereal time
|
|  references    :
|    vallado       2001, 189-193. alg 15, ex 3-5
|
 -----------------------------------------------------------------------------*/
void lstime(double lon, double jdut1, double& lst, double& gst)
{
    const double twopi = 2.0 * pi;

    gst = gstime(jdut1);
    lst = lon + gst;

    /* ------------------------ check quadrants --------------------- */
    lst = modfunc(lst, twopi);
    if (lst < 0.0)
        lst = lst + twopi;
}

/*------------------------------------------------------------------------------
|
|                           procedure newtonm
|
|  this procedure performs the newton rhapson iteration to find the
|    eccentric anomaly given the mean anomaly.  the true anomaly is also
|    calculated.
|
|  author        : david vallado                  303-344-6037    1 mar 2001
|
|  inputs          description                    range / units
|    ecc         - eccentricity                   0.0 to
|    m           - mean anomaly                   -2pi to 2pi rad
|
|  outputs       :
|    e0          - eccentric anomaly              0.0 to 2pi rad
|    nu          - true anomaly                   0.0 to 2pi rad
|
|  locals        :
|    e1          - eccentric anomaly, next value  rad
|    sinv        - sine of nu
|    cosv        - cosine of nu
|    ktr         - index
|    r1r         - cubic roots - 1 to 3
|    r1i         - imaginary component
|    r2r         -
|    r2i         -
|    r3r         -
|    r3i         -
|    s           - variables for parabolic solution
|    w           - variables for parabolic solution
|
|  coupling      :
|    atan2       - arc tangent function which also resloves quadrants
|    cubic       - solves a cubic polynomial
|    power       - raises a base number to an arbitrary power
|    sinh        - hyperbolic sine
|    cosh        - hyperbolic cosine
|    sgn         - returns the sign of an argument
|
|  references    :
|    vallado       2001, 72-75, alg 2, ex 2-1
|
 -----------------------------------------------------------------------------*/
void newtonm(double ecc, double m, double& e0, double& nu)
{
    const int numiter = 50;
    const double small = 0.00000001; // small value for tolerances

    double e1, sinv, cosv, r1r = 0.0;
    int ktr;

    /* -------------------------- hyperbolic  ----------------------- */
    if ((ecc - 1.0) > small) {
        /* ------------  initial guess ------------- */
        if (ecc < 1.6)
            if (((m < 0.0) && (m > -pi)) || (m > pi))
                e0 = m - ecc;
            else
                e0 = m + ecc;
        else if ((ecc < 3.6) && (fabs(m) > pi)) // just edges)
            e0 = m - sgn(m) * ecc;
        else
            e0 = m / (ecc - 1.0); // best over 1.8 in middle
        ktr = 1;
        e1 = e0 + ((m - ecc * sinh(e0) + e0) / (ecc * cosh(e0) - 1.0));
        while ((fabs(e1 - e0) > small) && (ktr <= numiter)) {
            e0 = e1;
            e1 = e0 + ((m - ecc * sinh(e0) + e0) / (ecc * cosh(e0) - 1.0));
            ktr++;
        }
        /* ---------  find true anomaly  ----------- */
        sinv = -(sqrt(ecc * ecc - 1.0) * sinh(e1)) / (1.0 - ecc * cosh(e1));
        cosv = (cosh(e1) - ecc) / (1.0 - ecc * cosh(e1));
        nu = atan2(sinv, cosv);
    }
    else {
        /* ---------------------- parabolic ------------------------- */
        if (fabs(ecc - 1.0) < small) {
            //kbn      cubic(1.0 / 3.0, 0.0, 1.0, -m, r1r, r1i, r2r, r2i, r3r, r3i);
            e0 = r1r;
            //kbn      if (fileout != null)
            //        fprintf(fileout, "roots %11.7f %11.7f %11.7f %11.7f %11.7f %11.7f\n",
            //                          r1r, r1i, r2r, r2i, r3r, r3i);
            /*
                  s  = 0.5 * (halfpi - atan(1.5 * m));
                  w  = atan(power(tan(s), 1.0 / 3.0));
                  e0 = 2.0 * cot(2.0* w );
            */
            ktr = 1;
            nu = 2.0 * atan(e0);
        }
        else {
            /* --------------------- elliptical --------------------- */
            if (ecc > small) {
                /* ------------  initial guess ------------- */
                if (((m < 0.0) && (m > -pi)) || (m > pi))
                    e0 = m - ecc;
                else
                    e0 = m + ecc;
                ktr = 1;
                e1 = e0 + (m - e0 + ecc * sin(e0)) / (1.0 - ecc * cos(e0));
                while ((fabs(e1 - e0) > small) && (ktr <= numiter)) {
                    ktr++;
                    e0 = e1;
                    e1 = e0 + (m - e0 + ecc * sin(e0)) / (1.0 - ecc * cos(e0));
                }
                /* ---------  find true anomaly  ----------- */
                sinv = (sqrt(1.0 - ecc * ecc) * sin(e1)) / (1.0 - ecc * cos(e1));
                cosv = (cos(e1) - ecc) / (1.0 - ecc * cos(e1));
                nu = atan2(sinv, cosv);
            }
            else {
                /* --------------------- circular --------------------- */
                ktr = 0;
                nu = m;
                e0 = m;
            }
        }
    }
    if (ktr > numiter)
        printf("newtonrhapson not converged in %3d iterations\n", numiter);
}


double modfunc(double x, double y)
{
    if (y)
        return x - int(x / y) * y;
    else
        return 0;
}

double sgn(double x)
{
    if (x)
        return fabs(x) / x;
    else
        return 0;
}
