///////////////////////////////////////////////////////////////////////////////
// Библиотека программных инструментов
// (с) ИПМ им.М.В.Келдыша, ЗАО "Проекттехника"
///////////////////////////////////////////////////////////////////////////////
// Средства чтения и анализа файловых каталогов
///////////////////////////////////////////////////////////////////////////////
#ifndef BNOTOOL_TDIR_H
#define BNOTOOL_TDIR_H
#include	<sys/types.h>
#include	<sys/stat.h>

#if defined(_WIN32) || defined(_WINDOWS) || defined(_Windows)
#include	"tdirent.h"
#define S_ISREG(mode) ((mode) == _S_IFREG)
#define S_ISDIR(mode) ((mode) == _S_IFDIR)
#define S_ISBLK(mode) false
#define S_ISCHR(mode) ((mode) == _S_IFCHR)
#define S_ISLNK(mode) false
#define S_ISFIFO(mode) ((mode) == _S_IFIFO)
#define S_ISSOCK(mode) false
#else
#include	<dirent.h>
#endif

#include "tstring.h"
///////////////////////////////////////////////////////////////////////////////
class TDirIter;

///////////////////////////////////////////////////////////////////////////////
// Описание одного элемента каталога
///////////////////////////////////////////////////////////////////////////////
class TDirEntry
{
public:
    TDirEntry(TString _p) : path(_p), n("")
    {
    }

    TString Path() const
    {
        return path;
    }

    TString Name() const
    {
        return n;
    }

private:
    TString path;
    const char* n;
    friend class TDirIter;
};

///////////////////////////////////////////////////////////////////////////////
// Класс чтения каталога
///////////////////////////////////////////////////////////////////////////////
class TDirIter
{
public:
    TDirIter(const char* dirSpec) : fd(0), eye(dirSpec)
    {
        init(dirSpec);
    };

    ~TDirIter()
    {
        close();
    }

    const TDirEntry& Item() const
    {
        return eye;
    }

    operator bool() const
    {
        return fd != 0;
    };

    void operator++()
    {
        next();
    }

private:
    DIR* fd;
    TDirEntry eye;

    void next();
    void init(const char* dirSpec);
    void close();
};

///////////////////////////////////////////////////////////////////////////////
// Класс информации о файле
///////////////////////////////////////////////////////////////////////////////
class TFileInfo
{
public:
    // конструктор из DirEntry
    TFileInfo(const TDirEntry& i) : gotStat(false), useName(true)
    {
        init(i);
    }

    // конструктор из дескриптора открытого файла
    TFileInfo(int _fd) : gotStat(false), useName(false)
    {
        fd = _fd;
    }

    // regular file ?
    bool IsReg()
    {
        prepIf();
        return S_ISREG(st.st_mode);
    }

    // directory ?
    bool IsDir()
    {
        prepIf();
        return S_ISDIR(st.st_mode);
    }

    // block-oriented device ?
    bool IsBlk()
    {
        prepIf();
        return S_ISBLK(st.st_mode);
    }

    // character-oriented device ?
    bool IsChr()
    {
        prepIf();
        return S_ISCHR(st.st_mode);
    }

    // sym-link ?
    bool IsLnk()
    {
        prepIf();
        return S_ISLNK(st.st_mode);
    }

    // FIFO ?
    bool IsFifo()
    {
        prepIf();
        return S_ISFIFO(st.st_mode);
    }

    // PIPE ?
    bool IsPipe()
    {
        return IsFifo();
    }

    // socket ?
    bool IsSock()
    {
        prepIf();
        return S_ISSOCK(st.st_mode);
    }

    // file size
    off_t Length() const { return Size(); }
    off_t Size() const { return st.st_size; }
private:
    TString name;
    int fd;
    bool gotStat;
    bool useName;
    struct stat st;

    void prepIf();

    void init(const TDirEntry& i)
    {
        name = i.Path() + "/" + i.Name();
    }
};

///////////////////////////////////////////////////////////////////////////////
inline void TDirIter::init(const char* dirSpec)
{
    fd = opendir(dirSpec);
    next();
}

//-----------------------------------------------------------------------------
inline void TDirIter::close()
{
    if (fd) {
        closedir(fd);
        fd = 0;
    }
}

//-----------------------------------------------------------------------------
inline void TDirIter::next()
{
    if (fd) {
        struct dirent* p = readdir(fd);
        if (p)
            eye.n = p->d_name;
        else
            close();
    }
}

//-----------------------------------------------------------------------------
inline void TFileInfo::prepIf()
{
    if (gotStat)
        return;
    if (stat(name, &st)) {
        // error
        return;
    }
    gotStat = true;
}

///////////////////////////////////////////////////////////////////////////////
#endif
