#include "AscNodeOrbit.h"
#include "Propagator.h"
#include "DrPeriod.h"
#include "GreenwichFrame.h"
#include "ECIFrame.h"
#include "XMLFunctions.h"
#include "mathconst.h"

static double GetZ(const Vect6& v) { return v.z; }
static double GetVZ(const Vect6& v) { return v.vz; }
static double GetY(const Vect6& v) { return v.y; }
static double GetVY(const Vect6& v) { return v.vy; }

AscNodeOrbit::AscNodeOrbit()
{
};

AscNodeOrbit::AscNodeOrbit(const PhasePoint6D& p)
{
    m_date = p.T;
    calculate_params(p);
}

void AscNodeOrbit::calc_asc_node(Propagator& prop, const TimeJD& date, const bool vzmode, const bool gcs_node, PhasePoint6D& result) const
{
    // std::string date_s = DateToStr(date);
    // std::cout << date_s << "\n";

    if (vzmode) {
        //if(!GetAscNodeYZ(*prop.GetPredictor()->GetPredictor(), prop.GetPredictor()->GetTime(), result, GetZ, GetVZ)){
        //	return false;
        //};
               
        SInitOrbit ascNodeOrbit;
        if (!jump_arg_of_lat(date, prop, 0, gcs_node, ascNodeOrbit)) {
            throw Exp() << "internal error in ascending node method 1";
        }

        result = ascNodeOrbit.GetPhasePointJ2000();

        // std::string date_s = DateToStr(result.T);
        // std::cout << date_s << "\n";

        if (DaysInterval(date, result.T) * 86400 > 600) {
            // TODO sometimes date lower result.T less than 600 sec
            //SInitOrbit ascNodeOrbit;
            //if (!jump_arg_of_lat(date, prop, 0, gcs_node, ascNodeOrbit)) {
            //    throw Exp() << "internal error in ascending node method 2";
            //}
            throw Exp() << "internal error in ascending node method 2 " << DaysInterval(date, result.T)*86400;
        }

    }
    else {
        if (!GetAscNodeYZ(*prop.GetPredictor()->GetPredictor(), prop.GetPredictor()->GetTime(), result, GetY, GetVY)) {
            throw Exp() << "internal error in ascending node method 3";
        }
    }
}

void AscNodeOrbit::calc_desc_node(Propagator& prop, const TimeJD& date, const bool vzmode, const bool gcs_node, PhasePoint6D& result) const
{
    if (vzmode) {
        //if(!GetAscNodeYZ(*prop.GetPredictor()->GetPredictor(), prop.GetPredictor()->GetTime(), result, GetZ, GetVZ)){
        //	return false;
        //};

        SInitOrbit descNodeOrbit;
        if (!jump_arg_of_lat(date, prop, 180 * deg2rad, gcs_node, descNodeOrbit, 1)) {
            throw Exp() << "internal error in descending node method 1";
        }

        result = descNodeOrbit.GetPhasePointJ2000();

        //std::cout << DateToStr(date) <<  " " << DateToStr(descNodeOrbit.GetPhasePointJ2000().T)  << "\n" ;
        if (DaysInterval(result.T, date) * 86400 > 1) {
            // if (!jump_arg_of_lat(date, prop, 180 * deg2rad, gcs_node, descNodeOrbit, 1)) {
            //     throw Exp() << "internal error in descending node method 2";
            // }
            // std::cout << DateToStr(date) << " " << DateToStr(descNodeOrbit.GetPhasePointJ2000().T) << "\n";
            throw Exp() << "internal error in descending node method 2 " << DaysInterval(date, result.T) * 86400;
        }

    }
    else {
        if (!GetAscNodeYZ(*prop.GetPredictor()->GetPredictor(), prop.GetPredictor()->GetTime(), result, GetY, GetVY)) {
            throw Exp() << "internal error in descending node method 3";
        }
    }
}

void AscNodeOrbit::calculate(Propagator& prop, const TimeJD& date, const bool vzmode, const bool gcs_node, bool calc_params)
{
    m_is_desc = false;

    PhasePoint6D j2000AscNodePoint;
    calc_asc_node(prop, date, vzmode, gcs_node, j2000AscNodePoint);
    calc_draconic_period(prop, j2000AscNodePoint.T, vzmode, gcs_node, m_drperiod, m_drperiodDt);
    
    m_date = j2000AscNodePoint.T;
        
    if (calc_params) {
        calculate_params(j2000AscNodePoint);
    }
}

void AscNodeOrbit::calculate_desc(Propagator& prop, const TimeJD& date, const bool vzmode, const bool gcs_node, bool calc_params)
{
    m_is_desc = true;

    PhasePoint6D j2000DescNodePoint;
    calc_desc_node(prop, date, vzmode, gcs_node, j2000DescNodePoint);
    m_date = j2000DescNodePoint.T;
        
    if (calc_params) {
        calculate_params(j2000DescNodePoint);
    }
}

void AscNodeOrbit::calculate_params(const PhasePoint6D& j2000AscNodePoint)
{
    PhasePoint6D GCSAscNodePoint = j2000AscNodePoint;
    GCSAscNodePoint.Convert(GreenwichFrame());

    PhasePoint6D ECIAscNodePoint = j2000AscNodePoint;
    ECIAscNodePoint.Convert(ECIFrame(j2000AscNodePoint.T));

    GeoCoordsPrec gcc;
    gcc.SetXYZ(GCSAscNodePoint.CoordsVel);

    m_latitude = gcc.Phi;
    m_longitude = gcc.Lam;
    m_altitude = gcc.H;

    m_kdj2000 = get_kep_data(j2000AscNodePoint.CoordsVel);
    m_kdeci = get_kep_data(ECIAscNodePoint.CoordsVel);
    m_j2000 = j2000AscNodePoint.CoordsVel;
    m_gcs = GCSAscNodePoint.CoordsVel;
}

double AscNodeOrbit::get_kep_time(double TA, double e, double w) const
{
    double tgE2 = tan(TA / 2.0) * sqrt((1 - e) / (1 + e));
    double E = atan(tgE2) * 2.0;
    return (E - e * sin(E)) / w;
}

double AscNodeOrbit::get_step(const Kepler& kep, double dst_u, bool positive) const
{
    return get_step(kep.GetPeriod(), kep.GetArgP(), kep.GetEcc(), kep.U, dst_u, positive);
}

double AscNodeOrbit::get_step(double period, double arg_of_per, double e, double u, double dst_u, bool positive) const
{
    double w = 86400 / period / 1000 * Pi2 / 1440 / 60;
    double t = get_kep_time((u - arg_of_per), e, w);
    double t2 = get_kep_time((dst_u - arg_of_per), e, w);
    double step = t2 - t;

    if (step > period * 1000 / 2) {
        step = step - period * 1000;
    }
    else if (step < -period * 1000 / 2) {
        step = period * 1000 + step;
    }

    if(positive && step < 0) {
        step += period * 1000;
    }

    return step;
}

bool AscNodeOrbit::jump_arg_of_lat(const TimeJD& date, Propagator& prop, double dstU, bool gcs_node, SInitOrbit& result, int type) const
{
    const int maxInter = 20;
    const double u_tol = 0.001;
    const double time_tol = 0.001;
    int iter = 0;

    PhasePoint6D p;
    prop.Propagate(date, p);
        
    //PhasePoint6D pt = p;
    //if (gcs_node) {
    //    pt.Convert(METEFrame(pt.T));
    //}
    
    if(type == 1) {
        Kepler kep;
        kep.SetXYZ(p.CoordsVel);
        const double step = get_step(kep, dstU, true);
        prop.Propagate(ShiftDate(p.T, step), p);
    }else {
        Kepler kep;
        kep.SetXYZ(p.CoordsVel);
        prop.Propagate(kep.GetTimeFromAN(p.T), p);
    }

    while (true) {
    
        // std::cout << DateToStr(p.T) << "\n";

        Kepler kep;
        if (gcs_node) {
            PhasePoint6D gcs = p;
            gcs.Convert(ECIFrame(gcs.T));
            kep.SetXYZ(gcs.CoordsVel);
        }
        else {
            kep.SetXYZ(p.CoordsVel);
        }
                
        if (fabs(Distance(0, kep.U, 0, dstU)) * rad2deg < u_tol) {
            break;
        }
        
        double step = get_step(kep, dstU, false);
        
        if (fabs(step) < time_tol) {
            break;
        }

        if (iter++ > maxInter) {
            return false;
        }

        prop.Propagate(ShiftDate(p.T, step), p);
    }
        
    prop.Propagate(p.T, result);

    return true;
}

// Определение драконического периода и его падения
// период в тыс.с 
// изм. периода в тыс.с/виток
void AscNodeOrbit::calc_draconic_period(Propagator& prop, const TimeJD& asc_node_date, const bool vzmode, const bool gcs_node, double& dracp, double& vdracp) const
{
    PhasePoint6D firstAscNode;
    prop.Propagate(asc_node_date, firstAscNode);

    const TimeJD firstAscDate = firstAscNode.T;

    J2000Frame f;
    if (firstAscNode.GetFrame()->GetType() != csIdJ2000) {
        firstAscNode.Convert(f);
    }

    Kepler kp;
    kp.SetXYZ(firstAscNode.CoordsVel);
    double period = kp.GetPeriod();

    PhasePoint6D secondAscNode;
    calc_asc_node(prop, ShiftDate(firstAscNode.T, -.5 * period * 1000), vzmode, gcs_node, secondAscNode);

    PhasePoint6D thirdAscNode;
    calc_asc_node(prop, ShiftDate(secondAscNode.T, -.5 * period * 1000), vzmode, gcs_node, thirdAscNode);

    double dracp1 = (firstAscDate - secondAscNode.T).GetDouble() * 86400 / 1000;
    double dracp2 = (secondAscNode.T - thirdAscNode.T).GetDouble() * 86400 / 1000;

    dracp = dracp1;
    vdracp = dracp1 - dracp2;
};

KepData AscNodeOrbit::get_kep_data(const Vect6& pos)
{
    Kepler kp;
    kp.SetXYZ(pos);

    KepData kepData;
    kepData.Init(kp);

    return kepData;
}

void AscNodeOrbit::set_rev(int rev)
{
    m_rev = rev;
}

void AscNodeOrbit::set_age(double age_days)
{
    m_age = age_days;
}

static void PutValue(std::ostream& ss, std::string conName, const Vect6& value)
{
    ss << "<" << conName << ">";

    PutValue(ss, "x", value.x);
    PutValue(ss, "y", value.y);
    PutValue(ss, "z", value.z);
    PutValue(ss, "vx", value.vx);
    PutValue(ss, "vy", value.vy);
    PutValue(ss, "vz", value.vz);

    ss << "</" << conName << ">";
}

void AscNodeOrbit::ToXml(std::ostream& outs) const
{
    PutValue(outs, "date", m_date);
    PutValue(outs, "latitude", m_latitude);
    PutValue(outs, "longitude", m_longitude);
    PutValue(outs, "altitude", m_altitude);
    PutValue(outs, "drperiod", m_drperiod);
    PutValue(outs, "drperiodDt", m_drperiodDt);
    PutValue(outs, "kepJ2000", m_kdj2000);
    PutValue(outs, "kepEci", m_kdeci);
    PutValue(outs, "vecJ2000", m_j2000);
}

