#include "Propagator.h"
#include "Consts.h"
#include "TextFunctions.h"
#include "METEFrame.h"
#include "J2000Frame.h"
#include "TimeFunctions.h"
#include "DVectors.h"
#include "PhasePoint.h"
#include "AstroBaseModels.h"
#include "BNOAstroBase.h"

Propagator::Propagator()
    :
    m_predictor(nullptr),
    m_has_orbit(false)
{
}

Propagator::Propagator(const SInitOrbit& initOrbit, bool variates)
    :
    m_predictor(nullptr),
    m_has_orbit(false)
{
    Init(initOrbit.GetPropagatorData(), initOrbit, variates);
}

Propagator::Propagator(ITraject* predictor)
    :
    m_predictor(predictor),
    m_has_orbit(false)
{
}

bool Propagator::Init(const SInitOrbit& initOrbit, bool variates)
{
    return Init(initOrbit.GetPropagatorData(), initOrbit, variates);
}

bool Propagator::Init(const PropInfoBase& propInfo, const SInitOrbit& initOrbit, bool variates)
{
    m_variates = variates;

    if (propInfo.GetType() == proptypes::ST_PROP) {
        if (!m_variates) {
            InitStPredictor(m_stPredictor, initOrbit, propInfo);
            m_stPredictor.GetPredictor()->SetAccuracy(
                propInfo.get_integrator_cfg().getEps(),
                propInfo.get_integrator_cfg().getStepInit(),
                propInfo.get_integrator_cfg().getStepMin(),
                propInfo.get_integrator_cfg().getPrecItfr());
            m_predictor = &m_stPredictor;
        }
        else {
            InitStvPredictor(m_stvPredictor, initOrbit, propInfo);
            m_stvPredictor.GetPredictor()->SetAccuracy(
                propInfo.get_integrator_cfg().getEps(),
                propInfo.get_integrator_cfg().getStepInit(),
                propInfo.get_integrator_cfg().getStepMin(),
                propInfo.get_integrator_cfg().getPrecItfr());
            m_predictor = &m_stvPredictor;
        }
    }
    else if (propInfo.GetType() == proptypes::KEPLER_PROP) {
        //	if (!variates){
        InitKepPredictor(m_kpPredictor, initOrbit);
        m_predictor = &m_kpPredictor;
        //	}else{
        //		TAssert(0);
        //	}
    }
    else if (propInfo.GetType() == proptypes::PROGNOZ_PROP) {
        //	if (!variates){
        InitPrognozPredictor(m_prPredictor, initOrbit);
        m_predictor = &m_prPredictor;
        //	}else{
        //		TAssert(0);
        //	}
    }
    else if (propInfo.GetType() == proptypes::SGP_PROP) {
        //	if (!variates){
        InitSGPPredictor(m_sgpPredictor, initOrbit);
        m_predictor = &m_sgpPredictor;
        //	}else{
        //		TAssert(0);
        //	}
    }
    else {
        TAssert(0);
    }

    m_initOrbit = initOrbit;
    m_has_orbit = true;

    return true;
}

void Propagator::InitSGPPredictor(SGPTrj& trj, const SInitOrbit& initOrbit)
{
    TimeTLE time(initOrbit.GetDate());

    const double xpdotp = 229.1831180523293;

    TLESet set;

    set.Satnum = -1;
    set.EpochYr = time.Year;
    set.EpochDays = time.Day;
    set.NDot = initOrbit.GetTleData().GetNDot() / (xpdotp * 1440.0);
    set.NDDot = initOrbit.GetTleData().GetNDDot() / (xpdotp * 1440.0 * 1440);
    set.BStar = initOrbit.GetTleData().GetBStar();
    set.ElementNum = initOrbit.GetTleData().GetSolution();
    set.Inclo = initOrbit.GetTleData().GetInclination() * deg2rad;
    set.Omegao = initOrbit.GetTleData().GetRAAN() * deg2rad;
    set.Ecco = initOrbit.GetTleData().GetEccentricity();
    set.Argpo = initOrbit.GetTleData().GetArgOfPerigee() * deg2rad;
    set.Mo = initOrbit.GetTleData().GetMeanAnomaly() * deg2rad;
    set.No = initOrbit.GetTleData().GetMeanMotion() / xpdotp;
    set.Set();

    trj.SetTime(set.GetTime());
    trj.Init(set);
}

void Propagator::InitStPredictor(TrjNum& trj, const SInitOrbit& initOrbit, const PropInfoBase& propInfo)
{
    SetMotionModel(propInfo, *trj.GetPredictor());

    Vect3 pos = initOrbit.GetCoortinatesJ2000();
    Vect3 vel = initOrbit.GetVelocitiesJ2000();

    DVector u(6);

    u[0] = pos.x;
    u[1] = pos.y;
    u[2] = pos.z;
    u[3] = vel.x;
    u[4] = vel.y;
    u[5] = vel.z;

    trj.SetTime(initOrbit.GetDate());
    trj.SetKOProp(initOrbit.GetS(), initOrbit.GetKappa());
    trj.Init(u);
}

void Propagator::InitStvPredictor(TrjNumVar& trj, const SInitOrbit& initOrbit, const PropInfoBase& propInfo)
{
    SetMotionModel(propInfo, *trj.GetPredictor());

    Vect3 pos = initOrbit.GetCoortinatesJ2000();
    Vect3 vel = initOrbit.GetVelocitiesJ2000();
    const CovMtx& covmtx = initOrbit.GetCovMtx();

    bool varkb = covmtx.HasKb() && covmtx.Dim() >= 7;
    bool varsp = covmtx.HasSp() && covmtx.Dim() >= 7;

    int size = !varkb && !varsp ? 6 : (varkb && varsp ? 8 : 7);
    DVector u(size);

    u[0] = pos.x;
    u[1] = pos.y;
    u[2] = pos.z;
    u[3] = vel.x;
    u[4] = vel.y;
    u[5] = vel.z;

    if (varkb && !varsp) {
        u[6] = initOrbit.GetS();
    }
    else if (!varkb && varsp) {
        u[6] = initOrbit.GetKappa();
    }
    else if (varkb && varsp) {
        u[6] = initOrbit.GetS();
        u[7] = initOrbit.GetKappa();
    }

    trj.SetTime(initOrbit.GetDate());
    trj.SetKOProp(initOrbit.GetS(), initOrbit.GetKappa(), varkb, varsp);
    trj.Init(u);
}

void Propagator::InitKepPredictor(KeplerTrj& trj, const SInitOrbit& initOrbit)
{
    Vect3 pos = initOrbit.GetCoortinatesJ2000();
    Vect3 vel = initOrbit.GetVelocitiesJ2000();

    DVector u(6);

    u[0] = pos.x;
    u[1] = pos.y;
    u[2] = pos.z;
    u[3] = vel.x;
    u[4] = vel.y;
    u[5] = vel.z;

    trj.SetTime(initOrbit.GetDate());
    trj.Init(u);
}

void Propagator::InitPrognozPredictor(PrognozTrj& trj, const SInitOrbit& initOrbit)
{
    Vect3 pos = initOrbit.GetCoortinatesECI();
    Vect3 vel = initOrbit.GetVelocitiesECI();

    DVector u(6);

    u[0] = pos.x;
    u[1] = pos.y;
    u[2] = pos.z;
    u[3] = vel.x;
    u[4] = vel.y;
    u[5] = vel.z;

    trj.SetPeriodChangePerRev(initOrbit.GetPeriodChangePerRev());
    trj.SetTime(initOrbit.GetDate());
    trj.Init(u);
}

void Propagator::SetMotionModel(const PropInfoBase& PropInfoBase, DigPredict& predictor)
{
    MotionMode mode;
    SetMotionModel(PropInfoBase, predictor, mode);
    predictor.MMode = mode;
}

void Propagator::SetMotionModel(const PropInfoBase& PropInfoBase, DigPredictVar& predictor)
{
    MotionMode mode;
    SetMotionModel(PropInfoBase, predictor, mode);
    predictor.MMode = mode;
}

void Propagator::SetMotionModel(const PropInfoBase& PropInfoBase, IPredict& predictor, MotionMode& mode)
{
    mode.IsAtm = IsAtmModelUse(PropInfoBase.GetAtmModel().GetModel().c_str());
    mode.LightMode = (TTypeLight)LightModelIdByKey(PropInfoBase.GetLightModel().c_str());
    mode.GrM = PropInfoBase.GetEarthModel().GetHarmonicsM();
    mode.GrN = PropInfoBase.GetEarthModel().GetHarmonicsN();

    if (mode.IsAtm) {
        ((BNOAstroBase*)GetAstroBase())->SetAtmosphereModel((TAtmosphereId)AtmModelIdByKey(PropInfoBase.GetAtmModel().GetModel().c_str()), (bool)KpModelIdByKey(PropInfoBase.GetAtmModel().GetKpModel().c_str()));
    }
    ((BNOAstroBase*)GetAstroBase())->SetEarthGravityModel((TGravityModelId)GravModelIdByKey(PropInfoBase.GetEarthModel().GetModel().c_str()));

    if (PropInfoBase.IsSun()) {
        mode.AddPlanet(plIdSun);
    }

    if (PropInfoBase.IsMoon()) {
        mode.AddPlanet(plIdMoon);
    }
}

bool Propagator::Propagate(const TimeJD& jd, SInitOrbit& res) const
{
    if (m_initOrbit.GetCovMtx().HasMtx() && m_variates) {
        PhasePoint6D pos;
        DSymmMatrix m;
        const CovMtx& srccov = m_initOrbit.GetCovMtx();
        if (Propagate(*m_predictor, jd, srccov.GetMtx(), pos, m)) {
            res = SInitOrbit(
                pos.CoordsVel,
                jd,
                m_initOrbit.GetS(),
                m_initOrbit.GetKappa(),
                m_initOrbit.GetPeriodChangePerRev(),
                m_initOrbit.GetPropagatorData(),
                CovMtx(m, srccov.HasKb(), srccov.HasSp()),
                0,
                m_initOrbit.GetId()
            );
            return true;
        }
        else {
            return false;
        }
    }
    else if (m_initOrbit.GetCovMtx().HasMtx() && !m_variates) {
        PhasePoint6D pos;
        if (Propagate(*m_predictor, jd, pos)) {
            res = SInitOrbit(
                pos.CoordsVel,
                jd,
                m_initOrbit.GetS(),
                m_initOrbit.GetKappa(),
                m_initOrbit.GetPeriodChangePerRev(),
                m_initOrbit.GetPropagatorData(),
                CovMtx(), //m_initOrbit.GetCovMtx(), 07/05/2019  m_initOrbit.GetCovMtx() removed. Why this was here if covariances not expexted?
                0,
                m_initOrbit.GetId()
            );
            return true;
        }
        else {
            return false;
        }
    }
    else {
        PhasePoint6D pos;
        if (Propagate(*m_predictor, jd, pos)) {
            res = SInitOrbit(
                pos.CoordsVel,
                jd,
                m_initOrbit.GetS(),
                m_initOrbit.GetKappa(),
                m_initOrbit.GetPeriodChangePerRev(),
                m_initOrbit.GetPropagatorData(),
                CovMtx(),
                0,
                m_initOrbit.GetId()
            );
            return true;
        }
        else {
            return false;
        }
    }
}

bool Propagator::Propagate(const TimeJD& jd, const IFrame& frame, PhasePoint6D& ret) const
{
    PhasePoint6D point;

    if (Propagate(*m_predictor, jd, frame, point)) {
        ret = point;
        return true;
    }
    else {
        return false;
    }
}

bool Propagator::Propagate(const TimeJD& jd, PhasePoint6D& ret)
{
    PhasePoint6D point;

    if (Propagate(*m_predictor, jd, point)) {
        ret = point;
        return true;
    }
    else {
        return false;
    }
}

bool Propagator::Propagate(const TimeJD& jd, PhasePoint6D& ret, DSymmMatrix& covm) const
{
    PhasePoint6D point;

    if (Propagate(*m_predictor, jd, m_initOrbit.GetCovMtx().GetMtx(), point, covm)) {
        ret = point;
        return true;
    }
    else {
        return false;
    }
}

bool Propagator::Propagate(ITraject& predictor, const TimeJD& jd, PhasePoint6D& ret)
{
    J2000Frame fr;

    Vect6 vec;
    if (predictor.Predict(jd, static_cast<const IFrame&>(fr), vec)) {
        ret = PhasePoint6D(fr, vec, jd);
        return true;
    }
    else {
        return false;
    }
}

bool Propagator::Propagate(ITraject& predictor, const TimeJD& jd, const IFrame& frame, PhasePoint6D& ret)
{
    Vect6 vec;
    if (predictor.Predict(jd, frame, vec)) {
        ret = PhasePoint6D(frame, vec, jd);
        return true;
    }
    else {
        return false;
    }
}

bool Propagator::Propagate(ITraject& predictor, const TimeJD& jd, const DSymmMatrix& srccov, PhasePoint6D& ret, DSymmMatrix& dstcov)
{
    J2000Frame fr;
    Vect6 vec;

    DMatrix mtx;
    if (predictor.PredictMatrix(jd, static_cast<const IFrame&>(fr), vec, mtx)) {
        ret = PhasePoint6D(fr, vec, jd);
        dstcov.RectSymmTRecmt(mtx, srccov);
        return true;
    }
    else {
        return false;
    }
}

bool Propagator::has_covmtx() const
{
    return m_has_orbit ? m_initOrbit.GetCovMtx().HasMtx() : false;
}
