#include "StateVector.h"
#include <utility>
#include "XMLFunctions.h"
#include "GreenwichFrame.h"
#include "PhasePoint.h"
#include "METEFrame.h"
#include "Kepler.h"
#include "ECIFrame.h"

#define BASEFRAME J2000Frame()

class SOVector
{
public:
    SOVector()
    {
    }

    bool Load(myXML::Contain* con)
    {
        std::map<std::string, myXML::Contain*> items;
        GetChildren(con, items);

        m_s = 0;
        m_k = 0;

        LoadXML(items, "X", m_vec.x);
        LoadXML(items, "Y", m_vec.y);
        LoadXML(items, "Z", m_vec.z);
        LoadXML(items, "VX", m_vec.vx);
        LoadXML(items, "VY", m_vec.vy);
        LoadXML(items, "VZ", m_vec.vz);
        LoadXML(items, "S", m_s, false);
        LoadXML(items, "K", m_k, false);

        return true;
    }

    Vect6 GetVec() const { return m_vec; }
    double GetBC() const { return m_s; }
    double GetSPC() const { return m_k; }

private:
    Vect6 m_vec;
    double m_s;
    double m_k;
};

SInitOrbit::SInitOrbit()
    :
    m_coordinates(0, 0, 0),
    m_velocities(0, 0, 0),
    m_date(0, 0),
    //m_srcType					(ORBSRCTYPE_UNK),
    m_ballisticCoefficient(0),
    m_solarPresureCoefficient(0)
{
}

SInitOrbit::~SInitOrbit()
{
}

SInitOrbit::SInitOrbit(
    const Vect6 pos, 
    const TimeJD& date, 
    const double ballc, 
    const double solarpc, 
    const double periodChangePerRev, 
    PropInfoBase propagator, 
    CovMtx covmtx, 
    const int rev, 
    TOrbitId baseid)
    :
    m_date(date),
    m_ballisticCoefficient(ballc),
    m_solarPresureCoefficient(solarpc),
    m_periodChangePerRev(periodChangePerRev),
    m_covmtx(std::move(covmtx)),
    //m_srcType					(ORBSRCTYPE_OPT),
    m_rev(rev),
    m_baseid(std::move(baseid)),
    m_propInfo(std::move(propagator)), rev_(rev)
{
    CreateFromPos(pos, "J2000");
}

bool SInitOrbit::CheckFrame(const std::string& frame)
{
    return !(frame != "ECI" && frame != "J2000" && frame != "METE" && frame != "GCS");
}

int SInitOrbit::ParseSrcType(const TOrbitId& id)
{
    int ret;
    std::string sep("::");
    size_t found = id.find(sep);
    if (found != std::string::npos) {
        std::string type = id.substr(found + sep.size());
        if (type == "tle") {
            ret = ORBSRCTYPE_TLE;
        }
        else if (type == "opt") {
            ret = ORBSRCTYPE_OPT;
        }
        else if (type == "optu") {
            ret = ORBSRCTYPE_OPT;
        }
        else {
            ret = ORBSRCTYPE_UNK;
        }
    }
    else {
        ret = ORBSRCTYPE_UNK;
    }

    return ret;
}

bool SInitOrbit::LoadNewStyle(myXML::Contain* con)
{
    std::string frame;

    std::map<std::string, myXML::Contain*> items;
    GetChildren(con, items);

    LoadXML(items, "Id", m_id, false);
    LoadXML(items, "BaseId", m_baseid, false);
    LoadXML(items, "Date", m_date);
    LoadXML(items, "Frame", frame);
    LoadXML(items, "Rev", m_rev, false);
    LoadXML(items, "MotionModel", m_propInfo);

    m_interval = -1;
    LoadXML(items, "Interval", m_interval, false);

    m_periodChangePerRev = 0;
    LoadXML(items, "PeriodChangePerRev", m_periodChangePerRev, false);

    if (!CheckFrame(frame)) {
        throw (Exp() << "Unsupported frame type:" << frame);
    }

    if (GetMapItem(items, std::string("Vector"))) {
        SOVector vec;
        LoadXML(items, "Vector", vec);
        m_covmtx.Load(con, true, true);
        m_ballisticCoefficient = vec.GetBC();
        m_solarPresureCoefficient = vec.GetSPC();
        CreateFromPos(vec.GetVec(), frame);
    }
    else if (GetMapItem(items, std::string("KeplerData"))) {
        m_ballisticCoefficient = 0;
        m_solarPresureCoefficient = 0;
        KepData kepData;
        LoadXML(items, "KeplerData", kepData);
        LoadXML(items, "BallisticCoefficient", m_ballisticCoefficient, false);
        LoadXML(items, "SolarPresureCoefficient", m_solarPresureCoefficient, false);
        m_covmtx.Load(con, true, true);
        CreateFromKepler(kepData, frame);
    }
    else if (GetMapItem(items, std::string("TleData"))) {
        LoadXML(items, "TleData", m_tleData);
    }
    else {
        throw (Exp() << "Unsupported orbit type");
    }

    return true;
}

bool SInitOrbit::Load(myXML::Contain* con)
{
    return LoadNewStyle(con);
}

Vect3 SInitOrbit::GetCoortinatesGSC() const
{
    PhasePoint6D p(static_cast<const IFrame&>(BASEFRAME), GetCoortinates(), GetVelocities(), GetDate());
    p.Convert(static_cast<const IFrame&>(GreenwichFrame()));
    Vect3 ret;
    ret[0] = p.CoordsVel.x;
    ret[1] = p.CoordsVel.y;
    ret[2] = p.CoordsVel.z;
    return ret;
}

Vect3 SInitOrbit::GetVelocitiesGSC() const
{
    PhasePoint6D p(static_cast<const IFrame&>(BASEFRAME), GetCoortinates(), GetVelocities(), GetDate());
    p.Convert(static_cast<const IFrame&>(GreenwichFrame()));
    Vect3 ret;
    ret[0] = p.CoordsVel.vx;
    ret[1] = p.CoordsVel.vy;
    ret[2] = p.CoordsVel.vz;
    return ret;
}

Vect3 SInitOrbit::GetCoortinatesJ2000() const
{
    PhasePoint6D p(static_cast<const IFrame&>(BASEFRAME), GetCoortinates(), GetVelocities(), GetDate());
    p.Convert(static_cast<const IFrame&>(J2000Frame()));
    Vect3 ret;
    ret[0] = p.CoordsVel.x;
    ret[1] = p.CoordsVel.y;
    ret[2] = p.CoordsVel.z;
    return ret;
}

Vect3 SInitOrbit::GetVelocitiesJ2000() const
{
    PhasePoint6D p(static_cast<const IFrame&>(BASEFRAME), GetCoortinates(), GetVelocities(), GetDate());
    p.Convert(static_cast<const IFrame&>(J2000Frame()));
    Vect3 ret;
    ret[0] = p.CoordsVel.vx;
    ret[1] = p.CoordsVel.vy;
    ret[2] = p.CoordsVel.vz;
    return ret;
}

Vect3 SInitOrbit::GetCoortinatesMETE() const
{
    PhasePoint6D p(static_cast<const IFrame&>(BASEFRAME), GetCoortinates(), GetVelocities(), GetDate());
    p.Convert(static_cast<const IFrame&>(METEFrame(GetDate())));
    Vect3 ret;
    ret[0] = p.CoordsVel.x;
    ret[1] = p.CoordsVel.y;
    ret[2] = p.CoordsVel.z;
    return ret;
}

Vect3 SInitOrbit::GetVelocitiesMETE() const
{
    PhasePoint6D p(static_cast<const IFrame&>(BASEFRAME), GetCoortinates(), GetVelocities(), GetDate());
    p.Convert(static_cast<const IFrame&>(METEFrame(GetDate())));
    Vect3 ret;
    ret[0] = p.CoordsVel.vx;
    ret[1] = p.CoordsVel.vy;
    ret[2] = p.CoordsVel.vz;
    return ret;
}

Vect3 SInitOrbit::GetCoortinatesECI() const
{
    PhasePoint6D p(static_cast<const IFrame&>(BASEFRAME), GetCoortinates(), GetVelocities(), GetDate());
    p.Convert(static_cast<const IFrame&>(ECIFrame(GetDate())));
    Vect3 ret;
    ret[0] = p.CoordsVel.x;
    ret[1] = p.CoordsVel.y;
    ret[2] = p.CoordsVel.z;
    return ret;
}

Vect3 SInitOrbit::GetVelocitiesECI() const
{
    PhasePoint6D p(static_cast<const IFrame&>(BASEFRAME), GetCoortinates(), GetVelocities(), GetDate());
    p.Convert(static_cast<const IFrame&>(ECIFrame(GetDate())));
    Vect3 ret;
    ret[0] = p.CoordsVel.vx;
    ret[1] = p.CoordsVel.vy;
    ret[2] = p.CoordsVel.vz;
    return ret;
}

PhasePoint6D SInitOrbit::GetPhasePointGSC() const
{
    return PhasePoint6D(static_cast<const IFrame&>(GreenwichFrame()), GetCoortinatesGSC(), GetVelocitiesGSC(), GetDate());
}

PhasePoint6D SInitOrbit::GetPhasePointJ2000() const
{
    return PhasePoint6D(static_cast<const IFrame&>(J2000Frame()), GetCoortinatesJ2000(), GetVelocitiesJ2000(), GetDate());
}

PhasePoint6D SInitOrbit::GetPhasePointMETE() const
{
    return PhasePoint6D(static_cast<const IFrame&>(METEFrame()), GetCoortinatesMETE(), GetVelocitiesMETE(), GetDate());
}

PhasePoint6D SInitOrbit::GetPhasePointECI() const
{
    return PhasePoint6D(static_cast<const IFrame&>(ECIFrame()), GetCoortinatesECI(), GetVelocitiesECI(), GetDate());
}

void SInitOrbit::CreateFromPos(Vect6 dst, const std::string& frame)
{
    PhasePoint6D pos;

    if (frame == "ECI") {
        pos = PhasePoint6D(static_cast<const IFrame&>(ECIFrame(GetDate())), dst, GetDate());
    }
    else if (frame == "J2000") {
        pos = PhasePoint6D(static_cast<const IFrame&>(J2000Frame()), dst, GetDate());
    }
    else if (frame == "METE") {
        pos = PhasePoint6D(static_cast<const IFrame&>(METEFrame(GetDate())), dst, GetDate());
    }
    else if (frame == "GCS") {
        pos = PhasePoint6D(static_cast<const IFrame&>(GreenwichFrame()), dst, GetDate());
    }

    pos.Convert(static_cast<const IFrame&>(BASEFRAME));

    m_coordinates.x = pos.CoordsVel.x;
    m_coordinates.y = pos.CoordsVel.y;
    m_coordinates.z = pos.CoordsVel.z;

    m_velocities.x = pos.CoordsVel.vx;
    m_velocities.y = pos.CoordsVel.vy;
    m_velocities.z = pos.CoordsVel.vz;

    Kepler kep;
    kep.SetXYZ(Vect6(m_coordinates, m_velocities));
    m_kepData.Init(kep);
}

void SInitOrbit::CreateFromKepler(const KepData& kd, const std::string& frame)
{
    Kepler kep;
    kep.A = kd.GetSemimajorAxis();
    kep.CosI = cos(kd.GetInclination());
    kep.Omega = kd.GetRAAN();
    kep.U = kd.GetArgOfLatitude();
    kep.Ex = kd.GetEccentricity() * cos(kd.GetArgOfPerigee());
    kep.Ey = kd.GetEccentricity() * sin(kd.GetArgOfPerigee());

    Vect6 dst;
    kep.GetXYZ(dst);

    CreateFromPos(dst, frame);
}

/*bool SInitOrbit::IsInitiated() const
{
    return m_srcType != ORBSRCTYPE_UNK;
}*/

/*bool SInitOrbit::AssignPropagator( const PropInfoBases& propagators )
{
    return GetMapItem(propagators,m_propagator,m_propInfo);
}*/

std::map<std::string, double> SInitOrbit::GetCovRvMap() const
{
    return GetCovMtx().GetCovRvMap();
}

std::map<std::string, double> SInitOrbit::GetCovOrbMap() const
{
    return GetCovMtx().GetCovOrbMap(GetPhasePointJ2000().CoordsVel);
}

std::map<std::string, double> SInitOrbit::GetCovRnbMap() const
{
    return GetCovMtx().GetCovRnbMap(GetPhasePointJ2000().CoordsVel);
}

double SInitOrbit::CalcLongitude() const
{
    GeoCoordsPrec geo;
    geo.SetXYZ(GetCoortinatesGSC());
    return geo.Lam;
}

void SInitOrbit::SetMotionModel(const MotionModel& motionModel)
{
    m_propInfo = motionModel;
}

void SInitOrbit::SetDate(const TimeJD& date)
{
    m_date = date;
    CreateFromPos(GetPhasePointJ2000().CoordsVel, "J2000");
}

void loadVec6(const jsonval& parent, Vect6& item)
{
    loadjson(parent, "x", item.x);
    loadjson(parent, "y", item.y);
    loadjson(parent, "z", item.z);
    loadjson(parent, "vx", item.vx);
    loadjson(parent, "vy", item.vy);
    loadjson(parent, "vz", item.vz);
}

void SInitOrbit::load(const jsonval& parent)
{
    std::string frame;

    loadjson(parent, "id", m_id, false);
    loadjson(parent, "date", m_date);

    loadjson(parent, "frame", frame);
    if (!CheckFrame(frame)) {
        throw (Exp() << "Unsupported frame type:" << frame);
    }

    loadjson(parent, "model", m_propInfo);

    m_interval = 0;
    loadjson(parent, "interval", m_interval, false);

    m_periodChangePerRev = 0;
    loadjson(parent, "dprev", m_periodChangePerRev, false);

    if (parent.FindMember("kep") != parent.MemberEnd()) {

        KepData kepData;
        loadjson(parent, "kep", kepData);

        CreateFromKepler(kepData, frame);
    }
    else {
        Vect6 vec;
        loadjson(parent, "vec", vec, &loadVec6);

        CreateFromPos(vec, frame);
    }

    double amr;
    if (loadjson(parent, "amr", amr, false)) {
        m_ballisticCoefficient = m_solarPresureCoefficient = amr * (1.44 * 767E-6);
    }
    else {
        m_ballisticCoefficient = 0;
        loadjson(parent, "bc", m_ballisticCoefficient, false);

        m_solarPresureCoefficient = 0;
        loadjson(parent, "spc", m_solarPresureCoefficient, false);
    }

    loadjson(parent, "cov", m_covmtx, false);
    loadjson(parent, "uncert", m_uncert, false);
}
