#include <cmath>
#include "ephi.h"

int debug = 0;

int bs_lower_bound(const double a[], const int n, const double x)
{
    int l = 0;
    int h = n; // Not n - 1
    while (l < h) {
        const int mid = (l + h) / 2;
        if (x <= a[mid]) {
            h = mid;
        }
        else {
            l = mid + 1;
        }
    }
    return l;
}

void hermite(int index, int ityp, const double* x, const double* y, const double* z, int nmax, int nval, double xp, double* yp, double* zp, int* ircode);

ei::ei()
    :
    jdi(nullptr),
    jdf(nullptr),
    outv(),
    jds(nullptr),
    nv(-1)
{
    for (auto& el : outv) {
        el = nullptr;
    }
}

ei::~ei()
{
    delete[] jdi;
    delete[] jdf;
    delete[] jds;

    for (auto& el : outv) {
        delete[] el;
    }
}

/*-----------------------------------------------------------------------
void cinterp_pvac (jdint, seceph, outvec, backvec,
aberout,aberback, corr12, corr23, ircode)

Interpolate position, velocity, aberration, and relativistic corrections
from cpf file arrays stored in memory.

jdint        Julian date (integer day)
seceph       Julian day  (seconds)
outvec       outv(i,1) ,...,outv(i,6): x,y,z,vx,vy,vz   outbound vector
backvec      backv(i,1),...,outv(i,6): x,y,z,vx,vy,vz   inbound vector
aberout      outbound aberration correction
aberback     inbound aberration correction
corr12       outbound relativistic correction
corr23       inbound relativistic correction
ircode       error code: 0=OK, 1,2=error (time out of range)

Author: R. L. Ricklefs / UT CSR
Date: 1 May, 2003
History:
-----------------------------------------------------------------------*/
int cinterp_pvac(
    const ei& ephi,
    double jdint,
    double seceph,
    double outvec[6],
    const int fvel,
    int* ircode)
{
    double z, vz; /* z is dummy for derivative field */

    const double itime = ((seceph / 86400.e0 - ephi.jdf[0]) + (jdint - ephi.jdi[0])) * 86400.e0;

    /* 3 or 6 elements? Handle velocity if present */
    const int pindex = bs_lower_bound((double*)ephi.jds, ephi.nv, itime);

    for (int j = 0; j < 6; j++) {
        /* Out */
        /* Position */
        if (j < 3) {
            if (fabs(ephi.outv[j + 3][0]) > 1.e-10) {
                /* Use velocities if they exists (hermite) */
                hermite(pindex, 2, ephi.jds, &ephi.outv[j][0], &ephi.outv[j + 3][0], ephi.nv, 10, itime, &outvec[j], &vz, ircode);
            }
            else {
                /* Otherwise, ignore them (fvel = -1) or create them (fvel = 1) */
                hermite(pindex, -fvel, ephi.jds, &ephi.outv[j][0], &z, ephi.nv, 10, itime, &outvec[j], &outvec[j + 3], ircode);
            }
        }
            /* Velocity */
            /* Interpolate if they exists */
        else if (fabs(ephi.outv[j][0]) > 1.e-10) {
            hermite(pindex, 1, ephi.jds, &ephi.outv[j][0], &z, ephi.nv, 10, itime, &outvec[j], &vz, ircode);
        }

        if (*ircode != 0) {
            return *ircode;
        }
    }

    return 0;
}
