#include "interpolate.h"
#include "Propagator.h"
#include <algorithm>

static double get_step(const double period, const double ecc)
{
    if (ecc < 0.1) {
        return std::max((double)(period / 100.0 * 60), 1.0);
    }

    if (ecc < 0.2) {
        return std::max((double)(period / 200.0 * 60), 1.0);
    }

    if (ecc < 0.4) {
        return std::max((double)(period / 400.0 * 60), 1.0);
    }

    if (ecc < 0.6) {
        return std::max((double)(period / 600.0 * 60), 1.0);
    }

    return std::max((double)(period / 1000.0 * 60), 1.0);
}

static double get_step(const SInitOrbit& orbit)
{
    const double period = orbit.GetKeplerData().GetPeriod() * 1000 / 60;
    const double ecc = orbit.GetKeplerData().GetEccentricity();
    return get_step(period, ecc) / 3;
}

XInterpolator::XInterpolator() = default;

XInterpolator::XInterpolator(const SInitOrbit& orbit, const TimeJD& beg, const TimeJD& end, const bool calc_rnb)
{
    create(orbit, beg, end, calc_rnb);
}

XInterpolator::XInterpolator(std::vector<IRes> xyz)
{
    std::sort(xyz.begin(), xyz.end(), [](const IRes& a, const IRes& b) {
        return a.get_date() < b.get_date();
    });

    create(xyz);
}

void XInterpolator::propagate_dir(
    const SInitOrbit& orbit,
    const TimeJD& beg,
    const TimeJD& end,
    const int dir,
    const bool calc_rnb,
    std::vector<IRes>& xyz
)
{
    const bool variates = orbit.GetCovMtx().HasMtx() && orbit.GetMotionModel().GetType() == proptypes::ST_PROP && calc_rnb;
    Propagator prop(orbit, variates);
    const double step = get_step(orbit);

    bool exit = false;
    TimeJD date = beg;
    while (!exit) {

        if (date >= end && dir == 1 || date <= end && dir == -1) {
            date = end;
            exit = true;
        }

        double age = DaysInterval(orbit.GetDate(), date);

        if (!variates) {
            PhasePoint6D pos;
            prop.Propagate(date, pos);
            xyz.emplace_back(pos.CoordsVel, date, age, orbit.GetId());
        }
        else {
            SInitOrbit norb;
            prop.Propagate(date, norb);
            std::map<std::string, double> covrnb = norb.GetCovMtx().GetCovRnbMap(norb.GetPhasePointJ2000().CoordsVel);
            xyz.emplace_back(
                norb.GetPhasePointJ2000().CoordsVel,
                Vect3(sqrt(covrnb["RR"]), sqrt(covrnb["NN"]), sqrt(covrnb["BB"])),
                date,
                age,
                orbit.GetId());
        }

        date = ShiftDate(date, dir * step);
    }
}

void XInterpolator::propagate(
    const SInitOrbit& orbit,
    const TimeJD& beg,
    const TimeJD& end,
    const int dir,
    const bool calc_rnb,
    std::vector<IRes>& xyz
)
{
    if (dir != 0) {
        propagate_dir(orbit, beg, end, dir, calc_rnb, xyz);
    }
    else {
        propagate_dir(orbit, ShiftDate(orbit.GetDate(), +get_step(orbit) / 2.0), end, +1, calc_rnb, xyz);
        propagate_dir(orbit, ShiftDate(orbit.GetDate(), -get_step(orbit) / 2.0), beg, -1, calc_rnb, xyz);
    }

    std::sort(xyz.begin(), xyz.end(), [](const IRes& a, const IRes& b) {
        return a.get_date() < b.get_date();
    });
}

void XInterpolator::calc_positions(
    const SInitOrbit& orbit,
    const TimeJD& beg,
    const TimeJD& end,
    const bool calc_rnb,
    std::vector<IRes>& xyz
)
{
    const int n = 100;
    const int step = get_step(orbit);

    const TimeJD beg_ext = ShiftDate(beg, -n * step);
    const TimeJD end_ext = ShiftDate(end, +n * step);

    if (orbit.GetDate() <= beg_ext) {
        propagate(orbit, beg_ext, end_ext, 1, calc_rnb, xyz);
    }
    else if (orbit.GetDate() >= end_ext) {
        propagate(orbit, end_ext, beg_ext, -1, calc_rnb, xyz);
    }
    else {
        propagate(orbit, beg_ext, end_ext, 0, calc_rnb, xyz);
    }
}

static void init(ei& ephi, const int size)
{
    ephi.jdi = new double[size]();
    memset(ephi.jdi, 0, size * sizeof(double));

    ephi.jdf = new double[size]();
    memset(ephi.jdf, 0, size * sizeof(double));

    ephi.jds = new double[size]();
    memset(ephi.jds, 0, size * sizeof(double));

    for (auto& j : ephi.outv) {
        j = new double[size]();
        memset(j, 0, size * sizeof(double));
    }

    ephi.nv = -1;
}

void fill(const Vect6& vec, const double jdi, const double jdf, ei& ephi)
{
    ephi.nv++;

    ephi.outv[0][ephi.nv] = vec.x * 1E6;
    ephi.outv[1][ephi.nv] = vec.y * 1E6;
    ephi.outv[2][ephi.nv] = vec.z * 1E6;
    //	ephi.outv[3][ephi.nv] = vec.vx*1E3; 
    //	ephi.outv[4][ephi.nv] = vec.vy*1E3;
    //	ephi.outv[5][ephi.nv] = vec.vz*1E3;

    ephi.jdi[ephi.nv] = jdi;
    ephi.jdf[ephi.nv] = jdf;

    if (ephi.jdf[ephi.nv] > 1.0e0) {
        ephi.jdi[ephi.nv] += 1.e0;
        ephi.jdf[ephi.nv] -= 1.e0;
    }
}

void fill_rnb(const Vect3& vec, const double jdi, const double jdf, ei& ephi)
{
    ephi.nv++;

    ephi.outv[0][ephi.nv] = vec.x * 1E6;
    ephi.outv[1][ephi.nv] = vec.y * 1E6;
    ephi.outv[2][ephi.nv] = vec.z * 1E6;

    ephi.jdi[ephi.nv] = jdi;
    ephi.jdf[ephi.nv] = jdf;

    if (ephi.jdf[ephi.nv] > 1.0e0) {
        ephi.jdi[ephi.nv] += 1.e0;
        ephi.jdf[ephi.nv] -= 1.e0;
    }
}

void finalize(ei& ephi)
{
    (ephi.nv)++;

    for (int i = 0; i < ephi.nv; i++) {
        ephi.jds[i] = ((ephi.jdf[i] - ephi.jdf[0]) + (ephi.jdi[i] - ephi.jdi[0])) * 86400.e0;
    }
}

void XInterpolator::create(const SInitOrbit& orbit, const TimeJD& beg, const TimeJD& end, const bool calc_rnb)
{
    std::vector<IRes> xyz;
    calc_positions(orbit, beg, end, calc_rnb, xyz);
    create(xyz);
}

void XInterpolator::create(const std::vector<IRes>& xyz)
{
    m_beg = xyz.front().get_date();
    m_end = xyz.back().get_date();

    init(m_ephi, xyz.size());

    for (const auto& p : xyz) {
        fill(p.get_pos(), p.get_date().GetDays(), p.get_date().GetFraction(), m_ephi);
    }

    finalize(m_ephi);
    m_positions = xyz;

    if (!xyz.empty() && xyz.front().has_rnb()) {
        init(m_rnb_ephi, xyz.size());

        for (const auto& p : xyz) {
            fill_rnb(p.get_rnb(), p.get_date().GetDays(), p.get_date().GetFraction(), m_rnb_ephi);
        }

        finalize(m_rnb_ephi);
    }
}

Vect6 XInterpolator::get_pos(const TimeJD& date, const bool tks) const
{
    if (!m_ephi.jdi) {
        throw Exp() << "interpolation not initialised\n";
    }

    if (!(date >= m_beg && date <= m_end)) {
        throw Exp()
            << "interpolation date out of range\n"
            << "date is " << DateToStr(date) << "\n"
            << "inerval is " << DateToStr(m_beg) << " - " << DateToStr(m_end) << "\n";
    }

    double outvec[6];
    const int fvel = 1;
    int ircode;
    const double jdi = date.GetDays();
    const double sec = date.GetFraction() * 86400;
    cinterp_pvac(m_ephi, jdi, sec, outvec, fvel, &ircode);

    if (ircode != 0) {
        throw Exp() << "interpolation fail\n";
    }

    if (!tks) {
        return {
            outvec[0],
            outvec[1],
            outvec[2],
            outvec[3],
            outvec[4],
            outvec[5]
        };
    }

    return {
        outvec[0] / 1E6,
        outvec[1] / 1E6,
        outvec[2] / 1E6,
        outvec[3] / 1E3,
        outvec[4] / 1E3,
        outvec[5] / 1E3
    };
}

Vect3 XInterpolator::get_rnb(const TimeJD& date, const bool tks) const
{
    if (!m_rnb_ephi.jdi) {
        throw Exp() << "interpolation not initialised\n";
    }

    if (!(date >= m_beg && date <= m_end)) {
        throw Exp()
            << "interpolation date out of range\n"
            << "date is " << DateToStr(date) << "\n"
            << "inerval is " << DateToStr(m_beg) << " - " << DateToStr(m_end) << "\n";
    }

    double outvec[6];
    const int fvel = 1;
    int ircode;
    const double jdi = date.GetDays();
    const double sec = date.GetFraction() * 86400;
    cinterp_pvac(m_rnb_ephi, jdi, sec, outvec, -1, &ircode);

    if (ircode != 0) {
        throw Exp() << "interpolation fail\n";
    }

    if (!tks) {
        return {
            outvec[0],
            outvec[1],
            outvec[2],
        };
    }

    return {
        outvec[0] / 1E6,
        outvec[1] / 1E6,
        outvec[2] / 1E6,
    };
}
