#include "orbBlocks.h"
#include "TimeFunctions.h"
#include "xalgorithm.h"
#include "orb_date.h"

OrbBlocksStore::OrbBlocksStore() = default;

OrbBlocksStore::OrbBlocksStore(const OrbBlocks& blocks)
    : m_blocks(blocks)
{
}

void OrbBlocksStore::init(
    const SInitOrbits& orbits,
    const TimeJD* beg,
    const TimeJD* end,
    const bool filter_orbits,
    const bool min_dist_forced,
    const bool interpolate,
    const bool calc_rnb
    )
{
    SInitOrbits sorted_orbits = orbits;

    std::sort(sorted_orbits.begin(), sorted_orbits.end(), [](const SInitOrbit& a, const SInitOrbit& b) {
        return a.GetDate() < b.GetDate();
    });

    m_beg_before_filter = beg ? *beg : sorted_orbits.front().GetDate();
    m_end_before_filter = end ? *end : sorted_orbits.back().GetDate();

    if (filter_orbits) {
        sorted_orbits = filter(sorted_orbits);
    }

    const TimeJD nbeg = beg ? *beg : sorted_orbits.front().GetDate();
    const TimeJD nend = end ? *end : sorted_orbits.back().GetDate();

    init_min_dist(sorted_orbits, nbeg, nend, min_dist_forced);

    if (interpolate) {
        for (auto& m_block : m_blocks) {
            m_block.init_by_orbit(calc_rnb);
        }
        m_initiated = true;
    }
}

void OrbBlocksStore::init(
    const SInitOrbits& orbits,
    const TimeJD& beg,
    const TimeJD& end,
    const bool filter_orbits,
    const bool min_dist_forced,
    const bool interpolate,
    const bool calc_rnb
    )
{
    init(orbits, &beg, &end, filter_orbits, min_dist_forced, interpolate, calc_rnb);
}

void OrbBlocksStore::init()
{
    for (auto& m_block : m_blocks) {
        m_block.init_by_data();
    }

    m_initiated = true;
}

SInitOrbits OrbBlocksStore::filter(const SInitOrbits& orbits) const
{
    SInitOrbits result;

    for (size_t i = 0; i < orbits.size(); i++) {
        const SInitOrbit& orb1 = orbits[i];

        const TimeJD orb1beg = ShiftDate(orb1.GetDate(), -orb1.GetInterval() * 86400);
        const TimeJD& orb1end = orb1.GetDate();

        bool found = false;
        for (size_t j = i + 1; j < orbits.size(); j++) {
            const SInitOrbit& orb2 = orbits[j];

            const TimeJD orb2beg = ShiftDate(orb2.GetDate(), -orb2.GetInterval() * 86400);
            const TimeJD& orb2end = orb2.GetDate();

            const bool eq1 = fabs(DaysInterval(orb1beg, orb2beg) * 86400 * 1000) < 1;
            const bool eq2 = fabs(DaysInterval(orb1end, orb2end) * 86400 * 1000) < 1;

            //fuzzy logic consumption analisys
            if (orb1beg < orb2beg && orb1end < orb2end && orb2beg < orb1end) {
                const double left = DaysInterval(orb1beg, orb2beg) * 86400;
                const double right = DaysInterval(orb1end, orb2end) * 86400;
                if (left > 0 && right > 0) {
                    const double ratio = left / right;
                    if (ratio < 0.1) {
                        //orbit2 consumes orbit1, so orbit1 no longer needed
                        found = true;
                        break;
                    }
                }
            }

            //strict logic consumption analisys
            if (((orb1beg > orb2beg) || eq1) && ((orb1end < orb2end) || eq2)) {
                //orbit2 consumes orbit1, so orbit1 no longer needed
                found = true;
                break;
            }
        }

        if (!found) {
            result.push_back(orb1);
        }
    }

    return result;
}

void loadjson(const jsonval& parent, OrbBlock& item)
{
    item.load(parent);
}

void OrbBlocksStore::load(const jsonval& parent)
{
    loadjson(parent, "records", m_blocks);
}

void OrbBlocksStore::init_min_dist(const SInitOrbits& orbits, const TimeJD& beg, const TimeJD& end, bool forced)
{
    TimeJD left = beg;
    //    std::string left_s = DateToStr(left);
    bool stop = false;

    for (size_t i = 0; i < orbits.size() && !stop; i++) {
        const SInitOrbit& orb1 = orbits[i];

        if (i == orbits.size() - 1) {
            m_blocks.emplace_back(orb1, left, end);
            break;
        }

        const TimeJD orb1beg = ShiftDate(orb1.GetDate(), -orb1.GetInterval() * 86400);
        const TimeJD& orb1end = orb1.GetDate();

        if (orb1end < left) {
            continue;
        }

        const SInitOrbit& orb2 = orbits[i + 1];

        TimeJD orb2beg = ShiftDate(orb2.GetDate(), -orb2.GetInterval() * 86400);

        if (orb1end < orb2beg) {
            //no intersection
            if (orb2beg < left) {
                //orbit1 if out of left and orbit2 has left border inside, so orbit1 no longer needed
                throw Exp() << "assert point 0";
            }

            TimeJD right;
            double minDist;
            calc_min_dist(orb1, orb2, orb1end, orb2beg, right, minDist);

            if (right > end) {
                right = end;
                stop = true;
            }

            m_blocks.push_back(OrbBlock(orb1, left, right, minDist, orb2.GetId()));
            left = right;
        }
        else {
            //intersection
            //bool eq = fabs(DaysInterval(orb1beg, orb2beg) * 86400 * 1000) < 1;
            if (orb1beg > orb2beg) {
                throw (Exp() << "assert point 1");
            }

            const TimeJD& right = orb2beg;
            //std::string right_s = DateToStr(right);
            if (left < right) {
                if (forced) {
                    TimeJD dummy;
                    double minDist;
                    calc_min_dist(orb1, orb2, orb1end, orb2beg, dummy, minDist);
                    m_blocks.push_back(OrbBlock(orb1, left, right, minDist, orb2.GetId()));
                }
                else {
                    m_blocks.push_back(OrbBlock(orb1, left, right));
                }
                left = right;
            }
        }
    }
}

class MinDistEvaluator
{
public:
    MinDistEvaluator(const XInterpolator& ip1, const XInterpolator& ip2, const TimeJD& base) :
        m_base(base),
        m_ip1(ip1),
        m_ip2(ip2)
    {
    }

    double calc(double time) const
    {
        const TimeJD date = ShiftDate(m_base, time * 86400);

        const Vect6 p1 = m_ip1.get_pos(date);
        const Vect6 p2 = m_ip2.get_pos(date);

        return sqrt(
            sqr(p1.x - p2.x) +
            sqr(p1.y - p2.y) +
            sqr(p1.z - p2.z)) / 1000.0;
    }

private:
    TimeJD m_base;
    const XInterpolator& m_ip1;
    const XInterpolator& m_ip2;
};

void OrbBlocksStore::calc_min_dist(const SInitOrbit& orb1, const SInitOrbit& orb2, const TimeJD& beg, const TimeJD& end, TimeJD& result_jd, double& result_dist) const
{
    const XInterpolator ip1(orb1, beg, end, false);
    const XInterpolator ip2(orb2, beg, end, false);
    MinDistEvaluator eval(ip1, ip2, beg);

    const double step = (int)(std::min(orb1.GetKeplerData().GetPeriod(), orb2.GetKeplerData().GetPeriod()) * 1000 / 20);
    double time = 0;
    const double time_end = DaysInterval(beg, end);
    double min = -1;
    double time_min = 0;

    bool exit = false;
    while (!exit) {
        if (time >= time_end) {
            time = time_end;
            exit = true;
        }

        const double dist = eval.calc(time);

        if (min == -1 || min > dist) {
            min = dist;
            time_min = time;
        }

        time += step / 86400.0;
    }

    const double tol = 60 / 86400.0;
    const double left = std::max(0.0, time_min - step / 86400.0 / 2.0);
    const double right = std::min(time_end, time_min + step / 86400.0 / 2.0);
    const double shift = golden_section_search(eval, left, right, tol);

    result_jd = ShiftDate(beg, shift * 86400);
    result_dist = eval.calc(shift);
}

Vect6 OrbBlocksStore::get_pos(const TimeJD& date, const bool tks) const
{
    for (const auto& block : m_blocks) {
        if (date >= block.get_beg() && date <= block.get_end()) {
            return block.get_interpolator()->get_pos(date, tks);
        }
    }

    throw Exp() << "no orbit found";
}

std::pair<Vect6, Vect3> OrbBlocksStore::get_pos_rnb(const TimeJD& date, const bool tks) const
{
    for (const auto& block : m_blocks) {
        if (date >= block.get_beg() && date <= block.get_end()) {
            return {
                block.get_interpolator()->get_pos(date, tks),
                block.get_interpolator()->get_rnb(date, tks)
            };
        }
    }

    throw Exp() << "no orbit found";
}

const SInitOrbit& OrbBlocksStore::get_orbit(const TimeJD& date) const
{
    for (const auto& block : m_blocks) {
        if (date >= block.get_beg() && date <= block.get_end()) {
            return block.get_orbit();
        }
    }

    throw Exp() << "no orbit found";
}

std::vector<orb_block_data> OrbBlocksStore::get_orb_data() const
{
    std::vector<orb_block_data> result;
    for (const auto& el : m_blocks) {
        result.emplace_back(el.get_data());
    }
    return result;
}

SInitOrbits OrbBlocksStore::calc_orbits(std::vector<TimeJD> dates, bool variates) const
{
    std::sort(dates.begin(), dates.end(), [](const TimeJD& a, const TimeJD& b) {
        return a < b;
    });

    SInitOrbits result;
    result.reserve(dates.size());

    size_t j = 0;
    for (auto const& block : get_blocks()) {
        std::vector<TimeJD> b_dates;

        for (size_t i = j; i < dates.size(); i++) {
            const TimeJD& date = dates[i];
            if (date >= block.get_beg() && date <= block.get_end()) {
                b_dates.push_back(date);
            }
            else if (date > block.get_end()) {
                j = i;
                break;
            }
        }

        if (b_dates.empty()) {
            continue;
        }

        std::vector<SInitOrbit> orbits = calc_orbits_on_dates(block.get_orbit(), b_dates, variates);
        result.insert(result.end(), orbits.begin(), orbits.end());
    }

    return result;
}
