#include "orb_date.h"
#include "Propagator.h"
#include <algorithm>

std::vector<TimeJD> generate_dates(const TimeJD& sdate, const TimeJD& edate, double step_sec)
{
    std::vector<TimeJD> result;

    result.push_back(sdate);

    if (sdate >= edate) {
        return result;
    }

    // old style check, need future analysis to choose right comparison
    //if (fabs(DaysInterval(sdate, edate)) * 86400 < 0.001) {
    //    return result;
    //}

    const auto n = (int)(DaysInterval(sdate, edate) * 86400 / step_sec);
    for (int i = 0; i < n; ++i) {
        result.emplace_back(ShiftDate(sdate, step_sec * (i + 1)));
    }

    if (result.back() != edate) {
        result.push_back(edate);
    }

    // old style check, need future analysis to choose right comparison
    //if (fabs(DaysInterval(result.back(), edate)) * 86400 > 0.001) {
    //    result.push_back(edate);
    //}

    return result;
}

void propagate_date(Propagator& prop, const TimeJD& date, IRes& result)
{
    PhasePoint6D pos;
    prop.Propagate(date, pos);
    result = IRes(pos.CoordsVel, date, DaysInterval(prop.GetInitOrbit().GetDate(), date), prop.GetInitOrbit().GetId());
}

void propagate_date(Propagator& prop, const TimeJD& date, SInitOrbit& result)
{
    SInitOrbit orbit;
    prop.Propagate(date, result);
    result.SetBaseOrbitId(prop.GetInitOrbit().GetId());
}

template <typename T>
void propagate_dir_prop(Propagator& prop, const std::vector<TimeJD>& dates, int dir, std::vector<T>& result)
{
    if (dir > 0) {
        for (const auto& date : dates) {
            T item;
            propagate_date(prop, date, item);
            result.push_back(item);
        }
    }
    else if (dir < 0) {
        for (auto it = dates.rbegin(); it != dates.rend(); ++it) {
            T item;
            propagate_date(prop, *it, item);
            result.push_back(item);
        }
    }
}

void propagate_dir(const SInitOrbit& orbit, const std::vector<TimeJD>& dates, int dir, bool variates, std::vector<IRes>& result)
{
    Propagator prop(orbit, variates);
    propagate_dir_prop(prop, dates, dir, result);
}

void propagate_dir(const SInitOrbit& orbit, const std::vector<TimeJD>& dates, int dir, bool variates, std::vector<SInitOrbit>& result)
{
    Propagator prop(orbit, variates);
    propagate_dir_prop(prop, dates, dir, result);
}

void split_dates(const std::vector<TimeJD>& dates, const TimeJD& date, std::vector<TimeJD>& dates_left, std::vector<TimeJD>& dates_right)
{
    for (const auto& it : dates) {
        if (it < date) {
            dates_left.push_back(it);
        }
        else {
            dates_right.push_back(it);
        }
    }
}

template <typename T>
void calc_orbit_on_dates(const SInitOrbit& orbit, std::vector<TimeJD> dates, bool variates, std::vector<T>& result)
{
    if (dates.empty()) {
        return;
    }

    std::sort(dates.begin(), dates.end(), [](const TimeJD& a, const TimeJD& b) {
        return a < b;
    });

    const TimeJD& beg_ext = dates.front();
    const TimeJD& end_ext = dates.back();

    if (orbit.GetDate() <= beg_ext) {
        propagate_dir(orbit, dates, 1, variates, result);
    }
    else if (orbit.GetDate() >= end_ext) {
        propagate_dir(orbit, dates, -1, variates, result);
    }
    else {
        std::vector<TimeJD> dates_left, dates_right;
        split_dates(dates, orbit.GetDate(), dates_left, dates_right);
        propagate_dir(orbit, dates_left, -1, variates, result);
        propagate_dir(orbit, dates_right, +1, variates, result);
    }
}

std::vector<IRes> calc_positions_on_dates(const SInitOrbit& orbit, const TimeJD& beg, const TimeJD& end, double step_sec)
{
    return calc_positions_on_dates(orbit, generate_dates(beg, end, step_sec));
}

std::vector<SInitOrbit> calc_orbits_on_dates(const SInitOrbit& orbit, const TimeJD& beg, const TimeJD& end, double step_sec, bool variates)
{
    return calc_orbits_on_dates(orbit, generate_dates(beg, end, step_sec), variates);
}

double calc_age(const SInitOrbit& orbit, const TimeJD& date)
{
    const TimeJD orbit_beg = ShiftDate(orbit.GetDate(), -orbit.GetInterval() * 86400);
    const TimeJD& orbit_end = orbit.GetDate();
        
    double age_days = 0;
    if (date < orbit_beg) {
        age_days = DaysInterval(orbit_beg, date);
    }
    else if (date > orbit_end) {
        age_days = DaysInterval(date, orbit_end);
    }
    return age_days;
}

std::vector<IRes> calc_positions_on_dates(const SInitOrbit& orbit, const std::vector<TimeJD>& dates)
{
    std::vector<IRes> result;
    calc_orbit_on_dates(orbit, dates, false, result);

    std::sort(result.begin(), result.end(), [](const IRes& a, const IRes& b) {
        return a.get_date() < b.get_date();
    });

    return result;
}

std::vector<SInitOrbit> calc_orbits_on_dates(const SInitOrbit& orbit, const std::vector<TimeJD>& dates, bool variates)
{
    std::vector<SInitOrbit> result;
    calc_orbit_on_dates(orbit, dates, variates, result);

    std::sort(result.begin(), result.end(), [](const SInitOrbit& a, const SInitOrbit& b) {
        return a.GetDate() < b.GetDate();
    });

    return result;
}
