#pragma once

#include <algorithm>
#include <cmath>
#include <functional>

template <typename T>
double golden_section_search(
    const T& f,
    double a,
    double b,
    const double time_accuracy)
{
    const double invphi = (std::sqrt(5.0) - 1) / 2.0;
    const double invphi2 = (3 - std::sqrt(5.0)) / 2.0;

    if (a > b) {
        std::swap(a, b);
    }

    double h = b - a;

    if (h <= time_accuracy) {
        return (a + b) / 2.0;
    }

    const int n = static_cast<int>(ceil(std::log(time_accuracy / h) / std::log(invphi)));

    double c = a + invphi2 * h;
    double d = a + invphi * h;
    double yc = f.calc(c);
    double yd = f.calc(d);

    for (int k = 0; k < n - 1; k++) {
        if (yc < yd) {
            b = d;
            d = c;
            yd = yc;
            h = invphi * h;
            c = a + invphi2 * h;
            yc = f.calc(c);
        }
        else {
            a = c;
            c = d;
            yc = yd;
            h = invphi * h;
            d = a + invphi * h;
            yd = f.calc(d);
        }
    }

    if (yc < yd) {
        return (a + d) / 2.0;
    }
    else {
        return (c + b) / 2.0;
    }
}

inline double golden_section_search_stdf(
    double a,
    double b,
    const double time_accuracy,
    const std::function<double(const double&)>& f
)
{
    const double invphi = (std::sqrt(5.0) - 1) / 2.0;
    const double invphi2 = (3 - std::sqrt(5.0)) / 2.0;

    if(a > b) {
        std::swap(a, b);
    }
    
    double h = b - a;

    if (h <= time_accuracy) {
        return (a + b) / 2.0;
    }

    const int n = static_cast<int>(ceil(std::log(time_accuracy / h) / std::log(invphi)));

    double c = a + invphi2 * h;
    double d = a + invphi * h;
    double yc = f(c);
    double yd = f(d);

    for (int k = 0; k < n - 1; k++) {
        if (yc < yd) {
            b = d;
            d = c;
            yd = yc;
            h = invphi * h;
            c = a + invphi2 * h;
            yc = f(c);
        }
        else {
            a = c;
            c = d;
            yc = yd;
            h = invphi * h;
            d = a + invphi * h;
            yd = f(d);
        }
    }

    if (yc < yd) {
        return (a + d) / 2.0;
    }
    else {
        return (c + b) / 2.0;
    }
}

template <typename T>
inline double bisection_search_stdf(
    double a,
    double b,
    const double accuracy,
    const std::function<T(const double&)>& f
)
{
    //f(a) should not be equal f(b) at the first step

    if (a > b) {
        std::swap(a, b);
    }

    while (fabs(b - a) > accuracy) {

        const double c = (a + b) / 2.0;
        if (f(a) != f(c)) {
            b = c;
        }
        else {
            a = c;
        }
    }

    return (a + b) / 2.0;
}
