/*
 * Copyright (c) 2010, Boris Popov <popov.b@gmail.com>
 *
 * Permission to use, copy, modify, and/or distribute
 * this software for any purpose with or without fee
 * is hereby granted, provided that the above copyright
 * notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "cbuffer.h"

/*---------------------------------------------------*/
static Result _GetCharFromStringBuffer(CBuffer);
static Result _GetCharFromFDBuffer(CBuffer);
static Result _GetCharFromSTDIN(CBuffer);
static CBuffer _InitCBuffer(const Buffer, const BufferType);
static void _DestroyCBuffer(CBuffer);
static CBufferFromString _InitCBufferFromString(const CString string);
static CBufferFromFD _InitCBufferFromFD(const FD);
static void _DestroyBufferFromString(CBufferFromString);
static void _DestroyBufferFromFD(CBufferFromFD);
static void _LoadBlock(CBufferFromFD);
static Result _GetResultFalse(void);
static void _DestroyBufferFromSTDIN(CBufferFromSTDIN);
static void _LoadBlockSTDIN(CBufferFromSTDIN);
static Result _GetResultNormalFrom(/*Result, */CBufferFrom);
static Size _ReadStdin(char*, Size);
/*---------------------------------------------------*/

/*---------------------------------------------------*/
CBuffer InitCBufferFromSTDIN()
{
    CBufferFromSTDIN buffer;
    buffer = (CBufferFromSTDIN)malloc(sizeof(struct _CBufferFromSTDIN));
    buffer->pos = 0;
    buffer->size = 8192;
    buffer->current_size = 0;
    buffer->buffer = (char*)calloc((buffer->size), sizeof(char));
    _LoadBlockSTDIN(buffer);
    return _InitCBuffer(buffer, BufferTypeSTDIN);
}

static void _LoadBlockSTDIN(CBufferFromSTDIN buffer)
{
    buffer->current_size = _ReadStdin(buffer->buffer, buffer->size);
    return;
}

static Size _ReadStdin(char* buffer, Size sizebuffer)
{
    Size size = 0;
    char* pbuffer = buffer;
    Size i;
    for (i = 0; i < sizebuffer; i++) {

        int i = fgetc(stdin);
        if (i == EOF) break;
        (*pbuffer) = (char)i;
        size++;
        pbuffer++;
    }
    return size;
}

CBuffer InitCBufferFromFD(const FD fd)
{
    CBufferFromFD cbuffer;
    if (fd == NULL) {

        return NULL;
    }

    cbuffer = _InitCBufferFromFD(fd);
    return _InitCBuffer(cbuffer, BufferTypeFD);
}

static Result _GetCharFromStringBuffer(CBuffer cbuffer)
{
    char* s;
    Result result;
    CBufferFromString buffer = (CBufferFromString)(cbuffer->buffer);

    if ((buffer->pos) < (buffer->size)) {

        result.type = TypeResultNORMAL;
        s = (char*)(buffer->string);
        s = s + (buffer->pos);
        result.ch = (*s);
        (buffer->pos)++;
    }

    else {

        result = _GetResultFalse();
    }

    return result;
}


static Result _GetResultFalse(void)
{
    Result result;
    result.type = TypeResultEOF;
    result.ch = '\0';
    return result;
}

static Result _GetResultNormalFrom(/*Result result, */CBufferFrom cbuffer)
{
    Result result;
    char* s;
    result.type = TypeResultNORMAL;
    s = (cbuffer->buffer);
    s = s + (cbuffer->pos);
    result.ch = (*s);
    (cbuffer->pos)++;
    return result;
}

static Result _GetCharFromFDBuffer(CBuffer cbuffer)
{
    Result result;
    CBufferFromFD buffer = (CBufferFromFD)(cbuffer->buffer);

    if ((buffer->pos) < (buffer->current_size)) {
        result = _GetResultNormalFrom(/*result, */(CBufferFrom)buffer);
    }

    else {

        buffer->pos = 0;
        _LoadBlock(buffer);

        if ((buffer->current_size) == 0) {

            result = _GetResultFalse();
        }

        else {

            result = _GetResultNormalFrom(/*result, */(CBufferFrom)buffer);
        }
    }

    return result;
}

Result GetChar(CBuffer buffer)
{
    if (buffer == NULL) {

        return _GetResultFalse();
    }

    if ((buffer->btype) == BufferTypeString) {

        return _GetCharFromStringBuffer(buffer);
    }

    else if ((buffer->btype) == BufferTypeSTDIN) {

        return _GetCharFromSTDIN(buffer);
    }

    else {

        return _GetCharFromFDBuffer(buffer);
    }
}

static Result _GetCharFromSTDIN(CBuffer cbuffer)
{
    Result result;
    CBufferFromSTDIN buffer = (CBufferFromSTDIN)(cbuffer->buffer);

    if ((buffer->pos) < (buffer->current_size)) {

        result = _GetResultNormalFrom(/*result, */(CBufferFrom)buffer);
    }

    else {

        buffer->pos = 0;
        _LoadBlockSTDIN(buffer);

        if ((buffer->current_size) == 0) {

            result = _GetResultFalse();
        }

        else {

            result = _GetResultNormalFrom(/*result, */(CBufferFrom)buffer);
        }
    }

    return result;
}

CBuffer InitCBufferFromString(const CString string)
{
    CBufferFromString buffer;
    if (string == NULL) {

        return NULL;
    }

    buffer = _InitCBufferFromString(string);
    return _InitCBuffer(buffer, BufferTypeString);
}

void DestroyCBuffer(CBuffer cbuffer)
{
    if (cbuffer == NULL) {

        return;
    }


    if ((cbuffer->btype) == BufferTypeString) {

        _DestroyBufferFromString((CBufferFromString)cbuffer->buffer);
    }

    else if ((cbuffer->btype) == BufferTypeFD) {

        _DestroyBufferFromFD((CBufferFromFD)cbuffer->buffer);
    }

    else {

        _DestroyBufferFromSTDIN((CBufferFromSTDIN)cbuffer->buffer);
    }
    _DestroyCBuffer(cbuffer);
    return;
}

static void _DestroyBufferFromSTDIN(CBufferFromSTDIN buffer)
{
    if (buffer == NULL) {

        return;
    }
    free(buffer->buffer);
    free(buffer);
    return;
}

static void _DestroyCBuffer(CBuffer cbuffer)
{
    free(cbuffer);
    return;
}

static CBuffer _InitCBuffer(const Buffer buffer, const BufferType bt)
{
    CBuffer temp;
    temp = (CBuffer)malloc(sizeof(struct _CBuffer));
    temp->btype = bt;
    temp->buffer = buffer;
    return temp;
}

static CBufferFromFD _InitCBufferFromFD(const FD fd)
{
    CBufferFromFD buffer;
    buffer = (CBufferFromFD)malloc(sizeof(struct _CBufferFromFD));
    buffer->fd = fd;
    buffer->size = 8192;
    buffer->buffer = (char*)calloc((buffer->size), sizeof(char));
    buffer->pos = 0;
    _LoadBlock(buffer);
    return buffer;
}

static void _LoadBlock(CBufferFromFD buffer)
{
    buffer->current_size = fread(buffer->buffer,
                                 sizeof(char),
                                 buffer->size,
                                 buffer->fd);
    return;
}

static void _DestroyBufferFromFD(CBufferFromFD buffer)
{
    if (buffer == NULL) {

        return;
    }
    free(buffer->buffer);
    free(buffer);
    return;
}

static CBufferFromString _InitCBufferFromString(const CString string)
{
    CBufferFromString buffer;
    buffer = (CBufferFromString)malloc(sizeof(struct _CBufferFromString));
    buffer->string = string;
    buffer->size = strlen(string);
    buffer->pos = 0;
    return buffer;
}

static void _DestroyBufferFromString(CBufferFromString buffer)
{
    if (buffer == NULL) {

        return;
    }
    free(buffer);
    return;
}

Bool HasItEof(const Result result)
{
    if (result.type == TypeResultNORMAL) {

        return Bool_FALSE;
    }
    return Bool_TRUE;
}

char GetCharFromResult(const Result result)
{
    return result.ch;
}

/*---------------------------------------------------*/
