/*
 * Copyright (c) 2010, Boris Popov <popov.b@gmail.com>
 *
 * Permission to use, copy, modify, and/or distribute
 * this software for any purpose with or without fee
 * is hereby granted, provided that the above copyright
 * notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "cbuffer.h"
#include "easysax_types.h"

#ifndef __easysax__h__
#define __easysax__h__

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

//     typedef void UserData;

struct _EasyParser
{
    CBuffer buffer;
    void (*FText)(Text, UserData*);
    void (*FOpenTag)(OpenTag, UserData*);
    void (*FCloseTag)(CloseTag, UserData*);
    void (*FAttribute)(AttValue, UserData*);
    void (*FError)(Error, UserData*);
    void (*FDeclar)(UserData*);
    Result result;
    Bool First;

    void* Demirror;
    void* RMDelimiter;

    UserData* userdata;
};

typedef struct _EasyParser* EasyParser;


/**
  * @addtogroup Parser
  * @{
  */

/** Make parser */
EasyParser InitEasyParser(CBuffer,
                          void (*FText)(Text, UserData*),
                          void (*FOpenTag)(OpenTag, UserData*),
                          void (*FCloseTag)(CloseTag, UserData*),
                          void (*FAttribute)(AttValue, UserData*),
                          void (*FDeclar)(UserData*),
                          void (*FError)(Error, UserData*));

/** Destroy Parser */
void DestroyEasyParser(EasyParser);

/** Parse */
Bool Parse(EasyParser, UserData* data);
/** @} */

#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif /* __easysax__h__ */


/* D O C U M E N T A T I O N */


/**
  \mainpage EasySAX


\section whatisit What is it?
Easy Simple Api for Xml.


\section whatkind How it works?
- make callback functions
- make buffer for parsing
- initialize parser
- parse
- callback functions are working


\section example For example:

  -# \ref string	"Parse from string"
  -# \ref file		"Parse from file"
  -# \ref stdin		"Parse from stdin"


\section homepage Home page:
http://launchpad.net/easysax


  \author Boris Popov <popov.b@gmail.com>, 2010 &copy;
 */


/**
  \page string	Parse from string

Include headers.
  \code
#include <easysax.h>
#include <stdio.h>
  \endcode

Make callback functions.
  \code
void PrintAtt (AttValue av) {

    printf ("att->%s<-", av.attribute);
    printf ("\n");
    return;
}

void PrintVal (AttValue av) {

    printf ("val->%s<-", av.value);
    printf ("\n");
    return;
}

void PrintA (AttValue av) {

    PrintAtt (av);
    PrintVal (av);
    return;
}

void PrintText (Text text) {

    printf ("text->%s<-", text);
    printf ("\n");
    return;
}

void PrintOpen (Text text) {

    printf ("open->%s<-", text);
    printf ("\n");
    return;
}

void PrintClose (Text text) {

    printf ("close->%s<-", text);
    printf ("\n");
    return;
}

void PrintD () {

    printf ("\n");
    printf ("*declaration*");
    printf ("\n");
    return;
}

void PrintE (Text text) {

    printf ("error->%s<-", text);
    printf ("\n");
    return;
}
  \endcode

  \code
int main () {

    CString string = "<root>text</root>";
  \endcode

  Make buffer.
  \code
  CBuffer buffer = InitCBufferFromString (string);
  if (buffer == NULL) {

      return 1;
  }
  \endcode

Make parser.
  \code
  EasyParser parser = \
    InitEasyParser (buffer, &PrintText, &PrintOpen, \
            &PrintClose, &PrintA, &PrintD, &PrintE);
  \endcode

  Parse.
  \code
    if (! Parse (parser)) {

        return 2;
    }
  \endcode

  Destroy objects.
  \code
    DestroyEasyParser (parser);
    DestroyCBuffer (buffer);
    return 0;
}
  \endcode
 */


/**
  \page file	Parse from file

Include headers.
  \code
#include <easysax.h>
#include <stdio.h>
  \endcode

Make callback functions.
  \code
void PrintAtt (AttValue av) {

    printf ("att->%s<-", av.attribute);
    printf ("\n");
    return;
}

void PrintVal (AttValue av) {

    printf ("val->%s<-", av.value);
    printf ("\n");
    return;
}

void PrintA (AttValue av) {

    PrintAtt (av);
    PrintVal (av);
    return;
}

void PrintText (Text text) {

    printf ("text->%s<-", text);
    printf ("\n");
    return;
}

void PrintOpen (Text text) {

    printf ("open->%s<-", text);
    printf ("\n");
    return;
}

void PrintClose (Text text) {

    printf ("close->%s<-", text);
    printf ("\n");
    return;
}

void PrintD () {

    printf ("\n");
    printf ("*declaration*");
    printf ("\n");
    return;
}

void PrintE (Text text) {

    printf ("error->%s<-", text);
    printf ("\n");
    return;
}
  \endcode

  \code
int main () {
  \endcode

  Make buffer.
  \code

  FILE* f = fopen ("in1.xml", "rb");

  if (f == NULL) {

      return 3;
  }

  CBuffer buffer = InitCBufferFromFD (f);

  if (buffer == NULL) {

      return 1;
  }
  \endcode

Make parser.
  \code
  EasyParser parser = \
    InitEasyParser (buffer, &PrintText, &PrintOpen, \
            &PrintClose, &PrintA, &PrintD, &PrintE);
  \endcode

  Parse.
  \code
    if (! Parse (parser)) {

        return 2;
    }
  \endcode

  Destroy objects.
  \code
    DestroyEasyParser (parser);
    DestroyCBuffer (buffer);
    fclose (f);
    return 0;
}
  \endcode
*/


/**
  \page stdin	Parse from stdin

Include headers.
  \code
#include <easysax.h>
#include <stdio.h>
  \endcode

Make callback functions.
  \code
void PrintAtt (AttValue av) {

    printf ("att->%s<-", av.attribute);
    printf ("\n");
    return;
}

void PrintVal (AttValue av) {

    printf ("val->%s<-", av.value);
    printf ("\n");
    return;
}

void PrintA (AttValue av) {

    PrintAtt (av);
    PrintVal (av);
    return;
}

void PrintText (Text text) {

    printf ("text->%s<-", text);
    printf ("\n");
    return;
}

void PrintOpen (Text text) {

    printf ("open->%s<-", text);
    printf ("\n");
    return;
}

void PrintClose (Text text) {

    printf ("close->%s<-", text);
    printf ("\n");
    return;
}

void PrintD () {

    printf ("\n");
    printf ("*declaration*");
    printf ("\n");
    return;
}

void PrintE (Text text) {

    printf ("error->%s<-", text);
    printf ("\n");
    return;
}
  \endcode

  \code
int main () {
  \endcode

  Make buffer.
  \code
  CBuffer buffer = InitCBufferFromSTDIN ();
  if (buffer == NULL) {

      return 1;
  }
  \endcode

Make parser.
  \code
  EasyParser parser = \
    InitEasyParser (buffer, &PrintText, &PrintOpen, \
            &PrintClose, &PrintA, &PrintD, &PrintE);
  \endcode

  Parse.
  \code
    if (! Parse (parser)) {

        return 2;
    }
  \endcode

  Destroy objects.
  \code
    DestroyEasyParser (parser);
    DestroyCBuffer (buffer);
    return 0;
}
  \endcode
 */