/*
 * Copyright (c) 2010, Boris Popov <popov.b@gmail.com>
 *
 * Permission to use, copy, modify, and/or distribute
 * this software for any purpose with or without fee
 * is hereby granted, provided that the above copyright
 * notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "easysaxatt.h"
#include "cstringakk.h"
#include "demirror.h"
#include "rmvdelimiter.h"

static void _PassWhiteSpace(EasyParserAtt);
static Bool _HasItEof(EasyParserAtt);
static Bool _Fetch(EasyParserAtt);
static Bool _IsItSpace(EasyParserAtt);
static char _GetChar(EasyParserAtt);
static Bool _IsItNormal(EasyParserAtt);
static Bool _IsItEq(EasyParserAtt);
static Bool _IsItAposOrQuot(EasyParserAtt);

static Bool _SaveAtt(EasyParserAtt, StringAkkN**);
static Bool _SaveVal(EasyParserAtt, StringAkkN**);

static void _InitAtt(EasyParserAtt);
static void _InitVal(EasyParserAtt);
static void _InitChar(char*);

static void _Out(EasyParserAtt);
static CString _Demirror(EasyParserAtt, CString);
static CString _RemoveDelimiter(EasyParserAtt, CString);

static CString _Demirror(EasyParserAtt parser, CString string)
{
    DeMirror demirror = (DeMirror)parser->Demirror;
    return (DecodeDeMirror(demirror, string));
}

static CString _RemoveDelimiter
(EasyParserAtt parser, CString string)
{
    RemoverDelimiter rd = (RemoverDelimiter)parser->RMDelimiter;
    return (RemoveDelimiter(rd, string));
}

EasyParserAtt InitEasyParserAtt
(CBuffer buffer, void (*Attribute)(AttValue, UserData*),
 void* demirror, void* rmdelimiter)
{
    EasyParserAtt parseratt = NULL;
    parseratt = malloc(sizeof(struct _EasyParserAtt));
    if (parseratt == NULL) return parseratt;
    parseratt->userdata = NULL;
    parseratt->buffer = buffer;
    parseratt->Attribute = Attribute;
    parseratt->att = NULL;
    parseratt->val = NULL;
    parseratt->Demirror = demirror;
    parseratt->RMDelimiter = rmdelimiter;
    return parseratt;
}

void DestroyEasyParserAtt(EasyParserAtt parseratt)
{
    if (parseratt != NULL) {

        _InitAtt(parseratt);
        _InitVal(parseratt);
        free(parseratt);
    }
    return;
}

Bool ParseAtt(EasyParserAtt parseratt, UserData* data)
{
    parseratt->userdata = data;


begin: {
        _PassWhiteSpace(parseratt);
        if (_HasItEof(parseratt)) return SUCCESS;
        if (_IsItNormal(parseratt)) goto att;
        return FAULT;
    }

att: {
        StringAkkN** sa = InitStringAkkN(4096);
        AddByteToStringAkkN(sa, _GetChar(parseratt));

        while (_Fetch(parseratt)) {

            /* space перед = */
            if (_IsItSpace(parseratt)) {

                _PassWhiteSpace(parseratt);
                if (!_IsItEq(parseratt)) {

                    DestoryStringAkkN(sa);
                    return FAULT;
                }
            }

            if (_IsItEq(parseratt)) {

                if (_SaveAtt(parseratt, sa)) {

                    DestoryStringAkkN(sa);
                    goto val;
                }
                break;
            }
            AddByteToStringAkkN(sa, _GetChar(parseratt));
        }
        DestoryStringAkkN(sa);
        return FAULT;
    }

val: {
        StringAkkN** sa;
        _PassWhiteSpace(parseratt);
        if (!_IsItAposOrQuot(parseratt)) {

            return FAULT;
        }

        sa = InitStringAkkN(4096);
        while (_Fetch(parseratt)) {

            if (_IsItAposOrQuot(parseratt)) {

                if (_SaveVal(parseratt, sa)) {

                    DestoryStringAkkN(sa);
                    _Out(parseratt);
                    goto begin;
                }
                break;
            }
            AddByteToStringAkkN(sa, _GetChar(parseratt));
        }
        DestoryStringAkkN(sa);
        return FAULT;
    }
}

static Bool _IsItNormal(EasyParserAtt parser)
{
    char ch = _GetChar(parser);
    if (!(IsItAposOrQuot(ch) || IsItEq(ch))) {

        return SUCCESS;
    }
    return FAULT;
}

static Bool _HasItEof(EasyParserAtt parser)
{
    return HasItEof(parser->result);
}

static void _PassWhiteSpace(EasyParserAtt parser)
{
    while (_Fetch(parser)) {

        if (!_IsItSpace(parser)) {

            break;
        }
    }
    return;
}

static Bool _Fetch(EasyParserAtt parser)
{
    parser->result = GetChar(parser->buffer);
    if (_HasItEof(parser)) {

        return FAULT;
    }
    return SUCCESS;
}

static Bool _IsItSpace(EasyParserAtt parser)
{
    char ch2 = _GetChar(parser);
    return IsItWhiteSpace(ch2);
}

static char _GetChar(EasyParserAtt parser)
{
    return GetCharFromResult(parser->result);
}

static Bool _IsItEq(EasyParserAtt parser)
{
    return IsItEq(_GetChar(parser));
}

static void _InitAtt(EasyParserAtt parser)
{
    _InitChar(parser->att);
    return;
}

static void _InitVal(EasyParserAtt parser)
{
    _InitChar(parser->val);
    return;
}

static void _InitChar(char* ch)
{
    if (ch != NULL) free(ch);
    ch = NULL;
    return;
}

static Bool _SaveAtt
(EasyParserAtt parser, StringAkkN** akk)
{
    _InitAtt(parser);
    parser->att =
        (char*)malloc
        ((strlen(GetStringFromStringAkkN(akk))) + 1);

    if ((parser->att) == NULL) {

        return FAULT;
    }

    *((parser->att) + (strlen(GetStringFromStringAkkN(akk)))) = '\0';
    strcpy(parser->att, (GetStringFromStringAkkN(akk)));
    return SUCCESS;
}

static Bool _SaveVal
(EasyParserAtt parser, StringAkkN** akk)
{
    _InitVal(parser);
    parser->val =
        (char*)malloc
        ((strlen(GetStringFromStringAkkN(akk))) + 1);

    if ((parser->val) == NULL) {

        return FAULT;
    }

    *((parser->val) + (strlen(GetStringFromStringAkkN(akk)))) = '\0';
    strcpy(parser->val, (GetStringFromStringAkkN(akk)));
    return SUCCESS;
}

static Bool _IsItAposOrQuot(EasyParserAtt parser)
{
    return IsItAposOrQuot(_GetChar(parser));
}

static void _Out(EasyParserAtt parser)
{
    if ((parser->Attribute) != NULL) {

        AttValue av;
        av.attribute = parser->att;
        av.value =
            _Demirror(parser,
                      _RemoveDelimiter(parser,
                                       parser->val));
        (parser->Attribute)(av, parser->userdata);
    }
    return;
}
