/*
 * Copyright 2008, 2009, 2010  Boris Popov <popov.b@gmail.com>
 */

#ifndef __xmldocument_hpp__
#define __xmldocument_hpp__

#include <string>
#include <list>
#include <queue>


//-----------------------------------------------
namespace myXML
{
    typedef std::string String;
    typedef const char* CString;
    typedef const String FileName;
    typedef String Attribute;
    typedef String Value;

    struct Attr
    {
        Attribute att;
        Value val;
    };

    typedef std::list<Attr> AttributesList;
    typedef std::list<Attr>::iterator AttributesListIterator;

    class AttrIter;

    class Attributes
    {
    public:
        Attributes();
        ~Attributes();

        AttrIter Begin();
        AttrIter End();

        void Clear();

        bool DelAtt(Attribute);
        void AddAtt(Attribute, Value);

        bool GetValueOfAttr(Attribute, Value&);

        bool HasItAttr() const;

        Attributes& operator=(const Attributes&);

    private:
        Attributes(const Attributes&);

        AttributesList m_ListOfAtt;
        AttributesListIterator m_GetAtt(Attribute);
    };

    //-----------------------------------------------


    //-----------------------------------------------
    class Declar
    {
    public:
        Declar();
        ~Declar();

        void UnSet();
        void Set();
        bool IsItEnabled() const;

        String Get();

        Attributes& GetAttributes();

    private:
        Attributes m_Attributes;

        static String part_one;
        static String part_three;

        enum SWITCH { ON, OFF };

        SWITCH SW;

        Declar(const Declar&);
        Declar& operator=(const Declar&);
    };

    //-----------------------------------------------


    //-----------------------------------------------
    class Declar;

    class Prolog
    {
    public:
        Prolog();
        ~Prolog();
        Declar& GetDeclar();

    private:
        Declar m_Declar;

        Prolog(const Prolog&);
        Prolog& operator=(const Prolog&);
    };

    //-----------------------------------------------


    //-----------------------------------------------
    class Builder;
    class Text;
    class Node;
    class BodyIter;
    class Contain;

    class Body
    {
        friend class Document;
        friend class Builder;

    public:
        Body();
        Body(Contain& c);
        ~Body();

        void DelNode(BodyIter&);
        void DelNode(Contain&);
        void DelNode(Text&);

        BodyIter BeginBodyIter();
        BodyIter EndBodyIter();

        Contain* GetRoot() { return Root; }

    private:
        Contain* Root;
        void delnode(Node*);

        Body(const Body&);
        Body& operator=(const Body&);
    };

    //-----------------------------------------------


    //-----------------------------------------------
    struct ParserResult;
    class Contain;
    class Body;
    class Prolog;

    class Document
    {
    public:
        Document(Contain&);
        Document();
        ~Document();

        Prolog& GetProlog();
        Body& GetBody();

        bool LoadFile(FileName);
        bool LoadStringX(CString);
        bool LoadStdin();
        String GetMessage() const;

        String Get();

        enum TypeOfNode { Container, Text };

    private:
        String message;
        void correct(ParserResult);
        Prolog m_Prolog;
        Body m_Body;

        Document& operator =(Document&);
        Document(const Document&);

        const char* m_InputString;
        void m_ParseString();
    };

    //-----------------------------------------------


    //-----------------------------------------------
    bool operator !=(myXML::BodyIter, myXML::BodyIter);
    bool operator ==(myXML::BodyIter, myXML::BodyIter);

    class BodyIter
    {
        friend class Body;
        friend class Contain;
        friend bool operator !=(myXML::BodyIter, myXML::BodyIter);
        friend bool operator ==(myXML::BodyIter, myXML::BodyIter);

    public:
        Node* operator *();

        BodyIter& operator ++();

        BodyIter(const BodyIter&);
        BodyIter& operator=(const BodyIter&);

    private:
        Node* pNode;

        std::queue<Node*> Queue;
        Node* plus_normal_iter();

        BodyIter();
        BodyIter(Node*);
    };

    //-----------------------------------------------


    //-----------------------------------------------
    bool operator !=(myXML::AttrIter, myXML::AttrIter);
    bool operator ==(myXML::AttrIter, myXML::AttrIter);

    class AttrIter
    {
        friend class Attributes;
        friend bool operator !=(myXML::AttrIter, myXML::AttrIter);
        friend bool operator ==(myXML::AttrIter, myXML::AttrIter);

    public:
        Attr& Get();
        AttrIter& operator --();
        AttrIter& operator ++();

        AttrIter();
        AttrIter(const AttrIter&);
        AttrIter& operator=(const AttrIter&);

    private:
        std::list<Attr>::iterator IPA;
    };

    //-----------------------------------------------


    //-----------------------------------------------
    class Node
    {
        friend class Body;
        friend class Contain;

    public:
        virtual ~Node() = 0;
        Node(Document::TypeOfNode);

        void Set(String);
        String Get() const;
        Document::TypeOfNode GetType() const;

    protected:
        Contain* pParent;
        String Content;
        Document::TypeOfNode TypeNode;

        enum Place { in, out, undef };

        Place place;

        virtual String build() = 0;

        size_t GetLevel() const;
    };

    //-----------------------------------------------


    //-----------------------------------------------
    class BuildOfTree;
    class ChildIter;

    class Contain : public Node
    {
        friend class Body;
        friend class BodyIter;
        friend class BuildOfTree;
        friend class Builder;

    public:
        Contain(String);

        Attributes& GetAttributes();

        ChildIter BeginChildIter();
        ChildIter EndChildIter();

        BodyIter BeginBodyIter();
        BodyIter EndBodyIter();

        Contain& operator <<(Contain&);
        Contain& operator <<(Text&);

        ~Contain();

    private:
        std::list<Node*> ListOfChildren;

        Attributes m_Attributes;

        Contain();
        Contain(const Contain&);
        Contain& operator =(const Contain&);

        String build();
        String get_string_attr();
    };

    //-----------------------------------------------


    //-----------------------------------------------
    class Text : public Node
    {
        friend class BuildOfTree;
        friend class Contain;

    public:
        Text(String);
        ~Text();

    private:
        Text();

        Text& operator =(Text&);
        Text(const Text&);

        String build();
    };

    //-----------------------------------------------


    //-----------------------------------------------
    typedef std::list<Node*>::iterator IteratorOnChild;

    bool operator !=(myXML::ChildIter, myXML::ChildIter);
    bool operator ==(myXML::ChildIter, myXML::ChildIter);

    class ChildIter
    {
        friend class Contain;
        friend bool operator !=(myXML::ChildIter, myXML::ChildIter);
        friend bool operator ==(myXML::ChildIter, myXML::ChildIter);

    public:
        ChildIter();
        ~ChildIter();

        Node* operator *();

        ChildIter& operator ++();
        ChildIter& operator --();

        ChildIter(const ChildIter&);
        ChildIter& operator=(const ChildIter&);

    private:
        IteratorOnChild m_IOC;
    };
}

//-----------------------------------------------
#endif
